/*
 * SMPFunction.cpp - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

//
// SMPFunction.cpp
//
// This module performs the fundamental data flow analyses needed for the
//   SMP project (Software Memory Protection) at the function level.
//


#include "SMPFunction_support.h"

// Comparison function for sorting.
bool LocalVarCompare(const LocalVar &LV1, const LocalVar &LV2) {
	return (LV1.offset < LV2.offset);
}

// Are the operands, SSA numbers, and SubtractAddend fields identical?
bool EqualInductionVars(const InductionVarTriple &IV1, const InductionVarTriple &IV2) {
	bool success = (IV1.InductionVar.GetSSANum() == IV2.InductionVar.GetSSANum())
		&& (IV1.SubtractAddend == IV2.SubtractAddend);
	if (success) {
		success = IsEqOp(IV1.InductionVar.GetOp(), IV2.InductionVar.GetOp());
		if (success) {
			success = (IsEqOp(IV1.Addend.GetOp(), IV2.Addend.GetOp())
				&& (IsEqOp(IV1.Multiplier.GetOp(), IV2.Multiplier.GetOp())));
		}
	}
	return success;
} // end of EqualInductionVars()

// Debug dump of induction variable.
void DumpInductionVar(const struct InductionVarTriple IndVar) {
	SMP_msg("InductionVar: Multiplier: ");
	IndVar.Multiplier.Dump();
	SMP_msg(" Induction Ref: ");
	IndVar.InductionVar.Dump();
	if (IndVar.SubtractAddend) {
		SMP_msg(" AddSubOperator: - ");
	}
	else {
		SMP_msg(" AddSubOperator: + ");
	}
	SMP_msg(" Induction Addend: ");
	IndVar.Addend.Dump();
	SMP_msg("\n");
	return;
} // end of DumpInductionVar()

// Debug dump of InductionVarFamily.
void DumpInductionVarFamily(const struct InductionVarFamily IVFamily) {
	SMP_msg(" BIVIncomingSSA: %d BIVIncomingDefAddr: %llx BIVInsideLoopDefAddrs:", 
		IVFamily.BIVIncomingSSANum, (uint64_t) IVFamily.BIVIncomingDefAddr);
	for (vector<STARS_ea_t>::const_iterator AddrIter = IVFamily.BIVInsideLoopDefAddrs.cbegin();
		AddrIter != IVFamily.BIVInsideLoopDefAddrs.cend();
		++AddrIter) {
		SMP_msg(" %llx", (uint64_t) (*AddrIter));
	}
	SMP_msg(" BIV: ");
	DumpInductionVar(IVFamily.BasicInductionVar);
	for (size_t index = 0; index < IVFamily.DependentInductionVars.size(); ++index) {
		struct DependentInductionVar DIV = IVFamily.DependentInductionVars[index];
		SMP_msg("DIVDefAddr: %llx DIVRef: ", (uint64_t) DIV.DIVDefAddr);
		DIV.DIV.Dump();
		SMP_msg(" DIV equation: ");
		DumpInductionVar(DIV.IVExpr);
	}
} // end of DumpInductionVarFamily()

// Debug dump of LoopComparison struct.
void DumpLoopComparison(const struct LoopComparison Expr) {
	Expr.Operand1.Dump();
	SMP_msg(" %s ", OperatorText[Expr.CompareOperator]);
	Expr.Operand2.Dump();
	SMP_msg(" CompareAddr: %llx ExitsLoop: %d\n", (uint64_t) Expr.CompareAddr, (int) Expr.ExitsLoop);
} // end of DumpLoopComparison()


// Emit annotation for induction variable.
void EmitInductionVarAnnotation(const struct InductionVarTriple IndVar) {
	return;
}

// Emit annotations for InductionVarFamily.
void EmitInductionVarFamilyAnnotations(const struct InductionVarFamily IVFamily) {
	return;
}


// Determine whether we have an incrementing or decrementing loop based on
//  the BasicInductionVar.
bool IsPositiveIncrementBIV(const struct InductionVarTriple BIV) {
	bool PositiveIncrement = true; // default
	STARSOpndTypePtr AddendOp = BIV.Addend.GetOp();
	if (AddendOp->IsImmedOp()) {
		STARS_uval_t ImmedValue = AddendOp->GetImmedValue();
		int SignedValue = (int) ImmedValue;
		if (SignedValue < 0) {
			PositiveIncrement = BIV.SubtractAddend;
		}
		else {
			PositiveIncrement = (!BIV.SubtractAddend);
		}
	}

	return PositiveIncrement;
} // end of IsPositiveIncrementBIV()


// *****************************************************************
// Class SMPFunction
// *****************************************************************

// Constructor
SMPFunction::SMPFunction(STARS_Function_t *Info, SMPProgram* pgm) {
	this->Program = pgm;
	this->FuncInfo = Info;
	this->FirstEA = this->FuncInfo->get_startEA();
	this->BlockCount = 0;
	this->LoopCount = 0;
	this->FuncProcessed = false;
	this->UseFP = false;
	this->StaticFunc = false;
	this->LibFunc = false;
	this->HasReturnInst = false;
	this->IndirectCalls = false;
	this->UnresolvedIndirectCalls = false;
	this->IndirectJumps = false;
	this->UnresolvedIndirectJumps = false;
	this->SystemCalls = false;
	this->DirectlyRecursive = false;
	this->MutuallyRecursive = false;
//	this->SetSharedChunks(false);
	this->UnsharedChunks = false;
	this->MultipleEntryPoints = false;
	this->CalledFromOrphanCode = false;
	this->TailCallChainFromOrphanCode = false;
	this->CallsNonReturningFunc = false;
	this->CalleeChainCallsNonReturningFunc = false;
	this->CallsAlloca = false;
	this->PushAfterLocalVarAlloc = false;
	this->LinkerStub = false;
	this->ThunkFunc = false;
	this->AnalyzedSP = false;
	this->STARSStackPtrAnalysisPerformed = false;
	this->StackAdjustmentComputed = false;
	this->BuiltRTLs = false;
	this->HasReducibleCFG = false;
	this->HasStructuredCFG = false;
	this->HasGoodLoopFollowBlocks = true;
	this->HasGoodSSA = false;
#if 1  // default to unsafe
	this->SetFuncSafe(false);
	this->SetSpecFuncSafe(false);
	this->SafeCallee = false;
	this->SpecSafeCallee = false;
	this->UnsafeForFastReturns = true;
#else // default to safe
	this->SetFuncSafe(true);
	this->SpecSafeFunc = true;
	this->SafeCallee = true;
	this->SpecSafeCallee = true;
	this->UnsafeForFastReturns = false;
#endif
	this->WritesAboveRA = false;
	this->NeedsStackReferent = true;
	this->SpecNeedsStackReferent = true;
	this->HasIndirectWrites = false;
	this->HasUnsafeIndirectWrites = false;
	this->AltersMemory = false;
	this->HasLoopInArgMemWrites = false;
	this->HasMemExprsFromCalleeLoops = false;
	this->PossibleIndirectCallTarget = false;
	this->PossibleTailCallTarget = false;
	this->ReturnTargetsComputed = false;
	this->OutgoingArgsComputed = false;
	this->GoodLocalVarTable = false;
	this->StackFrameExtendsPastStackTop = false;
	this->SetIsSpeculative(false);
	this->HasHashingCode = false;
	this->HasShortCircuitExpr = false;
	this->HasInArgCodePointer = false;
	this->HasInArgDataPointer = false;
	this->HasMallocCall = false;
	this->TranslatingSPARKLoop = false;
	this->PrintedSPARKUnstructuredMsg = false;
	this->PrintedCFG = false;
	this->PrintedTCFG = false;
	this->TypedDefs = 0;
	this->UntypedDefs = 0;
	this->TypedPhiDefs = 0;
	this->UntypedPhiDefs = 0;
	this->SafeBlocks = 0;
	this->UnsafeBlocks = 0;
	this->Size = 0;

#if 0
	// The sizes of the three regions of the stack frame other than the
	//  return address are stored in the function structure.
	this->LocalVarsSize = this->FuncInfo->GetFrameSize();
	this->CalleeSavedRegsSize = this->FuncInfo->GetSavedRegSize();
	this->IncomingArgsSize = (STARS_asize_t) this->FuncInfo->GetIncomingArgumentSize();

	// The return address size can be obtained in a machine independent
	//  way by calling get_frame_retsize(). 
	this->RetAddrSize = /* get_frame_retsize(this->GetFuncInfo()); */
			this->GetFuncInfo()->GetFrameReturnAddressSize();
#else  // compute values in MDFixFrameInfo() before their first use; avoid IDA Pro calls
	this->LocalVarsSize = 0;
	this->CalleeSavedRegsSize = 0;
	this->IncomingArgsSize = 0; // unused
	this->RetAddrSize = global_STARS_program->GetSTARS_ISA_Bytewidth();
#endif
	this->AllocSizeAfterFrameAlloc = 0;
	this->OutgoingArgsSize = 0;
	this->LocalVarsAllocInstr = STARS_BADADDR;
	this->LocalVarsDeallocInstr = STARS_BADADDR;
	this->AllocPointDelta = 0;
	this->MinStackDelta = 0;
	this->MaxStackDelta = 0;
	this->MinStackAccessOffset = 0;
	this->MaxStackAccessLimit = 0;
	this->NetStackDelta = CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA;
	this->PreAllocStackDelta = CALLING_CONVENTION_DEFAULT_PREFRAMEALLOC_STACK_DELTA;
	this->FramePointerStackDelta = 0;
	this->GlobalStackAdjustment = 0;
	this->SetLocalVarOffsetLimit(0);
	this->MaxDirectStackAccessDelta = 0;
	this->IDAReturnAddressOffset = 0;

	this->ReturnAddrStatus = FUNC_UNKNOWN;
	this->FastReturnStatus = SAFE_FAST_RETURN;

	this->InArgCount = 0;
	this->MaxInArgIndex = 0;
	this->TaintInArgPosBits = 0;
	this->SPARKConditionalNestingDepth = 0;
	this->MaxDirectStackAccessSSANum = 0;
	this->MaxRegSSANum = 0;
	this->InArgTypes.clear();
	this->InArgTypes.resize(STARS_MAX_ARGS_PER_FUNC, (unsigned short)UNINIT);

	this->Instrs.clear();
	this->Blocks.clear();
	this->DirectCallTargets.clear();
	this->IndirectCallTargets.clear();
	this->AllCallTargets.clear();
	this->AllCallSources.clear();
	this->InstBlockMap.clear();
	this->RPOBlocks.clear();
	this->IDom.clear();
	this->DomTree.clear();
	this->GlobalNames.clear();
	this->BlocksDefinedIn.clear();
	this->SSACounter.clear();
	this->SSAStack.clear();
	this->LocalVarTable.clear();
	this->SavedRegLoc.clear();
	this->ReturnRegTypes.clear();
	this->LiveOutSet.clear();
	this->KillSet.clear();
	this->GlobalDefAddrBySSA.clear();
	this->EdgeExpressions.clear();

	struct FineGrainedInfo TempFG;
	TempFG.SignMiscInfo = 0;
	TempFG.SizeInfo = 0;
	for (int RegIndex = STARS_x86_R_ax; RegIndex <= global_STARS_program->GetSTARS_MD_LAST_SAVED_REG_NUM(); ++RegIndex) {
		this->SavedRegLoc.push_back(0); // zero offset means reg not saved
		this->ReturnRegTypes.push_back(UNINIT);
		this->ReturnRegFGInfo.push_back(TempFG);
		this->IncomingRegTypes.push_back(UNINIT);
	}

	return;
} // end of SMPFunction() constructor

// Destructor
SMPFunction::~SMPFunction() {
	list<SMPInstr *>::iterator InstIter;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if (nullptr != CurrInst) delete CurrInst;
	}

	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		if (nullptr != CurrBlock) delete CurrBlock;
	}

	if (global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
		// DestroyLoopExprs() was not called from AnalyzeLoopIterations()
		//  when we saved the exprs for SPARK Ada translation.
		this->DestroyLoopExprs();
	}

	for (size_t i = 0; i < this->ShadowCFGBlocks.size(); ++i) {
		if (nullptr != this->ShadowCFGBlocks[i])
			delete this->ShadowCFGBlocks[i];
	}

#if 0 // crashes in STARS/IRDB ???
	for (size_t i = 0; i < this->LoopInductionVars.size(); ++i) {
		for (struct InductionVarFamily CurrFam : this->LoopInductionVars[i]) {
			if (nullptr != CurrFam.BIVInitExpr)
				delete CurrFam.BIVInitExpr;
			if (nullptr != CurrFam.BIVLimitExpr)
				delete CurrFam.BIVLimitExpr;
		}
	}
#endif

	for (size_t i = 0; i < this->LoopMemAddrExprsFromCalleeLoops.size(); ++i) {
		for (STARSExpression *CurrExpr : this->LoopMemAddrExprsFromCalleeLoops[i]) {
			if (nullptr != CurrExpr)
				delete CurrExpr;
		}
	}

	for (size_t i = 0; i < this->LoopMemAddrExprsFromCallees.size(); ++i) {
		for (STARSExpression *CurrExpr : this->LoopMemAddrExprsFromCallees[i]) {
			if (nullptr != CurrExpr)
				delete CurrExpr;
		}
	}

	return;
} // end of SMPFunction::destructor

void SMPFunction::DestroyLoopExprs(void) {
	size_t i;
	for (i = 0; i < this->LoopIterationsCountExprs.size(); ++i)
		if (nullptr != this->LoopIterationsCountExprs[i]) {
			delete this->LoopIterationsCountExprs[i];
			this->LoopIterationsCountExprs[i] = nullptr;
		}

	for (i = 0; i < this->LoopIterationsInitExprs.size(); ++i)
		if (nullptr != this->LoopIterationsInitExprs[i]) {
			delete this->LoopIterationsInitExprs[i];
			this->LoopIterationsInitExprs[i] = nullptr;
		}

	for (i = 0; i < this->LoopIterationsLimitExprs.size(); ++i)
		if (nullptr != this->LoopIterationsLimitExprs[i]) {
			delete this->LoopIterationsLimitExprs[i];
			this->LoopIterationsLimitExprs[i] = nullptr;
		}

	for (i = 0; i < this->LoopMemWriteRangeExprs.size(); ++i)
		this->LoopMemWriteRangeExprs[i].clear();
		
	for (i = 0; i < this->LoopMemWriteBoundsExprs.size(); ++i) {
		for (STARSExprBoundsPair CurrPair : this->LoopMemWriteBoundsExprs[i]) {
			if (nullptr != CurrPair.first) {
				delete CurrPair.first;
				CurrPair.first = nullptr;
			}
			if (nullptr != CurrPair.second) {
				delete CurrPair.second;
				CurrPair.second = nullptr;
			}
		}
	}

	for (i = 0; i < this->LoopMemWriteBoundsExprsExpanded.size(); ++i) {
		for (STARSExprBoundsPair CurrPair : this->LoopMemWriteBoundsExprsExpanded[i]) {
			if (nullptr != CurrPair.first) {
				delete CurrPair.first;
				CurrPair.first = nullptr;
			}
			if (nullptr != CurrPair.second) {
				delete CurrPair.second;
				CurrPair.second = nullptr;
			}
		}
	}

	return;
}
// Get a non-stale pointer to the STARS_Function_t info for the current function.
STARS_Function_t *SMPFunction::GetFuncInfo(void)  const {
	STARS_Function_t *myPtr = SMP_get_func(this->GetFirstFuncAddr());
	assert(nullptr != myPtr);
	return myPtr;
}

uint16_t SMPFunction::GetJumpToFollowNodeCounter(STARS_ea_t InstAddr) const {
	uint16_t Counter = 0;
	map<STARS_ea_t, uint16_t>::const_iterator MapIter = this->JumpToFollowNodeCounterMap.find(InstAddr);
	if (MapIter != this->JumpToFollowNodeCounterMap.end()) {
		Counter = MapIter->second;
	}
	return Counter;
} // end of SMPFunction::GetJumpToFollowNodeCounter()

STARSExprSetIter SMPFunction::FindInArgExprBySerialNumber(const std::size_t LoopPlusOne, const uint32_t SerialNumber) const {
	this->GetProg()->SetSearchExprSerialNumber(SerialNumber);
	return InArgsUsedInMemWrites[LoopPlusOne].find(this->GetProg()->GetSearchExpr());
}

// Reset the Processed flags in all blocks to false.
void SMPFunction::ResetProcessedBlocks(void) {
	for (SMPBasicBlock *CurrBlock : this->Blocks) {
		CurrBlock->SetProcessed(false);
	}
	return;
} // end of SMPFunction::ResetProcessedBlocks()

// Set SCCPVisited flag to false in all blocks
void SMPFunction::ResetSCCPVisitedBlocks(void) {
	for (SMPBasicBlock *CurrBlock : this->Blocks) {
		CurrBlock->SetSCCPVisited(false);
	}
	return;
} // end of SMPFunction::ResetSCCPVisitedBlocks()

// Set Visited flag to false in all TCFG blocks
void SMPFunction::ResetTCFGVisitedBlocks(void) {
	for (STARSTransCFGBlock *TCFGBlock : this->TCFGBlocks) {
		TCFGBlock->SetVisited(false);
	}
	return;
} // end of SMPFunction::ResetTCFGVisitedBlocks()

// Get first TCFG block # with OrigBlockNum as its GetOriginalBlockNum() value.
int SMPFunction::FindAnyTCFGBlockNumMatchingOrigBlockNum(const int OrigBlockNum) const {
	int TCFGBlockNum = SMP_BLOCKNUM_UNINIT;
	for (STARSTransCFGBlock *TCFGBlock : this->TCFGBlocks) {
		if (OrigBlockNum == TCFGBlock->GetOriginalBlockNum()) {
			TCFGBlockNum = TCFGBlock->GetNumber();
			break;
		}
	}
	assert(0 <= TCFGBlockNum);
	return TCFGBlockNum;
} // end of SMPFunction::FindAnyTCFGBlockNumMatchingOrigBlockNum()

// Return an iterator for the beginning of the LiveInSet. 
STARSOpndSetIter SMPFunction::GetFirstLiveIn(void) {
	return this->Blocks.front()->GetFirstLiveIn();
} // end of SMPBasicBlock::GetFirstLiveIn()

// Get termination iterator marker for the LiveIn set.
STARSOpndSetIter SMPFunction::GetLastLiveIn(void) {
	return this->Blocks.front()->GetLastLiveIn();
}

// Get iterator for the start of the LiveOut set.
STARSOpndSetIter SMPFunction::GetFirstLiveOut(void) {
	return this->LiveOutSet.begin();
}

// Get termination iterator marker for the LiveOut set.
STARSOpndSetIter SMPFunction::GetLastLiveOut(void) {
	return this->LiveOutSet.end();
}

// Get iterator for the start of the VarKill set.
STARSOpndSetIter SMPFunction::GetFirstVarKill(void) {
	return this->KillSet.begin();
}

// Get termination iterator marker for the VarKill set.
STARSOpndSetIter SMPFunction::GetLastVarKill(void) {
	return this->KillSet.end();
}

// Compute InRegs, OutRegs, LiveOut, Kill sets for function
void SMPFunction::ComputeGlobalSets(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		STARSOpndSetIter LVASetIter;
		pair<STARSOpndSetIter, bool> InsertResult;
		for (LVASetIter = CurrBlock->GetFirstVarKill(); LVASetIter != CurrBlock->GetLastVarKill(); ++LVASetIter) {
			STARSOpndTypePtr KilledOp = (*LVASetIter);
			InsertResult = this->KillSet.insert(KilledOp); // insert will often fail due to duplication, which is expected
		}
		if (CurrBlock->HasReturn()) {
			// LiveOut for return blocks are unioned to make the LiveOut for the function.
			for (LVASetIter = CurrBlock->GetFirstLiveOut(); LVASetIter != CurrBlock->GetLastLiveOut(); ++LVASetIter) {
				STARSOpndTypePtr LiveOutOp = (*LVASetIter);
				InsertResult = this->LiveOutSet.insert(LiveOutOp); // insert will often fail due to duplication, which is expected
			}
		}
	}
	return;
} // end of SMPFunction::ComputeGlobalSets()

// compute InputRegs (USEs) and OutputRegs (DEFs), only inherit from callees on InheritPass
bool SMPFunction::ComputeInOutRegs(bool InheritPass, bool &WritesMem, bool &CallChainNonReturning) {
	bool Changed = false;
	bool MemoryInput = false;
	bool MemoryOutput = false;
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();
	CallChainNonReturning = false;

	if (InheritPass) {
		if (VerboseOutput) {
			SMP_msg("INFO: InheritPass for ComputeInOutRegs(), function %s at %llx\n", this->GetFuncName(), (uint64_t) this->GetFirstFuncAddr());
		}
		// Look at all callees and union their bitsets in.
		size_t OldInputBitCount = this->InputRegs.count();
		size_t OldOutputBitCount = this->OutputRegs.count();
		size_t OldCalleePreservedRegsBitCount = this->CalleePreservedRegs.count();
		size_t NumCallees = this->GetNumCallTargets();
		for (size_t CalleeIndex = 0; CalleeIndex < NumCallees; ++CalleeIndex) {
			STARS_ea_t CalleeAddr = this->GetCallTargetAddr(CalleeIndex);
			if ((STARS_BADADDR != CalleeAddr) && (!STARS_IsExternalFunc(CalleeAddr))) {
				SMPFunction *CalleeFunc = this->GetProg()->FindFunction(CalleeAddr);
				if (nullptr != CalleeFunc) {
					this->InputRegs |= CalleeFunc->GetInputRegs();
					MemoryOutput |= CalleeFunc->AltersSPARKMemory();
					this->CalleeHasLoopInArgMemWrites |= CalleeFunc->UsesInArgsForLoopMemWrites();
					CallChainNonReturning |= (CalleeFunc->HasCallToNonReturningFunc() || CalleeFunc->HasCalleeChainWithNonReturningFunc());

					// Examine whether the callee preserves a reg before deciding how to record the OutputRegs info.
					std::bitset<1 + MD_LAST_REG_NO> TempCalleeOutputRegs = CalleeFunc->GetOutputRegs();
					std::bitset<1 + MD_LAST_REG_NO> TempCalleePreservedRegs = CalleeFunc->GetPreservedRegs();
					std::bitset<1 + MD_LAST_REG_NO> TempCalleeChainPreservedRegs = CalleeFunc->GetCalleePreservedRegs();
					for (size_t RegNo = 0; RegNo < TempCalleeOutputRegs.size(); ++RegNo) {
						if (TempCalleeOutputRegs[RegNo]) {
							// If RegNo is Preserved, record in CalleePreservedRegs.
							if (TempCalleePreservedRegs[RegNo]) {
								this->CalleePreservedRegs.set(RegNo);
							}
							else {
								// Altered, not preserved
								this->OutputRegs.set(RegNo);
							}
						}
						else if (TempCalleePreservedRegs[RegNo]) { // preserved in callee, not explicitly altered
							this->CalleePreservedRegs.set(RegNo);
						}
						else if (TempCalleeChainPreservedRegs[RegNo]) { // preserved in callee chain, not explicitly altered
							this->CalleePreservedRegs.set(RegNo);
						}
					}
				}
			}
		} // end for all callees
		Changed = ((OldInputBitCount < this->InputRegs.count()) || (OldOutputBitCount < this->OutputRegs.count()) || (OldCalleePreservedRegsBitCount < this->CalleePreservedRegs.count()));
	}

	// We need to record registers read and written in instructions, excluding calls with their
	//  conservative DEF and USE lists, and excluding the SSA Marker inst with its pseudo-DEFs.
	for (size_t BlockNum = 0; BlockNum < this->RPOBlocks.size(); ++BlockNum) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockNum];
		if (!InheritPass) {
			Changed = true;
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				if (STARS_IsSSAMarkerPseudoID(InstAddr))
					continue;

				SMPitype FlowType = CurrInst->GetDataFlowType();
				if ((DEFAULT != FlowType) && (JUMP != FlowType) && (INDIR_JUMP != FlowType)) {
					if ((CALL == FlowType) || (INDIR_CALL == FlowType) || (RETURN == FlowType)) {
						// Calls and returns change the stack pointer.
						this->OutputRegs.set((size_t) MD_STACK_POINTER_REG, true);
						this->InputRegs.set((size_t) MD_STACK_POINTER_REG, true);
					}
					continue;  // exclude CALL, INDIR_CALL, RETURN with their conservative USE and DEF lists
				}
				for (STARSDefUseIter UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
					STARSOpndTypePtr UseOp = UseIter->GetOp();
					if (UseOp->IsRegOp() || UseOp->IsFloatingPointRegOp()) {
						STARS_regnum_t RegNo = UseOp->GetReg();
						this->InputRegs.set((size_t) RegNo, true);
					}
					else if (UseOp->IsMemOp()) {
						MemoryInput = true;
					}
				} // end for all USEs

				for (STARSDefUseIter DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
					STARSOpndTypePtr DefOp = DefIter->GetOp();
					if (DefOp->IsRegOp() || DefOp->IsFloatingPointRegOp()) {
						STARS_regnum_t RegNo = DefOp->GetReg();
						assert(0 <= (int) RegNo);
						this->OutputRegs.set((size_t) RegNo, true);
					}
					else if (DefOp->IsMemOp()) {
						MemoryOutput = true;
					}
				} // end for all DEFs
			} // end for all insts in block
			if (!MemoryOutput && CurrBlock->HasMemoryWrite()) {
				SMP_msg("INFO: SPARK: Found new MemoryOutput using HasMemoryWrite\n");
				MemoryOutput = true;  // could be non-stack mem, not in DEFs
			}
		}
		else if (CurrBlock->HasCallInstruction()) { // InheritPass
			int LoopNum = (this->GetNumLoops() > 0) ? this->GetInnermostLoopNum(CurrBlock->GetNumber()) : -1;
			size_t LoopNumOffsetByOne = (size_t)(LoopNum + 1);
			// Note: LoopNum could be -1 if we are outside of any loop.
			list<size_t> LoopList;
			if (LoopNum >= 0) {
				this->BuildLoopList((int) BlockNum, LoopList);
			}
			string CurrFuncName(this->GetFuncName());

			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				// Find call instruction and determine callee.
				SMPInstr *CurrInst = (*InstIter);
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				SMPitype FlowType = CurrInst->GetDataFlowType();
				if ((CALL == FlowType) || (INDIR_CALL == FlowType) || ((RETURN == FlowType) && CurrInst->IsTailCall())) {
					STARS_ea_t CalleeAddr = CurrInst->GetCallTarget();
					if ((STARS_BADADDR != CalleeAddr) && (!STARS_IsExternalFunc(CalleeAddr))) {
						SMPFunction *CalleeFunc = this->GetProg()->FindFunction(CalleeAddr);
						if ((nullptr != CalleeFunc) && CalleeFunc->StackPtrAnalysisSucceeded() && CalleeFunc->HasStructuredControlFlow()) {
							if (VerboseOutput) {
								SMP_msg("INFO: SPARK: Function %s inheriting mem exprs from callee %s\n", CurrFuncName.c_str(), CalleeFunc->GetFuncName());
							}
							size_t NumCalleeLoops = CalleeFunc->GetNumLoops();
							for (size_t CalleeLoopIndexPlusOne = 0; CalleeLoopIndexPlusOne <= NumCalleeLoops; ++CalleeLoopIndexPlusOne) {
								// Get non-loop mem write exprs that trace back to InArgs, trace them back from actual arg in this function.
								size_t NumCalleeInArgMemAddrs = CalleeFunc->GetNumInArgsUsedInMemWrites(CalleeLoopIndexPlusOne);
								pair<STARSExprSetIter, bool> InsertResult;
								for (size_t InArgIndex = 0; InArgIndex < NumCalleeInArgMemAddrs; ++InArgIndex) {
									size_t MemExprWidth = CalleeFunc->GetInArgMemWriteWidth(CalleeLoopIndexPlusOne, InArgIndex);
									uint32_t ExprSerialNumber = CalleeFunc->GetInArgExprUsedInMemWrite(CalleeLoopIndexPlusOne, InArgIndex);
									STARSExprSetIter CurrentMemExprIter = CalleeFunc->FindInArgExprBySerialNumber(CalleeLoopIndexPlusOne, ExprSerialNumber);
									assert(CurrentMemExprIter != CalleeFunc->GetLastInArgExprSetIter(CalleeLoopIndexPlusOne));
									STARSExpression *CurrentMemAddrExpr = (*CurrentMemExprIter)->Clone();
									Changed |= this->InheritCalleeMemExpr(MemExprWidth, CurrentMemAddrExpr, CurrInst, LoopNum, LoopList);
								} // end for ( ... InArgIndex < NumCalleeInArgMemAddrs ...)
							}

							// Inherit all of the callee's inherited MemExprs that traced back to InArgs.
							for (size_t CalleeLoopIndexPlusOne = 0; CalleeLoopIndexPlusOne <= NumCalleeLoops; ++CalleeLoopIndexPlusOne) {
								size_t NumCalleeInheritedExprs = CalleeFunc->GetNumInheritedMemWriteExprs(CalleeLoopIndexPlusOne);
								for (size_t ExprIndex = 0; ExprIndex < NumCalleeInheritedExprs; ++ExprIndex) {
									size_t MemExprWidth = CalleeFunc->GetInheritedMemWriteWidth(CalleeLoopIndexPlusOne, ExprIndex);
									STARSExprSetIter ExprIter = CalleeFunc->GetInheritedMemWriteExpr(CalleeLoopIndexPlusOne, ExprIndex);
									STARSExpression *CurrentMemAddrExpr = (*ExprIter)->Clone();
									Changed |= this->InheritCalleeMemExpr(MemExprWidth, CurrentMemAddrExpr, CurrInst, LoopNum, LoopList);
								}
							}

							// Do the same for callee loop mem write range exprs.
							if (0 < CalleeFunc->GetNumLoopMemWriteExpandedExprBoundsEntries()) {
								for (size_t LoopIndex = 0; LoopIndex < NumCalleeLoops; ++LoopIndex) {

									STARSExprBoundsIter ExprIter;
									for (ExprIter = CalleeFunc->GetFirstLoopMemWriteExpandedExprBoundsIter(LoopIndex); ExprIter != CalleeFunc->GetLastLoopMemWriteExpandedExprBoundsIter(LoopIndex); ++ExprIter) {
										STARSExpression *LowerExpr = ExprIter->first;
										STARSExpression *UpperExpr = ExprIter->second;

										if ((nullptr != LowerExpr) && (nullptr != UpperExpr)) {
											// Set up a range expr so we can re-use this->InheritCalleeMemRangeExpr()
											STARSExpression *TempMemRangeExpr = new STARSExpression();
											TempMemRangeExpr->SetLeftTree(LowerExpr->Clone());
											TempMemRangeExpr->SetRightTree(UpperExpr->Clone());
											TempMemRangeExpr->SetParentFunc(TempMemRangeExpr->GetLeftTree()->GetParentFunc());
											TempMemRangeExpr->SetParentInst(TempMemRangeExpr->GetLeftTree()->GetParentInst());
											TempMemRangeExpr->SetOriginalParentInst(TempMemRangeExpr->GetLeftTree()->GetOriginalParentInst());
											TempMemRangeExpr->SetOperator(SMP_LESS_THAN); // LowerBound < UpperBound is range expr
											size_t MemWriteByteWidth = LowerExpr->FindOrigMemOpByteWidth();
											if (0 == MemWriteByteWidth) {
												MemWriteByteWidth = UpperExpr->FindOrigMemOpByteWidth();
											}
											if (0 == MemWriteByteWidth) {
												SMP_msg("ERROR: expr Inherited byte width of 0 from CalleeFunc %s ; expr dump follows.\n",
													CalleeFunc->GetFuncName());
												if (VerboseOutput)
													LowerExpr->Dump(0);
												MemWriteByteWidth = global_STARS_program->GetSTARS_ISA_Bytewidth(); // default
											}

											Changed |= this->InheritCalleeMemRangeExpr(TempMemRangeExpr, CurrInst, MemWriteByteWidth, LoopNumOffsetByOne, LoopList);
										}
									} // end for all expanded expr pairs in callee func for current LoopIndex
								} // end for all loops in CalleeFunc
							}

							// Do the same for callee looping string mem write range exprs.
							for (size_t LoopIndexPlusOne = 0; LoopIndexPlusOne <= CalleeFunc->GetNumLoops(); ++LoopIndexPlusOne) {
								size_t ExprIndexLimit = CalleeFunc->GetNumStringMemWriteRangeExprs(LoopIndexPlusOne);
								for (size_t i = 0; i < ExprIndexLimit; ++i) {
									STARSExprSetIter StringExprIter = CalleeFunc->GetStringMemWriteRangeExpr(LoopIndexPlusOne, i);
									STARSExpression *TempStringRangeExpr = (*StringExprIter);
									assert(nullptr != TempStringRangeExpr);
									STARSExpression *CalleeStringRangeExpr = TempStringRangeExpr->Clone();
									size_t MemWriteByteWidth = CalleeFunc->GetStringMemWriteRangeWidth(LoopIndexPlusOne, i);
									if (VerboseOutput) {
										SMP_msg("INFO: SPARK: Inheriting CalleeStringRangeExpr from CalleeAddr %llx at CallAddr %llx\n",
											(uint64_t)CalleeAddr, (uint64_t)InstAddr);
									}
									Changed |= this->InheritCalleeMemRangeExpr(CalleeStringRangeExpr, CurrInst, MemWriteByteWidth, LoopNumOffsetByOne, LoopList);
								}
							} // end for all loops in CalleeFunc
						}
					}
				}
			}
		}
	} // end for all blocks

	WritesMem = MemoryOutput;

	return Changed;
} // end of SMPFunction::ComputeInOutRegs()

// Return true if mem expr inherited and added to MemAddrExprWidthsFromCallees
bool SMPFunction::InheritCalleeMemExpr(size_t MemExprWidth, STARSExpression *CurrentMemAddrExpr, SMPInstr *CallInst, int LoopNum, const list<size_t> &LoopList) {
	// Trace actual arg register back, substitute source into clone of callee mem addr expr to produce
	//  the mem addr expr for this function.
	STARSOpndTypePtr InArgOp = CurrentMemAddrExpr->GetLeftOperand();
	STARS_ea_t InstAddr = CallInst->GetAddr();
	size_t LoopNumOffsetByOne = (size_t) (LoopNum + 1);
	bool changed = false;
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();
	if (!CurrentMemAddrExpr->HasLeftSubTree() && InArgOp->IsRegOp() && (!InArgOp->MatchesReg(MD_STACK_POINTER_REG))) { // simple LeftOperand == InArg case
		STARSDefUseIter UseIter = CallInst->FindUse(InArgOp);
		assert(UseIter != CallInst->GetLastUse());
		// Start changing values in our cloned expr to match this function.
		CurrentMemAddrExpr->SetParentFunc(this);
		CurrentMemAddrExpr->SetParentInst(CallInst);
		CurrentMemAddrExpr->SetLeftUseAddr(InstAddr);
		CurrentMemAddrExpr->SetLeftSSANum(UseIter->GetSSANum());
		CurrentMemAddrExpr->SetLeftPreLoopDefAddr(STARS_BADADDR);
		if (VerboseOutput) {
			SMP_msg("INFO: MemAddrExprFromCallee before expansion: Width: %zu", MemExprWidth);
			CurrentMemAddrExpr->Dump(0);
		}
		bool StoppedOnIV = false;
		bool changed = false;
		set<int> DummyLoopRegHashes;
		set<STARS_ea_t> StackPtrCopySet;
		int DepthCounter = 0;
		if (CurrentMemAddrExpr->ExpandExpr(InstAddr, (size_t) LoopNum, false, true, true, false, false, DummyLoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
			if (!StoppedOnIV) {
				// CurrentMemAddrExpr has been expanded successfully to an InArg or constant.
				CurrentMemAddrExpr->EvaluateConsts();
				bool Simplified = CurrentMemAddrExpr->SimplifyDriver();
				pair<STARSExprSetIter, bool> InsertResult = this->MemAddrExprsFromCallees[LoopNumOffsetByOne].insert(CurrentMemAddrExpr);
				if (InsertResult.second) { // new mem expr
					pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, MemExprWidth);
					this->MemAddrExprWidthsFromCallees[LoopNumOffsetByOne].push_back(InsertValue);
					if (VerboseOutput) {
						SMP_msg("INFO: MemAddExprFromCallee after expansion: Width: %zu", MemExprWidth);
						CurrentMemAddrExpr->Dump(0);
					}
					std::bitset<1 + MD_LAST_REG_NO> InArgRegNums;
					CurrentMemAddrExpr->ListInArgRegsUsed(InArgRegNums);
					// Mark all loops containing this block as having callee mem writes.
					for (list<size_t>::const_iterator LoopIter = LoopList.cbegin(); LoopIter != LoopList.cend(); ++LoopIter) {
						this->LoopHasCalleeMemWrites[*LoopIter] = true;
					}
					changed = true;
					this->MemRangeRegsBitmap |= InArgRegNums;

					if (LoopNum >= 0) {
						// If CurrentMemAddrExpr expands to a stack frame write in the loop's containing procedure,
						//  update the stack frame written maps.
						STARS_sval_t FinalStackPtrOffset;
						// CurrentMemAddrExpr has already expanded to the incoming stack pointer reg, so the
						//  expression is already normalized to a negative offset from the return address location.
						//  We pass in zero as the CurrentStackPtrOffset to avoid double-normalization.
						if (CurrentMemAddrExpr->IsStackPtrOffset(0, FinalStackPtrOffset)) {
							// Update for all loops that contain CallInst.
							int CallBlockNum = CallInst->GetBlock()->GetNumber();
							list<size_t> LoopList;
							this->BuildLoopList(CallBlockNum, LoopList);
							for (list<size_t>::const_iterator LoopIter = LoopList.cbegin(); LoopIter != LoopList.cend(); ++LoopIter) {
								size_t CurrLoopNum = (*LoopIter);
								this->UpdateStackBytesWrittenByLoop(FinalStackPtrOffset, MemExprWidth, CurrLoopNum);
								for (size_t RegNum = 0; RegNum < InArgRegNums.size(); ++RegNum) {
									if (InArgRegNums[RegNum]) {
										this->LoopMemRangeInArgRegsBitmap[CurrLoopNum].set(RegNum);
									}
								}
							}
						}
					}
				}
				else {
					SMP_msg("INFO: Expr insert collision at CallAddr %llx\n", (uint64_t)InstAddr);
				}
			}
			else { // StoppedOnIV
				SMP_msg("INFO: SPARK: Expr Expand() StoppedOnIV in InheritCalleeMemExpr() at %llx\n",
					(uint64_t) InstAddr);
				pair<STARSExprSetIter, bool> InsertResult = this->StoppedOnIVNonRangeExprs[LoopNumOffsetByOne].insert(CurrentMemAddrExpr);
				if (InsertResult.second) { // new Expr
					pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, MemExprWidth);
					this->StoppedOnIVNonRangeIterWidths[LoopNumOffsetByOne].push_back(InsertValue);
				}
			}
		}
		else {
			SMP_msg("ERROR: ExpandExpr() failure in InheritCalleeMemExpr() at CallAddr %llx ", (uint64_t) InstAddr);
			this->DumpFuncNameAndAddr();
		}
	}
	return changed;
} // end of SMPFunction::InheritCalleeMemExpr()

bool SMPFunction::InheritCalleeMemRangeExpr(STARSExpression *CalleeMemRangeExpr, SMPInstr *CallInst, size_t MemWriteByteWidth, size_t LoopNumPlusOne, const list<size_t> &LoopList) {
	// Again, deal with the simple to analyze case in which we can find the InArg on the lhs easily.
	bool FoundInArg = false;
	bool Changed = false;
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();
	STARS_ea_t InstAddr = CallInst->GetAddr();

	// Start changing values in our cloned expr to match this function.
	CalleeMemRangeExpr->SetParentFunc(this);
	CalleeMemRangeExpr->SetParentInst(CallInst);
	bitset<1 + MD_LAST_REG_NO> InArgRegNums;
	CalleeMemRangeExpr->ListInArgRegsUsed(InArgRegNums);
	bool FoundInArgReg = InArgRegNums.any();

	if (FoundInArgReg) {
		bool ReplacedInArgSSANum = false;
		for (size_t RegNum = 0; RegNum < InArgRegNums.size(); ++RegNum) {
			bool StackPointerReg = (RegNum == MD_STACK_POINTER_REG);
			if (!StackPointerReg && InArgRegNums[RegNum]) {
				STARSOpndTypePtr CurrInArgOp = CallInst->MakeRegOpnd((STARS_regnum_t) RegNum);
				STARSDefUseIter UseIter = CallInst->FindUse(CurrInArgOp);
				if (UseIter != CallInst->GetLastUse()) {
					// Replace all SSANum == 0 instances of InArgOp with SSANum of InArg in this function.
					// NOTE: Might we need to do the same for multiple InArgs, used as limits, etc.?
					CalleeMemRangeExpr->SubstituteSSANum(CallInst, UseIter->GetSSANum(), CurrInArgOp);
					ReplacedInArgSSANum = true;
				}
				else {
					SMP_msg("ERROR: SPARK: Callee loop range does not trace to InArgUSE at call site %llx\n", (uint64_t) InstAddr);
				}
			}
		} // end for all InArgRegNums
		if (!ReplacedInArgSSANum) {
			SMP_msg("ERROR: SPARK: Callee loop range expr did not see InArgSSANums replaced at call site %llx ", (uint64_t) InstAddr);
			this->DumpFuncNameAndAddr();
		}
		if (VerboseOutput) {
			SMP_msg("INFO: LoopCalleeMemRangeExpr before expansion:");
			CalleeMemRangeExpr->Dump(0);
		}
		bool StoppedOnIV = false;
		bool changed = false;
		set<int> DummyLoopRegHashes;
		set<STARS_ea_t> StackPtrCopySet;
		int DepthCounter = 0;
		if (CalleeMemRangeExpr->ExpandExpr(InstAddr, 0, false, true, true, false, false, DummyLoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
			if (!StoppedOnIV) {
				// CalleeMemRangeExpr has been expanded successfully to an InArg or constant.
				CalleeMemRangeExpr->EvaluateConsts();
				bool Simplified = CalleeMemRangeExpr->SimplifyDriver();
				pair<STARSExprSetIter, bool> InsertResult = this->LoopMemAddrExprsFromCalleeLoops[LoopNumPlusOne].insert(CalleeMemRangeExpr);
				if (InsertResult.second) { // new mem expr
					InArgRegNums.reset();
					this->HasMemExprsFromCalleeLoops = true;
					Changed = true;
					CalleeMemRangeExpr->ListInArgRegsUsed(InArgRegNums);
					pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, MemWriteByteWidth);
					this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].push_back(InsertValue);
					if (VerboseOutput) {
						SMP_msg("INFO: LoopCalleeMemRangeExpr after expansion:");
						CalleeMemRangeExpr->Dump(0);
					}
					this->MemRangeRegsBitmap |= InArgRegNums;
					// Mark all loops containing this block as having callee mem writes.
					for (list<size_t>::const_iterator LoopIter = LoopList.cbegin(); LoopIter != LoopList.cend(); ++LoopIter) {
						size_t CurrLoopNum = *LoopIter;
						this->LoopHasCalleeMemWrites[CurrLoopNum] = true;
						// Mark the InArg regs reached in the Expand() operation.
						for (size_t RegNum = 0; RegNum < InArgRegNums.size(); ++RegNum) {
							if (InArgRegNums[RegNum]) {
								this->LoopMemRangeInArgRegsBitmap[CurrLoopNum].set(RegNum);
							}
						}
					}
				}
				else {
					SMP_msg("INFO: Expr insert collision at CallAddr %llx\n", (uint64_t) InstAddr);
				}
			}
			else { // StoppedOnIV
				SMP_msg("ERROR: SPARK: Expr Expand() StoppedOnIV in InheritCalleeMemRangeExpr() at %llx ", (uint64_t) InstAddr);
				this->DumpFuncNameAndAddr();
				this->CalleeMemExprProblems[LoopNumPlusOne] = true;
				pair<STARSExprSetIter, bool> InsertResult = this->StoppedOnIVMemRangeExprs[LoopNumPlusOne].insert(CalleeMemRangeExpr);
				if (InsertResult.second) { // new Expr
					pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, MemWriteByteWidth);
					this->StoppedOnIVMemRangeIterWidths[LoopNumPlusOne].push_back(InsertValue);
				}
			}
		}
		else {
			SMP_msg("ERROR: SPARK: ExpandExpr() failure in InheritCalleeMemRangeExpr at CallAddr %llx ", (uint64_t) InstAddr);
			this->DumpFuncNameAndAddr();
			this->CalleeMemExprProblems[LoopNumPlusOne] = true;
		}
	}
	else if (!FoundInArgReg) {
		pair<STARSExprSetIter, bool> InsertResult = this->LoopMemAddrExprsFromCalleeLoops[LoopNumPlusOne].insert(CalleeMemRangeExpr);
		if (InsertResult.second) { // new mem expr
			this->HasMemExprsFromCalleeLoops = true;
			Changed = true;
			pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, MemWriteByteWidth);
			this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].push_back(InsertValue);
			if (VerboseOutput) {
				SMP_msg("INFO: Global static inherited LoopCalleeMemRangeExpr:");
				CalleeMemRangeExpr->Dump(0);
			}
			// Mark all loops containing this block as having callee mem writes.
			for (list<size_t>::const_iterator LoopIter = LoopList.cbegin(); LoopIter != LoopList.cend(); ++LoopIter) {
				size_t CurrLoopNum = *LoopIter;
				this->LoopHasCalleeMemWrites[CurrLoopNum] = true;
			}
		}
		else {
			SMP_msg("INFO: Expr insert collision at CallAddr %llx\n", (uint64_t) InstAddr);
		}
	}
	return Changed;
} // end of SMPFunction::InheritCalleeMemRangeExpr()

// Split relational RangeExpr into upper and lower bonds exprs and save them and their width
void SMPFunction::SplitAndSaveRelationalExpr(bool PositiveIncrement, std::size_t LoopNumPlusOne, size_t MemWidth, STARSExpression *RangeExpr) {
	assert(nullptr != RangeExpr);
	assert((0 <= LoopNumPlusOne) && (LoopNumPlusOne <= this->GetNumLoops()));
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();
	STARSExpression *LowerExpr;
	STARSExpression *UpperExpr;
	RangeExpr->SplitMemoryRangeExpr(PositiveIncrement, LowerExpr, UpperExpr);
	assert(nullptr != LowerExpr);
	assert(nullptr != UpperExpr);
	if (VerboseOutput) {
		SMP_msg("INFO: SPARK: RelationalExpr being split: \n");
		RangeExpr->Dump(0);
	}
	pair<STARSExprSetIter, bool> InsertResult = this->RelationalLowerBoundExprs[LoopNumPlusOne].insert(LowerExpr);
	pair<STARSExprSetIter, bool> InsertResult2 = this->RelationalUpperBoundExprs[LoopNumPlusOne].insert(UpperExpr);
	if (InsertResult.second || InsertResult2.second) { // new range if either one was new expr
		pair<STARSExprSetIter, STARSExprSetIter> BoundsIters(InsertResult.first, InsertResult2.first);
		pair<size_t, pair<STARSExprSetIter, STARSExprSetIter> > InsertValue(MemWidth, BoundsIters);
		this->RelationalMemWriteWidths[LoopNumPlusOne].push_back(InsertValue);
	}
	else {
		SMP_msg("INFO: SPARK: Duplicate RelationalExpr not saved.\n");
	}

	return;
} // end of SMPFunction::SplitAndSaveRelationalExpr()

// Build range expr for memory writes in looping string opcode
void SMPFunction::BuildLoopingStringMemExprs(SMPBasicBlock *CurrBlock, SMPInstr *CurrInst) {
	assert(CurrBlock->HasLoopingStringOpcode());
	unsigned short opcode = CurrInst->GetIDAOpcode();
	STARS_ea_t InstAddr = CurrInst->GetAddr();
	bool NonWritingLoopingStringOperation = ((opcode == STARS_NN_cmps) || (opcode == STARS_NN_scas) || (opcode == STARS_NN_outs));
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();
	if (!NonWritingLoopingStringOperation) {
		// Make Expr like: RDI < (RDI + (RCX*ByteWidth))
		int CurrBlockNum = CurrBlock->GetNumber();
		STARSOpndTypePtr FirstOpnd = CurrInst->GetOperand(0);
		uint16_t ByteWidth = FirstOpnd->GetByteWidth();
		STARSExpression *LowerBoundExpr = new STARSExpression();
		LowerBoundExpr->SetParentFunc(this);
		LowerBoundExpr->SetParentInst(CurrInst);
		LowerBoundExpr->SetOriginalParentInst(CurrInst);
		LowerBoundExpr->SetOperator(SMP_ASSIGN);
		STARSOpndTypePtr UseOp = CurrInst->MakeRegOpnd(STARS_x86_R_di);
		STARSDefUseIter UseIter = CurrInst->FindUse(UseOp);
		assert(UseIter != CurrInst->GetLastUse());
		LowerBoundExpr->SetLeftOperand(UseOp);
		LowerBoundExpr->SetLeftUseAddr(InstAddr);
		LowerBoundExpr->SetLeftSSANum(UseIter->GetSSANum());

		int LoopIndex = this->GetInnermostLoopNum(CurrBlockNum);
		size_t LoopNum = (size_t) LoopIndex;
		int LoopIndexPlusOne = LoopIndex + 1;
		bool InsideLoop = (0 <= LoopIndex);
		bool StoppedOnIV = false;
		bool changed = false;
		if (!InsideLoop) { // not in a loop
			LoopNum = 0;
		}
		set<int> LoopRegHashes;
		set<STARS_ea_t> StackPtrCopySet;
		int DepthCounter = 0;
		if (LowerBoundExpr->ExpandExpr(InstAddr, LoopNum, false, true, true, true, false, LoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
			if (StoppedOnIV) {
				this->SymbolicAnalysisProblems[LoopIndexPlusOne] = true;
				SMP_msg("ERROR: SPARK: LowerBoundExpr Expand() StoppedOnIV in BuildLoopingStringMemExprs() at %llx ", (uint64_t)InstAddr);
				this->DumpFuncNameAndAddr();
				pair<STARSExprSetIter, bool> InsertResult = this->StoppedOnIVMemRangeExprs[LoopIndexPlusOne].insert(LowerBoundExpr);
				if (InsertResult.second) { // new Expr
					pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, ByteWidth);
					this->StoppedOnIVMemRangeIterWidths[LoopIndexPlusOne].push_back(InsertValue);
				}
			}
			else {
				LowerBoundExpr->EvaluateConsts();
				LowerBoundExpr->SimplifyDriver();
				// The upper bound is RDI + RCX*ByteWidth. Optimize away ByteWidth == 1.
				//  Clone the expanded and simplified RDI from LowerBoundExpr to save time.
				STARSExpression *UpperBoundExpr = new STARSExpression();
				UpperBoundExpr->SetParentFunc(this);
				UpperBoundExpr->SetParentInst(CurrInst);
				UpperBoundExpr->SetOriginalParentInst(CurrInst);
				UpperBoundExpr->SetOperator(SMP_ADD);
				UpperBoundExpr->SetLeftTree(LowerBoundExpr->Clone());
				UpperBoundExpr->SetLeftPreLoopDefAddr(LowerBoundExpr->GetLeftPreLoopDefAddr());
				STARSOpndTypePtr CounterOp = CurrInst->MakeRegOpnd(STARS_x86_R_cx);
				STARSDefUseIter CounterUseIter = CurrInst->FindUse(CounterOp);
				assert(CounterUseIter != CurrInst->GetLastUse());
				if (1 == ByteWidth) { // Simplify RCX * 1 to RCX
					UpperBoundExpr->SetRightOperand(CounterOp);
					UpperBoundExpr->SetRightUseAddr(InstAddr);
					UpperBoundExpr->SetRightSSANum(CounterUseIter->GetSSANum());
				}
				else { // Create RCX*ByteWidth right tree.
					STARSExpression *RightTree = new STARSExpression();
					RightTree->SetParentFunc(this);
					RightTree->SetParentInst(CurrInst);
					RightTree->SetOriginalParentInst(CurrInst);
					RightTree->SetOperator(SMP_U_MULTIPLY);
					RightTree->SetLeftOperand(CounterOp);
					RightTree->SetLeftUseAddr(CurrInst->GetAddr());
					RightTree->SetLeftSSANum(CounterUseIter->GetSSANum());
					RightTree->SetRightOperand(CurrInst->MakeImmediateOpnd((STARS_uval_t)ByteWidth));
					UpperBoundExpr->SetRightTree(RightTree);
				}
				// Expand only the right side, as left side was expanded, simplified and cloned from lower bound.
				DepthCounter = 0;
				if (UpperBoundExpr->ExpandExpr(InstAddr, LoopNum, true, true, true, true, false, LoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
					if (!StoppedOnIV) {
						if (InsideLoop) {
							this->LoopRegHashSets[LoopNum] = LoopRegHashes;
						}
						if (UpperBoundExpr->HasRightSubTree()) {
							UpperBoundExpr->GetRightTree()->EvaluateConsts();
							UpperBoundExpr->GetRightTree()->SimplifyExpr(UpperBoundExpr);
							UpperBoundExpr->SimplifyDriver();
							// Success. Record exprs in container.
							STARSExpression *RangeExpr = new STARSExpression();
							RangeExpr->SetOperator(SMP_LESS_THAN);
							RangeExpr->SetParentFunc(this);
							RangeExpr->SetParentInst(LowerBoundExpr->GetParentInst());
							RangeExpr->SetOriginalParentInst(LowerBoundExpr->GetOriginalParentInst());
							// If LowerBoundExpr is just SMP_ASSIGN and an operand, then grab the operand as our LeftOperand.
							//  Otherwise, grab its left tree.
							if (LowerBoundExpr->GetOperator() == SMP_ASSIGN) {
								RangeExpr->SetLeftSSANum(LowerBoundExpr->GetLeftSSANum());
								RangeExpr->SetLeftPreLoopDefAddr(LowerBoundExpr->GetLeftPreLoopDefAddr());
								if (LowerBoundExpr->HasLeftSubTree()) {
									RangeExpr->SetLeftTree(LowerBoundExpr->GetLeftTree());
								}
								else {
									RangeExpr->SetLeftOperand(LowerBoundExpr->GetLeftOperand());
									RangeExpr->SetLeftUseAddr(LowerBoundExpr->GetLeftUseAddr());
								}
							}
							else { // complex expression
								RangeExpr->SetLeftTree(LowerBoundExpr);
							}
							// Ditto for UpperBoundExpr, except it cannot be a simple SMP_ASSIGN.
							RangeExpr->SetRightTree(UpperBoundExpr);
							RangeExpr->SetRightPreLoopDefAddr(UpperBoundExpr->GetLeftPreLoopDefAddr());
							pair<STARSExprSetIter, bool> InsertResult = this->StringMemWriteRangeExprs[(size_t)LoopIndexPlusOne].insert(RangeExpr);
							if (InsertResult.second) { // new memory range
								if (VerboseOutput) {
									SMP_msg("INFO: SPARK: LoopingStringRange at %llx\n", InstAddr);
									RangeExpr->Dump(0);
								}
								pair<size_t, STARSExprSetIter> InsertValue(ByteWidth, InsertResult.first);
								this->StringMemWriteRangeWidths[(size_t)LoopIndexPlusOne].push_back(InsertValue);
								// We have already computed the upper bound taking into account the ByteWidth, so pass
								//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
								this->SplitAndSaveRelationalExpr(true, (size_t)LoopIndexPlusOne, 0, RangeExpr);
							}
						}
						else {
							SMP_msg("ERROR: SPARK: No RightTree for BuildLoopingStringMemExprs() at %llx ",	(uint64_t) InstAddr);
							this->DumpFuncNameAndAddr();
							this->SymbolicAnalysisProblems[LoopIndexPlusOne] = true;
						}
					}
					else { // StoppedOnIV
						SMP_msg("ERROR: SPARK: UpperBoundExpr Expand() StoppedOnIV in BuildLoopingStringMemExprs() at %llx ", (uint64_t) InstAddr);
						this->DumpFuncNameAndAddr();
						this->SymbolicAnalysisProblems[LoopIndexPlusOne] = true;
						pair<STARSExprSetIter, bool> InsertResult = this->StoppedOnIVMemRangeExprs[LoopIndexPlusOne].insert(UpperBoundExpr);
						if (InsertResult.second) { // new Expr
							pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, ByteWidth);
							this->StoppedOnIVMemRangeIterWidths[LoopIndexPlusOne].push_back(InsertValue);
						}
					}
				}
				else { // error on UpperBoundExpr->ExpandExpr()
					this->SymbolicAnalysisProblems[LoopIndexPlusOne] = true;
					SMP_msg("ERROR: SPARK: UpperBoundExpr ExpandExpr() failed in BuildLoopingStringMemExprs() at %llx ", (uint64_t) InstAddr);
					this->DumpFuncNameAndAddr();
				}
			}
		}
		else {
			this->SymbolicAnalysisProblems[LoopIndexPlusOne] = true;
			SMP_msg("ERROR: SPARK: LowerBoundExpr ExpandExpr() failed in BuildLoopingStringMemExprs() at %llx ", (uint64_t) InstAddr);
			this->DumpFuncNameAndAddr();
		}
	}
	return;
} // end of SMPFunction::BuildLoopingStringMemExprs()

// Four methods to get values from the maps of global reg/SSA to FG info.
//  For local names, see corresponding methods in SMPBasicBlock.
unsigned short SMPFunction::GetDefSignMiscInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->GlobalDefFGInfoBySSA.end())
		return MapIter->second.SignMiscInfo;
	else
		return 0;
} // end of SMPFunction::GetDefSignMiscInfo()

unsigned short SMPFunction::GetStackDefSignMiscInfo(STARS_ea_t InstAddr) {
	map<STARS_ea_t, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->StackDefFGInfo.find(InstAddr);
	assert(MapIter != this->StackDefFGInfo.end());

	return MapIter->second.SignMiscInfo;
}

unsigned short SMPFunction::GetUseSignMiscInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->GlobalUseFGInfoBySSA.end())
		return MapIter->second.SignMiscInfo;
	else
		return 0;
} // end of SMPFunction::GetUseSignMiscInfo()

unsigned short SMPFunction::GetStackUseSignMiscInfo(STARS_ea_t InstAddr) {
	map<STARS_ea_t, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->StackUseFGInfo.find(InstAddr);
	assert(MapIter != this->StackUseFGInfo.end());

	return MapIter->second.SignMiscInfo;
}

unsigned short SMPFunction::GetDefWidthTypeInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->GlobalDefFGInfoBySSA.end())
		return MapIter->second.SizeInfo;
	else
		return 0;
} // end of SMPFunction::GetDefWidthTypeInfo()

unsigned short SMPFunction::GetUseWidthTypeInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->GlobalUseFGInfoBySSA.end())
		return MapIter->second.SizeInfo;
	else
		return 0;
} // end of SMPFunction::GetUseWidthTypeInfo()

struct FineGrainedInfo SMPFunction::GetDefFGInfo(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->GlobalDefFGInfoBySSA.end())
		return MapIter->second;
	else {
		struct FineGrainedInfo EmptyFG;
		EmptyFG.SignMiscInfo = 0;
		EmptyFG.SizeInfo = 0;
		return EmptyFG;
	}
} // end of SMPFunction::GetDefFGInfo()

struct FineGrainedInfo SMPFunction::GetUseFGInfo(int UseHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter != this->GlobalUseFGInfoBySSA.end())
		return MapIter->second;
	else {
		struct FineGrainedInfo EmptyFG;
		EmptyFG.SignMiscInfo = 0;
		EmptyFG.SizeInfo = 0;
		return EmptyFG;
	}
} // end of SMPFunction::GetUseFGInfo()

// Fetch SPARK Ada control flow type for jump at InstAddr; return FALL_THROUGH if none found.
ControlFlowType SMPFunction::GetControlFlowType(STARS_ea_t InstAddr) const {
	ControlFlowType JumpTypeCode = FALL_THROUGH;
	map<STARS_ea_t, unsigned short>::const_iterator MapIter = this->ControlFlowMap.find(InstAddr);
	if (MapIter != this->ControlFlowMap.end()) {
		JumpTypeCode = (ControlFlowType) MapIter->second;
	}
	return JumpTypeCode;
} // end of SMPFunction::GetControlFlowType()

// Set counter to zero, or insert zero counter if none found
void SMPFunction::ResetJumpToFollowNodeCounter(STARS_ea_t InstAddr) {
	map<STARS_ea_t, uint16_t>::iterator MapIter = this->JumpToFollowNodeCounterMap.find(InstAddr);
	if (MapIter != this->JumpToFollowNodeCounterMap.end()) { // found it
		MapIter->second = 0;
	}
	else {
		pair<STARS_ea_t, uint16_t> CounterItem(InstAddr, 0);
		pair<map<STARS_ea_t, uint16_t>::iterator, bool> InsertResult = this->JumpToFollowNodeCounterMap.insert(CounterItem);
		assert(InsertResult.second);
	}
	return;
} // end of SMPFunction::ResetJumpToFollowNodeCounter()

// Increment counter, or insert count of 1 if none found
void SMPFunction::IncrementJumpToFollowNodeCounter(STARS_ea_t InstAddr) {
	map<STARS_ea_t, uint16_t>::iterator MapIter = this->JumpToFollowNodeCounterMap.find(InstAddr);
	if (MapIter != this->JumpToFollowNodeCounterMap.end()) { // found it
		++MapIter->second;
	}
	else {
		pair<STARS_ea_t, uint16_t> CounterItem(InstAddr, 1);
		pair<map<STARS_ea_t, uint16_t>::iterator, bool> InsertResult = this->JumpToFollowNodeCounterMap.insert(CounterItem);
		assert(InsertResult.second);
	}
	return;
}

void SMPFunction::SetHasUnstructuredCFG(void) { 
	this->HasStructuredCFG = false;
	return;
} // end of SMPFunction::SetHasUnstructuredCFG()

// Add a caller to the list of all callers of this function.
void SMPFunction::AddCallSource(STARS_ea_t addr) {
	// Convert call instruction address to beginning address of the caller.
	STARS_Function_t *FuncInfo = SMP_get_func(addr);
	if (nullptr == FuncInfo) {
		SMP_msg("SERIOUS WARNING: Call location %llx not in a function.\n", (uint64_t) addr);
		return;
	}
	STARS_ea_t FirstAddr = FuncInfo->get_startEA();
	assert(STARS_BADADDR != FirstAddr);
	this->AllCallSources.insert(FirstAddr);
	this->AllCallSites.insert(addr);
	return;
} // end of SMPFunction::AddCallSource()

// Add a direct call target; return true if new target, false if target already added
bool SMPFunction::AddDirectCallTarget(STARS_ea_t addr) {
	assert(STARS_BADADDR != addr);
	pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->DirectCallTargets.insert(addr);
	if (InsertResult.second) { // new direct call target
		this->AllCallTargets.push_back(addr);
	}
	return InsertResult.second;
}

// Add an indirect call target; return true if new target, false if target already added
bool SMPFunction::AddIndirectCallTarget(STARS_ea_t addr) {
	this->SetHasIndirectCalls();
	assert(STARS_BADADDR != addr);
	pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->IndirectCallTargets.insert(addr);
	if (InsertResult.second) { // new indirect call target
		this->AllCallTargets.push_back(addr);
	}
	return InsertResult.second;
}

// Remove TargetAddr from DirectCallTargets and AllCallTargets.
set<STARS_ea_t>::iterator SMPFunction::RemoveDirectCallTarget(STARS_ea_t TargetAddr) {
	set<STARS_ea_t>::iterator TargetIter = this->DirectCallTargets.find(TargetAddr);
	set<STARS_ea_t>::iterator ReturnIter = TargetIter;
	if (TargetIter != this->DirectCallTargets.end()) {
		++ReturnIter;
		this->DirectCallTargets.erase(TargetIter);
		this->RebuildCallTargets(); // remove from AllCallTargets
	}
	return ReturnIter;
}

// Remove TargetAddr from IndirectCallTargets and AllCallTargets.
bool SMPFunction::RemoveIndirectCallTarget(STARS_ea_t TargetAddr) {
	set<STARS_ea_t>::iterator TargetIter = this->IndirectCallTargets.find(TargetAddr);
	bool RemovedTarget = false;
	if (TargetIter != this->IndirectCallTargets.end()) {
		this->IndirectCallTargets.erase(TargetIter);
		RemovedTarget = true;
		this->RebuildCallTargets(); // remove from AllCallTargets
	}
	return RemovedTarget;
}

// Add to map of CallAddr => set of arg-passing addrs
void SMPFunction::AddBufferCallWithInArg(const CallAddrArgDefAddrPair InsertVal) {
	STARS_ea_t CallAddr = InsertVal.first;
	STARS_ea_t ArgAddr = InsertVal.second;

	// Get the argument operand via the ArgAddr.
	SMPInstr *ArgDefInst = this->GetInstFromAddr(ArgAddr);
	assert(nullptr != ArgDefInst);
	STARSDefUseIter ArgDefIter = ArgDefInst->GetFirstNonFlagsDef();
	assert(ArgDefIter != ArgDefInst->GetLastDef());
	STARSOpndTypePtr ArgOp = ArgDefIter->GetOp();

	// Create the arg definition pair.
	STARSDefinition ArgDefn(ArgOp, ArgAddr);

	map<STARS_ea_t, STARSDefinitionSet >::iterator FindIter = this->BufferCallInArgsMap.find(CallAddr);
	if (FindIter == this->BufferCallInArgsMap.end()) {
		// Create a new set of ArgAddr definitions.
		STARSDefinitionSet ArgDefnSet;
		(void) ArgDefnSet.insert(ArgDefn);
		pair<STARS_ea_t, STARSDefinitionSet > InsertVal(CallAddr, ArgDefnSet);
		(void) this->BufferCallInArgsMap.insert(InsertVal);
	}
	else { // Add ArgDefn to the existing set.
		(void) FindIter->second.insert(ArgDefn);
	}
	return;
} // end of SMPFunction::AddBufferCallWithInArg()

// add map entry to LeaInstOpMap
void SMPFunction::AddLeaOperand(STARS_ea_t addr, STARSOpndTypePtr LeaOperand) {
	pair<STARS_ea_t, STARSOpndTypePtr> InsertValue(addr, LeaOperand);
	pair<map<STARS_ea_t, STARSOpndTypePtr>::iterator, bool> InsertResult;
	InsertResult = this->LeaInstOpMap.insert(InsertValue);
	if (!(InsertResult.second)) { // already existed; replace
		map<STARS_ea_t, STARSOpndTypePtr>::iterator FindIter = this->LeaInstOpMap.find(addr);
		assert(FindIter != this->GetLastLeaOperand());
		FindIter->second = LeaOperand;
	}
	return;
}

// Add input arguments to the NormalizedStackOpsMap.
void SMPFunction::AddNormalizedStackOperand(STARSOpndTypePtr OldOp, STARS_ea_t InstAddr, STARSOpndTypePtr NormalizedOp) {
	bool DuplicateCase = false; // e.g. inc [esp+8] will have [esp+8] as a DEF and a USE and maps will see [esp+8] twice
#if SMP_DEBUG_DATAFLOW_VERBOSE
	bool DebugFlag = (InstAddr == 0x8048463);
#endif
	pair<map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator, bool> InsertResult;
	pair<map<STARSDefinition, map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator, LessDefinition>::iterator, bool> InverseInsertResult;
	pair<STARSOpndTypePtr, STARS_ea_t> OldValue(OldOp, InstAddr);
	pair<STARSOpndTypePtr, STARS_ea_t> InverseValue(OldOp, InstAddr); // OldOp was NormalizedOp when it was inserted previously
	pair<pair<STARSOpndTypePtr, STARS_ea_t>, STARSOpndTypePtr> InsertValue(OldValue, NormalizedOp);
	pair<STARSOpndTypePtr, STARS_ea_t> InverseInsertValue(NormalizedOp, InstAddr);
	map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator OldIter = this->NormalizedStackOpsMap.begin();
	pair<STARSDefinition, map<pair<STARSOpndTypePtr, STARS_ea_t>, STARSOpndTypePtr, LessDefinition>::iterator> InverseInsertTriple(InverseInsertValue, OldIter);
	map<STARSDefinition, map<pair<STARSOpndTypePtr, STARS_ea_t>, STARSOpndTypePtr, LessDefinition>::iterator>::iterator InverseIter;

	// If this function calls alloca(), stack operands could be normalized more than once.
	//  Before we proceed, we update an old entry instead of inserting a new entry.
	if (this->CallsAlloca || this->HasPushAfterFrameAlloc()) {
		InverseIter = this->InverseNormalizedStackOpsMap.find(InverseValue);
		if (InverseIter != this->InverseNormalizedStackOpsMap.end()) {
			// We have our alloca() update case. We formerly mapped <A, InstAddr> to B.
			//  Now B is being normalized to C. All we want to do is change the original
			//  map entry so that we map <A, InstAddr> to C. In this manner, A is always the
			//  original un-normalized stack op, available for lookup from an RTL.
			OldIter = InverseIter->second; // OldIter points at map of <A, InstAddr> to B.
			OldIter->second = NormalizedOp; // Change B to C
			// Now we want to erase the Inverse map entry and insert a new one that maps
			//  <C, InstAddr> to OldIter instead of mapping <B, InstAddr> to OldIter.
			(void) this->InverseNormalizedStackOpsMap.erase(InverseIter);
			InverseInsertTriple.second = OldIter;
			InverseInsertResult = this->InverseNormalizedStackOpsMap.insert(InverseInsertTriple);
			assert(InverseInsertResult.second);
			return;
		}
		else {
			// We might have the final difficult case: We have a combination of CallsAlloca and the
			//  DuplicateCase described below (e.g. an increment of a stack location produces a DEF
			//  and a USE of the same location, causing duplicate mappings to be attempted). We need
			//  to detect the duplicate case here. What will happen is that, on the first call to this
			//  method, we will map <A, InstAddr> to B, and reverse-map <B, InstAddr> to A. On the second
			//  call to this method, we will detect the duplicate case and exit. On the third call, caused
			//  by CallsAlloca, we are asked to map <B, InstAddr> to C, and we will correctly hit the code
			//  just above, in the if-clause, to fix the A->B mapping to be an A->C mapping, and we will
			//  erase the reverse mapping of B->A and replace it with the C->A reverse mapping. On the
			//  fourth call to this method, we will not find a reverse mapping B->A any more, so the if-clause
			//  does not execute. We can only detect this case by finding an existing C->A reverse mapping
			//  and an existing A->C mapping to confirm our inference.
			pair<STARSOpndTypePtr, STARS_ea_t> TestInverseValue(NormalizedOp, InstAddr);
			InverseIter = this->InverseNormalizedStackOpsMap.find(TestInverseValue);
			if (InverseIter != this->InverseNormalizedStackOpsMap.end()) {
				// Found existing C->A inverse mapping. Is there an A->C mapping to confirm
				//  our interpretation of the situation?
				pair<STARSOpndTypePtr, STARS_ea_t> TestOldValue(InverseIter->second->first.first, InstAddr);
				map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator TestOldIter;
				TestOldIter = this->NormalizedStackOpsMap.find(TestOldValue);
				if (TestOldIter != this->NormalizedStackOpsMap.end()) {
					// We found a mapping from <A, InstAddr>.
					if (IsEqOp(NormalizedOp, TestOldIter->second)) {
						// The mapping is A->C as suspected.
						return; // duplication; nothing to do in either map.
					}
				}
			}
		}
	}
	// At this point, we have no inverse map entry to worry about, because we are
	//  normalizing this operand for the first time.
	InsertResult = this->NormalizedStackOpsMap.insert(InsertValue);
	OldIter = InsertResult.first;
	if (!(InsertResult.second)) {
		// Already had an entry. That should mean a rare case such as "inc [esp+8]" which
		//  produces a USE and a DEF of the same address. We can confirm that the map has
		//  the same normalized operand we were trying to insert. Otherwise, the collision
		//  is fatal.
		STARSOpndTypePtr OldOldOp = InsertResult.first->first.first;
		STARSOpndTypePtr OldNormalizedOp = InsertResult.first->second;
		assert(IsEqOp(OldOldOp, OldOp) && IsEqOp(OldNormalizedOp, NormalizedOp));
		DuplicateCase = true;
	}
	if (this->CallsAlloca || this->HasPushAfterFrameAlloc()) {
		// We need to add an entry to the inverse map.
		InverseInsertTriple.second = OldIter;
		InverseInsertResult = this->InverseNormalizedStackOpsMap.insert(InverseInsertTriple);
		assert(InverseInsertResult.second || DuplicateCase);
	}
#if SMP_DEBUG_DATAFLOW_VERBOSE
	if (DebugFlag) {
		map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator StackMapIter;
		SMP_msg("DEBUG: NormalizedStackOpsMap size: %zd\n", this->NormalizedStackOpsMap.size());
		for (StackMapIter = this->NormalizedStackOpsMap.begin(); StackMapIter != this->NormalizedStackOpsMap.end(); ++ StackMapIter) {
			STARSOpndTypePtr OldOp = StackMapIter->first.first;
			STARS_ea_t InstAddr = StackMapIter->first.second;
			SMP_msg("DEBUG: NormalizedStackOps: ");
			PrintOperand(OldOp);
			SMP_msg(" addr: %llx\n", (uint64_t) InstAddr);
		}
	}
#endif
	return;
} // SMPFunction::AddNormalizedStackOperand()

// Update MaxDirectStackAccessDelta
void SMPFunction::UpdateMaxDirectStackAccessOffset(STARS_sval_t NewOffset) {
	if (NewOffset > this->MaxDirectStackAccessDelta) {
		this->MaxDirectStackAccessDelta = NewOffset;
	}
	return;
}


// insert jump type into ControlFlowMap
void SMPFunction::SetControlFlowType(STARS_ea_t InstAddr, ControlFlowType JumpTypeCode) {
	map<STARS_ea_t, uint16_t>::iterator MapIter = this->ControlFlowMap.find(InstAddr);
	if (MapIter == this->ControlFlowMap.end()) { // no old entry; insert
		pair<STARS_ea_t, unsigned short> InsertPair(InstAddr, (unsigned short) JumpTypeCode);
		pair<map<STARS_ea_t, unsigned short>::iterator, bool> InsertResult = this->ControlFlowMap.insert(InsertPair);
		assert(InsertResult.second);
		MapIter = InsertResult.first;
	}
	else if (MapIter->second != ((uint16_t)JumpTypeCode)) { // old entry found; different; update
		// We permit SHORT_CIRCUIT_LOOP_EXIT or SHORT_CIRCUIT_INVERTED_LOOP_EXIT to take precedence
		bool NewShortCircuitExit = (SHORT_CIRCUIT_LOOP_EXIT == JumpTypeCode) || (SHORT_CIRCUIT_INVERTED_LOOP_EXIT == JumpTypeCode);
		bool OldLoopExit = IsLoopExitFlow((ControlFlowType)MapIter->second);
		bool OldShortCircuit = IsShortCircuitFlow((ControlFlowType)MapIter->second);
		bool ExitCase = NewShortCircuitExit && (OldLoopExit || OldShortCircuit);
		if (ExitCase) {
			MapIter->second = (uint16_t) JumpTypeCode;
		}
		else {
			SMP_msg("WARNING: SPARK Ada jump type changing from %s to %s at %llx: rejected\n", CFTTypeStrings[MapIter->second],
				CFTTypeStrings[JumpTypeCode], (uint64_t)InstAddr);
#if 0  // refuse to override past value
			MapIter->second = (uint16_t) JumpTypeCode;
#endif
		}
	}
	return;
} // end of SMPFunction::SetControlFlowType()

// Insert SCCP value for global name; change old entry if already found.
STARSSCCPMapIter SMPFunction::InsertGlobalConstValue(int DefHashValue, struct STARS_SCCP_Const_Struct NewConstEntry) {
#if STARS_DEBUG_FUNC_SCCP_VERBOSE
	if (0x807a530 == this->FirstEA) {
		SMP_msg("DEBUG: SCCP: Inserting for DefHashValue %x Type: %d Value: %llx\n",
			DefHashValue, (int) NewConstEntry.ConstType, (uint64_t) NewConstEntry.ConstValue);
	}
#endif
	STARSSCCPMapIter MapIter = this->FindConstValue(DefHashValue);
	if (MapIter == this->GetLastConstValueIter()) { // no old entry; insert
		pair<int, struct STARS_SCCP_Const_Struct> InsertPair(DefHashValue, NewConstEntry);
		pair<STARSSCCPMapIter, bool> InsertResult = this->ConstantDefs.insert(InsertPair);
		assert(InsertResult.second);
		MapIter = InsertResult.first;
	}
	else { // old entry found; update
		MapIter->second = NewConstEntry;
	}
	return MapIter;
} // end of SMPFunction::InsertGlobalConstValue()


// Return RTLop if not stack opnd; return normalized RTLop otherwise.
STARSOpndTypePtr SMPFunction::GetNormalizedOperand(STARS_ea_t InstAddr, const STARSOpndTypePtr &RTLop) {
	STARSOpndTypePtr NormOp = nullptr;
#if SMP_DEBUG_DATAFLOW_VERBOSE
	bool DebugFlag = (0x8048463 == InstAddr);
	if (DebugFlag) {
		map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator StackMapIter;
		SMP_msg("DEBUG: NormalizedStackOpsMap size: %zd\n", this->NormalizedStackOpsMap.size());
		for (StackMapIter = this->NormalizedStackOpsMap.begin(); StackMapIter != this->NormalizedStackOpsMap.end(); ++ StackMapIter) {
			STARSOpndTypePtr OldOp = StackMapIter->first.first;
			STARS_ea_t InstAddr = StackMapIter->first.second;
			SMP_msg("DEBUG: NormalizedStackOps: ");
			PrintOperand(OldOp);
			SMP_msg(" addr: %llx\n", (uint64_t) InstAddr);
		}
		SMP_msg(" RTLop to search for: ");
		PrintOperand(RTLop);
	}
#endif
	if (MDIsStackAccessOpnd(RTLop, this->UsesFramePointer())) {
		pair<STARSOpndTypePtr, STARS_ea_t> OldDefn(RTLop, InstAddr);
		map<STARSDefinition, STARSOpndTypePtr, LessDefinition>::iterator FindIter = this->NormalizedStackOpsMap.find(OldDefn);
		assert(this->NormalizedStackOpsMap.end() != FindIter);
		NormOp = FindIter->second;
	}
	else {
		NormOp = RTLop;
	}
	return NormOp;
} // end of SMPFunction::GetNormalizedOperand()

// Get reg type from all call sites; compute if called for the first time.
SMPOperandType SMPFunction::GetIncomingRegType(STARS_regnum_t RegNum) {
	SMPOperandType InRegType = UNINIT;
	if (RegNum < (STARS_regnum_t)this->IncomingRegTypes.size()) {
		InRegType = this->IncomingRegTypes[RegNum];
	}
	return InRegType;
} // end of SMPFunction::GetIncomingRegType()

// return true if LiveInOp found in MarkerInst DEFs
bool SMPFunction::GetMarkerInstDefType(STARSOpndTypePtr &LiveInOp, SMPOperandType &MarkerDefType) {
	bool success = this->IsLiveIn(LiveInOp);
	if (success) {
		SMPInstr *MarkerInst = this->Instrs.front();
		success = (MarkerInst->IsMarkerInst());
		if (success) {
			STARSDefUseIter DefIter = MarkerInst->FindDef(LiveInOp);
			success = (DefIter != MarkerInst->GetLastDef());
			if (success) {
				MarkerDefType = DefIter->GetType();
			}
		}
	}
	return success;
} // end of SMPFunction::GetMarkerInstDefType()

// Eight methods to set values into the maps of global reg/stack/SSA to FG info.
//  For local names, see corresponding methods in SMPBasicBlock.
void SMPFunction::UpdateDefSignMiscInfo(int DefHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->GlobalDefFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = NewInfo;
		NewFGInfo.SizeInfo = 0;
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFGInfo);
		MapResult = this->GlobalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SignMiscInfo |= NewInfo;
	}

	return;
} // end of SMPFunction::UpdateDefSignMiscInfo()

void SMPFunction::UpdateStackDefSignMiscInfo(STARS_ea_t InstAddr, unsigned short NewInfo) {
	map<STARS_ea_t, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->StackDefFGInfo.find(InstAddr);
	assert(MapIter != this->StackDefFGInfo.end());
	// found; just OR in the new bits.
	MapIter->second.SignMiscInfo |= NewInfo;

	return;
} // end of SMPFunction::UpdateStackDefSignMiscInfo()

void SMPFunction::UpdateUseSignMiscInfo(int UseHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->GlobalUseFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = NewInfo;
		NewFGInfo.SizeInfo = 0;
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFGInfo);
		MapResult = this->GlobalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SignMiscInfo |= NewInfo;
	}

	return;
} // end of SMPFunction::UpdateUseSignMiscInfo()

void SMPFunction::UpdateStackUseSignMiscInfo(STARS_ea_t InstAddr, unsigned short NewInfo) {
	map<STARS_ea_t, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->StackUseFGInfo.find(InstAddr);
	assert(MapIter != this->StackUseFGInfo.end());
	// found; just OR in the new bits.
	MapIter->second.SignMiscInfo |= NewInfo;

	return;
} // end of SMPFunction::UpdateStackUseSignMiscInfo()

void SMPFunction::UpdateDefWidthTypeInfo(int DefHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->GlobalDefFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = 0;
		NewFGInfo.SizeInfo = NewInfo;
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFGInfo);
		MapResult = this->GlobalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SizeInfo |= NewInfo;
	}

	return;
} // end of SMPFunction::UpdateDefWidthTypeInfo()

void SMPFunction::UpdateUseWidthTypeInfo(int UseHashValue, unsigned short NewInfo) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->GlobalUseFGInfoBySSA.end()) {
		// Not found; insert first.
		struct FineGrainedInfo NewFGInfo;
		NewFGInfo.SignMiscInfo = 0;
		NewFGInfo.SizeInfo = NewInfo;
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFGInfo);
		MapResult = this->GlobalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just OR in the new bits.
		MapIter->second.SizeInfo |= NewInfo;
	}

	return;
} // end of SMPFunction::UpdateUseWidthTypeInfo()

void SMPFunction::UpdateDefFGInfo(int DefHashValue, struct FineGrainedInfo NewFG) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter == this->GlobalDefFGInfoBySSA.end()) {
		// Not found; insert it.
		pair<int, struct FineGrainedInfo> MapItem(DefHashValue, NewFG);
		MapResult = this->GlobalDefFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just put in the new bits.
		MapIter->second.SignMiscInfo |= NewFG.SignMiscInfo;
		MapIter->second.SizeInfo |= NewFG.SizeInfo;
	}

	return;
} // end of SMPFunction::UpdateDefFGInfo()

void SMPFunction::UpdateUseFGInfo(int UseHashValue, struct FineGrainedInfo NewFG) {
	map<int, struct FineGrainedInfo>::iterator MapIter;
	pair<map<int, struct FineGrainedInfo>::iterator, bool> MapResult;

	MapIter = this->GlobalUseFGInfoBySSA.find(UseHashValue);
	if (MapIter == this->GlobalUseFGInfoBySSA.end()) {
		// Not found; insert it.
		pair<int, struct FineGrainedInfo> MapItem(UseHashValue, NewFG);
		MapResult = this->GlobalUseFGInfoBySSA.insert(MapItem);
		assert(MapResult.second); // Was not previously found, insertion must work.
	}
	else { // found; just put in the new bits.
		MapIter->second.SignMiscInfo |= NewFG.SignMiscInfo;
		MapIter->second.SizeInfo |= NewFG.SizeInfo;
	}

	return;
} // end of SMPFunction::UpdateUseFGInfo()

// Reset the signedness bits to zero for DEF.
void SMPFunction::ClearDefSignedness(int DefHashValue) {
	map<int, struct FineGrainedInfo>::iterator MapIter;

	MapIter = this->GlobalDefFGInfoBySSA.find(DefHashValue);
	if (MapIter != this->GlobalDefFGInfoBySSA.end()) {
		MapIter->second.SignMiscInfo &= (~FG_MASK_SIGNEDNESS_BITS);
	}
	return;
} // end of SMPFunction::ClearDefSignedness()

// Erase a range of instructions from the Instrs list, usually corresponding
//  the the range of a basic block.
void SMPFunction::EraseInstRange(STARS_ea_t FirstAddr, STARS_ea_t LastAddr) {
	list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
	SMPInstr *CurrInst;
	STARS_ea_t InstAddr;

	bool FoundFirstAddr = false;
	bool FoundLastAddr = false;
	while (InstIter != this->Instrs.end()) {
		CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		if (InstAddr == FirstAddr) {
			FoundFirstAddr = true;
		}
		if ((FoundFirstAddr) && (!CurrInst->IsMarkerInst())) {
			InstIter = this->Instrs.erase(InstIter);
			if (InstAddr == LastAddr) {
				FoundLastAddr = true;
				break;
			}
		}
		else {
			++InstIter;
		}
	}
	assert(FoundFirstAddr);
	assert(FoundLastAddr);
	return;
} // end of SMPFunction::EraseInstRange()

// For instruction address UseAddr, compute the reaching defs for operand TempOp,
//  placing them into the TempReachingDefs list.
void SMPFunction::ComputeTempReachingDefs(const STARSOpndTypePtr &TempOp, STARS_ea_t UseAddr) {
	this->TempReachingDefs.clear();
	SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(UseAddr);
	assert(nullptr != CurrBlock);
	STARSDefinitionSet::iterator ReachesInIter;
	pair<set<STARS_ea_t, LessAddr>::iterator, bool> InsertResult;

	// Start with the matching members of the ReachesIn set for the current basic block.
	for (ReachesInIter = CurrBlock->GetFirstReachesIn(); ReachesInIter != CurrBlock->GetLastReachesIn(); ++ReachesInIter) {
		pair<STARSOpndTypePtr, STARS_ea_t> ReachesInDef = *ReachesInIter;
		if (IsEqOp(TempOp, ReachesInDef.first)) {
			InsertResult = this->TempReachingDefs.insert(ReachesInDef.second);
			assert(InsertResult.second);
		}
	}

	// Now, see if any def in the block hides the ReachesIn defs before we get to UseAddr.
	vector<SMPInstr *>::iterator InstIter;
	for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
		SMPInstr *CurrInst = *InstIter;
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->IsMarkerInst())
			continue;
		if (InstAddr >= UseAddr)
			break;
		set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(TempOp);
		if (DefIter != CurrInst->GetLastDef()) {
			// Found a def. All previous defs of TempOp are hidden from UseAddr by this def.
			this->TempReachingDefs.clear();
			InsertResult = this->TempReachingDefs.insert(InstAddr);
			assert(InsertResult.second);
		}
	}
	return;
} // end of SMPFunction::ComputeTempReachingDefs()

// Find all the saved stack deltas (if any) for the def addrs in the TempReachesDefs list for TempOp.
//  Put the entries matching TempOp into TempStackDeltaReachesList.
void SMPFunction::ComputeTempStackDeltaReachesList(const STARSOpndTypePtr &TempOp) {
	bool FoundOperand = false;
	set<pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> >, LessStackDeltaCopy>::iterator CopyIter;
	this->TempStackDeltaReachesList.clear();
	for (CopyIter = this->StackPtrCopySet.begin(); CopyIter != this->StackPtrCopySet.end(); ++CopyIter) {
		pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> > CopyEntry = *CopyIter;
		if (IsEqOp(TempOp, CopyEntry.first)) {
			set<STARS_ea_t, LessAddr>::iterator FindReachDefIter;
			FoundOperand = true; // help us save time later by exiting loop
			// Match address at which stack ptr copy was made to a reaching def address for TempOp.
			FindReachDefIter = this->TempReachingDefs.find(CopyEntry.second.first);
			if (FindReachDefIter != this->TempReachingDefs.end()) {
				// Found a StackPtrCopySet entry for TempOp, AND we found the DefAddr
				//  in the TempReachingDefs set. 
				this->TempStackDeltaReachesList.push_back(CopyEntry.second); // push back a pair<ea_t, STARS_sval_t>
			}
		}
		else if (FoundOperand) {
			// We have found the operand, but have now moved past it in the iteration of StackPtrCopySet.
			//  Save time by exiting the loop.
			break;
		}
	}
	return;
} // end of SMPFunction::ComputeTempStackDeltaReachesList()

// Find the largest stack delta in the TempStackDeltaReachesList.
// Return true if only one value was found in the list.
bool SMPFunction::FindReachingStackDelta(STARS_sval_t &StackDelta) {
	bool UniqueDelta = true;

	if (this->TempStackDeltaReachesList.empty()) {
		StackDelta = 0;
		return false;
	}
	else {
		StackDelta = this->TempStackDeltaReachesList.front().second;
	}

	list<pair<STARS_ea_t, STARS_sval_t> >::iterator DeltaIter;
	for (DeltaIter = this->TempStackDeltaReachesList.begin(); DeltaIter != this->TempStackDeltaReachesList.end(); ++DeltaIter) {
		STARS_sval_t NewDelta = DeltaIter->second;
		if (NewDelta != StackDelta) {
			UniqueDelta = false;
			if (NewDelta > StackDelta) {
				StackDelta = NewDelta;
			}
		}
	}
	return UniqueDelta;
} // end of SMPFunction::FindReachingStackDelta()

// Find any apparent stack adjustment after the call instruction at CallAddr,
//  confining our search to the basic block containing CallAddr.
STARS_sval_t SMPFunction::GetStackAdjustmentForCallee(STARS_ea_t CallAddr) {
	STARS_sval_t CalleeAdjustment = 0;

	SMPBasicBlock *CallBlock = this->GetBlockFromInstAddr(CallAddr);
	assert(nullptr != CallBlock);
	STARS_sval_t BlockAnalysisDelta = CallBlock->ComputeStackAdjustmentAfterCall(CallAddr);
	if (0 != BlockAnalysisDelta) {
		CalleeAdjustment = BlockAnalysisDelta;
		SMP_msg("INFO: Block analysis produced callee adjustment of %ld bytes after %llx\n", 
			(long) CalleeAdjustment, (uint64_t) CallAddr);
	}

	return CalleeAdjustment;
} // end of SMPFunction::GetStackAdjustmentForCallee()

// Get stack delta from a callee function that is unable to provide the info from
//  its own analyses (e.g. analyses failed or have not been performed yet, due to
//  a mutually recursive clique in the call graph). We have three approaches in
//  this case: Use a default value, consult IDA Pro's analyses, or see if we can
//  detect a stack adjustment after the call instruction, from which we could infer
//  the stack delta of the callee. We choose the latter approach, and find the smallest
//  adjustment among all call sites for the callee.
STARS_sval_t SMPFunction::GetStackDeltaForCallee(STARS_ea_t CallTargetAddr) {
	STARS_sval_t CalleeDelta = CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA;

	SMPFunction *CalleeFunc = this->GetProg()->FindFunction(CallTargetAddr);
	if (nullptr != CalleeFunc) {
		STARS_sval_t GlobalAdjustment = CalleeFunc->ComputeGlobalStackAdjustment(true);
		if (0 != GlobalAdjustment) {
			CalleeDelta -= GlobalAdjustment;
			SMP_msg("INFO: Global stack adjustment analysis produced callee delta of %ld bytes after %llx\n",
				(long) CalleeDelta, (uint64_t) CallTargetAddr);
		}
	}

	return CalleeDelta;
} // end of SMPFunction::GetStackDeltaForCallee()

// Compute a consistent (or smallest) stack adjustment seen program-wide after all calls to the current function.
//  Do not return a non-zero value unless more than one call site can be used as evidence.
STARS_sval_t SMPFunction::ComputeGlobalStackAdjustment(bool AllowSingleCaller) {
	bool FoundZeroAdjustment = false;
	STARS_sval_t GlobalAdjustment = 0;
	STARS_sval_t NegativeAdjustment = -10000; // record negative adjustments detected
	STARS_sval_t PositiveAdjustment = 10000; // record positive adjustments detected
	std::size_t NumCallSites = this->AllCallSites.size();

	if (!this->FuncInfo->HasReturnPoints()) { // never returns to caller for adjustments to be possible
		this->GlobalStackAdjustment = 0;
		this->StackAdjustmentComputed = true;
	}

	// Use cached value if already computed.
	if (this->StackAdjustmentComputed) {
		return this->GlobalStackAdjustment;
	}

	bool DebugFlag = (0x8089910 == this->GetFirstFuncAddr());
	if (DebugFlag)
		SMP_msg("DEBUG: NumCallSites = %zu\n", NumCallSites);

	if (AllowSingleCaller || (1 < NumCallSites)) { // if only one call site, it is dangerous to draw conclusions about seeming "adjustments."
		set<STARS_ea_t>::iterator CallSiteIter;
		for (CallSiteIter = this->AllCallSites.begin(); CallSiteIter != this->AllCallSites.end(); ++CallSiteIter) {
			STARS_ea_t CallSiteAddr = (*CallSiteIter);
			STARS_Function_t *CurrFunc = SMP_get_func(CallSiteAddr);
			assert(nullptr != CurrFunc);
			STARS_ea_t CallerFirstAddr = CurrFunc->get_startEA();
			SMPFunction *CallerFunc = this->GetProg()->FindFunction(CallerFirstAddr);
			assert(nullptr != CallerFunc);
			STARS_sval_t CurrentAdjustment = CallerFunc->GetStackAdjustmentForCallee(CallSiteAddr);
			if (DebugFlag)
				SMP_msg("DEBUG: CurrentAdjustment = %ld at CallSiteAddr %llx\n", (long) CurrentAdjustment,
				(uint64_t) CallSiteAddr);

			// See if CurrentAdjustment is a new, lowest positive value for GlobalAdjustment.
			if ((0 < CurrentAdjustment) && (CurrentAdjustment < PositiveAdjustment)) {
				PositiveAdjustment = CurrentAdjustment;
			}
			else if ((0 > CurrentAdjustment) && (CurrentAdjustment > NegativeAdjustment)) {
				NegativeAdjustment = CurrentAdjustment;
			}
			else if (0 == CurrentAdjustment) {
				FoundZeroAdjustment = true;
				break; // Any zero adjustment found invalidates non-zero inferences
			}
		}
	}

	// See if we consistently had positive or negative adjustments
	if (FoundZeroAdjustment) {
		GlobalAdjustment = 0; // cannot be a clear non-zero indication if we found any zeroes
	}
	else if (PositiveAdjustment < 10000) { // found at least one positive adjustment
		if (NegativeAdjustment > -10000) { // found at least one negative adjustment; bad
			GlobalAdjustment = 0; // inconsistent; reset to zero
		}
		else {
			GlobalAdjustment = PositiveAdjustment;
		}
	}
	else if (NegativeAdjustment > -10000) { // found negative but no positive adjustments
		GlobalAdjustment = NegativeAdjustment;
	}
	else { // did not find negative or positive adjustments
		GlobalAdjustment = 0;
	}

	this->StackAdjustmentComputed = true; // signal caching of the value for future speed
	this->GlobalStackAdjustment = GlobalAdjustment; // cache the value
	return GlobalAdjustment;
} // end of SMPFunction::ComputeGlobalStackAdjustment()

// Use IDA Pro stack pointer deltas instead of doing our own analysis.
bool SMPFunction::UseIDAStackPointerDeltas(void) {
	list<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	bool IDATraceFlag = false;
#if SMP_COMPARE_IDA_STARS_STACK_POINTER_DELTAS
	IDATraceFlag = (0 == strcmp("do_length", this->GetFuncName()));
#endif

	InstIter = this->Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	++InstIter; // skip marker pseudo-instruction
#endif
	while (InstIter != this->Instrs.end()) {
		CurrInst = *InstIter;
		STARS_sval_t IDAProDelta = SMP_get_spd(this->GetFuncInfo(), CurrInst->GetAddr());
		CurrInst->SetStackPtrOffset(IDAProDelta);
		++InstIter;
		if (IDATraceFlag) {
			SMP_msg("INFO: IDA Pro stack delta trace: %ld at %llx\n", (long) IDAProDelta, (uint64_t) CurrInst->GetAddr());
		}
	}
	return true;
} // end of SMPFunction::UseIDAStackPointerDeltas()

// Analyze changes to the stack pointer over all instructions.
bool SMPFunction::AnalyzeStackPointerDeltas(void) {
	list<pair<SMPBasicBlock *, STARS_sval_t> > WorkList;
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	STARS_sval_t CurrentDelta = 0;
	STARS_sval_t DeltaIncrement = 0; // change when reprocessing a block in alloca()-calling function
	bool ConsistentNetDelta = true; // Net change to stack pointer is consistent at all RETURN locations
	bool ConflictingValuesSeen = false; // At least one block was entered with multiple deltas
	bool StackPointerRestoreSeen = false; // Stack pointer restored; must become true if ConflictingValuesSeen
	bool ReturnSeen = false;
	bool IDAProSucceeded = this->AnalyzedSP;
	bool DebugFlag = false;
	bool TraceFlag = false;
	bool IDATraceFlag = false;

#if SMP_COMPARE_IDA_STARS_STACK_POINTER_DELTAS
	DebugFlag = (0 == strcmp("_dl_profile_fixup", this->GetFuncName()));
	TraceFlag = (0 == strcmp("_dl_profile_fixup", this->GetFuncName()));
	bool DebugFuncFound = (0x43ac70 == this->GetFirstFuncAddr());
	DebugFlag = DebugFlag || DebugFuncFound;
	TraceFlag = TraceFlag || DebugFuncFound;
#endif

	if (!this->HasGoodRTLs()) {
		SMP_msg("INFO: Using IDA Pro stack pointer deltas for BADRTLS function %s .\n", this->GetFuncName());
		// (void) this->UseIDAStackPointerDeltas();
		this->AnalyzedSP = false;
		return false; // leave it unsolved
	}

#if 0
	// Temporarily pull the functions that call alloca out of the stack pointer delta computations, so
	//  that we can focus on solving other problems.
	if (this->CallsAlloca || this->HasPushAfterFrameAlloc()) {
		if (!this->AnalyzedSP) {
			(void) this->UseIDAStackPointerDeltas();
			return false; // leave it unsolved
		}
		else {
			SMP_msg("INFO: Using IDA Pro stack pointer deltas for alloca-calling function %s .\n", this->GetFuncName());
			return this->UseIDAStackPointerDeltas();
		}
	}
#endif

	// NOTE: The below comment assumes that we cannot have SSA form yet, because we need stack pointer
	//  deltas in order to incorporate stack locations into SSA form. Dealing with this phase ordering
	//  problem is the motivation for the reaching definitions analysis discussed below.

	// In order to precisely track stack deltas, we need to deal with instruction sequences that save the stack pointer
	//  and then restore it later. This requires a reaching definitions data flow analysis that includes, at a minimum,
	//  all stack definitions (normalized by stack delta, so that we do not confuse [esp+20] and [esp+20] where the values
	//  of esp are not the same). We also need to keep track of stack pointer saves in both registers and in stack locations.
	// In order for the information about saved stack pointer copies to be available as soon as we need them in the stack
	//  delta analysis, we have to perform both stack delta analysis and reaching definitions analysis at the same time. Luckily,
	//  both analyses are well suited to being performed as forward analyses starting from the entry basic block.
	//
	// Data structures for the reaching definitions analysis include a ReachesIn and a ReachesOut set for each basic block, a
	//  VarKill set for each block, and a DownExposedDefs set for each block. The VarKill set is shared with the later Live Variable
	//  Analysis (LVA), so we compute the VarKill and the UpExposed sets (UpExposed is only used by LVA) on the first pass through
	//  each block. The VarKill and all other LVA sets are sets of operands. The ReachesIn, ReachesOut, and DownExposedDefs sets
	//  are sets of definitions, where a definition is a pair<operand, instruction address>. The StackPtrCopySet is a triple of
	//   <operand, instruction address, stack delta>, arranged as a pair of pairs <operand, <addr, delta> >
	//
	// Algorithm: We maintain a WorkList of pairs <basic block pointer, incoming stack delta to that block>
	//
	// All sets are empty at the beginning.
	// Add the entry basic block to the WorkList, with IncomingDelta of zero.
	// while (WorkList is not empty) do
	//    de-queue first block from WorkList, obtain IncomingDelta
	//    Compute ReachesIn as the union of the ReachesOut of all predecesssor blocks
	//    if (block has not already been processed) then
	//       mark block as processed
	//       for each inst in block (forward iteration) do
	//           for each USE in inst do
	//              if USE operand not in VarKill set for block then
	//                  add USE operand to UpExposed set for block
	//              endif
	//              if USE operand is a stack pointer value AND it will produce DEF that is a stack pointer value then
	//                  if DEF is stack pointer register then  { a stack pointer value that was saved is being restored }
	//                      retrieve new stack pointer delta from saved value in StackPtrCopySet, looking it up in
	//                          that set using the reaching definitions for the USE operand. If inconsistent  ******
	//                  else { stack pointer value is being saved somewhere besides the stack pointer register }
	//                      add stack delta to StackPtrCopySet for DEF that is receiving it in current inst
	//                  endif
	//               endif
	//            endfor { each USE }
	//            for each DEF in inst do
	//               if register or stack operand then 
	//                  add to VarKill set
	//                  update DownExposedDefs set (insert, or replace current def for this operand)
	//               endif
	//            endfor { each DEF }
	//            Store IncomingDelta for current instruction
	//            Get change in delta for current instruction
	//            Add current change to IncomingDelta
	//       endfor { each inst }
	//       At end of block, make ReachesOut set be (ReachesIn minus VarKill) union DownExposedDefs
	//       For each successor block, add pairs <block pointer, IncomingDelta> to end of WorkList
	//    else { block has already been processed at least once}
	//       if IncomingDelta from WorkList is inconsistent with old IncomingDelta then
	//          if function calls alloca() then
	//             if new IncomingDelta makes stack frame look larger than old IncomingDelta then
	//                ignore new IncomingDelta and just process reaching definitions sets below
	//             else
	//                use new IncomingDelta and re-process deltas in this block to converge to
	//                 smallest stack frame, which means we are basically ignoring alloca()'s as much as possible.
	//             endif
	//          else
	//             Set AnalyzedSP to false, emit error message, clear WorkList and bail out of this function.
	//          endif
	//       endif { inconsistent IncomingDelta values }
	//       Recompute ReachesIn as union of ReachesOut of all predecessor blocks
	//       if ReachesIn set changed then
	//          recompute ReachesOut without examining instructions unless alloca() case requires iterating through instructions
	//       endif
	//       if any change in deltas or reaching definitions sets, then add block to end of WorkList along with all successor blocks.
	//    endif
	// end while

	// Mark all blocks as unprocessed
	this->ResetProcessedBlocks();

	this->AnalyzedSP = true;

	// Put the entry block on the work list.
	assert(!(this->Blocks.empty()));
	pair<SMPBasicBlock *, STARS_sval_t> WorkingPair(this->Blocks.front(), CurrentDelta);
	WorkList.push_back(WorkingPair);

	// While blocks exist on the work list
	//  if block already processed, confirm that we are re-entering
	//    the block with the same stack pointer delta as previously,
	//    and pop it off the work list
	//    otherwise declare the stack pointer to be un-analyzeable;
	//  else
	//     iterate through all instructions in the block, analyzing
	//     the stack pointer delta of each inst and accumulating current delta
	//     At the end of the block, put the successor blocks on the work list.
	// For both cases, maintain and update reaching definitions sets, and the
	//  UpExposed and VarKill sets that are used by LVA as well as reaching defs analysis.
	bool ReprocessingAllocaBlocks = false;
	bool ReachesInChanged;
	bool ReachesOutChanged = false;
	do {
		SMPBasicBlock *CurrBlock = WorkList.front().first;
		STARS_sval_t IncomingDelta = WorkList.front().second;

		if (TraceFlag) {
			SMP_msg("TRACE: Starting block %d with IncomingDelta of %lld\n", CurrBlock->GetNumber(), (int64_t) IncomingDelta);
		}
#if SMP_COMPARE_IDA_STARS_STACK_POINTER_DELTAS
		if (DebugFlag && IDAProSucceeded) {
			STARS_ea_t InstAddr = CurrBlock->GetFirstAddr();
			if (!STARS_IsSSAMarkerPseudoID(InstAddr)) {
				STARS_sval_t IDAProDelta = SMP_get_spd(this->GetFuncInfo(), InstAddr);
				if (IDAProDelta != IncomingDelta) {
					intmax_t IDADelta = (intmax_t) IDAProDelta, OurDelta = (intmax_t) IncomingDelta;
					if (sizeof(STARS_sval_t) == 8) {
						SMP_msg("ERROR: At inst %p block %d entry IDA Pro has stack pointer delta of %jd and we compute %jd BlockProcessed: %d ",
							InstAddr, CurrBlock->GetNumber(), IDADelta, OurDelta, CurrBlock->IsProcessed());
						this->DumpFuncNameAndAddr();
					}
					else {
						assert(sizeof(STARS_sval_t) == 4);
						SMP_msg("ERROR: At inst %p block %d entry IDA Pro has stack pointer delta of %d and we compute %d BlockProcessed: %d ",
							InstAddr, CurrBlock->GetNumber(), IDAProDelta, IncomingDelta, CurrBlock->IsProcessed());
						this->DumpFuncNameAndAddr();
					}
				}
			}
		}
#endif

		if (0 < IncomingDelta) {
			SMP_msg("ERROR: Stack delta of %ld implies stack underflow ", (long) IncomingDelta);
			this->DumpFuncNameAndAddr();
			this->AnalyzedSP = false;
			WorkList.clear();
			break;
		}

		if (CurrBlock->IsProcessed()) { // already processed
			ReachesOutChanged = false;
#if 0
			ReachesInChanged = CurrBlock->ComputeReachesInSet();
			if (ReachesInChanged) {
				ReachesOutChanged = CurrBlock->ComputeReachesOutSet();
			}
#else
			if (CurrBlock->IsReachesOutStale()) {
				ReachesOutChanged = CurrBlock->ComputeReachesOutSet();
			}
#endif
			if (ReachesOutChanged) {
				// Push the successor blocks onto the work list
				STARS_sval_t SuccIncomingDelta = CurrBlock->GetOutgoingStackDelta();
				list<SMPBasicBlock *>::iterator SuccIter;
				for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
					pair<SMPBasicBlock *, STARS_sval_t> SuccPair (*SuccIter, SuccIncomingDelta);
					WorkList.push_back(SuccPair);
					if (TraceFlag) {
						int SuccBlockNum = (*SuccIter)->GetNumber();
						SMP_msg("TRACE: Pushing SuccBlock %d from CurrBlock %d with IncomingDelta %d\n",
							SuccBlockNum, CurrBlock->GetNumber(), (int) SuccIncomingDelta);
					}
				}
			}
			InstIter = CurrBlock->GetFirstInst();
			STARS_sval_t PrevIncomingDelta = (*InstIter)->GetStackPtrOffset();
			if (IncomingDelta == PrevIncomingDelta) {
				// No error, already processed.
				WorkList.pop_front(); // discard already processed block.
			}
#if 1
			else if (this->CallsAlloca || this->HasPushAfterFrameAlloc()) 
#else
			else 
#endif
			{
				ConflictingValuesSeen = true;
				// Calls to alloca() become additional stack allocations, which can produce
				//  multiple possible stack deltas for an instruction if different paths
				//  to the instruction do not hit the same alloca() calls, so it is not
				//  an error to have conflicting deltas in the functions that call alloca().
				//  We want to converge to the smallest magnitude deltas, which are the greatest
				//  values because the deltas are negative. This is the opposite of IDA Pro, which
				//  seems to use the largest magnitude stack deltas it has seen.
				if (PrevIncomingDelta >= IncomingDelta) {
					// Old incoming delta should be retained.
					WorkList.pop_front(); // discard already processed block.
				}
				else {
					CurrBlock->SetProcessed(false);
					ReprocessingAllocaBlocks = true;
					DeltaIncrement = IncomingDelta - PrevIncomingDelta;
					if (0 != (DeltaIncrement % 32)) { // 32 bytes is our alloca fudge increment
						SMP_msg("WARNING: DeltaIncrement of %ld not alloca multiple for block %d at %llx\n",
							(long)DeltaIncrement, CurrBlock->GetNumber(), (uint64_t) CurrBlock->GetFirstAddr());
#if SMP_VERBOSE_DEBUG_FUNC
						if ((8 >= DeltaIncrement) && (0 <= DeltaIncrement))
							this->Dump(false);
#endif
					}
					continue;  // Make the loop come around and process this block again, using
							   //  the new incoming delta. Because we do this only when it decreases
							   //  the stack size as seen by this block, no infinite loop is possible.
				}
			}
#if 1
			else {
				this->AnalyzedSP = false;
				SMP_msg("ERROR: Stack delta: PrevIncoming is %ld NewIncoming is %ld at %llx block %d ",
					(long) PrevIncomingDelta, (long) IncomingDelta, (uint64_t) (*InstIter)->GetAddr(), CurrBlock->GetNumber());
				this->DumpFuncNameAndAddr();
				WorkList.clear();
			}
#endif
		}
		else { // not already processed
			// ReprocessingAllocaBlocks => Reaching definitions sets have already been computed; just need to do stack delta analysis
			ReachesOutChanged = false;
#if 0
			ReachesInChanged = CurrBlock->ComputeReachesInSet();
			if (ReachesInChanged && ReprocessingAllocaBlocks) {
				// Because block is not truly being processed for the first time, the ReachesOut set can be
				//  recomputed without processing instructions, as the DEDefs set and VarKill set will never
				//  change after the first pass through the block.
				ReachesOutChanged = CurrBlock->ComputeReachesOutSet();
			}

			if (CurrBlock->IsReachesOutStale()) {
				ReachesOutChanged = CurrBlock->ComputeReachesOutSet();
			}
#endif
			CurrBlock->SetProcessed(true);
			WorkList.pop_front();
			for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				CurrInst = (*InstIter);
				if (CurrInst->IsMarkerInst()) {
					continue; // skip marker instruction
				}
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				if (InstAddr == this->GetFirstFrameAllocInstAddr()) {
					// Record the reset point for frame deallocations
					this->PreAllocStackDelta = IncomingDelta;
				}

				CurrInst->SetStackPtrOffset(IncomingDelta);

				// Search for tail calls, defined strictly as having an incoming stack delta of zero and
				//  being jumps to far chunks not included in the current function.
				if (0 == IncomingDelta) {
					if (CurrInst->IsBranchToOtherFunc()) {
						CurrInst->SetTailCall();
#if 0
						SMP_msg("INFO: Found tail call at %llx from %s: %s\n", (uint64_t) InstAddr, this->GetFuncName(),
								CurrInst->GetDisasm());
#endif
					}
					else if (INDIR_JUMP == CurrInst->GetDataFlowType()) {
						// We have an indirect jump with an incoming delta of zero. This
						//  is probably an indirect tail call, unless the target address has
						//  been resolved to addresses in the current function (e.g. switch table).
						if (this->HasUnresolvedIndirectJumps() && (STARS_BADADDR == CurrInst->GetCallTarget())) {
							CurrInst->SetTailCall();
#if 0
							SMP_msg("INFO: Found indirect tail call at %llx from %s: %s\n", (uint64_t) InstAddr, this->GetFuncName(),
									CurrInst->GetDisasm());
#endif
						}
					}
				}

#if SMP_COMPARE_IDA_STARS_STACK_POINTER_DELTAS
				if (DebugFlag && IDAProSucceeded && !(this->CallsAlloca || this->HasPushAfterFrameAlloc())) {
					STARS_sval_t IDAProDelta = SMP_get_spd(this->GetFuncInfo(), InstAddr);
					if ((IDAProDelta != IncomingDelta) && (!CurrInst->MDIsHaltInstr())) {
						// IDA Pro special-cases the HALT instruction to make it appear that the
						//  incoming stack delta is zero. We do no such special case delta adjudstment,
						//  so we suppress error messages, as our delta will be non-zero.
						SMP_msg("ERROR: At %llx IDA Pro has stack pointer delta of %lld and we compute %lld ", 
							(uint64_t) InstAddr, (long long) IDAProDelta, (long long) IncomingDelta);
						this->DumpFuncNameAndAddr();
					}
				}
				if (TraceFlag) {
					SMP_msg("INFO: Stack delta trace: %lld at %llx\n", (long long) IncomingDelta, (uint64_t) InstAddr);
				}
#endif

				// As soon as the stack ptr offset has been set for the current instruction, we can normalize
				//  all of its stack DEFs and USEs.
				bool StackOpsChanged = CurrInst->MDNormalizeStackOps(this->UsesFramePointer(), this->GetFramePtrStackDelta(), ReprocessingAllocaBlocks, DeltaIncrement);

				// Dataflow equation for upward exposed variables: If a variable has not been
				//  killed yet in this block, starting from the top of the block, and it is used
				//  in the current instruction, then it is upwardly exposed.
				set<DefOrUse, LessDefUse>::iterator CurrUse;
				if (!ReprocessingAllocaBlocks) { // Only compute on first pass through block
					for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
						STARSOpndTypePtr UseOp = CurrUse->GetOp();
						// CanonicalizeOpnd(UseOp);
						if (MDIsDataFlowOpnd(UseOp, this->UsesFramePointer())) {
							// We have a register or stack operand. If stack operand, it is normalized, i.e. EBP-4 might be ESP-8,
							//  where the ESP-8 refers to the value of ESP upon entry to the function, not its current value.
							//  This normalization makes each stack location uniquely named (no aliases at different code locations due
							//  to different values of ESP at different code locations).
							// We only track certain kinds of operands in our data flow analyses.
							// Only add non-immediate operands that are not already killed in this block.
							//  o_near and o_far operands are code addresses in immediate form, e.g.
							//  call _printf might be call 0x8048040, with o_near = 0x8048040.
							if (!(CurrBlock->MDAlreadyKilled(UseOp))) {
								CurrBlock->AddUpExposed(UseOp);
							}
						}
					}
				}

				// Find stack pointer saves and restores.
				bool StackPtrSaved;
				STARS_sval_t SavedDelta;
				STARSOpndTypePtr CopyOperand = nullptr;
				bool SavedDeltaHasNewValue = false;
				bool ErrorFlag = false;
				if (CurrInst->MDIsStackPtrSaveOrRestore(this->UsesFramePointer(), this->GetFramePtrStackDelta(), StackPtrSaved, SavedDelta, CopyOperand, ErrorFlag)) {
					// NOTE: If CopyOperand is a stack location, it is normalized.
					if (StackPtrSaved) {
						// Insert new entry into the StackPtrCopySet. For the ReprocessingAllocaBlocks case, this might be
						//  just a tricky update of the delta for an existing item in the set.
						bool DeltaInserted = this->AddToStackPtrCopySet(CopyOperand, InstAddr, SavedDelta);
						if (TraceFlag) {
							SMP_msg("INFO: Stack delta saved: %ld at %llx\n", (long) SavedDelta, (uint64_t) InstAddr);
							PrintOperand(CopyOperand);
							SMP_msg("\n");
						}
					}
					else { // stack pointer was restored from saved value
						StackPointerRestoreSeen = true;
						SavedDeltaHasNewValue = true; // no need to compute effect of restore instruction later
						if (ReprocessingAllocaBlocks) {
							// Now that the stack pointer has been restored, the effect of the alloca() should
							//  be undone. We no longer need to adjust delta values for the rest of the block.
							DeltaIncrement = 0;
						}
					}
				} // end if (CurrInst->MDIsStackPtrSaveOrRestore())
				else if (ErrorFlag) {
					this->AnalyzedSP = false;
					WorkList.clear();
					SMP_msg("ERROR: ErrorFlag=true from MDIsStackPtrSaveOrRestore() at %llx ",	(uint64_t) InstAddr);
					this->DumpFuncNameAndAddr();
					break;
				}
				else if (CurrInst->MDIsLeaveInstr()) {
					// LEAVE is a restoration of a stack pointer, not processed by CurrInst->MDIsStackPtrSaveOrRestore()
					StackPointerRestoreSeen = true; 
				}

				// Update VarKill and DownExposedDefs sets for DEFs in current instruction.
				// Dataflow equation for killed variables: If a variable is defined in any
				//  instruction in the block, it is killed by this block (i.e. prior definitions
				//  of that variable will not make it through the block).
				if (!ReprocessingAllocaBlocks) { // Only compute on first pass through block
					set<DefOrUse, LessDefUse>::iterator CurrDef;
					for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
						STARSOpndTypePtr DefOp = CurrDef->GetOp();
						if (MDIsDataFlowOpnd(DefOp, this->UsesFramePointer())) {
							// We have a register or stack operand. If stack operand, it is normalized, i.e. EBP-4 might be ESP-8,
							//  where the ESP-8 refers to the value of ESP upon entry to the function, not its current value.
							//  This normalization makes each stack location uniquely named (no aliases at different code locations due
							//  to different values of ESP at different code locations).
							CurrBlock->AddVarKill(DefOp);
							CurrBlock->UpdateDownExposedDefs(DefOp, InstAddr);
						}
					}
				}

				if (SavedDeltaHasNewValue) {
					IncomingDelta = SavedDelta; // from restore instruction
				}
				else {
					CurrentDelta = CurrInst->AnalyzeStackPointerDelta(IncomingDelta, this->GetFramePtrStackDelta());
					if (SMP_STACK_POINTER_BITWISE_AND_CODE == ((STARS_uval_t) CurrentDelta)) {
						// For now, we ignore instructions that AND a constant into the stack pointer.
						CurrentDelta = 0;
						SMP_msg("WARNING: Stack pointer bitwise AND ignored at %llx\n",
							(uint64_t) CurrInst->GetAddr());
					}
					else if (SMP_STACK_DELTA_ERROR_CODE == ((STARS_uval_t) CurrentDelta)) {
						this->AnalyzedSP = false;
						SMP_msg("ERROR: Stack delta unanalyzeable at %llx ", (uint64_t) InstAddr);
						this->DumpFuncNameAndAddr();
						WorkList.clear();
						break;
					}
					if (TraceFlag) {
						SMP_msg("TRACE: CurrentDelta is %lld at %llx\n", (long long) CurrentDelta, (uint64_t) InstAddr);
					}
					SMPitype FlowType = CurrInst->GetDataFlowType();
					IncomingDelta += CurrentDelta;
					if ((RETURN == FlowType) && (!CurrInst->IsCondTailCall()) && (!CurrInst->IsTailCall())) {
						// We hope to see a consistent outgoing delta from all RETURN points.
						//  We special-case the conditional jump used as tail call, because it must be followed
						//  by a real return instruction later. If the jump is taken, it acts as a return, but
						//  it has not yet popped the stack.
						// Also, a regular tail call always has the stack delta at zero and does not match
						//  the stack delta of actual return instructions elsewhere in the function.
						if (ReturnSeen) { // This is not the first RETURN seen.
							if (IncomingDelta != this->NetStackDelta) { // Inconsistent
								SMP_msg("ERROR: Inconsistent stack deltas at return instruction at %p : Previous: %lld Current: %lld ",
									CurrInst->GetAddr(), (long long) this->NetStackDelta, (long long) IncomingDelta);
								this->DumpFuncNameAndAddr();
								ConsistentNetDelta = false;
								this->AnalyzedSP = false;
								WorkList.clear();
								break;
							}
						}
						else { // First RETURN statement seen.
							ReturnSeen = true;
							this->NetStackDelta = IncomingDelta;
							SMP_msg("INFO: First stack delta seen at return instruction at %p : Current: %lld\n",
								CurrInst->GetAddr(), (long long) IncomingDelta);
#if SMP_AUDIT_STACK_POINTER_DELTAS
							if (CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA != IncomingDelta) {
								SMP_msg("WARNING: Stack delta not %d after return instruction at %llx\n", 
									CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA, (uint64_t) CurrInst->GetAddr());
							}
#endif
						}
						// If we permitted inconsistent stack deltas previously, then the stack pointer has to
						//  have been restored, e.g. if we allocate a frame with sub esp,32 and then we later
						//  have paths that pass through an alloca() call, a push, etc., then the alloca() or
						//  push will not be undone by add esp,32. It must be undone by something like mov esp,ebp.
						if (ConflictingValuesSeen && !StackPointerRestoreSeen) {
							SMP_msg("ERROR: Inconsistent stack deltas seen, no stack pointer restore before return instruction at %llx ", (uint64_t) CurrInst->GetAddr());
							this->DumpFuncNameAndAddr();
							this->AnalyzedSP = false;
							WorkList.clear();
							break;
						}
					}
				} // end if (SavedDeltaHasNewValue) ... else ...
			} // end for each instruction in WorkList block
			if (CurrBlock->IsReachesOutStale()) {
				ReachesOutChanged = CurrBlock->ComputeReachesOutSet();
			}
			// Push the successor blocks onto the work list if anything changed
			if (this->AnalyzedSP) { // if we do not have an error already
				CurrBlock->SetOutgoingStackDelta(IncomingDelta); // record incoming delta for all successors
				if (ReachesOutChanged || (!ReprocessingAllocaBlocks)) { // if anything changed (deltas or reaching defs ReachOut set)
					list<SMPBasicBlock *>::iterator SuccIter;
					if (DebugFlag && (0 == IncomingDelta) && (0 < this->GetLocalVarsSize())) {
						SMP_msg("ERROR: Pushing WorkList items with IncomingDelta of zero. Dumping Block:\n");
						CurrBlock->Dump(false);
					}
					for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
						pair<SMPBasicBlock *, STARS_sval_t> SuccPair (*SuccIter, IncomingDelta);
						WorkList.push_back(SuccPair);
						if (TraceFlag) {
							int SuccBlockNum = (*SuccIter)->GetNumber();
							SMP_msg("TRACE: Pushing SuccBlock %d from CurrBlock %d with IncomingDelta %d\n",
								SuccBlockNum, CurrBlock->GetNumber(), (int) IncomingDelta);
						}
					}
				}
			}
		} // end if block already processed ... else ...
		ReprocessingAllocaBlocks = false; // reset to default before processing next worklist element
	} while (!WorkList.empty());

	this->STARSStackPtrAnalysisPerformed = true;
	if (this->AnalyzedSP) {
// Don't let net effect of a function on the stack pointer be greater than this value.
//  Anything greater indicates some sort of alloca() analyzeability failure, or similar failure.
#define STARS_NET_STACK_DELTA_ERROR_THRESHOLD 16
		if (CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA != ((size_t) this->NetStackDelta)) {
			SMP_msg("WARNING: Non-default stack ptr delta %ld for function: %s\n", (long) this->NetStackDelta, this->GetFuncName());
			// See if all of our callers agree that we have a non-standard net effect on the stack
			//  pointer, which is indicated by all of them making a stack pointer adjustment after
			//  we return to them.
			STARS_sval_t StackAdjustment = this->ComputeGlobalStackAdjustment(true);
			STARS_sval_t FinalNetDelta = this->NetStackDelta + StackAdjustment;
			if (CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA != ((size_t)FinalNetDelta)) {
				if (STARS_NET_STACK_DELTA_ERROR_THRESHOLD < this->NetStackDelta) {
					SMP_msg("ERROR: Func cannot have NetStackDelta of %lld \n", (int64_t) this->NetStackDelta);
					this->AnalyzedSP = false;
				}
				else if (this->CallsAlloca && ((STARS_sval_t)CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA < FinalNetDelta)) {
					if (this->GetNumCallers() != 0) { // should have good stack adjustment
						SMP_msg("ERROR: Func cannot have NetStackDelta of %lld and FinalNetDelta of %lld for func that calls alloca\n",
							(int64_t) this->NetStackDelta, (int64_t)FinalNetDelta);
						this->AnalyzedSP = false;
					}
				}
			}
		}
		if (this->StackAdjustmentComputed 
			&& (this->GlobalStackAdjustment != (((STARS_sval_t) CALLING_CONVENTION_DEFAULT_FUNCTION_STACK_DELTA) - this->NetStackDelta))) {
			// Oops. When program graph cycles caused us to try to compute the GlobalStackAdjustment as our best guess
			//  for this function's effect on the stack delta, we told our callers that these three values would cancel out.
			//  They do not. Our callers have now been using a bad stack delta for their call instructions. Too late for
			//  anything but a diagnostic message.
				SMP_msg("ERROR: Earlier GlobalStackAdjustment computation %ld does not agree with current NetStackDelta result ",
					(long) this->GlobalStackAdjustment);
				this->DumpFuncNameAndAddr();
		}
	}


	if (IDAProSucceeded) {
		if (!this->AnalyzedSP) {
			SMP_msg("ERROR: Stack Ptr Delta Analysis succeeded in IDA, failed in STARS ");
			this->DumpFuncNameAndAddr();
		}
	}
	else {
		if (this->AnalyzedSP) {
			SMP_msg("SUCCESS: Stack Ptr Delta Analysis failed in IDA, succeeded in STARS ");
			this->DumpFuncNameAndAddr();
		}
	}
	if (!this->AnalyzedSP) {
		;
		// (void) this->UseIDAStackPointerDeltas();
	}
	else {
		// Success, so try to find saved/restored register pairs so that we do not
		//  conservatively say that the function kills registers that it in fact preserves.
		this->DetectThunkFunction();
		this->FindPreservedRegs();
		this->ComputeGlobalSets();
	}

	// Cannot keep the reaching defs around on huge benchmarks, or we run out of memory.
	//  Once we have SSA form, we can obtain reaching defs info on the fly if we want it.
	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		(*BlockIter)->FreeReachingDefsMemory();
	}

	return this->AnalyzedSP;
} // end of SMPFunction::AnalyzeStackPointerDeltas()

// Insert the arguments into the StackPtrCopySet; or, if a matching entry already exists
//  with a StackDelta of greater magnitude than StackDelta, update just the StackDelta.
// Return true if StackDelta was inserted, false if it was used to update an old entry.
bool SMPFunction::AddToStackPtrCopySet(STARSOpndTypePtr CopyOp, STARS_ea_t InstAddr, STARS_sval_t StackDelta) {
	bool NewInsertion;
	pair<STARS_ea_t, STARS_sval_t> InsertStackDefn(InstAddr, StackDelta);
	pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> > InsertStackDefnOp(CopyOp, InsertStackDefn);
	set<pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> >, LessStackDeltaCopy>::iterator FindIter;
	pair<set<pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> >, LessStackDeltaCopy>::iterator, bool> InsertResult;

	FindIter = this->StackPtrCopySet.find(InsertStackDefnOp);
	if (FindIter == this->StackPtrCopySet.end()) {
		// Not already present; insert new triple.
		NewInsertion = true;
		InsertResult = this->StackPtrCopySet.insert(InsertStackDefnOp);
		assert(InsertResult.second);
	}
	else {
		// Already there; see if delta needs to be updated.
		NewInsertion = false;
		pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> > OldStackDefnOp(*FindIter);
		// Favor a smaller stack frame for the alloca-calling functions, e.g. favor -24 over -32 as a delta.
		if (StackDelta > OldStackDefnOp.second.second) {
			// Replace the old entry with a new one.
			this->StackPtrCopySet.erase(FindIter);
			InsertResult = this->StackPtrCopySet.insert(InsertStackDefnOp);
			assert(InsertResult.second);
		}
	}

	return NewInsertion;
} // end of SMPFunction::AddToStackPtrCopySet()

// Determined which regs are not killed by func or its callees; set bits in PreservedRegsBitmap
void SMPFunction::FindPreservedRegs(void) {
	// First, find candidates from the pushes in the entry block.
	bool EntryBlockProcessed = false;
	bool ReturnBlockSeen = false;
	SMPBasicBlock *CurrBlock;
	map<uint32_t, STARS_sval_t> PushedRegsList, PoppedRegsList;
	list<SMPBasicBlock *>::iterator BlockIter;
	vector<SMPInstr *>::iterator InstIter;
	bool success = true;
	for (BlockIter = this->Blocks.begin(); success && (BlockIter != this->Blocks.end()); ++BlockIter) {
		CurrBlock = (*BlockIter);
		if (!EntryBlockProcessed) { // Must be first block
			EntryBlockProcessed = true;
			for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (CurrInst->MDIsPushInstr() || CurrInst->MDIsEnterInstr()) {
					// Get list of regs saved, and their relative stack offsets, from the RTLs.
					//  Could be a PushAllRegs instruction, so there might be more than one.
					success = CurrInst->GetPushedRegsList(PushedRegsList);
					if (!success)
						break;
				}
			}
		}
		if (EntryBlockProcessed && CurrBlock->HasReturn()) {
			// See if we have pops that match the pushes.
			size_t MatchCount = 0;
			for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (CurrInst->MDIsPopInstr() || CurrInst->MDIsLeaveInstr()) {
					// Get list of regs restored, and their relative stack offsets, from the RTLs.
					//  Could be a PopAllRegs instruction, so there might be more than one.
					success = CurrInst->GetPoppedRegsList(!ReturnBlockSeen, PoppedRegsList);
					if (success)
						++MatchCount;
					else
						break;
				}
			}
			if (ReturnBlockSeen) { // this is not the first return block
				if (MatchCount != PoppedRegsList.size()) {
					// Did not have same number of successes as prior return blocks. Inconsistent.
					success = false;
					break;
				}
			}
			ReturnBlockSeen = true;
		}
	} // end for all blocks

	if (success) {
		// Try to match elements from the two lists.
		map<uint32_t, STARS_sval_t>::iterator PushIter, PopIter;
		while (!PushedRegsList.empty() && (!PoppedRegsList.empty())) {
			PushIter = PushedRegsList.begin();
			PopIter = PoppedRegsList.begin();
			uint32_t PushedReg = PushIter->first;
			uint32_t PoppedReg = PopIter->first;
			if (PushedReg == PoppedReg) {
				if (PushIter->second == PopIter->second) { // match
					this->PreservedRegsBitmap.set(PushedReg);
				}
				PushedRegsList.erase(PushIter);
				PoppedRegsList.erase(PopIter);
			}
			else if (PushedReg < PoppedReg) {
				PushedRegsList.erase(PushIter);
			}
			else { // must be PushedReg > PoppedReg
				PoppedRegsList.erase(PopIter);
			}
		}
	}

	PushedRegsList.clear();
	PoppedRegsList.clear();
	return;
} // end of SMPFunction::FindPreservedRegs()

// Find failure to preserve callee-saved regs
void SMPFunction::AuditCallingConvention(void) {
	string FuncName(this->GetFuncName());
	// Exclude special cases that are not required to preserve callee-saved regs.
	if (IsStartupFuncName(FuncName) || (0 == strcmp("main", FuncName.c_str())) || this->IsThunkFunction())
		return;

	STARSOpndSetIter KillIter;
	for (KillIter = this->GetFirstVarKill(); KillIter != this->GetLastVarKill(); ++KillIter) {
		STARSOpndTypePtr KillOp = (*KillIter);
		if (KillOp->IsRegOp()) {
			STARS_regnum_t RegNum = KillOp->GetReg();
			if (!this->IsRegPreserved(RegNum)) {
				// Killed and not saved+restored
				if (global_STARS_program->IsCalleeSavedReg(RegNum)) {
					SMP_msg("WARNING: Callee-saved reg %u killed and not preserved in func at %llx\n",
						RegNum, (uint64_t) this->GetFirstFuncAddr());
				}
			}
		}
	}
	return;
} // SMPFunction::AuditCallingConvention()

void SMPFunction::FindAllAllocsAndDeallocs(void) {
	bool FoundAllocInstr = false;
	bool FoundDeallocInstr = false;
	bool FoundAlloca = false;
	bool DebugFlag = false;
#if SMP_DEBUG_FRAMEFIXUP
	DebugFlag |= (0 == strcmp("_dl_profile_fixup", this->GetFuncName()));
#endif

	list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	++InstIter;  // skip marker instruction
#endif
	for (; InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t addr = CurrInst->GetAddr();

		// Keep the most recent instruction in the DeallocInstr
		//  in case we reach the return without seeing a dealloc.
		if (!FoundDeallocInstr) {
			this->LocalVarsDeallocInstr = addr;
		}

		if (addr == this->LocalVarsAllocInstr) {
			FoundAllocInstr = true;
		}
		else if (FoundAllocInstr) {
			// We can now start searching for the DeallocInstr.
			if (CurrInst->MDIsFrameDeallocInstr(UseFP, this->GetLocalVarsSize())) {
				// Keep saving the most recent addr that looks
				//  like the DeallocInstr until we reach the
				//  end of the function. Last one to look like
				//  it is used as the DeallocInstr.
#if SMP_DEBUG_CONTROLFLOW
				SMP_msg("Returned from MDIsFrameDeallocInstr()\n");
#endif
				this->LocalVarsDeallocInstr = addr;
				FoundDeallocInstr = true;
			}
			else if (CurrInst->MDIsFrameAllocInstr()) {
				FoundAlloca = true;
				if (DebugFlag)
					SMP_msg("INFO: FoundAlloca set to true at %llx\n", (uint64_t)addr);
				if (CurrInst->HasAllocaRTL()) {
					if (DebugFlag)
						SMP_msg("INFO: Alloca RTL found at %llx\n", (uint64_t)addr);
					CurrInst->SetAllocaCall();
				}
			}
			else if (CurrInst->MDIsPushInstr()) {
				this->PushAfterLocalVarAlloc = true;
			}
		}
		else if (CurrInst->HasAllocaRTL()) {
			CurrInst->SetAllocaCall();
			// Don't set this->CallsAlloca() if the alloca() is before the frame alloc
		}
	}
	this->CallsAlloca = FoundAlloca;

	return;
} // end of SMPFunction::FindAllAllocsAndDeallocs()

// Compute FramePointerStackDelta as soon as possible so that it is available for SyncAllRTs().
void SMPFunction::FindFramePointerDelta(void) {
	bool FirstBlockProcessed = false;
	bool FPSaved = false;  // have seen push of frame pointer reg
	bool SPintoFP = false; // have seen copy of stack pointer into frame pointer
	STARS_sval_t IncomingDelta = 0;
	STARS_sval_t CurrentDelta;
	list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	++InstIter;  // skip marker instruction
#endif
	while (!FirstBlockProcessed && (InstIter != this->Instrs.end())) {
		SMPInstr *CurrInst = (*InstIter);
		// Accumulate stack delta values.
		CurrentDelta = CurrInst->AnalyzeStackPointerDelta(IncomingDelta, this->GetFramePtrStackDelta());
		if (SMP_STACK_POINTER_BITWISE_AND_CODE == ((STARS_uval_t) CurrentDelta)) {
			// For now, we ignore instructions that AND a constant into the stack pointer.
			CurrentDelta = 0;
		}
		else if (SMP_STACK_DELTA_ERROR_CODE == ((STARS_uval_t) CurrentDelta)) {
			this->AnalyzedSP = false;
			break; // error exit
		}
		// Look for initialization of frame pointer, record its stack delta
		FirstBlockProcessed = CurrInst->IsLastInBlock();
		if (!FPSaved) { // still looking for "push <framepointerreg>"
			if (CurrInst->MDIsPushInstr() && CurrInst->GetOperand(0)->MatchesReg(MD_FRAME_POINTER_REG)) {
				FPSaved = true;
			}
		}
		else if (!SPintoFP) { // found "push <framepointerreg>", looking for "fp := sp"
			if ((CurrInst->GetIDAOpcode() == MD_MOVE_INSTRUCTION) && CurrInst->IsAnalyzeable()
				&& (CurrInst->GetFirstDef()->GetOp()->MatchesReg(MD_FRAME_POINTER_REG))
				&& (CurrInst->GetFirstUse()->GetOp()->MatchesReg(MD_STACK_POINTER_REG))) {
				SPintoFP = true;
				this->FramePointerStackDelta = IncomingDelta;
				FirstBlockProcessed = true; // stop looking
				assert(this->UsesFramePointer());
			}
		}
		IncomingDelta += CurrentDelta;
		++InstIter;
	}

	return;
} // end of SMPFunction::FindFramePointerDelta()

// Figure out the different regions of the stack frame, and find the
//  instructions that allocate and deallocate the local variables space
//  on the stack frame.
// The stack frame info will be used to emit stack
//  annotations when Analyze() reaches the stack allocation
//  instruction that sets aside space for local vars.
// Set the address of the instruction at which these
//  annotations should be emitted. This should normally
//  be an instruction such as:  sub esp,48
//  However, for a function with no local variables at all,
//  we will need to determine which instruction should be
//  considered to be the final instruction of the function
//  prologue and return its address.
// Likewise, we find the stack deallocating instruction in
//  the function epilogue.
void SMPFunction::SetStackFrameInfo(void) {

	// Now, find the boundaries between stack frame objects.
	this->BuildLocalVarTable();

	// Get callee-saved regs info for remediation use.
	if ((STARS_BADADDR != this->GetFirstFrameAllocInstAddr()) && this->StackPtrAnalysisSucceeded()) {
		this->MDFindSavedRegs();
	}

	return;
} // end of SMPFunction::SetStackFrameInfo()

// IDA Pro defines the sizes of regions in the stack frame in a way
//  that suits its purposes but not ours. The frsize field of the func_info_t
//  structure measures the distance between the stack pointer and the
//  frame pointer (ESP and EBP in the x86). This region includes some
//  of the callee-saved registers. So, the frregs field only includes
//  the callee-saved registers that are above the frame pointer.
//  x86 standard prologue on gcc/linux:
//    push ebp      ; save old frame pointer
//    mov ebp,esp   ; new frame pointer = current stack pointer
//    push esi      ; callee save reg
//    push edi      ; callee save reg
//    sub esp,34h   ; allocate 52 bytes for local variables
//
//  Notice that EBP acquires its final frame pointer value AFTER the
//  old EBP has been pushed. This means that, of the three callee saved
//  registers, one is above where EBP points and two are below.
//  IDA Pro is concerned with generating readable addressing expressions
//  for items on the stack. None of the callee-saved regs will ever
//  be addressed in the function; they will be dormant until they are popped
//  off the stack in the function epilogue. In order to create readable
//  disassembled code, IDA defines named constant offsets for locals. These
//  offsets are negative values (x86 stack grows downward from EBP toward
//  ESP). When ESP_relative addressing occurs, IDA converts a statement:
//    mov eax,[esp+12]
//  into the statement:
//    mov eax,[esp+3Ch+var_30]
//  Here, 3Ch == 60 decimal is the distance between ESP and EBP, and
//  var_30 is defined to have the value -30h == -48 decimal. So, the
//  "frame size" in IDA Pro is 60 bytes, and a certain local can be
//  addressed in ESP-relative manner as shown, or as [ebp+var_30] for
//  EBP-relative addressing. The interactive IDA user can then edit
//  the name var_30 to something mnemonic, such as "virus_size", and IDA
//  will replace all occurrences with the new name, so that code references
//  automatically become [ebp+virus_size]. As the user proceeds
//  interactively, he eventually produces very understandable code.
// This all makes sense for producing readable assembly text. However,
//  our analyses have a compiler perspective as well as a memory access
//  defense perspective. SMP distinguishes between callee saved regs,
//  which should not be overwritten in the function body, and local
//  variables, which can be written. We view the stack frame in logical
//  pieces: here are the saved regs, here are the locals, here is the
//  return address, etc. We don't care which direction from EBP the
//  callee-saved registers lie; we don't want to lump them in with the
//  local variables. We also don't like the fact that IDA Pro will take
//  the function prologue code shown above and declare frregs=4 and
//  frsize=60, because frsize no longer matches the stack allocation
//  statement sub esp,34h == sub esp,52. We prefer frsize=52 and frregs=12.
// So, the task of this function is to fix these stack sizes in our
//  private data members for the function, while leaving the IDA database
//  alone because IDA needs to maintain its own definitions of these
//  variables.
// Fixing means we will update the data members LocalVarsSize and
//  CalleeSavedRegsSize.
// NOTE: This function is both machine dependent and platform dependent.
//  The prologue and epilogue code generated by gcc-linux is as discussed
//  above, while on Visual Studio and other Windows x86 compilers, the
//  saving of registers other than EBP happens AFTER local stack allocation.
//  A Windows version of the function would expect to see the pushing
//  of ESI and EDI AFTER the sub esp,34h statement.

// How many irrelevant insts in the first basic block should we see before deciding that frame allocation is finished?
#define STARS_POST_CALL_NONALLOC_INSTR_LIMIT 10  // after function call seen
#define STARS_NONALLOC_INSTR_LIMIT 30            // after this many irrelevant insts even if no calls

bool SMPFunction::MDFixFrameInfo(void) {
	int SavedRegsSize = 0;
	int OtherPushesSize = 0;  // besides callee-saved regs
	int OtherAllocsSize = 0;  // additions to stack pointer, other odd instructions in prologue
	long AllocValue = 0;
	bool EBPSaved = false;  // detected push of frame pointer reg
	bool ESPintoEBP = false;  // detected initialization of frame pointer reg with stack pointer reg value
	bool FoundAllocInstr = false; // found simple alloc, e.g. subtraction from stack pointer
	bool DebugFlag = (0 == strcmp("__libc_csu_init", this->GetFuncName()));
	bool CallSeen = false;
	uint16_t PostCallNonAllocInstructions = 0;
	uint16_t NonAllocatingInstructions = 0;
	STARS_ea_t CallAddr = STARS_BADADDR;
	STARS_ea_t SaveAddr = this->GetStartAddr(); // keeps most recent best value for stack frame allocation instr
	list<pair<STARS_ea_t, int> > AllocPointsList; // list of pair<InstAddr, bytes allocated on stack at that addr>

	// Iterate through the first basic block in the function. If we find
	//  a frame allocating Instr in it, then we have local vars. If not,
	//  we don't, and LocalVarsSize should have been zero. Count the callee
	//  register saves leading up to the local allocation. Set data members
	//  according to what we found if the values of the data members would
	//  change.
	SMPBasicBlock *CurrBlock = this->Blocks.front();
	vector<SMPInstr *>::iterator CurrIter = CurrBlock->GetFirstInst();
#if SMP_USE_SSA_FNOP_MARKER
	++CurrIter;  // skip marker instruction
#endif
	for ( ; CurrIter != CurrBlock->GetLastInst(); ++CurrIter) {
		SMPInstr *CurrInstr = (*CurrIter);
		STARS_ea_t InstAddr = CurrInstr->GetAddr();
		if (!CurrInstr->IsAnalyzeable())
			continue;

		if (FoundAllocInstr && (DEFAULT != CurrInstr->GetDataFlowType())) {
			// After we find the frame allocation instruction, a call or any other
			//  control flow instruction should terminate our processing. A stack
			//  frame adjustment could follow a call, which could falely look like another
			//  frame allocation instruction.
			break;
		}
		if (CALL == CurrInstr->GetDataFlowType()) {
			CallSeen = true;
			CallAddr = InstAddr;
		}
		if (CurrInstr->MDIsPushInstr()) {
			if (!FoundAllocInstr)
				SaveAddr = InstAddr;
			// We will make the gcc-linux assumption that a PUSH in
			//  the first basic block, prior to the stack allocating
			//  instruction, is a callee register save. To make this
			//  more robust, we ensure that the register is from
			//  the callee saved group of registers, and that it has
			//  not been defined thus far in the function (else it might
			//  be a push of an outgoing argument to a call that happens
			//  in the first block when there are no locals). **!!!!**
			size_t PushWidth = CurrInstr->GetPushedOpndByteSize();
			pair<STARS_ea_t, int> AllocPair(InstAddr, (int) PushWidth);
			AllocPointsList.push_back(AllocPair);
			if (CurrInstr->MDUsesCalleeSavedReg() && !CurrInstr->HasSourceMemoryOperand()) {
				SavedRegsSize += PushWidth;
				if (DebugFlag) SMP_msg("libc_csu_init SavedRegsSize: %d  %s\n", SavedRegsSize,
					CurrInstr->GetDisasm());
			}
			else {
				// Pushes of outgoing args can be scheduled so that
				//  they are mixed with the pushes of callee saved regs.
				OtherPushesSize += PushWidth;
				if (DebugFlag) SMP_msg("libc_csu_init OtherPushesSize: %d  %s\n", OtherPushesSize,
					CurrInstr->GetDisasm());
			}
			if (!EBPSaved) { // still looking for "push ebp"
				if (CurrInstr->GetOperand(0)->MatchesReg(MD_FRAME_POINTER_REG)) {
					EBPSaved = true;
				}
			}
		}
		else if (CurrInstr->MDIsFrameAllocInstr() && (!CurrInstr->HasAllocaRTL())) {
			if (DebugFlag) SMP_msg("libc_csu_init allocinstr: %s\n", CurrInstr->GetDisasm());
			// Get the size being allocated.
			set<DefOrUse, LessDefUse>::iterator CurrUse;
			for (CurrUse = CurrInstr->GetFirstUse(); CurrUse != CurrInstr->GetLastUse(); ++CurrUse) {
				// Find the immediate operand.
				if (CurrUse->GetOp()->IsImmedOp()) {
					// Get its value into LocalVarsSize.
					AllocValue = (signed long) CurrUse->GetOp()->GetImmedValue();
					// One compiler might have sub esp,24 and another
					//  might have add esp,-24. Take the absolute value.
					if (0 > AllocValue)
						AllocValue = -AllocValue;
					SaveAddr = InstAddr;  // ImmedOp, so now we know this is an alloc and not an alloca() call
					FoundAllocInstr = true;
					pair<STARS_ea_t, int> AllocPair(InstAddr, (int) AllocValue);
					AllocPointsList.push_back(AllocPair);
					break; // only want to process the immediate operand
				} // end if (o_imm == ...)
			} // end for all uses
		} // end if (push) .. elsif frame allocating instr
		else if (CurrInstr->HasStackPointerDEF() && (DEFAULT == CurrInstr->GetDataFlowType())) {
			// Handle all other possible changes to stack pointer register from non-control-flow instructions.
			//  This includes the alloca() inlined call pattern (sub esp,reg) which is assigned a default size.
			if (CurrInstr->MDIsLeaveInstr() && (STARS_BADADDR != this->LocalVarsAllocInstr)) {
				SMP_msg("ERROR: MDFixFrameInfo() found LEAVE opcode at %llx before finding a frame allocation. ", (uint64_t) InstAddr);
				this->DumpFuncNameAndAddr();
				break;  // should not find a LEAVE opcode before the frame allocation
			}
			STARS_sval_t CurrentDelta = CurrInstr->AnalyzeStackPointerDelta(0, 0);
			if (SMP_STACK_POINTER_BITWISE_AND_CODE == ((STARS_uval_t) CurrentDelta)) {
				// For now, we ignore instructions that AND a constant into the stack pointer.
				SMP_msg("WARNING: Stack pointer bitwise AND ignored at %llx in MDFixFrameInfo ", (uint64_t) InstAddr);
				this->DumpFuncNameAndAddr();
			}
			else if (SMP_STACK_DELTA_ERROR_CODE == ((STARS_uval_t) CurrentDelta)) {
				SMP_msg("ERROR: Stack delta unanalyzeable at %llx in MDFixFrameInfo ", (uint64_t) InstAddr);
				this->DumpFuncNameAndAddr();
			}
			else if (0 < CurrentDelta) { // deallocating stack; might be stack adjustment after a call instruction
				if (CallSeen) { // match the stack adjustment to the allocations before the call and remove them
					STARS_sval_t AllocBytesFound = 0;
					list<pair<STARS_ea_t, int> >::reverse_iterator ListRevIter = AllocPointsList.rbegin();
					while (ListRevIter != AllocPointsList.rend()) {
						STARS_ea_t AllocAddr = (*ListRevIter).first;
						int CurrAllocBytes = (*ListRevIter).second;
						AllocBytesFound += (STARS_sval_t) CurrAllocBytes;
						if (AllocBytesFound >= CurrentDelta) { // found enough bytes to match call adjustment
							++ListRevIter;
							if (ListRevIter != AllocPointsList.rend()) {
								SaveAddr = (*ListRevIter).first;
								OtherPushesSize -= CurrentDelta; // Adjust allocated bytes to remove arg pushes before call inst
								SMP_msg("INFO: Removed arg pushes from stack allocs in %s\n", this->GetFuncName());
							}
							else { // reached beginning of function; no allocs, only arg pushes
								AllocValue = OtherAllocsSize = 0;
								SavedRegsSize = OtherPushesSize = 0;
								SMP_msg("INFO: No stack allocs other than arg pushes ");
								this->DumpFuncNameAndAddr();
							}
							if (AllocBytesFound > CurrentDelta) {
								SMP_msg("ERROR: Arg pushes did not match stack adjustment ");
								this->DumpFuncNameAndAddr();
							}
							break; // exit while loop
						}
						++ListRevIter;
					}
					if (AllocBytesFound < CurrentDelta) {
						AllocValue = OtherAllocsSize = 0;
						SavedRegsSize = OtherPushesSize = 0;
						SMP_msg("ERROR: Not enough stack allocs to equal call adjustment in first block ");
						this->DumpFuncNameAndAddr();
					}
				}
				else if (FoundAllocInstr) {
					SMP_msg("WARNING: Stack deallocation at %llx in first block with no call inst. ", (uint64_t) InstAddr);
					this->DumpFuncNameAndAddr();
				}
				else { // Bad; no allocs found, but dealloc found
					SMP_msg("ERROR: Stack deallocation at %llx in first block with no alloc instr or call found. ",	(uint64_t) InstAddr);
					this->DumpFuncNameAndAddr();
				}
				break; // exit after stack deallocation of any kind
			}
			else { // success; normal allocation
				int AllocSize = (int) (0 - CurrentDelta);
				if (FoundAllocInstr)
					this->AllocSizeAfterFrameAlloc += AllocSize;
				else
					OtherAllocsSize += AllocSize;
				pair<STARS_ea_t, int> AllocPair(InstAddr, AllocSize);
				AllocPointsList.push_back(AllocPair);
				SMP_msg("INFO: Miscellaneous stack pointer delta of %d found at %llx in MDFixFrameInfo\n", CurrentDelta, (uint64_t) InstAddr);
			}
		}
		else if (EBPSaved && (!ESPintoEBP)) { // found "push ebp", looking for "mov ebp,esp"
			if ((CurrInstr->GetIDAOpcode() == STARS_NN_mov)
				&& (CurrInstr->GetFirstDef()->GetOp()->MatchesReg(MD_FRAME_POINTER_REG))
				&& (CurrInstr->GetFirstUse()->GetOp()->MatchesReg(MD_STACK_POINTER_REG))) {
				ESPintoEBP = true;
			}
		}
		else {
			++NonAllocatingInstructions;
			if (CallSeen) {
				++PostCallNonAllocInstructions;
				if (STARS_POST_CALL_NONALLOC_INSTR_LIMIT <= PostCallNonAllocInstructions) {
					SMP_msg("INFO: Stopping search of first block for frame allocations after call at %llx\n",
						(uint64_t) InstAddr);
					break;
				}
			}
			if (STARS_NONALLOC_INSTR_LIMIT <= NonAllocatingInstructions) {
				SMP_msg("INFO: Stopping search of first block for frame allocations at %llx\n",
					(uint64_t) InstAddr);
				break;
			}
		}
	} // end for all instructions in the first basic block

	this->LocalVarsSize = (STARS_asize_t) AllocValue + (STARS_asize_t) OtherAllocsSize;
	this->CalleeSavedRegsSize = (uint16_t) SavedRegsSize + OtherPushesSize;
	this->LocalVarsAllocInstr = SaveAddr;

	// If we found ESPintoEBP, we also found EBPSaved first, and we need to set
	//  this->UseFP to true.
	this->UseFP = ESPintoEBP;
	if (ESPintoEBP) {
		this->FindFramePointerDelta();
	}

	return true;
} // end of SMPFunction::MDFixFrameInfo()


// IDA Pro is sometimes confused by a function that uses the frame pointer
//  register for other purposes. For the x86, a function that uses EBP
//  as a frame pointer would begin with: push ebp; mov ebp,esp to save
//  the old value of EBP and give it a new value as a frame pointer. The
//  allocation of local variable space would have to come AFTER the move
//  instruction. A function that begins: push ebp; push esi; sub esp,24
//  is obviously not using EBP as a frame pointer. IDA is apparently
//  confused by the push ebp instruction being the first instruction
//  in the function. We will reset UseFP to false in this case.
// The inverse problem happens with a function that begins with instructions
//  other than push ebp; mov ebp,esp; ... etc. but eventually has those
//  instructions in the first basic block. For example, a C compiler generates
//  for the first block of main():
//    lea ecx,[esp+arg0]
//    and esp, 0xfffffff0
//    push dword ptr [ecx-4]
//    push ebp
//    mov ebp,esp
//    push ecx
//    sub esp,<framesize>
//
//  This function is obviously using EBP as a frame pointer, but IDA Pro marks
//  the function as not using a frame pointer. We will reset UseFP to true in
//  this case.
// NOTE: This logic should work for both Linux and Windows x86 prologues.
// NOTE As of July, 2015, this logic was incorporated into MDFixFrameInfo().
#if 0
bool SMPFunction::MDFixUseFP(void) {
	bool OldUseFP = this->UsesFramePointer();
	bool HasLocals = (0 < this->GetLocalVarsSize());
	list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
	STARS_ea_t addr;

#if SMP_USE_SSA_FNOP_MARKER
	++InstIter;  // skip marker instruction
#endif
	SMPInstr *CurrInst;

#if 0
	if (!(this->UseFP)) {
#endif
		// See if we can detect the instruction "push ebp" followed by the instruction
		//  "mov ebp,esp" in the first basic block. The instructions do not have to be
		//  consecutive. If we find them, we will reset UseFP to true.
		bool FirstBlockProcessed = false;
		bool EBPSaved = false;
		bool ESPintoEBP = false;
		do {
			CurrInst = (*InstIter);
			addr = CurrInst->GetAddr();
			FirstBlockProcessed = CurrInst->IsLastInBlock();
			if (!EBPSaved) { // still looking for "push ebp"
				if (CurrInst->MDIsPushInstr() && CurrInst->GetOperand(0)->MatchesReg(MD_FRAME_POINTER_REG)) {
					EBPSaved = true;
				}
			}
			else if (!ESPintoEBP) { // found "push ebp", looking for "mov ebp,esp"
				if ((CurrInst->GetIDAOpcode() == STARS_NN_mov) 
					&& (CurrInst->GetFirstDef()->GetOp()->MatchesReg(MD_FRAME_POINTER_REG))
					&& (CurrInst->GetFirstUse()->GetOp()->MatchesReg(MD_STACK_POINTER_REG))) {
					ESPintoEBP = true;
					FirstBlockProcessed = true; // exit loop
				}
			}
			// We must get EBP set to its frame pointer value before we reach the
			//  local frame allocation instruction (i.e. the subtraction of locals space
			//   from the stack pointer).
			if (HasLocals) {
				FirstBlockProcessed |= (addr >= this->LocalVarsAllocInstr);
			}
			++InstIter;
		} while (!FirstBlockProcessed);
		// If we found ESPintoEBP, we also found EBPSaved first, and we need to change
		//  this->UseFP to true and return true. Otherwise, return false.
		this->UseFP = ESPintoEBP;
		bool changed = (ESPintoEBP != OldUseFP);
		if (changed)
			SMP_msg("INFO: MDFixUseFP toggled UseFP for %s\n", this->GetFuncName());
		return (changed);
#if 0
	} // end if (!(this->UseFP))

	// At this point, this->UseFP must have been true on entry to this method and we will
	//  check whether it should be reset to false.
	while (addr <= this->LocalVarsAllocInstr) {
		set<DefOrUse, LessDefUse>::iterator CurrDef = CurrInst->GetFirstDef();
		while (CurrDef != CurrInst->GetLastDef()) {
			if (CurrDef->GetOp()->MatchesReg(MD_FRAME_POINTER_REG))
				return false; // EBP got set before locals were allocated
			++CurrDef;
		}
		++InstIter;
		CurrInst = (*InstIter);
		addr = CurrInst->GetAddr();
	}
	// If we found no defs of the frame pointer before the local vars
	//  allocation, then the frame pointer register is not being used
	//  as a frame pointer, just as a general callee-saved register.
	this->UseFP = false;
	SMP_msg("INFO: MDFixUseFP reset UseFP to false for %s\n", this->GetFuncName());
	return true;
#endif
} // end of SMPFunction::MDFixUseFP()
#endif

// Find the callee-saved reg offsets (negative offset from return address)
//  for all registers pushed onto the stack before the stack frame allocation
//  instruction.
void SMPFunction::MDFindSavedRegs(void) {
	list<SMPInstr *>::iterator InstIter;
	int RegIndex;
	STARS_Function_t *CurrFunc = SMP_get_func(this->GetStartAddr());
	assert(nullptr != CurrFunc);

	for (SMPInstr *CurrInst : this->Instrs) {
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (!(CurrInst->MDIsPushInstr())) {
			if (InstAddr != this->LocalVarsAllocInstr) {
				continue;
			}
		}

		STARS_sval_t CurrOffset = CurrInst->GetStackPtrOffset();
		if (CurrInst->GetIDAOpcode() == STARS_NN_push) {
			STARSOpndTypePtr PushedReg = CurrInst->GetPushedOpnd();
			if (PushedReg->IsRegOp()) {
				RegIndex = (int) PushedReg->GetReg();
				if (RegIndex > global_STARS_program->GetSTARS_MD_LAST_SAVED_REG_NUM()) {
					SMP_msg("WARNING: Skipping save of register %d\n", RegIndex);
					continue;
				}
				if (this->SavedRegLoc[(size_t) RegIndex] == 0) {
					this->SavedRegLoc[(size_t) RegIndex] = CurrOffset - PushedReg->GetByteWidth();
				}
				else {
					SMP_msg("WARNING: Multiple saves of register %d\n", RegIndex);
				}
			} // end if register push operand
		} // end if PUSH instruction
		else if (STARS_NN_pusha == CurrInst->GetIDAOpcode()) {
			// NOTE: There is no push-all of the 64-bit regs, so these sizes are correct.
			this->SavedRegLoc[(std::size_t) STARS_x86_R_ax] = CurrOffset - 4;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_cx] = CurrOffset - 8;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_dx] = CurrOffset - 12;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_bx] = CurrOffset - 16;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_sp] = CurrOffset - 20;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_bp] = CurrOffset - 24;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_si] = CurrOffset - 28;
			this->SavedRegLoc[(std::size_t) STARS_x86_R_di] = CurrOffset - 32;
			break; // all regs accounted for
		}
		else if (CurrInst->MDIsEnterInstr()) {
			this->SavedRegLoc[(std::size_t) STARS_x86_R_bp] = CurrOffset - global_STARS_program->GetSTARS_ISA_Bytewidth();
		}

		if (InstAddr == this->LocalVarsAllocInstr)
			break;
	} // end for all instructions

	return;
} // end of SMPFunction::MDFindSavedRegs()

// Compute the ReturnRegTypes[] as the meet over all register types
//  at all return instructions. Return true if any type was updated.
bool SMPFunction::MDFindReturnTypes(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	bool changed = false;

	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		if (CurrBlock->HasReturn()) {
			// Get the types of all registers at the RETURN point.
			//  Calculate the meet function over them.
			InstIter = --(CurrBlock->GetLastInst());
			CurrInst = (*InstIter);
			assert(RETURN == CurrInst->GetDataFlowType());
			set<DefOrUse, LessDefUse>::iterator CurrUse;
			for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
				STARSOpndTypePtr UseOp = CurrUse->GetOp();
				if ((! UseOp->IsRegOp()) || (global_STARS_program->GetSTARS_MD_LAST_SAVED_REG_NUM() < UseOp->GetReg()))
					continue;
				SMPOperandType OldType = this->ReturnRegTypes[(size_t)UseOp->GetReg()];
				bool TypeErrorFlag = false;
				SMPOperandType NewType = SMPTypeMeet(OldType, CurrUse->GetType(), TypeErrorFlag);
				if (TypeErrorFlag) {
					SMP_msg("ERROR: TypeMeet error in MDFindReturnTypes() return at %llx ", (uint64_t) CurrInst->GetAddr());
					this->DumpFuncNameAndAddr();
				}
				if (NewType != OldType) {
					this->ReturnRegTypes[UseOp->GetReg()] = NewType;
					changed = true;
				}
				// Get the FGInfo for each USE as well.
				int UseSSANum = CurrUse->GetSSANum();
				int DefHashIndex = HashGlobalNameAndSSA(UseOp, UseSSANum);
				struct FineGrainedInfo RegDefFG;
				if (this->IsGlobalName(UseOp)) {
					RegDefFG = this->GetDefFGInfo(DefHashIndex);
				}
				else {
					RegDefFG = CurrBlock->GetDefFGInfo(DefHashIndex);
				}
				// Meet function for FGInfo is a simple bitwise OR operation.
				this->ReturnRegFGInfo[UseOp->GetReg()].SignMiscInfo |= RegDefFG.SignMiscInfo;
				this->ReturnRegFGInfo[UseOp->GetReg()].SizeInfo |= RegDefFG.SizeInfo;
			} // for all USEs in the RETURN instruction
		} // end if current block has a RETURN
	} // end for all blocks
	return changed;
} // end of SMPFunction::MDFindReturnTypes()

// Fill IncomingRegTypes[]
void SMPFunction::MDFindIncomingTypes(void) {
	// For each register USE in the marker inst at the top of the function,
	//  see if we have a consistent type inference at all call sites (i.e.
	//  the USEs of the call instructions that call this function).
	set<STARS_ea_t>::iterator CallSiteIter;
	for (CallSiteIter = this->AllCallSites.begin(); CallSiteIter != this->AllCallSites.end(); ++CallSiteIter) {
		STARS_ea_t CallSiteAddr = (*CallSiteIter);
		STARS_Function_t *CurrFunc = SMP_get_func(CallSiteAddr);
		assert(nullptr != CurrFunc);
		STARS_ea_t CallerFirstAddr = CurrFunc->get_startEA();
		SMPFunction *CallerFunc = this->GetProg()->FindFunction(CallerFirstAddr);
		if (nullptr == CallerFunc) {
			SMP_msg("FATAL ERROR: Program has no function at %llx which should be a caller of %llx\n",
				(uint64_t) CallerFirstAddr, (uint64_t) this->GetFirstFuncAddr());
		}
		assert(nullptr != CallerFunc);
		SMPInstr *CallSiteInst = CallerFunc->GetInstFromAddr(CallSiteAddr);
		// Iterate through USE regs in our marker inst, match them to USEs in call inst.
		SMPInstr *MarkerInst = this->Instrs.front();
		assert(MarkerInst->IsMarkerInst());
		for (set<DefOrUse, LessDefUse>::iterator MarkerDefIter = MarkerInst->GetFirstDef(); MarkerDefIter != MarkerInst->GetLastDef(); ++MarkerDefIter) {
			STARSOpndTypePtr MarkerDefOp = MarkerDefIter->GetOp();
			if (MarkerDefOp->IsRegOp()) {
				// Find same USE in caller inst.
				STARS_regnum_t RegNum = MarkerDefOp->GetReg();
				set<DefOrUse, LessDefUse>::iterator CallSiteUseIter = CallSiteInst->FindUse(MarkerDefOp);
				if ((CallSiteUseIter != CallSiteInst->GetLastUse()) && (RegNum < (decltype(RegNum))this->IncomingRegTypes.size())) {
					// Found it. See if we have type info.
					SMPOperandType CallSiteType = CallSiteUseIter->GetType();
					SMPOperandType OldType = this->IncomingRegTypes[RegNum];
					bool TypeErrorFlag = false;
					this->IncomingRegTypes[RegNum] = SMPTypeMeet(CallSiteType, OldType, TypeErrorFlag);
					if (TypeErrorFlag) {
						SMP_msg("ERROR: TypeMeet error in MDFindIncomingTypes() ");
						this->DumpFuncNameAndAddr();
					}
					// Note this type meet function is used in an aggressive way here, as
					//  a mixture of UNINIT and NUMERIC at various call sites will produce NUMERIC, for example.
					//  We are not demanding that all call sites have a consistent type with no UNINIT instances.

					// Get the FGInfo (from call site in CallerFunc) for each USE as well.
					int UseSSANum = CallSiteUseIter->GetSSANum();
					int DefHashIndex = HashGlobalNameAndSSA(MarkerDefOp, UseSSANum);
					struct FineGrainedInfo RegDefFG;
					if (CallerFunc->IsGlobalName(MarkerDefOp)) {
						RegDefFG = CallerFunc->GetDefFGInfo(DefHashIndex);
					}
					else {
						RegDefFG = CallSiteInst->GetBlock()->GetDefFGInfo(DefHashIndex);
					}
					// Meet function for FGInfo is a simple bitwise OR operation.
					(void) MarkerInst->UpdateDefOpFGInfo(MarkerDefOp, RegDefFG);
				}
			}
		}
	}
	return;
} // end of SMPFunction::MDFindIncomingTypes()

// Determine boundaries in the stack frame.
void SMPFunction::BuildLocalVarTable(void) {
	this->BuildStackAccessTables();
	return;
} // end of SMPFunction::BuildLocalVarTable()

// Build tables to characterize stack accesses.
void SMPFunction::BuildStackAccessTables(void) {

	STARS_Function_t *FuncPtr = SMP_get_func(this->GetStartAddr());
	if (nullptr == FuncPtr) {
		SMP_msg("FATAL ERROR in SMPFunction::SemiNaiveLocalVarID; no func ptr\n");
	}
	assert(nullptr != FuncPtr);

	// If AnalyzedSP is false, that is all we can do.
	if (!this->AnalyzedSP) {
		this->OutgoingArgsSize = 0;
		this->MinStackDelta = 0;
		this->AllocPointDelta = 0;
		return;
	}

	// Calculate min and max stack operand offsets accessed.
	list<SMPInstr *>::iterator InstIter;
	InstIter = this->Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	if ((*InstIter)->IsMarkerInst())
		++InstIter;  // skip marker instruction
#endif
	for (; InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t addr = CurrInst->GetAddr();
		// Find the min and max stack offsets in DEFs and USEs.
		if (CurrInst->HasDestMemoryOperand() || CurrInst->MDIsPushInstr() || CurrInst->MDIsEnterInstr()) {
			set<DefOrUse, LessDefUse>::iterator CurrDef;
			for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
				STARSOpndTypePtr TempOp = CurrDef->GetOp();
				if ((!TempOp->IsMemNoDisplacementOp()) && (!TempOp->IsMemDisplacementOp()))
					continue;
				this->UpdateMinMaxStackOffsets(CurrInst, TempOp);
			} // end for all DEFs
		}
		if (CurrInst->HasSourceMemoryOperand() || CurrInst->MDIsPopInstr() || CurrInst->MDIsLeaveInstr() || CurrInst->MDIsLoadEffectiveAddressInstr()) {
			if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
				STARSOpndTypePtr TempOp = CurrInst->GetLeaMemUseOp();
				if (nullptr != TempOp) {
					if (((TempOp->IsMemNoDisplacementOp()) || (TempOp->IsMemDisplacementOp())) && (!(CurrInst->IsRegClearIdiom() || CurrInst->IsNop())))  {
						this->UpdateMinMaxStackOffsets(CurrInst, TempOp);
					}
				}
			}
			else {
				set<DefOrUse, LessDefUse>::iterator CurrUse;
				for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
					STARSOpndTypePtr TempOp = CurrUse->GetOp();
					if ((!TempOp->IsMemNoDisplacementOp()) && (!TempOp->IsMemDisplacementOp()))
						continue;
					this->UpdateMinMaxStackOffsets(CurrInst, TempOp);
				} // end for all USEs
			}
		}
	} // end for all instructions
	if (0 == this->MaxStackAccessLimit) {
		// Never accessed any incoming args. However, we know the return address is on the stack,
		//  and it is almost never accessed, so we want to record its presence.
		this->MaxStackAccessLimit = MD_DEFAULT_RETURN_ADDRESS_SIZE;
	}
	if (this->MinStackAccessOffset > this->MinStackDelta) {
		// Some functions allocate space that is not visibly accessed. We still want to make
		//  our stack frame maps of maximum size, and MinStackDelta is used for normalizing offsets.
		this->MinStackAccessOffset = this->MinStackDelta;
	}
	else if (this->MinStackAccessOffset < this->MinStackDelta) {
		// x86-64 leaf functions are often optimized by not allocating a stack frame. Instead,
		//  negative displacements from the stack pointer are used to access locals. So the
		//  stack pointer (reflected in MinStackDelta) never goes down as far as the bottom of
		//  the frame (reflected by MinStackAccessOffset). We need to record that such a function
		//  has been detected so that we don't fail assertions unnecessarily later.
		this->SetStackFrameExtendsPastStackTop();
	}

	if (!this->AnalyzedSP || (STARS_BADADDR == this->LocalVarsAllocInstr)) {
		SMP_msg("FindOutgoingArgsSize not called for %s ", this->GetFuncName());
		SMP_msg("AnalyzedSP: %d CallsAlloca: %d LocalVarsAllocInstr: %llx \n",
			this->AnalyzedSP, this->CallsAlloca, (uint64_t) this->LocalVarsAllocInstr);
		return;
	}
	

	// Now, do the mapping of instruction stack accesses to the table entries.
	unsigned short BitWidthMask;
	bool DebugFlag = false;
	int SignedOffset;

	if ((0 <= this->MinStackDelta) || (0 <= this->AllocPointDelta)) {
		// No allocations; sometimes happens in library functions.
		this->OutgoingArgsSize = 0;
		this->AllocPointDelta = 0;
		if ((this->MinStackDelta > this->MaxStackDelta) || (0 < this->MinStackDelta)) {
			this->MinStackDelta = 0;
		}
	}
	assert(0 >= this->MinStackDelta);

	// Allocate vectors of stack frame entries, one for each byte of the stack frame.
	// One vector will be for negative stack offsets (e.g. saved regs, local vars) and
	//  the other vector will be for nonnegative offsets (e.g. return address, inargs).
	// Two more vectors will hold fine-grained stack access info.
	//  These will be our memory maps for analyzing stack usage.
	STARS_sval_t AccessTableEntriesCount = this->MaxStackAccessLimit - this->MinStackAccessOffset;
	for (int i = this->MinStackAccessOffset; i < this->MaxStackAccessLimit; ++i) {
		struct StackFrameEntry TempEntry;
		TempEntry.VarPtr = nullptr;
		TempEntry.offset = (long) i;
		TempEntry.Read = false;
		TempEntry.Written = false;
		TempEntry.AddressTaken = false;
		TempEntry.ESPRelativeAccess = false;
		TempEntry.EBPRelativeAccess = false;
		TempEntry.IndexedAccess = false;
		struct FineGrainedInfo TempFineGrained;
		TempFineGrained.SignMiscInfo = 0;
		TempFineGrained.SizeInfo = 0;
		if (i < 0) {
			TempEntry.EntryType = STARS_STACK_UNKNOWN;
			this->NegativeOffsetStackFrameMap.push_back(TempEntry);
			this->NegativeOffsetFineGrainedStackTable.push_back(TempFineGrained);
		}
		else {
			if (i < ((int) MD_DEFAULT_RETURN_ADDRESS_SIZE))
				TempEntry.EntryType = STARS_STACK_RETURN_ADDRESS;
			else
				TempEntry.EntryType = STARS_STACK_INARG;
			this->PositiveOffsetStackFrameMap.push_back(TempEntry);
			this->PositiveOffsetFineGrainedStackTable.push_back(TempFineGrained);
		}
	}

	if (0 < this->AllocPointDelta) {
		SMP_msg("FATAL ERROR: AllocPointDelta = %ld ", (long) this->AllocPointDelta);
		this->DumpFuncNameAndAddr();
	}
	assert(0 >= this->AllocPointDelta);

	// Iterate through all instructions and record stack frame accesses in the StackFrameMaps.
	InstIter = this->Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	if ((*InstIter)->IsMarkerInst())
		++InstIter;  // skip marker instruction
#endif
	for (; InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		STARS_sval_t sp_delta = CurrInst->GetStackPtrOffset();
		if (0 < sp_delta) {
			// Stack underflow.
			SMP_msg("WARNING: Stack underflow at %llx %s sp_delta: %ld\n", (uint64_t) InstAddr,
				CurrInst->GetDisasm(), (long)sp_delta);
#if 0
			this->OutgoingArgsComputed = false;
			this->OutgoingArgsSize = 0;
			return;
#endif
		}
		// assert(0 >= sp_delta);
		STARS_ea_t offset;
		std::size_t DataSize;
		bool UsedFramePointer;
		bool IndexedAccess;
		bool SignedMove;
		bool UnsignedMove;
		if (CurrInst->HasDestMemoryOperand() || CurrInst->MDIsPushInstr()) {
			set<DefOrUse, LessDefUse>::iterator CurrDef;
			for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
				STARSOpndTypePtr TempOp = CurrDef->GetOp();
				if ((!TempOp->IsMemNoDisplacementOp()) && (!TempOp->IsMemDisplacementOp()))
					continue;
				if (this->MDGetStackOffsetAndSize(CurrInst, TempOp, this->MinStackAccessOffset, offset, DataSize, UsedFramePointer,
					IndexedAccess, SignedMove, UnsignedMove)) {
					SignedOffset = (int) offset;
					if (IndexedAccess && ((0 > SignedOffset) || ((SignedOffset + (int) DataSize) > ((int) AccessTableEntriesCount)))) {
						continue; // Indexed expressions can be within frame even when offset is outside frame
					}
					assert(0 <= SignedOffset);
					SignedOffset += this->MinStackAccessOffset;

					bool ESPRelative = (!(UsedFramePointer || CurrInst->HasFPNormalizedToSP()));
					if (SignedOffset < 0) {
						bool OutOfRange = ((offset + DataSize - 1) >= this->NegativeOffsetStackFrameMap.size());
						if (OutOfRange) {
							if (IndexedAccess)
								continue; // Indexed expressions can be within frame even when offset is outside frame
							else if (this->CallsAlloca) {
								SMP_msg("ERROR: Stack access out of range at %llx due to alloca issues. ",	(uint64_t)InstAddr);
								this->DumpFuncNameAndAddr();
								this->AnalyzedSP = false;
								this->OutgoingArgsSize = 0;
								this->MinStackDelta = 0;
								this->AllocPointDelta = 0;
								return;
							}
						}
						assert((offset + DataSize - 1) < this->NegativeOffsetStackFrameMap.size());
						for (int j = 0; j < (int) DataSize; ++j) { // offset has zero-based index into negative offset vectors
							this->NegativeOffsetStackFrameMap[offset + j].Written = true;
							this->NegativeOffsetStackFrameMap[offset + j].IndexedAccess = IndexedAccess;
							if (ESPRelative) {
								this->NegativeOffsetStackFrameMap[offset + j].ESPRelativeAccess = true;
							}
							else {
								this->NegativeOffsetStackFrameMap[offset + j].EBPRelativeAccess = true;
							}
						}
					}
					else {
						assert((SignedOffset + DataSize - 1) < this->PositiveOffsetStackFrameMap.size());
						for (int j = 0; j < (int) DataSize; ++j) {
							this->PositiveOffsetStackFrameMap[SignedOffset + j].Written = true;
							this->PositiveOffsetStackFrameMap[SignedOffset + j].IndexedAccess = IndexedAccess;
							if (ESPRelative) {
								this->PositiveOffsetStackFrameMap[SignedOffset + j].ESPRelativeAccess = true;
							}
							else {
								this->PositiveOffsetStackFrameMap[SignedOffset + j].EBPRelativeAccess = true;
							}
						}
					}
					struct FineGrainedInfo StackDefFG;
					BitWidthMask = ComputeOperandBitWidthMask(TempOp, DataSize);
					StackDefFG.SizeInfo = BitWidthMask;
					StackDefFG.SignMiscInfo = FG_MASK_WRITTEN;
					if (SignedOffset < 0) { // offset has zero-based index into negative offset vectors
						assert(offset < this->NegativeOffsetFineGrainedStackTable.size());
						this->NegativeOffsetFineGrainedStackTable[offset].SizeInfo |= BitWidthMask;
						this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_WRITTEN;
						if (IndexedAccess) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
							StackDefFG.SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
						}
						if (ESPRelative) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_SP_RELATIVE;
							StackDefFG.SignMiscInfo |= FG_MASK_SP_RELATIVE;
						}
						else {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_FP_RELATIVE;
							StackDefFG.SignMiscInfo |= FG_MASK_FP_RELATIVE;
						}
					}
					else { // SignedOffset >= 0
						assert((size_t)SignedOffset < this->PositiveOffsetFineGrainedStackTable.size());
						this->PositiveOffsetFineGrainedStackTable[SignedOffset].SizeInfo |= BitWidthMask;
						this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_WRITTEN;
						if (IndexedAccess) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
							StackDefFG.SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
						}
						if (ESPRelative) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_SP_RELATIVE;
							StackDefFG.SignMiscInfo |= FG_MASK_SP_RELATIVE;
						}
						else {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_FP_RELATIVE;
							StackDefFG.SignMiscInfo |= FG_MASK_FP_RELATIVE;
						}
					}
					// We will process the signedness of stores later, so that loads can take precedence
					//  over stores in determining signedness in the table. We go ahead and process
					//  signedness for the separate DEF and USE maps by InstAddr.
					if (SignedMove) {
						StackDefFG.SignMiscInfo |= FG_MASK_SIGNED;
					}
					else if (UnsignedMove) {
						StackDefFG.SignMiscInfo |= FG_MASK_UNSIGNED;
					}
					if (!CurrInst->MDIsPushInstr()) { // could just exclude push-all, which had multiple DEFs per InstAddr
						// Insert the StackDefFG into the map of InstAddr to DEF FG info.
						pair<map<STARS_ea_t, struct FineGrainedInfo>::iterator, bool> InsertResult;
						pair<STARS_ea_t, struct FineGrainedInfo> InsertValue(InstAddr, StackDefFG);
						InsertResult = this->StackDefFGInfo.insert(InsertValue);
						if (!InsertResult.second) {
							SMP_msg("FATAL ERROR: Duplicate item in StackDefFGInfo at %llx ", (uint64_t) InstAddr);
							this->DumpFuncNameAndAddr();
							SMP_msg("Inst dump follows:\n");
							CurrInst->Dump();
							this->DumpDotCFG();
							this->Dump(false);
						}
						assert(InsertResult.second);
					}

					// Find saves of registers.
					if ((0 == CurrInst->GetBlock()->GetNumber()) && CurrInst->MDIsPushInstr()) { // push in entry block
						STARSOpndTypePtr PushedOp = CurrInst->GetPushedOpnd();
						if (PushedOp->IsRegOp()) {
							if (SignedOffset < 0) {
								for (int j = 0; j < (int)DataSize; ++j) {
									this->NegativeOffsetStackFrameMap[offset + j].EntryType = STARS_STACK_CALLEE_SAVED_REG;
								}
							}
							else { // odd to have callee-saved regs up here
								for (int j = 0; j < (int)DataSize; ++j) {
									this->PositiveOffsetStackFrameMap[SignedOffset + j].EntryType = STARS_STACK_CALLEE_SAVED_REG;
								}
							}
						}
					}
				} // end if MDGetStackOffsetAndSize()
			} // end for all DEFs
		} // end if DestMemoryOperand

		if (CurrInst->HasSourceMemoryOperand()) {
			set<DefOrUse, LessDefUse>::iterator CurrUse;
			for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
				STARSOpndTypePtr TempOp = CurrUse->GetOp();
				if ((!TempOp->IsMemNoDisplacementOp()) && (!TempOp->IsMemDisplacementOp()))
					continue;
				if (this->MDGetStackOffsetAndSize(CurrInst, TempOp, this->MinStackAccessOffset, offset, DataSize, UsedFramePointer,
					IndexedAccess, SignedMove, UnsignedMove)) {
					SignedOffset = (int) offset;
					if (IndexedAccess && ((0 > SignedOffset) || ((SignedOffset + (int) DataSize) > ((int) AccessTableEntriesCount)))) {
						continue; // Indexed expressions can be within frame but offset is outside frame
					}
					assert(0 <= SignedOffset);
					SignedOffset += this->MinStackAccessOffset;

					bool ESPRelative = (!(UsedFramePointer || CurrInst->HasFPNormalizedToSP()));
					if (SignedOffset < 0) { // offset has zero-based index into negative offset vectors
						bool OutOfRange = ((offset + DataSize - 1) >= this->NegativeOffsetStackFrameMap.size());
						if (OutOfRange) {
							if (IndexedAccess)
								continue; // Indexed expressions can be within frame even when offset is outside frame
							else if (this->CallsAlloca) {
								SMP_msg("ERROR: Stack access out of range at %llx due to alloca issues. ", (uint64_t)InstAddr);
								this->DumpFuncNameAndAddr();
								this->AnalyzedSP = false;
								this->OutgoingArgsSize = 0;
								this->MinStackDelta = 0;
								this->AllocPointDelta = 0;
								return;
							}
						}
						if ((offset + DataSize - 1) >= this->NegativeOffsetStackFrameMap.size()) {
							// We are about to fail an assertion. Emit some useful debugging info.
							SMP_msg("FATAL ERROR: Stack access: (offset+DataSize-1) out of range: offset %llu DataSize %zu NegativeOffsetStackFrameMap size %zu InstAddr %llx ",
								(uint64_t)offset, DataSize, this->NegativeOffsetStackFrameMap.size(), (uint64_t)InstAddr);
							this->DumpFuncNameAndAddr();
							CurrInst->Dump();
							this->DumpDotCFG();
							this->Dump(false);
							this->GoodLocalVarTable = false; // See if we can do other analyses, like DEADREGS annotations, instead of bailing out.
							return; 
						}
						assert((offset + DataSize - 1) < this->NegativeOffsetStackFrameMap.size());
						for (int j = 0; j < (int) DataSize; ++j) {
							this->NegativeOffsetStackFrameMap[offset + j].Read = true;
							this->NegativeOffsetStackFrameMap[offset + j].IndexedAccess = IndexedAccess;
							if (ESPRelative) {
								this->NegativeOffsetStackFrameMap[offset + j].ESPRelativeAccess = true;
							}
							else {
								this->NegativeOffsetStackFrameMap[offset + j].EBPRelativeAccess = true;
							}
						}
					}
					else {
						assert((SignedOffset + DataSize - 1) < this->PositiveOffsetStackFrameMap.size());
						for (int j = 0; j < (int)DataSize; ++j) {
							this->PositiveOffsetStackFrameMap[SignedOffset + j].Read = true;
							this->PositiveOffsetStackFrameMap[SignedOffset + j].IndexedAccess = IndexedAccess;
							if (ESPRelative) {
								this->PositiveOffsetStackFrameMap[SignedOffset + j].ESPRelativeAccess = true;
							}
							else {
								this->PositiveOffsetStackFrameMap[SignedOffset + j].EBPRelativeAccess = true;
							}
						}
					}
					struct FineGrainedInfo StackUseFG;
					BitWidthMask = ComputeOperandBitWidthMask(TempOp, DataSize);
					StackUseFG.SizeInfo = BitWidthMask;
					StackUseFG.SignMiscInfo = FG_MASK_READ;
					if (SignedOffset < 0) { // offset has zero-based index into negative offset vectors
						assert(offset < this->NegativeOffsetFineGrainedStackTable.size());
						this->NegativeOffsetFineGrainedStackTable[offset].SizeInfo |= BitWidthMask;
						this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_READ;
						if (IndexedAccess) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
							StackUseFG.SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
						}
						if (ESPRelative) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_SP_RELATIVE;
							StackUseFG.SignMiscInfo |= FG_MASK_SP_RELATIVE;
						}
						else {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_FP_RELATIVE;
							StackUseFG.SignMiscInfo |= FG_MASK_FP_RELATIVE;
						}
						if (SignedMove) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_SIGNED;
							StackUseFG.SignMiscInfo |= FG_MASK_SIGNED;
						}
						else if (UnsignedMove) {
							this->NegativeOffsetFineGrainedStackTable[offset].SignMiscInfo |= FG_MASK_UNSIGNED;
							StackUseFG.SignMiscInfo |= FG_MASK_UNSIGNED;
						}
					}
					else { // SignedOffset >= 0
						assert(SignedOffset < (decltype(SignedOffset))this->PositiveOffsetFineGrainedStackTable.size());
						this->PositiveOffsetFineGrainedStackTable[SignedOffset].SizeInfo |= BitWidthMask;
						this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_READ;
						if (IndexedAccess) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
							StackUseFG.SignMiscInfo |= FG_MASK_INDEXED_ACCESS;
						}
						if (ESPRelative) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_SP_RELATIVE;
							StackUseFG.SignMiscInfo |= FG_MASK_SP_RELATIVE;
						}
						else {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_FP_RELATIVE;
							StackUseFG.SignMiscInfo |= FG_MASK_FP_RELATIVE;
						}
						if (SignedMove) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_SIGNED;
							StackUseFG.SignMiscInfo |= FG_MASK_SIGNED;
						}
						else if (UnsignedMove) {
							this->PositiveOffsetFineGrainedStackTable[SignedOffset].SignMiscInfo |= FG_MASK_UNSIGNED;
							StackUseFG.SignMiscInfo |= FG_MASK_UNSIGNED;
						}
					}
					// Insert the StackUseFG into the map of InstAddr to USE FG info.
					pair<map<STARS_ea_t, struct FineGrainedInfo>::iterator, bool> InsertResult;
					pair<STARS_ea_t, struct FineGrainedInfo> InsertValue(InstAddr, StackUseFG);
					InsertResult = this->StackUseFGInfo.insert(InsertValue);
					assert(InsertResult.second);
				} // end if MDGetStackOffsetAndSize()
			} // end for all USEs
		} // end if SourceMemoryOperand
		// NOTE: Detect taking the address of stack locations. **!!**
	} // end for all instructions

	this->GoodLocalVarTable = true;
	this->OutgoingArgsComputed = true;

	// If function is a leaf function, set OutgoingArgsSize to zero and return.
	// If function has no local frame allocation, ditto.
	if ((this->IsLeaf() && !(this->IsDirectlyRecursive())) || (this->AllocPointDelta == 0)) {
		this->OutgoingArgsSize = 0;
		if (0 == this->AllocPointDelta)
			return;
	}
	else {

		// For non-leaf functions, set the OutgoingArgsSize to the write-only, ESP-relative
		//  region of the bottom of the NegativeOffsetStackFrameMap.
		bool OutgoingArgsRegionFinished = false;
		bool IndexedOutgoingArgs = false; // Any indexed accesses to outgoing args?
		std::size_t FramePadSize = 0;
		std::size_t AlignmentPadSize = 0; // bottom of frame, unused space below outargs
		for (std::size_t MapIndex = 0; MapIndex < this->NegativeOffsetStackFrameMap.size(); ++MapIndex) {
			// Some of the bottom of the stack frame might be below the local frame allocation.
			//  These are pushes that happened after allocation, etc. We skip over these
			//  locations and define the outgoing args region to start strictly at the bottom
			//  of the local frame allocation.
			struct StackFrameEntry TempEntry = this->NegativeOffsetStackFrameMap[MapIndex];
#if 0
			if (DebugFlag) {
				SMP_msg("StackFrameMap entry %zu: offset: %ld Read: %d Written: %d ESP: %d EBP: %d\n",
					MapIndex, TempEntry.offset, TempEntry.Read, TempEntry.Written,
					TempEntry.ESPRelativeAccess, TempEntry.EBPRelativeAccess);
			}
#endif
			if (TempEntry.offset < this->AllocPointDelta)
				continue; // don't look below the stack frame allocation region
			if (STARS_STACK_CALLEE_SAVED_REG == this->NegativeOffsetStackFrameMap[MapIndex].EntryType)
				break;

			if (OutgoingArgsRegionFinished) {
				// We are just processing the stack frame padding.
				if (!TempEntry.Read && !TempEntry.Written) {
					// Could be stack frame padding.
					++FramePadSize;
				}
				else {
					break; // No more padding region
				}
			}
			else if ((this->OutgoingArgsSize == 0) && (!TempEntry.Read) && (!TempEntry.Written)) {
				// We have not started accumulating outgoing args bytes, we have reached the
				//  AllocPointDelta, yet we find space that is neither written nor read. This
				//  empty space at the bottom of the stack frame could just be for stack alignment
				//  purposes, especially in the new x86-64 ABI, so it should not prevent us from
				//  finding outgoing args space above it.
				++AlignmentPadSize;
			}
			else if (TempEntry.Read || TempEntry.EBPRelativeAccess || !TempEntry.Written
				|| !TempEntry.ESPRelativeAccess) {
				OutgoingArgsRegionFinished = true;
				if (!TempEntry.Read && !TempEntry.Written) {
					// Could be stack frame padding.
					++FramePadSize;
				}
				else {
					break; // No padding region
				}
			}
			else {
				this->OutgoingArgsSize++;
				if (TempEntry.IndexedAccess) {
					IndexedOutgoingArgs = true;
				}
			}
		}

		// Add in the alignment padding below the written outargs region.
		if (this->OutgoingArgsSize > 0) {
			this->OutgoingArgsSize += AlignmentPadSize;
		}

		// If any outgoing arg was accessed using an index register, then we don't know how high
		//  the index register value went. It could potentially consume the so-called padding
		//  region, which might be just the region we did not detect direct accesses to because
		//  the accesses were indirect. To be safe, we expand the outgoing args region to fill
		//  the padding region above it in this indexed access case.
		if (IndexedOutgoingArgs) {
			this->OutgoingArgsSize += FramePadSize;
		}

		// Mark the outgoing args entries in the map.
		// Could have stack accesses below stack frame allocation. Don't go below the frame allocation for outargs
		//  !!!!****!!!! Might need to change for Windows and other stack frame conventions.
		int AllocIndex = ((int) this->NegativeOffsetStackFrameMap.size()) + (int) this->AllocPointDelta;
		// e.g. if 20 bytes in map but -16 is AllocPointDelta, then bottom four bytes are below the frame.
		assert(0 <= AllocIndex);
		for (size_t index = (size_t) AllocIndex; index < (AllocIndex + this->OutgoingArgsSize); ++index) {
			this->NegativeOffsetStackFrameMap[index].EntryType = STARS_STACK_OUTARG;
		}
	}

	// Mark the local vars entries in the map.
	bool LocalFrameMarked = false;
	int AllocIndex = ((int) this->NegativeOffsetStackFrameMap.size()) + (int) this->AllocPointDelta;
	// e.g. if 20 bytes in map but -16 is AllocPointDelta, then bottom four bytes are below the frame.
	assert(0 <= AllocIndex);
	for (size_t index = AllocIndex + this->OutgoingArgsSize; index < this->NegativeOffsetStackFrameMap.size(); ++index) {
		if (this->NegativeOffsetStackFrameMap[index].EntryType == STARS_STACK_UNKNOWN) {
			this->NegativeOffsetStackFrameMap[index].EntryType = STARS_STACK_LOCAL_FRAME;
			LocalFrameMarked = true;
		}
		else if (LocalFrameMarked) {
			// Marked some local frame entries, and now we are running up into something besides the local frame,
			//  like callee-saved regs. We don't want to go above them and mark a second local frame area, which might
			//  be a save from memory, for example: push [ecx+4] ; push ebx; push edx; sub esp,32 should be treated as
			//  4 unknown bytes at top, 8 callee-saved reg bytes, then 32 local frame bytes. We don't want to keep
			//  processing and mark the 4 unknown bytes as another local var frame.
			break;
		}
	}

	return;

} // end of SMPFunction::BuildStackAccessTables()

// Update MinStackAccessOffset and MaxStackAccessLimit if TempOp is stack access
void SMPFunction::UpdateMinMaxStackOffsets(SMPInstr *CurrInst, const STARSOpndTypePtr &TempOp) {
	STARS_ea_t offset;
	std::size_t DataSize;
	bool UsedFramePointer;
	bool IndexedAccess;
	bool SignedMove;
	bool UnsignedMove;

	if (this->MDGetStackOffsetAndSize(CurrInst, TempOp, this->MinStackDelta, offset, DataSize, UsedFramePointer,
		IndexedAccess, SignedMove, UnsignedMove)) {
		int SignedOffset = (int) offset + (int) this->MinStackDelta; // Don't want zero-based for min/max finding
		if (((STARS_sval_t) SignedOffset) < this->MinStackAccessOffset) {
			this->MinStackAccessOffset = (STARS_sval_t) SignedOffset;
		}
		if (((STARS_sval_t)(SignedOffset + (int) DataSize)) > this->MaxStackAccessLimit) {
			this->MaxStackAccessLimit = (STARS_sval_t)(SignedOffset + (int) DataSize);
		}
	}
	return;
} // end of SMPFunction::UpdateMinMaxStackOffsets()


// If TempOp reads or writes to a stack location, return the offset (relative to the initial
//  stack pointer value) and the size in bytes of the data access. Also return whether the
//  access was frame-pointer-relative, and whether signedness can be inferred due to a load
//  from the stack being zero-extended or sign-extended.
// NOTE: This function assumes that offsets are already normalized. i.e. the TempOp argument
//  should always come from a DEF or USE that has been normalized to the stack delta at function entry.
// NOTE: TempOp must be of type o_displ or o_phrase, as no other operand type could be a
//  stack memory access.
// BaseValue is either this->MinStackAccessOffset, or this->MinStackDelta (when this->MinStackAccessOffset is still
//  being computed).
// Return true if a stack memory access was found in TempOp, false otherwise.
bool SMPFunction::MDGetStackOffsetAndSize(SMPInstr *Instr, const STARSOpndTypePtr &TempOp, STARS_sval_t BaseValue, STARS_ea_t &offset, std::size_t &DataSize, bool &FP,
										  bool &Indexed, bool &Signed, bool &Unsigned) {
	int BaseReg;
	int IndexReg;
	uint16_t ScaleFactor;
	int SignedOffset;
	STARS_sval_t sp_delta = Instr->GetStackPtrOffset();
	STARS_ea_t InstAddr = Instr->GetAddr(); // helps debugging

	if (nullptr == TempOp)
		return false;
	if (!((TempOp->IsMemNoDisplacementOp()) || (TempOp->IsMemDisplacementOp())))
		return false;
	MDExtractAddressFields(TempOp, BaseReg, IndexReg, ScaleFactor, offset);

	if (TempOp->IsMemNoDisplacementOp()) {
		assert(offset == 0);  // implicit zero, as in [esp] ==> [esp+0]
	}

	SignedOffset = (int) offset;  // avoid sign errors during adjustment arithmetic

	if ((BaseReg == STARS_x86_R_sp) || (IndexReg == STARS_x86_R_sp)) {
		// ESP-relative constant offset
		if (!Instr->AreDefsNormalized()) {
			SignedOffset += sp_delta; // base offsets from entry ESP value
		}
		SignedOffset -= BaseValue; // convert to StackFrameMap index
		offset = (STARS_ea_t) SignedOffset; // write back to outgoing argument
		// Get size of data written
		DataSize = TempOp->GetByteWidth();
		FP = false;
		Indexed = ((BaseReg != STARS_x86_R_none) && (IndexReg != STARS_x86_R_none)); // two regs used
		unsigned short opcode = Instr->GetIDAOpcode();
		Unsigned = (opcode == STARS_NN_movzx);
		Signed = (opcode == STARS_NN_movsx);
		if ((0 > SignedOffset) && (!Indexed) && (BaseValue == this->MinStackAccessOffset) && (!this->DoesStackFrameExtendPastStackTop())) {
			if (this->GoodLocalVarTable) { // No phase ordering problem; should not happen.
				// Consider asserting here.
				SMP_msg("ERROR: Negative offset in MDGetStackOffsetAndSize at %llx for inst dump: ", (uint64_t)Instr->GetAddr());
				this->DumpFuncNameAndAddr();
				Instr->Dump();
			}
			else {
				// BuildStackAccessTables() calls UpdateMinMaxStackOffsets() which calls
				//  the current method, and also calls the current method directly.
				//  DoesStackFrameExtendPastStackTop() will be false until that phase completes, so not
				//  an error. We can set it as soon as we see it is needed.
				this->SetStackFrameExtendsPastStackTop();
			}
		}
		return true;
	}
	else if (this->UseFP && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG))) {
		SignedOffset += this->GetFramePtrStackDelta(); // base offsets from entry ESP value, e.g. ebp-4 == esp-8
		SignedOffset -= BaseValue; // convert to StackFrameMap index
		offset = (STARS_ea_t) SignedOffset;
		DataSize = TempOp->GetByteWidth();
		FP = true;
		Indexed = ((BaseReg != STARS_x86_R_none) && (IndexReg != STARS_x86_R_none)); // two regs used
#if 0
		assert(Indexed || (!this->StackPtrAnalysisSucceeded()) || !this->HasSTARSStackPtrAnalysisCompleted()); // Else we should never get here with unnormalized stack operands
#else
		if (!(Indexed || (!this->StackPtrAnalysisSucceeded()) || !this->HasSTARSStackPtrAnalysisCompleted())) {
			SMP_msg("WARNING: Unnormalized FP-relative stack offset at %llx after stack analysis succeeded.\n",
				(uint64_t) Instr->GetAddr());
		}
#endif
		unsigned short opcode = Instr->GetIDAOpcode();
		Unsigned = (opcode == STARS_NN_movzx);
		Signed = (opcode == STARS_NN_movsx);
		if ((0 > SignedOffset) && (!Indexed) && (BaseValue == this->MinStackAccessOffset)) {
			// Consider asserting here.
			SMP_msg("ERROR: Negative offset %d in MDGetStackOffsetAndSize: frregs: %d MinStackDelta: %ld Inst dump: \n",
				SignedOffset, this->CalleeSavedRegsSize, (long) this->MinStackDelta);
			Instr->Dump();
		}
		return true;
	}
	else {
		return false;
	}
} // end of SMPFunction::MDGetStackOffsetAndSize()
		
// Return fine grained stack entry for stack op TempOp from instruction at InstAddr
bool SMPFunction::MDGetFGStackLocInfo(STARS_ea_t InstAddr, const STARSOpndTypePtr &TempOp, struct FineGrainedInfo &FGEntry) {
	int BaseReg;
	int IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	int SignedOffset;

	assert((TempOp->IsMemNoDisplacementOp()) || (TempOp->IsMemDisplacementOp()));
	MDExtractAddressFields(TempOp, BaseReg, IndexReg, ScaleFactor, offset);

	SignedOffset = (int) offset;

	if (TempOp->IsMemNoDisplacementOp()) {
		assert(SignedOffset == 0);  // implicit zero, as in [esp] ==> [esp+0]
	}
	if ((BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG)) {
		// ESP-relative constant offset
		SignedOffset -= this->MinStackAccessOffset; // convert to StackFrameMap index
	}
	else if (this->UseFP && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG))) {
		assert(false); // should never get here with unnormalized stack operand
		SignedOffset += this->GetFramePtrStackDelta(); // base offsets from entry ESP value
		SignedOffset -= this->MinStackAccessOffset; // convert to StackFrameMap index
	}
	else {
		return false;
	}
	// We did not return false, so we should have a good offset. Use it to
	//  pass back the fine grained stack table entry for that offset.
	if (SignedOffset >= (int) this->NegativeOffsetFineGrainedStackTable.size()) {
		SignedOffset -= (int) this->NegativeOffsetFineGrainedStackTable.size();
		assert(SignedOffset < ((int) this->PositiveOffsetFineGrainedStackTable.size()));
		FGEntry = this->PositiveOffsetFineGrainedStackTable[(std::size_t) SignedOffset];
	}
	else {
		assert(SignedOffset >= 0);
		FGEntry = this->NegativeOffsetFineGrainedStackTable[(std::size_t) SignedOffset];
	}
	return true;
} // end of SMPFunction::MDGetFGStackLocInfo()

// Return true if we update fine grained stack entry for stack op TempOp from instruction at InstAddr
bool SMPFunction::MDUpdateFGStackLocInfo(STARS_ea_t InstAddr, const STARSOpndTypePtr &TempOp, struct FineGrainedInfo NewFG) {
	int BaseReg;
	int IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	int SignedOffset;
	struct FineGrainedInfo OldFG, UnionFG;

	assert((TempOp->IsMemNoDisplacementOp()) || (TempOp->IsMemDisplacementOp()));
	MDExtractAddressFields(TempOp, BaseReg, IndexReg, ScaleFactor, offset);

	SignedOffset = (int) offset;

	if (TempOp->IsMemNoDisplacementOp()) {
		assert(SignedOffset == 0);  // implicit zero, as in [esp] ==> [esp+0]
	}
	if ((BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG)) {
		// ESP-relative constant offset
		SignedOffset -= this->MinStackAccessOffset; // convert to StackFrameMap index
	}
	else if (this->UseFP && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG))) {
		assert(false); // should never get here with unnormalized stack operands
		SignedOffset += this->GetFramePtrStackDelta(); // base offsets from entry ESP value
		SignedOffset -= this->MinStackAccessOffset; // convert to StackFrameMap index
	}
	else {
		return false;
	}
	// We did not return false, so we should have a good offset. Use it to
	//  retrieve the fine grained stack table entry for that offset.
	if (SignedOffset >= (int) this->NegativeOffsetFineGrainedStackTable.size()) { // in Positive offset table
		SignedOffset -= (int) this->NegativeOffsetFineGrainedStackTable.size();
		assert(SignedOffset < ((int) this->PositiveOffsetFineGrainedStackTable.size()));
		OldFG = this->PositiveOffsetFineGrainedStackTable[(size_t) SignedOffset];
		UnionFG.SignMiscInfo = OldFG.SignMiscInfo | NewFG.SignMiscInfo;
		UnionFG.SizeInfo = OldFG.SizeInfo | NewFG.SizeInfo;
		if ((OldFG.SignMiscInfo != UnionFG.SignMiscInfo) || (OldFG.SizeInfo != UnionFG.SizeInfo)) {
			// The signs they are a-changin'. Or maybe the sizes.
			this->PositiveOffsetFineGrainedStackTable[(size_t)SignedOffset].SignMiscInfo |= NewFG.SignMiscInfo;
			this->PositiveOffsetFineGrainedStackTable[(size_t)SignedOffset].SizeInfo |= NewFG.SizeInfo;
		}
	}
	else if (!this->DoesStackFrameExtendPastStackTop() && this->OutgoingArgsComputed && (((std::size_t) SignedOffset) < this->OutgoingArgsSize)) {
		// We don't want to update the outgoing args region, as it will not be consistent
		//  over multiple function calls. NOTE: We could fine tune this by seeing if we
		//  call multiple target functions or not; if only one, then outgoing args region
		//  would be consistent in the absence of varargs targets.
		return false;
	}
	else { // local var access in NegativeOffsetFineGrainedStackTable
		OldFG = this->NegativeOffsetFineGrainedStackTable[(size_t)SignedOffset];
		UnionFG.SignMiscInfo = OldFG.SignMiscInfo | NewFG.SignMiscInfo;
		UnionFG.SizeInfo = OldFG.SizeInfo | NewFG.SizeInfo;
		if ((OldFG.SignMiscInfo != UnionFG.SignMiscInfo) || (OldFG.SizeInfo != UnionFG.SizeInfo)) {
			// The signs they are a-changin'. Or maybe the sizes.
			this->NegativeOffsetFineGrainedStackTable[(size_t)SignedOffset].SignMiscInfo |= NewFG.SignMiscInfo;
			this->NegativeOffsetFineGrainedStackTable[(size_t)SignedOffset].SizeInfo |= NewFG.SizeInfo;
		}
	}
	return true;
} // end of SMPFunction::MDUpdateFGStackLocInfo()

// retrieve from GlobalDefAddrBySSA or return BADADDR
STARS_ea_t SMPFunction::GetGlobalDefAddrForRegHash(int RegHashValue) {
	STARS_ea_t DefAddr = STARS_BADADDR; // STARS_BADADDR means we did not find it
	map<int, STARS_ea_t>::iterator MapResult = this->GlobalDefAddrBySSA.find(RegHashValue);
	if (MapResult != this->GlobalDefAddrBySSA.end()) { // Found it.
		DefAddr = (STARS_ea_t) MapResult->second;
	}
	return DefAddr;
} // end of SMPFunction::GetGlobalDefAddrForRegHash()

// retrieve DEF addr from GlobalDefAddrBySSA or return STARS_BADADDR
STARS_ea_t SMPFunction::GetGlobalDefAddr(const STARSOpndTypePtr &DefOp, int SSANum) {
	STARS_ea_t DefAddr = STARS_BADADDR; // STARS_BADADDR means we did not find it
	bool RegDef = (DefOp->IsRegOp());

	if (RegDef) {
		int HashedName = HashGlobalNameAndSSA(DefOp, SSANum);
		DefAddr = this->GetGlobalDefAddrForRegHash(HashedName);
	}
	else if (MDIsDirectStackAccessOpnd(DefOp, this->UsesFramePointer())) {
		int64_t HashedName = HashGlobalStackNameAndSSA(DefOp, SSANum, this->UsesFramePointer());
		map<int64_t, STARS_ea_t>::iterator MapResult = this->GlobalStackDefAddrBySSA.find(HashedName);
		if (MapResult != this->GlobalStackDefAddrBySSA.end()) { // Found it.
			DefAddr = (STARS_ea_t) MapResult->second;
		}
	}
	else if (DefOp->IsMemOp()) {
		// Until we get indirect memory operands into the GlobalDefAddrBySSA map,
		//  do a linear search. Search only blocks with indirect memory writes.
		list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
		set<DefOrUse, LessDefUse>::iterator DefIter;
		SMPInstr *CurrInst = (*InstIter);
		if (CurrInst->IsMarkerInst()) {
			if (0 == SSANum) { // Live-in-to-func stack locations get DEF at marker inst.
				DefIter = CurrInst->FindDef(DefOp);
				if (DefIter != CurrInst->GetLastDef()) {
					// Found it. Must be SSA 0.
					assert(0 == DefIter->GetSSANum());
					DefAddr = CurrInst->GetAddr();
					return DefAddr;
				}
			}
			++InstIter;
		}

		for (size_t BlockNum = 0; BlockNum < this->GetNumBlocks(); ++BlockNum) {
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
			if (CurrBlock->HasIndirectMemWrite()) {
				vector<SMPInstr *>::const_iterator InstVecIter;
				for (InstVecIter = CurrBlock->GetFirstConstInst(); InstVecIter != CurrBlock->GetLastConstInst(); ++InstVecIter) {
					CurrInst = (*InstIter);
					if (CurrInst->HasDestMemoryOperand()) {
						STARSOpndTypePtr MemDefOp = CurrInst->GetMemDef();
						if (IsEqOp(DefOp, MemDefOp)) {
							DefIter = CurrInst->FindDef(MemDefOp);
							assert(DefIter != CurrInst->GetLastDef());
							int DefSSANum = DefIter->GetSSANum();
							if (DefSSANum == SSANum) { // found it
								DefAddr = CurrInst->GetAddr();
								break;
							}
						}
					}
					else if (CurrInst->MDIsPushInstr()) { // merge if and else branches here ???
						DefIter = CurrInst->FindDef(DefOp);
						if (DefIter != CurrInst->GetLastDef()) {
							int DefSSANum = DefIter->GetSSANum();
							if (DefSSANum == SSANum) { // found it
								DefAddr = CurrInst->GetAddr();
								break;
							}
						}
					}
				} // end for all insts in current block
			}
		} // end for all blocks
#if 0 // speed up
		for ( ; InstIter != this->Instrs.end(); ++InstIter) {
			CurrInst = (*InstIter);
			if (CurrInst->HasDestMemoryOperand()) {
				STARSOpndTypePtr MemDefOp = CurrInst->GetMemDef();
				if (IsEqOp(DefOp, MemDefOp)) {
					DefIter = CurrInst->FindDef(MemDefOp);
					assert(DefIter != CurrInst->GetLastDef());
					int DefSSANum = DefIter->GetSSANum();
					if (DefSSANum == SSANum) { // found it
						DefAddr = CurrInst->GetAddr();
						break;
					}
				}
			}
			else if (CurrInst->MDIsPushInstr()) { // merge if and else branches here ???
				DefIter = CurrInst->FindDef(DefOp);
				if (DefIter != CurrInst->GetLastDef()) {
					int DefSSANum = DefIter->GetSSANum();
					if (DefSSANum == SSANum) { // found it
						DefAddr = CurrInst->GetAddr();
						break;
					}
				}
			}
		}
#endif
	}
	return DefAddr;
} // end of SMPFunction::GetGlobalDefAddr()

// Track global DefOp+SSANum to its DEF block, return the inst iter within that block.
vector<SMPInstr *>::iterator SMPFunction::GetBlockInstIterBySSA(const STARSOpndTypePtr &DefOp, int SSANum) {
	assert(this->IsGlobalName(DefOp));
	STARS_ea_t DefAddr = this->GetGlobalDefAddr(DefOp, SSANum);
	assert(STARS_BADADDR != DefAddr);
	if (STARS_IsBlockNumPseudoID(DefAddr)) {
		size_t BlockNum = STARS_GetBlockNumFromPseudoID(DefAddr);
		SMPBasicBlock *PhiBlock = this->GetBlockByNum(BlockNum);
		return PhiBlock->GetFirstInst();
	}
	else {
		SMPBasicBlock *DefBlock = this->GetBlockFromInstAddr(DefAddr);
		return DefBlock->GetInstIterFromAddr(DefAddr);
	}
} // end of SMPFunction::GetBlockInstIterBySSA()

int SMPFunction::GetBlockNumForPhiDef(const STARSOpndTypePtr &DefOp, int SSANum) {
	std::size_t BlockIndex;
	for (BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
		set<SMPPhiFunction, LessPhi>::iterator PhiIter = CurrBlock->FindPhi(DefOp);
		if (PhiIter != CurrBlock->GetLastPhi()) {
			if (PhiIter->GetDefSSANum() == SSANum) {
				return CurrBlock->GetNumber();
			}
		}
	}
	return (int) STARS_BADADDR;
} // end of SMPFunction::GetBlockNumForPhiDef()

// Retrieve block iterator for InstAddr from InstBlockMap; assert if failure
SMPBasicBlock *SMPFunction::GetBlockFromInstAddr(STARS_ea_t InstAddr) {
	if (STARS_IsBlockNumPseudoID(InstAddr)) {
		int BlockNum = STARS_GetBlockNumFromPseudoID(InstAddr);
		return this->GetBlockByNum((size_t) BlockNum);
	}
	else {
		map<STARS_ea_t, SMPBasicBlock *>::iterator MapEntry;
		MapEntry = this->InstBlockMap.find(InstAddr);
		assert(MapEntry != this->InstBlockMap.end());
		return MapEntry->second;
	}
} // end of SMPFunction::GetBlockFromInstAddr()

// Retrieve block iterator for block starting with InstAddr from TCFGBlocks; assert if failure
STARSTransCFGBlock *SMPFunction::GetTCFGBlockFromFirstInstAddr(STARS_ea_t InstAddr) const {
	if (STARS_IsBlockNumPseudoID(InstAddr)) {
		int BlockNum = STARS_GetBlockNumFromPseudoID(InstAddr);
		return this->GetTCFGBlockByNum((size_t)BlockNum);
	}
	else { // linear search of TCFGBlocks
		for (STARSTransCFGBlock *TCFGBlock : this->TCFGBlocks) {
			if (TCFGBlock->GetFirstNonMarkerAddr() == InstAddr) {
				return TCFGBlock;
			}
		}
		assert(false);
	}
} // end of SMPFunction::GetTCFGBlockFromInstAddr()

// -1 or -2 InstAddr is returned as follow block; >=0 InstAddr is converted to TCFG block num
int SMPFunction::GetTCFGBlockNumFromFirstInstAddr(STARS_ea_t InstAddr) const {
	int BlockNum = (int)InstAddr;
	if (0 <= BlockNum) {
		BlockNum = this->GetTCFGBlockFromFirstInstAddr(InstAddr)->GetNumber();
	}
	return BlockNum;
} // end of SMPFunction::GetTCFGBlockNumFromFirstInstAddr()

// return -1 if not in InstBlockMap, block # otherwise
int SMPFunction::GetBlockNumFromInstAddr(STARS_ea_t InstAddr) {
	map<STARS_ea_t, SMPBasicBlock *>::iterator MapEntry = this->InstBlockMap.find(InstAddr);
	if (MapEntry != this->InstBlockMap.end()) {
		return MapEntry->second->GetNumber();
	}
	else {
		return SMP_BLOCKNUM_UNINIT;
	}
}

// Find FindBlock in Blocks, return iterator.
list<SMPBasicBlock *>::iterator SMPFunction::GetBlockIter(SMPBasicBlock *FindBlock) {
	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		if (FindBlock == (*BlockIter)) {
			break;
		}
	}
	return BlockIter;
}

// <0 maps to itself; real block nums map to FirstAddr
STARS_ea_t SMPFunction::GetFirstAddrFromTCFGFollowBlockNum(const int TCFGFollowBlockNum) const {
	STARS_ea_t FirstAddr = (STARS_ea_t)TCFGFollowBlockNum; // map -2 and -1 to themselves
	if (0 <= TCFGFollowBlockNum) {
		FirstAddr = this->GetTCFGBlockByNum(TCFGFollowBlockNum)->GetFirstAddr();
	}
	return FirstAddr;
} // end of SMPFunction::GetFirstAddrFromTCFGFollowBlockNum()

// Retrieve inst pointer for InstAddr; assert if failure on block find.
SMPInstr *SMPFunction::GetInstFromAddr(STARS_ea_t InstAddr) {
	SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(InstAddr);
	SMPInstr *CurrInst = CurrBlock->FindInstr(InstAddr);
	return CurrInst;
}

int SMPFunction::GetLoopNumFromTestBlockNum(int BlockNum) const {
	int LoopNum = -1;
	for (std::size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		if (BlockNum == this->LoopTestBlocksByLoopNum[LoopIndex]) {
			LoopNum = (int) LoopIndex;
			break;
		}
	}
	return LoopNum;
}

// find innermost loop # for BlockNum
int SMPFunction::GetInnermostLoopNum(const int BlockNum) const {
	int LoopNum = -1;
	size_t MinBlockCount = this->GetNumBlocks() + 1;
	bool FuncHasLoops = (this->LoopCount > 0);
	if ((BlockNum >= 0) && FuncHasLoops && this->IsBlockInAnyLoop(BlockNum)) {
		// Find loop containing this BlockNum with the fewest number of blocks.
		for (size_t LoopIndex = 0; LoopIndex < this->FuncLoopsByBlock[(size_t) BlockNum].GetNumBits(); ++LoopIndex) {
			if (this->FuncLoopsByBlock[(size_t) BlockNum].GetBit(LoopIndex)) { // BlockNum is in loop #LoopIndex
				size_t CurrBlockCount = this->FuncBlocksByLoop[LoopIndex].CountSetBits();
				if (CurrBlockCount < MinBlockCount) {
					MinBlockCount = CurrBlockCount;
					LoopNum = (int) LoopIndex;
				}
			}
		}
	}
	return LoopNum;
} // end of SMPFunction::GetInnermostLoopNum()

// find outermost loop # for BlockNum
int SMPFunction::GetOutermostLoopNum(const int BlockNum) const {
	int LoopNum = -1;
	size_t MaxBlockCount = 0;
	if ((BlockNum >= 0) && (this->LoopCount > 0) && this->IsBlockInAnyLoop(BlockNum)) {
		// Find loop containing this BlockNum with the fewest number of blocks.
		for (size_t LoopIndex = 0; LoopIndex < this->FuncLoopsByBlock[(size_t) BlockNum].GetNumBits(); ++LoopIndex) {
			if (this->FuncLoopsByBlock[(size_t) BlockNum].GetBit(LoopIndex)) {
				size_t CurrBlockCount = this->FuncBlocksByLoop[LoopIndex].CountSetBits();
				if (CurrBlockCount > MaxBlockCount) {
					MaxBlockCount = CurrBlockCount;
					LoopNum = (int) LoopIndex;
				}
			}
		}
	}
	return LoopNum;
}

// Find innermost loop # whose head block dominates BlockNum but does not contain it
int SMPFunction::GetInnermostDominatingLoopNum(const int BlockNum) {
	int InnerDominatingLoopNum = -1;
	int MaxRPOHeadBlockNum = -1;
	for (size_t LoopNum = 0; LoopNum < this->GetNumLoops(); ++LoopNum) {
		int HeadBlockNum = this->LoopHeadBlockNumbers[LoopNum];
		this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		if (this->DoesBlockDominateBlock(HeadBlockNum, BlockNum)) {
			if (!this->IsBlockInLoop(BlockNum, LoopNum)) {
				if (HeadBlockNum > MaxRPOHeadBlockNum) {
					MaxRPOHeadBlockNum = HeadBlockNum;
					InnerDominatingLoopNum = (int)LoopNum;
				}
			}
		}
	}

	return InnerDominatingLoopNum;
} // end of SMPFunction::GetInnerMostDominatingLoopNum()

// Given block # and PhiDef STARSOpndType and SSANum, return the Phi iterator or assert.
set<SMPPhiFunction, LessPhi>::iterator SMPFunction::GetPhiIterForPhiDef(std::size_t BlockNumber, const STARSOpndTypePtr &DefOp, int SSANum) {
	SMPBasicBlock *DefBlock = this->RPOBlocks[BlockNumber];
	set<SMPPhiFunction, LessPhi>::iterator PhiIter = DefBlock->FindPhi(DefOp);
	assert(PhiIter != DefBlock->GetLastPhi());
	return PhiIter;
}

// push_back new entries on switch data structures
bool SMPFunction::AddSwitchTableInfo(STARS_ea_t IndirJumpAddr, struct SwitchTableInfo TableInfo) {
	bool success = true;
	size_t SwitchIndex = this->GetSwitchJumpMapSize(); // first index is zero, etc.
	assert(this->GetSwitchInfoArraySize() == SwitchIndex);  // data structure sizes must always be in sync
	pair<STARS_ea_t, size_t> JumpMapItem(IndirJumpAddr, SwitchIndex);
	pair<map<STARS_ea_t, size_t>::iterator, bool> InsertResult = this->SwitchJumpMap.insert(JumpMapItem);
	assert(InsertResult.second); // should never insert twice for the same IndirJumpAddr
	// Now we assert that each field of TableInfo is sane.
	assert(TableInfo.FollowNodeNum != SMP_BLOCKNUM_UNINIT);
	assert(!TableInfo.CaseBlockNums.empty());
	assert(TableInfo.IndexValue.size() == TableInfo.CaseBlockNums.size());
	this->SwitchInfoArray.push_back(TableInfo);
	return success;
} // end of SMPFunction::AddSwitchTableInfo()

// Is DestOp within the outgoing args area? Assume it must be an ESP-relative
//  DEF operand in order to be a write to the outgoing args area.
// NOTE: DestOp should be already normalized to the entry stack delta.
bool SMPFunction::IsInOutgoingArgsRegion(const STARSOpndTypePtr &DestOp) {
	bool OutArgWrite = false;
	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;

	if (this->IsLeaf())
		return false;

	MDExtractAddressFields(DestOp, BaseReg, IndexReg, ScaleFactor, offset);
	if ((BaseReg != STARS_x86_R_sp) && (IndexReg != STARS_x86_R_sp))
		return false;
	if (((BaseReg == STARS_x86_R_sp) && (IndexReg != STARS_x86_R_none))
		|| ((IndexReg == STARS_x86_R_sp) && (BaseReg != STARS_x86_R_none))
		|| (0 < ScaleFactor)) {

#if 0
		SMP_msg("WARNING: WritesToOutgoingArgs called with indexed write.");
		PrintOperand(DestOp);
#endif
		return false;
	}

	if (!this->OutgoingArgsComputed) {
		OutArgWrite = true; // be conservative
	}
	else {
		int SignedOffset = (int) offset;
		SignedOffset -= this->MinStackDelta; // convert to zero-based from bottom of stack frame
		int BelowFrameBytes = this->AllocPointDelta - this->MinStackDelta;
		assert(BelowFrameBytes >= 0);
		int OutArgsLimit = (BelowFrameBytes + this->OutgoingArgsSize);
#if 0
		OutArgWrite = (SignedOffset >= BelowFrameBytes) && (SignedOffset < OutArgsLimit);
#else
		OutArgWrite = (SignedOffset >= 0) && (SignedOffset < OutArgsLimit);
#endif
	}
	return OutArgWrite;
} // end of SMPFunction::IsInOutgoingArgsRegion()

// Does SourceOp from SourceInst fall within incoming args area?
bool SMPFunction::IsInIncomingArgsRegion(SMPInstr *SourceInst, const STARSOpndTypePtr &SourceOp) const {
	bool InArg = MDIsStackAccessOpnd(SourceOp, false) && SourceInst->AreDefsNormalized();
	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;

	if (InArg) {
		MDExtractAddressFields(SourceOp, BaseReg, IndexReg, ScaleFactor, offset);
		int SignedOffset = (int) offset; // will be greater than ISA_ByteWidth if SourceOp is above the return address
		if (((BaseReg != STARS_x86_R_sp) && (IndexReg != STARS_x86_R_sp)) 
			|| (SignedOffset < ((int) global_STARS_program->GetSTARS_ISA_Bytewidth()))) {
			InArg = false;
		}
	}

	return InArg;
} // end of SMPFunction::IsInIncomingArgsRegion()

bool SMPFunction::IsLiveIn(const STARSOpndTypePtr &CurrOp) const {
	return (this->Blocks.front()->IsLiveIn(CurrOp));
}

// Is DestOp a direct memory access above the local vars frame?
bool SMPFunction::WritesAboveLocalFrame(const STARSOpndTypePtr &DestOp, bool OpNormalized, STARS_ea_t InstAddr) {
	bool InArgWrite = false;
	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	long SignedOffset;

	MDExtractAddressFields(DestOp, BaseReg, IndexReg, ScaleFactor, offset);
	SignedOffset = (long) offset;
	bool IndexedWrite = (((IndexReg != STARS_x86_R_none) && (BaseReg != STARS_x86_R_none))
		|| (0 < ScaleFactor));
	bool ESPrelative = (BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG);
	bool EBPrelative = this->UseFP && (0 != offset) && (!IndexedWrite) && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG));
	assert(!EBPrelative || !OpNormalized); // stack operands should be normalized by now
	if (!(ESPrelative || EBPrelative))
		return false;
	if (IndexedWrite) {
		SMP_msg("WARNING: WritesAboveLocalFrame called with indexed write.");
		PrintOperand(DestOp);
		return false;
	}

	// The next statement omits a complication: The possibility that OpNormalized is false,
	//  and an ESPRelative access is above the stack frame. For the purposes of determining
	//  whether a function is safe, this is irrelevant, because !OpNormalized would indicate
	//  that AnalyzedSP is false, which will make the function unsafe anyway. Future uses for
	//  other purposes need to fix this.
	InArgWrite = (ESPrelative && OpNormalized && (SignedOffset >= 0))
		|| (EBPrelative && (SignedOffset > 0));

	if (InArgWrite && OpNormalized && (0 == SignedOffset)) {
		SMP_msg("DANGER: Write to saved return address detected at %llx in function that begins at %llx\n",
			(uint64_t) InstAddr, (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, RETURN_ADDRESS_WRITE);
	}

	return InArgWrite;
}// end of SMPFunction::WritesAboveLocalFrame()

// Is StackOp direct stack access to caller's frame, incoming args on stack, or return address?
bool SMPFunction::AccessAboveLocalFrame(const STARSOpndTypePtr &StackOp, bool OpNormalized, STARS_ea_t InstAddr, bool WriteAccess){
	bool InArgAccess = false;
	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	long SignedOffset;

	MDExtractAddressFields(StackOp, BaseReg, IndexReg, ScaleFactor, offset);
	SignedOffset = (long) offset;
	bool IndexedWrite = (((IndexReg != STARS_x86_R_none) && (BaseReg != STARS_x86_R_none))
		|| (0 < ScaleFactor));
	bool ESPrelative = (BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG);
	bool EBPrelative = this->UseFP && (0 != offset) && (!IndexedWrite) && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG));
	assert(!EBPrelative || !OpNormalized); // stack operands should be normalized by now
	if (!(ESPrelative || EBPrelative))
		return false;
	if (IndexedWrite) {
		SMP_msg("WARNING: AccessAboveLocalFrame called with indexed operand.");
		PrintOperand(StackOp);
		return false;
	}

	// The next statement omits a complication: The possibility that OpNormalized is false,
	//  and an ESPRelative access is above the stack frame. For the purposes of determining
	//  whether a function is safe, this is irrelevant, because !OpNormalized would indicate
	//  that AnalyzedSP is false, which will make the function unsafe anyway. Future uses for
	//  other purposes need to fix this.
	InArgAccess = (ESPrelative && OpNormalized && (SignedOffset >= 0))
		|| (EBPrelative && (SignedOffset > 0));

	if (InArgAccess && OpNormalized && (0 == SignedOffset)) {
		if (WriteAccess) {
			SMP_msg("DANGER: Write to saved return address detected at %llx in function that begins at %llx\n",
				(uint64_t) InstAddr, (uint64_t) this->GetFirstFuncAddr());
			this->SetUnsafeForFastReturns(true, RETURN_ADDRESS_WRITE);
		}
		else {
			SMP_msg("INFO: Read of saved return address detected at %llx in function that begins at %llx\n",
				(uint64_t) InstAddr, (uint64_t) this->GetFirstFuncAddr());
			this->SetUnsafeForFastReturns(true, RETURN_ADDRESS_READ);
		}
	}

	return InArgAccess;
} // end of SMPFunction::AccessAboveLocalFrame()

// Get stack access type from stack frame map
StackAccessType SMPFunction::GetStackAccessType(const STARSOpndTypePtr &StackOp, bool OpNormalized, STARS_ea_t InstAddr, bool WriteAccess) {
	StackAccessType AccessVal = STARS_STACK_UNKNOWN;
	if (this->NegativeOffsetStackFrameMap.empty()) {
		SMP_msg("WARNING: IsAccessToCalleeSavedReg() called with empty stack frame map at %llx\n",
			(uint64_t) InstAddr);
		return AccessVal;
	}

	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	long SignedOffset;

	MDExtractAddressFields(StackOp, BaseReg, IndexReg, ScaleFactor, offset);
	SignedOffset = (long) offset;
	bool ESPrelative = (BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG);
	bool EBPrelative = this->UseFP && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG));
	assert(!EBPrelative || !OpNormalized); // stack operands should be normalized by now
	if (!(ESPrelative || EBPrelative))
		return AccessVal;

	if (((IndexReg != STARS_x86_R_none) && (BaseReg != STARS_x86_R_none)) || (0 < ScaleFactor)) {
		SMP_msg("WARNING: GetStackAccessType() called at %llx with indexed operand.", (uint64_t) InstAddr);
		PrintOperand(StackOp);
		return AccessVal;
	}

	// TODO: Handle unnormalized operands. Not sure why they should exist at this point.
	if (!OpNormalized)
		return AccessVal;

	if (SignedOffset >= 0) {
		size_t TableIndex = (size_t) SignedOffset;
		if (TableIndex < this->PositiveOffsetStackFrameMap.size())
			AccessVal = this->PositiveOffsetStackFrameMap[TableIndex].EntryType;
		if (WriteAccess) {
			if (STARS_STACK_RETURN_ADDRESS == AccessVal)
				SMP_msg("DANGER: WriteAccess to return address at %llx\n", (uint64_t) InstAddr);
			else if (STARS_STACK_INARG == AccessVal)
				SMP_msg("DANGER: WriteAccess to incoming arg at %llx\n", (uint64_t) InstAddr);
		}
	}
	else {
		int TableIndex = ((int) this->NegativeOffsetStackFrameMap.size()) + SignedOffset;
		assert(0 <= TableIndex);
		if (((size_t) TableIndex) < this->NegativeOffsetStackFrameMap.size()) {
			AccessVal = this->NegativeOffsetStackFrameMap[TableIndex].EntryType;
			if (WriteAccess) {
				if (STARS_STACK_CALLEE_SAVED_REG == AccessVal)
					SMP_msg("DANGER: WriteAccess to callee-saved reg at %llx\n", (uint64_t) InstAddr);
			}
		}
	}
	return AccessVal;
} // end of SMPFunction::GetStackAccessType()

// Is StackOp direct stack access to a callee-saved reg?
bool SMPFunction::IsAccessToCalleeSavedReg(const STARSOpndTypePtr &StackOp, bool OpNormalized, STARS_ea_t InstAddr, bool WriteAccess) {
	StackAccessType AccessVal = this->GetStackAccessType(StackOp, OpNormalized, InstAddr, WriteAccess);
	bool SavedRegAccess = (STARS_STACK_CALLEE_SAVED_REG == AccessVal);

	return SavedRegAccess;
} // end of SMPFunction::IsAccessToCalleeSavedReg()


// Is DestOp an indexed write above the local vars frame?
bool SMPFunction::IndexedWritesAboveLocalFrame(const STARSOpndTypePtr &DestOp) {
	bool InArgWrite = false;
	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	int SignedOffset;

	MDExtractAddressFields(DestOp, BaseReg, IndexReg, ScaleFactor, offset);
	bool ESPrelative = (BaseReg == MD_STACK_POINTER_REG) || (IndexReg == MD_STACK_POINTER_REG);
	bool EBPrelative = this->UseFP && ((BaseReg == MD_FRAME_POINTER_REG) || (IndexReg == MD_FRAME_POINTER_REG));
	assert(!EBPrelative || !this->StackPtrAnalysisSucceeded() || !this->HasSTARSStackPtrAnalysisCompleted()); // stack operands should be normalized by now
	if (!(ESPrelative || EBPrelative))
		return false;

	SignedOffset = (int) offset;
	InArgWrite = (ESPrelative && (SignedOffset > 0))
		|| (EBPrelative && (SignedOffset > 0));

	return InArgWrite;
} // end of SMPFunction::IndexedWritesAboveLocalFrame()

// Is CurrOp found anywhere in the StackPtrCopySet, regardless of which address and stack delta
//  values are associated with it?
bool SMPFunction::IsInStackPtrCopySet(const STARSOpndTypePtr &CurrOp) {
	if (nullptr == CurrOp)
		return false;

	bool found = false;
	// Set is composed of triples, so we have to iterate through it and compare operands.
	set<pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> >, LessStackDeltaCopy>::iterator CopyIter;
	for (CopyIter = this->StackPtrCopySet.begin(); CopyIter != this->StackPtrCopySet.end(); ++CopyIter) {
		pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> > CurrCopy = *CopyIter;
		STARSOpndTypePtr CopyOp = CurrCopy.first;
		if (IsEqOp(CopyOp, CurrOp)) {
			// Found it.
			found = true;
			break;
		}
		else if (CopyOp->GetOpType() > CurrOp->GetOpType()) {
			// already moved past its spot; not found
			break;
		}
	}

	return found;
} // end of SMPFunction::IsInStackPtrCopySet()

// Is definition CurrOp at DefAddr in the StackPtrCopySet, with any stack offset associated with it?
bool SMPFunction::IsDefnInStackPtrCopySet(const STARSOpndTypePtr &CurrOp, const STARS_ea_t &DefAddr) const {
	if (nullptr == CurrOp)
		return false;

	bool found = false;
	// Set is composed of triples, so we have to iterate through it and compare operands.
	set<pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> >, LessStackDeltaCopy>::const_iterator CopyIter;
	for (CopyIter = this->StackPtrCopySet.begin(); CopyIter != this->StackPtrCopySet.end(); ++CopyIter) {
		pair<STARSOpndTypePtr, pair<STARS_ea_t, STARS_sval_t> > CurrCopy = *CopyIter;
		STARSOpndTypePtr CopyOp = CurrCopy.first;
		if (IsEqOp(CopyOp, CurrOp)) {
			STARS_ea_t CopyAddr = CurrCopy.second.first;
			if (CopyAddr == DefAddr) {
				// Found it.
				found = true;
				break;
			}
		}
		else if (CopyOp->GetOpType() > CurrOp->GetOpType()) {
			// already moved past its spot; not found
			break;
		}
	}

	return found;
} // end of SMPFunction::IsDefnInStackPtrCopySet()


// Find evidence of calls to alloca(), which appear as stack space allocations (i.e.
//  subtractions [of unknown values(?)] from the stack pointer) AFTER the local frame allocation instruction
//  for this function.
// Return true if such an allocation is found and false otherwise.
bool SMPFunction::FindAlloca(void) {
	bool FoundAlloca = false;
	bool FoundFrameAlloc = false;
	list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
	SMPInstr *CurrInst;
	STARS_ea_t InstAddr;
#if SMP_USE_SSA_FNOP_MARKER
	++InstIter;  // skip marker instruction
#endif
	for ( ; InstIter != this->Instrs.end(); ++InstIter) {
		CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		if (InstAddr == this->LocalVarsAllocInstr)
			FoundFrameAlloc = true;
		else if (FoundFrameAlloc) {
			if (CurrInst->MDIsFrameAllocInstr()) {
				FoundAlloca = true;
				if (CurrInst->HasAllocaRTL()) {
					CurrInst->SetAllocaCall();
				}
			}
			else if (CurrInst->MDIsPushInstr()) {
				this->PushAfterLocalVarAlloc = true;
			}
		}
	}
	return FoundAlloca;
} // end of SMPFunction::FindAlloca()

#if 1
// Emit the annotations describing the regions of the stack frame.
void SMPFunction::EmitStackFrameAnnotations(FILE *AnnotFile, SMPInstr *Instr) {
	STARS_ea_t addr = Instr->GetAddr();

#if 0
	if (0 < IncomingArgsSize) {
		SMP_fprintf(AnnotFile, "%18llx %6llu INARGS STACK esp + %ld %s \n",
			(uint64_t) addr, (uint64_t) IncomingArgsSize,
			(long) (this->GetLocalVarsSize() + CalleeSavedRegsSize + RetAddrSize),
			Instr->GetDisasm());
	}
#endif
	if (0 < this->RetAddrSize) {
		SMP_fprintf(AnnotFile, "%18llx %6d MEMORYHOLE STACK esp + %lu ReturnAddress \n",
			(uint64_t) addr, RetAddrSize, (unsigned long)(this->GetLocalVarsSize() + this->CalleeSavedRegsSize));
	}
	if (this->GoodLocalVarTable && (!this->NegativeOffsetStackFrameMap.empty())) {
		// Walk through the NegativeOffsetStackFrameMap and emit annotations for the regions.
		bool LocalFrameEmitted = false; // have we emitted the LOCALFRAME annotation?
		unsigned long ParentReferentID = 0;
		size_t Limit = this->NegativeOffsetStackFrameMap.size();
		uint32_t CurrentRegionBytes = 0;
		StackAccessType CurrentRegionType = STARS_STACK_RETURN_ADDRESS; // start with previous type
		// Could have stack accesses below stack frame allocation. Don't look below the frame allocation, because
		//  we are emitting these annotations at the frame allocation instruction. !!!!****!!!! Might need to change
		//  for Windows and other stack frame conventions.
		int AllocIndex = ((int) this->NegativeOffsetStackFrameMap.size()) + (int) this->AllocPointDelta;
		// e.g. if 20 bytes in map but -16 is AllocPointDelta, then bottom four bytes are below the frame.
		assert(0 <= AllocIndex);
		for (int index = Limit - 1; index >= AllocIndex; --index) {
			StackAccessType NewType = this->NegativeOffsetStackFrameMap[index].EntryType;
			if (NewType == CurrentRegionType) {
				++CurrentRegionBytes;
			}
			else { // changing regions; time for annotation for previous region
				if (CurrentRegionType == STARS_STACK_CALLEE_SAVED_REG) {
					SMP_fprintf(AnnotFile, "%18llx %6u MEMORYHOLE STACK esp + %lu CalleeSavedRegs \n",
						(uint64_t) addr, CurrentRegionBytes, (unsigned long) (index + 1 - AllocIndex));
				}
				else if (CurrentRegionType == STARS_STACK_LOCAL_FRAME) {
					ParentReferentID = global_STARS_program->GetDataReferentID();
					global_STARS_program->IncrementDataReferentID();
					STARS_asize_t TempRegionBytes = (STARS_asize_t) CurrentRegionBytes;
					STARS_asize_t TempOutArgsSize = (STARS_asize_t) this->OutgoingArgsSize;
					if (this->GetLocalVarsSize() != (TempRegionBytes + TempOutArgsSize)) {
						if (!this->HasPushAfterFrameAlloc()) {
							// Suspect incorrect function boundaries in most of these cases.
							SMP_msg("ERROR: LocalVarsSize: %lu not sum of CurrentRegionBytes: %lu  and OutArgsSize: %lu at %llx ",
								(unsigned long) this->GetLocalVarsSize(), (unsigned long)TempRegionBytes, (unsigned long)TempOutArgsSize, (uint64_t)addr);
							this->DumpFuncNameAndAddr();
#if 1
							this->Dump(false);
							this->DumpDotCFG();
#endif
						}
					}
					SMP_fprintf(AnnotFile, "%18llx %6u DATAREF STACK %lu esp + 0 PARENT LocalFrame LOCALFRAME\n",
						(uint64_t) addr, CurrentRegionBytes + this->OutgoingArgsSize, (unsigned long) ParentReferentID);
					LocalFrameEmitted = true;
				}
				else if (CurrentRegionType == STARS_STACK_OUTARG) {
					if (!LocalFrameEmitted) { // entire local stack frame is outgoing args region; unusual case
						ParentReferentID = global_STARS_program->GetDataReferentID();
						global_STARS_program->IncrementDataReferentID();
						SMP_fprintf(AnnotFile, "%18llx %6u DATAREF STACK %lu esp + 0 PARENT LocalFrame LOCALFRAME\n",
							(uint64_t) addr, CurrentRegionBytes, (unsigned long)ParentReferentID);
						LocalFrameEmitted = true;
					}
					SMP_fprintf(AnnotFile, "%18llx %6zu DATAREF STACK %lu esp + %d CHILDOF %lu OFFSET %d OutArgsRegion OUTARGS\n",
						(uint64_t) addr, CurrentRegionBytes, global_STARS_program->GetDataReferentID(),
						(unsigned long) (index + 1 - AllocIndex), ParentReferentID, 0);
					global_STARS_program->IncrementDataReferentID();
				}
				else if (CurrentRegionType == STARS_STACK_RETURN_ADDRESS) {
					; // just getting started
				}
				else { // something odd is happening
					;
				}
				CurrentRegionType = NewType;
				CurrentRegionBytes = 1;
			}
		} // end for index in range of NegativeOffsetStackAccessTable
		if (CurrentRegionBytes > 0) { // Get last region as we exited the loop.
			if (CurrentRegionType == STARS_STACK_CALLEE_SAVED_REG) {
				SMP_fprintf(AnnotFile, "%18llx %6u MEMORYHOLE STACK esp + 0 CalleeSavedRegs \n",
					(uint64_t) addr, CurrentRegionBytes);
			}
			else if (CurrentRegionType == STARS_STACK_LOCAL_FRAME) {
				ParentReferentID = global_STARS_program->GetDataReferentID();
				global_STARS_program->IncrementDataReferentID();
				SMP_fprintf(AnnotFile, "%18llx %6u DATAREF STACK %lu esp + 0 PARENT LocalFrame LOCALFRAME\n",
					(uint64_t) addr, CurrentRegionBytes, (unsigned long) ParentReferentID);
				LocalFrameEmitted = true;
			}
			else if (CurrentRegionType == STARS_STACK_OUTARG) {
				if (!LocalFrameEmitted) { // entire local stack frame is outgoing args region; unusual case
					ParentReferentID = global_STARS_program->GetDataReferentID();
					global_STARS_program->IncrementDataReferentID();
					SMP_fprintf(AnnotFile, "%18llx %6u DATAREF STACK %lu esp + 0 PARENT LocalFrame LOCALFRAME\n",
						(uint64_t) addr, CurrentRegionBytes, (unsigned long)ParentReferentID);
					LocalFrameEmitted = true;
				}
				SMP_fprintf(AnnotFile, "%18llx %6zu DATAREF STACK %lu esp + 0 CHILDOF %lu OFFSET %d OutArgsRegion OUTARGS\n",
					(uint64_t) addr, CurrentRegionBytes, global_STARS_program->GetDataReferentID(), ParentReferentID, 0);
				global_STARS_program->IncrementDataReferentID();
			}
		}

	} // end if (0 < LocalVarsSize)
	return;
} // end of SMPFunction::EmitStackFrameAnnotations() 

#else
// Emit the annotations describing the regions of the stack frame.
void SMPFunction::EmitStackFrameAnnotations(FILE *AnnotFile, SMPInstr *Instr) {
	STARS_ea_t addr = Instr->GetAddr();

#if 0
	if (0 < IncomingArgsSize) {
		SMP_fprintf(AnnotFile, "%18llx %6llu INARGS STACK esp + %ld %s \n",
				(uint64_t) addr, (uint64_t) IncomingArgsSize,
				(long) (this->GetLocalVarsSize() + CalleeSavedRegsSize + RetAddrSize),
				Instr->GetDisasm());
	}
#endif
	if (0 < this->RetAddrSize) {
		SMP_fprintf(AnnotFile, "%18llx %6d MEMORYHOLE STACK esp + %lu ReturnAddress \n",
				(uint64_t) addr, RetAddrSize, (unsigned long) (this->GetLocalVarsSize() + this->CalleeSavedRegsSize));
	}
	if (0 < this->CalleeSavedRegsSize) {
		SMP_fprintf(AnnotFile, "%18llx %6u MEMORYHOLE STACK esp + %lu CalleeSavedRegs \n",
				(uint64_t) addr, this->CalleeSavedRegsSize, (unsigned long) this->GetLocalVarsSize());
	}
	if ((0 < this->GetLocalVarsSize()) && this->GoodLocalVarTable) {
		unsigned long ParentReferentID = global_STARS_program->GetDataReferentID();
		global_STARS_program->IncrementDataReferentID();
		SMP_fprintf(AnnotFile, "%18llx %6lu DATAREF STACK %lu esp + %d PARENT LocalFrame LOCALFRAME\n",
				(uint64_t) addr, (unsigned long) this->GetLocalVarsSize(), ParentReferentID, 0);
		if (this->AnalyzedSP && !this->CallsAlloca && (STARS_BADADDR != this->LocalVarsAllocInstr)) {
			// We can only fine-grain the stack frame if we were able to analyze the stack
			if (this->OutgoingArgsSize > 0) {
				SMP_fprintf(AnnotFile, "%18llx %6zu DATAREF STACK %lu esp + %d CHILDOF %lu OFFSET %d OutArgsRegion OUTARGS\n",
					(uint64_t) addr, this->OutgoingArgsSize, global_STARS_program->GetDataReferentID(), 0, ParentReferentID, 0);
				global_STARS_program->IncrementDataReferentID();
			}
#if SMP_DEBUG_STACK_GRANULARITY
			SMP_msg("LocalVarTable of size %d for function %s\n", this->LocalVarTable.size(),
				this->GetFuncName());
#endif
			for (std::size_t i = 0; i < this->LocalVarTable.size(); ++i) {
#if SMP_DEBUG_STACK_GRANULARITY
				SMP_msg("Entry %d offset %ld size %d name %s\n", i, this->LocalVarTable[i].offset,
					this->LocalVarTable[i].size, this->LocalVarTable[i].VarName);
#endif
				// Don't emit annotations for incoming or outgoing args or anything else
				//  above or below the current local frame.
				if ((this->LocalVarTable[i].offset >= (long) this->FuncInfo->GetFrameSize())
					|| (this->LocalVarTable[i].offset < (long) this->OutgoingArgsSize))
					continue;
				SMP_fprintf(AnnotFile, "%18llx %6zu DATAREF STACK %lu esp + %ld CHILDOF %lu OFFSET %ld LOCALVAR %s \n",
					(uint64_t) addr, this->LocalVarTable[i].size, global_STARS_program->GetDataReferentID(),
					this->LocalVarTable[i].offset, ParentReferentID,
					this->LocalVarTable[i].offset, this->LocalVarTable[i].VarName);
				global_STARS_program->IncrementDataReferentID();
			}
		} // end if (this->AnalyzedSP and not Alloca .... )
	} // end if (0 < LocalVarsSize)
	return;
} // end of SMPFunction::EmitStackFrameAnnotations() 
#endif

// Fill MemDefAddrs with inst addrs that DEF MemUseOp, tracing back from UseAddr.
// Stop if we encounter a DEF of an address reg in MemUseOp.
void SMPFunction::FindMatchingMemDEFAddrs(STARS_ea_t UseAddr, SMPBasicBlock *CurrBlock, STARSOpndTypePtr &MemUseOp, std::list<STARS_ea_t> &MemDefAddrs, set<int> &AddressRegs) {
	set<int> CommonAddressRegs; // result of set intersection
	bool UseAddrSeen = (STARS_BADADDR == UseAddr);
	set<int> DefRegs; // regs in DEF set of CurrInst

	// Trace back from UseAddr to find MemDEFs that match MemUseOp.
	//  If UseAddr == STARS_BADADDR, that means the mthod has been called recursively and it should
	//  trace back from the last instruction in CurrBlock. By ensuring that an inst addr is < UseAddr,
	//  we accomplish both cases without special cases in the code.
	if (!CurrBlock->IsProcessed()) {
		for (vector<SMPInstr *>::const_reverse_iterator InstIter = CurrBlock->GetRevInstCBegin(); InstIter != CurrBlock->GetRevInstCEnd(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			if (InstAddr == UseAddr) {
				UseAddrSeen = true;
			}
			else if (UseAddrSeen) {
				DefRegs.clear();
				if (!AddressRegs.empty()) {
					CurrInst->GetDEFRegs(DefRegs);
					(void) set_intersection(AddressRegs.begin(), AddressRegs.end(), DefRegs.begin(),
						DefRegs.end(), std::inserter(CommonAddressRegs, CommonAddressRegs.begin()));
					if (!CommonAddressRegs.empty()) {
						// Intersection is not empty. An address reg is being redefined; stop search.
						return;
					}
				}
				if (CurrInst->HasDestMemoryOperand()) {
					STARSOpndTypePtr MemDefOp = CurrInst->GetMemDef();
					assert(nullptr != MemDefOp);
					if (IsEqOpIgnoreBitwidth(MemUseOp, MemDefOp)) {
						MemDefAddrs.push_back(InstAddr);
					}
				}
			}
		}
		// If we are beginning our first recursion, mark descendants in the CFG (not 
		//  including back edges) as processed to avoid cycling.
		if (STARS_BADADDR != UseAddr) {
			CurrBlock->DepthFirstMarkNonBackEdgeSuccessors();
		}
		else {
			CurrBlock->SetProcessed(true);
		}
		// Recurse into predecessors, passing in STARS_BADADDR as the UseAddr.
		for (list<SMPBasicBlock *>::iterator PredIter = CurrBlock->GetFirstPred(); PredIter != CurrBlock->GetLastPred(); ++PredIter) {
			SMPBasicBlock *PredBlock = (*PredIter);
			this->FindMatchingMemDEFAddrs(STARS_BADADDR, PredBlock, MemUseOp, MemDefAddrs, AddressRegs);
		}
	}
		
	return;
} // end of SMPFunction::FindMatchingMemDEFAddrs()

// Find first inst in loop, return its stack delta.
STARS_sval_t SMPFunction::GetLoopIncomingStackDelta(std::size_t LoopNum) const {
	assert(LoopNum < this->LoopHeadBlockNumbers.size());
	int HeadBlockNum = this->LoopHeadBlockNumbers[LoopNum];
	assert(SMP_BLOCKNUM_UNINIT != HeadBlockNum);
	SMPBasicBlock *HeadBlock = this->RPOBlocks[(size_t) HeadBlockNum];
	assert(nullptr != HeadBlock);
	return HeadBlock->GetIncomingStackDelta();
} // SMPFunction::GetLoopIncomingStackDelta()

// Mark functions with multiple entry points. These will be unsafe for fast returns and
//  are probably IDA Pro disassembly problems rather than true multi-entry functions.
void SMPFunction::DetectMultiEntryFunction(void) {
	bool IsIndirectCallTarget = false;
	bool IsTailCallTarget = false;
	this->MultipleEntryPoints = this->GetFuncInfo()->IsMultiEntry(this->HasIndirectJumps());
	STARS_ea_t FirstAddr = this->GetFirstFuncAddr();
	STARS_Function_t *FuncPtr = SMP_get_func(FirstAddr);

	bool TargetSuccess = FuncPtr->AnalyzeInstAsCallTarget(this, IsIndirectCallTarget, IsTailCallTarget);

	if (!IsIndirectCallTarget) {
		IsIndirectCallTarget = this->GetProg()->IsCodeXrefFromData(FirstAddr);
	}

	this->PossibleIndirectCallTarget |= IsIndirectCallTarget;
	this->PossibleTailCallTarget = IsTailCallTarget;

	// Augment with our own analysis if no multi-entry detected yet.
	if (!(this->MultipleEntryPoints || this->IsPossibleIndirectCallTarget())) {
		list<SMPInstr *>::iterator Instructions = Instrs.begin();
		SMPInstr *CurrInst;
#if SMP_USE_SSA_FNOP_MARKER
		++Instructions;  // skip marker instruction
#endif
		STARS_ea_t FirstAddr = (*Instructions)->GetAddr();
		for (; Instructions != Instrs.end(); ++Instructions) {
			CurrInst = (*Instructions);
			STARS_ea_t address = CurrInst->GetAddr();
			// If we do not already have a MultiEntryFunction determination, search
			//  for addresses beyond the entry point that have their address taken.
			// We exclude code addresses found in jump tables.
			bool AddrFoundInJumpTable = (this->JumpTableTargets.find(address) != this->JumpTableTargets.end());
			if ((address != FirstAddr) && (!AddrFoundInJumpTable)) {
				// Could optimize this later by searching the CodeAddressesTaken set
				//  for the range of addresses in this function.
				if (global_STARS_program->IsCodeAddressTaken(address)) {
					this->MultipleEntryPoints = true;
					SMP_msg("INFO: Func at %llx becoming multi-entry because code addr %llx found in data.\n",
						(uint64_t)FirstAddr, (uint64_t)address);
					break; // don't need to look for more
				}
				else if (this->GetProg()->IsCodeXrefFromData(address)) {
					this->MultipleEntryPoints = true;
					SMP_msg("INFO: Func at %llx becoming multi-entry because code addr %llx found in CodeXrefsFromData.\n",
						(uint64_t)FirstAddr, (uint64_t)address);
					break; // don't need to look for more
				}
			}
		} // end for all instructions
	}
	return;
} // end of SMPFunction::DetectMultiEntryFunction()

// Audit and fix the IDA Pro code cross references for jumps and jump targets.
void SMPFunction::MDAuditJumpXrefs(void) {

	// ask interface to do this, probably only done for idapro
	GetFuncInfo()->UpdateXrefs();
} // end of SMPFunction::MDAuditJumpXrefs()

// Rebuild AllCallTargets as the union of the direct and indirect call targets.
void SMPFunction::RebuildCallTargets(void) {
	set<STARS_ea_t>::iterator TargetIter;
	this->AllCallTargets.clear();
	for (TargetIter = this->DirectCallTargets.begin(); TargetIter != this->DirectCallTargets.end(); ++TargetIter) {
		STARS_ea_t TargetAddr = (*TargetIter);
		this->AllCallTargets.push_back(TargetAddr);
	}
	for (TargetIter = this->IndirectCallTargets.begin(); TargetIter != this->IndirectCallTargets.end(); ++TargetIter) {
		STARS_ea_t TargetAddr = (*TargetIter);
		this->AllCallTargets.push_back(TargetAddr);
	}
	return;
} // end of SMPFunction::RebuildCallTargets()

	// Main data flow analysis driver. Goes through the function and
//  fills all objects for instructions, basic blocks, and the function
//  itself.
void SMPFunction::AnalyzeFunc(void) {
	set<STARS_ea_t> FragmentWorkList;  // Distant code fragments that belong to this function and need processing
	STARS_ea_t InstAddr; // grab address to help in debugging, conditional breakpoints, etc.

#if SMP_DEBUG_CONTROLFLOW
	SMP_msg("Entering SMPFunction::Analyze.\n");
#endif

	// Get some basic info from the FuncInfo structure.
	this->Size = this->FuncInfo->GetFuncSize();
	this->UseFP = this->FuncInfo->FunctionUsesFP(); 
	this->StaticFunc = this->FuncInfo->IsStaticFunction(); 
	this->LibFunc = this->FuncInfo->IsLibraryFunction(); 
	this->AnalyzedSP = this->FuncInfo->IsStackPointerAnalyzed(); 

#if SMP_DEBUG_CONTROLFLOW
	SMP_msg("SMPFunction::AnalyzeFunc: got basic info.\n");
#endif

	// Detect exception throwing code.
	if (0 == strcmp("__cxa_throw", this->GetFuncName())) {
		this->GetProg()->SetProgramThrowsExceptions();
	}

	// Determine if we are dealing with shared chunks.
	GetFuncInfo()->MarkSharedChunks();
#if STARS_AUDIT_JUMP_XREFS
	this->MDAuditJumpXrefs();
#endif

	STARS_Function_t *MySTARSFunc = this->GetFuncInfo();
	MySTARSFunc->BuildFuncIR(this);

#if KLUDGE_VFPRINTF_FAMILY
	if (!this->HasSharedChunks() && (0 != strstr(this->GetFuncName(), "printf"))) {
		GetFuncInfo()->SetSharedChunks(true);
		SMP_msg("INFO: Kludging function %s\n", this->GetFuncName());
	}
#endif

#if SMP_IDAPRO52_WORKAROUND
	if (!this->HasSharedChunks() && (0 == strcmp(this->GetFuncName(), "error_for_asm"))) {
		GetFuncInfo()->SetSharedChunks(true);
		SMP_msg("Kludging function %s\n", this->GetFuncName());
	}
#endif

	// Now that we have all instructions and basic blocks, link each instruction
	//  to its basic block. 
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			CurrInst->SetBlock(CurrBlock->GetThisBlock());


			// jdh hack:  redo syscall analysis as block may not be set 
			// when syscall analysis was done first time.
			if (CurrInst->GetDataFlowType() == INDIR_CALL) {
				if (CurrInst->MDIsSystemCall()) {
					this->SetHasSystemCalls();
				}
			}

			SMPitype InstDataFlowType = CurrInst->GetDataFlowType();
			if ((CALL == InstDataFlowType) || (INDIR_CALL == InstDataFlowType)) {
				CurrBlock->SetHasCallInst();
				STARS_ea_t CallTarget = CurrInst->GetCallTarget();
				if (0 == CallTarget) {
					// Some libc functions have a call to a debug function that is set to a "call 0" 
					//  when the library is not compiled with debug options. The basic block is
					//  unreachable and likely will be removed later by SCCP, but it is nice to remove
					//  it before it clutters up SSA, and just in case it has some weird code pattern
					//  before it that obscures the fact that it is unreachable.
					CurrBlock->SetUnreachableBlock(true);
					SMP_msg("INFO: Marking basic block unreachable due to call 0 instruction at %llx\n", (uint64_t) InstAddr);
				}
				else {
					if (CALL == InstDataFlowType) {
						if (STARS_BADADDR != CallTarget) {
							bool NewDirectTarget = this->AddDirectCallTarget(CallTarget);
						}
						else {
							SMP_msg("ERROR: Direct call with bad target at %llx ", (uint64_t) InstAddr);
							this->DumpFuncNameAndAddr();
						}
					}
					else { // INDIR_CALL
						this->SetHasIndirectCalls();
						if (STARS_BADADDR == CallTarget) {
							if (!CurrInst->MDIsSystemCall()) {
								this->SetHasUnresolvedIndirectCalls();
							}
						}
						else {
							bool NewIndirectTarget = this->AddIndirectCallTarget(CallTarget);
						}
					}
				}
			} // end if CALL or INDIR_CALL
		} // end for each inst
		CurrBlock->Analyze();
		// Set up the map of instructions to basic blocks.
		for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			pair<STARS_ea_t, SMPBasicBlock *> MapItem(InstAddr, CurrBlock);
			this->InstBlockMap.insert(MapItem);
			if ( CurrInst->MDIsPossibleStringLoopingOpcode() && CurrInst->MDHasAnyRepeatPrefix() ) {
				CurrBlock->SetHasLoopingStringOpcode();
			}
		}

	} // end for each block

	// Set up basic block links and map of instructions to blocks.
	this->SetLinks();
	this->RPONumberBlocks();
	this->DomSoFar.AllocateBits(this->GetNumBlocks());
	this->TCFGDomSoFar.AllocateBits(4 * this->GetNumBlocks());
	this->DetectLinkerStubFunction();
	if (this->GetProg()->IsFuncAddrCalledFromOrphanedCode(this->GetFirstFuncAddr())) {
		this->SetIsCalledFromOrphanedCode();
	}

	FragmentWorkList.clear();
	return;
} // end of SMPFunction::AnalyzeFunc()

#define STARS_DEBUG_LINKER_STUBS 1
// Determine whether func is a linker stub, e.g. a PLT stub
void SMPFunction::DetectLinkerStubFunction(void) {
#if STARS_DEBUG_LINKER_STUBS
	if (0x405c40 == this->GetFirstFuncAddr()) {
		SMP_msg("DEBUG: Linker stub: NumBlocks: %zd NumInstrs: %zd \n", this->GetNumBlocks(), this->Instrs.size());
		SMP_msg("DEBUG: Block 0 indir jump: %d Last block 1 tail call: %d \n", 
			this->RPOBlocks[0]->HasIndirectJump(),
			this->RPOBlocks[this->GetNumBlocks() - 1]->EndsWithTailCall());
	}
#endif
	// We will address only GCC PLT stub forms at this time.
	if ((2 >= this->GetNumBlocks()) && (6 >= this->Instrs.size())) {
		// Each block of a PLT stub ends with an indirect jump, and there
		//  are not too many other instructions.
		//  In particular, it would be bizarre for any other function besides a linker stub to
		//  have two basic blocks, with the second block being reached only by an indirect jump from
		//  the first block (which happens through the PLT, but we are not checking that here), and then
		//  the second block ending with a tail call jump, with no return instruction in the func.
		bool IDA_PLT = (this->RPOBlocks[0]->HasIndirectJump()
			&& ((1 == this->GetNumBlocks()) || this->RPOBlocks[this->GetNumBlocks() - 1]->EndsWithTailCall()));
		bool IRDB_PLT = ((1 == this->GetNumBlocks()) && this->RPOBlocks[0]->EndsWithTailCall());
		if (IDA_PLT || IRDB_PLT) {
			this->LinkerStub = true;
			SMP_msg("INFO: Linker stub function found at %llx \n", (uint64_t) this->GetFirstFuncAddr());
		}
		else if (this->Instrs.size() < 6) { // Must have few instructions
			// Try the lower-level pattern match:
			//  indirect jump
			//  push value
			//  unconditional jump
			// e.g., from objdump:
			// <__gnat_free_process@plt>:  jmpq   *0x4661ba(%rip)        # 0x575e30
			// <__gnat_free_process@plt + 6>:    pushq  $0x2f2
			// <__gnat_free_process@plt + 11> : jmpq   0x10cd40
			bool IndirJumpFirst = false;
			bool PushSecond = false;
			bool JumpThird = false;
			uint16_t InstCount = 0;
			for (list<SMPInstr *>::iterator InstIter = this->GetFirstInstIter(); InstIter != this->GetLastInstIter(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (!CurrInst->IsMarkerInst()) {
					if (!IndirJumpFirst) {
						IndirJumpFirst = (CurrInst->GetDataFlowType() == INDIR_JUMP) || CurrInst->IsTailCall();
					}
					else if (!PushSecond) {
						PushSecond = CurrInst->MDIsPushInstr();
					}
					else if (!JumpThird) {
						JumpThird = (CurrInst->GetDataFlowType() == JUMP) || CurrInst->IsTailCall();
					}
					++InstCount;
				}
			} // end for each inst
			if ((3 == InstCount) && IndirJumpFirst && PushSecond && JumpThird) {
				this->LinkerStub = true;
				SMP_msg("INFO: Linker stub function found at %llx by 3-inst pattern\n", (uint64_t) this->GetFirstFuncAddr());
			}
		}
	}
	return;
} // end of SMPFunction::DetectLinkerStubFunction()

// Detect function that just puts its return address into a reg and returns.
void SMPFunction::DetectThunkFunction(void) {
	if (1 == this->GetNumBlocks() && this->HasSTARSStackPtrAnalysisCompleted() && this->StackPtrAnalysisSucceeded()) {
		bool FoundReturnAddressRead = false;
		SMPBasicBlock *CurrBlock = this->RPOBlocks[0];
		for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			if (CurrInst->IsMarkerInst())
				continue;
			if (!FoundReturnAddressRead && CurrInst->HasSourceMemoryOperand() && CurrInst->MDIsMoveInstr()) {
				STARSOpndTypePtr MemSrcOp = CurrInst->GetMemUse();
				if (MDIsDirectStackAccessOpnd(MemSrcOp, false)) {
					// Must be normalized, because stack ptr analysis succeeded.
					int BaseReg, IndexReg;
					uint16_t Scale;
					STARS_ea_t Offset;
					MemSrcOp->MDExtractAddressFields(BaseReg, IndexReg, Scale, Offset);
					FoundReturnAddressRead = (0 == Offset);
				}
			}
			else if (FoundReturnAddressRead && CurrInst->HasReturnOpcode()) {
				this->ThunkFunc = true;
				break;
			}
		}
	}
	return;
} // end of SMPFunction::DetectThunkFunction()

// Perform analyses that might need some info from other functions in the call graph.
void SMPFunction::AdvancedAnalysis(void) {
	list<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;

	// IDA Pro has trouble with functions that do not have any local
	//  variables. Unfortunately, the C library has plenty of these
	//  functions. IDA usually claims that frregs is zero and frsize
	//  is N, when the values should have been reversed. We can attempt
	//  to detect this and fix it. IDA Pro also sometimes has trouble with
	//  functions that allocate the stack frame, and then push registers
	//  later before making a call, because it wants to include register
	//  pushes below the stack frame as being part of the stack frame,
	//  even when they are temporary saves and restores. __brk in the
	//  Gnu stdclib is an example as of November of 2012.
	// NOTE: We are no longer fixing IDA Pro frame info as of July, 2015.
	//  We just compute our own frame info, which has always proved to be at
	//  least as good as IDA Pro info.
	bool FrameInfoFixed = this->MDFixFrameInfo();
#if SMP_DEBUG_CONTROLFLOW
	SMP_msg("Returned from MDFixFrameInfo()\n");
#endif

	this->FindAllAllocsAndDeallocs();

	InstIter = this->Instrs.begin();
	if ((*InstIter)->IsMarkerInst()) {
		++InstIter; // skip marker inst
	}
	for ( ; InstIter != this->Instrs.end(); ++InstIter) {
		CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr(); // for debugging breakpoints
		// We can finally search for stack loads now that UseFP has been fixed by
		//  MDFixFrameInfo(). Otherwise, we would do this in SMPInstr::Analyze(),
		//  but the UseFP flag is not ready that early.
		CurrInst->MDFindLoadFromStack(this->UsesFramePointer());

		// Fix up machine dependent quirks in the def and use lists.
		//  This used to be called from within SMPInstr.Analyze(), but info such as UseFP
		//  is not available that early.
		CurrInst->MDFixupDefUseLists();

		if (CurrInst->HasGoodRTL())
			CurrInst->SyncAllRTs(this->UsesFramePointer(), this->GetFramePtrStackDelta());

		// Detect indirect memory references.
		CurrInst->AnalyzeIndirectRefs(this->UsesFramePointer());
	}

	// Audit the call instructions and call targets.
	//  !!!!****!!!! NOTE: Not sure the address range checks in this code are valid
	//   for functions with scattered chunks.
	if ((!this->AllCallTargets.empty()) || this->HasUnresolvedIndirectCalls()) {
		bool FoundInternalCallTarget = false;
		vector<STARS_ea_t>::iterator CurrTarget = this->AllCallTargets.begin();
		set<STARS_ea_t>::iterator CurrDirectTarget, CurrIndirectTarget;
		while (CurrTarget != this->AllCallTargets.end()) {
			if (this->IsInstIDInFunc(*CurrTarget)) {
				// Found a call target that is within the function.
				FoundInternalCallTarget = true;
				if (this->GetFirstFuncAddr() == *CurrTarget) { // Direct recursion, not a pseudo-jump
					this->DirectlyRecursive = true;
				}
				CurrTarget = this->AllCallTargets.erase(CurrTarget);
			}
			else {
				++CurrTarget;
			}
		}
		if (FoundInternalCallTarget) {
			// We have to mark the pseudo-call instructions and audit the direct and
			//  indirect call target vectors.

			// Audit direct call targets.
			CurrDirectTarget = this->DirectCallTargets.begin();
			set<STARS_ea_t>::iterator CopyOfIterator;
			bool CallTargetRemoved = false;
			while (CurrDirectTarget != this->DirectCallTargets.end()) {
				STARS_ea_t TargetAddr = (*CurrDirectTarget);
				if (this->IsInstIDInFunc(TargetAddr) && (this->FuncInfo->get_startEA() != TargetAddr)) {
					// Found a call target that is within the function, but not recursive call to func entry.
					CopyOfIterator = this->RemoveDirectCallTarget(TargetAddr);
					CurrDirectTarget = CopyOfIterator;
					CallTargetRemoved = true;
				}
				else {
					++CurrDirectTarget;
				}
			}
			// Audit indirect call targets.
			CurrIndirectTarget = this->IndirectCallTargets.begin();
			while (CurrIndirectTarget != this->IndirectCallTargets.end()) {
				STARS_ea_t TargetAddr = (*CurrIndirectTarget);
				if (this->IsInstIDInFunc(TargetAddr)) {
					// Found a call target that is within the function.
					CopyOfIterator = CurrIndirectTarget;
					++CopyOfIterator; // point to element after element that will be erased
					this->IndirectCallTargets.erase(CurrIndirectTarget);
					CurrIndirectTarget = CopyOfIterator;
					CallTargetRemoved = true;
				}
				else {
					++CurrIndirectTarget;
				}
			}
#if 1
			// Find calls used as jumps.
			list<SMPInstr *>::iterator InstIter = this->Instrs.begin();
			while (InstIter != this->Instrs.end()) {
				SMPInstr *CurrInst = (*InstIter);
				SMPitype InstFlow = CurrInst->GetDataFlowType();
				if ((CALL == InstFlow) || (INDIR_CALL == InstFlow)) {
					CurrInst->AnalyzeCallInst(this->GetFirstFuncAddr());
				}
				++InstIter;
			}
#endif
		} // end if (FoundInternalCallTarget)
	}

	// Figure out the stack frame and related info.
	(void) this->AnalyzeStackPointerDeltas();
	this->StackPtrCopySet.clear(); // free memory
	this->TempStackDeltaReachesList.clear();
	this->TempReachingDefs.clear();

	if (this->StackPtrAnalysisSucceeded() && this->HasGoodRTLs() && (!this->HasUnresolvedIndirectJumps())) {
		this->AuditCallingConvention();
#define RECOMPUTE_VARKILL_SETS_AFTER_NORMALIZATION 1
#if RECOMPUTE_VARKILL_SETS_AFTER_NORMALIZATION
		if (this->CallsAlloca) {
			list<SMPBasicBlock *>::iterator BlockIter;
			for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
				(*BlockIter)->InitKilledExposed(UseFP, false);
			}
		}
#endif
		this->LiveVariableAnalysis(false);
	}

#define RECOMPUTE_LVA_SETS_AFTER_NORMALIZATION 0
#if RECOMPUTE_LVA_SETS_AFTER_NORMALIZATION
	// Recompute LVA sets if we have stack ops in any block's LVA sets
	//  and we went through alloca() reprocessing in AnalyzeStackPointerDeltas() above.
	if (this->CallsAlloca) {
		this->RecomputeStackLVA();
	}
#endif

	// Calculate min and max stack point deltas, and AllocPointDelta.
	this->MinStackDelta = 20000; // Final value should be negative or zero
	this->MaxStackDelta = -1000; // Final value should be zero.
	InstIter = this->Instrs.begin();
	if ((*InstIter)->IsMarkerInst()) {
		++InstIter; // skip marker inst
	}
	for (; InstIter != this->Instrs.end(); ++InstIter) {
		CurrInst = (*InstIter);
		STARS_ea_t addr = CurrInst->GetAddr();
		STARS_sval_t sp_delta = CurrInst->GetStackPtrOffset();
		if (sp_delta < this->MinStackDelta)
			this->MinStackDelta = sp_delta;
		if (sp_delta > this->MaxStackDelta)
			this->MaxStackDelta = sp_delta;
		if (addr == this->LocalVarsAllocInstr) {
			// Total stack pointer delta is sp_delta for the next instruction,
			//  because IDA updates the sp delta AFTER each instruction.
			list<SMPInstr *>::iterator NextInstIter = InstIter;
			++NextInstIter;
			if (NextInstIter != this->Instrs.end()) {
				sp_delta = (*NextInstIter)->GetStackPtrOffset();
				this->AllocPointDelta = sp_delta;
			}
		}
	}

	bool WritesMem = false;
	bool CallChainNonReturning = false;
	bool Changed = this->ComputeInOutRegs(false, WritesMem, CallChainNonReturning);
	this->AltersMemory |= WritesMem;
	return;
} // end of SMPFunction::AdvancedAnalysis()

// Fix call inst DEFs and USEs, return true if changed
bool SMPFunction::AdvancedAnalysis2(void) {
	bool FuncChanged = false;
#if 0   // We want to process directly recursive calls to see what is LiveIn and not LiveIn
	if (!this->IsLeaf()) {
#endif
		list<SMPInstr *>::iterator InstIter;
		SMPInstr *CurrInst;

		// Fixup DEF and USE lists for call insts.
		InstIter = this->Instrs.begin();
		if ((*InstIter)->IsMarkerInst()) {
			++InstIter; // skip marker inst
		}
		for (; InstIter != this->Instrs.end(); ++InstIter) {
			CurrInst = (*InstIter);
			STARS_ea_t addr = CurrInst->GetAddr();
			bool InstChanged = CurrInst->MDFixupCallDefUseLists();
			FuncChanged = FuncChanged || InstChanged;
		}
#if 0 // debug crashes first
		if (FuncChanged) {
			// LVA sets are stale because call instructions had DEFs and USEs updated.
			this->LiveVariableAnalysis(true);
		}
#endif
#if 0
	}
#endif

	return FuncChanged;
} // end of SMPFunction::AdvancedAnalysis2()

// Perform analyses that might need some info from other functions in the call graph.
void SMPFunction::AdvancedAnalysis3(void) {
	list<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;

#if SMP_DEBUG_CONTROLFLOW
	SMP_msg("SMPFunction::AdvancedAnalysis3: set stack frame info.\n");
#endif
	if (this->StackPtrAnalysisSucceeded()) {
		if (!(this->HasSharedChunks())) {
			this->SetStackFrameInfo();
		} // end if not shared chunks
		else { // has shared chunks; still want to compute stack frame info
#ifdef SMP_DEBUG_FUNC
			SMP_msg("INFO: %s has shared chunks \n", this->GetFuncName());
#endif
			// Figure out the stack frame and related info.
			this->SetStackFrameInfo();
		}
	}

#if SMP_COUNT_MEMORY_ALLOCATIONS
	SMPInstCount += ((unsigned long) this->Instrs.size());
	SMPBlockCount += ((unsigned long) this->Blocks.size());
	SMPLocalVarCount += ((unsigned long) this->LocalVarTable.size());
#endif
	return;
} // end of SMPFunction::AdvancedAnalysis3()

// Count call targets that have not been processed.
std::size_t SMPFunction::UnprocessedCalleesCount(void) {
	std::size_t UnprocessedTargetsCount = 0;

	std::size_t TargetIndex;
	for (TargetIndex = 0; TargetIndex < this->AllCallTargets.size(); ++TargetIndex) {
		SMPFunction *CurrTarget = this->GetProg()->FindFunction(this->AllCallTargets[TargetIndex]);
		if (nullptr == CurrTarget) {
#if 0
			// Bad call targets are removed in AdvancedAnalysis(), which comes later.
			SMP_msg("ERROR: NULL CallTarget in UnprocessedCalleesCount() at TargetIndex %zu \n", TargetIndex);
#endif
		}
		else if (!(CurrTarget->IsFuncProcessed())) {
			++UnprocessedTargetsCount;
		}
	}
	return UnprocessedTargetsCount;
} // end of SMPFunction::UnprocessedCalleesCount()

STARS_ea_t SMPFunction::GetFirstUnprocessedCallee(void) {
	STARS_ea_t CalleeAddr = STARS_BADADDR;
	std::size_t TargetIndex;
	for (TargetIndex = 0; TargetIndex < this->AllCallTargets.size(); ++TargetIndex) {
		STARS_ea_t TargetAddr = this->AllCallTargets[TargetIndex];
		SMPFunction *CurrTarget = this->GetProg()->FindFunction(TargetAddr);
		if ((nullptr != CurrTarget) && (!(CurrTarget->IsFuncProcessed()))) {
			CalleeAddr = TargetAddr;
			break;
		}
	}
	return CalleeAddr;
} // end of SMPFunction::GetFirstUnprocessedCallee()

// Is the code starting at TargetAddr a non-shared chunk that jumps back into our function?
//  If so, it can be incorporated into our function rather than treated as a separate function.
//  This method is called only when we see a jump outside our function, and it is looking for
//  code fragments that are not really functions (i.e. don't have a stack frame, jump straight back
//  into our function after executing a few instructions, not a chunk shared among other functions).
//  These code fragments are found in the locking and unlocking code of the gcc stdlib, for example.
bool SMPFunction::FindDistantCodeFragment(STARS_ea_t TargetAddr) {
	return GetFuncInfo()->FindDistantCodeFragment(this,TargetAddr);
} // end of SMPFunction::FindDistantCodeFragment()

// Free memory that is no longer needed after loop 2 of SMPProgram::Analyze().
void SMPFunction::FreeUnusedMemory2(void) {
	std::size_t UnusedElements;
	std::size_t CurrSize;

#if 0
	// Go through vector containers and resize to current capacity, if the vector
	//  has been fully computed by the time SMPProgram:Analyze() loop 2 completes.
	CurrSize = this->DirectCallTargets.size();
	UnusedElements = this->DirectCallTargets.capacity() - CurrSize;
	if (0 < UnusedElements) {
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<STARS_ea_t>(this->DirectCallTargets).swap(this->DirectCallTargets);
#else
		this->DirectCallTargets.resize(CurrSize);
#endif
	}

	CurrSize = this->IndirectCallTargets.size();
	UnusedElements = this->IndirectCallTargets.capacity() - CurrSize;
	if (0 < UnusedElements) {
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<STARS_ea_t>(this->IndirectCallTargets).swap(this->IndirectCallTargets);
#else
		this->IndirectCallTargets.resize(CurrSize);
#endif
	}
#endif

	CurrSize = this->AllCallTargets.size();
	UnusedElements = this->AllCallTargets.capacity() - CurrSize;
	if (0 < UnusedElements) {
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<STARS_ea_t>(this->AllCallTargets).swap(this->AllCallTargets);
#else
		this->AllCallTargets.resize(CurrSize);
#endif
	}

	CurrSize = this->SavedRegLoc.size();
	UnusedElements = this->SavedRegLoc.capacity() - CurrSize;
	if (0 < UnusedElements) {
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<int>(this->SavedRegLoc).swap(this->SavedRegLoc);
#else
		this->SavedRegLoc.resize(CurrSize);
#endif
	}

	CurrSize = this->RPOBlocks.size();
	UnusedElements = this->RPOBlocks.capacity() - CurrSize;
	if (0 < UnusedElements) {
		list<SMPBasicBlock *>::iterator DummyIter = this->Blocks.end();
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<SMPBasicBlock *>(this->RPOBlocks).swap(this->RPOBlocks);
#else
		this->RPOBlocks.resize(CurrSize, DummyIter);
#endif
	}

	CurrSize = this->LocalVarTable.size();
	UnusedElements = this->LocalVarTable.capacity() - CurrSize;
	if (0 < UnusedElements) {
		struct LocalVar DummyVar;
		DummyVar.offset = 0;
		DummyVar.size = 0;
		UnusedStructCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<struct LocalVar>(this->LocalVarTable).swap(this->LocalVarTable);
#else
		this->LocalVarTable.resize(CurrSize, DummyVar);
#endif
	}

	return;
} // end of SMPFunction::FreeUnusedMemory2()

// Free memory that is no longer needed after loop 3 of SMPProgram::Analyze().
void SMPFunction::FreeUnusedMemory3(void) {
	std::size_t UnusedElements;
	std::size_t CurrSize;

	// Go through vector containers and resize to current capacity, if the vector
	//  has been fully computed by the time SMPProgram:Analyze() loop 2 completes.
	CurrSize = this->ReturnRegTypes.size();
	UnusedElements = this->ReturnRegTypes.capacity() - CurrSize;
	if (0 < UnusedElements) {
		UnusedIntCount += (unsigned long) UnusedElements;
#if SMP_SHRINK_TO_FIT
		std::vector<SMPOperandType>(this->ReturnRegTypes).swap(this->ReturnRegTypes);
#else		
		this->ReturnRegTypes.resize(CurrSize);
#endif
	}

	return;
} // end of SMPFunction::FreeUnusedMemory3()

// Free memory that is no longer needed after type inference (loop 4 of SMPProgram::Analyze()).
void SMPFunction::FreeUnusedMemory4(void) {
	// this->KillSet.clear();  // needed for EmitFuncSPARKAda()
	// this->LiveOutSet.clear();
	this->BlocksDefinedIn.clear();

#if SMP_SHRINK_TO_FIT
#if 0
	STARSOpndSet(this->KillSet).swap(this->KillSet);
	STARSOpndSet(this->LiveOutSet).swap(this->LiveOutSet);
#endif
#endif

	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		(*BlockIter)->FreeUnusedMemory4();
	}
	return;
} // end of SMPFunction::FreeUnusedMemory4()

// Free SSA data structures that are no longer needed when all SSA numbers have
//  been recorded in DEFs and USEs.
void SMPFunction::FreeSSAMemory(void) {
#if 0 // dominator info is useful for lots of CFG analyses
	this->IDom.clear();
	this->DomTree.clear();
#endif
	// this->BlocksDefinedIn.clear();
	this->SSACounter.clear();
	this->SSAStack.clear();

#if SMP_SHRINK_TO_FIT
#if 0
	vector<int>(this->IDom).swap(this->IDom);
	vector<pair<int, list<int> > >(this->DomTree).swap(this->DomTree);
#endif
	vector<list<int> >(this->BlocksDefinedIn).swap(this->BlocksDefinedIn);
	vector<int>(this->SSACounter).swap(this->SSACounter);
	vector<list<int> >(this->SSAStack).swap(this->SSAStack);
#endif

	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		(*BlockIter)->FreeSSAMemory();
	}
	return;
} // end of SMPFunction::FreeSSAMemory()

// For each instruction, mark the non-flags-reg DEFs as having live
//  metadata (mmStrata needs to fetch and track this metadata for this
//  instruction) or dead metadata (won't be used as addressing reg, won't
//  be stored to memory, won't be returned to caller, won't be passed to a callee).
void SMPFunction::AnalyzeMetadataLiveness(void) {
	bool changed;
	int BaseReg;
	int IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	list<SMPInstr *>::iterator InstIter;
	list<SMPInstr *>::reverse_iterator RevInstIter;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	set<DefOrUse, LessDefUse>::iterator NextUse;
	bool DebugFlag = false;
	bool UseFP = this->UsesFramePointer();
	int IterationCount = 0;

#if SMP_DEBUG_DATAFLOW
	if (0 == strcmp("uw_frame_state_for", this->GetFuncName())) {
		DebugFlag = true;
	}
#endif

	do {
		changed = false;
		++IterationCount;
		bool SafeMemDest;
		if (DebugFlag) {
			SMP_msg("AnalyzeMetadataLiveness iteration count: %d \n", IterationCount);
		}
		for (RevInstIter = this->Instrs.rbegin(); RevInstIter != this->Instrs.rend(); ++RevInstIter) {
			SMPInstr *CurrInst = (*RevInstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			SafeMemDest = false;  // true for some SafeFunc instructions
#if 0
			// Skip the SSA marker instruction.
			if (CurrInst->IsMarkerInst())
				continue;
#endif

			if (DebugFlag) {
				SMP_msg("Inst addr: %llx \n", (uint64_t) CurrInst->GetAddr());
			}
			CurrDef = CurrInst->GetFirstDef();
			while (CurrDef != CurrInst->GetLastDef()) {
				if (DEF_METADATA_UNANALYZED == CurrDef->GetMetadataStatus()) {
					STARSOpndTypePtr DefOp = CurrDef->GetOp();
					// Handle special registers never used as address regs.
					if (DefOp->MatchesReg(X86_FLAGS_REG) || DefOp->MDIsSpecialRegOpType()) {
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_UNUSED);
						changed = true;
					}
					else if (MDIsStackOrFramePointerReg(DefOp, UseFP)) {
						// Stack pointer register DEFs always have live
						//  metadata, but we don't need to propagate back
						//  through particular DEF-USE chains.
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_USED);
						changed = true;
					}
					else if (DefOp->IsMemOp()) {
						// DEF is a memory operand. The addressing registers
						//  therefore have live metadata, and the memory metadata is live.
						// EXCEPTION: If the function is Safe, then direct stack writes
						//  to local variables (above the outgoing args area of the frame)
						//  are not live metadata, and there will be no indirect local frame
						//  writes, by definition of "safe." So, for safe funcs, only
						//  the o_mem (globals) and indirect writes are live metadata.
						if (this->IsSafe() && MDIsStackAccessOpnd(DefOp, this->UseFP)
							&& (!this->WritesAboveLocalFrame(DefOp, CurrInst->AreDefsNormalized(), InstAddr))
							&& (!this->IsInOutgoingArgsRegion(DefOp))) {
							++CurrDef;
							SafeMemDest = true;
							continue;
						}
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_USED);
						changed = true;
						MDExtractAddressFields(DefOp, BaseReg, IndexReg,
							ScaleFactor, offset);
						if (STARS_x86_R_none != BaseReg) {
							STARSOpndTypePtr BaseOp = CurrInst->MakeRegOpnd(MDCanonicalizeSubReg((STARS_regnum_t) BaseReg));
							if (MDIsStackOrFramePointerReg(BaseOp, UseFP)) {
								; // do nothing; DEF handled by case above
							}
							else {
								CurrUse = CurrInst->FindUse(BaseOp);
								if (CurrUse == CurrInst->GetLastUse()) {
									SMP_msg("FATAL ERROR: BaseReg %d not in USE list at %llx for %s\n",
										BaseOp->GetReg(), (uint64_t) CurrInst->GetAddr(), CurrInst->GetDisasm());
									this->DumpDotCFG();
									this->Dump(false);
									assert(CurrUse != CurrInst->GetLastUse()); // kablooey!
								}
								if (this->IsGlobalName(BaseOp)) {
									changed |= this->PropagateGlobalMetadata(BaseOp,
										DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
								}
								else {
									changed |= CurrInst->GetBlock()->PropagateLocalMetadata(BaseOp,
										DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
								}
							}
						} // end if STARS_x86_R_none != BaseReg
						if (STARS_x86_R_none != IndexReg) {
							STARSOpndTypePtr IndexOp = CurrInst->MakeRegOpnd(MDCanonicalizeSubReg((STARS_regnum_t) IndexReg));
							if (MDIsStackOrFramePointerReg(IndexOp, UseFP)) {
								; // do nothing; DEF handled by case above
							}
							else {
								CurrUse = CurrInst->FindUse(IndexOp);
								if (CurrUse == CurrInst->GetLastUse()) {
									SMP_msg("FATAL ERROR: IndexReg %d not in USE list at %llx for %s\n",
										IndexOp->GetReg(), (uint64_t) CurrInst->GetAddr(),	CurrInst->GetDisasm());
									this->DumpDotCFG();
									this->Dump(false);
								}
								assert(CurrUse != CurrInst->GetLastUse()); // could be kablooey! if FATAL ERROR above
								if (0 != ScaleFactor) {
									; // mmStrata knows scaled reg is NUMERIC
									// ... its metadata is not fetched
								}
								else if (this->IsGlobalName(IndexOp)) {
									changed |= this->PropagateGlobalMetadata(IndexOp,
										DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
								}
								else {
									changed |= CurrInst->GetBlock()->PropagateLocalMetadata(IndexOp,
										DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
								}
							}
						} // end if STARS_x86_R_none != IndexReg
					} // end if X86_FLAGS_REG .. else if stack ptr ... 
				} // end if unanalyzed metadata usage
				++CurrDef;
			} // end while processing DEFs
			if ((RETURN == CurrInst->GetDataFlowType())
				|| (CurrInst->IsTailCall())   // quasi-return
				|| (CurrInst->IsBranchToOtherFunc())
				|| (CALL == CurrInst->GetDataFlowType())
				|| (INDIR_CALL == CurrInst->GetDataFlowType())) {
				// The EAX and EDX registers can be returned to the caller,
				//  which might use their metadata. They show up as USEs
				//  of the return instruction. Some library functions
				//  pass return values in non-standard ways. e.g. through
				//  EBX or EDI, so we treat all return regs the same.
				// For CALL instructions, values can be passed in caller-saved
				//  registers, unfortunately, so the metadata is live-in.
				CurrUse = CurrInst->GetFirstUse();
				while (CurrUse != CurrInst->GetLastUse()) {
					NextUse = CurrUse;
					++NextUse;
					STARSOpndTypePtr ReturnOp = CurrUse->GetOp();
					if (DebugFlag) {
						SMP_msg("ReturnOp: ");
						PrintOperand(ReturnOp);
						SMP_msg("\n");
					}
					if (ReturnOp->IsRegOp() &&
						(!MDIsStackOrFramePointerReg(ReturnOp, UseFP)) &&
						(!ReturnOp->MatchesReg(X86_FLAGS_REG))) {
						if (this->IsGlobalName(ReturnOp)) {
							changed |= this->PropagateGlobalMetadata(ReturnOp,
									DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
						}
						else {
							changed |= CurrInst->GetBlock()->PropagateLocalMetadata(ReturnOp,
									DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
						}
					}
					CurrUse = NextUse;
				} // end while all USEs
			} // end if return or call
			else if (CurrInst->HasDestMemoryOperand() 
				|| CurrInst->MDIsPushInstr()) {
				// Memory writes cause a lot of metadata usage.
				//  Addressing registers in the memory destination
				//  have live metadata used in bounds checking. The
				//  register being stored to memory could end up being
				//  used in some other bounds checking, unless we 
				//  have precise memory tracking and know that it
				//  won't.
				// We handled the addressing registers above, so we
				//  handle the register written to memory here.
				// The same exception applies as above: If the destination
				//  memory operand is not a stack write, then safe functions
				//  do not need to track the metadata.
				// If we push a register and have callees, the metadata could
				//  be live, if the callee gets its incoming args from our push
				//  instructions.
				if (SafeMemDest && !(CurrInst->MDIsPushInstr() && !this->IsLeaf())) {
					continue;  // go to next instruction
				}
				CurrUse = CurrInst->GetFirstUse();
				while (CurrUse != CurrInst->GetLastUse()) {
					NextUse = CurrUse;
					++NextUse;
					STARSOpndTypePtr UseOp = CurrUse->GetOp();
					// NOTE: **!!** To be less conservative, we
					//  should propagate more for exchange category
					//  instructions.
					if ((UseOp->IsRegOp()) && (!MDIsStackOrFramePointerReg(UseOp, UseFP)) && (!UseOp->MatchesReg(X86_FLAGS_REG))) {
						if (this->IsGlobalName(UseOp)) {
							changed |= this->PropagateGlobalMetadata(UseOp,
									DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
						}
						else {
							changed |= CurrInst->GetBlock()->PropagateLocalMetadata(UseOp,
									DEF_METADATA_USED, CurrUse->GetSSANum(), InstAddr);
						}
					} // end if register
					CurrUse = NextUse;
				} // end while all USEs
			} // end if call or return else if memdest ...
		} // end for all instructions
	} while (changed);

	// All DEFs that still have status DEF_METADATA_UNANALYZED can now
	//  be marked as DEF_METADATA_UNUSED.
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
#if 0   // important to know if incoming values are dead metadata
		if (CurrInst->IsMarkerInst())
			continue;
#endif
		CurrDef = CurrInst->GetFirstDef();
		while (CurrDef != CurrInst->GetLastDef()) {
			SMPMetadataType CurrStatus = CurrDef->GetMetadataStatus();
			if (DEF_METADATA_UNANALYZED == CurrStatus) {
				CurrDef = CurrInst->SetDefMetadata(CurrDef->GetOp(),
					DEF_METADATA_UNUSED);
				assert(CurrDef != CurrInst->GetLastDef());
				++DeadMetadataCount;
			}
			else if (DEF_METADATA_USED <= CurrStatus) {
				++LiveMetadataCount;
			}
			else {
				;  // must be UNUSED, but only because it is the flags register. Don't count.
			}
			++CurrDef;
		}
	}

	return;
} // end of SMPFunction::AnalyzeMetadataLiveness() 

// Propagate the metadata Status for UseOp/SSANum to its global DEF.
// Return true if successful.
bool SMPFunction::PropagateGlobalMetadata(const STARSOpndTypePtr &UseOp, SMPMetadataType Status, int SSANum, STARS_ea_t UseAddr) {
	bool changed = false;
	bool UseFP = this->UsesFramePointer();

	if ((0 > SSANum) || (!MDIsDataFlowOpnd(UseOp, UseFP)))
		return false;

	// Find the DEF of UseOp with SSANum.
	STARS_ea_t DefAddr;
	SMPBasicBlock *UseBlock;
	SMPBasicBlock *DefBlock;

	if (!STARS_IsBlockNumPseudoID(UseAddr)) { // UseAddr is an inst addr
		UseBlock = this->GetBlockFromInstAddr(UseAddr);
	}
	else { // UseAddr is a block number
		UseBlock = this->GetBlockByNum(STARS_GetBlockNumFromPseudoID(UseAddr));
	}

	STARSOpndTypePtr DefMoveOp = nullptr;
	DefAddr = UseBlock->GetUltimateDefAddr(UseOp, UseAddr, SSANum, false, false, false, DefMoveOp);

	if (STARS_BADADDR == DefAddr) {
		return changed;
	}

	if (STARS_LIVEIN_PSEUDO_ID == DefAddr) {
		// Seems to only happen on STARS/IRDB. Need to debug later.  !!!!****!!!!
		return changed;
	}

	if (!STARS_IsBlockNumPseudoID(DefAddr)) { // found a DEF inst.
		SMPInstr *CurrInst = this->GetInstFromAddr(DefAddr);
		STARS_ea_t InstAddr = DefAddr;
		DefBlock = this->GetBlockFromInstAddr(DefAddr);
		set<DefOrUse, LessDefUse>::iterator CurrDef = CurrInst->FindDef(UseOp);
		set<DefOrUse, LessDefUse>::iterator CurrUse;
		assert(CurrDef != CurrInst->GetLastDef());
		assert(SSANum == CurrDef->GetSSANum());
		if (Status != CurrDef->GetMetadataStatus()) {
			CurrDef = CurrInst->SetDefMetadata(UseOp, Status);
			changed = (CurrDef != CurrInst->GetLastDef());
			bool PropThroughUses = changed;

			// If source operand was memory, we have two cases.
			//  (1) The instruction could be a load, in which
			//  case we should simply terminate the
			//  propagation, because the prior DEF of a memory
			//  location is always considered live metadata
			//  already, and we do not want to propagate liveness
			//  to the address regs in the USE list.
			//  EXCEPTION: For safe funcs, we propagate liveness
			//   for stack locations.
			//  (2) We could have an arithmetic operation such
			//  as reg := reg arithop memsrc. In this case, we
			//  still do not want to propagate through the memsrc,
			//  (with the same safe func EXCEPTION),
			//  but the register is both DEF and USE and we need
			//  to propagate through the register.
			if (CurrInst->HasSourceMemoryOperand()) {
				if (this->IsSafe()) {
					STARSOpndTypePtr MemSrcOp = CurrInst->MDGetMemUseOp();
					assert(! MemSrcOp->IsVoidOp());
					if (MDIsDirectStackAccessOpnd(MemSrcOp, this->UseFP)) {
						// We have a SafeFunc stack access. This is
						//  the EXCEPTION case where we want to
						//  propagate metadata liveness for a memory
						//  location.
						CurrUse = CurrInst->FindUse(MemSrcOp);
						assert(CurrUse != CurrInst->GetLastUse());
						if (this->IsGlobalName(MemSrcOp)) {
							changed |= this->PropagateGlobalMetadata(MemSrcOp,
								Status, CurrUse->GetSSANum(), InstAddr);
						}
						else {
							changed |= CurrInst->GetBlock()->PropagateLocalMetadata(MemSrcOp,
								Status, CurrUse->GetSSANum(), InstAddr);
						}
					} // end if stack access operand
				} // end if SafeFunc
				if (3 == CurrInst->GetOptType()) { // move inst
					PropThroughUses = false; // load address regs are not live metadata
				}
				else if ((5 == CurrInst->GetOptType())
					|| (STARS_NN_and == CurrInst->GetIDAOpcode())
					|| (STARS_NN_or == CurrInst->GetIDAOpcode())
					|| (STARS_NN_xor == CurrInst->GetIDAOpcode())) {
					// add, subtract, and, or with memsrc
					// Find the DEF reg in the USE list.
					CurrUse = CurrInst->FindUse(UseOp);
					assert(CurrUse != CurrInst->GetLastUse());
					changed |= this->PropagateGlobalMetadata(UseOp,
						Status, CurrUse->GetSSANum(), InstAddr);
					PropThroughUses = false;
				}
			} // end if memory source

			// Now, propagate the metadata status to all the
			//  non-memory, non-flags-reg, non-special-reg 
			//  (i.e. regular registers) USEs.
			if (PropThroughUses) {
				CurrUse = CurrInst->GetFirstUse();
				while (CurrUse != CurrInst->GetLastUse()) {
					STARSOpndTypePtr CurrUseOp = CurrUse->GetOp();
					// NOTE: **!!** To be less conservative, we
					//  should propagate less for exchange category
					//  instructions.
					if ((CurrUseOp->IsRegOp()) && (!MDIsStackOrFramePointerReg(CurrUseOp, UseFP))
						&& (!CurrUseOp->MatchesReg(X86_FLAGS_REG))) {

						if (this->IsGlobalName(CurrUseOp)) {
							changed |= this->PropagateGlobalMetadata(CurrUseOp,
								Status, CurrUse->GetSSANum(), InstAddr);
						}
						else {
							changed |= CurrInst->GetBlock()->PropagateLocalMetadata(CurrUseOp,
								Status, CurrUse->GetSSANum(), InstAddr);
						}
					}
					++CurrUse;
				} // end while all USEs
			}
		}
	}

	else { // Found a DEF block number in DefAddr.
		// Check the Phi functions
		DefBlock = this->GetBlockByNum(STARS_GetBlockNumFromPseudoID(DefAddr));
		set<SMPPhiFunction, LessPhi>::iterator DefPhi;
		DefPhi = DefBlock->FindPhi(UseOp);
		assert(DefPhi != DefBlock->GetLastPhi());
		assert(SSANum == DefPhi->GetDefSSANum());
		if (Status != DefPhi->GetDefMetadata()) {
			DefPhi = DefBlock->SetPhiDefMetadata(UseOp, Status);
			changed = true;
			// If the Phi DEF has live metadata, then the Phi
			//  USEs each have live metadata. Propagate.
			int UseSSANum;
			for (std::size_t index = 0; index < DefPhi->GetPhiListSize(); ++index) {
				UseSSANum = DefPhi->GetUseSSANum(index);
				// UseSSANum can be -1 in some cases because
				//  we conservatively make EAX and EDX be USEs
				//  of all return instructions, when the function
				//  might have a void return type, making it
				//  appear as if an uninitialized EAX or EDX
				//  could make it to the return block.
				if (0 <= UseSSANum) {
					changed |= this->PropagateGlobalMetadata(UseOp,	Status, UseSSANum, DefAddr);
				}
			}
		}

	} // end if (DefAddr is inst addr) else ... [DefAddr is block number]

	return changed;
} // end of SMPFunction::PropagateGlobalMetadata()

// Find consecutive DEFs of the same type and mark the second one redundant.
void SMPFunction::FindRedundantMetadata(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	bool changed = false;

	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		changed |= CurrBlock->FindRedundantLocalMetadata(this->IsSafe());
	}
	return;
} // end of SMPFunction::FindRedundantMetadata()

// Perform SCCP to find constant values for DEFs, store in this->ConstantDefs
void SMPFunction::SparseConditionalConstantPropagation(void) {
	// We perform the SCCP (Sparse Conditional Constant Propagation) algorithm
	//  as found in Cooper & Torczon, "Engineering a Compiler."
#if STARS_DEBUG_FUNC_SCCP
	bool DebugFlag = (0x8049d20 == this->GetFirstFuncAddr());
#endif

	// CFGWorkList := { all edges from pseudo-entry block to entry blocks }
	//  We do not have a pseudo-entry block, so we special case by starting processing at
	//  the first basic block, block number 0, which is our entry block.
	list<pair<int, int> > CFGWorkList; // edges from pair.first = source block number to pair.second = dest block number
	pair<int, int> InitEdge(-1, 0); // -1 is pseudo-block-number
	CFGWorkList.push_back(InitEdge);
	std::size_t BlockNumLimit = this->Blocks.size();
	vector<STARSBitSet> ExecutedEdgeBitSet(BlockNumLimit); // records which edges have been executed in SCCP; row = DestBlockNum, col (bit) = SrcBlockNum
	vector<STARSBranchConst> CurrentBranchStatus;
	CurrentBranchStatus.resize(BlockNumLimit);

	// for each edge e in the CFG
	//  mark e as unexecuted
	for (std::size_t EdgeIndex = 0; EdgeIndex < BlockNumLimit; ++EdgeIndex) {
		ExecutedEdgeBitSet[EdgeIndex].AllocateBits(BlockNumLimit); // allocate and zero all bits
		CurrentBranchStatus[EdgeIndex] = STARS_BRANCH_UNKNOWN;
	}
	this->ResetSCCPVisitedBlocks();  // records which blocks have been visited in SCCP algorithm

	// SSAWorkList := { empty set }
	list<pair<int, int> > SSAWorkList; // pair.first = block number, pair.second = name+SSA hash
	SSAWorkList.clear();

	// for each ref def x in the procedure
	//   Value(x) = TOP
	//  We currently implement this by having this->ConstantDefs contain no entry for defs with value TOP

	// while ((CFGWorkList is not empty) or (SSAWorkList is not empty))
	while (!(CFGWorkList.empty() && SSAWorkList.empty())) {

	//   if (CFGWorkList is not empty) then
	//       remove an edge e = (m, n) from the CFGWorkList
	//       if (e is marked as unexecuted) then
	//          mark e as executed
	//          EvaluateAllPhisInBlock(n)
	//          if (e is only edge into n marked as executed) then [i.e. first visit]
	//              for (each instruction i in block n) do
	//                  if (i is an assignment) then
	//                      EvaluateAssign(i)
	//                  else if (i is a conditional branch) then
	//                      EvaluateConditional(i)
	//                  endif
	//              endfor
	//              Put block successors on CFGWorkList, based on conditional branch evaluation if any
	//          endif
	//       endif
	//   endif

	//   if (CFGWorkList is not empty) then
		if (!(CFGWorkList.empty())) {
	//       remove an edge e = (m, n) from the CFGWorkList
			pair<int, int> CurrentEdge = CFGWorkList.front();
			CFGWorkList.pop_front();
	//       if (e is marked as unexecuted) then
			int SrcBlockNum = CurrentEdge.first;
			int DestBlockNum = CurrentEdge.second;
			bool UnexecutedEdge = (0 > CurrentEdge.first);
#if STARS_DEBUG_FUNC_SCCP_VERBOSE
			if (0x807a530 == this->FirstEA) {
				SMP_msg("DEBUG: SCCP: Processing CFG edge from %d to %d\n", SrcBlockNum, DestBlockNum);
			}
#endif
			if (!UnexecutedEdge) {
				UnexecutedEdge = (!ExecutedEdgeBitSet[(size_t)DestBlockNum].GetBit((std::size_t) SrcBlockNum));
			}
#if STARS_DEBUG_FUNC_SCCP
			if (DebugFlag) {
				SMP_msg("SCCP: Processing edge from block %d to block %d: UnexecutedEdge = %d\n", SrcBlockNum, DestBlockNum, UnexecutedEdge);
			}
#endif
			if (UnexecutedEdge) {
	//          mark e as executed
				if (0 <= SrcBlockNum) {
					ExecutedEdgeBitSet[(std::size_t) DestBlockNum].SetBit((std::size_t) SrcBlockNum);
				}
	//          EvaluateAllPhisInBlock(n)
				this->EvaluateAllPhiConstants(DestBlockNum, ExecutedEdgeBitSet, SSAWorkList);
	//          if (e is only edge into n marked as executed) then
				SMPBasicBlock *CurrBlock = this->GetBlockByNum(DestBlockNum);
				if (!(CurrBlock->IsSCCPVisited())) {
	//              for (each instruction i in block n) do
	//                  if (i is an assignment) then
	//                      EvaluateAssign(i)
	//                  else if (i is a conditional branch) then
	//                      EvaluateConditional(i)
	//                  endif
	//              endfor
					if (CurrBlock->IsUnreachableBlock()) { // must have a call 0 instruction
						SMP_msg("ERROR: Unreachable block is being visited by SCCP at %llx ", (uint64_t) CurrBlock->GetFirstAddr());
						this->DumpFuncNameAndAddr();
					}
					enum STARSBranchConst BranchEval = STARS_BRANCH_UNKNOWN;
					CurrBlock->SCCPEvaluateConstants(BranchEval, CFGWorkList, SSAWorkList); // also marks block as SCCP visited
					CurrentBranchStatus[DestBlockNum] = BranchEval;
	//          endif
				}
	//       endif
			}
	//   endif
		}


	//
	//   if (SSAWorkList is not empty) then
	//       remove an edge e = (s, d) from SSAWorkList
	//       c := CFG node that uses d
	//       if (any edge entering c is marked as executable) then
	//           if (d is a phi function argument) then
	//               EvaluatePhi(d)
	//           else
	//              for (each instruction i in block c) do
	//                  if (i is an assignment that uses d) then
	//                      EvaluateAssign(i)
	//                  else if (i is a conditional branch that uses d) then
	//                      EvaluateConditional(i)
	//                  endif
	//              endfor
	//           endif
	//       endif
	//   endif

	//   if (SSAWorkList is not empty) then
		if (!(SSAWorkList.empty())) {
	//       remove an edge e = (s, d) from SSAWorkList
			pair<int, int> SSAEdge = SSAWorkList.front();
			SSAWorkList.pop_front();
			int BlockNum = SSAEdge.first;
			int DefHashValue = SSAEdge.second;
	//       c := CFG node that uses d
			assert(0 <= BlockNum);
			SMPBasicBlock *CurrBlock = this->GetBlockByNum((std::size_t) BlockNum);
#if STARS_DEBUG_FUNC_SCCP_VERBOSE
			if (0x807a530 == this->FirstEA) {
				SMP_msg("DEBUG: SCCP: Processing SSA item block %d DefHash %x\n", BlockNum, DefHashValue);
			}
#endif
			STARSOpndTypePtr DefOp = (*(this->GetFirstInstIter()))->MakeRegOpnd((STARS_regnum_t)(DefHashValue & 0x0000ffff));
			unsigned int TempSSANum = (unsigned int) (DefHashValue & 0xffff0000);
			int DefSSANum = (int)(TempSSANum >> 16);
			bool LocalName = CurrBlock->IsLocalName(DefOp);
	//       if (any edge entering c is marked as executable) then
	//           if (d is a phi function argument) then
	//               EvaluatePhi(d)
	//           else
	//              for (each instruction i in block c) do
	//                  if (i is an assignment that uses d) then
	//                      EvaluateAssign(i)
	//                  else if (i is a conditional branch that uses d) then
	//                      EvaluateConditional(i)
	//                  endif
	//              endfor
	//           endif
	//       endif
			// NOTE: We differ slightly from the published implementation. We don't maintain an SSA def-use chain explicitly,
			//  so we add the current block to the SSAWorkList when we find a new constant DEF, and search that block for
			//  further USEs of that operand with that SSA number.
			assert(CurrBlock->IsSCCPVisited()); // a local name is only added to SSAWorkList from within the block
			vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst();
			if (!LocalName) {
				// Speed up search by starting at the DEF inst.
				InstIter = this->GetBlockInstIterBySSA(DefOp, DefSSANum);
			}
			else {
				STARS_ea_t DefAddr = CurrBlock->GetLocalDefAddrForRegHash(DefHashValue);
				InstIter = CurrBlock->GetInstIterFromAddr(DefAddr);
			}
			bool FoundReDEF = false;
			while (InstIter != CurrBlock->GetLastInst()) {
				SMPInstr *CurrInst = (*InstIter);
				assert(nullptr != CurrInst);
				set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->FindUse(DefOp);
				if (UseIter != CurrInst->GetLastUse()) { // operand is USEd; check SSANum
					int UseSSANum = UseIter->GetSSANum();
					if (UseSSANum == DefSSANum) {
						SMPitype DataFlowType = CurrInst->GetDataFlowType();
						if (DEFAULT == DataFlowType) {
							CurrInst->SCCPEvaluateAssignment(SSAWorkList);
						}
						else if (COND_BRANCH == DataFlowType) {
							enum STARSBranchConst BranchEval;
							CurrInst->SCCPEvaluateCondBranch(BranchEval);
							CurrentBranchStatus[BlockNum] = BranchEval;
							CurrBlock->SCCPHandleSuccessors(BranchEval, CFGWorkList);
						}
					}
				}
				set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->FindDef(DefOp);
				if (DefIter != CurrInst->GetLastDef()) { // check for re-DEF
					int NewDefSSANum = DefIter->GetSSANum();
					if (NewDefSSANum > DefSSANum) { // re-DEF; we can stop searching for USEs of DefOp/DefSSANum in this block
						FoundReDEF = true;
						break;
					}
				}
				++InstIter;
			} // end while not last inst in block
			// See if we need to search for other blocks that use DefOp/DefSSANum
			if (!(LocalName || FoundReDEF)) { // we have global name that is not redefined
				list<SMPBasicBlock *>::iterator SuccIter;
				for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
					SMPBasicBlock *SuccBlock = (*SuccIter);
					if (SuccBlock->IsSCCPVisited() && SuccBlock->IsLiveIn(DefOp)) {
						this->ResetProcessedBlocks();
						SuccBlock->SCCPGlobalPropagationHelper(DefOp, DefSSANum, ExecutedEdgeBitSet, CFGWorkList, SSAWorkList);
					}
				}
			}
	//   endif
		}
#if 0
		SSAWorkList.clear(); // temporary stub
#endif
	// endwhile
	}

	// Go back over the basic blocks and detect never-visited blocks. Label these as unreachable.
	list<SMPBasicBlock *>::iterator BlockIter;
	bool FoundUnreachableBlock = false;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		if (!(CurrBlock->IsSCCPVisited())) {
			vector<SMPInstr *>::iterator FirstInstIter = CurrBlock->GetFirstInst();
			SMPInstr *FirstInst = (*FirstInstIter);
			if (!this->GetProg()->IsCodeXrefFromData(FirstInst->GetAddr())) {
				CurrBlock->SetUnreachableBlock(true);
				FoundUnreachableBlock = true;
				SMP_fprintf(global_STARS_program->GetInfoAnnotFile(), "%18llx %6zu INSTR UNREACHABLE %s \n",
					(uint64_t) FirstInst->GetAddr(), FirstInst->GetSize(), FirstInst->GetDisasm());
#if STARS_DEBUG_FUNC_SCCP
				STARS_ea_t BlockAddr = CurrBlock->GetFirstAddr();
				SMP_msg("INFO: SCCP found unreachable block at %llx\n", (uint64_t) BlockAddr);
#endif
				this->GetProg()->AddBlockToRemovalList(CurrBlock);
#if STARS_SCCP_CONVERT_UNREACHABLE_BLOCKS
				CurrBlock->SCCPNullifyUnreachableBlock();
#endif
			}
		}
#if STARS_SCCP_GATHER_STATISTICS
		else { 
			vector<SMPInstr *>::iterator LastInstIter = CurrBlock->GetLastInst();
			--LastInstIter;
			SMPInstr *LastInst = (*LastInstIter);
			STARS_ea_t BranchAddr = LastInst->GetAddr();
			STARSBranchConst BranchEval = CurrentBranchStatus[CurrBlock->GetNumber()];
			if (STARS_BRANCH_ALWAYS_TAKEN == BranchEval) {
				SMP_fprintf(global_STARS_program->GetInfoAnnotFile(), "%18llx %6zu INSTR BRANCHALWAYSTAKEN %s \n",
					(uint64_t) BranchAddr, LastInst->GetSize(), LastInst->GetDisasm());
				this->AddResolvedBranch(BranchAddr);
			}
			else if (STARS_BRANCH_NEVER_TAKEN == BranchEval) {
				SMP_fprintf(global_STARS_program->GetInfoAnnotFile(), "%18llx %6zu INSTR BRANCHNEVERTAKEN %s \n",
					(uint64_t) BranchAddr, LastInst->GetSize(), LastInst->GetDisasm());
				this->AddResolvedBranch(BranchAddr);
			}
			if (CurrBlock->HasCallInstruction()) {
				CurrBlock->SCCPGatherStatistics();
			}
		}
#endif
	}

	return;
} // end of SMPFunction::SparseConditionalConstantPropagation() 

// emit debug output if SCCP found constant call target for an indir call or jump
void SMPFunction::AuditSCCPForIndirectTargets(void) {
	list<SMPInstr *>::iterator InstIter = this->GetFirstInstIter();
	++InstIter; // skip SSA Marker inst
	for (; InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		SMPitype CurrDataFlow = CurrInst->GetDataFlowType();
		if (((CurrDataFlow == INDIR_CALL) || (CurrDataFlow == INDIR_JUMP)) && (!CurrInst->IsInterruptCall())) {
			const STARSOpndTypePtr &TargetOp = CurrInst->GetControlFlowTarget();
			STARSOpndTypePtr SearchOp = CloneIfSubwordReg(TargetOp);
			CanonicalizeOpnd(SearchOp);
			if (TargetOp->IsRegOp()) { // can only find SCCP constant values for regs
				set<DefOrUse, LessDefUse>::iterator TargetIter = CurrInst->FindUse(SearchOp);
				int UseSSANum = TargetIter->GetSSANum();
				int UseHashValue = HashGlobalNameAndSSA(TargetOp, UseSSANum);
				STARS_SCCP_Const_Struct ConstStruct;
				ConstStruct.ConstType = STARS_CONST_TOP; // default to no const found
				if (CurrInst->GetBlock()->IsLocalName(TargetOp)) { // local name
					map<int, struct STARS_SCCP_Const_Struct>::iterator ConstValIter = CurrInst->GetBlock()->FindLocalConstValue(UseHashValue);
					if (ConstValIter != CurrInst->GetBlock()->GetLastLocalConstValueIter()) { // Has current const val entry
						ConstStruct = ConstValIter->second;
					}
				}
				else { // global name
					STARSSCCPMapIter ConstValIter = CurrInst->GetBlock()->GetFunc()->FindConstValue(UseHashValue);
					if (ConstValIter != CurrInst->GetBlock()->GetFunc()->GetLastConstValueIter()) { // Has current const val entry
						ConstStruct = ConstValIter->second;
					}
				}
				if (ConstStruct.ConstType == STARS_CONST_HAS_VALUE) {
					SMP_msg("INFO: SCCP found constant value of %llx for indirect target at %llx\n",
						(uint64_t) ConstStruct.ConstValue, (uint64_t) CurrInst->GetAddr());
					pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->ConstantIndirCalls.insert(CurrInst->GetAddr());
					assert(InsertResult.second);
				}
			}
		}
	}
	return;
} // end of SMPFunction::AuditSCCPForIndirectTargets()

// part of SCCP processing; propagate const DEFs into Phi USEs and Phi DEFs
void SMPFunction::EvaluateAllPhiConstants(int BlockNum, const vector<STARSBitSet> &ExecutedEdgeBitSet, list<pair<int, int> > &SSAWorkList) {
	// For all Phi DEFs of the type we are tracking for const values:
	//   If Phi DEF const value is not already the lattice bottom value:
	//     Accumulate const DEF values only for Phi USEs that correspond to incoming
	//      edges that are executed.
	//     If we have a consistent const value, set Phi DEF const value to it; if
	//      we have conflicting const values, set Phi DEF const value to type lattice bottom.
	//     If we set the Phi DEF const value to a new value, then propagate along SSA edges.

	// For all Phi DEFs of the type we are tracking for const values:
	SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	for (PhiIter = CurrBlock->GetFirstPhi(); PhiIter != CurrBlock->GetLastPhi(); ++PhiIter) {
		STARSOpndTypePtr PhiOp = PhiIter->GetAnyOp();
		if (!(PhiOp->IsRegOp()))   // !!!!****!!!! Add stack locations also in safe functions
			continue;
		int DefSSANum = PhiIter->GetDefSSANum();
		int DefHashValue = HashGlobalNameAndSSA(PhiOp, DefSSANum);
	//   If Phi DEF const value is not already the lattice bottom value:
		STARSSCCPMapIter ConstIter = this->FindConstValue(DefHashValue);
		STARSConstantValueType DefConstValType = STARS_CONST_TOP; // default; no entry in map
		if (ConstIter != this->GetLastConstValueIter()) { // found entry in map
			DefConstValType = ConstIter->second.ConstType;
		}
		if (DefConstValType != STARS_CONST_BOTTOM) {
	//     Accumulate const DEF values only for Phi USEs that correspond to incoming
	//      edges that are executed.
			STARS_uval_t ConstUseVal = 1;
			bool ConstUseValueSeen = false; // value seen on executed incoming edge Phi USE
			bool ConstNonValueSeen = false; // one executed incoming edge Phi USE had no value
			list<SMPBasicBlock *>::iterator PredIter;
			std::size_t PredIndex = 0;
			STARSConstantValueType DefFinalConstValType = DefConstValType;
			for (PredIter = CurrBlock->GetFirstPred(); PredIter != CurrBlock->GetLastPred(); ++PredIter, ++PredIndex) {
				int IncomingBlockNum = (*PredIter)->GetNumber();
				bool ExecutedIncomingEdge = ExecutedEdgeBitSet[BlockNum].GetBit(IncomingBlockNum);
				if (ExecutedIncomingEdge) {
					int UseSSANum = PhiIter->GetUseSSANum(PredIndex);
					int UseHashValue = HashGlobalNameAndSSA(PhiOp, UseSSANum);
	//     If we have a consistent const value, set Phi DEF const value to it; if
	//      we have conflicting const values, set Phi DEF const value to type lattice bottom.
					ConstIter = this->FindConstValue(UseHashValue);
					STARSConstantValueType UseConstValType = STARS_CONST_TOP; // default; no entry in map
					if (ConstIter != this->GetLastConstValueIter()) { // found entry in map
						UseConstValType = ConstIter->second.ConstType;
					}
					else {
						ConstNonValueSeen = true;
					}
					if (UseConstValType == STARS_CONST_HAS_VALUE) {
						if (ConstUseValueSeen) { // check for consistency of values
							if (ConstUseVal != ConstIter->second.ConstValue) { // inconsistent const values
								DefFinalConstValType = STARS_CONST_BOTTOM;
								break; // no need to see more Phi USEs
							}
						}
						else { // this is the first const value we have seen
							DefFinalConstValType = STARS_CONST_HAS_VALUE; // so far, anyway
							ConstUseValueSeen = true;
							ConstUseVal = ConstIter->second.ConstValue; // only value seen so far
						}
					}
					else if (UseConstValType == STARS_CONST_BOTTOM) {
						// Any BOTTOM value in a USE makes the DEF also become BOTTOM.
						DefFinalConstValType =  STARS_CONST_BOTTOM;
						break; // no need to see more Phi USEs
					}
					// else must be STARS_CONST_TOP, which is a don't-care case
				} // end if (ExecutedIncomingEdge)
			} // end for PredIter iteration through predecessor blocks

			// NOTE: It seems at first glance to be an error to propagate a new constant value
			//  from a Phi USE to the Phi DEF when there were Phi USEs that we ignored because
			//  their incoming edges were not executed. It seems that we are saying that:
			//  EDX20 := phi(EDX19, EDX21) produces EDX20 := 0 just because EDX19 == 0 even though
			//  EDX21 is still the lattice TOP value, having not been determined yet. In fact, this
			//  is the key feature of SCCP. On at least one path, EDX can have the value zero, so a
			//  path through the current block could occur with EDX == 0, and if that causes only one
			//  of the current block's successors to be added to the CFG WorkList, then the other successor
			//  still has a chance to be added later after the value of EDX21 is determined, because
			//  EvaluateAllPhiConstants() will be executed each time we enter the current block from
			//  an entry in the CFG WorkList. So, even if we have a test at the bottom of the block
			//  that determines the successor block based on EDX20 directly:
			//   test edx,edx
			//   jz label2
			//  all we are determining at the present time is that one possible path leads to label2, and
			//  the block at label2 is reachable and should be added to the work list. If EDX21 is later
			//  determined to have a non-zero value, then EvaluateAllPhiConstants() will be called again,
			//  producing the BOTTOM lattice value for EDX20, and both successor blocks of the current block
			//  will be placed on the CFG WorkList. Thus, even though the current invocation seems to lead to
			//  deciding that the branch to label2 is always taken, by using the SSAWorkList to propagate EDX==0
			//  down to the bottom of the block, this is not a final determination.

	//     If we set the Phi DEF const value to a new value, then propagate along SSA edges.
			if (!ConstNonValueSeen && (DefFinalConstValType != DefConstValType)) { // const entry is changing
				struct STARS_SCCP_Const_Struct NewConstEntry;
				NewConstEntry.ConstType = DefFinalConstValType;
				NewConstEntry.ConstValue = ConstUseVal;
				if (DefConstValType == STARS_CONST_TOP) { // there was no old map entry; insert new one
					pair<int, struct STARS_SCCP_Const_Struct> NewMapEntry(DefHashValue, NewConstEntry);
					pair<map<int, struct STARS_SCCP_Const_Struct>::iterator, bool> InsertResult = this->ConstantDefs.insert(NewMapEntry);
					assert(InsertResult.second);
				}
				else { // old map entry needs to be changed
					this->ConstantDefs[DefHashValue] = NewConstEntry;
				}
				// Propagate along SSA edges.
				pair<int, int> SSAEdge(BlockNum, DefHashValue);
				SSAWorkList.push_back(SSAEdge);
			} // end if entry is changing
		} // end if previous DEF value was not already BOTTOM
	} // end for all Phi functions
	return;
} // end of SMPFunction::EvaluateAllPhiConstants()

// Do we not care if DEF underflowed, due to how it is used?
bool SMPFunction::IsBenignUnderflowDEF(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, int &IdiomCode) {
	bool benign = false;
	list<SMPInstr *>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator DefIter, UseIter;
	int UseSSANum;
	SMPOperandType DefType;

	// We are looking to suppress overflow and underflow warnings on the following
	//  code sequence: PTR1-PTR2+1 gets a loop invariant code motion optimization
	//  that pulls  temp := 1-PTR2 out of the loop, and leaves temp2 := PTR1+temp
	//  inside the loop. The hoisted subtraction could underflow, and the addition
	//  that is not hoisted could overflow. The net effect of these two instructions
	//  is benign, however, so we want to suppress underflow and overflow checks on
	//  both of them, but only if we can match the pair of instructions.
	
	// We know that DefOp/DefAddr/DefSSANum refer to a subtraction instruction that
	//  produces a NEGATEDPTR result. We only need to find the paired addition instruction
	//  that USEs the same SSA name to produce a PTROFFSET result to prove that we have
	//  a case of benign underflow and overflow. If we find such a pair, we will mark
	//  both of their DEF results as benign overflows to suppress overflow checks.

	// PAINFUL: Linear search of instructions. Need to address this in the future.
	//  Perhaps we should have a map of UseHashValue to InstAddr, but that is more
	//  memory consumption. Sure would be useful, though.
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		UseIter = CurrInst->FindUse(DefOp);
		if (UseIter == CurrInst->GetLastUse()) {
			continue;
		}
		UseSSANum = UseIter->GetSSANum();
		if (UseSSANum == DefSSANum) {
			// Only remaining question: Do we produce a PTROFFSET in CurrInst? (If we do,
			//  that implies we had an addition, so we don't need to check that.)
			DefIter = CurrInst->GetFirstNonFlagsDef();
			if (DefIter != CurrInst->GetLastDef()) {
				DefType = DefIter->GetType();
				// NOTE: Make this more general. What if we just move the NEGATEDPTR into a register
				//  and then the next instruction, with different SSA name, produces the PTROFFSET?
				//  !!!!!*****!!!!!
				if (IsEqType(DefType, PTROFFSET)) {
					// Found a pair. Mark both DEFs as benign and return true.
					benign = true;
					IdiomCode = 4;
					// Note that we have two possibilities for the addition. The NEGATEDPTR could be
					//  both the DEF and a USE, e.g. add negptr,ptr1; or the NEGATEDPTR could be
					//  just a USE, e.g. add reg,negptr, so that reg is overwritten and becomes a
					//  PTROFFSET. It really does not matter. The point is that we want to ignore
					//  overflow on this addition, and also on the subtraction that produced the
					//  NEGATEDPTR, so we mark the DEF in each instruction as benignly overflowing.
					STARSOpndTypePtr UseInstDefOp = DefIter->GetOp();
					DefIter = CurrInst->SetDefNoOverflow(UseInstDefOp, true);
					SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
					(void) DefInst->SetDefNoOverflow(DefOp, true);
					break;
				}
			}
		}
	}

	return benign;
} // end of SMPFunction::IsBenignUnderflowDEF()

bool SMPFunction::HasIntErrorCallSink(const STARSOpndTypePtr &DefOp, int DefSSANum, STARS_ea_t DefAddr, std::string &SinkString, bool &FoundAnyCall) {
	this->ResetProcessedBlocks(); // prepare for recursion through blocks
	SinkString.clear();
	SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(DefAddr);
	assert(CurrBlock != nullptr);

	bool FoundSink = CurrBlock->IsCriticalSink(DefOp, DefSSANum, SinkString, FoundAnyCall);

	return FoundSink;
} // end of SMPFunction::HasIntErrorCallSink()

// Recompute LVA and SSA and all dependent data structures now that unreachable blocks have been removed.
void SMPFunction::RecomputeSSA(void) {
	this->FreeSSAMemory();
	this->LoopCount = 0;
	this->LoopTestBlocksByLoopNum.clear();
	this->LoopTypesByLoopNum.clear();
	this->FuncLoopsByBlock.clear();
	this->FuncBlocksByLoop.clear();
	this->LoopNests.clear();
	this->GlobalDefAddrBySSA.clear();
	this->LoopInductionVars.clear();
#if 1 // need to debug
	this->RPONumberBlocks(); // Re-do the RPO numbering now that a block is removed.
#endif
	this->CallerSavedLocalRegsBitmap.reset(); // reset all bits to zero
	this->LiveVariableAnalysis(true);
	this->ComputeSSA();
	return;
} // end of SMPFunction::RecomputeSSA()

// Compute SSA form data structures across the function.
void SMPFunction::ComputeSSA(void) {
	bool DebugFlag = false;
	bool DumpFlag = false;
#if SMP_DEBUG_DATAFLOW
	DumpFlag |= (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
	DebugFlag |= (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
#endif

#if 1
	if (DumpFlag)
		this->Dump(false);
#endif
	if (DebugFlag) SMP_msg("Computing global names.\n");
	this->ComputeGlobalNames();
	if (DebugFlag) SMP_msg("Computing IDoms.\n");
	this->ComputeIDoms();
	if (DebugFlag) SMP_msg("Computing Dom frontiers.\n");
	this->ComputeDomFrontiers();
	if (DebugFlag) SMP_msg("Computing blocks defined in.\n");
	this->ComputeBlocksDefinedIn();

	if (DebugFlag) SMP_msg("Inserting Phi functions.\n");
	this->InsertPhiFunctions();
	if (DebugFlag) SMP_msg("Building dominator tree.\n");
	this->BuildDominatorTree();
	this->HasReducibleCFG = this->TestCFGReducibility();
	if (this->HasReducibleControlFlow()) {
		this->HasStructuredCFG = (!this->HasUnresolvedIndirectJumps());
		if (!this->HasStructuredControlFlow()) {
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to unresolved indirect jumps ");
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
		}
	}
	else {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to irreducible CFG ");
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
	}
	if (DebugFlag) SMP_msg("Computing SSA renumbering.\n");
	this->SSARenumber();
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	int LoopHeadBlockNum = SMP_BLOCKNUM_UNINIT;
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		this->FindLoopHeadsAndTails(CurrBlock);

		if (DumpFlag) CurrBlock->Dump(false);

		if (DebugFlag) SMP_msg("Computing local names.\n");
		CurrBlock->SetLocalNames();
		if (DebugFlag) SMP_msg("Computing local SSA renumbering.\n");
		CurrBlock->SSALocalRenumber();
		if (DumpFlag) CurrBlock->Dump(false);

		bool SingleExprBlock = CurrBlock->FindSingleExpression();
		if (SingleExprBlock) {
			CurrBlock->SetSingleExpression();
		}
	}
	this->HasGoodSSA = true;

	// We have to mark dead regs after SetLocalNames() has run for all blocks
	//  so that the CallerSavedLocalRegs bitmap is complete.
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		if (DebugFlag) SMP_msg("Marking dead registers.\n");
		CurrBlock = (*BlockIter);
		CurrBlock->MarkDeadRegs();
	}

	this->DetectMultiEntryFunction();

	// Analyze system calls, if any.
	if (this->HasSystemCalls()) {
		this->AnalyzeSystemCalls();
	}

#if STARS_BUILD_LOOP_BITSET
	// Now that we know how many loops we have, we can allocate the loops data structures.
	assert(this->BlockCount == ((int) this->Blocks.size()));
	this->FuncLoopsByBlock.resize(this->BlockCount);
	for (std::size_t BlockIndex = 0; ((int) BlockIndex) < this->BlockCount; ++BlockIndex) {
		this->FuncLoopsByBlock[BlockIndex].AllocateBits(this->LoopCount);
	}

	this->FuncBlocksByLoop.resize(this->LoopCount);
	this->LoopNests.resize(this->LoopCount);
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		this->FuncBlocksByLoop[LoopIndex].AllocateBits(this->BlockCount);
		this->LoopNests[LoopIndex].AllocateBits(this->LoopCount);
	}
	this->LoopTypesByLoopNum.resize(this->LoopCount);
	this->LoopTestBlocksByLoopNum.resize(this->LoopCount);

	if (this->LoopCount > 0) {
		this->DetectLoops();
		if (this->HasStructuredControlFlow()) {
			this->DetectLoopInvariantDEFs();
			if (global_STARS_program->ShouldSTARSPerformDeepLoopAnalyses()) {
				this->DetectLoopInductionVars2();
			}
		}
	}
	bool FoundSafeWrites = this->AnalyzeMemWriteSafety();
	if (FoundSafeWrites) {
		SMP_msg("INFO: Found safe indirect memory writes in func at %p\n", this->GetFirstFuncAddr());
	}
#endif

#if 0  // Need less conservative USEs on call insts to get good uninit var results
	this->DetectUninitializedVars();
#endif

#if SMP_DEBUG_DATAFLOW
	if (DumpFlag)
		this->Dump(false);
#endif

	// Compute the loop nests.
	this->ComputeLoopNests();

	// Emit loop nest annotations.
	this->EmitLoopNestAnnotations();

	// Once SSA numbers have been set into all DEFs, USES, and DU-chains, then
	//  the SSA numbering data structures will no longer be used and can be
	//  de-allocated.
	this->FreeSSAMemory();
	return;
} // end of SMPFunction::ComputeSSA()

// Detect loop head and/or tail block for loops and record in data structures
void SMPFunction::FindLoopHeadsAndTails(SMPBasicBlock *CurrBlock) {
	bool LoopTail = false;
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	int HeaderBlockNum = SMP_BLOCKNUM_UNINIT;
	list<SMPBasicBlock *>::iterator SuccIter;
	for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
		int SuccBlockNum = (*SuccIter)->GetNumber();
		int OuterLoopHeadBlockNum = SMP_BLOCKNUM_UNINIT;
		this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		if (this->DoesBlockDominateBlock(SuccBlockNum, CurrBlock->GetNumber())) {
			CurrBlock->SetLoopTailBlock();
			(*SuccIter)->SetLoopHeaderBlock();
			HeaderBlockNum = SuccBlockNum;
			if (!LoopTail) {
				LoopTail = true;
				CurrBlock->SetLoopHeaderBlockNumberForTailBlock(SuccBlockNum);
			}
			else {
				CurrBlock->SetLoopDoubleTailBlock();
				SMP_msg("INFO: LoopDoubleTailBlock: block %d ", CurrBlock->GetNumber());
				this->DumpFuncNameAndAddr();
				// Which loop header block is outermost? It dominates the other header.
				this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				if (this->DoesBlockDominateBlock(SuccBlockNum, CurrBlock->GetLoopHeaderNumber())) {
					CurrBlock->SetOuterLoopHeaderBlockNumberForDoubleTailBlock(SuccBlockNum);
				}
				else {
					CurrBlock->SetOuterLoopHeaderBlockNumberForDoubleTailBlock(CurrBlock->GetLoopHeaderNumber());
					CurrBlock->SetLoopHeaderBlockNumberForTailBlock(SuccBlockNum);
				}
			}

			size_t LoopIndex;
			if (this->IsBlockNumLoopHeader(HeaderBlockNum, LoopIndex)) {
				// Multiple branches of an if-then-elsif-else can return to the top of a loop,
				//  so we can encounter loop tail blocks multiple times for a loop. We found
				//  the LoopHeadBlockNum already in our list, so this must be one of these cases.
				// See if we can provide a better follow node number from the current analysis.
				assert((0 <= LoopIndex) && (LoopIndex < this->LoopExitTargets.size()));
			}
			else {
				this->LoopHeadBlockNumbers.push_back(HeaderBlockNum);
				set<int> DummySet;
				DummySet.insert(FollowBlockNum);
				this->LoopExitTargets.push_back(DummySet);
				map<int, STARSBitSet> DummyMap;
				this->LoopExitTargetsReachabilitySets.push_back(DummyMap);
				++this->LoopCount;
				assert(this->LoopHeadBlockNumbers.size() == this->LoopCount);
				assert(this->LoopExitTargetsReachabilitySets.size() == this->LoopCount);
				// If this is a double tail block, we need to arrange Loop N == inner, loop N-1 == outer.
				if (CurrBlock->IsDoubleLoopTailBlock()) {
					OuterLoopHeadBlockNum = CurrBlock->GetOuterLoopHeaderNumberForDoubleTailBlock();
					int InnerLoopHeadBlockNum = CurrBlock->GetLoopHeaderNumber();
					if (OuterLoopHeadBlockNum != this->LoopHeadBlockNumbers[this->LoopCount - 1]) {
						assert(InnerLoopHeadBlockNum == this->LoopHeadBlockNumbers[this->LoopCount - 1]);
#if 0 // incorrect, unnecessary?
						this->LoopHeadBlockNumbers[this->LoopCount - 2] = InnerLoopHeadBlockNum;
						this->LoopHeadBlockNumbers[this->LoopCount - 1] = OuterLoopHeadBlockNum;
#endif
					}
				}
			}

		}
	}

	return;
} // end of SMPFunction::FindLoopHeadsAndTails()

// Is BlockNum already in this->LoopHeadBlockNumbers? If so, put index in LoopNum
bool SMPFunction::IsBlockNumLoopHeader(const int BlockNum, std::size_t &LoopNum) const {
	size_t LoopIndex;
	bool FoundHeader = false;
	for (LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		if (BlockNum == this->LoopHeadBlockNumbers[LoopIndex]) {
			FoundHeader = true;
			LoopNum = LoopIndex;
			break;
		}
	}
	return FoundHeader;
} // end of SMPFunction::IsBlockNumLoopHeader()
