/*
 * SMPFunction.cpp - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#include "SMPFunction_support.h"

// Find candidate block # for if-else follow node for HeadBlockNum; return -1 otherwise
int SMPFunction::FindConditionalFollowNode2(int HeadBlockNum) {
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	int ElseBlockNum = SMP_BLOCKNUM_UNINIT;
	int ThenBlockNum = SMP_BLOCKNUM_UNINIT;
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();

	// Guard against pathological code like jz $+2 which jumps and falls through
	//  to the same ensuing instruction.
	SMPBasicBlock *HeadBlock = this->RPOBlocks[HeadBlockNum];
	assert(nullptr != HeadBlock);
	if (1 == HeadBlock->GetNumSuccessors()) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("INFO: SPARK: Unstructured due to failure in FindConditionalFollowNode due to single successor in HeadBlockNum %d ", HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}

	// Screen out branches that go outside their current loop in an unstructured way.
	if (this->FindMultiLoopContinue(HeadBlockNum)) {
		return FollowBlockNum;
	}

	// A well-structured candidate would be the biggest RPO-number block that has >= 2 predecessors
	//  and has HeadBlockNum as its IDom. 
	// EXCEPTION #1: Return and loop-back instructions interrupt control flow. See below.
	// EXCEPTION #2: Short circuit conditionals can trick the IDom algorithm into choosing
	//   one of the coalesced blocks in the short circuit expression as the follow node, when the follow
	//   node should come after the entire short circuit expr. This happens if one branch ends in a 
	//   return or loop-back and does not converge into a follow node.

	STARS_ea_t HeadLastAddr = HeadBlock->GetLastAddr();
	bool HeadBlockIsInALoop = this->IsBlockInAnyLoop(HeadBlockNum);
	ControlFlowType LastCFType = this->GetControlFlowType(HeadLastAddr);
	bool ShortCircuit = (IsShortCircuitFlow(LastCFType));

	// Go through DomTree[HeadBlockNum] in reverse order until we find a well-structured candidate or fail.
#if 1
	if (!ShortCircuit)
#endif
	for (int CurrBlockNum : this->DomTree[HeadBlockNum].second) {
		if (2 <= this->RPOBlocks[(size_t) CurrBlockNum]->GetNumPredsMinusBackEdges()) { // candidate
			if (!(ShortCircuit && this->GetCFGBlockByNum((size_t)CurrBlockNum)->IsCoalesced())) {
				// Follow node needs to be at the same loop nesting level as the head block, else we
				//  could just be finding the convergence of two paths inside some unrelated loop.
				if (this->AreBlocksInSameLoops(HeadBlockNum, CurrBlockNum)) {
					FollowBlockNum = CurrBlockNum;
					break;
				}
			}
		}
	}

	// Audit result to find unstructured code. If any block >= HeadBlockNum and < FollowBlockNum
	//  in RPO order is dominated by HeadBlockNum but has a successor that exits the loops containing
	//  HeadBlockNum, then there is a jump that spans loops inside the conditional statement,
	//  which cannot be structured without transformation.
	//
	// Note that if the exiting of the loop is a branch to the FollowBlock, then we can potentially
	//  have:
	//  loop
	//    if cond1 then
	//       [code1]
	//       exit when cond2;
	//    else
	//       [code2]
	//       exit when cond3;
	//    endif;
	//  endloop;
	//  This case needs analysis below and is not a problem. The conditional follow node will be the same
	//    as the loop follow node.
	//
	if (HeadBlockIsInALoop) {
		int InnermostLoopNum = this->GetInnermostLoopNum(HeadBlockNum);
		assert(0 <= InnermostLoopNum);
		int InnerMostFollowBlockNum = SMP_BLOCKNUM_UNINIT;
		bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)InnermostLoopNum, InnerMostFollowBlockNum);
		if (MultiExitLoop) {
			InnerMostFollowBlockNum = SMP_BLOCKNUM_UNINIT; // error
		}
		bool HitLoopFollowBlock = false;
		// If we did not find a FollowBlockNum, this will be a zero-trip loop.
		for (int BlockIndex = HeadBlockNum; BlockIndex < FollowBlockNum; ++BlockIndex) {
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			if (this->DoesBlockDominateBlock(HeadBlockNum, BlockIndex)) { // inside conditional
				SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t)BlockIndex);
				list<SMPBasicBlock *>::const_iterator SuccIter;
				for (SuccIter = CurrBlock->GetFirstConstSucc(); SuccIter != CurrBlock->GetLastConstSucc(); ++SuccIter) {
					int SuccBlockNum = (*SuccIter)->GetNumber();
					if (this->IsBlockLoopExitTarget((size_t)InnermostLoopNum, SuccBlockNum)) {
						HitLoopFollowBlock = true;
						FollowBlockNum = MultiExitLoop ? SMP_BLOCKNUM_COMMON_RETURN : SuccBlockNum;
					}
					else if (!this->IsBlockInLoop(SuccBlockNum, (size_t)InnermostLoopNum)) {
						// We are exiting the innermost loop without finding InnerMostFollowBlockNum.
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("INFO: SPARK: Unstructured due to Loop-spanning branch to block %d from block %zu ", SuccBlockNum, BlockIndex);
							this->DumpFuncNameAndAddr();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
						this->DumpDotCFG();
						if (VerboseOutput)
							this->Dump(false);
						return FollowBlockNum;
					}
				}
#if 0  // audit below
				if (HitLoopFollowBlock)
					return FollowBlockNum;
#endif
			}
		}
	}

	if (0 <= FollowBlockNum) {
		if (this->FindUnstructuredIfThenElse(HeadBlockNum, ThenBlockNum, ElseBlockNum, FollowBlockNum)) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
		}
		return FollowBlockNum;
	}

#if 0
	// If we failed, see if it was because of both branches of an if-then-else
	//  terminating with a RETURN. A call to a function that does not return, such
	//  as exit() or abort(), also behaves like a return here; both are control flow sinks.
	if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) {
		IDomIndex = (int) this->IDom.size() - 1;
		size_t ReturnCounter = 0;
		int BlockNum = (int) this->RPOBlocks.size();
		int SaveBlockNum = SMP_BLOCKNUM_UNINIT;
		--BlockNum;
		while (BlockNum > HeadBlockNum) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[(size_t) BlockNum];
			if (CurrBlock->HasReturn() || CurrBlock->HasNonReturningCall()) {
				if (this->DoesBlockDominateBlock(HeadBlockNum, BlockNum)) {
					++ReturnCounter;
					SaveBlockNum = BlockNum;
				}
			}
			--BlockNum;
		}
	
		if (ReturnCounter > 1) {
			FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
		}
		else if (1 == ReturnCounter) {
			// We must have an if-then with a return at the end of the then-clause.
			//  Confirm this pattern in the CFG first.
			assert(0 <= SaveBlockNum);
			SMPBasicBlock *CurrBlock = this->RPOBlocks[(size_t) HeadBlockNum];
			// HeadBlockNum should have two successors. One dominates SaveBlockNum, and
			//  that is the then-clause direction; the other successor is the target of
			//  jumping around the then-clause and is our follow node.
			assert(2 == CurrBlock->GetNumSuccessors());
			list<SMPBasicBlock *>::const_iterator SuccBlock = CurrBlock->GetCondNonFallThroughSucc();
			// Normal case is to jump around the then-clause.
			int SuccBlockNum = (*SuccBlock)->GetNumber();
			if (!this->DoesBlockDominateBlock(SuccBlockNum, SaveBlockNum)) {
				FollowBlockNum = SuccBlockNum;
			}
			else { // test other branch; not sure how this could fail unless we have unstructured code.
				SuccBlock = CurrBlock->GetFallThroughSucc();
				int SuccBlockNum = (*SuccBlock)->GetNumber();
				if (!this->DoesBlockDominateBlock(SuccBlockNum, SaveBlockNum)) {
					FollowBlockNum = SuccBlockNum;
				}
			}
		}
	}
#endif

	// A similar case to the above is when both branches of an if-then-else flow (via back edges)
	//   into a loop header block for a loop that contains both branches.
	if ((SMP_BLOCKNUM_UNINIT == FollowBlockNum) && HeadBlockIsInALoop) {
		// See if we can find two blocks, each dominated by HeadBlockNum,
		//  each being loop tail blocks, with the same loop header block
		//  as a successor block, with the loop header block dominating HeadBlockNum.
		size_t LoopBackCounter = 0;
		int BlockNum = (int) this->RPOBlocks.size();
		int SaveLoopHeadBlockNum = SMP_BLOCKNUM_UNINIT;
		int LoopHeadBlockNum = SMP_BLOCKNUM_UNINIT;
		--BlockNum;
		while (BlockNum > HeadBlockNum) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[(size_t)BlockNum];
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			if (CurrBlock->IsLoopTailBlock() && this->DoesBlockDominateBlock(HeadBlockNum, BlockNum)) {
				if (CurrBlock->HasLoopHeadAsSuccessor(LoopHeadBlockNum)) {
					// Are we looping back to the loop that contains our conditional HeadBlockNum?
					this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
					if (this->DoesBlockDominateBlock(LoopHeadBlockNum, HeadBlockNum)) {
						if (0 == LoopBackCounter) {
							SaveLoopHeadBlockNum = LoopHeadBlockNum;
						}
						else {
							if (SaveLoopHeadBlockNum != LoopHeadBlockNum) {
#if 0
								SMP_msg("WARNING: Inconsistent loop-back HeadBlockNums for conditional block %d in func at %llx\n",
									HeadBlockNum, (uint64_t) this->GetFirstFuncAddr());
								this->SetHasUnstructuredCFG();  // Inconsistent loop head block nums; irrecoverable problem
								return SMP_BLOCKNUM_UNINIT;
#else
								LoopBackCounter = 0;
								FollowBlockNum = SMP_BLOCKNUM_UNINIT;
								break; // try to succeed using TrackConditionalBranchTerminus() below
#endif
							}
						}
						++LoopBackCounter;
					}
					else {
						; // Test for loop being an inner loop within conditional??
					}
				}
			}
			--BlockNum;
		} // end of while (BlockNum > HeadBlockNum)
		if (2 <= LoopBackCounter) { // success; 2 or more if-elsif-else branches loop back to loop header
			// Screen out unstructured code. If we are jumping outside the loop containing the
			//  conditional HeadBlockNum from two or more blocks dominated by HeadBlockNum, then
			//  the code is doing a multi-level loop back, which is unstructured.
			if (this->AreBlocksInSameLoops(HeadBlockNum, SaveLoopHeadBlockNum)) {
				// FollowBlockNum = SaveLoopHeadBlockNum;
				FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
			}
			else { // unstructured, multi-level loop back
				if (!this->PrintedSPARKUnstructuredMsg) {
					SMP_msg("INFO: SPARK: Unstructured due to multi-level loop backs to block %d ", SaveLoopHeadBlockNum);
					this->DumpFuncNameAndAddr();
					this->PrintedSPARKUnstructuredMsg = true;
				}
				FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
				return FollowBlockNum;
			}
		}
	}

	if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) {
		// See if we can track the two branches of the COND_BRANCH to a sensible follow node.
		SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t) HeadBlockNum);
		assert(nullptr != CurrBlock);
		vector<SMPInstr *>::const_iterator LastInstIter = CurrBlock->GetLastConstInst();
		--LastInstIter;
		SMPInstr *BranchInst = (*LastInstIter);
		if (BranchInst->IsBranchToFarChunk()) {
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("INFO: SPARK: Unstructured due to failure in FindConditionalFollowNode for FarBranch in HeadBlockNum %d ", HeadBlockNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
			FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
		}
		else {
			assert(2 == CurrBlock->GetNumSuccessors());
			int ElseIDom = SMP_BLOCKNUM_UNINIT;
			int ThenIDom = SMP_BLOCKNUM_UNINIT;
			bool IfThenElseCase = true;

			if (!ShortCircuit) {
				ElseBlockNum = (*(CurrBlock->GetCondNonFallThroughSucc()))->GetNumber();
				assert(SMP_BLOCKNUM_UNINIT != ElseBlockNum);
				ThenBlockNum = (*(CurrBlock->GetCondOtherSucc(ElseBlockNum)))->GetNumber();
				assert(SMP_BLOCKNUM_UNINIT != ThenBlockNum);

				// For an if-then-else, both successors should be immediately dominated by the branch header block.
				//  If one is immediately dominated but the other is not, then the non-dominated block is the follow block
				//  and we just have an if-then structure, e.g.
				//    branch_block
				//        |
				//        +-----+
				//        |     |
				//   |    |    then_block
				//   |    |     |
				//   +--block2--+
				//   
				//   block2 is reachable from somewhere above branch_block. We don't want to start
				//   at block2 and search downwards, because block2 is already the follow block for
				//   branch_block.
				ThenIDom = this->IDom[(size_t)ThenBlockNum];
				ElseIDom = this->IDom[(size_t)ElseBlockNum];
				IfThenElseCase = ((ThenIDom == ElseIDom) && (ThenIDom == HeadBlockNum));
			}
			else { // ShortCircuit
				STARSCFGBlock *HeadCFGBlock = this->GetCFGBlockByNum((size_t)HeadBlockNum);
				ThenBlockNum = HeadCFGBlock->GetExpr()->GetFallThroughBlockNum();
				ElseBlockNum = HeadCFGBlock->GetExpr()->GetNonFallThroughBlockNum();
				// Fudge IDoms so short circuit branches follow regular branch pattern.
				ThenIDom = ElseIDom = HeadBlockNum;
			}

			if (IfThenElseCase) {
				// Screen out remaining if-then cases by ensuring that neither branch
				//  can reach the other without following a back edge. It is sufficient
				//  to check whether the lower RPO-numbered block can reach the other.
				bool HitDeadEnd = false;
				if (ThenBlockNum < ElseBlockNum) {
					SMPBasicBlock *ThenBlock = this->GetBlockByNum(ThenBlockNum);
					this->ResetSCCPVisitedBlocks(); // SCCP is long since done; reuse its visited flag in each block
					IfThenElseCase = (!ThenBlock->IsBlockReachableWithoutBackEdges(ElseBlockNum, HitDeadEnd));
				}
				else {
					SMPBasicBlock *ElseBlock = this->GetBlockByNum(ElseBlockNum);
					this->ResetSCCPVisitedBlocks(); // SCCP is long since done; reuse its visited flag in each block
					IfThenElseCase = (!ElseBlock->IsBlockReachableWithoutBackEdges(ThenBlockNum, HitDeadEnd));
				}
			}
			if (!IfThenElseCase) {
				// See if we fit the pattern above.
				if (ThenIDom == HeadBlockNum)
					FollowBlockNum = ElseBlockNum;
				else if (ElseIDom == HeadBlockNum)
					FollowBlockNum = ThenBlockNum;
				else {
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("INFO: SPARK: Unstructured due to failure in FindConditionalFollowNode: no successors with IDom of HeadBlockNum %d ", HeadBlockNum);
						this->DumpFuncNameAndAddr();
						this->PrintedSPARKUnstructuredMsg = true;
					}
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
				}
			}
			else {
				STARSBitSet ThenBlocksSeen, ElseBlocksSeen;
				int BlockAlreadySeenCounter = 0;
				ThenBlocksSeen.AllocateBits(this->GetNumBlocks());
				ElseBlocksSeen.AllocateBits(this->GetNumBlocks());

				// If the ThenBlock is a loop header block, then the candidate ThenFollowBlock is just
				//  the follow block for that loop. We will probably end up with -1 as our final
				//  return value, as the ElseBlock will produce a different candidate.
				int ThenFollowBlockNum = SMP_BLOCKNUM_UNINIT;
				SMPBasicBlock *ThenBlock = this->GetBlockByNum(ThenBlockNum);
				if (ThenBlock->IsLoopHeaderBlock()) {
					int ThenLoopNum = this->GetLoopNumFromHeaderBlockNum(ThenBlockNum);
					assert(0 <= ThenLoopNum);
					assert(((size_t) ThenLoopNum) < this->LoopExitTargets.size());
					bool MultiExitThenLoop = this->GetLoopFollowBlockNum((size_t)ThenLoopNum, ThenFollowBlockNum);
					if (MultiExitThenLoop) {
						ThenFollowBlockNum = SMP_BLOCKNUM_UNINIT; // error
					}
				}
				else {
					ThenFollowBlockNum = this->TrackConditionalBranchTerminus(HeadBlockNum, ThenBlockNum, ThenBlocksSeen, BlockAlreadySeenCounter);
					BlockAlreadySeenCounter = 0;
				}
				int ElseFollowBlockNum = SMP_BLOCKNUM_UNINIT;
				SMPBasicBlock *ElseBlock = this->GetBlockByNum(ElseBlockNum);
				if (ElseBlock->IsLoopHeaderBlock()) {
					int ElseLoopNum = this->GetLoopNumFromHeaderBlockNum(ElseBlockNum);
					assert(0 <= ElseLoopNum);
					assert(((size_t) ElseLoopNum) < this->LoopExitTargets.size());
					bool MultiExitElseLoop = this->GetLoopFollowBlockNum((size_t)ElseLoopNum, ElseFollowBlockNum);
					if (MultiExitElseLoop) {
						ElseFollowBlockNum = SMP_BLOCKNUM_UNINIT; // error
					}
				}
				else {
					ElseFollowBlockNum = this->TrackConditionalBranchTerminus(HeadBlockNum, ElseBlockNum, ElseBlocksSeen, BlockAlreadySeenCounter);
				}
				if ((SMP_BLOCKNUM_UNINIT != ThenFollowBlockNum) && (SMP_BLOCKNUM_UNINIT != ElseFollowBlockNum)) {
					// No errors; compute the meet function.
					if ((0 <= ThenFollowBlockNum) && (0 <= ElseFollowBlockNum)) {
						if (ThenFollowBlockNum == ElseFollowBlockNum) {
							// Why didn't we get this follow block from the dominator info?
							FollowBlockNum = ThenFollowBlockNum;
							if (ShortCircuit)
								SMP_msg("INFO: Got FollowBlockNum of %d for ShortCircuit block %d in %s via consistent termini.\n",
								FollowBlockNum, HeadBlockNum, this->GetFuncName());
							else
								SMP_msg("WARNING: Got FollowBlockNum of %d for block %d in %s via consistent termini after failing via dominator tree.\n",
								FollowBlockNum, HeadBlockNum, this->GetFuncName());
						}
						else {
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("INFO: SPARK: Unstructured due to inconsistent if-else termini block numbers: %d and %d for branch head: %d ",
									ThenFollowBlockNum, ElseFollowBlockNum, HeadBlockNum);
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal, or branches actually don't meet
						}
					}
					else if (0 <= ThenFollowBlockNum) {
						FollowBlockNum = ThenFollowBlockNum;
						SMP_msg("INFO: Got FollowBlockNum of %d for block %d in %s via THEN terminus after failing via dominator tree.\n",
							FollowBlockNum, HeadBlockNum, this->GetFuncName());
					}
					else {
						FollowBlockNum = ElseFollowBlockNum;
						SMP_msg("INFO: Got FollowBlockNum of %d for block %d in %s via ELSE terminus after failing via dominator tree.\n",
							FollowBlockNum, HeadBlockNum, this->GetFuncName());
					}
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT;
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("INFO: SPARK: Unstructured due to no FollowBlockNum for block %d due to bad then or else terminus after failing via dominator tree ", HeadBlockNum);
						this->DumpFuncNameAndAddr();
						this->PrintedSPARKUnstructuredMsg = true;
					}
				}
			}
		}
	}

	if (0 <= FollowBlockNum) {
		if (this->FindUnstructuredIfThenElse(HeadBlockNum, ThenBlockNum, ElseBlockNum, FollowBlockNum)) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
		}
	}

	return FollowBlockNum;
} // end of SMPFunction::FindConditionalFollowNode2()

// Find candidate block # for if-else follow node for HeadBlockNum; return -1 otherwise
int SMPFunction::FindConditionalFollowNode3(const int HeadBlockNum, bool &IfThenCase, bool &OddIfThenCase, bool &IfThenElseCase) {
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	int NFTBlockNum = SMP_BLOCKNUM_UNINIT;
	int FTBlockNum = SMP_BLOCKNUM_UNINIT;
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();

	// Guard against pathological code like jz $+2 which jumps and falls through
	//  to the same ensuing instruction.
	SMPBasicBlock *HeadBlock = this->RPOBlocks[HeadBlockNum];
	assert(nullptr != HeadBlock);
	if (1 == HeadBlock->GetNumSuccessors()) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode due to single successor in HeadBlockNum %d ", HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}
	else if (0 == HeadBlock->GetNumSuccessors()) { // must branch and fall-through outside function
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode for FarBranch in HeadBlockNum %d ", HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}

	// Screen out branches that go outside their current loop in an unstructured way.
	if (this->FindMultiLoopContinue(HeadBlockNum)) {
		return FollowBlockNum;
	}

	STARS_ea_t HeadLastAddr = HeadBlock->GetLastAddr();
	bool HeadBlockIsInALoop = this->IsBlockInAnyLoop(HeadBlockNum);

	// LastCFType is incomplete at this point, as the current method is part
	//  of the analysis. But SHORT_CIRCUIT_BRANCH locations have already been detected.
	ControlFlowType LastCFType = this->GetControlFlowType(HeadLastAddr);
	int InnermostLoopNum = HeadBlockIsInALoop ? this->GetInnermostLoopNum(HeadBlockNum) : -1;
	bool ShortCircuit = (IsShortCircuitFlow(LastCFType));

	assert(2 == HeadBlock->GetNumSuccessors());

	// In the discussion below, FTBlockNum and NFTBlockNum are the Fall-Through and Non-Fall-Through successor
	//  RPO (Reverse Post-Order) block numbers of the COND_BRANCH block.

	// Proposed redesign, not yet implemented:
	// The simple new algorithm depends entirely on the reachable blocks from each of the
	//  two successors of the COND_BRANCH (FTBlockNum and NFTBlockNum). Without following back edges, and staying within
	//  the region dominated by HeadBlockNum, we mark reachable blocks in bitsets for each
	//  of the two COND_BRANCH successor blocks. Then we have the following cases:
	//  0. If one of the COND_BRANCH successor blocks dominates HeadBlock, we have a back edge that
	//     cannot be followed, so we cannot have an intersection of traces from the COND_BRANCH successor
	//     blocks. So, set the code for structured lack of intersection (SMP_BLOCKNUM_COMMON_RETURN) as
	//     the follow block.
	//  1: Smaller RPO-numbered block out of FTBlockNum and NFTBlockNum reaches the other: The larger RPO
	//     number is the FollowBlockNum. If FTBlockNum is smaller, this is the normal IfThenCase. If NFTBlockNum
	//     is smaller, this is the OddIfThenCase. If the larger RPO block can reach the smaller RPO block, we
	//     would have a catastrophic failure of RPO numbering (assert and abort; should never happen).
	//  2. If not case 1, we assume the IfThenElseCase for now.
	//    2A. If the reachability bitsets intersect, then the lowest RPO block # in the intersection is
	//        the FollowBlockNum.
	//    2B. If the reachability bitsets do not intersect, we have SMP_BLOCKNUM_COMMON_RETURN (i.e. -2)
	//        as the FollowBlockNum.
	//     EXCEPTION: Intersection point was about to happen, but it is just out of the region dominated
	//     by HeadBlockNum, e.g. conditional join point is also a loop follow block for an enclosing loop or
	//     is a join point for an enclosing conditional. This can be detected by seeing if the end of each
	//     of the two traces has only one successor, and the successor is the same for each trace. This could
	//     happen for an IfThenCase or OddIfThenCase as well as for an IfThenElseCase.
	//
	// NOTE: One consequence of the above classification is that source code that might look like an IfThenCase
	//  will be classified as an IfThenElseCase as in the following example:
	//
	//  if (cond1) then
	//    [code here]
	//    return;
	//  end if;
	//  [continuation code here; might go for eons until end of function]
	//
	// Our SPARK Ada translation of this sequence, due to the return statement making the "end if;" point not
	//  reachable from the top of the THEN-clause, will be:
	//
	//  if (cond1) then
	//    [code here]
	//    return;
	//  else
	//    [continuation code here; might go for eons until end of function]
	//  end if;
	//
	// This is a perfectly valid translation of the source code and its resulting assembly language code.

	// Old cases before redesign:
	//  1. No intersection of the two bitsets. FollowBlockNum is -2 (SMP_BLOCKNUM_COMMON_RETURN)
	//     indicating the lack of intersection, with each trace ending in a block with no successors
	//     (e.g. return instructions or calls to non-returning functions or halt instructions or LOOP_BACKs).
	//
	//     If NFTBlockNum is not dominated by HeadBlockNum, but FTBlockNum is dominated by HeadBlockNum,
	//     then the FTBlockNum branch is part of an IF-THEN. If vice versa, the NFTBlockNum is part of an OddIfThenCase.
	//     If neither FTBlockNum nor NFTBlockNum is dominated by HeadBlockNum, the code is unstructured.
	//
	//     EXCEPTION: Intersection point was about to happen, but it is just out of the region dominated
	//     by HeadBlockNum, e.g. conditional join point is also a loop follow block for an enclosing loop or
	//     is a join point for an enclosing conditional. This can be detected by seeing if the end of each
	//     of the two traces has only one successor, and the successor is the same for each trace.
	//  2. Intersection point is the non-fall-through successor of the HeadBlock: simple if-then case,
	//     with the intersection point being the FollowBlockNum.
	//  3. Intersection point is the fall-through successor of the HeadBlock: OddIfThenCase,
	//     with the intersection point being the FollowBlockNum.
	//  4. Intersection point is neither successor of the HeadBlock, but is dominated by HeadBlock. Subcases:
	//   4A. Blocks reachable from both traces continue after the intersection point, but no block with
	//       RPO number greater than the intersection block is reachable from only one trace. Structured,
	//       with the intersection point being the FollowBlockNum.
	//   4B. Like 4A, but at least one block with RPO number greater than the intersection point is reachable
	//       from only one of the two traces. UNSTRUCTURED.
	//   4C: Conditional is inside a loop, and 2 or more loop tail blocks for that loop are reachable from the
	//       two traces. The decompilation of the conditional will need to continue until the loop backs and loop
	//       boundaries are reached, so the follow block needs to be -2 (SMP_BLOCKNUM_COMMON_RETURN).
	//  5. Intersection point is not dominated by HeadBlock. Some path is joining the conditional at the 
	//     intersection point. The intersection point is part of some other structure, such as a containing
	//     loop or containing conditional that shares the intersection point as a common follow node. In any case,
	//     the intersection point must be the follow node for the current conditional. NOTE: Isn't this the EXCEPTION
	//     to Case 2 discussed above? How could it be hit down here?

	if (!ShortCircuit) {
		NFTBlockNum = (*(HeadBlock->GetCondNonFallThroughSucc()))->GetNumber();
		assert(SMP_BLOCKNUM_UNINIT != NFTBlockNum);
		FTBlockNum = (*(HeadBlock->GetCondOtherSucc(NFTBlockNum)))->GetNumber();
		assert(SMP_BLOCKNUM_UNINIT != FTBlockNum);
	}
	else { // ShortCircuit
		STARSCFGBlock *HeadCFGBlock = this->GetCFGBlockByNum((size_t)HeadBlockNum);
		FTBlockNum = HeadCFGBlock->GetExpr()->GetFallThroughBlockNum();
		NFTBlockNum = HeadCFGBlock->GetExpr()->GetNonFallThroughBlockNum();
	}

	STARSBitSet FTBlocksSeen;
	STARSBitSet NFTBlocksSeen;
	FTBlocksSeen.AllocateBits(this->GetNumBlocks());
	NFTBlocksSeen.AllocateBits(this->GetNumBlocks());

	SMPBasicBlock *NFTBlock = this->GetBlockByNum((size_t)NFTBlockNum);
	SMPBasicBlock *FTBlock = this->GetBlockByNum((size_t)FTBlockNum);
	uint16_t LoopBackCount, ForwardCount;
	int FTLimitBlockNum = FTBlock->GetSingleDomFrontierBlockNum(InnermostLoopNum, LoopBackCount, ForwardCount);
	int NFTLimitBlockNum = NFTBlock->GetSingleDomFrontierBlockNum(InnermostLoopNum, LoopBackCount, ForwardCount);
	this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
	this->MarkReachableBlocks(FTBlockNum, -1, HeadBlockNum, true, FTLimitBlockNum, FTBlocksSeen);
	this->MarkReachableBlocks(NFTBlockNum, -1, HeadBlockNum, true, NFTLimitBlockNum, NFTBlocksSeen);
	STARSBitSet CommonBlocksSeen;
	CommonBlocksSeen.AllocateBits(this->GetNumBlocks());
	CommonBlocksSeen = STARSBitSetIntersection(FTBlocksSeen, NFTBlocksSeen);
	size_t CommonBlocksCount = CommonBlocksSeen.CountSetBits();
	int FirstCommonBlockNum = CommonBlocksSeen.FindLowestBitSet();
	bool IntersectionSeen = (0 < CommonBlocksCount);
	
	this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
	bool FTNotDominated = (!this->DoesBlockDominateBlock(HeadBlockNum, FTBlockNum));
	bool NFTNotDominated = (!this->DoesBlockDominateBlock(HeadBlockNum, NFTBlockNum));
	bool DominanceProblem = (FTNotDominated || NFTNotDominated);

	int LastFTReachableBlockNum = FTBlocksSeen.FindHighestBitSet();
	int LastNFTReachableBlockNum = NFTBlocksSeen.FindHighestBitSet();
	int LastBlockNumReached = (LastFTReachableBlockNum >= LastNFTReachableBlockNum) ? LastFTReachableBlockNum : LastNFTReachableBlockNum;
#if 0
	bool IfThenElseCase = true;
#else
	IfThenCase = (FirstCommonBlockNum == NFTBlockNum) || (!FTNotDominated && NFTNotDominated);
	OddIfThenCase = (FirstCommonBlockNum == FTBlockNum) || (!NFTNotDominated && FTNotDominated);
	IfThenElseCase = (!(IfThenCase || OddIfThenCase));
#endif
	bool Case0 = false;

	// Based on CommonBlocksCount, FirstCommonBlockNum, FTBlockNum, and NFTBlockNum, we
	//  can find which case in the comment above applies.
	int NextFTBlockNum = SMP_BLOCKNUM_UNINIT;
	int NextNFTBlockNum = SMP_BLOCKNUM_UNINIT;
	if (!IntersectionSeen) { // Case 1 or Case 0.
		FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN; // Case 1

		// Detect the Case 1 exception subcase.
		if (!FTNotDominated) { // FTBlock was dominated by HeadBlock
			bool ErrorSeen = this->FindNextBlockAfterReachableBlocks(InnermostLoopNum, FTBlocksSeen, NextFTBlockNum);
			if (ErrorSeen) {
				FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			}
		}
		else { // FTBlock was outside region dominated by HeadBlock
			NextFTBlockNum = FTBlockNum; // First block outside dominated region
			IfThenElseCase = false;
		}
		if (!NFTNotDominated) { // NFTBlock was dominated by HeadBlock
			bool ErrorSeen = this->FindNextBlockAfterReachableBlocks(InnermostLoopNum, NFTBlocksSeen, NextNFTBlockNum);
			if (ErrorSeen) {
				FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			}
		}
		else { // NFTBlock was outside region dominated by HeadBlock
			NextNFTBlockNum = NFTBlockNum; // First block outside dominated region
			IfThenElseCase = false;
		}

		bool NotLoopBack = ((NextFTBlockNum > HeadBlockNum) || (SMP_BLOCKNUM_COMMON_RETURN == NextFTBlockNum));
		if ((NextFTBlockNum == NextNFTBlockNum) && NotLoopBack) { // Exception to Case 1
			FollowBlockNum = NextFTBlockNum;
			if (NFTNotDominated) {
				IfThenCase = true;
			}
			else if (FTNotDominated) {
				OddIfThenCase = true;
			}
			else {
				IfThenElseCase = true;
			}
		}
		else if (FTNotDominated && OddIfThenCase) {
			FollowBlockNum = FTBlockNum;
		}
		else if (NFTNotDominated && IfThenCase) {
			FollowBlockNum = NFTBlockNum;
		}

		if (FollowBlockNum == SMP_BLOCKNUM_COMMON_RETURN) { // Not the exception to Case 1; search for Case 0.
			if (HeadBlockIsInALoop && DominanceProblem) { // Maybe one COND_BRANCH successor is our own loop head.
				if (NFTBlock->IsLoopHeaderBlock()) {
					int NFTLoopNum = this->GetLoopNumFromHeaderBlockNum(NFTBlockNum);
					Case0 = (NFTLoopNum == InnermostLoopNum);
				}
				if (!Case0 && FTBlock->IsLoopHeaderBlock()) {
					int FTLoopNum = this->GetLoopNumFromHeaderBlockNum(FTBlockNum);
					Case0 = (FTLoopNum == InnermostLoopNum);
				}
				// Case0 ==> SMP_BLOCKNUM_COMMON_RETURN for FollowBlock; use boolean flag below to avoid
				//  falsely declaring a CFG to be unstructured due to the DominanceProblem.
			}
		}
	}
	else if (FirstCommonBlockNum == NFTBlockNum) { // Case 2; IfThenCase
		FollowBlockNum = FirstCommonBlockNum;
		IfThenCase = true;
		OddIfThenCase = false;
		IfThenElseCase = false;
	}
	else if (FirstCommonBlockNum == FTBlockNum) { // Case 3; OddIfThenCase
		FollowBlockNum = FirstCommonBlockNum;
		IfThenCase = false;
		OddIfThenCase = true;
		IfThenElseCase = false;
	}
	else { // if-then-else with at least one intersection point
		IfThenCase = false;
		OddIfThenCase = false;
		IfThenElseCase = true;
		this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		if (this->DoesBlockDominateBlock(HeadBlockNum, FirstCommonBlockNum)) { // Case 4
			set<int> TailBlocksSeen;
			if (HeadBlockIsInALoop) {
				int FirstFTReachableBlockNum = FTBlocksSeen.FindLowestBitSet();
				int FirstNFTReachableBlockNum = NFTBlocksSeen.FindLowestBitSet();
				for (int BlockIndex = FirstFTReachableBlockNum; BlockIndex <= LastFTReachableBlockNum; ++BlockIndex) {
					if (FTBlocksSeen.GetBit((size_t)BlockIndex)) {
						SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t)BlockIndex);
						if (CurrBlock->IsLoopTailBlock()) {
							// We don't care about loop backs that come after we converged at FirstCommonBlockNum,
							//  and we don't care about loop backs within inner loops that are contained inside the
							//  conditional.
							int CurrLoopHeadBlockNum = CurrBlock->GetLoopHeaderNumber();
							assert(0 <= CurrLoopHeadBlockNum);
							if (this->AreBlocksInSameLoops(HeadBlockNum, CurrLoopHeadBlockNum)) { // not contained inner loop
								this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
								if (!this->DoesBlockDominateBlock(FirstCommonBlockNum, BlockIndex)) { // not after intersection
									TailBlocksSeen.insert(BlockIndex);
								}
							}
						}
					}
				}
				for (int BlockIndex = FirstNFTReachableBlockNum; BlockIndex <= LastNFTReachableBlockNum; ++BlockIndex) {
					if (NFTBlocksSeen.GetBit((size_t)BlockIndex)) {
						if (this->GetBlockByNum((size_t)BlockIndex)->IsLoopTailBlock()) {
							// We don't care about loop backs that come after we converged at FirstCommonBlockNum.
							this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
							if (!this->DoesBlockDominateBlock(FirstCommonBlockNum, BlockIndex)) {
								TailBlocksSeen.insert(BlockIndex);
							}
						}
					}
				}
			}
			if (2 <= TailBlocksSeen.size()) { // Case 4C
				FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
			}
			else if (LastFTReachableBlockNum != LastNFTReachableBlockNum) { // Case 4B
				FollowBlockNum = SMP_BLOCKNUM_UNINIT;
				if (!this->PrintedSPARKUnstructuredMsg) {
					SMP_msg("ERROR: SPARK: Unstructured due to non-common-reached block after HeadBlockNum %d and intersection block %d ",
						HeadBlockNum, FirstCommonBlockNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
					this->PrintedSPARKUnstructuredMsg = true;
				}
			}
			else {
				bool NotCommonBlockFound = false;
				for (int BlockIndex = FirstCommonBlockNum; BlockIndex < LastFTReachableBlockNum; ++BlockIndex) {
					if (FTBlocksSeen.GetBit((size_t)BlockIndex) != (NFTBlocksSeen.GetBit((size_t)BlockIndex))) {
						NotCommonBlockFound = true;
						break;
					}
				}
				if (NotCommonBlockFound) { // Case 4B
					FollowBlockNum = SMP_BLOCKNUM_UNINIT;
				}
				else { // Case 4A
					FollowBlockNum = FirstCommonBlockNum;
				}
			}
		}
		else { // Case 5
			// !!!!****!!!! Should we demand that all other predecessors of FirstCommonBlockNum
			//  dominate HeadBlockNum, proving that we truly have containing structures that prevented
			//  HeadBlockNum from dominating FirstCommonBlockNum?
			FollowBlockNum = FirstCommonBlockNum;
		}
	}
	// Detect unstructured entry points.
	if (IfThenElseCase) {
		// Both sides of the conditional should be dominated by HeadBlock.
		if (DominanceProblem && (!Case0)) {
			// Some path is entering one side of our conditional at the FTBlock or at the NFTBlock.
			FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to IfThenElse BranchBlock %d not dominating its successors ", HeadBlockNum);
				this->DumpFuncNameAndAddr();
				this->DumpDotCFG();
				this->PrintedSPARKUnstructuredMsg = true;
			}
		}
		if (0 <= FollowBlockNum) {
			if (this->FindUnstructuredIfThenElse(HeadBlockNum, FTBlockNum, NFTBlockNum, FollowBlockNum)) {
				FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
			}
		}
	}
	// All THEN and ELSE blocks should be able to reach the FollowBlockNum.
	if (0 <= FollowBlockNum) {
		// Eliminate blocks after the FollowBlockNum from the bitsets.
		for (int BlockIndex = FollowBlockNum; BlockIndex <= LastBlockNumReached; ++BlockIndex) {
			FTBlocksSeen.ResetBit((size_t)BlockIndex);
			NFTBlocksSeen.ResetBit((size_t)BlockIndex);
		}
		list<size_t> ReachedBlocksList;
		this->BuildBlockListFromBitset(FTBlocksSeen, ReachedBlocksList);
		this->BuildBlockListFromBitset(NFTBlocksSeen, ReachedBlocksList);
		this->ResetSCCPVisitedBlocks();
		if (this->FindUnstructuredBlockList(ReachedBlocksList, HeadBlockNum, FollowBlockNum)) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error
		}
	}

	if (SMP_BLOCKNUM_UNINIT != FollowBlockNum) {
		// Find CFG problems that are obvious just from the reachability bitsets.
		this->ResetSCCPVisitedBlocks();
		if (this->DetectConditionalReachabilityErrors(HeadBlockNum, FTBlockNum, NFTBlockNum, FTNotDominated, NFTNotDominated, NextFTBlockNum, NextNFTBlockNum, FTBlocksSeen, NFTBlocksSeen, CommonBlocksSeen)) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT; // error
		}
	}

	if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) {
		this->Dump(!VerboseOutput);
	}
	return FollowBlockNum;
} // end of SMPFunction::FindConditionalFollowNode3()

// Find candidate block # for if-else follow node for HeadBlockNum; return -1 otherwise
int SMPFunction::FindConditionalFollowNode4(const int HeadBlockNum, bool &IfThenCase, bool &OddIfThenCase, bool &IfThenElseCase, bool &ShortCircuit) {
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	int NFTBlockNum = SMP_BLOCKNUM_UNINIT;
	int FTBlockNum = SMP_BLOCKNUM_UNINIT;
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();

	// Guard against pathological code like jz $+2 which jumps and falls through
	//  to the same ensuing instruction.
	STARSTransCFGBlock *HeadBlock = this->TCFGBlocks[HeadBlockNum];
	assert(nullptr != HeadBlock);
	if (1 == HeadBlock->GetNumSuccs()) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode4 due to single successor in HeadBlockNum %d ", HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}
	else if (0 == HeadBlock->GetNumSuccs()) { // must branch and fall-through outside the current function
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode4 for FarBranch in HeadBlockNum %d ", HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}

	// Screen out branches that go outside their current loop in an unstructured way.
	if (this->FindMultiLoopContinue(HeadBlockNum)) {
		return FollowBlockNum;
	}

	// We can determine all of our analysis outputs using dominance frontiers. The dominance frontier of
	//  a basic block is the set of blocks that are one step beyond being dominated by the block. A block
	//  can have itself, or a smaller RPO-numbered block, in its dominance frontier due to a loop back.
	//
	// A dominance frontier can be the empty set if every path from the block continues on to dead ends
	//  (e.g. return statements) without intersecting another path starting at a different dominating block.
	//  Given that we need to find intersection points to analyze conditionals (see logic of FindConditionalFollowNode3()),
	//  the DomFrontier gives us a much quicker analysis of intersection points.
	//
	// In a simple, structured if-then-else, the COND_BRANCH statement has two successors, the fall-through block (FTBlock)
	//  and the non-fall-through block (NFTBlock, a.k.a. the branch target block). The FTBlock and NFTBlock will
	//  each have the same DomFrontier set, with cardinality 1, and the block in that set is the ENDIF intersection
	//  point.
	//
	// In a simple, structured if-then, the FTBlock will have the NFTBlock as its lone DomFrontier block.
	//
	// In a simple, structured odd-if-then, the NFTBlock will have the FTBlock as its lone DomFrontier block.
	//
	// Now we need to decide what to do with the non-simple cases. If the DomFrontier of either the FTBlock or
	//  NFTBlock is the empty set, then we have the dead end case. We can classify the conditional as an if-then-else
	//  without loss of precision. Note that the original source code might have looked like an if-then:
	//
	//  if (cond1) {
	//     ...
	//     return;
	//  }
	//  [code here]
	//
	//  and we are transforming it into an if-then-else in SPARK Ada:
	//
	//  if (cond1) then
	//     ...
	//     return;
	//  else
	//     [code here]
	//  end if;
	//
	// The [code here] will continue until it hits the follow block or boundary of an enclosing
	// statement, such as a loop. We will use the SMP_BLOCKNUM_COMMON_RETURN code (== -2) for this case.
	//
	// The DomFrontier sets might include loop-back targets, which are also signaled with the -2 code.
	//
	// If either DomFrontier set has multiple forward successor blocks, or otherwise does not fit
	//  the above patterns (e.g. DomFrontier blocks are not the same), we have an unstructured conditional.
	//  Either we are bypassing the natural ENDIF point, or some path is branching into a THEN or ELSE clause.
	//  We use the SMP_BLOCKNUM_UNINIT (== -1) code for this result. EXCEPTION: If we can structure the 
	//  unstructured code by cloning, we need not return the error code.
	//

	STARS_ea_t HeadLastAddr = HeadBlock->GetLastAddr();
	int OriginalHeadBlockNum = HeadBlock->GetOriginalBlockNum();
	bool HeadBlockIsInALoop = this->IsBlockInAnyLoop(OriginalHeadBlockNum);

	// LastCFType is incomplete at this point, as the current method is part
	//  of the analysis. But SHORT_CIRCUIT_BRANCH locations have already been detected.
	ControlFlowType LastCFType = this->GetControlFlowType(HeadLastAddr);
	int InnermostLoopNum = HeadBlockIsInALoop ? this->GetInnermostLoopNum(OriginalHeadBlockNum) : -1;
	ShortCircuit = (IsShortCircuitFlow(LastCFType));

	assert(2 == HeadBlock->GetNumSuccs());

	if (!ShortCircuit) {
		NFTBlockNum = HeadBlock->GetCondNonFallThroughSuccBlockNum();
		assert(SMP_BLOCKNUM_UNINIT != NFTBlockNum);
		FTBlockNum = (*(HeadBlock->GetCondOtherSucc(NFTBlockNum)))->GetNumber();
		assert(SMP_BLOCKNUM_UNINIT != FTBlockNum);
	}
	else { // ShortCircuit
		STARSCFGBlock *HeadCFGBlock = this->GetClonedCFGBlockByNum((size_t)HeadBlockNum);
		FTBlockNum = HeadCFGBlock->GetExpr()->GetFallThroughBlockNum();
		assert(SMP_BLOCKNUM_UNINIT != FTBlockNum);
		NFTBlockNum = HeadCFGBlock->GetExpr()->GetNonFallThroughBlockNum();
		assert(SMP_BLOCKNUM_UNINIT != NFTBlockNum);
	}

	// Find loop control flow, e.g. LOOP_CONTINUE, which needs to be handled elsewhere.
	if ((NFTBlockNum <= HeadBlockNum) || (FTBlockNum <= HeadBlockNum)) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode4 due to loop back edge from CondBranch in OrigHeadBlockNum %d ", OriginalHeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		return FollowBlockNum;
	}

	STARSTransCFGBlock *NFTBlock = this->GetTCFGBlockByNum((size_t)NFTBlockNum);
	STARSTransCFGBlock *FTBlock = this->GetTCFGBlockByNum((size_t)FTBlockNum);

	// In the if-then-else case, FTBlock and NFTBlock must both be dominated
	//  by HeadBlock. The if-then case can have the endif block not dominated
	//  by HeadBlock.
	bool FTNotDominated = (1 != FTBlock->GetNumPredsMinusBackEdges());
	bool NFTNotDominated = (1 != NFTBlock->GetNumPredsMinusBackEdges());
	this->TCFGDomSoFar.ResetAllBits(); // Prepare for TCFGDomTree search, memoize dominated blocks seen so far.
	// In the short-circuit case, multiple edges from short-circuit tests flow to 
	//  each of FTBlock and NFTBlock, so the pred count code above is not sufficient.
	if (ShortCircuit) {
		FTNotDominated = (!this->DoesTCFGBlockDominateTCFGBlock(HeadBlockNum, FTBlockNum));
		NFTNotDominated = (!this->DoesTCFGBlockDominateTCFGBlock(HeadBlockNum, NFTBlockNum));
	}

	uint16_t FTLoopBackCount, FTForwardCount;
	uint16_t NFTLoopBackCount, NFTForwardCount;
	int FTDomFrontierBlockNum = FTBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, FTLoopBackCount, FTForwardCount);
	int NFTDomFrontierBlockNum = NFTBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, NFTLoopBackCount, NFTForwardCount);
	bool FTDeadEnd = (FTForwardCount == 0);
	bool NFTDeadEnd = (NFTForwardCount == 0);
	bool TwoDeadEnds = (FTDeadEnd && NFTDeadEnd);
	bool OneDeadEnd = (!TwoDeadEnds) && (FTDeadEnd || NFTDeadEnd);
	bool NoDeadEnds = (!(FTDeadEnd || NFTDeadEnd));
	bool TooMany = ((FTForwardCount > 1) || (NFTForwardCount > 1));
	bool NonConvergence = (FTDomFrontierBlockNum != NFTDomFrontierBlockNum);

	if (TooMany) {
		// Find tight connection (premature merge) between branches of IfThenElse; difficult for our UnlinkPred()
		//  and EnforceDominanceOnPreds() methods to generalize to handle these, and they are rare, so
		//  leave them as unstructured. These cause our definition of FTNotDominated and NFTNotDominated
		//  above (based on Pred count) to be incorrect; we tried to avoid expensive dominance checks using
		//  the predecessor count and the connection from FT to NFT or NFT to FT ruins that short cut.
		bool TightError = (FTNotDominated && FTBlock->IsBlockPred(NFTBlock) && (1 < NFTForwardCount));
		if (!TightError) {
			TightError = (NFTNotDominated && NFTBlock->IsBlockPred(FTBlock) && (1 < FTForwardCount));
		}
		if (!TightError) {
			// Tight merge might not be tight enough if it passes through a short circuit block set. Check.
			STARSCFGBlock *FTCFGBlock = this->GetClonedCFGBlockByNum((size_t)FTBlockNum);
			if (NFTNotDominated && FTCFGBlock->IsExprHeadBlock()) {
				TightError = ((FTCFGBlock->GetExpr()->GetFallThroughBlockNum() == NFTBlockNum)
					|| (FTCFGBlock->GetExpr()->GetNonFallThroughBlockNum() == NFTBlockNum));
			}
			if (!TightError) {
				STARSCFGBlock *NFTCFGBlock = this->GetClonedCFGBlockByNum((size_t)NFTBlockNum);
				if (FTNotDominated && NFTCFGBlock->IsExprHeadBlock()) {
					TightError = ((NFTCFGBlock->GetExpr()->GetFallThroughBlockNum() == FTBlockNum)
						|| (NFTCFGBlock->GetExpr()->GetNonFallThroughBlockNum() == FTBlockNum));
				}
			}
		}
		if (TightError) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = true;
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to THEN and ELSE clauses merge prematurely for HeadBlock %d ", HeadBlockNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
			return FollowBlockNum;
		}
	}

	// Find the simple unstructured case in which a side entry into the THEN clause or ELSE clause
	//  does not enter at the FTBlock or NFTBlock, but a little later:
	//
	//       B1 [ends with COND_BRANCH]
	//   |----+--------|
	//   |             |
	//   |            FTB
	//   |             |
	//   |             B2 ------+ [side entry here from up above B1]
	//   |             |
	//   |             |
	//   +-----+-------+
	//         |
	//        NFTB
	//
	// The DomFrontiers are not convergent, because the FTBlock has B2 as its DomFrontier.
	//  However, the fact that B2 has NFTB as its DomFrontier reveals the shape of the CFG;
	//  if we clone B2 through NFTB, we can eliminate the side entry and have a simple if-then statement.
	//  Note that FTB and B2 must have smaller RPO numbers than NFTB in this CFG; FTB is dominated
	//  by B1 but NFTB is not dominated by B1.
	//
	// Detect this pattern before getting into other cloning attempts.
	//  We can swap the FT and NFT labels without loss of generality.
	//
	bool DominanceIssue = (FTNotDominated || NFTNotDominated);
	bool SimpleProblem = (NonConvergence && NoDeadEnds && (!TooMany));
	if (SimpleProblem) {
		// See if we can get convergence by looking at the DomFrontier for the smaller RPO
		//  successor and then looking at the DomFrontier for the DomFrontier block.
		bool FTFirst = (FTBlockNum < NFTBlockNum);
		int SmallerRPOBlock = FTFirst ? FTBlockNum : NFTBlockNum;
		int LargerRPOBlock = FTFirst ? NFTBlockNum : FTBlockNum;

		IfThenCase = false;
		OddIfThenCase = false;
		IfThenElseCase = false;

		if (FTFirst && (FTDomFrontierBlockNum < NFTBlockNum) && HeadBlock->IsBlockInDomFrontier(FTDomFrontierBlockNum)) {
			// We have the shape we are looking for IF FTDomFrontierBlockNum (B2 in drawing above) has NFTBlockNum
			//  as its sole DomFrontier set entry.
			STARSTransCFGBlock *FTDomFrontierBlock = this->GetTCFGBlockByNum((size_t)FTDomFrontierBlockNum);
			assert(nullptr != FTDomFrontierBlock);
			uint16_t LoopBackCount, ForwardCount;
			int FTDomDomBlockNum = FTDomFrontierBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, LoopBackCount, ForwardCount);
			if (FTDomDomBlockNum == NFTBlockNum) { // bingo!
				set<int> BlockNumsToClone;
				if (this->IsCloningSafe(FTDomFrontierBlockNum, NFTBlockNum, BlockNumsToClone)) {
					// regular if-then case if we clone from FTDomFrontierBlockNum to a limit of NFTBlockNum.
					SMP_msg("INFO: SPARK: if-then DomDom cloning opportunity for HeadBlock %d from %d to %d ",
						HeadBlockNum, FTDomFrontierBlockNum, NFTBlockNum);
					this->DumpFuncNameAndAddr();

					FollowBlockNum = NFTBlockNum;
					// The block to unlink from is more complicated than in other cloning cases. It is the predecessor
					//  of FTDomFrontierBlockNum on the path from FTBlock to FTDomFrontierBlock, so it will be 
					//  dominated by HeadBlockNum. The side entry path is not dominated by HeadBlockNum.
					//  If more than one predecessor of FTDomFrontierBlock is dominated by the HeadBlock,
					//  we need more expensive analysis. Do that only when we encounter an example CFG.
					bool FoundUnlinkPred = false;
					int UnlinkBlockNum = SMP_BLOCKNUM_UNINIT;
					for (STARSTransCFGBlock *PredBlock : FTDomFrontierBlock->GetConstPredList()) {
						int PredTCFGBlockNum = PredBlock->GetNumber();
						if ((PredTCFGBlockNum >= FTBlockNum) && (PredTCFGBlockNum < FTDomFrontierBlockNum) && this->DoesTCFGBlockDominateTCFGBlock(HeadBlockNum, PredTCFGBlockNum)) {
							// Candidate. Is it the only candidate?
							if (FoundUnlinkPred) { // previously found another candidate
								SMP_msg("ERROR: SPARK: DomDom cloning foiled by >1 candidates for UnlinkBlockNum: %d and %d ",
									UnlinkBlockNum, PredTCFGBlockNum);
								this->DumpFuncNameAndAddr();
								UnlinkBlockNum = SMP_BLOCKNUM_UNINIT;
								FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
								break;
							}
							else {
								UnlinkBlockNum = PredTCFGBlockNum;
								FoundUnlinkPred = true;
							}
						}
					} // end for all Preds of FTDomFrontierBlock
					if (0 < UnlinkBlockNum) {
						if (DominanceIssue) {
#if 0
							SMP_msg("INFO: SPARK: DomDom cloning foiled by DominanceIssue at HeadBlock %d ", HeadBlockNum);
							this->DumpFuncNameAndAddr();
						}
						else {
#else
							SMP_msg("INFO: SPARK: DomDom cloning continuing despite DominanceIssue at HeadBlock %d ", HeadBlockNum);
							this->DumpFuncNameAndAddr();
						}
						if (true) {

#endif
							pair<int, int> UnlinkLimitPair(UnlinkBlockNum, FollowBlockNum);
							pair<int, pair<int, int> > CloningWorkListItem(FTDomFrontierBlockNum, UnlinkLimitPair);
							this->SPARKCloningWorkList.push_back(CloningWorkListItem);
							SMP_msg("INFO: SPARK: CLONING 11: Cloning from %d to %d would solve nonconvergent DomFrontiers for HeadBlock %d ",
								FTDomFrontierBlockNum, FollowBlockNum, HeadBlockNum);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							IfThenCase = true;
						}
					}
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // not cloning
				}
			}
		}
		else if (!FTFirst && (NFTDomFrontierBlockNum < FTBlockNum) && HeadBlock->IsBlockInDomFrontier(NFTDomFrontierBlockNum)) {
			// We have the odd-if-then shape we are looking for IF NFTDomFrontierBlockNum (B2 in drawing above) has FTBlockNum
			//  as its sole DomFrontier set entry.
			STARSTransCFGBlock *NFTDomFrontierBlock = this->GetTCFGBlockByNum((size_t)NFTDomFrontierBlockNum);
			assert(nullptr != NFTDomFrontierBlock);
			uint16_t LoopBackCount, ForwardCount;
			int NFTDomDomBlockNum = NFTDomFrontierBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, LoopBackCount, ForwardCount);
			if (NFTDomDomBlockNum == FTBlockNum) { // bingo!
				set<int> BlockNumsToClone;
				if (this->IsCloningSafe(NFTDomFrontierBlockNum, FTBlockNum, BlockNumsToClone)) {
					// odd-if-then case if we clone from NFTDomFrontierBlockNum to a limit of FTBlockNum.
					SMP_msg("INFO: SPARK: odd-if-then DomDom cloning opportunity for HeadBlock %d from %d to %d ",
						HeadBlockNum, NFTDomFrontierBlockNum, FTBlockNum);
					this->DumpFuncNameAndAddr();

					FollowBlockNum = FTBlockNum;
					// The block to unlink from is more complicated than in other cloning cases. It is the predecessor
					//  of NFTDomFrontierBlockNum on the path from NFTBlock to NFTDomFrontierBlock, so it will have an
					//  RPO number between those two block numbers, while the side entry comes from a different RPO
					//  numbering region in the CFG. If more than one predecessor of NFTDomFrontierBlock has an eligible
					//  RPO block number, we need more expensive analysis. Do that only when we encounter an example CFG.
					bool FoundUnlinkPred = false;
					int UnlinkBlockNum = SMP_BLOCKNUM_UNINIT;
					for (STARSTransCFGBlock *PredBlock : NFTDomFrontierBlock->GetConstPredList()) {
						int PredTCFGBlockNum = PredBlock->GetNumber();
						if ((PredTCFGBlockNum > NFTBlockNum) && (PredTCFGBlockNum < NFTDomFrontierBlockNum)) {
							// Candidate. Is it the only candidate?
							if (FoundUnlinkPred) { // previously found another candidate
								SMP_msg("ERROR: SPARK: DomDom cloning foiled by >1 candidates for UnlinkBlockNum: %d and %d ",
									UnlinkBlockNum, PredTCFGBlockNum);
								this->DumpFuncNameAndAddr();
								UnlinkBlockNum = SMP_BLOCKNUM_UNINIT;
								FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
								break;
							}
							else {
								UnlinkBlockNum = PredTCFGBlockNum;
								FoundUnlinkPred = true;
							}
						}
					} // end for all Preds of NFTDomFrontierBlock
					if (0 < UnlinkBlockNum) {
						pair<int, int> UnlinkLimitPair(UnlinkBlockNum, FollowBlockNum);
						pair<int, pair<int, int> > CloningWorkListItem(NFTDomFrontierBlockNum, UnlinkLimitPair);
						this->SPARKCloningWorkList.push_back(CloningWorkListItem);
						SMP_msg("INFO: SPARK: CLONING 12: Cloning from %d to %d would solve nonconvergent DomFrontiers for HeadBlock %d ",
							NFTDomFrontierBlockNum, FollowBlockNum, HeadBlockNum);
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
						OddIfThenCase = true;
					}
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
		}
	}

	if (this->SPARKCloningWorkList.empty()) { // no special case cloning from above
		if ((FTForwardCount == 1) && (NFTForwardCount == 1) && (!NonConvergence)) {
			FollowBlockNum = FTDomFrontierBlockNum;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = true;
		}
		else if (TwoDeadEnds) {
			assert(SMP_BLOCKNUM_COMMON_RETURN == FTDomFrontierBlockNum);
			assert(SMP_BLOCKNUM_COMMON_RETURN == NFTDomFrontierBlockNum);
			FollowBlockNum = FTDomFrontierBlockNum;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = true;
		}
		else if (FTDomFrontierBlockNum == NFTBlockNum) {
			FollowBlockNum = FTDomFrontierBlockNum;
			IfThenCase = true;
			OddIfThenCase = false;
			IfThenElseCase = false;
		}
		else if (NFTDomFrontierBlockNum == FTBlockNum) {
			FollowBlockNum = NFTDomFrontierBlockNum;
			IfThenCase = false;
			OddIfThenCase = true;
			IfThenElseCase = false;
		}
		else if (OneDeadEnd) {
			FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = true;
		}
		else if (TooMany || NonConvergence) {
			FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = false;
			if (NonConvergence && (0 < FTDomFrontierBlockNum) && (0 < NFTDomFrontierBlockNum)) {
				bool CloningCure = this->SolveConditionalDomFrontierProblemViaCloning(HeadBlockNum, FTBlockNum, NFTBlockNum, FTDomFrontierBlockNum, NFTDomFrontierBlockNum, FollowBlockNum);
				if (CloningCure) {
					IfThenElseCase = true;
				}
			}
			if ((SMP_BLOCKNUM_UNINIT == FollowBlockNum) && (!this->PrintedSPARKUnstructuredMsg)) {
				SMP_msg("ERROR: SPARK: Unstructured due to failure in FindConditionalFollowNode4 due to non-convergent DomFrontiers in HeadBlockNum %d ", HeadBlockNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
		}
		else if ((FTLoopBackCount > 0) || (NFTLoopBackCount > 0)) {
			FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
			IfThenCase = false;
			OddIfThenCase = false;
			IfThenElseCase = true;
		}

		// Check for the dominance problem on if-then-else case, which indicates an entry to
		//  the top of the THEN or ELSE cases that bypassed HeadBlock. Likewise, FTBlock must 
		//  be dominated in an if-then and NFTBlock must be dominated on an odd if-then.
		bool IfThenElseError = (IfThenElseCase && DominanceIssue);
		bool IfThenError = (IfThenCase && FTNotDominated);
		bool OddIfThenError = (OddIfThenCase && NFTNotDominated);
		bool DominanceError = (IfThenElseError || IfThenError || OddIfThenError);
		if (DominanceError) {
			// See if we can salvage any of these errors by checking for straightforward translation patterns.

			if (IfThenElseError && FTDeadEnd && NFTNotDominated && (!FTNotDominated)) {
				// We can put the FTDeadEnd branch inside an if-then and branch to the NFTBlock/ENDIF.
				IfThenCase = true;
				IfThenElseCase = false;
				FollowBlockNum = NFTBlockNum;
			}
			else if (IfThenElseError && NFTDeadEnd && FTNotDominated && (!NFTNotDominated)) {
				// We can put the NFTDeadEnd branch inside an odd if-then and fall through to the FTBlock/ENDIF.
				OddIfThenCase = true;
				IfThenElseCase = false;
				FollowBlockNum = FTBlockNum;
			}
			else if (IfThenElseError && (!NonConvergence) && FTNotDominated && (!NFTNotDominated)) {
				// FTNotDominated is only problem; we have convergence. Clone starting at FTBlockNum.
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(FTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(FTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 1: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d NoDeadEnds %d ", FTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (IfThenElseError && (!NonConvergence) && NFTNotDominated && (!FTNotDominated)) {
				// NFTNotDominated is only problem; we have convergence. Clone starting at NFTBlockNum.
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(NFTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(NFTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 2: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d NoDeadEnds %d ", NFTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (IfThenElseError && NFTDeadEnd && NFTNotDominated && (!FTNotDominated)) {
				// NFTNotDominated is only problem; lack of convergence is because of dead end. Clone starting at NFTBlockNum through its dead end.
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(NFTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(NFTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 9: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d NoDeadEnds %d ", NFTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (IfThenElseError && FTDeadEnd && FTNotDominated && (!NFTNotDominated)) {
				// FTNotDominated is only problem; lack of convergence is because of dead end. Clone starting at FTBlockNum through its dead end.
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(FTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(FTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 10: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d ", FTBlockNum, FollowBlockNum, HeadBlockNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (IfThenElseError && FTNotDominated && NFTNotDominated && (FTDomFrontierBlockNum == NFTDomFrontierBlockNum)) {
				// Side entrances into both FTBlock and NFTBlock. Cloning can structure it.
				//  The FTBlock and NFTBlock already have convergence at their DomFrontiers, just have dominance problem.
				//  We can relax the NoDeadEnds requirement and clone through dead ends later.
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				set<int> BlockNumsToClone2;
				bool SafeCloning = this->IsCloningSafe(NFTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) { // see if second cloning will also be safe.
					this->ResetTCFGVisitedBlocks();
					SafeCloning = this->IsCloningSafe(FTBlockNum, FollowBlockNum, BlockNumsToClone2);
					if (SafeCloning) {
						SafeCloning = this->IsDoubleCloningSafe(FTBlockNum, NFTBlockNum);
						if (SafeCloning) {
							pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
							pair<int, pair<int, int> > CloningWorkListItem(NFTBlockNum, UnlinkLimitPair);
							this->SPARKCloningWorkList.push_back(CloningWorkListItem);
							pair<int, int> UnlinkLimitPair2(HeadBlockNum, FollowBlockNum);
							pair<int, pair<int, int> > CloningWorkListItem2(FTBlockNum, UnlinkLimitPair2);
							this->SPARKCloningWorkList.push_back(CloningWorkListItem2);
							SMP_msg("INFO: SPARK: CLONING 6: Cloning doubly from %d and %d to %d would solve DomFrontier non-convergence for HeadBlock %d NoDeadEnds %d ShortCircuitFlag %d ", FTBlockNum, NFTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds, ShortCircuit);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
						}
						else {
							SMP_msg("INFO: SPARK: Cloning doubly from %d and %d to %d unsafe due to reachability for HeadBlock %d NoDeadEnds %d ShortCircuitFlag %d ", FTBlockNum, NFTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds, ShortCircuit);
						}
					}
				}
				if (!SafeCloning) {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (OddIfThenError) { // NFTNotDominated is the problem; clone starting at NFT
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(NFTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(NFTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 7: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d NoDeadEnds %d ", NFTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else if (IfThenError) { // FTNotDominated is the problem; clone starting at FT
				this->ResetTCFGVisitedBlocks();
				set<int> BlockNumsToClone;
				bool SafeCloning = this->IsCloningSafe(FTBlockNum, FollowBlockNum, BlockNumsToClone);
				if (SafeCloning) {
					pair<int, int> UnlinkLimitPair(HeadBlockNum, FollowBlockNum);
					pair<int, pair<int, int> > CloningWorkListItem(FTBlockNum, UnlinkLimitPair);
					this->SPARKCloningWorkList.push_back(CloningWorkListItem);
					SMP_msg("INFO: SPARK: CLONING 8: Cloning from %d to %d would solve DominanceProblem for HeadBlock %d NoDeadEnds %d ", FTBlockNum, FollowBlockNum, HeadBlockNum, NoDeadEnds);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
				else {
					FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning
				}
			}
			else {
				if (!this->PrintedSPARKUnstructuredMsg) {
					SMP_msg("ERROR: SPARK: Unstructured due to Dominance error in FindConditionalFollowNode4 at HeadBlockNum %d IfThenCase: %d OddIfThenCase: %d IfThenElseCase: %d ", HeadBlockNum, IfThenCase, OddIfThenCase, IfThenElseCase);
					this->DumpFuncNameAndAddr();
					this->PrintedSPARKUnstructuredMsg = true;
				}
				this->DumpDotCFG();
				IfThenCase = false;
				OddIfThenCase = false;
				IfThenElseCase = false;
				FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			}
		}
	}

	// Check sanity.
	if (0 <= FollowBlockNum) {
		assert(FollowBlockNum > HeadBlockNum); // RPO numbering should endure this relation
	}

	return FollowBlockNum;
} // end of SMPFunction::FindConditionalFollowNode4()

// return true if cloning blocks would make DonFrontiers converge
bool SMPFunction::SolveConditionalDomFrontierProblemViaCloning(const int HeadBlockNum, const int FTBlockNum, const int NFTBlockNum, const int FTDomFrontierBlockNum, const int NFTDomFrontierBlockNum, int &FollowBlockNum) {
	assert((0 < FTDomFrontierBlockNum) && (0 < NFTDomFrontierBlockNum));
	assert((0 < FTBlockNum) && (0 < NFTBlockNum));
	bool CloningStructuresConditional = false;
	FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	size_t IncomingWorkListSize = this->SPARKCloningWorkList.size();

	// We are looking for a conditional structure where exactly one of the if-then-else branches has a side entry
	//  that perturbs the TCFGDomFrontier checks, e.g.:
	//
	//        B1            B0
	//  +------+-----+      |
	//  |            |      |
	// B2           B3      |
	//  |            |      |
	// B4           B5------+
	//  |            |
	//  +------+-----+
	//         |
	//        B6
	//
	// Instead of B2 and B3 having B6 as their lone and common DomFrontier block, B3 has B5 as its DomFrontier
	//  block. If we clone block B5 as B5' we can have a clean if-then-else with B6 as the common DomFrontier
	//  block:
	//
	//        B1            B0
	//  +------+-----+      |
	//  |            |      |
	// B2           B3      |
	//  |            |      |
	// B4           B5      B5'
	//  |            |      |
	//  +------+-----+------+
	//         |
	//        B6
	//
	// Detection is simple. B2 has {B6} as its DomFrontier, while B3 has {B5}. We check the DomFrontier of
	//  B5 and find {B6} the convergence point with the other branch. We checked B5 because it has a lower
	//  RPO block number than B6; no point in checking B6 to see if {B5} is its DomFrontier as we are not
	//  looking to follow back edges to find a join point in a conditional.
	//
	// There could be a lot of code between B5 and B6 (another conditional, a loop, a switch). As long as that
	//  code does not bypass B6, the DomFrontier of B5 will be {B6}. Any more side entries that complicate
	//  the picture will prevent a simple DomFrontier detection like this. All code from B5 to B6 will need to
	//  be cloned.
	//
	// We can also search for DomFrontiers that stopped early on both sides, which converge later at the join point,
	//  which can be solved by cloning on both paths.
	//
	// We can also have the case of the FTBlock and NFTBlock having the side entrances, in which case they should
	//  already have equal DomFrontiers. That is detected without calling this method.

	STARSTransCFGBlock *FTDomFrontierBlock = this->GetTCFGBlockByNum(FTDomFrontierBlockNum);
	STARSTransCFGBlock *NFTDomFrontierBlock = this->GetTCFGBlockByNum(NFTDomFrontierBlockNum);
	uint16_t LoopBackCount = 0;
	uint16_t ForwardCount = 0;
	int OrigHeadBlockNum = this->GetOrigBlockNumFromTCFGBlockNum(HeadBlockNum);
	bool HeadBlockIsInALoop = this->IsTCFGBlockInAnyLoop(HeadBlockNum);
	int InnermostLoopNum = HeadBlockIsInALoop ? this->GetInnermostLoopNum(OrigHeadBlockNum) : -1;
	int NewFTDomFrontierBlockNum = FTDomFrontierBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, LoopBackCount, ForwardCount);
	bool GoodFTCount = (ForwardCount == 1);
	LoopBackCount = 0;
	ForwardCount = 0;
	int NewNFTDomFrontierBlockNum = NFTDomFrontierBlock->GetSingleDomFrontierBlockNumInsideLoop(InnermostLoopNum, LoopBackCount, ForwardCount);
	bool GoodNFTCount = (ForwardCount == 1);
	if (FTDomFrontierBlockNum < NFTDomFrontierBlockNum) {
		if (GoodFTCount && (NewFTDomFrontierBlockNum == NFTDomFrontierBlockNum)) {
			SMP_msg("INFO: SPARK: CLONING 3: Cloning from %d to %d would solve DomFrontier non-convergence for HeadBlock %d ", FTDomFrontierBlockNum, NFTDomFrontierBlockNum, HeadBlockNum);
			this->DumpFuncNameAndAddr();
			this->DumpDotCFG();

			FollowBlockNum = NFTDomFrontierBlockNum;
			CloningStructuresConditional = true;
			this->ResetTCFGVisitedBlocks();
			set<int> BlockNumsToClone;
			bool SafeCloning = this->IsCloningSafe(FTDomFrontierBlockNum, FollowBlockNum, BlockNumsToClone);
			// Guard against size blowup.
			if (SafeCloning && ((BlockNumsToClone.size() + this->TCFGBlocks.size()) < this->TCFGDomSoFar.GetNumBits())) {
				pair<int, int> UnlinkLimitPair(FTBlockNum, FollowBlockNum);
				pair<int, pair<int, int> > CloningWorkListItem(FTDomFrontierBlockNum, UnlinkLimitPair);
				this->SPARKCloningWorkList.push_back(CloningWorkListItem);
			}
			BlockNumsToClone.clear();
		}
	}
	else if (GoodNFTCount && (NewNFTDomFrontierBlockNum == FTDomFrontierBlockNum)) { // NFTDomFrontierBlockNum < FTDomFrontierBlockNum
		SMP_msg("INFO: SPARK: CLONING 4: Cloning from %d to %d would solve DomFrontier non-convergence for HeadBlock %d ", NFTDomFrontierBlockNum, FTDomFrontierBlockNum, HeadBlockNum);
		this->DumpFuncNameAndAddr();
		this->DumpDotCFG();
		this->ResetTCFGVisitedBlocks();
		FollowBlockNum = FTDomFrontierBlockNum;
		CloningStructuresConditional = true;
		this->ResetTCFGVisitedBlocks();
		set<int> BlockNumsToClone;
		bool SafeCloning = this->IsCloningSafe(NFTDomFrontierBlockNum, FollowBlockNum, BlockNumsToClone);
		// Guard against size blowup.
		if (SafeCloning && ((BlockNumsToClone.size() + this->TCFGBlocks.size()) < this->TCFGDomSoFar.GetNumBits())) {
			pair<int, int> UnlinkLimitPair(NFTBlockNum, FollowBlockNum);
			pair<int, pair<int, int> > CloningWorkListItem(NFTDomFrontierBlockNum, UnlinkLimitPair);
			this->SPARKCloningWorkList.push_back(CloningWorkListItem);
		}
		BlockNumsToClone.clear();
	}
	else if (GoodFTCount && GoodNFTCount && (NewFTDomFrontierBlockNum == NewNFTDomFrontierBlockNum)) {
		// Both branches have a side entry, and they converge neatly if we clone on both branches.
		this->ResetTCFGVisitedBlocks();
		set<int> BlockNumsToClone;
		set<int> BlockNumsToClone2;
		FollowBlockNum = NewFTDomFrontierBlockNum;
		bool SafeCloning = this->IsCloningSafe(FTDomFrontierBlockNum, FollowBlockNum, BlockNumsToClone);
		if (SafeCloning) {
			this->ResetTCFGVisitedBlocks();
			SafeCloning = this->IsCloningSafe(NFTDomFrontierBlockNum, FollowBlockNum, BlockNumsToClone2);
			if (SafeCloning) {
				pair<int, int> UnlinkLimitPair(NFTBlockNum, FollowBlockNum);
				pair<int, pair<int, int> > CloningWorkListItem(NFTDomFrontierBlockNum, UnlinkLimitPair);
				this->SPARKCloningWorkList.push_back(CloningWorkListItem);
				pair<int, int> UnlinkLimitPair2(FTBlockNum, FollowBlockNum);
				pair<int, pair<int, int> > CloningWorkListItem2(FTDomFrontierBlockNum, UnlinkLimitPair2);
				SMP_msg("INFO: SPARK: CLONING 5: Cloning doubly from %d and %d to %d would solve DomFrontier non-convergence for HeadBlock %d ", FTDomFrontierBlockNum, NFTDomFrontierBlockNum, NewFTDomFrontierBlockNum, HeadBlockNum);
				this->DumpFuncNameAndAddr();
				this->DumpDotCFG();
				this->ResetTCFGVisitedBlocks();
				CloningStructuresConditional = true;
			}
		}
	}

	// Don't return true or set FollowBlockNum until we are ready to clone.
	if (this->SPARKCloningWorkList.size() <= IncomingWorkListSize) { // no new item added to work list; not ready to clone
		FollowBlockNum = SMP_BLOCKNUM_UNINIT; // Not cloning yet
		CloningStructuresConditional = false; // Not cloning yet
	}

	return CloningStructuresConditional;
} // end of SMPFunction::SolveConditionalDomFrontierProblemViaCloning()

// Is it safe to clone path from StartTCFGBlockNum, stopping when LimitBlockNum is reached?
// NOTE: Call ResetSCCPVisitedBlocks() first
bool SMPFunction::IsCloningSafe(const int StartTCFGBlockNum, const int LimitTCFGBlockNum, set<int> &BlockNumsToClone) const {
	assert(0 < StartTCFGBlockNum);
	STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum((size_t)StartTCFGBlockNum);
	// We start our cloning with the manageable cases, so we exclude cloning loops or switch statements.
	bool TooBig = (this->TCFGBlocks.size() >= (3 * this->RPOBlocks.size())); // limit cloning explosion
	bool Convergence = (0 < LimitTCFGBlockNum);
	bool DeadEnd = (SMP_BLOCKNUM_COMMON_RETURN == LimitTCFGBlockNum);
	bool LoopProblem = CurrTCFGBlock->IsLoopHeaderBlock();
	bool SwitchProblem = CurrTCFGBlock->HasIndirectJump();
#if 0
	bool Safe = ((StartTCFGBlockNum < LimitTCFGBlockNum) && (!CurrTCFGBlock->IsLoopHeaderBlock()) && (!CurrTCFGBlock->HasIndirectJump()) && (!TooBig));
#else
	bool Safe = (Convergence || DeadEnd) && ((StartTCFGBlockNum != LimitTCFGBlockNum) && (!LoopProblem) && (!SwitchProblem) && (!TooBig));
#endif
	if (!Safe) {
		SMP_msg("ERROR: SPARK: IsCloningSafe() fails for StartBlock %d LimitBlock %d LoopProblem %d SwitchProblem %d TooBig %d ", 
			StartTCFGBlockNum, LimitTCFGBlockNum, LoopProblem, SwitchProblem, TooBig);
		this->DumpFuncNameAndAddr();
	}

	if (Safe && (!CurrTCFGBlock->IsVisited())) {
		BlockNumsToClone.insert(StartTCFGBlockNum);
		CurrTCFGBlock->SetVisited(true);
		// recurse into successors
		for (STARSTransCFGBlock *SuccTCFGBlock : CurrTCFGBlock->GetConstSuccList()) {
			int SuccTCFGBlockNum = SuccTCFGBlock->GetNumber();
#if 0
			if ((SuccTCFGBlockNum > StartTCFGBlockNum) && (SuccTCFGBlockNum < LimitTCFGBlockNum)) { // no back edges
#else
			if ((SuccTCFGBlockNum > StartTCFGBlockNum) && (SuccTCFGBlockNum != LimitTCFGBlockNum)) { // no back edges
#endif
				Safe = this->IsCloningSafe(SuccTCFGBlockNum, LimitTCFGBlockNum, BlockNumsToClone);
				if (!Safe)
					break; // bail out on first problem encountered
			}
		} // end for all successors
	}

	// One last check for code explosion
	TooBig = (this->TCFGDomSoFar.GetNumBits() < (BlockNumsToClone.size() + this->TCFGBlocks.size()));
	if (TooBig) {
		Safe = false;
		SMP_msg("ERROR: SPARK: TooBig2: IsCloningSafe() fails for StartBlock %d LimitBlock %d ", StartTCFGBlockNum, LimitTCFGBlockNum);
		this->DumpFuncNameAndAddr();
	}

	if (!Safe) {
		BlockNumsToClone.clear();
		SMP_msg("ERROR: SPARK: IsCloningSafe() fails for StartBlock %d LimitBlock %d ", StartTCFGBlockNum, LimitTCFGBlockNum);
		this->DumpFuncNameAndAddr();
	}

	return Safe;
} // end of SMPFunction::IsCloningSafe()

// Is it safe to double clone paths from StartBlockNum1 & 2, without either path running into the other start block?
bool SMPFunction::IsDoubleCloningSafe(const int StartTCFGBlockNum1, const int StartTCFGBlockNum2) {
	// We only need to test from the lesser RPO number to the greater RPO number.
	bool Reachable = true; // in case block nums are passed in equal
	if ((0 < StartTCFGBlockNum1) && (0 < StartTCFGBlockNum2)) {
		bool HitDeadEnd = false; // don't care; argument placeholder
		if (StartTCFGBlockNum1 < StartTCFGBlockNum2) {
			STARSTransCFGBlock *TCFGBlock1 = this->GetTCFGBlockByNum((size_t)StartTCFGBlockNum1);
			this->ResetTCFGVisitedBlocks();
			Reachable = TCFGBlock1->IsBlockReachableWithoutBackEdges(StartTCFGBlockNum2, HitDeadEnd);
		}
		else if (StartTCFGBlockNum2 < StartTCFGBlockNum1) {
			STARSTransCFGBlock *TCFGBlock2 = this->GetTCFGBlockByNum((size_t)StartTCFGBlockNum2);
			this->ResetTCFGVisitedBlocks();
			Reachable = TCFGBlock2->IsBlockReachableWithoutBackEdges(StartTCFGBlockNum1, HitDeadEnd);
		}
	}

	return (!Reachable);
} // end of SMPFunction::IsDoubleCloningSafe()

// Clone blocks on path from StartingBlockNum to <LimitBlockNum within InnermostLoopNum; unlink cloned StartingBlockNum from pred UnlinkBlockNum.
//  Uses SCCPVisited flags to avoid redundant visits of blocks.
//  Push cloned blocks onto end of TCFGBlocks.
//  Update ClonedShadowCFGBlocks in lockstep with TCFGBlocks.
void SMPFunction::MarkBlocksAndClone(const int StartingBlockNum, const int UnlinkBlockNum, const int LimitBlockNum, const int InnermostLoopNum, bool FirstIter) {
	size_t FirstClonedBlockIndex = 0;
	if (FirstIter) {
		FirstClonedBlockIndex = this->TCFGBlocks.size();
	}

	bool InsideLoop = (0 <= InnermostLoopNum);
	bool ReachedLimit = (StartingBlockNum == LimitBlockNum);
	if (!ReachedLimit) {
		STARSTransCFGBlock *StartBlock = this->GetTCFGBlockByNum((size_t)StartingBlockNum);
		STARSCFGBlock *StartShadowBlock = this->GetClonedCFGBlockByNum((size_t)StartingBlockNum);
		assert(nullptr != StartBlock);
		assert(nullptr != StartShadowBlock);
		bool PassesLoopTest = ((!InsideLoop) || this->IsTCFGBlockInLoop(StartingBlockNum, (size_t)InnermostLoopNum));
		if (!StartBlock->IsVisited() && PassesLoopTest) {
			SMPBasicBlock *OrigStartBlock = StartBlock->GetOriginalBlock();
			OrigStartBlock->IncrementSPARKTranslationLimit();
			StartBlock->SetVisited(true);

			// Clone the block and push onto TCFGBlocks list, unlinking from predecessor UnlinkBlockNum if found.
			STARSTransCFGBlock *BlockClone = new STARSTransCFGBlock(*StartBlock);
			int NewBlockNum = (int)this->TCFGBlocks.size();
			BlockClone->SetBlockNum(NewBlockNum);
			this->TCFGBlocks.push_back(BlockClone);

			// Repeat for ShadowCFG.
			STARSCFGBlock *ShadowBlockClone = StartShadowBlock->DeepCopyExpr();
			ShadowBlockClone->SetBlockNum(NewBlockNum);
			// Constructor sets OriginalBlockNum to same value as BlockNum, so update it.
			ShadowBlockClone->SetOriginalBlockNum(StartShadowBlock->GetOriginalBlockNum());
			this->ClonedShadowCFGBlocks.push_back(ShadowBlockClone);

			// Clone OddIfThenBranches entry if needed.
			if (this->IsOddIfThenBranch(StartingBlockNum)) {
				pair<set<int>::iterator, bool> InsertResult = this->OddIfThenBranches.insert(NewBlockNum);
				assert(InsertResult.second);
			}

			// NOTE: ShadowCFG does not maintain Pred and Succ links.
			if (FirstIter) {
				this->ResetTCFGDomSoFar(); // prepare for dom tests on UnlinkBlockNum in UnlinkPred()
											// and EnforceDominanceOnPreds
				BlockClone->UnlinkPred(UnlinkBlockNum); // Unlink cloned block from any pred dominated by UnlinkBlockNum
				// Unlink source block from any predecessor not dominated by UnlinkBlockNum, to eliminate unstructured edge.
				bool PredErased = StartBlock->EnforceDominanceOnPreds(UnlinkBlockNum, NewBlockNum);
#if 0
				if (!PredErased) {
					SMP_msg("WARNING: SPARK: !PredErased in MarkBlocksAndClone for UnlinkBlockNum %d StartingBlockNum %d ", UnlinkBlockNum, StartingBlockNum);
				}
#else
				assert(PredErased);
#endif

				// The beginning of the path to clone is where some "side entrance" to some structure occurs.
				//  That means that entries in the JumpFollowNodesMap could be pointing to the beginning block.
				//  They should all be redirected to the clone of the beginning block. If a map entry starts with
				//  a block dominated by the StartingBlockNum, we need to clone the entire entry, which we do later.
				map<int, int>::iterator JumpMapIter = this->JumpFollowNodesMap.begin();
				while (JumpMapIter != this->JumpFollowNodesMap.end()) {
					if (JumpMapIter->second == StartingBlockNum) {
						int SourceBlockNum = JumpMapIter->first;
						if (!this->DoesTCFGBlockDominateTCFGBlock(UnlinkBlockNum, SourceBlockNum)) {
							JumpMapIter->second = NewBlockNum;
							SMP_msg("INFO: SPARK: Updating JumpMap entry from %d %d to %d %d after cloning ",
								SourceBlockNum, StartingBlockNum, SourceBlockNum, NewBlockNum);
							this->DumpFuncNameAndAddr();
						}
					}
					++JumpMapIter;
				} // end while not end of JumpMap
			} // end if FirstIter

			pair<int, int> MapItem(StartingBlockNum, NewBlockNum);
			pair<int, int> ReverseMapItem(NewBlockNum, StartingBlockNum);
			pair<map<int, int>::iterator, bool> InsertResult = this->CloningMap.insert(MapItem);
			assert(InsertResult.second); // no conflict should be possible
			InsertResult = this->CloningReverseMap.insert(ReverseMapItem);
			assert(InsertResult.second); // no conflict should be possible

			for (STARSTransCFGBlock *SuccBlock : StartBlock->GetConstSuccList()) {
				int SuccBlockNum = SuccBlock->GetNumber();
				if (SuccBlockNum > StartingBlockNum) { // no back edges
					// recurse
					this->MarkBlocksAndClone(SuccBlockNum, UnlinkBlockNum, LimitBlockNum, InnermostLoopNum, false);
				}
				if (SuccBlockNum == LimitBlockNum) {
					// Link current cloned block to its non-cloned successor's pred list
					SuccBlock->AddPred(BlockClone);
				}
			}
		}
	}

	if (!this->CloningMap.empty() && FirstIter) {
		// We are finished cloning on this path. Update the JumpFollowNodesMap.
		// Update JumpFollowNodesMap. For a current pair <sourceblocknum, targetblocknum> we have four possibilities:
		//  1. Neither block was cloned; do nothing
		//  2. Only targetblock was cloned; if sourceblocknum not dominated by unlinkblocknum, handled in FirstIter code above, else we leave entry alone.
		//  3. Only sourceblock was cloned; clone entry with new sourceblocknum+old targetblocknum. Merging back into CFG from cloned region.
		//  4. Both blocks in the pair were cloned; clone entry with two new blocknums from CloningMap.
		// Add cloned entries to the JumpFollowNodesMap
		map<int, int> CloneMapItems; // append these later to JumpFollowNodesMap; don't insert while iterating over it
		for (pair<int, int> JumpMapItem : this->JumpFollowNodesMap) {
			int SourceTCFGBlockNum = JumpMapItem.first;
			int TargetTCFGBlockNum = JumpMapItem.second;
			map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(SourceTCFGBlockNum);
			bool ClonedSource = (CloningMapIter != this->CloningMap.cend());
			map<int, int>::const_iterator CloningMapIter2 = this->CloningMap.find(TargetTCFGBlockNum);
			bool ClonedTarget = (CloningMapIter2 != this->CloningMap.cend());

			if (ClonedSource) { // Cases 3 & 4.
				pair<int, int> CloningMapItem(*CloningMapIter);
				int NewTCFGSourceBlockNum = CloningMapItem.second;
				assert(0 < NewTCFGSourceBlockNum);

				// Find the new TCFG block num for the follow node, if any.
				int NewTCFGTargetBlockNum = ClonedTarget ? CloningMapIter2->second : TargetTCFGBlockNum;
				pair<int, int> NewJumpMapItem(NewTCFGSourceBlockNum, NewTCFGTargetBlockNum);
				pair<map<int, int>::iterator, bool> JumpInsertResult = CloneMapItems.insert(NewJumpMapItem);
				assert(JumpInsertResult.second);
			}
			else if (ClonedTarget) {
				// Case 2.
#if 0
				if (UnlinkBlockNum != SourceTCFGBlockNum) {
					this->SetHasUnstructuredCFG();
					SMP_msg("ERROR: SPARK: Cannot update jump table entry %d to %d with UnlinkBlockNum %d ",
						SourceTCFGBlockNum, TargetTCFGBlockNum, UnlinkBlockNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
#endif
				; // do nothing
			}
		} // end for all JumpFollowNodesMap entries
		if (!CloneMapItems.empty()) {
			// Append the cloned items to JumpFollowNodesMap.
			for (pair<int, int> CloneItem : CloneMapItems) {
				pair<map<int, int>::iterator, bool> JumpInsertResult = this->JumpFollowNodesMap.insert(CloneItem);
				if (JumpInsertResult.second) {
					SMP_msg("INFO: SPARK: Adding JumpMapItem after cloning; head block %d follow block %d ", CloneItem.first, CloneItem.second);
				}
				else {
					SMP_msg("INFO: SPARK: Adding redundant JumpMapItem after cloning; old %d %d new %d %d ",
						JumpInsertResult.first->first, JumpInsertResult.first->second, CloneItem.first, CloneItem.second);
				}
				this->DumpFuncNameAndAddr();
			} // end for all CloneMapItems
			CloneMapItems.clear();
		}

		// Update block numbers only in the short-circuit blocks that were just cloned.
		//  This step is needed now because of the double-cloning possibility. If we have
		//  two items in the SPARKCloningWorkList, we need to keep track of block numbers
		//  at each step, not just after renumbering all blocks in UpdateAfterCloning().
		for (size_t ClonedBlockIndex = FirstClonedBlockIndex; ClonedBlockIndex < this->TCFGBlocks.size(); ++ClonedBlockIndex) {
			STARSCFGBlock *ClonedCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
			STARSCondExpr *CurrExpr = ClonedCFGBlock->GetExpr();
			if (nullptr != CurrExpr) {
				CurrExpr->ResetAllExprsVisited(); // prepare to avoid redundant updates
			}
		} // end for all ClonedBlockIndex values only

		for (size_t ClonedBlockIndex = FirstClonedBlockIndex; ClonedBlockIndex < this->TCFGBlocks.size(); ++ClonedBlockIndex) {
			STARSCFGBlock *ClonedCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
			bool ExprHeadBlock = ClonedCFGBlock->IsExprHeadBlock();
			bool UpdateBlockNums = (ClonedCFGBlock->IsCoalesced() || ExprHeadBlock);
			if (UpdateBlockNums) {
				// Part of short circuit expr; need to adjust Left, right, CoalescedInto, FT and NFT block numbers.
				//  This is the entire purpose of maintaining a cloned ShadowCFG. When
				//  a short-circuit expr is cloned, it appears in two places in the CFG
				//  and has two different FTBlock #'s and two different NFTBlock #'s in the TCFG.
				// Because all exprs were deep-copied on creation, we have to recurse through subexprs.
				this->UpdateCFGExprBlockNums(ClonedCFGBlock->GetExpr());

				// Update CoalescedBlockNum when not in the Head block.
				if (!ExprHeadBlock) {
					map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(ClonedCFGBlock->GetCoalescedBlockNum());
					assert(CloningMapIter != this->CloningMap.cend());
					int NewCoalescedBlockNum = CloningMapIter->second;
					ClonedCFGBlock->SetCoalescedBlockNum(NewCoalescedBlockNum);
				}

				// Update MyCFGBlock pointer, even though it seems to only be used to access the GetFunc()
				//  method, which works fine without updating.
				STARSCFGBlock *NewCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
				assert(nullptr != NewCFGBlock);
				ClonedCFGBlock->GetExpr()->SetCFGBlock(NewCFGBlock);
			}
		} // end for all ClonedBlockIndex values only
	}

	return;
} // end of SMPFunction::MarkBlocksAndClone()

// Update data structures, set pred & succ links, etc. after cloning
bool SMPFunction::UpdateAfterCloning(void) {
	bool success = true;

	int LowestNewBlockNum = 2 + (int)this->TCFGBlocks.size();
	for (pair<int, int> ReverseMapItem : this->CloningReverseMap) {
		int NewTCFGBlockNum = ReverseMapItem.first;
		if (NewTCFGBlockNum < LowestNewBlockNum) {
			LowestNewBlockNum = NewTCFGBlockNum;
		}
	} // end for all CloningReverseMap items

	// Clear all the old Dominance frontiers.
	for (STARSTransCFGBlock *TCFGBlock : this->TCFGBlocks) {
		TCFGBlock->ClearDomFrontiersAfterCloning();
	}

#if 1
	// Clone predecessor and successor links to newly cloned blocks.
	// NOTE: ShadowCFG does not maintain Pred and Succ links.
	for (pair<int, int> ReverseMapItem : this->CloningReverseMap) {
		int NewTCFGBlockNum = ReverseMapItem.first;
		int SourceTCFGBlockNum = ReverseMapItem.second;
		STARSTransCFGBlock *SourceTCFGBlock = this->GetTCFGBlockByNum((size_t)SourceTCFGBlockNum);
		assert(nullptr != SourceTCFGBlock);
		STARSTransCFGBlock *NewTCFGBlock = this->GetTCFGBlockByNum((size_t)NewTCFGBlockNum);

		// Redirect predecessor links to newly cloned blocks.
		list<STARSTransCFGBlock *> NewPredList;
		for (STARSTransCFGBlock *SourcePredTCFGBlock : SourceTCFGBlock->GetConstPredList()) {
			int SourcePredNum = SourcePredTCFGBlock->GetNumber();
			// The SourcePredNum might be in the cloning maps, or it might be the block outside
			//  the cloning region that is being redirected to the cloned blocks subgraph. If the former,
			//  we need to look up the new cloned block num that corresponds to the old block num; if the
			//  latter, we have already updated links in MarkBlocksAndClone().
			map<int, int>::const_iterator MapIter = this->CloningMap.find(SourcePredNum);
			if (this->CloningMap.cend() != MapIter) { // part of the cloned subgraph
				pair<int, int> CloningMapItem = (*MapIter);
				int NewPredBlockNum = CloningMapItem.second;
				STARSTransCFGBlock *NewPredTCFGBlock = this->GetTCFGBlockByNum((size_t)NewPredBlockNum);
				NewPredList.push_back(NewPredTCFGBlock);
			}
		} // end for all source pred blocks
		// Replace old Preds list with new list.
		if (!NewPredList.empty()) { // not the head of the cloned group, which was handled in MarkBlocksAndClone().
			NewTCFGBlock->ReplacePredsAfterCloning(NewPredList);
		}

		// Clone successor links.
		NewTCFGBlock->ClearSuccsAfterCloning();
		for (STARSTransCFGBlock *SourceSuccTCFGBlock : SourceTCFGBlock->GetConstSuccList()) {
			int SourceSuccNum = SourceSuccTCFGBlock->GetNumber();
			// The SourceSuccNum might be in the cloning maps, or it might be the dom frontier block outside
			//  the cloning region that is being used as the LimitBlockNum for cloning. If the former,
			//  we need to look up the new cloned block num that corresponds to the old block num; if the
			//  latter, we just use the old block num.
			map<int, int>::const_iterator MapIter = this->CloningMap.find(SourceSuccNum);
			if (this->CloningMap.cend() == MapIter) { // before (e.g. loop-back) or after the cloned subgraph
				NewTCFGBlock->AddSucc(SourceSuccTCFGBlock); // add original successor block, not cloned anyway
				if (!SourceSuccTCFGBlock->HasPred(NewTCFGBlock)) { // probably loop-back to before cloned region
					SourceSuccTCFGBlock->AddPred(NewTCFGBlock);
				}
			}
			else { // part of the cloned subgraph
				pair<int, int> CloningMapItem = (*MapIter);
				int NewSuccBlockNum = CloningMapItem.second;
				STARSTransCFGBlock *NewSuccTCFGBlock = this->GetTCFGBlockByNum((size_t)NewSuccBlockNum);
				NewTCFGBlock->AddSucc(NewSuccTCFGBlock);
			}
		} // end for all source succ blocks
	}
#endif


	// Renumber the TCFGBlocks in RPO order, keeping maps to use to update data structures afterwards.
	size_t OldTCFGBlocksSize = this->TCFGBlocks.size();
	this->RPONumberTCFGBlocks();
	assert(this->CloningMap.size() == this->TCFGBlocks.size());
	if (this->CloningMap.size() != OldTCFGBlocksSize) {
		for (int BlockNum = 0; BlockNum < ((int)this->CloningMap.size()); ++BlockNum) {
			if (this->CloningMap.find(BlockNum) == this->CloningMap.end()) {
				SMP_msg("ERROR: SPARK: BlockNum %d not found in CloningMap after renumbering TCFG ", BlockNum);
				this->DumpFuncNameAndAddr();
			}
		}
		this->AuditTCFG();
	}
	assert(this->CloningMap.size() == OldTCFGBlocksSize); // didn't miss any blocks

	// Recompute TCFGIdom.
	this->TCFGIDom.clear();
	this->ComputeTCFGIDoms();
	// Recompute the DomFrontier for each TCFG block.
	this->ComputeTCFGDomFrontiers();
	// Recompute TCFGDomTree.
	this->TCFGDomTree.clear();
	this->BuildTCFGDominatorTree();

	// Update the TCFG block numbers in OddIfThenBranches.
	if (!this->OddIfThenBranches.empty()) {
		set<int> TempOddIfThenBranches;
		for (int BranchTCFGBlockNum : this->OddIfThenBranches) {
			map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(BranchTCFGBlockNum);
			assert(CloningMapIter != this->CloningMap.cend()); // all blocks are in the CloningMap after renumbering
			pair<set<int>::iterator, bool> InsertResult = TempOddIfThenBranches.insert(CloningMapIter->second);
		}
		this->OddIfThenBranches.clear();
		this->OddIfThenBranches = TempOddIfThenBranches;
	}

	// Update the TCFG block numbers in LoopExitTargets, LoopExitTargetsReachabilitySets, and 
	//  MultiLoopExitTargetsConvergenceMap.
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		set<int> TempExitTargetsSet(this->LoopExitTargets[LoopIndex]); // make a copy
		this->LoopExitTargets[LoopIndex].clear(); // clear original, ready to insert new block numbers
		for (int ExitTargetTCFGBlockNum : TempExitTargetsSet) {
			if (0 > ExitTargetTCFGBlockNum) // could be do-forever loop
				continue;
			map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(ExitTargetTCFGBlockNum);
			assert(CloningMapIter != this->CloningMap.cend()); // all blocks are in the CloningMap after renumbering
			int NewTCFGBlockNum = CloningMapIter->second;
			assert(0 < NewTCFGBlockNum);
			pair<set<int>::iterator, bool> InsertResult = this->LoopExitTargets[LoopIndex].insert(NewTCFGBlockNum);
			assert(InsertResult.second);
		} // end for all old entries copied into TempExitTargetsSet
	} // end for all loops

	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		set<int> TempExitTargetsSet(this->MultiLoopExitTargetsConvergenceMap[LoopIndex]); // make a copy
		this->MultiLoopExitTargetsConvergenceMap[LoopIndex].clear(); // clear original, ready to insert new block numbers
		for (int ExitTargetTCFGBlockNum : TempExitTargetsSet) {
			if (ExitTargetTCFGBlockNum < 0) { // code like -2 for SMP_BLOCKNUM_COMMON_RETURN
				pair<set<int>::iterator, bool> InsertResult = this->MultiLoopExitTargetsConvergenceMap[LoopIndex].insert(ExitTargetTCFGBlockNum);
				assert(InsertResult.second);
			}
			else {
				map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(ExitTargetTCFGBlockNum);
				assert(CloningMapIter != this->CloningMap.cend()); // all blocks are in the CloningMap after renumbering
				int NewTCFGBlockNum = CloningMapIter->second;
				assert(0 < NewTCFGBlockNum);
				pair<set<int>::iterator, bool> InsertResult = this->MultiLoopExitTargetsConvergenceMap[LoopIndex].insert(NewTCFGBlockNum);
				assert(InsertResult.second);
			}
		} // end for all old entries copied into TempExitTargetsSet
	} // end for all loops

	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		map<int, STARSBitSet> TempExitTargetsMap(this->LoopExitTargetsReachabilitySets[LoopIndex]); // make a copy
		this->LoopExitTargetsReachabilitySets[LoopIndex].clear(); // clear original, ready to insert new block numbers
		for (pair<int, STARSBitSet> ExitTargetMapItem : TempExitTargetsMap) {
			int ExitTargetTCFGBlockNum = ExitTargetMapItem.first;
			map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(ExitTargetTCFGBlockNum);
			assert(CloningMapIter != this->CloningMap.cend()); // all blocks are in the CloningMap after renumbering
			int NewExitTargetTCFGBlockNum = CloningMapIter->second;
			assert(0 < NewExitTargetTCFGBlockNum);
			STARSBitSet NewReachabilitySet;
			NewReachabilitySet.AllocateBits(this->TCFGBlocks.size());
			int LowestBitIndex = ExitTargetMapItem.second.FindLowestBitSet();
			int HighestBitIndex = ExitTargetMapItem.second.FindHighestBitSet();
			for (int OldBlockNum = LowestBitIndex; OldBlockNum <= HighestBitIndex; ++OldBlockNum) {
				if (ExitTargetMapItem.second.GetBit((size_t)OldBlockNum)) {
					map<int, int>::const_iterator NewCloningMapIter = this->CloningMap.find(OldBlockNum);
					assert(NewCloningMapIter != this->CloningMap.cend()); // all blocks are in the CloningMap after renumbering
					int NewBlockNum = NewCloningMapIter->second;
					assert(0 <= NewBlockNum);
					NewReachabilitySet.SetBit((size_t)NewBlockNum);
				}
			} // end for OldBlockNums in range
			pair<int, STARSBitSet> NewEntry(NewExitTargetTCFGBlockNum, NewReachabilitySet);
			pair<map<int, STARSBitSet>::iterator, bool> InsertResult = this->LoopExitTargetsReachabilitySets[LoopIndex].insert(NewEntry);
			assert(InsertResult.second);
		} // end for all old entries copied into TempExitTargetsMap
		TempExitTargetsMap.clear();
	} // end for all loops

	// Update the block numbers in ClonedShadowCFGBlocks, using the maps of old and new TCFG
	//  block numbers that were created above.
	vector<STARSCFGBlock *> TempShadowCFG; // for reordering
	TempShadowCFG.resize(this->ClonedShadowCFGBlocks.size(), nullptr);
	for (size_t ClonedBlockIndex = 0; ClonedBlockIndex < this->ClonedShadowCFGBlocks.size(); ++ClonedBlockIndex) {
		STARSCFGBlock *ClonedCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
		STARSCondExpr *CurrExpr = ClonedCFGBlock->GetExpr();
		if (nullptr != CurrExpr) {
			CurrExpr->ResetAllExprsVisited(); // prepare to avoid redundant updates
		}
	}
	for (size_t ClonedBlockIndex = 0; ClonedBlockIndex < this->ClonedShadowCFGBlocks.size(); ++ClonedBlockIndex) {
		STARSCFGBlock *ClonedCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
		bool IsExprHeadBlock = ClonedCFGBlock->IsExprHeadBlock();
		bool UpdateBlockNums = (ClonedCFGBlock->IsCoalesced() || IsExprHeadBlock);
		if (UpdateBlockNums) {
			// Part of short circuit expr; need to adjust Left, right, FT and NFT block numbers.
			//  This is the entire purpose of maintaining a cloned ShadowCFG. When
			//  a short-circuit expr is cloned, it appears in two places in the CFG
			//  and has two different FTBlock #'s and two different NFTBlock #'s in the TCFG.
			// Because all exprs were deep-copied on creation, we have to recurse through subexprs.
			this->UpdateCFGExprBlockNums(ClonedCFGBlock->GetExpr());

			// Update CoalescedBlockNum when not in the Head block.
			if (!IsExprHeadBlock) {
				map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(ClonedCFGBlock->GetCoalescedBlockNum());
				assert(CloningMapIter != this->CloningMap.cend());
				int NewCoalescedBlockNum = CloningMapIter->second;
				ClonedCFGBlock->SetCoalescedBlockNum(NewCoalescedBlockNum);
			}

			// Update MyCFGBlock pointer, even though it seems to only be used to access the GetFunc()
			//  method, which works fine without updating.
			STARSCFGBlock *NewCFGBlock = this->GetClonedCFGBlockByNum(ClonedBlockIndex);
			assert(nullptr != NewCFGBlock);
			ClonedCFGBlock->GetExpr()->SetCFGBlock(NewCFGBlock);
		}

		// Place ClonedCFGBlock into its new position, matching the TCFG block ordering.
		map<int, int>::iterator CloneMapIter = this->CloningMap.find((int)ClonedBlockIndex);
		assert(this->CloningMap.end() != CloneMapIter);
		int NewCFGBlockNum = CloneMapIter->second;
		assert(0 <= NewCFGBlockNum);
		assert(((size_t)NewCFGBlockNum) < TempShadowCFG.size());
		ClonedCFGBlock->SetBlockNum(NewCFGBlockNum);
		TempShadowCFG[(size_t)NewCFGBlockNum] = ClonedCFGBlock;
	} // end for all blocks in this->ClonedShadowCFGBlocks

	// Now, copy the Temp vector into the final cloned vector.
	this->ClonedShadowCFGBlocks.clear();
	this->ClonedShadowCFGBlocks = TempShadowCFG;
	// TempShadowCFG.clear();

	// Now that we added items to JumpFollowNodesMap for cloned branches, we need to renumber
	//  the completed map based on the new RPO numbering of TCFG.
	map<int, int> TempJumpFollowNodesMap;
	for (pair<int, int> JumpMapPair : this->JumpFollowNodesMap) {
		int SourceBlockNum = JumpMapPair.first;
		int TargetBlockNum = JumpMapPair.second;
		map<int, int>::const_iterator CloningMapIter = this->CloningMap.find(SourceBlockNum);
		assert(this->CloningMap.cend() != CloningMapIter);
		int NewSourceBlockNum = CloningMapIter->second;
		int NewTargetBlockNum = TargetBlockNum;
		if (0 <= TargetBlockNum) { // might be -2 or -1 codes (e.g. SMP_BLOCKNUM_COMMON_RETURN)
			CloningMapIter = this->CloningMap.find(TargetBlockNum);
			assert(this->CloningMap.cend() != CloningMapIter);
			NewTargetBlockNum = CloningMapIter->second;
		}
		pair<int, int> NewJumpMapPair(NewSourceBlockNum, NewTargetBlockNum);
		pair<map<int, int>::iterator, bool> InsertResult = TempJumpFollowNodesMap.insert(NewJumpMapPair);
		assert(InsertResult.second);
		SMP_msg("INFO: SPARK: Altering JumpMap entry after cloning; old pair %d %d new pair %d %d ",
			SourceBlockNum, TargetBlockNum, NewSourceBlockNum, NewTargetBlockNum);
		this->DumpFuncNameAndAddr();
	} // end for all old entries in JumpFollowNodesMap
	this->JumpFollowNodesMap.clear();
	this->JumpFollowNodesMap = TempJumpFollowNodesMap;

	return success;
} // end of SMPFunction::UpdateAfterCloning()

// Update left, right, FT and NFT block nums in CurrExpr; recurse into subexprs.
void SMPFunction::UpdateCFGExprBlockNums(STARSCondExpr *CurrExpr) {
	if (!CurrExpr->IsVisited()) {
		int FTBlockNum = CurrExpr->GetFallThroughBlockNum();
		int NFTBlockNum = CurrExpr->GetNonFallThroughBlockNum();
		// For an if-then, the LimitBlockNum on the cloning could be the FTBlockNum or NFTBlockNum,
		//  so it might not be found in the CloningMap. We could be more robust by actually matching
		//  the FTBlockNum or NFTBlockNum that is missing from the CloningMap to a LimitBlockNum in
		//  the SPARKCloningWorkList, and asserting that it must be found in the work list or the CloningMap.
		map<int, int>::iterator CloneMapIter = this->CloningMap.find(FTBlockNum);
		int NewFTBlockNum = (this->CloningMap.end() != CloneMapIter) ? CloneMapIter->second : FTBlockNum;
		CloneMapIter = this->CloningMap.find(NFTBlockNum);
		int NewNFTBlockNum = (this->CloningMap.end() != CloneMapIter) ? CloneMapIter->second : NFTBlockNum;
		CurrExpr->SetFTBlockNum(NewFTBlockNum);
		CurrExpr->SetNFTBlockNum(NewNFTBlockNum);
		// Likewise, update the Left and Right blocknums in the expression.
		int LeftBlockNum = CurrExpr->GetLeftBlockNum();
		if (0 <= LeftBlockNum) {
			map<int, int>::iterator CloneMapIter = this->CloningMap.find(LeftBlockNum);
			assert(this->CloningMap.end() != CloneMapIter);
			int NewLeftBlockNum = CloneMapIter->second;
			bool NegateLeft = CurrExpr->IsLeftExprNegated();
			CurrExpr->SetLeftExprBlockNum(NewLeftBlockNum, NegateLeft);
		}
		int RightBlockNum = CurrExpr->GetRightBlockNum();
		if (0 <= RightBlockNum) {
			map<int, int>::iterator CloneMapIter = this->CloningMap.find(RightBlockNum);
			assert(this->CloningMap.end() != CloneMapIter);
			int NewRightBlockNum = CloneMapIter->second;
			CurrExpr->SetRightExprBlockNum(NewRightBlockNum);
		}

		CurrExpr->SetVisited(true);
	}

	// Recurse into left and right subexprs if they exist.
	if (nullptr != CurrExpr->GetLeftExpr()) {
		this->UpdateCFGExprBlockNums(CurrExpr->GetLeftExpr());
	}
	if (nullptr != CurrExpr->GetRightExpr()) {
		this->UpdateCFGExprBlockNums(CurrExpr->GetRightExpr());
	}

	return;
} // end of SMPFunction::UpdateCFGExprBlockNums()

// Check TCFG sanity after cloning; return true if sane, false if problem seen.
bool SMPFunction::AuditTCFG(void) const {
	bool Sane = true;
	// Make sure all Pred and Succ links are reciprocal and block numbers are in range.
	int BlockNumLimit = (int)this->TCFGBlocks.size();
	for (const STARSTransCFGBlock *TCFGBlock : this->TCFGBlocks) {
		int CurrBlockNum = TCFGBlock->GetNumber();

		if ((0 > CurrBlockNum) || (CurrBlockNum > BlockNumLimit)) {
			Sane = false;
			SMP_msg("ERROR: TCFG: Block %d has out-of-range number ", CurrBlockNum);
			this->DumpFuncNameAndAddr();
		}

		for (const STARSTransCFGBlock *PredBlock : TCFGBlock->GetConstPredList()) {
			int PredBlockNum = PredBlock->GetNumber();
			if (!PredBlock->HasSucc(TCFGBlock)) {
				Sane = false;
				SMP_msg("ERROR: TCFG: Block %d has non-reciprocal PredBlock %d ", CurrBlockNum, PredBlockNum);
				this->DumpFuncNameAndAddr();
			}
			if ((0 > PredBlockNum) || (PredBlockNum > BlockNumLimit)) {
				Sane = false;
				SMP_msg("ERROR: TCFG: Block %d has out-of-range PredBlockNum %d ", CurrBlockNum, PredBlockNum);
				this->DumpFuncNameAndAddr();
			}
		} // end for all predecessors

		for (const STARSTransCFGBlock *SuccBlock : TCFGBlock->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();
			if (!SuccBlock->HasPred(TCFGBlock)) {
				Sane = false;
				SMP_msg("ERROR: TCFG: Block %d has non-reciprocal SuccBlock %d ", CurrBlockNum, SuccBlockNum);
				this->DumpFuncNameAndAddr();
			}
			if ((0 > SuccBlockNum) || (SuccBlockNum > BlockNumLimit)) {
				Sane = false;
				SMP_msg("ERROR: TCFG: Block %d has out-of-range SuccBlockNum %d ", CurrBlockNum, SuccBlockNum);
				this->DumpFuncNameAndAddr();
			}
		} // end for all successors
	} // end for all TCFGBlocks

	if (this->TCFGBlocks.size() > this->TCFGDomSoFar.GetNumBits()) {
		Sane = false;
		SMP_msg("ERROR: TCFG: Block count %zu has grown larger than TCFGDomSoFar size %zu ", this->TCFGBlocks.size(), this->TCFGDomSoFar.GetNumBits());
		this->DumpFuncNameAndAddr();
	}

	// Audit the JumpFollowNodesMap.
	for (pair<int, int> JumpFollowNode : this->JumpFollowNodesMap) {
		int SourceBlockNum = JumpFollowNode.first;
		int TargetBlockNum = JumpFollowNode.second;
		assert(0 <= SourceBlockNum);
		if (0 > TargetBlockNum) {
			assert(SMP_BLOCKNUM_COMMON_RETURN == TargetBlockNum);
			continue;
		}
		if (TargetBlockNum <= SourceBlockNum) {
			// Back edge? Must go to loop header node.
			STARSTransCFGBlock *TargetBlock = this->GetTCFGBlockByNum((size_t)TargetBlockNum);
			if (!TargetBlock->IsLoopHeaderBlock()) {
				Sane = false;
				SMP_msg("ERROR: TCFG: Bad JumpFollowNodesMap entry %d , %d ", SourceBlockNum, TargetBlockNum);
				this->DumpFuncNameAndAddr();
			}
		}
	}

	return Sane;
} // end of SMPFunction::AuditTCFG()

// Get unique successor to BlocksSeen; return false if unable to do so.
bool SMPFunction::FindNextBlockAfterReachableBlocks(const int ContainingLoopNum, const STARSBitSet &BlocksSeen, int &NextBlockNum) {
	// If we don't encounter any loops, we just go to the end of the BlocksSeen trace and
	//  get the unique successor to the last block. If we see a loop along the way, we 
	//  look at the exit targets. All exit targets must be in the BlocksSeen trace. If
	//  the LoopFollowNode is also in the trace, continue with it and skip over the blocks
	//  within the loop. If not, then the LoopFollowNode is the NextBlockNum we are searching for.
	NextBlockNum = SMP_BLOCKNUM_UNINIT;
	int LowestBlockNum = BlocksSeen.FindLowestBitSet();
	int HighestBlockNum = BlocksSeen.FindHighestBitSet();
	int HighestNonLoopBlockNum = 0;
	bool ErrorSeen = (0 > LowestBlockNum);
	int BlockNum = LowestBlockNum;
	bool UsedLoopFollowBlock = false;
	bool ContainingLoopIsMultiExitTargetLoop = (ContainingLoopNum < 0) ? false : this->DoesLoopHaveMultipleExitTargets((size_t)ContainingLoopNum);
	while (!ErrorSeen && (BlockNum <= HighestBlockNum)) {
		size_t BlockIndex = (size_t)BlockNum;
		if (BlocksSeen.GetBit(BlockIndex)) {
			int InnermostLoopNum = this->GetInnermostLoopNum(BlockNum);
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
			if (InnermostLoopNum != ContainingLoopNum) { // Entered into loop within conditional statement
				if (CurrBlock->IsLoopHeaderBlock()) { // jump to follow block
					int LoopNum = this->GetLoopNumFromHeaderBlockNum(BlockNum);
					int LoopFollowBlockNum;
					bool MultiExitTargetLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, LoopFollowBlockNum);
					if (0 > LoopFollowBlockNum) {
						ErrorSeen = true;
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to bad loop follow block for loop %d LoopHeaderBlock %d in conditional ", LoopNum, BlockNum);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						break;
					}
					if (BlocksSeen.GetBit((size_t)LoopFollowBlockNum)) {
						// Loop and its follow node are contained inside the current conditional branch. Continue
						//  with the follow node and beyond.
						BlockNum = LoopFollowBlockNum;
					}
					else {
						// Loop is in current conditional branch, but its follow block is not. If the
						//  code is structured, it will be the case that the loop follow block is also
						//  the follow block for the containing conditional statement, and the SPARK Ada
						//  decompilation will have "end loop;" falling directly into "end if;". The exception
						//  is the MultiExitTargetLoop inner loop case, when there is also a containing loop
						//  where the conditional needs to get a follow node of -2 to make translation stop at 
						//  every path that hits a containing loop boundary, in case the follow node is shared among
						//  more than two structures.
						if (MultiExitTargetLoop && (ContainingLoopNum >= 0)) {
							NextBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
						}
						else {
							NextBlockNum = LoopFollowBlockNum;
						}
						UsedLoopFollowBlock = true;
						break;
					}
				}
				else {
					++BlockNum; // Don't bother with loop blocks other than the header block
				}
			}
			else {
				HighestNonLoopBlockNum = BlockNum;
				++BlockNum;
			}
		}
		else {
			++BlockNum;
		}
	} // end while (BlockNum <= HighestBlockNum)

	if (!ErrorSeen) {
		if (!UsedLoopFollowBlock) {
			if (ContainingLoopIsMultiExitTargetLoop && (SMP_BLOCKNUM_UNINIT == NextBlockNum)) {
				NextBlockNum = SMP_BLOCKNUM_COMMON_RETURN; // Conditional translation must stop at loop boundary.
			}
			else {
				// Still need convergence point.
				SMPBasicBlock *LastBlock = this->GetBlockByNum((size_t)HighestNonLoopBlockNum);
				assert(nullptr != LastBlock);
				if (1 == LastBlock->GetNumSuccessorsMinusBackEdges()) {
					NextBlockNum = LastBlock->GetFirstNonBackEdgeSuccNum();
				}
				else {
					NextBlockNum = HighestNonLoopBlockNum;
				}
			}
		}
		ErrorSeen = (SMP_BLOCKNUM_UNINIT == NextBlockNum);
	}

	return ErrorSeen;
} // end of SMPFunction::FindNextBlockAfterReachableBlocks()

// Track CurrBlockNum until we reach block that BranchHeadBlockNum doesn't dominate, or dead end. Return block num, possibly SMP_BLOCKNUM_COMMON_RETURN.
//  A dead end is a block that returns or that calls a non-returning function.
//  We follow successors and we don't follow loop back edges.
int SMPFunction::TrackConditionalBranchTerminus(int BranchHeadBlockNum, int CurrBlockNum, STARSBitSet &BlocksSeen, int &BlockAlreadySeenCounter) {
	int TerminusBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
	SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t) CurrBlockNum);
	assert(CurrBlock != nullptr);
	bool BranchHeadIsInALoop = this->IsBlockInAnyLoop(BranchHeadBlockNum);
	if (CurrBlock->HasReturn() || CurrBlock->HasNonReturningCall()) {
		TerminusBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
	}
	else {
		list<int> WorkListBlockNums;
		STARS_ea_t LastAddr = CurrBlock->GetLastAddr();
		ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
		bool IsLoopBack = ((LastCFType == LOOP_BACK) || (LastCFType == INVERTED_LOOP_BACK));
		if (IsLoopBack) {
			// Don't follow back edge, but get the other successor, if conditional.
			vector<SMPInstr *>::const_reverse_iterator LastInstIter = CurrBlock->GetRevInstCBegin();
			if (COND_BRANCH == (*LastInstIter)->GetDataFlowType()) {
				assert(CurrBlock->IsLoopTailBlock());
				if (!CurrBlock->IsDoubleLoopTailBlock()) {
					for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc(); SuccIter != CurrBlock->GetLastConstSucc(); ++SuccIter) {
						int SuccBlockNum = (*SuccIter)->GetNumber();
						if (SuccBlockNum != CurrBlock->GetLoopHeaderNumber()) {
							WorkListBlockNums.push_back(SuccBlockNum);
						}
					}
				}
			}
		}
		else if (IsInvertedLoopExitFlow(LastCFType) && this->AreBlocksInSameLoops(BranchHeadBlockNum, CurrBlockNum)) {
			// COND_BRANCH taken would remain in the loop; fall-through will exit the loop.
#if 0	// Could do INVERTED_LOOP_EXIT by falling out of any block in the loop, not just the header block.
			assert(CurrBlock->IsLoopHeaderBlock());
#endif
#if 0
			list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFallThroughSucc();
			assert(SuccIter != CurrBlock->GetLastConstSucc());
			int SuccBlockNum = (*SuccIter)->GetNumber();
			WorkListBlockNums.push_back(SuccBlockNum);
#else
			TerminusBlockNum = CurrBlockNum; // First good candidate, as conditional cannot span loop boundaries.
#endif
		}
		else { // add all successors to the work list
			for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc(); SuccIter != CurrBlock->GetLastConstSucc(); ++SuccIter) {
				int SuccBlockNum = (*SuccIter)->GetNumber();
				if (!BlocksSeen.GetBit((size_t)SuccBlockNum)) { // Don't go backwards
					WorkListBlockNums.push_back(SuccBlockNum);
				}
			}
		}
#define STARS_COND_BRANCH_WORKLIST_LIMIT 200
#define STARS_COND_BRANCH_ALREADY_SEEN_LIMIT 50
		while (!WorkListBlockNums.empty()) { // recurse
			int SuccBlockNum = WorkListBlockNums.back();
			WorkListBlockNums.pop_back();
			bool BlockAlreadyInList = IsIntInList(WorkListBlockNums, SuccBlockNum);
			bool BlockAlreadyProcessed = BlocksSeen.GetBit((size_t) SuccBlockNum);
			if (BlockAlreadyProcessed)
				++BlockAlreadySeenCounter;
#if 0
			if (BlockAlreadyProcessed || BlockAlreadyInList) {
#else
			if (BlockAlreadyInList) {
#endif
				// We are looping around the same blocks, which should 
				//  not happen for a conditional branch. Terminate.
				SMP_msg("ERROR: Looping on branch terminus block numbers: %d and %d for branch head: %d current block: %d Bitset: %d InList: %d ",
					CurrBlockNum, SuccBlockNum, BranchHeadBlockNum, CurrBlockNum, BlockAlreadyProcessed, BlockAlreadyInList);
				this->DumpFuncNameAndAddr();
				TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
				WorkListBlockNums.clear();
				this->SetHasUnstructuredCFG(); // recovery of blocks into if-elsif-elsif-else is not possible
				break;
			}
			else if (STARS_COND_BRANCH_WORKLIST_LIMIT < WorkListBlockNums.size()) {
				SMP_msg("ERROR: WorkList explosion on branch terminus block numbers: %d and %d for branch head: %d current block: %d ",
					CurrBlockNum, SuccBlockNum, BranchHeadBlockNum, CurrBlockNum);
				this->DumpFuncNameAndAddr();
				TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
				WorkListBlockNums.clear();
				this->SetHasUnstructuredCFG(); // recovery of blocks into if-elsif-elsif-else is not possible
				break;
			}
			else if (STARS_COND_BRANCH_ALREADY_SEEN_LIMIT < BlockAlreadySeenCounter) {
				SMP_msg("ERROR: BlockAlreadySeen limit reached on branch terminus block numbers: %d and %d for branch head: %d current block: %d ",
					CurrBlockNum, SuccBlockNum, BranchHeadBlockNum, CurrBlockNum);
				this->DumpFuncNameAndAddr();
				TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
				WorkListBlockNums.clear();
				this->SetHasUnstructuredCFG(); // recovery of blocks into if-elsif-elsif-else is not possible
				break;
			}

			BlocksSeen.SetBit((size_t) SuccBlockNum);
			bool OutsideTheLoop = BranchHeadIsInALoop && (!this->AreBlocksInSameLoops(BranchHeadBlockNum, SuccBlockNum));
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			if (OutsideTheLoop || (!this->DoesBlockDominateBlock(BranchHeadBlockNum, SuccBlockNum))) {
				// We reached a block that is not dominated by BranchHeadBlockNum, or exited the loop.
				//  Check for consistency with previous such blocks.
				if (0 <= TerminusBlockNum) { // had previous terminus
					if (SuccBlockNum != TerminusBlockNum) {
						SMP_msg("ERROR: Inconsistent branch terminus block numbers: %d and %d for branch head: %d current block: %d ",
							TerminusBlockNum, SuccBlockNum, BranchHeadBlockNum, CurrBlockNum);
						this->DumpFuncNameAndAddr();
						TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
						WorkListBlockNums.clear();
					}
				}
				else { // first good terminus
					TerminusBlockNum = SuccBlockNum;
				}
			}
			else { // still in path dominated by BranchHeadBlockNum
				int NewTerminusBlockNum = this->TrackConditionalBranchTerminus(BranchHeadBlockNum, SuccBlockNum, BlocksSeen, BlockAlreadySeenCounter);
				// Perform the meet function over the old and new terminus block numbers.
				if (NewTerminusBlockNum == SMP_BLOCKNUM_UNINIT) {
					SMP_msg("ERROR: Bad branch terminus block number from recursion for branch head: %d current block: %d ",
						BranchHeadBlockNum, CurrBlockNum);
					this->DumpFuncNameAndAddr();
					TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
					WorkListBlockNums.clear();
				}
				else if (TerminusBlockNum == SMP_BLOCKNUM_COMMON_RETURN) {
						TerminusBlockNum = NewTerminusBlockNum; //  new block num takes precedence; might be unchanged
				}
				else if (NewTerminusBlockNum == SMP_BLOCKNUM_COMMON_RETURN) {
					; // old block num takes precedence
				}
				else if (NewTerminusBlockNum != TerminusBlockNum) {
					SMP_msg("ERROR: Inconsistent branch terminus block numbers: %d and %d for branch head: %d current block: %d ",
						TerminusBlockNum, NewTerminusBlockNum, BranchHeadBlockNum, CurrBlockNum);
					this->DumpFuncNameAndAddr();
					TerminusBlockNum = SMP_BLOCKNUM_UNINIT; // error signal
					WorkListBlockNums.clear();
				}
			}
		} // end while work list is not empty
	}

	// Corner case: Escaped domination by head block or escaped loop, and hit return block.
	if (0 <= TerminusBlockNum) {
		SMPBasicBlock *TerminusBlock = this->GetBlockByNum((size_t)TerminusBlockNum);
		if (TerminusBlock->HasReturn() || TerminusBlock->HasNonReturningCall()) {
			TerminusBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
		}
	}

	return TerminusBlockNum;
} // end of SMPFunction::TrackConditionalBranchTerminus()

// Do IF and ELSE branches merge before FollowBlockNum? Do all IF and ELSE blocks reach FollowBlockNum?
bool SMPFunction::FindUnstructuredIfThenElse(const int HeadBlockNum, const int ThenBlockNum, const int ElseBlockNum, const int FollowBlockNum) {
	bool UnstructuredClauses = false;

	if ((SMP_BLOCKNUM_UNINIT != ElseBlockNum) && (SMP_BLOCKNUM_UNINIT != ThenBlockNum)) {
		STARSBitSet ThenBlocksSeen, ElseBlocksSeen;
		int BlockAlreadySeenCounter = 0;
		ThenBlocksSeen.AllocateBits(this->GetNumBlocks());
		ElseBlocksSeen.AllocateBits(this->GetNumBlocks());

		// Follow no back edges, start at ThenBlock, stop at FollowBlockNum, record blocks seen.
		list<int> WorkList;
		ThenBlocksSeen.SetBit((size_t)ThenBlockNum);
		WorkList.push_back(ThenBlockNum);
		while (!WorkList.empty()) { // RPO ordering, no back edges, ensures termination
			int CurrBlockNum = WorkList.front();
			WorkList.pop_front();
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(CurrBlockNum);
			for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc(); SuccIter != CurrBlock->GetLastConstSucc(); ++SuccIter) {
				int SuccBlockNum = (*SuccIter)->GetNumber();
				if (!ThenBlocksSeen.GetBit((size_t)SuccBlockNum)) { // don't follow an unstructured loop
					if (SuccBlockNum > CurrBlockNum) { // no back edge
						if (SuccBlockNum != FollowBlockNum) {
							ThenBlocksSeen.SetBit((size_t)SuccBlockNum);
							WorkList.push_back(SuccBlockNum);
						}
					}
				}
			}
		} // end while WorkList is not empty

		WorkList.push_back(ElseBlockNum);
		ElseBlocksSeen.SetBit((size_t)ElseBlockNum);
		while (!WorkList.empty()) { // RPO ordering, no back edges, ensures termination
			int CurrBlockNum = WorkList.front();
			WorkList.pop_front();
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(CurrBlockNum);
			for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc(); SuccIter != CurrBlock->GetLastConstSucc(); ++SuccIter) {
				int SuccBlockNum = (*SuccIter)->GetNumber();
				if (!ElseBlocksSeen.GetBit((size_t)SuccBlockNum)) { // don't follow an unstructured loop
					if (SuccBlockNum > CurrBlockNum) { // no back edge
						if (SuccBlockNum != FollowBlockNum) {
							ElseBlocksSeen.SetBit((size_t)SuccBlockNum);
							WorkList.push_back(SuccBlockNum);
						}
					}
				}
			}
		} // end while WorkList is not empty

		// A structured if-then-else should have no blocks reached from both THEN and ELSE
		//  clauses, except the FollowBlock, which we avoided marking.
		for (size_t BitIndex = 0; BitIndex < ThenBlocksSeen.GetNumBits(); ++BitIndex) {
			if (ThenBlocksSeen.GetBit(BitIndex) && ElseBlocksSeen.GetBit(BitIndex)) {
				UnstructuredClauses = true;
				if (!this->PrintedSPARKUnstructuredMsg) {
					SMP_msg("ERROR: SPARK: Unstructured due to THEN and ELSE clauses merge prematurely at block %zu for HeadBlock %d FollowBlock %d ",
						BitIndex, HeadBlockNum, FollowBlockNum);
					this->DumpFuncNameAndAddr();
					this->PrintedSPARKUnstructuredMsg = true;
				}
				break;
			}
		}
		if (!UnstructuredClauses && (0 < FollowBlockNum)) {
			// Every block in the THEN clause should reach the follow block. Ditto for every
			//  block in the ELSE clause.
			list<size_t> BlockNumList;
			this->BuildBlockListFromBitset(ThenBlocksSeen, BlockNumList);
			this->BuildBlockListFromBitset(ElseBlocksSeen, BlockNumList);
			UnstructuredClauses = this->FindUnstructuredBlockList(BlockNumList, HeadBlockNum, FollowBlockNum);
		}
	}

	return UnstructuredClauses;
} // end of SMPFunction::FindUnstructuredIfThenElse()

// Do any blocks fail to reach their FollowBlockNum?
bool SMPFunction::FindUnstructuredBlockList(const list<size_t> &BlockList, const int CondHeadBlockNum, const int FollowBlockNum) {
	assert(0 < FollowBlockNum);
	bool Unstructured = false;
	int InnermostContainingLoopNum = this->GetInnermostLoopNum(CondHeadBlockNum);

	for (size_t BlockNum : BlockList) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		bool HitDeadEnd = false;

		// Every block in a loop can make it back to the header block. If the header
		//  block can reach FollowBlockNum, then all blocks in the loop can also reach
		//  FollowBlockNum, but some might have to follow a back edge to the loop header
		//  block in order to do so. We can save time and avoid special-case following of
		//  back edges by only testing the header blocks.
		// NOTE: If the header block is not inside the conditional statement, meaning that the
		//  conditional statement is contained in the loop, then we would ignore all blocks in
		//  the conditional statement and analyze nothing; a mistake. We need to separate internal loops from
		//  containing loops. For internal loops, the header block should reach the conditional
		//  follow node and other loop blocks can be ignored. For containing loops, we do not ignore
		//  any blocks; each must reach the conditional follow node.
		bool LoopDontCareBlock = (this->IsBlockInAnyLoop((int)BlockNum) && (!CurrBlock->IsLoopHeaderBlock()));
		if (LoopDontCareBlock) {
			// Don't ignore blocks from the containing loop.
			int InnermostLoopNum = this->GetInnermostLoopNum((int)BlockNum);
			LoopDontCareBlock = (InnermostContainingLoopNum != InnermostLoopNum);
		}
		if (!LoopDontCareBlock) {
			this->ResetSCCPVisitedBlocks();
			if (!CurrBlock->IsBlockReachableWithoutBackEdges(FollowBlockNum, HitDeadEnd)) {
				if (!HitDeadEnd) {
					Unstructured = true;
					if (Unstructured && (!this->PrintedSPARKUnstructuredMsg)) {
						SMP_msg("ERROR: SPARK: Unstructured due to conditional block %zu unable to reach follow block %d ", BlockNum, FollowBlockNum);
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
						this->PrintedSPARKUnstructuredMsg = true;
					}
					break;
				}
			}
		}
	} // end for all blocks in BlockList

	return Unstructured;
} // end of SMPFunction::FindUnstructuredBlockList()

// Does COND_BRANCH behave as an unstructured multi-level loop continue statement?
bool SMPFunction::FindMultiLoopContinue(const int BranchTCFGBlockNum) {
	assert(0 <= BranchTCFGBlockNum);
	bool Unstructured = false;
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();
	int OrigBlockNum = this->GetOrigBlockNumFromTCFGBlockNum(BranchTCFGBlockNum);
	if (VerboseOutput) {
		SMP_msg("INFO: SPARK: FindMultiLoopContinue() called with BranchTCFGBlockNum %d OrigBlockNum %d ", BranchTCFGBlockNum, OrigBlockNum);
		this->DumpFuncNameAndAddr();
	}
	if (!this->IsBlockInAnyLoop(OrigBlockNum))
		return Unstructured;

	int InnermostLoopNum = this->GetInnermostLoopNum(OrigBlockNum);
	STARSTransCFGBlock *BranchTCFGBlock = this->GetTCFGBlockByNum((size_t)BranchTCFGBlockNum);
	assert(0 <= InnermostLoopNum);

	// Each successor of the COND_BRANCH at the end of BranchBlock
	//  must be in one of the following three categories:
	//  A) Inside the innermost loop containing BranchBlock.
	//  B) An exit target block for the innermost loop containing BranchBlock.
	//  C) The header block for an outer loop that has the same (single) follow block as the
	//      innermost loop.
	//
	// The SPARK Ada translations of these cases are straightforward:
	//  A) if-then-else within the innermost loop.
	//  B) exit when (condition) or exit when (inverted condition).
	//  C) if (inverted condition) then ... endif; endloop; endloop;
	//
	// Note that category C precludes having any code inside the endif-endloop-endloop series.
	//  This condition is guaranteed by having the same follow nodes; a follow node comes
	//  right after an endloop, so if an inner loop has the same follow node as an outer loop,
	//  their endloop statements are consecutive.
	//
	// An example of a COND_BRANCH inside a loop that is not in categories A or B but is within
	//  category C is seen at the end of basic block #5 in the func at 0x4031b0 in bzip2.psexe.
	// An example that fails all three category tests, and is thus an unstructured multi-level
	//  loop continue statement, is seen at basic block #8 in the func at 0x437bf3 in busybox.psexe.

	for (STARSTransCFGBlock *SuccBlock : BranchTCFGBlock->GetConstSuccList()) {
		int SuccBlockNum = SuccBlock->GetNumber();
		int OrigSuccBlockNum = this->GetOrigBlockNumFromTCFGBlockNum(SuccBlockNum);
		if (SuccBlock->IsOnlyInternalDirectJump()) {
			// Follow the jump to the real successor.
			SuccBlock = (*(SuccBlock->GetFirstConstSucc()));
			SuccBlockNum = SuccBlock->GetNumber();
			OrigSuccBlockNum = this->GetOrigBlockNumFromTCFGBlockNum(SuccBlockNum);
		}
		if (!this->IsTCFGBlockInLoop(SuccBlockNum, (size_t)InnermostLoopNum)) {
			// Failed category A test.
			if (!this->IsBlockLoopExitTarget((size_t)InnermostLoopNum, OrigSuccBlockNum)) {
				// Failed category B test. Needs to be header block for outer loop.
				int InnerLoopFollowBlockNum = SMP_BLOCKNUM_UNINIT;
				int OuterLoopFollowBlockNum = SMP_BLOCKNUM_UNINIT;
				if (SuccBlockNum < BranchTCFGBlockNum) { // back edge; now, to which loop header?
					bool MultiExitInnerLoop = this->GetLoopFollowBlockNum((size_t)InnermostLoopNum, InnerLoopFollowBlockNum);
					if (SuccBlock->IsLoopHeaderBlock()) {
						int OuterLoopNum = this->GetLoopNumFromHeaderBlockNum(OrigSuccBlockNum);
						assert(0 <= OuterLoopNum);
						bool MultiExitOuterLoop = this->GetLoopFollowBlockNum((size_t)OuterLoopNum, OuterLoopFollowBlockNum);
						if (MultiExitInnerLoop || MultiExitOuterLoop) {
							Unstructured = true;
						}
					}
					if (OuterLoopFollowBlockNum != InnerLoopFollowBlockNum) {
						// Failed category C test.
						Unstructured = true;
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to multi-level loop continue statement at TCFG block %d ", BranchTCFGBlockNum);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						break;
					}
				}
				else { //How could SuccBlock not be in the loop and not be an exit target block for the loop and not a back edge?
					SMP_msg("ERROR: SPARK: SuccBlockNum %d for BranchTCFGBlockNum %d uncategorized ", SuccBlockNum, BranchTCFGBlockNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
					this->Dump(true);
				}
				if (!Unstructured) {
					SMP_msg("INFO: SPARK: Passed Category C test for multi-loop continue at BranchTCFGBlockNum %d ", BranchTCFGBlockNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
			}
		}
	} // end for all successors to BranchTCFGBlock

	return Unstructured;
} // end of SMPFunction::FindMultiLoopContinue()

// Detect conditional statement structuring errors that are obvious just from the reachability bitsets.
bool SMPFunction::DetectConditionalReachabilityErrors(const int HeadBlockNum, const int FTBlockNum, const int NFTBlockNum, const bool FTNotDominated, const bool NFTNotDominated, const int NextFTBlockNum, const int NextNFTBlockNum, const STARSBitSet &FTBlocksSeen, const STARSBitSet &NFTBlocksSeen, const STARSBitSet &CommonBlocksSeen) {
	bool Unstructured = false;
	bool IfThenElseCaseDom = (!(FTNotDominated || NFTNotDominated)); // no dominance problem; might be if-then-else or if-then

	// All blocks reachable by either FTBlock or NFTBlock should either precede the intersection point or
	//  should be in CommonBlocksSeen (i.e. RPO block number greater than intersection point RPO number
	//  should imply that we are merely continuing downward from the intersection point). This is only true
	//  for the IfThenElse case, not for an IfThen.
	int IntersectionBlockNum = CommonBlocksSeen.FindLowestBitSet();
	bool TracesIntersect = (0 < IntersectionBlockNum);
	if (TracesIntersect) {
		bool IfThenCase = (IntersectionBlockNum == NFTBlockNum);
		bool OddIfThenCase = (IntersectionBlockNum == FTBlockNum);
		bool IfThenElseCase = (!(IfThenCase || OddIfThenCase));
		if (IfThenElseCase) {
			assert(IfThenElseCaseDom);
			for (size_t BlockNum = (size_t)(IntersectionBlockNum + 1); BlockNum < this->GetNumBlocks(); ++BlockNum) {
				if (FTBlocksSeen.GetBit(BlockNum)) {
					bool success = NFTBlocksSeen.GetBit(BlockNum);
					if (!success) {
						Unstructured = true;
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to non-common-reached block %d after HeadBlockNum %d and intersection block %d ", 
								BlockNum, HeadBlockNum, IntersectionBlockNum);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						break;
					}
				}
			}
		}
	}
	if (0 < this->GetNumLoops()) {
		if (!Unstructured) {
			// Ensure that any loop contained inside the FT clause has its exit target
			//  blocks dominated by the loop head block; a "back door" into a loop exit target block
			//  that does not come from the loop must be restructured.
			int LowestBlockSeen = FTBlocksSeen.FindLowestBitSet();
			int HighestBlockSeen = FTBlocksSeen.FindHighestBitSet();
			int LimitBlockNum = TracesIntersect ? IntersectionBlockNum : HighestBlockSeen;
			if (0 <= LowestBlockSeen) {
				for (int BlockNum = LowestBlockSeen; BlockNum <= LimitBlockNum; ++BlockNum) {
					if (FTBlocksSeen.GetBit((size_t)BlockNum)) {
						if (this->ConditionalReachabilityLoopErrorHelper(HeadBlockNum, BlockNum)) {
							Unstructured = true;
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to exit targets not all dominated by loop head block in FT clause starting at block %d ", FTBlockNum);
								this->DumpFuncNameAndAddr();
								this->DumpDotCFG();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							break;
						}
					}
				}
			}
		}
		if (!Unstructured) {
			// Ensure that any loop contained inside the NFT clause has its exit target
			//  blocks dominated by the loop head block; a "back door" into a loop exit target block
			//  that does not come from the loop must be restructured.
			int LowestBlockSeen = NFTBlocksSeen.FindLowestBitSet();
			int HighestBlockSeen = NFTBlocksSeen.FindHighestBitSet();
			int LimitBlockNum = TracesIntersect ? IntersectionBlockNum : HighestBlockSeen;
			if (0 <= LowestBlockSeen) {
				for (int BlockNum = LowestBlockSeen; BlockNum <= LimitBlockNum; ++BlockNum) {
					if (NFTBlocksSeen.GetBit((size_t)BlockNum)) {
						if (this->ConditionalReachabilityLoopErrorHelper(HeadBlockNum, BlockNum)) {
							Unstructured = true;
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to exit targets not all dominated by loop head block in NFT clause starting at block %d ", NFTBlockNum);
								this->DumpFuncNameAndAddr();
								this->DumpDotCFG();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							break;
						}
					}
				}
			}
		}
	} // end if we have loops

	// In the If-Then case, the smaller RPO block # out of FTBlockNum and NFTBlockNum begins the THEN branch,
	//  and the larger RPO block # must be the FollowBlockNum. Every block on the path beginning at the THEN
	//  branch and concluding with a dead end block or the FollowBlock (whichever is encountered first)
	//  should be present in the corresponding reachable bitset. If any block is not present in that bitset, it is
	//  because that block is not dominated by the HeadBlockNum, which caused MarkReachableBlocks to terminate its
	//  trace. That means that the THEN branch is being entered from outside the if-then statement, making the 
	//  CFG unstructured.
	if (!IfThenElseCaseDom && (!Unstructured)) {
		bool OddIfThenCase = FTNotDominated; // FTBlock is FollowNode
		int LowestBlockNum = OddIfThenCase ? NFTBlockNum : FTBlockNum;
		int HighestBlockNum = OddIfThenCase ? FTBlockNum : NFTBlockNum;
		list<int> WorkList;
		WorkList.push_front(LowestBlockNum);
		STARSBitSet WorkListBlocksSeen;
		WorkListBlocksSeen.AllocateBits(this->GetNumBlocks());
		while (!WorkList.empty()) {
			int BlockNum = WorkList.front();
			WorkList.pop_front();
			size_t BlockIndex = (size_t)BlockNum;
			if (WorkListBlocksSeen.GetBit(BlockIndex)) { // already processed (we see join points more than once)
				continue;
			}
			else {
				WorkListBlocksSeen.SetBit(BlockIndex);
			}
			bool BlockReached = OddIfThenCase ? NFTBlocksSeen.GetBit(BlockIndex) : FTBlocksSeen.GetBit(BlockIndex);
			if (BlockReached) {
				SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
				// Add non-back-edge successors to the WorkList, limited by HighestBlockNum.
				for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc();
					SuccIter != CurrBlock->GetLastConstSucc();
					++SuccIter) {
					int SuccBlockNum = (*SuccIter)->GetNumber();
					// Only successors between the current block and the follow node go onto the work list.
					if ((SuccBlockNum > BlockNum) && (SuccBlockNum < HighestBlockNum)) {
						WorkList.push_back(SuccBlockNum);
					}
				} // end for all successor blocks
			}
			else {
				Unstructured = true;
				if (!this->PrintedSPARKUnstructuredMsg) {
					SMP_msg("ERROR: SPARK: Unstructured due to then-branch block %d not dominated and reached from head block %d ",
						BlockNum, HeadBlockNum);
					this->DumpFuncNameAndAddr();
					this->PrintedSPARKUnstructuredMsg = true;
				}
				break; // break out of while loop
			}
		} // end while WorkList is not empty
	}

	if (!Unstructured) {
		// One last possible problem: A bypass of the intersection point:
		//         [block 1]
		//  |----------+------------|
		//  |                       |
		// THEN block            ELSE block
		//  |                       |
		//  |----------+------------|------|
		//             |                   |
		//        [int. block]         [block 5]
		//             |                   |
		//             |                   |
		//        [block 6]----------------|
		//
		// It is possible to bypass the intersection block from the ELSE clause
		//  by going to block 5 and then on to block 6. So the intersection block
		//  can no longer be used as a structured conditional follow node.
		//
		// We can detect this case by finding the downward-reachable blocks from
		//  the intersection block, imposing no dominance tests, and seeing if
		//  any downwards (no back edges) paths from the top of the ELSE clause
		//  reach one of those blocks without passing through the intersection block.
		//
		// The logic above works equally well for an IF-THEN in which a path through
		//  the THEN clause bypasses the follow node. The conditional need not be an
		//  IF-THEN-ELSE.
		//
		// The complication we must deal with is a follow node that is not the intersection
		//  block, because the intersection block would have been not dominated by the
		//  conditional head block, so it is not present in the reachable bitsets, e.g.
		//  the conditional follow node is also a follow node for an enclosing statement.
		//  In that case, we need an artificial intersection block, i.e. the block that
		//  would have been the intersection block if not for the dominance test in
		//  MarkReachableBlocks().
		int CurrIntersectionBlockNum = IntersectionBlockNum;
		if (!TracesIntersect) {
			if (NextFTBlockNum == NextNFTBlockNum) {
				CurrIntersectionBlockNum = NextFTBlockNum; // artificial intersection block; about to happen before dominance cutoff
			}
		}
		if (0 < CurrIntersectionBlockNum) {
			STARSBitSet CurrReachableBlocks;
			CurrReachableBlocks.AllocateBits(this->RPOBlocks.size());
			int LimitBlockNum = (int)(this->GetNumBlocks() + 1);
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			this->MarkReachableBlocks(CurrIntersectionBlockNum, -1, HeadBlockNum, false, LimitBlockNum, CurrReachableBlocks);
			if (CurrReachableBlocks.IsAnyBitSet()) {
				int BlockLimit = CurrReachableBlocks.FindHighestBitSet();
				if (!FTNotDominated) {
					Unstructured = this->CanBlockBypassBlock(FTBlockNum, CurrIntersectionBlockNum, BlockLimit, CurrReachableBlocks);
					if (Unstructured && (!this->PrintedSPARKUnstructuredMsg)) {
						SMP_msg("ERROR: SPARK: Unstructured due to bypass of Intersection block %d from FTBlock %d ", IntersectionBlockNum, FTBlockNum);
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
						this->PrintedSPARKUnstructuredMsg = true;
					}
				}
				if (!NFTNotDominated && (!Unstructured)) {
					Unstructured = this->CanBlockBypassBlock(NFTBlockNum, CurrIntersectionBlockNum, BlockLimit, CurrReachableBlocks);
					if (Unstructured && (!this->PrintedSPARKUnstructuredMsg)) {
						SMP_msg("ERROR: SPARK: Unstructured due to bypass of Intersection block %d from NFTBlock %d ", IntersectionBlockNum, NFTBlockNum);
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
						this->PrintedSPARKUnstructuredMsg = true;
					}
				}
			}
		}
	}

	return Unstructured;
} // end of SMPFunction::DetectConditionalReachabilityErrors()

// Back door edge into loop exit targets (for multi-exit loops)?
bool SMPFunction::ConditionalReachabilityLoopErrorHelper(const int CondHeadBlockNum, const int BlockNum) {
	bool Unstructured = false;

	SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t)BlockNum);
	if (CurrBlock->IsLoopHeaderBlock()) {
		int InnerLoopNum = this->GetLoopNumFromHeaderBlockNum(BlockNum);
		assert(0 <= InnerLoopNum);
		bool MultiExitTargetInnerLoop = this->DoesLoopHaveMultipleExitTargets(InnerLoopNum);
		for (int ExitTargetBlockNum : this->LoopExitTargets[(size_t)InnerLoopNum]) {
			// If we have multiple exit targets, then they should all be contained
			//  within the ELSE-branch, with their common successor being the loop
			//  follow node. This loop follow node can be a common follow node for enclosing
			//  statements, including the current conditional statement and outer 
			//  conditional statements, in which case the inner HeadBlock does not
			//  dominate it. However, with multiple exit targets, the exit targets
			//  themselves must be dominated by the conditional HeadBlockNum in order
			//  to be a structured CFG.
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			if (MultiExitTargetInnerLoop && (!this->DoesBlockDominateBlock(CondHeadBlockNum, ExitTargetBlockNum))) {
				Unstructured = true;
				break;
			}
		}
	}

	return Unstructured;
} // end of SMPFunction::ConditionalReachabilityLoopErrorHelper()

// Find guarded loops and fill the GuardToLoopMap and LoopToGuardMap
void SMPFunction::FindGuardedLoops(void) {
	assert(this->HasStructuredControlFlow());
	bool CloningHasOccurred = (this->TCFGBlocks.size() > this->RPOBlocks.size());
	// We look for the following code structure:
	// if (cond) then
	//   [straight-line code]
	//   loop
	//     [loop body]
	//   end loop;
	//   [any code]
	// end if;
	//
	// The key is detecting the sequence of a BRANCH_IF_THEN, then straight-line code, then a loop header block.
	//  When this sequence is detected, we map the first address in the loop header block to the BRANCH_IF_THEN
	//  address, and vice versa, in the two maps.
	// NOTE: We have already found the follow block for the if-then statement, and it needs to NOT be the loop
	//  header block, else we have "end if;" before the loop starts.
	for (size_t BlockIndex = 0; BlockIndex < this->TCFGBlocks.size(); ++BlockIndex) {
		STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum(BlockIndex);
		if (CurrTCFGBlock->HasConditionalBranch()) {
			STARS_ea_t LastInstAddr = CurrTCFGBlock->GetLastAddr();
			assert(STARS_BADADDR != LastInstAddr);
			ControlFlowType LastCFType = this->GetControlFlowType(LastInstAddr);
			if (LastCFType == BRANCH_IF_THEN) {
				SMPInstr *BranchInst = this->GetInstFromAddr(LastInstAddr);
				bool OddIfThenCase = BranchInst->IsOddIfThenCase(); // jumps to THEN case instead of falling through to it
				list<STARSTransCFGBlock *>::const_iterator FallThroughBlockIter = OddIfThenCase ? CurrTCFGBlock->GetCondNonFallThroughSucc() : CurrTCFGBlock->GetFallThroughSucc();
				assert(FallThroughBlockIter != CurrTCFGBlock->GetLastConstSucc());
				bool FoundGuardedLoop = false;
				STARS_ea_t LoopAddr = STARS_BADADDR;
				STARSTransCFGBlock *FallThroughBlock = (*FallThroughBlockIter);
				if (FallThroughBlock->IsLoopHeaderBlock()) {
					// The simplest case: no [straight-line code] between COND_BRANCH and loop header.
					FoundGuardedLoop = true;
					LoopAddr = FallThroughBlock->GetFirstNonMarkerAddr();
				}
				else {
					// FallThroughBlock might be a block that only falls through to the loop header block,
					//  the slightly less simple case shown as [straight-line code] in the comments above.
					if (1 == FallThroughBlock->GetNumSuccs()) {
						STARS_ea_t LastFallThroughBlockAddr = FallThroughBlock->GetLastAddr();
						ControlFlowType FTLastCFType = this->GetControlFlowType(LastFallThroughBlockAddr);
						if (FALL_THROUGH == FTLastCFType) {
							// Just falling through to next block. Is it a loop header block?
							list<STARSTransCFGBlock *>::const_iterator NextBlockIter = FallThroughBlock->GetFallThroughSucc();
							if (NextBlockIter == FallThroughBlock->GetLastConstSucc()) {
								// Probably special case of direct jump that acts as fall-through control flow.
								SMPInstr *LastInst = this->GetInstFromAddr(LastFallThroughBlockAddr);
								if (LastInst->IsCondTailCall()) {
									// Special case of only one successor; not a guarded loop.
									continue; // Go on to the next block.
								}
								if (JUMP != LastInst->GetDataFlowType()) {
									// Don't know what is going on, but it is not a guarded loop.
									SMP_msg("ERROR: FindGuardedLoop(): LastInst flow type not JUMP ");
									this->DumpFuncNameAndAddr();
									SMP_msg("ERROR: FindGuardedLoop(): LastInst flow type is ");
									DumpDataFlowType(LastInst->GetDataFlowType());
									continue; // Go on to the next block.
								}
								NextBlockIter = FallThroughBlock->GetFirstConstSucc();
								assert(NextBlockIter != FallThroughBlock->GetLastConstSucc());
							}
							STARSTransCFGBlock *NextBlock = (*NextBlockIter);
							if (NextBlock->IsLoopHeaderBlock()) {
								// The less simple case: some [straight-line code] between COND_BRANCH and loop header.

								// To prevent being deceived by the case in which we fall through 
								//  to the tail block, which the compiler has placed so that the tail block falls through
								//  to the loop header block, we require that the FallThroughBlock must dominate the loop
								//  header block. Simple, fast screening before concluding.
								this->TCFGDomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
								bool CorrectDomTreeStructure = this->DoesTCFGBlockDominateTCFGBlock((int)BlockIndex, NextBlock->GetNumber());
								if (!CorrectDomTreeStructure)
									continue; // move on to next block
								// Now, we ensure that we don't have:
								//  if (cond) then
								//     fall-through block
								//  end if;
								//  loop
								//     loop body
								//  end loop;
								//
								// We don't allow the follow block of the if-then to be the loop header block.
								map<int, int>::iterator MapIter = this->JumpFollowNodesMap.find((int)BlockIndex);
								assert(MapIter != this->JumpFollowNodesMap.end());
								int FollowTCFGBlockNum = MapIter->second;
								if (FollowTCFGBlockNum > NextBlock->GetNumber()) { // Guarded loop has nice convergence of end-if after end-loop.
									FoundGuardedLoop = true;
									LoopAddr = NextBlock->GetFirstNonMarkerAddr();
								}
							}
						}
					}
				}
				if (FoundGuardedLoop) {
					assert(STARS_BADADDR != LoopAddr);
					assert(STARS_BADADDR != LastInstAddr);
					pair<STARS_ea_t, STARS_ea_t> GuardToLoopValue(LastInstAddr, LoopAddr);
					pair<map<STARS_ea_t, STARS_ea_t>::iterator, bool> InsertResult;
					pair<STARS_ea_t, STARS_ea_t> LoopToGuardValue(LoopAddr, LastInstAddr);
					InsertResult = this->GuardToLoopMap.insert(GuardToLoopValue);
					assert(InsertResult.second || CloningHasOccurred); // should never visit COND_BRANCH or loop address twice w/o cloning
					InsertResult = this->LoopToGuardMap.insert(LoopToGuardValue);
					assert(InsertResult.second || CloningHasOccurred); // should never visit COND_BRANCH or loop address twice w/o cloning
					SMP_msg("INFO: SPARK: Found guarded loop at %llx TCFGBlockIndex %zu with guard at %llx\n", (uint64_t) LoopAddr, BlockIndex, (uint64_t) LastInstAddr);
				}
			}
		} // end if block has conditional branch
	} // end for all basic blocks

	return;
} // end of SMPFunction::FindGuardedLoops()

// Are we already translating a SPARK_LOOP when we encounter another loop?
bool SMPFunction::IsSPARKLoopInTranslationStack(void) const {
	list<SPARKTranslationCFType>::const_reverse_iterator CFIter;
	bool FoundLoop = false;
	for (CFIter = this->SPARKControlStack.crbegin(); CFIter != this->SPARKControlStack.crend(); ++CFIter) {
		if (SPARK_LOOP == (*CFIter)) {
			FoundLoop = true;
			break;
		}
	}
	return FoundLoop;
} // end of SMPFunction::IsSPARKLoopInTranslationStack()

// Remove duplicate elements in this->SPARKLoopWorkList.
void SMPFunction::CleanUpSPARKLoopWorkList(void) {
	STARSBitSet LoopsSeen;
	LoopsSeen.AllocateBits(this->GetNumLoops());
	list<pair<int, pair<int, int> > >::iterator WorkIter = this->SPARKLoopWorkList.begin();
	while (WorkIter != this->SPARKLoopWorkList.end()) {
		int LoopNum = (*WorkIter).first;
		assert(0 <= LoopNum);
		int HeadBlockNum = WorkIter->second.first;
		assert(0 <= HeadBlockNum);
		if (LoopsSeen.GetBit((size_t)LoopNum)) {
			WorkIter = this->SPARKLoopWorkList.erase(WorkIter);
			SMP_msg("SPARK: WARNING: LoopWorkList duplicate removed for loop %d in %s\n", LoopNum, this->GetFuncName());
		}
		else {
			LoopsSeen.SetBit((size_t) LoopNum);
			++WorkIter;
		}
	}
	return;
} // end of SMPFunction::CleanUpSPARKLoopWorkList()

// Find memory writes (DEFs) with possible aliases
void SMPFunction::AliasAnalysis2(void) {
	// First task: Mark which memory DEFs MIGHT be aliased because an
	//  indirect memory write occurs somewhere in the DEF-USE chain.
	//  Memory DEF-USE chains with no possible aliasing can be subjected
	//  to type inference and type-based optimizing annotations, e.g. a
	//  register spill to memory followed by retrieval from spill memory
	//  followed by NUMERIC USEs should be typed as a continuous NUMERIC
	//  chain if there is no possibility of aliasing.

	// Preparatory step: For each indirect write, mark the block containing the
	//  write instruction. If there are no indirect writes in the function,
	//  leave all DEFs marked as unaliased and exit.

	for (list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		bool CurrBlockHasIndirectWrite = false;
		SMPBasicBlock *CurrBlock = (*BlockIter);
		for (vector<SMPInstr *>::iterator BlkInstIter = CurrBlock->GetFirstInst(); BlkInstIter != CurrBlock->GetLastInst(); ++BlkInstIter) {
			SMPInstr *CurrInst = (*BlkInstIter);
			if (CurrInst->HasIndirectMemoryWrite()) {
				this->HasIndirectWrites = true;
				// Until we get true aliasing analysis, any indirect write
				//  is classified as may-be-aliased.
				if (!CurrBlockHasIndirectWrite) { // first one
					++STARS_IndirectMemWriteBlockCount;
					CurrBlockHasIndirectWrite = true;
				}
				CurrBlock->SetMaybeAliased(true);
				if (CurrInst->DetectUnsafeMemWrite()) {
					this->HasUnsafeIndirectWrites = true;
					CurrInst->SetUnsafeMemWrite();
					++STARS_UnsafeIndirectMemWriteCount;
				}
				else {
					++STARS_SafeIndirectMemWriteCount;
					SMP_msg("INFO: Found safe indirect write in AliasAnalysis at %p\n", CurrInst->GetAddr());
				}
			}
		} // end for all insts in block
	} // end for all blocks in function

	if (!(this->HasIndirectWrites))
		return;

	// Algorithm: Maintain a set of live memory DEFs and their addresses. All LiveIn
	//  memory DEFs start out in the set, with the live-in pseudo-address. For each inst,
	//  if a memory DEF is redefined, replace the memory DEF in the tracking set with
	//  the current memory DEF. If current inst has an indirect write, set the MaybeAliased
	//  flag for each DEF in the tracking set and remove them from the set. Add the current
	//  memory DEF as the new (lone) DEF in the tracking set.
	// Do all of the above starting at the first RPO basic block and follow into successors,
	//  marking each block as processed on exit and not processing any block twice. When exiting
	//  a block, remove local memory DEFs from the tracking set.
	this->ResetProcessedBlocks();

	// Iterate through blocks starting with block 0.
	for (size_t BlockNum = 0; BlockNum < this->GetNumBlocks(); ++BlockNum) {
		this->ChainAliasHelper(BlockNum);
	}

	return;
} // end of SMPFunction::AliasAnalysis2()

// recursive helper for AliasAnalysis2()
void SMPFunction::ChainAliasHelper(size_t BlockNum) {
	SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);

	if (CurrBlock->HasIndirectMemWrite()) {
		STARSDefinitionSet MemTrackingSet;
		// The MemTrackingSet is only relevant for blocks that have indirect writes.
		//  It should hold live memory operands only.
		// Add the other LiveIn memory operands to the tracking set.
		STARS_ea_t BlockPseudoAddr = STARS_MakeBlockNumPseudoID(BlockNum);
		for (STARSOpndSetIter LiveInIter = CurrBlock->GetFirstLiveIn(); LiveInIter != CurrBlock->GetLastLiveIn(); ++LiveInIter) {
			STARSOpndTypePtr LiveInOp = (*LiveInIter);
			if (LiveInOp->IsMemOp()) {
				// If it is a PhiOp, we use the BlockPseudoAddr. Otherwise, it
				//  should only be LiveOut from one predecessor block, and we
				//  trace it to its DefAddr.
				if (CurrBlock->FindPhi(LiveInOp) != CurrBlock->GetLastPhi()) {
					pair<STARSOpndTypePtr, STARS_ea_t> InsertPair(LiveInOp, BlockPseudoAddr);
					MemTrackingSet.insert(InsertPair);
				}
				else {
					bool AlreadyMarkedMaybeAliased = false;
					STARS_ea_t LiveInDefAddr = CurrBlock->TraceLiveInOpToDefAddr(LiveInOp, AlreadyMarkedMaybeAliased);
					if (STARS_BADADDR == LiveInDefAddr) {
						SMP_msg("FATAL ERROR: ChainAliasHelper bad LiveInDefAddr in block %zu ", BlockNum);
						this->DumpFuncNameAndAddr();
						SMP_msg("Operand: ");
						PrintOperand(LiveInOp);
						SMP_msg("\n");
						this->Dump(false);
						assert(STARS_BADADDR != LiveInDefAddr); // kablooey!
					}
					if (!AlreadyMarkedMaybeAliased) {
						pair<STARSOpndTypePtr, STARS_ea_t> InsertPair(LiveInOp, LiveInDefAddr);
						MemTrackingSet.insert(InsertPair);
					}
				}
			}
		} // end for all LiveIn operands

		// Iterate through instructions for blocks with indirect memory writes. 
		if (!MemTrackingSet.empty()) {
			for (vector<SMPInstr *>::iterator BlkInstIter = CurrBlock->GetFirstInst(); BlkInstIter != CurrBlock->GetLastInst(); ++BlkInstIter) {
				SMPInstr *CurrInst = (*BlkInstIter);
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				if (CurrInst->HasIndirectMemoryWrite()) {
					// Mark the DEF for each operand in the MemTrackingSet as MaybeAliased,
					//  then remove it from the tracking set.
					STARSDefinitionSet::iterator TrackingIter = MemTrackingSet.begin();
					while (TrackingIter != MemTrackingSet.end()) {
						STARSOpndTypePtr LocalOp = (*TrackingIter).first;
						STARS_ea_t DefAddr = (*TrackingIter).second;
						if (STARS_IsBlockNumPseudoID(DefAddr)) { // Phi DEF
							// We need to mark the DEFs for each Phi USE as MaybeAliased.
							size_t DefBlockNum = STARS_GetBlockNumFromPseudoID(DefAddr);
							SMPBasicBlock *DefBlock = this->GetBlockByNum(DefBlockNum);
							assert(nullptr != DefBlock);
							this->ResetSCCPVisitedBlocks(); // overloaded use for MarkAllPhiUsesAliased()
							DefBlock->MarkAllPhiUsesAliased(LocalOp);
						}
						else {
							SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
							// Make sure that the local operand is USEd after the indirect write,
							//  else its live range does not overlap and there is no alias.
							STARSDefUseIter DefIter = DefInst->FindDef(LocalOp);
							assert(DefIter != DefInst->GetLastDef());
							bool DeadMemOp = CurrBlock->IsMemOpDead(LocalOp, BlkInstIter, DefIter->GetSSANum());
							if (!DeadMemOp) {
								STARSDefUseIter DefIter = DefInst->SetDefIndWrite(LocalOp, true);
							}
						}
						++TrackingIter;
					} // end while TrackingIter is not at end of MemTrackingSet
					MemTrackingSet.clear();

					// Now, initialize the tracking set with just this latest memory operand.
					STARSOpndTypePtr DefOp = CurrInst->GetMemDef();
					STARSDefinition InsertPair(DefOp, InstAddr);
					pair<STARSDefinitionSet::iterator, bool> InsertResult = MemTrackingSet.insert(InsertPair);
					assert(InsertResult.second);
				}
				else if (CurrInst->HasDestMemoryOperand()) {
					// Replace re-DEFed operands in MemTrackingSet, or add new operand.
					STARSOpndTypePtr DefOp = CurrInst->GetMemDef();
					STARSDefinitionSet::iterator TrackingIter = MemTrackingSet.begin();
					while (TrackingIter != MemTrackingSet.end()) {
						STARSOpndTypePtr LocalOp = (*TrackingIter).first;
						if (IsEqOpIgnoreBitwidth(DefOp, LocalOp)) {
							TrackingIter = MemTrackingSet.erase(TrackingIter);
							STARSDefinition InsertPair(DefOp, InstAddr);
							pair<STARSDefinitionSet::iterator, bool> InsertResult = MemTrackingSet.insert(InsertPair);
							// assert(InsertResult.second);
							TrackingIter = InsertResult.first;
							// It is conceivable that two different DEFs of a memory operand, at
							//  two different addresses, are being re_DEFed, so we continue the loop.
						}
						++TrackingIter;
					} // end while TrackingIter is not at end of MemTrackingSet
				}
			} // end for all insts in block
		}
	}

	return;
} // end of SMPFunction::ChainAliasHelper()

// Find memory writes (DEFs) with possible aliases
void SMPFunction::AliasAnalysis(void) {
	// First task: Mark which memory DEFs MIGHT be aliased because an
	//  indirect memory write occurs somewhere in the DEF-USE chain.
	//  Memory DEF-USE chains with no possible aliasing can be subjected
	//  to type inference and type-based optimizing annotations, e.g. a
	//  register spill to memory followed by retrieval from spill memory
	//  followed by NUMERIC USEs should be typed as a continuous NUMERIC
	//  chain if there is no possibility of aliasing.

	// Preparatory step: For each indirect write, mark all def-use chains
	//  (maintained at the basic block level) that include the indirect
	//  write instruction. If there are no indirect writes in the function,
	//  leave all DEFs marked as unaliased and exit.

	for (list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		for (vector<SMPInstr *>::iterator BlkInstIter = CurrBlock->GetFirstInst(); BlkInstIter != CurrBlock->GetLastInst(); ++BlkInstIter) {
			SMPInstr *CurrInst = (*BlkInstIter);
			if (CurrInst->HasIndirectMemoryWrite()) {
				this->HasIndirectWrites = true;
				// Until we get true aliasing analysis, any indirect write
				//  is classified as may-be-aliased.
				CurrBlock->SetMaybeAliased(true);
				if (CurrInst->DetectUnsafeMemWrite()) {
					this->HasUnsafeIndirectWrites = true;
					CurrInst->SetUnsafeMemWrite();
					++STARS_UnsafeIndirectMemWriteCount;
				}
				else {
					++STARS_SafeIndirectMemWriteCount;
					SMP_msg("INFO: Found safe indirect write in AliasAnalysis at %p\n", CurrInst->GetAddr());
				}
			}
		} // end for all insts in block
	} // end for all blocks in function

#if 1
	if (!(this->HasIndirectWrites))
		return;
#endif
	// Step one: Find only the memory DEFs to start with.
	bool FoundIndWrite = false;
	for (list<SMPInstr *>::iterator InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t InstAddr = CurrInst->GetAddr();
		if (CurrInst->HasDestMemoryOperand()) {
			// Starting with the DEF instruction, traverse the control flow
			//  until we run into (A) the re-definition of the operand, including
			//  a re-definition of any of its addressing registers, or (B) an
			//  indirect write. Return false if condition A terminates the
			//  search, and true if condition B terminates the search.
			this->ResetProcessedBlocks();
			STARSOpndTypePtr MemDefOp = CurrInst->MDGetMemDefOp();
			assert(! MemDefOp->IsVoidOp());
			set<DefOrUse, LessDefUse>::iterator CurrMemDef = CurrInst->FindDef(MemDefOp);
			assert(CurrMemDef != CurrInst->GetLastDef());
			int SSANum = CurrMemDef->GetSSANum();
			FoundIndWrite = this->FindPossibleChainAlias(CurrInst, MemDefOp, SSANum);
			if (FoundIndWrite) {
				// Mark the DEF as aliased.
				CurrMemDef = CurrInst->SetDefIndWrite(CurrMemDef->GetOp(), true);
			}
		} // end if inst has dest memory operand
		if (global_STARS_program->ShouldSTARSShadowFuncPtrs() || global_STARS_program->ShouldSTARSMaximizeCFGImprovement()) {

#if 0
			if (!CurrInst->HasDestMemoryOperand() && CurrInst->IsMarkerInst()) { // analyze memory DEFs in marker inst
				set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstDef();
				while (DefIter != CurrInst->GetLastDef()) {
					STARSOpndTypePtr DefOp = DefIter->GetOp();
					if (DefOp->IsMemOp()) {
						int SSANum = DefIter->GetSSANum();
						this->ResetProcessedBlocks();
						FoundIndWrite = this->FindPossibleChainAlias(CurrInst, DefOp, SSANum);
						if (FoundIndWrite) {
							// Mark the DEF as aliased.
							DefIter = CurrInst->SetDefIndWrite(DefOp, true);
						}
					}
					++DefIter;
				}
			}
#endif
		}
	} // end for all instructions

	return;
} // end of SMPFunction::AliasAnalysis()

// Does the DefOp DEF_USE chain have an indirect mem write starting at CurrInst?
bool SMPFunction::FindPossibleChainAlias(SMPInstr *CurrInst, const STARSOpndTypePtr &DefOp, int SSANum) {
#if 0
	bool DebugFlag = false;
	if (0 == strcmp("sdissect", this->GetFuncName())) {
		// Next line is just a good place to set a break point.
		DebugFlag = true;
	}
#endif

	// Starting with the DEF instruction, traverse the control flow
	//  until we run into (A) the re-definition of the operand, including
	//  a re-definition of any of its addressing registers, or (B) an
	//  indirect write. Return false if condition A terminates the
	//  search, and true if condition B terminates the search.
	SMPBasicBlock *CurrBlock = CurrInst->GetBlock();
	if (!(CurrBlock->IsProcessed())) {
		CurrBlock->SetProcessed(true);
	}
	else
		return false; // block already processed

	// Proceed by cases:
	STARS_ea_t DefAddr = CurrInst->GetAddr();
	vector<SMPInstr *>::iterator InstIter = CurrBlock->GetInstIterFromAddr(DefAddr);
	bool IndWriteFound = false, ReDef = false;
	++InstIter;
	// Case 1: Local name. Return the IndWrite flag for the local Def-Use
	//  chain begun by CurrInst.
	bool UseAfterIndirectWrite = CurrBlock->HasUseAfterIndWrite(DefOp, InstIter, IndWriteFound, ReDef);
	bool LiveOutFlag = CurrBlock->IsLiveOut(DefOp);
	if (CurrBlock->IsLocalName(DefOp)) {
		return (UseAfterIndirectWrite);
	}

	// Case 2: Global name.
#if 1
	if (UseAfterIndirectWrite) {
		return true;
	}
	else if (IndWriteFound) {
		// We found an indirect write, but no USE of DefOp after it.
		//  If DefOp is LiveOut, then there is a USE of DefOp in a
		//  successor block, so DefOp can be aliased.
		return (LiveOutFlag && !ReDef);
	}
#else
	// Case 2A: If Def-Use chain within this block for this memory operand
	//  has its IndWrite flag set to true, then stop and return true.
	else if (CurrBlock->GetGlobalDUChainIndWrite(DefOp, DefAddr)) {
		return true;
	}

	// Case 2B: Else if Def-Use chain is not the last chain in this block
	//  for this operand, then there must be a later redefinition of the
	//  memory operand (with new SSA number assigned) later in this block.
	//  Because we did not fall into case 2A, we know there is no IndWrite
	//  within the current memory operand's chain, so we return false.
	else if (!CurrBlock->IsLastGlobalChain(DefOp, DefAddr)) {
		return false;
	}

	// Case 2C: Else if current memory operand is NOT LiveOut, even though
	//  this is the last def-use chain in the block, then there is no more
	//  traversing of the control flow graph to be done. The chain has ended
	//  without encountering an IndWrite, so return false.
	else if (!(CurrBlock->IsLiveOut(DefOp))) {
		return false;
	}
#endif

	// Case 2D: We have passed all previous checks, so we must have a memory
	//  operand that reaches the end of the block without encountering an
	//  IndWrite and is LiveOut. Its may-alias status will be determined by
	//  following the control flow graph for all successor blocks and examining
	//  the def-use chains in those blocks.
	list<SMPBasicBlock *>::iterator SuccBlock;
	SuccBlock = CurrBlock->GetFirstSucc();
	bool FoundAliasedWrite = false;
	if (LiveOutFlag && !ReDef) {
		do {
			if ((*SuccBlock)->IsLiveIn(DefOp)) {
				FoundAliasedWrite = this->FindChainAliasHelper(SuccBlock, DefOp);
			}
			++SuccBlock;
		} while (!FoundAliasedWrite && (SuccBlock != CurrBlock->GetLastSucc()));
	}

	return FoundAliasedWrite;
} // end of SMPFunction::FindPossibleChainAlias()

// recursive helper for global DU-chains that traverse CFG
bool SMPFunction::FindChainAliasHelper(list<SMPBasicBlock *>::iterator BlockIter, const STARSOpndTypePtr &DefOp) {
	bool DebugFlag = false;
	SMPBasicBlock *CurrBlock = (*BlockIter);
	if (0 == strcmp("mem2chunk_check", this->GetFuncName())) {
		// Next line is just a good place to set a break point.
		DebugFlag = true;
	}

	if (!(CurrBlock->IsProcessed())) {
		CurrBlock->SetProcessed(true);
	}
	else
		return false; // block already processed

	// The LVA sets can be used to decide whether it is possible that
	//  the incoming DU chain overlaps a may-alias write. We can express
	//  the decision making in a truth table:
	//
	//  Case #    LiveIn?   Killed?   AliasedWrite in block?  Action to take
	//  -------   -------   -------   ----------------------  --------------
	//    1          N          N                N             return false
	//    2          N          N                Y             return false
	//    3          N          Y                N             return false
	//    4          N          Y                Y             return false
	//    5          Y          N                N             recurse into successors
	//    6          Y          N                Y             return true
	//    7          Y          Y                N             return false
	//    8          Y          Y                Y             check location of aliased write
	//
	// In the last case, if there is an aliased write before the
	// incoming DEF is killed and after it is used, then the
	// incoming DU chain overlaps an aliased write, otherwise
	// it does not.


	// If not LiveIn, incoming DU chain does not run through this block
	//  at all, so return false.
	if (!(CurrBlock->IsLiveIn(DefOp)))
		return false;  // cases 1-4

	bool killed = CurrBlock->IsVarKill(DefOp);
	bool BlockHasAliasedWrite = CurrBlock->MaybeAliasedWrite();

	if (BlockHasAliasedWrite) {
		// If DefOp is LiveIn and is not killed, then any aliased
		//  write in the block overlaps the incoming DU chain.
		if (!killed) {
			return true;  // case 6
		}
		// If DefOp is LiveIn and is killed, then the location
		//  of the aliased write is the determining factor.
		else { // case 8; depends on finding indirect write, then USE, before re-DEF.
			vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst();
			bool IndWriteFound, ReDef;
			return CurrBlock->HasUseAfterIndWrite(DefOp, InstIter, IndWriteFound, ReDef);
		}
	}
	else {
		// If killed, no aliased write, then cannot overlap an aliased write.
		if (killed)
			return false; // case 7
		else {
			// Need to recurse into all successors, because we passed through
			//  the block without seeing an aliased write and without killing
			//  the DefOp.
			list<SMPBasicBlock *>::iterator SuccBlock;
			SuccBlock = CurrBlock->GetFirstSucc();
			bool FoundAliasedWrite = false;
			while (!FoundAliasedWrite && (SuccBlock != CurrBlock->GetLastSucc())) {
				FoundAliasedWrite = this->FindChainAliasHelper(SuccBlock, DefOp);
				++SuccBlock;
			};

			if (DebugFlag) {
				SMP_msg("FindChainAliasHelper is returning %d\n", FoundAliasedWrite);
			}
			return FoundAliasedWrite;
		}
	}
	assert(false); // statement should be unreachable
	return false;
} // end of SMPFunction::FindChainAliasHelper()

// Remove a basic block and its instructions.
void SMPFunction::RemoveBlock(SMPBasicBlock *CurrBlock, list<SMPBasicBlock *>::iterator &BlockIter, bool IBTarget) {
	if (IBTarget) { 
		// Block could be IBTarget and thus actually be reachable. Cover our bases by emitting an IBT annotation.
		SMPInstr *FirstInst = (*(CurrBlock->GetFirstInst()));
		global_STARS_program->PrintUnknownCodeXref(CurrBlock->GetFirstAddr(), FirstInst->GetSize(), ZST_UNREACHABLEBLOCK);

		// It cannot hurt to add INSTR BELONGTO annotations to the main annotations file.
		STARS_ea_t FuncAddr = this->GetFirstFuncAddr();
		FILE *AnnotFile = global_STARS_program->GetAnnotFile();
		assert(nullptr != AnnotFile);
		for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			SMP_fprintf(AnnotFile, "%18llx %6zu INSTR BELONGTO %llx \n",
				(uint64_t) InstAddr, CurrInst->GetSize(), (uint64_t) FuncAddr);
		}
	}

	// Remove this block from the predecessors list of its successors.
	list<SMPBasicBlock *>::iterator SuccIter;
	STARS_ea_t TempAddr = CurrBlock->GetFirstAddr();
	for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
		(*SuccIter)->ErasePred(TempAddr);
	}

	// Remove this block from the successors list of its predecessors.
	list<SMPBasicBlock *>::iterator PredIter;
	for (PredIter = CurrBlock->GetFirstPred(); PredIter != CurrBlock->GetLastPred(); ++PredIter) {
		(*PredIter)->EraseSucc(TempAddr);
	}

	// Transfer the unreachable block to the program-wide container of unreachable code.
	this->GetProg()->AddUnreachableBlock(CurrBlock);

	// Remove the unreachable instructions from the function inst list.
	vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst();
	STARS_ea_t FirstBadAddr = (*InstIter)->GetAddr();
	InstIter = --(CurrBlock->GetLastInst()); // get last real instruction
	STARS_ea_t LastBadAddr = (*InstIter)->GetAddr();
	this->EraseInstRange(FirstBadAddr, LastBadAddr);

	// Remove the block from the blocks list.
	BlockIter = this->Blocks.erase(BlockIter);
	this->BlockCount -= 1;
	return;
} // end of SMPFunction::RemoveBlock()

// Func is empty, so add all blocks that call it to Program->BlocksPendingRemoval.
void SMPFunction::RemoveCallingBlocks(void) const {
	for (set<STARS_ea_t>::iterator CallSiteIter = this->AllCallSites.begin(); CallSiteIter != this->AllCallSites.end(); ++CallSiteIter) {
		STARS_ea_t CallInstAddr = (*CallSiteIter);
		STARS_ea_t CallingFuncAddr = STARS_BADADDR;
		if (this->GetProg()->IsInstAddrStillInFunction(CallInstAddr, CallingFuncAddr)) {
			SMPFunction *CallingFunc = this->GetProg()->FindFunction(CallingFuncAddr);
			if (nullptr == CallingFunc) {
				SMP_msg("ERROR: Cannot find function with start addr %llx and call inst at %llx ", 
					(uint64_t) CallingFuncAddr, (uint64_t) CallInstAddr);
				this->DumpFuncNameAndAddr();
			}
			else {
				SMPBasicBlock *CallingBlock = CallingFunc->GetBlockFromInstAddr(CallInstAddr);
				assert(nullptr != CallingBlock);
				this->GetProg()->AddBlockToRemovalList(CallingBlock);
			}
		}
	}
	return;
} // end of SMPFunction::RemoveCallingBlocks()

// Link basic blocks to their predecessors and successors.
void SMPFunction::SetLinks(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	list<SMPBasicBlock *> UnresolvedBranchWorkList;
	STARS_ea_t InstAddr;
	STARS_ea_t FirstAddr = this->GetFirstFuncAddr();
	bool DebugFlag = ((0x717be0 == FirstAddr) || (0x13f264 == FirstAddr));
	if (DebugFlag) {
		SMP_msg("SetLinks called for ");
		this->DumpFuncNameAndAddr();
	}

	// Set successors of each basic block, also setting up the predecessors in the
	//  process.
	for (SMPBasicBlock *CurrBlock : this->Blocks) {
		int BlockNum = CurrBlock->GetNumber();
		vector<SMPInstr *>::iterator InstIter = (--(CurrBlock->GetLastInst()));
		SMPInstr *CurrInst = (*InstIter);
		InstAddr = CurrInst->GetAddr();
		STARS_ea_t BlockAddr = CurrBlock->GetFirstAddr();
		if (DebugFlag) {
			SMP_msg("DEBUG: SetLinks processing block %d from %llx to %llx\n",
				BlockNum, (uint64_t) BlockAddr, (uint64_t) InstAddr);
		}
		bool CondTailCall = false;
		if (CurrBlock->HasReturn()) {
			if (!(CurrInst->IsCondTailCall())) {
				// We either have a return instruction or an unconditional
				//  tail call instruction. We don't want to link to the
				//  tail call target, and there is no link for a return
				if (DebugFlag) {
					SMP_msg("DEBUG: SetLinks not linking successors for return block\n");
				}
				continue;
			}
			else {
				// We have a conditional tail call. We don't want to
				//  link to the tail call target, but we do want fall
				//  through to the next instruction.
				CondTailCall = true;
				if (DebugFlag) {
					SMP_msg("DEBUG: SetLinks linking successors for CondTailCall block\n");
				}
			}
		}
		
		// Last instruction in block; set successors
		SMPitype FlowType = CurrInst->GetDataFlowType();
		if (DebugFlag) {
			SMP_msg("DEBUG: SetLinks: FlowType ");
			DumpDataFlowType(FlowType);
		}
		if (HALT == FlowType) {
			if (DebugFlag) {
				SMP_msg("DEBUG: SetLinks not linking successors for HALT block\n");
			}
			continue;
		}
		bool CallFlag = (CALL == FlowType);
		bool IndirJumpFlag = (INDIR_JUMP == FlowType);
		bool IndirCallFlag = (INDIR_CALL == FlowType);
		bool SysCallFlag = (IndirCallFlag && CurrInst->MDIsSystemCall());
		// NOTE: Due to phase re-ordering, we cannot yet identify tail calls,
		//  so CondTailCall and TailCallFlag will always be false, which is harmless.
		//  SMPInstr::SetTailCall() will do a little cleanup later.
		bool TailCallFlag = CondTailCall && CurrInst->IsCondTailCall();
		bool LinkedToTarget = false;
		bool FixedCallFlag = (CallFlag && CurrInst->IsFixedCallJump());
		if (DebugFlag) {
			SMP_msg("DEBUG: SysCallFlag %d TailCallFlag %d FixedCallFlag %d\n", SysCallFlag, TailCallFlag, FixedCallFlag);
		}
		list<STARS_InstructionID_t> SuccList;
		global_STARS_program->GetBlockSuccessorTargets((CallFlag || IndirCallFlag || TailCallFlag), CurrInst->GetInstID(), CurrInst->GetSize(), SuccList);
		if (FixedCallFlag) { // Need to find successor from fixed call jump
			assert(SuccList.size() <= 1); // jump should have <= 1 target successor, no fall-through
			STARS_ea_t FallThroughAddr = CurrBlock->FindFixedCallFallThrough(); // get fall-through from push
			map<STARS_ea_t, SMPBasicBlock *>::iterator MapEntry = this->InstBlockMap.find(FallThroughAddr);
			if (MapEntry == this->InstBlockMap.end()) {
				SMP_msg("WARNING: FixedCallPush at %llx not linked to FallThroughAddr at %llx\n",
					(uint64_t) InstAddr, (uint64_t) FallThroughAddr);
			}
			else {
				SMPBasicBlock *Target = MapEntry->second;
				// Make target block a successor of current block.
				CurrBlock->LinkToSucc(Target);
				// Make current block a predecessor of target block.
				Target->LinkToPred(CurrBlock);
				LinkedToTarget = true;
				if (DebugFlag) {
					SMP_msg("DEBUG: SetLinks linking to successor block at %llx\n",
						(uint64_t) FallThroughAddr);
				}
			}
			if (!LinkedToTarget) {
				SMP_msg("ERROR: SetLinks unable to link to successor of fixed-call block ending at %llx ",
					(uint64_t) InstAddr);
				this->DumpFuncNameAndAddr();
			}
		}
		else {
			for (list<STARS_InstructionID_t>::iterator IDIter = SuccList.begin(); IDIter != SuccList.end(); ++IDIter) {
				STARS_ea_t TargetAddr = IDIter->GetIDWithinFile();

				map<STARS_ea_t, SMPBasicBlock *>::iterator MapEntry = this->InstBlockMap.find(TargetAddr);
				if (MapEntry == this->InstBlockMap.end()) {
					; // do nothing; probably a tail call (not yet identified)
#if 1
					if (DebugFlag) {
						SMP_msg("WARNING: TargetAddr %llx not found in map for %s\n", (uint64_t) TargetAddr,
							this->GetFuncName());
						SMP_msg(" Referenced from %s at %llx\n", CurrInst->GetDisasm(), (uint64_t)InstAddr);
					}
#endif
				}
				else {
					SMPBasicBlock *Target = MapEntry->second;
					// Make target block a successor of current block.
					CurrBlock->LinkToSucc(Target);
					// Make current block a predecessor of target block.
					Target->LinkToPred(CurrBlock);
					LinkedToTarget = true;
					if (DebugFlag) {
						SMP_msg("DEBUG: SetLinks linking to successor block %d at %llx\n",
							Target->GetNumber(), (uint64_t)TargetAddr);
					}
#if SMP_USE_SWITCH_TABLE_INFO
					if (IndirJumpFlag) {
						pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->JumpTableTargets.insert(TargetAddr);
						if (!InsertResult.second) {
							SMP_msg("WARNING: code address %llx found in multiple jump tables?\n", (uint64_t) TargetAddr);
						}
#if SMP_DEBUG_SWITCH_TABLE_INFO
						SMP_msg("Switch table link: jump at %x target at %x\n",
							CurrInst->GetAddr(), TargetAddr);
#else
						;
#endif
					}
#endif
				}
			} // end for all IDs in SuccList
		} // end if (FixedCallFlag) ... else [loop through SuccList]

		if (IndirJumpFlag && (!LinkedToTarget)) {
			this->UnresolvedIndirectJumps = true;
			++UnresolvedIndirectJumpCount;
			UnresolvedBranchWorkList.push_back(CurrBlock);
			SMP_msg("WARNING: Unresolved indirect jump at %llx FuncAddr: %llx\n",
				(uint64_t) CurrInst->GetAddr(), (uint64_t) this->GetFirstFuncAddr());
		}
		else if (IndirCallFlag && (!LinkedToTarget) && (!SysCallFlag)) {
			this->SetHasUnresolvedIndirectCalls();
			SMP_msg("WARNING: Unresolved indirect call at %llx FuncAddr: %llx\n",
				(uint64_t) CurrInst->GetAddr(), (uint64_t) this->GetFirstFuncAddr());
		}
		else if (IndirJumpFlag && LinkedToTarget) {
			++ResolvedIndirectJumpCount;
		}
	} // end for all blocks

	// Mark all blocks that can be reached from the entry block, so we can find the unreachable ones.
	this->ResetProcessedBlocks();
	this->Blocks.front()->DepthFirstMark(false);
	// We have two cases: (1) Unresolved indirect branches could be targeting the unmarked blocks, making
	//  these blocks reachable, in which case we should link the unresolved branches to the unmarked blocks;
	//  or (2) there are no unresolved branches, in which case the unmarked blocks are unreachable within
	//  the function. They might be reachable from outside the function using exception handling jumps, but
	//  that still would not allow us to link them into the CFG of this function properly, so in any case we
	//  are deleting those unreachable blocks and not emitting annotations for them.
	// NOTE: An odd new gcc recursion optimization uses indirect calls within the function, so
	//  they can behave like indirect jumps. However, we don't want to link unresolved calls to unmarked blocks
	//  at this time.
	bool HellNodeCase = (!UnresolvedBranchWorkList.empty() && (this->HasUnresolvedIndirectCalls() || this->HasUnresolvedIndirectJumps()));
	bool AddedMissingLinks = false;
	bool changed;
	bool FirstIteration = true;
	do {
		changed = false;
		list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin();
		while (BlockIter != this->Blocks.end()) {
			SMPBasicBlock *CurrBlock = (*BlockIter);
			STARS_ea_t BlockAddr = CurrBlock->GetFirstAddr();
			int BlockNum = CurrBlock->GetNumber();
			if (CurrBlock->IsProcessed()) {
				if (CurrBlock->IsUnreachableBlock()) { // has call 0 instruction, must be unreachable
					this->GetProg()->AddBlockToRemovalList(CurrBlock);
					if (DebugFlag) {
						SMP_msg("DEBUG: Block %d unreachable due to call 0 instruction?\n", BlockNum);
					}
				}
				++BlockIter;
			}
			else {
				// Block cannot be reached from entry node, even after we have added links
				//  on previous loop iterations.
				bool CatchBlockFound = global_STARS_program->IsCatchBlockAddr(BlockAddr);
				if (DebugFlag) {
					SMP_msg("DEBUG: SetLinks found unreachable block %d at %llx Catch: %d\n", 
						BlockNum, (uint64_t) BlockAddr, CatchBlockFound);
				}
				if (!HellNodeCase) {
					if (CurrBlock->AllNops())
						SMP_msg("INFO: Removing all nops block %d at %llx\n", BlockNum, (uint64_t) BlockAddr);
					else {
						CurrBlock->SetUnreachableBlock(true);
						SMP_msg("INFO: Function is Removing unreachable block %d at %llx\n", BlockNum, (uint64_t) BlockAddr);
					}

					bool MightBeIndirectTarget = true;
					this->RemoveBlock(CurrBlock, BlockIter, MightBeIndirectTarget);

#if 0  // Exception handling code requires something more delicate than this. Later checks for stack adjustment etc. can look at these blocks.
					// Finally, call destructors on the block and insts removed.
					InstIter = CurrBlock->GetFirstInst();
					while (InstIter != CurrBlock->GetLastInst()) {
						SMPInstr *DeadInst = (*InstIter);
						++InstIter;
						if (nullptr != DeadInst) delete DeadInst;
					}
					delete CurrBlock;
#endif
				}
				else { // HellNodeCase
					if (!CatchBlockFound) {
						if (DebugFlag) {
							SMP_msg("DEBUG: SetLinks adding hell node links for unreachable block %d at %llx\n",
								BlockNum, (uint64_t)BlockAddr);
						}
						// Block must be reachable only through an unresolved indirect branch.
						// Make each unresolved indirect branch link to the block so it is reachable.
						AddedMissingLinks = true;
						for (list<SMPBasicBlock *>::iterator WorkIter = UnresolvedBranchWorkList.begin(); WorkIter != UnresolvedBranchWorkList.end(); ++WorkIter) {
							SMPBasicBlock *WorkBlock = (*WorkIter);
							WorkBlock->LinkToSucc(CurrBlock);
						}
					}
					else {
						SMP_msg("INFO: SetLinks not adding link to block %d because %llx is a catch block addr.\n", BlockNum, (uint64_t)BlockAddr);
					}
					// Mark CurrBlock as now being reachable, along with the blocks it dominates.
					CurrBlock->DepthFirstMark(true); // yes, even if it is a catch block
					++BlockIter;
					// Block could be IBTarget and thus actually be reachable. Cover our bases by emitting an IBT annotation.
					SMPInstr *FirstInst = (*(CurrBlock->GetFirstInst()));
					global_STARS_program->PrintUnknownCodeXref(BlockAddr, FirstInst->GetSize(), ZST_UNREACHABLEBLOCK);
				}
				changed = true;
			} // end if (processed) ... else ...
		} // end loop through blocks
		FirstIteration = false;
	} while (changed);
	if (HellNodeCase && (!AddedMissingLinks)) {
		SMP_msg("SERIOUS WARNING: SetLinks: Function at %llx has unresolved indirect branches but no unreachable blocks.\n",
			(uint64_t) FirstAddr);
	}

#if 0
	// If we have any blocks that are all no-ops and have no predecessors, remove those
	//  blocks. They are dead and make the CFG no longer a lattice. Any blocks that have
	//  no predecessors but are not all no-ops should also be removed with a different
	//  log message.
	// NOTE: Prior to construction of hell nodes in functions with unresolved indirect jumps,
	//  we cannot conclude that a block with no predecessors is unreachable. Also, the block
	//  order might be such that removal of a block makes an already processed block
	//  unreachable, so we have to iterate until there are no more changes.
	bool NoPredecessors;
	bool OnlyPredIsItself;
	list<SMPBasicBlock *>::iterator CurrPred;
#if SMP_USE_SWITCH_TABLE_INFO
	if (!(this->HasUnresolvedIndirectJumps() || this->HasUnresolvedIndirectCalls())) {
#else
	if (!(this->HasIndirectJumps() || this->HasIndirectCalls())) {
#endif
		bool changed;
		do {
			changed = false;
			BlockIter = this->Blocks.begin();
			++BlockIter; // don't delete the top block, no matter what.
			while (BlockIter != this->Blocks.end()) {
				CurrBlock = (*BlockIter);
				OnlyPredIsItself = false;
				CurrPred = CurrBlock->GetFirstPred();
				NoPredecessors = (CurrPred == CurrBlock->GetLastPred());
				if (!NoPredecessors) {
					if ((*CurrPred)->GetFirstAddr() == CurrBlock->GetFirstAddr()) { // self-recursion
						++CurrPred; // any more preds besides itself?
						OnlyPredIsItself = (CurrPred == CurrBlock->GetLastPred());
							// Only predecessor was the self-recursion if no more preds
					}
				}
				if (NoPredecessors || OnlyPredIsItself) {
					if (CurrBlock->AllNops())
						SMP_msg("Removing all nops block at %x\n", CurrBlock->GetFirstAddr());
					else
						SMP_msg("Removing block with no predecessors at %x\n", CurrBlock->GetFirstAddr());
					// Remove this block from the predecessors list of its successors.
					list<SMPBasicBlock *>::iterator SuccIter;
					STARS_ea_t TempAddr = CurrBlock->GetFirstAddr();
					for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
						(*SuccIter)->ErasePred(TempAddr);
					}
					// Transfer the unreachable block to the program-wide container of unreachable code.
					this->GetProg()->AddUnreachableBlock(CurrBlock);
					// Remove the unreachable instructions from the function inst list.
					vector<SMPInstr *>::iterator InstIter;
					InstIter = CurrBlock->GetFirstInst();
					STARS_ea_t FirstBadAddr = (*InstIter)->GetAddr();
					InstIter = CurrBlock->GetLastInst();
					--InstIter; // get last real instruction
					STARS_ea_t LastBadAddr = (*InstIter)->GetAddr();
					this->EraseInstRange(FirstBadAddr, LastBadAddr);

					// Finally, remove the block from the blocks list.
					BlockIter = this->Blocks.erase(BlockIter);
					this->BlockCount -= 1;
					changed = true;
				}
				else {
					++BlockIter;
				}
			} // end while all blocks after the first one
		} while (changed);
	} // end if not unresolved indirect jumps or indirect calls
	else if (this->HasUnresolvedIndirectJumps()) {
		// Make each unresolved indirect branch have each block with no predecessor as a target,
		//  so that the resulting CFG has a proper structure.
		BlockIter = this->Blocks.begin();
		++BlockIter; // The top block is expected to have no predecessors, which is not a CFG problem.
		bool AddedMissingLinks = false;
		while (BlockIter != this->Blocks.end()) {
			CurrBlock = (*BlockIter);
			OnlyPredIsItself = false;
			CurrPred = CurrBlock->GetFirstPred();
			NoPredecessors = (CurrPred == CurrBlock->GetLastPred());
			if (!NoPredecessors) {
				if ((*CurrPred)->GetFirstAddr() == CurrBlock->GetFirstAddr()) { // self-recursion
					++CurrPred; // any more preds besides itself?
					OnlyPredIsItself = (CurrPred == CurrBlock->GetLastPred());
						// Only predecessor was the self-recursion if no more preds
				}
			}
			if (NoPredecessors || OnlyPredIsItself) {
				// Block must be reachable only through an unresolved indirect branch.
				// Make each unresolved indirect branch link to the block so it is reachable.
				list<SMPBasicBlock *>::iterator WorkIter;
				AddedMissingLinks = true;
				for (WorkIter = UnresolvedBranchWorkList.begin(); WorkIter != UnresolvedBranchWorkList.end(); ++ WorkIter) {
					SMPBasicBlock *WorkBlock = (*WorkIter);
					WorkBlock->LinkToSucc(CurrBlock);
				}
			}
			++BlockIter;
		} // end for all blocks
		if (!AddedMissingLinks) {
			SMP_msg("SERIOUS WARNING: Function at %x has unresolved indirect branches but no unreachable blocks.\n", this->GetFirstFuncAddr());
		}
	}
#endif

	return;
} // end of SMPFunction::SetLinks()

// Populate TCFG for SPARK translation before any cloning occurs
void SMPFunction::InitializeTCFG(void) {
	this->TCFGBlocks.clear();
	// Copy the current block info.
	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		STARSTransCFGBlock *TransBlock = new STARSTransCFGBlock(this);
		assert(nullptr != TransBlock);
		// Set all TCFG block fields to match current CFG block.
		TransBlock->SetBlockNum((int)BlockIndex);
		TransBlock->SetOriginalBlockNum((int)BlockIndex);
		this->TCFGBlocks.push_back(TransBlock);
	}

	// Set up predecessor and successor links.
	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		for (SMPBasicBlock *PredBlock : this->GetBlockByNum(BlockIndex)->GetConstPredList()) {
			int PredBlockNum = PredBlock->GetNumber();
			this->TCFGBlocks[BlockIndex]->AddPred(this->TCFGBlocks[(size_t)PredBlockNum]);
		}
		for (SMPBasicBlock *SuccBlock : this->GetBlockByNum(BlockIndex)->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();
			this->TCFGBlocks[BlockIndex]->AddSucc(this->TCFGBlocks[(size_t)SuccBlockNum]);
		}
	}

	return;
} // end of SMPFunction::InitializeTCFG()

// Number all basic blocks in reverse postorder (RPO) and set RPOBlocks vector to
//  access them.
void SMPFunction::RPONumberBlocks(void) {
#if SMP_DEBUG_DATAFLOW
	bool DebugFlag = false;
	DebugFlag = (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
	if (DebugFlag) SMP_msg("Entered RPONumberBlocks\n");
#endif
	int CurrNum = 0;
	list<SMPBasicBlock *> WorkList;
	this->ResetProcessedBlocks();
	bool Renumbering = (!(this->RPOBlocks.empty()));
	this->RPOBlocks.clear();  // we can re-do the RPO numbering as we remove unreachable blocks.

#if 1
	list<SMPBasicBlock *> PostorderBlockList;
	this->CreatePostorderBlockList(PostorderBlockList);
	this->ResetProcessedBlocks();
	for (list<SMPBasicBlock *>::const_reverse_iterator BlockIter = PostorderBlockList.crbegin();
		BlockIter != PostorderBlockList.crend();
		++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		CurrBlock->SetNumber(CurrNum);
		CurrBlock->SetProcessed(true);
		++CurrNum;
		this->RPOBlocks.push_back(CurrBlock);
	}
#else
	// Number the first block with 0.
	list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin();
	if ((BlockIter == this->Blocks.end()) || (0 == this->BlockCount)) {
		// Empty function. Happens when we remove basic blocks with "call 0" instructions.
		//  This means the entire function is only reachable when the code is compiled
		//  with some debug option, which was not present.
		this->BlockCount = 0;
		return;
	}

	SMPBasicBlock *CurrBlock = (*BlockIter);
	CurrBlock->SetNumber(CurrNum);
	CurrBlock->SetProcessed(true);
	this->RPOBlocks.push_back(CurrBlock);
	++CurrNum;
	// Push the first block's successors onto the work list.
	list<SMPBasicBlock *>::iterator CurrSucc = CurrBlock->GetFirstSucc();
	while (CurrSucc != CurrBlock->GetLastSucc()) {
		WorkList.push_back(*CurrSucc);
		++CurrSucc;
	}

	// Use the WorkList to iterate through all blocks in the function
	list<SMPBasicBlock *>::iterator CurrListItem = WorkList.begin();
	bool change;
	while (!WorkList.empty()) {
		change = false;
		while (CurrListItem != WorkList.end()) {
			CurrBlock = (*CurrListItem);
			STARS_ea_t CurrAddr = CurrBlock->GetFirstAddr();
			if (CurrBlock->IsProcessed()) {
				// Duplicates get pushed onto the WorkList because a block
				//  can be the successor of multiple other blocks. If it is
				//  already numbered, it is a duplicate and can be removed
				//  from the list.
				CurrListItem = WorkList.erase(CurrListItem);
				change = true;
				continue;
			}
			if (CurrBlock->AllPredecessorsProcessed()) {
				// Ready to be numbered.
				CurrBlock->SetNumber(CurrNum);
				CurrBlock->SetProcessed(true);
#if 0
				SMP_msg("Set RPO number %d\n", CurrNum);
				if (DebugFlag && (7 == CurrNum))
					this->Dump(false);
#endif
				this->RPOBlocks.push_back(CurrBlock);
				++CurrNum;
				change = true;
				// Push its unnumbered successors onto the work list.
				CurrSucc = CurrBlock->GetFirstSucc();
				while (CurrSucc != CurrBlock->GetLastSucc()) {
					if (!(*CurrSucc)->IsProcessed())
						WorkList.push_back(*CurrSucc);
					++CurrSucc;
				}
				CurrListItem = WorkList.erase(CurrListItem);
			}
			else {
				++CurrListItem;
			}
		} // end while (CurrListItem != WorkList.end())
		if (change) {
			// Reset CurrListItem to beginning of work list for next iteration.
			CurrListItem = WorkList.begin();
		}
		else {
			// Loops can cause us to not be able to find a WorkList item that has
			//  all predecessors numbered. Take the WorkList item with the lowest address
			//  and number it so we can proceed.
			CurrListItem = WorkList.begin();
			STARS_ea_t LowAddr = (*CurrListItem)->GetFirstAddr();
			list<SMPBasicBlock *>::iterator SaveItem = CurrListItem;
			++CurrListItem;
			while (CurrListItem != WorkList.end()) {
				if (LowAddr > (*CurrListItem)->GetFirstAddr()) {
					SaveItem = CurrListItem;
					LowAddr = (*CurrListItem)->GetFirstAddr();
				}
				++CurrListItem;
			}
			// SaveItem should now be numbered.
			(*SaveItem)->SetNumber(CurrNum);
			(*SaveItem)->SetProcessed(true);
#if SMP_DEBUG_DATAFLOW
			SMP_msg("Picked LowAddr %x and set RPO number %d\n", LowAddr, CurrNum);
#endif
			this->RPOBlocks.push_back(*SaveItem);
			++CurrNum;
			// Push its unnumbered successors onto the work list.
			CurrSucc = (*SaveItem)->GetFirstSucc();
			while (CurrSucc != (*SaveItem)->GetLastSucc()) {
				if (!(*CurrSucc)->IsProcessed())
					WorkList.push_back(*CurrSucc);
				++CurrSucc;
			}
			CurrListItem = WorkList.erase(SaveItem);
			CurrListItem = WorkList.begin();
		} // end if (change) ... else ...
	} // end while work list is nonempty

#endif

	// Prior to construction of hell nodes for functions with indirect jumps, there
	//  could still be unnumbered blocks because they appear to be unreachable
	//  (no predecessors from SetLinks() because they are reached only via indirect
	//  jumps). We need to number these and push them on the RPOBlocks vector so
	//  that the vector contains all the blocks.
	// NOTE: Odd new gcc recursion optimization seems to use indirect calls to reach
	//  some blocks within a recursive function, operating somewhat like an indirect
	//  jump.
	if (!Renumbering && (CurrNum < this->BlockCount)) { // already did this during the first numbering
		if (this->HasIndirectJumps() || this->HasIndirectCalls()) {
			for (SMPBasicBlock *CurrBlock : this->Blocks) {
				if (SMP_BLOCKNUM_UNINIT == CurrBlock->GetNumber()) {
					SMP_msg("WARNING: Numbering indirectly reachable block at %llx\n", (uint64_t) CurrBlock->GetFirstAddr());
					CurrBlock->SetNumber(CurrNum);
					CurrBlock->SetProcessed(true);
					this->RPOBlocks.push_back(CurrBlock);
					++CurrNum;
				}
			}
		}
	}
	// If we still have unnumbered blocks, it is not because of indirect jumps or calls.
	//  We have some mysterious dead code. NOTE: Probably just try/catch blocks reachable from
	//  addresses stored in the EH_FRAME section of the binary.
	if (this->BlockCount > ((int) this->RPOBlocks.size())) {
		SMP_msg("SERIOUS WARNING: RPONumberBlocks method: Function %s has BlockCount %d and RPOBlocks size %zu\n",
			this->GetFuncName(), this->BlockCount, this->RPOBlocks.size());
		for (SMPBasicBlock *CurrBlock : this->Blocks) {
			if (!(CurrBlock->IsProcessed())) {
				SMP_msg("WARNING: Numbering apparently unreachable block at %llx\n", (uint64_t) CurrBlock->GetFirstAddr());
				CurrBlock->SetNumber(CurrNum);
				CurrBlock->SetProcessed(true);
				this->RPOBlocks.push_back(CurrBlock);
				++CurrNum;
			}
		}
	}

	if (((long long) this->BlockCount) > ((long long) STARS_MaxBlockCount)) {
		assert(0 < this->BlockCount);
		STARS_MaxBlockCount = (unsigned long) this->BlockCount;
		assert(this->BlockCount <= STARS_BLOCKNUM_MASK);
	}

	return;
} // end of SMPFunction::RPONumberBlocks()

void SMPFunction::CreatePostorderBlockList(list<SMPBasicBlock *> &PostorderList) {
	this->ResetProcessedBlocks();
	// Do a depth-first traversal, marking blocks visited, using those markings to avoid
	//  taking back edges.
	SMPBasicBlock *HeadBlock = this->Blocks.front();
	HeadBlock->SetProcessed(true);
	this->PostorderBlockListHelper(HeadBlock, PostorderList);

	return;
} // end of SMPFunction::CreatePostorderBlockList()

void SMPFunction::PostorderBlockListHelper(SMPBasicBlock *CurrBlock, list<SMPBasicBlock *> &PostorderList) {
	for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetFirstConstSucc();
		SuccIter != CurrBlock->GetLastConstSucc();
		++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		if (!SuccBlock->IsProcessed()) {
			SuccBlock->SetProcessed(true);
			this->PostorderBlockListHelper(SuccBlock, PostorderList); // recurse
		}
	}
	// After all successors are visited, push this block on the list.
	PostorderList.push_back(CurrBlock);

	return;
} // end of SMPFunction::PostorderBlockListHelper()

// Number TCFG basic blocks in reverse post-order and place pointers in TCFGBlocks. Called after cloning.
void SMPFunction::RPONumberTCFGBlocks(void) {
	int CurrNum = 0;
	list<STARSTransCFGBlock *> WorkList;
	this->ResetTCFGVisitedBlocks();
	size_t TCFGBlockCount = this->TCFGBlocks.size();
	bool Renumbering = (0 < TCFGBlockCount);
	assert(Renumbering); // Should only be here if we are renumbering after cloning
	vector<STARSTransCFGBlock *> TCFGBlocksCopy; // Make vector for rearranging

	list<STARSTransCFGBlock *> PostorderTCFGBlockList;
	this->CreatePostorderTCFGBlockList(PostorderTCFGBlockList);
	this->ResetTCFGVisitedBlocks();
	// Do a reverse iteration over the PostorderTCFGBlockList to produce RPO numbers.
	size_t PostorderListSize = PostorderTCFGBlockList.size();
	if (PostorderListSize < TCFGBlockCount) {
		// Find which blocks were missed.
		STARSBitSet OldBlockNumsListed;
		OldBlockNumsListed.AllocateBits(TCFGBlockCount);
		for (STARSTransCFGBlock *ListBlock : PostorderTCFGBlockList) {
			int ListBlockNum = ListBlock->GetNumber();
			assert(0 <= ListBlockNum);
			OldBlockNumsListed.SetBit((size_t)ListBlockNum);
		} // end for all blocks in postorder list
		for (size_t BitIndex = 0; BitIndex < TCFGBlockCount; ++BitIndex) {
			if (!OldBlockNumsListed.GetBit(BitIndex)) {
				SMP_msg("ERROR: SPARK: Old TCFG block %zu not in postorder list before renumbering ", BitIndex);
				this->DumpFuncNameAndAddr();
			}
		} // end for all BitIndex values
	}
	list<STARSTransCFGBlock *>::const_reverse_iterator BlockIter = PostorderTCFGBlockList.crbegin();
	this->CloningMap.clear();
	this->CloningReverseMap.clear();
	while (BlockIter != PostorderTCFGBlockList.crend()) {
		STARSTransCFGBlock *CurrTCFGBlock = (*BlockIter);

		// Maintain map and reverse map for old and new TCFG block numbers.
		int OldTCFGNum = CurrTCFGBlock->GetNumber();
		assert(0 <= OldTCFGNum);
		pair<int, int> CloningMapItem(OldTCFGNum, CurrNum);
		pair<int, int> CloningReverseMapItem(CurrNum, OldTCFGNum);
		pair<map<int, int>::iterator, bool> InsertResult = this->CloningMap.insert(CloningMapItem);
		assert(InsertResult.second);
		InsertResult = this->CloningReverseMap.insert(CloningReverseMapItem);
		assert(InsertResult.second);

		CurrTCFGBlock->SetBlockNum(CurrNum);
		assert(((size_t)OldTCFGNum) < TCFGBlockCount);
		assert(((size_t)CurrNum) < TCFGBlockCount);
		CurrTCFGBlock->SetVisited(true);
		++CurrNum;
		TCFGBlocksCopy.push_back(CurrTCFGBlock);
		++BlockIter;
	} // end while reverse BlockIter has not reached the end of PostorderTCFGBlockList

	// If there are unreachable TCFG blocks, it is either because we have an
	//  unresolved indirect jump (in which case we don't have structured code and
	//  cannot translate to SPARK Ada; we will emit a stub instead) or because we
	//  have exception handling code reachable only through run-time mechanisms
	//  not visible within this function in the binary (in which case we do not
	//  translate these blocks to SPARK Ada and never reach them by following
	//  links to successors during translation). So, unlike RPONumberBlocks(), this
	//  method does not have to handle blocks that were not reached by the list
	//  creation above.

	// Copy from temporary vector to final vector.
	this->TCFGBlocks.clear();
	this->TCFGBlocks = TCFGBlocksCopy;

	return;
} // end of SMPFunction::RPONumberTCFGBlocks()

// Create postorder traversal, no back edges followed.
void SMPFunction::CreatePostorderTCFGBlockList(std::list<STARSTransCFGBlock *> &PostorderList) {
	this->ResetTCFGVisitedBlocks();
	// Do a depth-first traversal, marking blocks visited, using those markings to avoid
	//  taking back edges.
	STARSTransCFGBlock *HeadBlock = this->TCFGBlocks[0];
	HeadBlock->SetVisited(true);
	this->PostorderTCFGBlockListHelper(HeadBlock, PostorderList);

	return;
} // end of SMPFunction::CreatePostorderTCFGBlockList()

// recursive DFS block visitor
void SMPFunction::PostorderTCFGBlockListHelper(STARSTransCFGBlock *CurrBlock, std::list<STARSTransCFGBlock *> &PostorderList) {
	for (STARSTransCFGBlock *SuccTCFGBlock : CurrBlock->GetConstSuccList()) {
		if (!SuccTCFGBlock->IsVisited()) {
			SuccTCFGBlock->SetVisited(true);
			this->PostorderTCFGBlockListHelper(SuccTCFGBlock, PostorderList); // recurse
		}
	}
	// After all successors are visited, push this block on the list.
	PostorderList.push_back(CurrBlock);

	return;
} // end of SMPFunction::PostorderTCFGBlockListHelper()

// Find greatest RPO block in chain from StartingBlockNum, no back edges followed.
int SMPFunction::FindLastBlockProcessedInChain(const int StartingBlockNum) const {
	assert(0 <= StartingBlockNum);
	assert(((size_t)StartingBlockNum) < this->GetNumBlocks());
	int HighestRPOBlockFound = StartingBlockNum;

	// Follow all successors, find highest block that IsProcessed().
	SMPBasicBlock *StartingBlock = this->GetBlockByNum((size_t) StartingBlockNum);
	assert(StartingBlock->IsProcessed());

	if (!StartingBlock->IsSCCPVisited()) {
		StartingBlock->SetSCCPVisited(true);
		for (list<SMPBasicBlock *>::const_iterator SuccIter = StartingBlock->GetFirstConstSucc();
			SuccIter != StartingBlock->GetLastConstSucc();
			++SuccIter) {
			SMPBasicBlock *SuccBlock = (*SuccIter);
			if (SuccBlock->IsProcessed()) {
				int SuccBlockNum = SuccBlock->GetNumber();
				if (SuccBlockNum <= StartingBlockNum) // back edge
					continue;
				int TempHighestFound = this->FindLastBlockProcessedInChain(SuccBlockNum);
				if (HighestRPOBlockFound < TempHighestFound)
					HighestRPOBlockFound = TempHighestFound;
			}
		}
	}

	return HighestRPOBlockFound;
} // end of SMPFunction::FindLastBlockProcessedInChain()


// return block # of block that falls through to CurrBlockNum; SMP_BLOCKNUM_UNINIT if none
int SMPFunction::GetFallThroughPredBlockNum(int CurrTCFGBlockNum) {
	assert((0 <= CurrTCFGBlockNum) && (CurrTCFGBlockNum < (int)this->TCFGBlocks.size()));
	int FTPredTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
	STARSTransCFGBlock *CurrTCFGBlock = this->TCFGBlocks[(size_t) CurrTCFGBlockNum];
	for (STARSTransCFGBlock *PredBlock : CurrTCFGBlock->GetConstPredList()) {
		list<STARSTransCFGBlock *>::const_iterator SuccIter = PredBlock->GetFallThroughSucc();
		if (SuccIter != PredBlock->GetLastSucc()) { // has a fall-through successor
			if ((*SuccIter)->GetNumber() == CurrTCFGBlockNum) { // found it
				FTPredTCFGBlockNum = PredBlock->GetNumber();
				break;
			}
		}
	}
	return FTPredTCFGBlockNum;
} // end of SMPFunction::GetFallThroughPredBlockNum()

void SMPFunction::SetMarkerInstDefs(void) {
	// Create DEFs in the marker instruction for all names in the LiveInSet
	//  of the first block. These are the names for the function that
	//  would otherwise look like USEs of uninitialized variables later.
	// Note that the LiveVariableAnalysis work does not actually populate
	//  a LiveInSet for the first block, so we simulate it with its
	//  dataflow equation, UpExposed union (LiveOut minus VarKill).
	STARSOpndSetIter UpExposedIter, LiveOutIter;
	list<SMPInstr *>::iterator MarkerInst = this->Instrs.begin();
	SMPBasicBlock *FirstBlock = this->Blocks.front();
	for (UpExposedIter = FirstBlock->GetFirstUpExposed();
		UpExposedIter != FirstBlock->GetLastUpExposed();
		++UpExposedIter) {
		// Add DEF with SSANum of 0.
		(*MarkerInst)->AddDef(*UpExposedIter, UNINIT, 0);
		// Add to the VarKill and LiveIn sets.
		FirstBlock->AddVarKill(*UpExposedIter); // "killed" by marker inst def
		FirstBlock->AddLiveIn(*UpExposedIter);
	}
	for (LiveOutIter = FirstBlock->GetFirstLiveOut();
		LiveOutIter != FirstBlock->GetLastLiveOut();
		++LiveOutIter) {
		if (!(FirstBlock->IsVarKill(*LiveOutIter))) {
			// Add DEF with SSANum of 0.
			(*MarkerInst)->AddDef(*LiveOutIter, UNINIT, 0);
			// Add to the VarKill and LiveIn sets.
			FirstBlock->AddVarKill(*LiveOutIter); // "killed" by marker inst def
			FirstBlock->AddLiveIn(*LiveOutIter);
		}
	}
}

// Perform live variable analysis on all blocks in the function.
// See chapter 9 of Cooper/Torczon, Engineering a Compiler, 1st edition, for the algorithm.
void SMPFunction::LiveVariableAnalysis(bool Recomputing) {
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
#if SMP_DEBUG_DATAFLOW
	bool DebugFlag = (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
#endif
#if SMP_DEBUG_DATAFLOW_VERBOSE
	SMP_msg("LiveVariableAnalysis for %s\n", this->GetFuncName());
#endif

	if (Recomputing) {
		for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
			CurrBlock = (*BlockIter);

			// Clear out old Phi functions before they are recomputed later.
			CurrBlock->ClearPhiFunctions();

			// Initialize the Killed and UpwardExposed sets for each block.
			CurrBlock->InitKilledExposed(this->UsesFramePointer(), false);

			// Clear the loop bits for each block.
			CurrBlock->ResetLoopHeaderBlock();
			CurrBlock->ResetLoopTailBlock();
		}
	}

	bool changed;
	// Iterate over each block, updating LiveOut sets until no more changes are made.
	// NOTE: LVA is more efficient when computed over a reverse post-order list of blocks
	//  from the inverted CFG. We have an RPO list from the forward CFG, so it is just as
	//  good to simply iterate through the blocks in layout order.
#if 1
	do {
		changed = false;
		for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
			CurrBlock = (*BlockIter);
			changed |= CurrBlock->UpdateLiveOut(false);
		}
	} while (changed);
#else // Use reverse postorder
	do {
		changed = false;
		for (std::size_t index = 0; index < this->RPOBlocks.size(); ++index) {
			CurrBlock = this->RPOBlocks[index];
			changed |= CurrBlock->UpdateLiveOut(false);
		}
	} while (changed);
#endif

	this->SetMarkerInstDefs();

#if SMP_DEBUG_DATAFLOW_VERBOSE
	if (DebugFlag) SMP_msg("Exiting LiveVariableAnalysis\n");
#endif
	return;
} // end of SMPFunction::LiveVariableAnalysis()

// After normalization, re-do stack ops in LVA sets.
void SMPFunction::RecomputeStackLVA(void) {
	// Step 1: Erase old stack ops in all LVA sets.
	bool StackOpsRemoved = false;
	bool UseFP = this->UsesFramePointer();

	list<SMPBasicBlock *>::iterator BlockIter;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		StackOpsRemoved = ((*BlockIter)->RemoveStackOpsFromLVASets() || StackOpsRemoved);
	}

	if (!StackOpsRemoved)
		return;

	// Step 2: Re-analyze LVA VarKill, UpExposed, and LiveOut stack ops.
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		(*BlockIter)->InitKilledExposed(UseFP, true);
	}
	bool changed = true;
	while (changed) {
		bool ChangedOnThisIter = false;
		for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
			ChangedOnThisIter = ((*BlockIter)->UpdateLiveOut(true) || ChangedOnThisIter);
		}
		changed = ChangedOnThisIter;
	}

	// Step 3: Redo the stack ops in the SSA marker rung.
	SMPBasicBlock *FirstBlock = this->GetBlockByNum(0);
	vector<SMPInstr *>::iterator MarkerIter = FirstBlock->GetFirstInst();
	SMPInstr *MarkerInst = (*MarkerIter);
	assert(MarkerInst->IsMarkerInst());
	MarkerInst->ClearDefs();
	this->SetMarkerInstDefs();

	// Step 4: Redo the LiveIn sets.
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		(*BlockIter)->ComputeLiveInSet(true);
	}

	return;
} // SMPFunction::RecomputeStackLVA()


// Return the IDom index that is the end of the intersection prefix of the Dom sets of
//  the two blocks designated by the RPO numbers passed in.
// See Cooper & Torczon, "Engineering a Compiler" 1st edition figure 9.8.
int SMPFunction::IntersectDoms(int block1, int block2) const {
	int finger1 = block1;
	int finger2 = block2;
	while (finger1 != finger2) {
		while (finger1 > finger2)
			finger1 = this->IDom[(size_t)finger1];
		while (finger2 > finger1)
			finger2 = this->IDom[(size_t)finger2];
	}
	return finger1;
} // end of SMPFunction::IntersectDoms()

// Find Dom intersection (as TCFGIDom[] index) for 2 TCFG blocks
int SMPFunction::IntersectTCFGDoms(int block1, int block2) const {
	assert(((size_t)block1) < this->TCFGIDom.size());
	assert(((size_t)block2) < this->TCFGIDom.size());
	int finger1 = block1;
	int finger2 = block2;
	while (finger1 != finger2) {
		while (finger1 > finger2)
			finger1 = this->TCFGIDom[(size_t)finger1];
		while (finger2 > finger1)
			finger2 = this->TCFGIDom[(size_t)finger2];
	}
	return finger1;
} // end of SMPFunction::IntersectTCFGDoms()

// Compute immediate dominators of all blocks into IDom[] vector.
void SMPFunction::ComputeIDoms(void) {
	bool DebugFlag = false;
#if SMP_DEBUG_DATAFLOW
	DebugFlag = (0 == strcmp("_ZN6soplex7NameSetC2Eiidd", this->GetFuncName()));
	if (DebugFlag) SMP_msg("Entered ComputeIDoms\n");
#endif
	// Initialize the IDom[] vector to uninitialized values for all blocks.
	this->IDom.reserve(this->BlockCount);
	this->IDom.assign(this->BlockCount, SMP_BLOCKNUM_UNINIT);
	if (DebugFlag) {
		SMP_msg("BlockCount = %d RPOBlocks size = %zu\n", this->BlockCount, this->RPOBlocks.size());
	}
	if (this->BlockCount != ((int) this->RPOBlocks.size())) {
		SMP_msg("SERIOUS WARNING: Function %s has BlockCount of %d and RPOBlocks size of %zu\n",
			this->GetFuncName(), this->BlockCount, this->RPOBlocks.size());
	}
	this->IDom[0] = 0; // First block is dominated only by itself
	bool changed;
	do {
		changed = false;
		for (std::size_t RPONum = 1; RPONum < (std::size_t) this->BlockCount; ++RPONum) {
			if (DebugFlag) SMP_msg("RPONum %zu\n", RPONum);
#if 0
			if (DebugFlag) {
				SMP_msg("RPOBlocks vector size: %d\n", this->RPOBlocks.size());
				for (std::size_t index = 0; index < this->RPOBlocks.size(); ++index) {
					SMP_msg("RPOBlocks entry %d is %d\n", index, RPOBlocks[index]->GetNumber());
				}
			}
#endif
			// To avoid infinite loops on blocks that dominate themselves but otherwise have no 
			//  predecessors (probably reachable only through indirect jumps), we stop processing
			//  the blocks once the IDom becomes the top (entry) block. This probably saves time
			//  on other blocks as well.
			if (0 == this->IDom[RPONum])
				continue;

			SMPBasicBlock *CurrBlock = this->RPOBlocks[RPONum];
			// if (DebugFlag) SMP_msg("CurrBlock: %x\n", CurrBlock._Ptr);
			list<SMPBasicBlock *>::iterator CurrPred;
			// Initialize NewIdom to the first processed predecessor of block RPONum.
			int NewIdom = SMP_BLOCKNUM_UNINIT;
			for (CurrPred = CurrBlock->GetFirstPred(); CurrPred != CurrBlock->GetLastPred(); ++CurrPred) {
				int PredNum = (*CurrPred)->GetNumber();
				if (DebugFlag) SMP_msg("Pred: %d\n", PredNum);
				// **!!** See comment below about unreachable blocks.
				if (SMP_BLOCKNUM_UNINIT == PredNum)
					continue;
				int PredIDOM = this->IDom[(size_t)PredNum];
				if (DebugFlag) SMP_msg("Pred IDom: %d\n", PredIDOM);
				if (SMP_BLOCKNUM_UNINIT != PredIDOM) {
					NewIdom = PredNum;
					break;
				}
			}
			if (NewIdom == SMP_BLOCKNUM_UNINIT) {
				SMP_msg("WARNING: Failure on NewIdom in ComputeIDoms for %s\n", this->GetFuncName());
				if (this->HasIndirectJumps() || this->HasIndirectCalls()) {
					// Might be reachable only through indirect jumps.
					NewIdom = 0; // make it dominated by entry block
					SMP_msg("WARNING: Assuming block %d at address %llx is reachable indirectly.\n",
						CurrBlock->GetNumber(), (uint64_t) CurrBlock->GetFirstAddr());
				}
				else {
					// Might be exception handling code, reachable only by call stack walking.
					NewIdom = 0; // make it be dominated by entry block
					SMP_msg("WARNING: Assuming block %d at address %llx is reachable by exception handling.\n",
						CurrBlock->GetNumber(), (uint64_t) CurrBlock->GetFirstAddr());
				}
			}
			assert(NewIdom != SMP_BLOCKNUM_UNINIT);
			// Loop through all predecessors of block RPONum except block NewIdom.
			//  Set NewIdom to the intersection of its Dom set and the Doms set of
			//  each predecessor that has had its Doms set computed.
			for (CurrPred = CurrBlock->GetFirstPred(); CurrPred != CurrBlock->GetLastPred(); ++CurrPred) {
				int PredNum = (*CurrPred)->GetNumber();
				if (DebugFlag) SMP_msg("PredNum: %d\n", PredNum);
				// **!!** We can avoid failure on unreachable basic blocks
				//  by executing a continue statement if PredNum is -1. Long term solution
				//  is to prune out unreachable basic blocks, or better yet, create hell nodes
				//  if the function has indirect jumps.
				if (PredNum == SMP_BLOCKNUM_UNINIT)
					continue;
				int PredIDOM = this->IDom[PredNum];
				if (DebugFlag) SMP_msg("PredIDOM: %d\n", PredIDOM);
				if ((SMP_BLOCKNUM_UNINIT == PredIDOM) || (NewIdom == PredIDOM)) {
					// Skip predecessors that have uncomputed Dom sets, or are the
					//  current NewIdom.
					continue;
				}
				if (DebugFlag) SMP_msg("Old NewIdom value: %d\n", NewIdom);
				NewIdom = this->IntersectDoms(PredNum, NewIdom);
				if (DebugFlag) SMP_msg("New NewIdom value: %d\n", NewIdom);
			}
			// If NewIdom is not the value currently in vector IDom[], update the
			//  vector entry and set changed to true.
			if (NewIdom != this->IDom[RPONum]) {
				if (DebugFlag) SMP_msg("IDOM changed from %d to %d\n", this->IDom[RPONum], NewIdom);
				this->IDom[RPONum] = NewIdom;
				changed = true;
			}
		}
	} while (changed);
	return;
} // end of SMPFunction::ComputeIDoms()

// Compute immediate dominators of all TCFG blocks into TCFGIDom[]
void SMPFunction::ComputeTCFGIDoms(void) {
	// Initialize the TCFGIDom[] vector to uninitialized values for all blocks.
	size_t TCFGBlockCount = this->TCFGBlocks.size();
	this->TCFGIDom.reserve(TCFGBlockCount);
	this->TCFGIDom.assign(TCFGBlockCount, SMP_BLOCKNUM_UNINIT);
	this->TCFGIDom[0] = 0; // First block is dominated only by itself
	bool changed;
	do {
		changed = false;
		for (std::size_t RPONum = 1; RPONum < TCFGBlockCount; ++RPONum) {
			// To avoid infinite loops on blocks that dominate themselves but otherwise have no 
			//  predecessors (probably reachable only through indirect jumps), we stop processing
			//  the blocks once the IDom becomes the top (entry) block. This probably saves time
			//  on other blocks as well.
			if (0 == this->TCFGIDom[RPONum])
				continue;

			STARSTransCFGBlock *CurrTCFGBlock = this->TCFGBlocks[RPONum];
			// Initialize NewIdom to the first processed predecessor of block RPONum.
			int NewIdom = SMP_BLOCKNUM_UNINIT;
			for (STARSTransCFGBlock *CurrPred : CurrTCFGBlock->GetConstPredList()) {
				int PredNum = CurrPred->GetNumber();
				// **!!** See comment below about unreachable blocks.
				if (SMP_BLOCKNUM_UNINIT == PredNum)
					continue;
				int PredIDOM = this->TCFGIDom[(size_t)PredNum];
				if (SMP_BLOCKNUM_UNINIT != PredIDOM) {
					NewIdom = PredNum;
					break;
				}
			}
			if (NewIdom == SMP_BLOCKNUM_UNINIT) {
				SMP_msg("WARNING: Failure on NewIdom in ComputeTCFGIDoms ");
				this->DumpFuncNameAndAddr();
				if (this->HasIndirectJumps() || this->HasIndirectCalls()) {
					// Might be reachable only through indirect jumps.
					NewIdom = 0; // make it dominated by entry block
					SMP_msg("WARNING: Assuming TCFG block %d is reachable indirectly.\n", CurrTCFGBlock->GetNumber());
				}
				else {
					// Might be exception handling code, reachable only by call stack walking.
					NewIdom = 0; // make it be dominated by entry block
					SMP_msg("WARNING: Assuming block %d is reachable by exception handling.\n", CurrTCFGBlock->GetNumber());
				}
			}
			assert(NewIdom != SMP_BLOCKNUM_UNINIT);
			// Loop through all predecessors of TCFG block RPONum except block NewIdom.
			//  Set NewIdom to the intersection of its Dom set and the Doms set of
			//  each predecessor that has had its Doms set computed.
			for (STARSTransCFGBlock *CurrPred : CurrTCFGBlock->GetConstPredList()) {
				int PredNum = CurrPred->GetNumber();
				// **!!** We can avoid failure on unreachable basic blocks
				//  by executing a continue statement if PredNum is -1. Long term solution
				//  is to prune out unreachable basic blocks, or better yet, create hell nodes
				//  if the function has indirect jumps.
				if (PredNum == SMP_BLOCKNUM_UNINIT)
					continue;
				int PredIDOM = this->TCFGIDom[PredNum];
				if ((SMP_BLOCKNUM_UNINIT == PredIDOM) || (NewIdom == PredIDOM)) {
					// Skip predecessors that have uncomputed Dom sets, or are the
					//  current NewIdom.
					continue;
				}
				NewIdom = this->IntersectTCFGDoms(PredNum, NewIdom);
			}
			// If NewIdom is not the value currently in vector IDom[], update the
			//  vector entry and set changed to true.
			if (NewIdom != this->TCFGIDom[RPONum]) {
				this->TCFGIDom[RPONum] = NewIdom;
				changed = true;
			}
		}
	} while (changed);
	return;
} // end of SMPFunction::ComputeTCFGIDoms()

// Compute dominance frontier sets for each block.
void SMPFunction::ComputeDomFrontiers(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *RunnerBlock;
	SMPBasicBlock *CurrBlock;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		// We look only at join points in the CFG, as per Cooper/Torczon chapter 9.
		if (1 < CurrBlock->GetNumPreds()) { // join point; more than 1 predecessor
			int runner;
			list<SMPBasicBlock *>::iterator PredIter;
			SMPBasicBlock *CurrPred;
			for (PredIter = CurrBlock->GetFirstPred(); PredIter != CurrBlock->GetLastPred(); ++PredIter) {
				CurrPred = (*PredIter);
				// For each predecessor, we run up the IDom[] vector and add CurrBlock to the
				//  DomFrontier for all blocks that are between CurrPred and IDom[CurrBlock],
				//  not including IDom[CurrBlock] itself.
				runner = CurrPred->GetNumber();
				while (runner != this->IDom[(size_t)CurrBlock->GetNumber()]) {
					// Cooper/Harvey/Kennedy paper does not quite agree with the later
					//  text by Cooper/Torczon. Text (1st ed.) says that the start node has no IDom
					//  in the example on pages 462-463, but it shows an IDOM for the
					//  root node in Figure 9.9 of value == itself. The first edition text
					//  on p.463 seems correct, as the start node dominates every node and
					//  thus should have no dominance frontier.
					if (SMP_TOP_BLOCK == runner)
						break;
					RunnerBlock = this->RPOBlocks[(size_t)runner];
					RunnerBlock->AddToDomFrontier(CurrBlock->GetNumber());
					runner = this->IDom[(size_t)runner];
				}
			} // end for all predecessors
		} // end if join point
	} // end for all blocks
	return;
} // end of SMPFunction::ComputeDomFrontiers()

// Compute dominance frontiers for all TCFG blocks
void SMPFunction::ComputeTCFGDomFrontiers(void) {
	STARSTransCFGBlock *RunnerBlock;	
	for (STARSTransCFGBlock *CurrTCFGBlock : this->TCFGBlocks) {
		// We look only at join points in the CFG, as per Cooper/Torczon chapter 9.
		if (1 < CurrTCFGBlock->GetNumPreds()) { // join point; more than 1 predecessor
			int runner;
			int CurrTCFGBlockNum = CurrTCFGBlock->GetNumber();
			assert(((size_t)CurrTCFGBlockNum) < this->TCFGIDom.size());
			int CurrIDom = this->TCFGIDom[(size_t)CurrTCFGBlockNum];
			for (STARSTransCFGBlock *CurrPred : CurrTCFGBlock->GetConstPredList()) {
				// For each predecessor, we run up the TCFGIDom[] vector and add CurrBlock (a.k.a. runner) to the
				//  DomFrontier for all blocks that are between CurrPred and TCFGIDom[CurrTCFGBlock] (a.k.a. CurrIDom),
				//  not including TCFGIDom[CurrTCFGBlock] itself.
				runner = CurrPred->GetNumber();
				assert(((size_t)runner) < this->TCFGBlocks.size());
				while (runner != CurrIDom) {
					// Cooper/Harvey/Kennedy paper does not quite agree with the later
					//  text by Cooper/Torczon. Text (1st ed.) says that the start node has no IDom
					//  in the example on pages 462-463, but it shows an IDOM for the
					//  root node in Figure 9.9 of value == itself. The first edition text
					//  on p.463 seems correct, as the start node dominates every node and
					//  thus should have no dominance frontier.
					if (SMP_TOP_BLOCK == runner)
						break;
					RunnerBlock = this->TCFGBlocks[(size_t)runner];
					RunnerBlock->AddToDomFrontier(CurrTCFGBlockNum);
					runner = this->TCFGIDom[(size_t)runner];
				}
			} // end for all predecessors
		} // end if join point
	} // end for all blocks
	return;
} // end of SMPFunction::ComputeTCFGDomFrontiers()

// Compute the GlobalNames set, which includes all operands that are used in more than
//  one basic block. It is the union of all UpExposedSets of all blocks.
void SMPFunction::ComputeGlobalNames(void) {
	STARSOpndSetIter SetIter;
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	unsigned int index = 0;
	this->GlobalNames.clear();
	if (this->Blocks.size() < 2)
		return; // cannot have global names if there is only one block

#if SMP_DEBUG_DATAFLOW
	bool DebugFlag = false;
	DebugFlag = (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
#endif

	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		for (SetIter = CurrBlock->GetFirstUpExposed(); SetIter != CurrBlock->GetLastUpExposed(); ++SetIter) {
			STARSOpndTypePtr TempOp = *SetIter;
			// The GlobalNames set will have the complete collection of operands that we are
			//  going to number in our SSA computations. We now assign an operand number
			//  within the STARSOpndType structure for each, so that we can index into the
			//  BlocksUsedIn[] vector, for example. This operand number is not to be
			//  confused with SSA numbers.
			// We use the operand number field op_t.n for the lower 8 bits, and the offset
			//  fields op_t.offb:op_t.offo for the upper 16 bits. We are overwriting IDA
			//  values here, but operands in the data flow analysis sets should never be
			//  inserted back into the IDA database anyway.
			SetGlobalIndex(TempOp, (std::size_t) index);

#if SMP_DEBUG_DATAFLOW
			if (DebugFlag) {
				SMP_msg("Global Name: ");
				PrintListOperand(TempOp);
			}
#endif
			pair<STARSOpndSetIter, bool> InsertResult;
			InsertResult = this->GlobalNames.insert(TempOp);
			if (!InsertResult.second) {
				// Already in GlobalNames, so don't assign an index number.
				;
#if SMP_DEBUG_DATAFLOW
				if (DebugFlag) {
					SMP_msg(" already in GlobalNames.\n");
				}
#endif
			}
			else {
				++index;
#if SMP_DEBUG_DATAFLOW
				if (DebugFlag) {
					SMP_msg(" inserted as index %d\n", ExtractGlobalIndex(TempOp));
				}
#endif
			}
		} // for each upward exposed item in the current block
	} // for each basic block

	SMP_msg("INFO: GlobalNames size is %zu for func at %llx\n", this->GlobalNames.size(),
		(uint64_t) this->GetFirstFuncAddr());
	assert(16777215 >= this->GlobalNames.size()); // index fits in 24 bits
	return;
} // end of SMPFunction::ComputeGlobalNames()

// For each item in GlobalNames, record the blocks that DEF the item.
void SMPFunction::ComputeBlocksDefinedIn(void) {
	// Loop through all basic blocks and examine all DEFs. For Global DEFs, record
	//  the block number in BlocksDefinedIn. The VarKill set records DEFs without
	//  having to examine every instruction.
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;

	this->BlocksDefinedIn.clear();
	for (std::size_t i = 0; i < this->GlobalNames.size(); ++i) {
		list<int> TempList;
		this->BlocksDefinedIn.push_back(TempList);
	}
#if SMP_DEBUG_DATAFLOW_VERBOSE
	SMP_msg("Number of GlobalNames: %d\n", this->GlobalNames.size());
	SMP_msg("Size of BlocksDefinedIn: %d\n", this->BlocksDefinedIn.size());
#endif
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		STARSOpndSetIter KillIter;
		CurrBlock = (*BlockIter);
		for (KillIter = CurrBlock->GetFirstVarKill(); KillIter != CurrBlock->GetLastVarKill(); ++KillIter) {
			// If killed item is not a block-local item (it is global), record it.
			STARSOpndTypePtr KillOp = (*KillIter);
			STARSOpndSetIter NameIter = this->GlobalNames.find(KillOp);
			if (NameIter != this->GlobalNames.end()) { // found in GlobalNames set
				// We have a kill of a global name. Get index from three 8-bit fields.
				unsigned int index = ExtractGlobalIndex(*NameIter);
				if (index >= this->GlobalNames.size()) {
					// We are about to assert false.
					SMP_msg("ComputeBlocksDefinedIn: Bad index: %u limit: %zu\n", index,
						this->GlobalNames.size());
					SMP_msg("Block number %d\n", CurrBlock->GetNumber());
					SMP_msg("Killed item: ");
					PrintListOperand(*KillIter);
					SMP_msg("\n");
					SMP_msg("This is a fatal error.\n");
				}
				assert(index < this->GlobalNames.size());
				// index is a valid subscript for the BlocksDefinedIn vector. Push the
				//  current block number onto the list of blocks that define this global name.
				this->BlocksDefinedIn[index].push_back(CurrBlock->GetNumber());
				(*KillIter)->SetOpGlobalIndex(index);
			}			
		} // end for all VarKill operands
		// Set the GlobalNameIndex into each LiveOutSet item.
		STARSOpndSetIter LiveOutIter;
		for (LiveOutIter = CurrBlock->GetFirstLiveOut(); LiveOutIter != CurrBlock->GetLastLiveOut(); ++LiveOutIter) {
			STARSOpndTypePtr LiveOutOp = (*LiveOutIter);
			STARSOpndSetIter NameIter = this->GlobalNames.find(LiveOutOp);
			assert(NameIter != this->GlobalNames.end()); // found in GlobalNames set
			unsigned int index = ExtractGlobalIndex(*NameIter);
			(*LiveOutIter)->SetOpGlobalIndex(index);
		}
	} // end for all blocks
	return;
} // end of SMPFunction::ComputeBlocksDefinedIn()

// Compute the phi functions at the entry point of each basic block that is a join point.
void SMPFunction::InsertPhiFunctions(void) {
	STARSOpndSetIter NameIter;
	list<int> WorkList;  // list of block numbers
	bool DebugFlag = false;
#if SMP_DEBUG_DATAFLOW
	DebugFlag = (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
#endif
	if (DebugFlag) SMP_msg("GlobalNames size: %zu\n", this->GlobalNames.size());
	for (NameIter = this->GlobalNames.begin(); NameIter != this->GlobalNames.end(); ++NameIter) {
		int CurrNameIndex = (int) (ExtractGlobalIndex(*NameIter));
		if (DebugFlag) SMP_msg("CurrNameIndex: %d\n", CurrNameIndex);
#if 0
		DebugFlag = (DebugFlag && (6 == CurrNameIndex));
#endif
		// Initialize the work list to all blocks that define the current name.
		WorkList.clear();
		for (int BlockNum :this->BlocksDefinedIn[(size_t) CurrNameIndex]) {
			WorkList.push_back(BlockNum);
		}

		// Iterate through the work list, inserting phi functions for the current name
		//  into all the blocks in the dominance frontier of each work list block.
		//  Insert into the work list each block that had a phi function added.
		while (!WorkList.empty()) {
#if SMP_DEBUG_DATAFLOW_VERBOSE
			if (DebugFlag) SMP_msg("WorkList size: %d\n", WorkList.size());
#endif
			list<int>::iterator WorkIter = WorkList.begin();
			while (WorkIter != WorkList.end()) {
				set<int>::iterator DomFrontIter;
#if SMP_DEBUG_DATAFLOW_VERBOSE
				if (DebugFlag) SMP_msg("WorkIter: %d\n", *WorkIter);
#endif
				if (DebugFlag && (*WorkIter > this->BlockCount)) {
					SMP_msg("ERROR: WorkList block # %d out of range ", *WorkIter);
					this->DumpFuncNameAndAddr();
				}
				SMPBasicBlock *WorkBlock = this->RPOBlocks[*WorkIter];
				for (DomFrontIter = WorkBlock->GetFirstDomFrontier();
					DomFrontIter != WorkBlock->GetLastDomFrontier();
					++DomFrontIter) {
#if SMP_DEBUG_DATAFLOW_VERBOSE
					if (DebugFlag) SMP_msg("DomFront: %d\n", *DomFrontIter);
#endif
					if (DebugFlag && (*DomFrontIter > this->BlockCount)) {
						SMP_msg("ERROR: DomFront block # %d out of range ", *DomFrontIter);
						this->DumpFuncNameAndAddr();
					}
					SMPBasicBlock *PhiBlock = this->RPOBlocks[*DomFrontIter];
					// Before inserting a phi function for the current name in *PhiBlock,
					//  see if the current name is LiveIn for *PhiBlock. If not, there
					//  is no need for the phi function. This check is what makes the SSA
					//  a fully pruned SSA.
					if (PhiBlock->IsLiveIn(*NameIter)) {
						std::size_t NumPreds = PhiBlock->GetNumPreds();
						DefOrUse CurrRef(*NameIter);
						SMPPhiFunction CurrPhi(CurrNameIndex, CurrRef);
						for (std::size_t NumCopies = 0; NumCopies < NumPreds; ++NumCopies) {
							CurrPhi.PushBack(CurrRef); // inputs to phi
						}
						if (PhiBlock->AddPhi(CurrPhi)) {
							// If not already in Phi set, new phi function was inserted.
							WorkList.push_back(PhiBlock->GetNumber());
#if SMP_DEBUG_DATAFLOW_VERBOSE
							if (DebugFlag) SMP_msg("Added phi for name %d at top of block %d\n", CurrNameIndex, PhiBlock->GetNumber());
#endif
						}
					}
					else {
						if (DebugFlag) {
							SMP_msg("Global %d not LiveIn for block %d\n", CurrNameIndex, PhiBlock->GetNumber());
						}
					}
				} // end for all blocks in the dominance frontier
				// Remove current block number from the work list
				if (DebugFlag) {
					SMP_msg("Removing block %d from work list.\n", *WorkIter);
				}
				WorkIter = WorkList.erase(WorkIter);
			} // end for all block numbers in the work list
		} // end while the work list is not empty
		if (DebugFlag) SMP_msg("WorkList empty.\n");
	} // end for all elements of the GlobalNames set
	return;
} // end of SMPFunction::InsertPhiFunctions()

// Build the dominator tree.
void SMPFunction::BuildDominatorTree(void) {
	std::size_t index;
	// First, fill the DomTree vector with the parent numbers filled in and the child lists
	//  left empty.
	for (index = 0; index < this->IDom.size(); ++index) {
		pair<int, list<int> > DomTreeEntry;
		DomTreeEntry.first = this->IDom[index];
		DomTreeEntry.second.clear();
		this->DomTree.push_back(DomTreeEntry);
	}
	// Now, push the children onto the appropriate lists.
	for (index = 0; index < this->IDom.size(); ++index) {
		// E.g. if block 5 has block 3 as a parent, then we fetch the number 3
		//  using the expression this->DomTree[index].first, which was just
		//  initialized in the previous loop. Then we go to DomTree entry 3 and push
		//  the number 5 on its child list.
		int parent = this->DomTree[index].first;
		if (parent != (int) index) // block can dominate itself, but not in DomTree!
			this->DomTree[(size_t)parent].second.push_back((int) index);
	}

	return;
} // end of SMPFunction::BuildDominatorTree()

// Build the TCFGDomTree structure
void SMPFunction::BuildTCFGDominatorTree(void) {
	std::size_t index;
	// First, fill the TCFGDomTree vector with the parent numbers filled in and the child lists
	//  left empty.
	for (index = 0; index < this->TCFGIDom.size(); ++index) {
		pair<int, list<int> > DomTreeEntry;
		DomTreeEntry.first = this->TCFGIDom[index];
		DomTreeEntry.second.clear();
		this->TCFGDomTree.push_back(DomTreeEntry);
	}
	// Now, push the children onto the appropriate lists.
	for (index = 0; index < this->TCFGIDom.size(); ++index) {
		// E.g. if block 5 has block 3 as a parent, then we fetch the number 3
		//  using the expression this->TCFGDomTree[index].first, which was just
		//  initialized in the previous loop. Then we go to TCFGDomTree entry 3 and push
		//  the number 5 on its child list.
		int parent = this->TCFGDomTree[index].first;
		if (parent != (int)index) // block can dominate itself, but not in DomTree!
			this->TCFGDomTree[(size_t)parent].second.push_back((int)index);
	}

	return;
} // end of SMPFunction::BuildTCFGDominatorTree()

// Does basic block HeadBlockNum dominate basic block TailBlockNum?
bool SMPFunction::DoesBlockDominateBlock(const int HeadBlockNum, const int TailBlockNum) {
	if (HeadBlockNum == TailBlockNum)
		return true;
	else if ((HeadBlockNum < 0) || (TailBlockNum < 0)) {
		return false;
	}
	else if (HeadBlockNum > TailBlockNum) { // RPO numbering ensures no domination
		return false;
	}
	else {
		// Recurse downward from HeadBlockNum in the dominator tree until we find TailBlockNum
		//  or return false if we never find it.
		bool FoundIt = false;
		assert(((size_t)HeadBlockNum) < this->DomTree.size());
		assert(((size_t)TailBlockNum) < this->DomTree.size());

		// Memoization optimization: If all Predecessors of TailBlockNum have already been found to be dominated
		//  by HeadBlockNum, then TailBlockNum is dominated by HeadBlockNum.
		SMPBasicBlock *TailBlock = this->GetBlockByNum((size_t)TailBlockNum);
		bool AllPredsDominated = true;
		// NOTE: Need to pass the pred list in, because when we walk the DomTree we keep
		//  getting the same pred list, as TailBlockNum remains constant.
		for (SMPBasicBlock *PredBlock : TailBlock->GetConstPredList()) {
			int PredBlockNum = PredBlock->GetNumber();
			if (!this->DomSoFar.GetBit((size_t)PredBlockNum)) {
				AllPredsDominated = false;
				break;
			}
		}
		if (AllPredsDominated) {
			this->DomSoFar.SetBit((size_t)TailBlockNum);
			return true;
		}

		for (int ChildBlockNum : this->DomTree[HeadBlockNum].second) {
			this->DomSoFar.SetBit((size_t)ChildBlockNum); // memoize everything encountered in DomTree walk
			assert(((size_t)ChildBlockNum) < this->DomTree.size());
			if (this->DoesBlockDominateBlock(ChildBlockNum, TailBlockNum)) { // recurse depth-first
				FoundIt = true;
				this->DomSoFar.SetBit((size_t)TailBlockNum);
				break;
			}
		}
		return FoundIt;
	}
} // end of SMPFunction::DoesBlockDominateBlock()

// Does TCFG block HeadBlockNum dominate TCFG block TailBlockNum?
bool SMPFunction::DoesTCFGBlockDominateTCFGBlock(const int HeadBlockNum, const int TailBlockNum) {
	if (HeadBlockNum == TailBlockNum)
		return true;
	else if ((HeadBlockNum < 0) || (TailBlockNum < 0)) {
		return false;
	}
	else if (HeadBlockNum > TailBlockNum) { // RPO numbering ensures no domination
		return false;
	}
	else {
		// Recurse downward from HeadBlockNum in the dominator tree until we find TailBlockNum
		//  or return false if we never find it.
		bool FoundIt = false;
		assert(((size_t)HeadBlockNum) < this->TCFGDomTree.size());
		assert(((size_t)TailBlockNum) < this->TCFGDomTree.size());

		// Memoization optimization: If all Predecessors of TailBlockNum have already been found to be dominated
		//  by HeadBlockNum, then TailBlockNum is dominated by HeadBlockNum.
		STARSTransCFGBlock *TailBlock = this->GetTCFGBlockByNum((size_t)TailBlockNum);
		bool AllPredsDominated = true;
		// NOTE: Need to pass the pred list in, because when we walk the DomTree we keep
		//  getting the same pred list, as TailBlockNum remains constant.
		for (STARSTransCFGBlock *PredBlock : TailBlock->GetConstPredList()) {
			int PredBlockNum = PredBlock->GetNumber();
			if (!this->TCFGDomSoFar.GetBit((size_t)PredBlockNum)) {
				AllPredsDominated = false;
				break;
			}
		}
		if (AllPredsDominated) {
			this->TCFGDomSoFar.SetBit((size_t)TailBlockNum);
			return true;
		}

		for (int ChildBlockNum : this->TCFGDomTree[HeadBlockNum].second) {
			this->TCFGDomSoFar.SetBit((size_t)ChildBlockNum); // memoize everything encountered in DomTree walk
			assert(((size_t)ChildBlockNum) < this->TCFGDomTree.size());
			if (this->DoesTCFGBlockDominateTCFGBlock(ChildBlockNum, TailBlockNum)) { // recurse depth-first
				FoundIt = true;
				this->TCFGDomSoFar.SetBit((size_t)TailBlockNum);
				break;
			}
		}
		return FoundIt;
	}
} // end of SMPFunction::DoesTCFGBlockDominateTCFGBlock()

#define STARS_DFS_UNVISITED -1
#define STARS_DFS_IN_PROGRESS 0
#define STARS_DFS_COMPLETED 1

// return true if CFG is a reducible graph.
bool SMPFunction::TestCFGReducibility(void) {
	this->DFSMarkers.clear();
	for (std::size_t BlockNum = 0; BlockNum < (std::size_t) this->BlockCount; ++BlockNum) {
		this->DFSMarkers.push_back(STARS_DFS_UNVISITED);
	}

	bool Reducible = this->CFGReducibilityHelper(SMP_TOP_BLOCK);
	this->DFSMarkers.clear();

	return Reducible;
} // end of SMPFunction::TestCFGReducibility()

// recursive depth-first-search helper for TestCFGReducibility()
bool SMPFunction::CFGReducibilityHelper(std::size_t BlockNumber) {
	bool Reducible = true;

	this->DFSMarkers[BlockNumber] = STARS_DFS_IN_PROGRESS;
	SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockNumber];
	for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
		std::size_t SuccBlockNum = (size_t) SuccBlock->GetNumber();
		if (SuccBlockNum < this->DFSMarkers.size()) {
			int DFSStatus = this->DFSMarkers[SuccBlockNum];
			if (DFSStatus == STARS_DFS_UNVISITED) {
				Reducible = this->CFGReducibilityHelper(SuccBlockNum);
			}
			else if (DFSStatus == STARS_DFS_IN_PROGRESS) {
				// back edge; reducibility depends on dominator relationship
				this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				if (this->DoesBlockDominateBlock((int)BlockNumber, (int)SuccBlockNum)) {
					// forward edge; no problem
					;
				}
				else {
					this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
					if (this->DoesBlockDominateBlock((int)SuccBlockNum, (int)BlockNumber)) {
						// reducible back edge; no problem
						;
					}
					else { // neither block dominates the other; irreducible back edge
						Reducible = false;
						SMP_msg("SERIOUS WARNING: Function is irreducible because of back edge from block at %llx to %llx ",
							(uint64_t)SuccBlock->GetFirstAddr(), (uint64_t)CurrBlock->GetFirstAddr());
						this->DumpFuncNameAndAddr();
					}
				}
			}
			else {
				assert(DFSStatus == STARS_DFS_COMPLETED);
				// forward edge or cross edge; no problem
			}
		}
		else {
			Reducible = false;
			SMP_msg("ERROR: Block %d out of range in CFGReducibilityHelper for parent block %u ", SuccBlock->GetNumber(), BlockNumber);
			this->DumpFuncNameAndAddr();
		}
		if (!Reducible) {
			break;
		}
	} // end for all successors

	this->DFSMarkers[BlockNumber] = STARS_DFS_COMPLETED;

	return Reducible;
} // end of SMPFunction::CFGReducibilityHelper()

bool SMPFunction::IsBlockInAnyLoop(int BlockNum) const {
	if (0 == this->LoopCount)
		return false;
	assert(((size_t) BlockNum) < this->GetNumBlocks());
	bool FoundInsideLoop = this->FuncLoopsByBlock[(size_t) BlockNum].IsAnyBitSet();
	return FoundInsideLoop;
} // end of SMPFunction::IsBlockInAnyLoop()

// Is block (with block # BlockNum) inside loop # LoopNum?
bool SMPFunction::IsBlockInLoop(int BlockNum, std::size_t LoopNum) const {
	if ((0 == this->LoopCount) || (0 > BlockNum))
		return false;
	assert(((size_t) BlockNum) < this->GetNumBlocks());
	return ((LoopNum < this->LoopCount) && (this->FuncLoopsByBlock[(size_t) BlockNum].GetBit(LoopNum)));
} // end of SMPFunction::IsBlockInLoop()

bool SMPFunction::AreBlocksInSameLoops(const int BlockNum1, const int BlockNum2) const {
	if (0 == this->LoopCount)
		return true;
	assert(((size_t) BlockNum1) < this->GetNumBlocks());
	assert(((size_t) BlockNum2) < this->GetNumBlocks());
	// Valid short cut based on how we detect and number loops?
	bool SameLoops = (this->FuncLoopsByBlock[(std::size_t) BlockNum1].FindLowestBitSet() == this->FuncLoopsByBlock[(std::size_t) BlockNum2].FindLowestBitSet())
		&& (this->FuncLoopsByBlock[(std::size_t) BlockNum1].FindHighestBitSet() == this->FuncLoopsByBlock[(std::size_t) BlockNum2].FindHighestBitSet());
	if (SameLoops) {
		bool SameCount = (this->FuncLoopsByBlock[(std::size_t) BlockNum1].CountSetBits() == this->FuncLoopsByBlock[(std::size_t) BlockNum2].CountSetBits());
		SameLoops = SameCount;
		if (!SameCount) {
			SMP_msg("INFO: LOOP: SameCount failure on loops for blocks %d and %d ", BlockNum1, BlockNum2);
			this->DumpFuncNameAndAddr();
			// this->DumpDotCFG();
		}
	}
	return SameLoops;
} // end of SMPFunction::AreBlocksInSameLoops()

bool SMPFunction::IsTCFGBlockInAnyLoop(int BlockNum) const {
	if (0 == this->LoopCount)
		return false;
	int OrigBlockNum = this->TCFGBlocks[(size_t)BlockNum]->GetOriginalBlockNum();
	assert(((size_t)OrigBlockNum) < this->GetNumBlocks());
	bool FoundInsideLoop = this->FuncLoopsByBlock[(size_t)OrigBlockNum].IsAnyBitSet();
	return FoundInsideLoop;
} // end of SMPFunction::IsTCFGBlockInAnyLoop()

// Is block (with block # BlockNum) inside loop # LoopNum?
bool SMPFunction::IsTCFGBlockInLoop(int BlockNum, std::size_t LoopNum) const {
	if ((0 == this->LoopCount) || (0 > BlockNum))
		return false;
	int OrigBlockNum = this->TCFGBlocks[(size_t)BlockNum]->GetOriginalBlockNum();
	assert(((size_t)OrigBlockNum) < this->GetNumBlocks());
	return ((LoopNum < this->LoopCount) && (this->FuncLoopsByBlock[(size_t)OrigBlockNum].GetBit(LoopNum)));
} // end of SMPFunction::IsTCFGBlockInLoop()

bool SMPFunction::AreTCFGBlocksInSameLoops(const int BlockNum1, const int BlockNum2) const {
	if (0 == this->LoopCount)
		return true;
	assert(0 <= BlockNum1);
	assert(0 <= BlockNum2);
	int OrigBlockNum1 = this->TCFGBlocks[(size_t)BlockNum1]->GetOriginalBlockNum();
	int OrigBlockNum2 = this->TCFGBlocks[(size_t)BlockNum2]->GetOriginalBlockNum();
	assert(((size_t)OrigBlockNum1) < this->GetNumBlocks());
	assert(((size_t)OrigBlockNum2) < this->GetNumBlocks());
	// Valid short cut based on how we detect and number loops?
	bool SameLoops = (this->FuncLoopsByBlock[(std::size_t) OrigBlockNum1].FindLowestBitSet() == this->FuncLoopsByBlock[(std::size_t) OrigBlockNum2].FindLowestBitSet())
		&& (this->FuncLoopsByBlock[(std::size_t) OrigBlockNum1].FindHighestBitSet() == this->FuncLoopsByBlock[(std::size_t) OrigBlockNum2].FindHighestBitSet());
	return SameLoops;
} // end of SMPFunction::AreTCFGBlocksInSameLoops()


// Does going from BlockNum1 to BlockNum2 exit a loop?
bool SMPFunction::DoesEdgeExitLoop(const int BlockNum1, const int BlockNum2) const {
	bool ExitsLoop = false;
	if ((0 == this->GetNumLoops()) || (0 > BlockNum2)) // BlockNum2 might be code like SMP_BLOCKNUM_COMMON_RETURN
		return ExitsLoop;
	assert(((size_t)BlockNum1) < this->GetNumBlocks());
	assert(((size_t)BlockNum2) < this->GetNumBlocks());
	for (size_t LoopIndex = 0; LoopIndex < this->GetNumLoops(); ++LoopIndex) {
		if (this->FuncLoopsByBlock[(size_t)BlockNum1].GetBit(LoopIndex)) {
			if (!this->FuncLoopsByBlock[(size_t)BlockNum2].GetBit(LoopIndex)) {
				ExitsLoop = true;
				break;
			}
		}
	}
	return ExitsLoop;
} // end of SMPFunction::DoesEdgeExitLoop()

// build list of loop numbers that BlockNum is part of.
void SMPFunction::BuildLoopList(int BlockNum, list<std::size_t> &LoopList) const {
	std::size_t LoopIndex;
	assert((BlockNum >= 0) && (BlockNum < this->BlockCount));
	for (LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		if (this->FuncLoopsByBlock[(size_t) BlockNum].GetBit(LoopIndex)) {
			LoopList.push_back(LoopIndex);
		}
	}
	return;
} // end of SMPFunction::BuildLoopList()

// Build list of Block numbers contained in LoopNum.
void SMPFunction::BuildLoopBlockList(const size_t LoopNum, list<std::size_t> &BlockList) {
	assert(LoopNum < this->LoopCount);
	this->BuildBlockListFromBitset(this->FuncBlocksByLoop[LoopNum], BlockList);
	return;
} // end of SMPFunction::BuildLoopBlockList()

// each bit set in CurrBitSet is a block # for BlockList
void SMPFunction::BuildBlockListFromBitset(const STARSBitSet &CurrBitSet, list<std::size_t> &BlockList) const {
	for (size_t BlockIndex = 0; BlockIndex < (size_t)this->BlockCount; ++BlockIndex) {
		if (CurrBitSet.GetBit(BlockIndex)) {
			BlockList.push_back(BlockIndex);
		}
	}
	return;
} // end of SMPFunction::BuildBlockListFromBitset()

void SMPFunction::BuildLoopIterationCountExprs(const bool VerboseOutput, FILE *InfoAnnotFile) {
	std::bitset<1 + MD_LAST_REG_NO> TempBitset(0);

	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		this->LoopMemRangeInArgRegsBitmap[LoopIndex] = TempBitset;
		this->LoopAnalyzedBIVIters[LoopIndex] = this->LoopInductionVars[LoopIndex].end();
		++STARS_LoopIterationExprFailures; // Decrement if we succeed

		// Find the condition that terminates the loop.
		struct LoopComparison CurrentLoopComparisonExpr;
		CurrentLoopComparisonExpr.CompareOperator = SMP_NULL_OPERATOR;
		int TestBlockNum = this->LoopTestBlocksByLoopNum[LoopIndex];
		int LoopType = this->LoopTypesByLoopNum[LoopIndex];
		bool LowerLimit = false, UpperLimit = false; // Loop terminates by becoming greater than UpperLimit, or less than LowerLimit?
		STARSDefUseIter ExitUse1, ExitUse2; // iterators for the UseOps involved in conditional loop exit
		STARS_uval_t ExitConstUse2 = 0; // const value for exit comparison instead of ExitUse2
		bool CompareAgainstConst = false; // compare ExitUse1 against ExitConstUse2 if true, else ExitUse1 vs. ExitUse2
		if ((SMP_BLOCKNUM_UNINIT != TestBlockNum) && (DoesLoopTestAtTop(LoopType) || DoesLoopTestAtBottom(LoopType))) {
			SMPBasicBlock *TestBlock = this->GetBlockByNum(TestBlockNum);
			int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
			bool MultiExitLoop = this->GetLoopFollowBlockNum(LoopIndex, FollowBlockNum);
			list<SMPBasicBlock *>::const_iterator TestTargBlockIter = TestBlock->GetCondNonFallThroughSucc();
			if (TestTargBlockIter == TestBlock->GetLastConstSucc()) {
				SMP_msg("ERROR: LOOP: Could not find NonFallThroughSuccessor in test block %d loop %u ", TestBlockNum, LoopIndex);
				this->DumpFuncNameAndAddr();
				this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
				this->LoopAnalysisProblems[LoopIndex] = true;
				continue;
			}

			int TestTargBlockNum = (*TestTargBlockIter)->GetNumber();
			bool BranchExitsLoop = ((TestTargBlockNum == FollowBlockNum) || (!this->IsBlockInLoop(TestTargBlockNum, LoopIndex)));
			CurrentLoopComparisonExpr.ExitsLoop = BranchExitsLoop;
			SMPoperator BranchOperator = SMP_NULL_OPERATOR;
			SMPoperator SetBitOperator = SMP_NULL_OPERATOR;
			STARS_ea_t DecrementAddr = STARS_BADADDR;
			SMPInstr *CompareOrTestInst = TestBlock->FindBranchCompareOrTest(BranchOperator, SetBitOperator, DecrementAddr, false, SMP_SSA_UNINIT, TestBlock->GetRevInstCBegin());
			if (nullptr != CompareOrTestInst) {
				CompareOrTestInst->GetCompareOrTestExprAndReduceSelfTest(CurrentLoopComparisonExpr);
				CurrentLoopComparisonExpr.CompareOperator = BranchOperator;
				CurrentLoopComparisonExpr.CompareAddr = CompareOrTestInst->GetAddr();
				this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
			}
			else if (STARS_BADADDR != DecrementAddr) {
				// Found a decrement rather than a compare or test.
				SMPInstr *DecrementInst = this->GetInstFromAddr(DecrementAddr);
				assert(nullptr != DecrementInst);
				if ((BranchOperator == SMP_EQUAL) || (BranchOperator == SMP_NOT_EQUAL)) {
					CurrentLoopComparisonExpr.CompareOperator = BranchOperator;
					CurrentLoopComparisonExpr.CompareAddr = DecrementAddr;
					STARSDefUseIter DecrementDEF = DecrementInst->GetFirstNonFlagsDef();
					CurrentLoopComparisonExpr.Operand1 = (*DecrementDEF);
					STARSOpndTypePtr ZeroOp = DecrementInst->MakeImmediateOpnd(0);
					DefOrUse ZeroUse(ZeroOp);
					CurrentLoopComparisonExpr.Operand2 = ZeroUse;
					this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
				}
				else {
					SMP_msg("ERROR: LOOP: Decrement with BranchOperator %d in test block %d DecrementAddr %llx loop %u ",
						BranchOperator, TestBlockNum, (uint64_t)DecrementAddr, LoopIndex);
					this->DumpFuncNameAndAddr();
					this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
					this->LoopAnalysisProblems[LoopIndex] = true;
					continue;
				}
			}
			else {
				SMP_msg("ERROR: LOOP: Could not find compare or test in test block %d loop %zu ",
					TestBlockNum, LoopIndex);
				this->DumpFuncNameAndAddr();
				this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
				this->LoopAnalysisProblems[LoopIndex] = true;
				continue;
			}
		}
		else {
			SMP_msg("ERROR: LOOP: Could not do iteration analysis on loop of type %d loop %zu ", LoopType, LoopIndex);
			this->DumpFuncNameAndAddr();
			this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
			this->LoopAnalysisProblems[LoopIndex] = true;
			continue; // cannot analyze loop iterations
		}

		if (LoopIndex < this->LoopInductionVars.size()) {
			SMP_msg("INFO: LOOP: Analyzing loop iteration exprs for loop %zu ", LoopIndex);
			this->DumpFuncNameAndAddr();
			for (STARSInductionVarFamilyIter IVarVecIter = this->LoopInductionVars[LoopIndex].begin();
				IVarVecIter != this->LoopInductionVars[LoopIndex].end();
				++IVarVecIter) {

				int OutsideDefSSANum = (*IVarVecIter).BIVIncomingSSANum;
				if (OutsideDefSSANum != SMP_SSA_UNINIT) {
					// We have a valid basic induction variable.
					STARSOpndTypePtr BIVDefOp = (*IVarVecIter).BasicInductionVar.InductionVar.GetOp();
					// Normalize the LoopComparison struct so that the BIVDefOp is the LeftOperand.
					SMPoperator RelationalOperator = CurrentLoopComparisonExpr.CompareOperator;
					bool BIVIsLeftOp = IsEqOpIgnoreBitwidth(CurrentLoopComparisonExpr.Operand1.GetOp(), BIVDefOp);
					bool BIVIsRightOp = false;
					if (!BIVIsLeftOp)
						BIVIsRightOp = IsEqOpIgnoreBitwidth(CurrentLoopComparisonExpr.Operand2.GetOp(), BIVDefOp);
					if (BIVIsRightOp) { // invert comparison and swap operands
						CurrentLoopComparisonExpr.CompareOperator = InvertRelationalOperatorForExprSwap(RelationalOperator);
						RelationalOperator = CurrentLoopComparisonExpr.CompareOperator;
						CurrentLoopComparisonExpr.Operand2 = CurrentLoopComparisonExpr.Operand1;
						CurrentLoopComparisonExpr.Operand1 = this->LoopComparisonExprs.back().Operand2;
						this->LoopComparisonExprs.pop_back();
						this->LoopComparisonExprs.push_back(CurrentLoopComparisonExpr);
					}
					// assert(BIVIsLeftOp || BIVIsRightOp); // not true for BIV decrement or increment in lieu of a compare opcode
					if (!(BIVIsLeftOp || BIVIsRightOp))
						continue;
					bool PositiveIncrement = IsPositiveIncrementBIV((*IVarVecIter).BasicInductionVar);


					//  First, we need to find a loop-invariant initializer for the incoming value.
					STARS_ea_t OutsideDefAddr = (*IVarVecIter).BIVIncomingDefAddr;
					assert(STARS_BADADDR != OutsideDefAddr);
					// if (!(STARS_IsBlockNumPseudoID(OutsideDefAddr) || STARS_IsSSAMarkerPseudoID(OutsideDefAddr))) {
					if (!(STARS_IsBlockNumPseudoID(OutsideDefAddr))) {
						SMPInstr *OutsideDefInst = this->GetInstFromAddr(OutsideDefAddr);
						assert(nullptr != OutsideDefInst);
						STARSExpression *InitExpr = OutsideDefInst->CreateDefExpr(BIVDefOp);
						// Replace constants and simplify the init expression.
						InitExpr->EvaluateConsts();
						bool ParentChanged = InitExpr->SimplifyDriver();
						bool StoppedOnIV = false;
						bool changed = false;
						set<int> LoopRegHashes;
						set<STARS_ea_t> InitStackPtrCopySet;
						set<STARS_ea_t> LimitStackPtrCopySet;
						int DepthCounter = 0;
						bool Expanded = InitExpr->ExpandExpr(InitExpr->GetParentInst()->GetAddr(), LoopIndex, false, false, false, false, true, LoopRegHashes, StoppedOnIV, changed, InitStackPtrCopySet, DepthCounter);
						if (Expanded) {
							assert(!StoppedOnIV);
							InitExpr->EvaluateConsts();
							ParentChanged = InitExpr->SimplifyDriver();
							if (VerboseOutput) {
								SMP_msg("INFO: LOOP: Init EXPR:");
								InitExpr->Dump(0);
							}

							// Next, we need an expr for the loop limit test.
							STARSExpression *LimitExpr = this->CreateLimitExpr(LoopIndex, *IVarVecIter, CurrentLoopComparisonExpr);
							if (nullptr != LimitExpr) {
								LimitExpr->EvaluateConsts();
								ParentChanged = LimitExpr->SimplifyDriver();
								DepthCounter = 0;
								Expanded = LimitExpr->ExpandExpr(LimitExpr->GetParentInst()->GetAddr(), LoopIndex, true, false, false, false, false, LoopRegHashes, StoppedOnIV, changed, LimitStackPtrCopySet, DepthCounter);
								if (Expanded) {
									assert(!StoppedOnIV);
									LimitExpr->EvaluateConsts();
									ParentChanged = LimitExpr->SimplifyDriver();
									if (VerboseOutput) {
										SMP_msg("INFO: LOOP: Limit EXPR:");
										LimitExpr->Dump(0);
									}

									// Next, create the iterations count expr.
									STARS_sval_t LimitIncrease;
									STARSExpression *IterationCountExpr = this->CreateIterationsExpr(LoopIndex, *IVarVecIter, PositiveIncrement, InitExpr->Clone(), LimitExpr->Clone());

									if (nullptr != IterationCountExpr) {
										// See if we have a static memory address used in a compare or test instruction
										//  to terminate the loop, and the address is not used in the loop because
										//  this->LoopExecutesWithLimitValue[LoopIndex] is false. This could be the case
										//  in which we have: cmp rbx,staticaddr and staticaddr points to a global var
										//  that is adjacent to the var being read or written. In that case, staticaddr
										//  needs to be adjusted by the move_globals defense as the global var that is
										//  read or written in the loop is moved. It does NOT need to follow the adjacent
										//  variable in which the sentinel address lies, as the adjacent variable is getting
										//  separated by move_globals from the currently accessed variable. The annotation
										//  will identify the last address actually read or written in the loop.
										if (!this->LoopExecutesWithLimitValue[LoopIndex]) {
											// Limit value is not used. Might be the sentinel case.
											STARSOpndTypePtr AddendOp = (*IVarVecIter).BasicInductionVar.Addend.GetOp();
											if (AddendOp->IsImmedOp()) {
												STARS_sval_t IncDecValue = (STARS_sval_t)AddendOp->GetImmedValue();
												if ((0 < IncDecValue) && (*IVarVecIter).BasicInductionVar.SubtractAddend) {
													IncDecValue = (0 - IncDecValue);
												}
												// Get the compare instruction. See if it has a GlobalAddr directly referenced.
												STARS_ea_t CmpInstAddr = this->LoopComparisonExprs[LoopIndex].CompareAddr;
												bool GlobalAddrFound = false;
												STARS_ea_t GlobalAddr = STARS_BADADDR;
												if (this->LoopComparisonExprs[LoopIndex].Operand1.GetOp()->IsImmedOp()) {
													GlobalAddr = (STARS_ea_t) this->LoopComparisonExprs[LoopIndex].Operand1.GetOp()->GetImmedValue();
												}
												else if (this->LoopComparisonExprs[LoopIndex].Operand2.GetOp()->IsImmedOp()) {
													GlobalAddr = (STARS_ea_t)  this->LoopComparisonExprs[LoopIndex].Operand2.GetOp()->GetImmedValue();
												}
												if ((STARS_BADADDR != GlobalAddr) && IsImmedGlobalAddress(GlobalAddr)) {
													SMPInstr *CmpInst = this->GetInstFromAddr(CmpInstAddr);
													assert(nullptr != CmpInst);
													STARS_sval_t Var1Loc = (STARS_sval_t)GlobalAddr;
													Var1Loc -= IncDecValue;
													// Emit annotation with Var1Loc and IncDecValue for CopySetInstAddr.
													SMP_fprintf(InfoAnnotFile, "%18llx %6u INSTR SENTINEL BASE %llx OFFSET %lld ZZ %s\n",
														(uint64_t)CmpInstAddr, CmpInst->GetSize(), (uint64_t)Var1Loc, (int64_t)IncDecValue, CmpInst->GetDisasm());
												}
											}
										}

										// Second SENTINEL case: If we have taken the address of a global as the LimitExpr,
										//  and the Limit value is a sentinel and memory is not therefore
										//  accessed on an iteration with the BIV == the Limit value, then
										//  there is a potential problem in the move_globals transformation.
										//  We could use the address of global var2 as a sentinel when looping
										//  through global var1. If we move global var1 away from global var2, then
										//  we need to adjust the limit value to follow it and not remain
										//  pointing at global var2.
										if (!this->LoopExecutesWithLimitValue[LoopIndex]) {
											// Limit value is not used. Might be the sentinel case.
											if (1 == LimitStackPtrCopySet.size()) { // simple case
												STARSOpndTypePtr AddendOp = (*IVarVecIter).BasicInductionVar.Addend.GetOp();
												assert(AddendOp->IsImmedOp());
												STARS_sval_t IncDecValue = (STARS_sval_t)AddendOp->GetImmedValue();
												// If we have a loop IncDecValue of k, and we start in var1 and an
												//  address in var2 is the sentinel, then the last memory value accessed
												//  will be k bytes below the sentinel. We need move_globals to instrument
												//  the instruction that takes the address of the sentinel location so that
												//  it is logically location_in_var1+k rather than logically being
												//  location_in_var2, i.e. as var1 is moved, the sentinel address computation
												//  will remain location_in_var1+k and make no reference to var2.
												// The annotation we emit will therefore need to record two numbers:
												//  location_in_var1 and k.
												STARS_ea_t CopySetInstAddr = (*LimitStackPtrCopySet.cbegin());
												SMPInstr *CopySetInst = this->GetInstFromAddr(CopySetInstAddr);
												assert(nullptr != CopySetInst);
												STARS_uval_t GlobalAddr;
												if (CopySetInst->IsLoadGlobalStaticDataAddress(GlobalAddr)) {
													STARS_sval_t Var1Loc = (STARS_sval_t)GlobalAddr;
													Var1Loc -= IncDecValue;
													// Emit annotation with Var1Loc and IncDecValue for CopySetInstAddr.
													SMP_fprintf(InfoAnnotFile, "%18llx %6u INSTR SENTINEL BASE %llx OFFSET %lld ZZ\n",
														(uint64_t)CopySetInstAddr, CopySetInst->GetSize(), (uint64_t)Var1Loc, (int64_t)IncDecValue);
												}
											}
										}
										ParentChanged = IterationCountExpr->SimplifyDriver();
										this->LoopIterationsInitExprs[LoopIndex] = InitExpr;
										this->LoopIterationsLimitExprs[LoopIndex] = LimitExpr;
										this->LoopIterationsCountExprs[LoopIndex] = IterationCountExpr;
										SMP_msg("INFO: LOOP: Computed IterationCountExpr for loop %d in function at %llx\n",
											LoopIndex, (uint64_t) this->GetFirstFuncAddr());
										// Save the InitExpr and LimitExpr in case this BIV
										//  is encountered in symbolic ExpandExpr() work later
										//  and we need to substitute the bounds of it.
										(*IVarVecIter).BIVInitExpr = InitExpr;
										(*IVarVecIter).BIVLimitExpr = LimitExpr;
										this->LoopAnalyzedBIVIters[LoopIndex] = IVarVecIter;
										--STARS_LoopIterationExprFailures; // undo top of loop increment
										++STARS_LoopIterationExprSuccesses;
										if (VerboseOutput) {
											SMP_msg("INFO: LOOP: IterationCount EXPR:");
											IterationCountExpr->Dump(0);
										}
										break; // Found the primary BIV for this LoopIndex
									}
									else {
										SMP_msg("ERROR: LOOP: Failure to create IterationCountExpr for loop %zu ", LoopIndex);
										this->DumpFuncNameAndAddr();
										this->LoopAnalysisProblems[LoopIndex] = true;
										if (nullptr != InitExpr)
											delete InitExpr;
										if (nullptr != LimitExpr)
											delete LimitExpr;
									}
								}
								else {
									SMP_msg("ERROR: LOOP: Failure on ExpandExpr() for LimitExpr in AnalyzeLoopIterations() for loop %zu ", LoopIndex);
									this->DumpFuncNameAndAddr();
									this->LoopAnalysisProblems[LoopIndex] = true;
									if (nullptr != InitExpr)
										delete InitExpr;
									if (nullptr != LimitExpr)
										delete LimitExpr;
								}
							}
							else {
								SMP_msg("ERROR: LOOP: Failure to create LimitExpr for loop %zu ", LoopIndex);
								this->DumpFuncNameAndAddr();
								this->LoopAnalysisProblems[LoopIndex] = true;
								if (nullptr != InitExpr)
									delete InitExpr;
							}
						}
						else {
							SMP_msg("ERROR: LOOP: Failure on ExpandExpr() for InitExpr in AnalyzeLoopIterations() for loop %zu ", LoopIndex);
							this->DumpFuncNameAndAddr();
							this->LoopAnalysisProblems[LoopIndex] = true;
							if (nullptr != InitExpr)
								delete InitExpr;
						}
					}
				}
			} // end for all IV iters in current LoopIndex
		}
	} // end for all LoopIndex values

	return;
} // end of SMPFunction::BuildLoopIterationCountExprs()

// Analyze how many times each loop iterates
void SMPFunction::AnalyzeLoopIterations(void) {
	bool DebugFlag = false;
	DebugFlag = (0x804f596 == this->GetFirstFuncAddr());
	// DebugFlag = (0x402730 == this->GetFirstFuncAddr());
	bool VerboseOutput = DebugFlag || global_stars_interface->VerboseLoopsMode();
	FILE *InfoAnnotFile = global_STARS_program->GetInfoAnnotFile();
	if (0 < this->GetNumLoops()) {
		// Get ready to mark memory writes by loop.
		STARSBitSet TempPositiveBitSet;
		TempPositiveBitSet.AllocateBits(this->MaxStackAccessLimit);
		STARSBitSet TempNegativeBitSet;
		TempNegativeBitSet.AllocateBits(0 - this->MinStackAccessOffset);
		for (unsigned i = 0; i < this->LoopCount; ++i) {
			this->NegativeOffsetStackBytesWrittenByLoop.push_back(TempNegativeBitSet);
			this->PositiveOffsetStackBytesWrittenByLoop.push_back(TempPositiveBitSet);
		}
	}

	this->LoopIterationsInitExprs.resize(this->LoopCount, nullptr);
	this->LoopIterationsLimitExprs.resize(this->LoopCount, nullptr);
	this->LoopIterationsCountExprs.resize(this->LoopCount, nullptr);
	this->LoopMemWriteRangeExprs.resize(this->LoopCount);
	this->LoopMemWriteBoundsExprs.resize(this->LoopCount);
	this->LoopMemWriteBoundsExprsExpanded.resize(this->LoopCount);
	this->StringMemWriteRangeExprs.resize(this->LoopCount + 1);
	this->StringMemWriteRangeWidths.resize(this->LoopCount + 1);
	this->LoopRegHashSets.resize(this->LoopCount);
	this->LoopRegSourceExprPairs.resize(this->LoopCount);
	this->LoopIVUsedAsStringWriteCounter.resize(this->LoopCount);
	this->NonStackFrameLoopMemWrites.resize(this->LoopCount, false);
	this->LoopMemRangeInArgRegsBitmap.resize(this->LoopCount);
	this->OutputRegsByLoop.resize(this->LoopCount);
	this->CalleePreservedRegsByLoop.resize(this->LoopCount);
	std::bitset<1 + MD_LAST_REG_NO> TempBitset(0);
	this->MemRangeRegsBitmap = TempBitset;
	this->LoopAnalyzedBIVIters.resize(this->LoopCount);
	this->LoopWritesGlobalStaticMemory.resize(this->LoopCount, false);
	this->LoopReadsGlobalStaticMemory.resize(this->LoopCount, false);
	this->LoopHasCalleeMemWrites.resize(this->LoopCount, false);
	this->LoopUsesStackPtrRegs.resize(this->LoopCount, true); // default to the most common case
	this->LoopHasPreconditions.resize(this->LoopCount, true); // default to the most common case
	this->LoopExecutesWithLimitValue.resize(this->LoopCount, false);
	this->LoopAnalysisProblems.resize(this->LoopCount, false);
	this->CalleeMemExprProblems.resize(this->LoopCount + 1, false);
	this->SymbolicAnalysisProblems.resize(this->LoopCount + 1, false);
	this->LoopIncrementValue.resize(this->LoopCount, 0); // default to invalid case
	this->LoopMemAddrExprsFromCallees.resize(this->LoopCount);
	this->LoopMemAddrExprWidthsFromCallees.resize(this->LoopCount);
	this->LoopMemExprsExpandToStackOffsets.resize(this->LoopCount, false);
	this->LoopMemAddrExprWidthsFromCalleeLoops.resize(this->LoopCount + 1);
	this->LoopMemAddrExprsFromCalleeLoops.resize(this->LoopCount + 1);
	this->MemAddrExprsFromCallees.resize(this->LoopCount + 1);
	this->MemAddrExprWidthsFromCallees.resize(this->LoopCount + 1);
	this->NonStackFrameCalleeMemWrites.resize(this->LoopCount + 1);
	this->InArgsUsedInMemWrites.resize(this->LoopCount + 1);
	this->InArgsUsedInMemWriteByteWidths.resize(this->LoopCount + 1);
	this->StoppedOnIVMemRangeExprs.resize(this->LoopCount + 1);
	this->StoppedOnIVNonRangeExprs.resize(this->LoopCount + 1);
	this->StoppedOnIVMemRangeIterWidths.resize(this->LoopCount + 1);
	this->StoppedOnIVNonRangeIterWidths.resize(this->LoopCount + 1);
	this->RelationalLowerBoundExprs.resize(this->LoopCount + 1);
	this->RelationalUpperBoundExprs.resize(this->LoopCount + 1);
	this->RelationalMemWriteWidths.resize(this->LoopCount + 1);

	if (!this->HasStructuredControlFlow()) {
		SMP_msg("INFO: LOOP: AnalyzeLoopIterations() terminating early due to unstructured control flow ");
		this->DumpFuncNameAndAddr();
		return; // just allocate data structures above to avoid crashes, then exit.
	}

	// Prepare for limited interprocedural constant propagation
	//  by finding constant InArg values.
	this->ComputeInArgConstValues();

	// Fill in LoopIterationsCountExprs[] for each LoopIndex; emit SENTINEL annotations as discovered along the way.
	this->BuildLoopIterationCountExprs(VerboseOutput, InfoAnnotFile);

	// See if we had any fatal errors in analysis.
	for (size_t LoopIndex = 0; LoopIndex < this->GetNumLoops(); ++LoopIndex) {
		if (nullptr == this->LoopIterationsCountExprs[LoopIndex]) {
			this->LoopAnalysisProblems[LoopIndex] = true;
			if (this->DoesLoopWriteMemory(LoopIndex)) {
				SMP_msg("ERROR: LOOP: Failure to create iteration count expr for loop %d ", LoopIndex);
				this->DumpFuncNameAndAddr();
				this->DumpDotCFG();
			}
			else {
				SMP_msg("SERIOUS WARNING: LOOP: Failure to create iteration count expr for non-mem-writing loop %d starting at %llx ",
					LoopIndex, (uint64_t)(this->GetBlockByNum((size_t) this->LoopHeadBlockNumbers[LoopIndex])->GetFirstAddr()));
				this->DumpFuncNameAndAddr();
			}
		}
	}

	// Second loop: Create InitExpr and LimitExpr for every basic
	//  induction var that was not involved in the loop compare-and-branch and was
	//  therefore not processed in the previous loop.
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		// bool AttemptAnalysis = (!this->LoopAnalysisProblems[LoopIndex] && (LoopIndex < this->LoopInductionVars.size()));
		bool AttemptAnalysis = (LoopIndex < this->LoopInductionVars.size());
		if (AttemptAnalysis) {
			if (VerboseOutput) {
				SMP_msg("INFO: LOOP: Analyzing secondary BIVs for function %s loop %zu\n", this->GetFuncName(), LoopIndex);
			}
			for (STARSInductionVarFamilyIter IVarVecIter = this->LoopInductionVars[LoopIndex].begin();
				IVarVecIter != this->LoopInductionVars[LoopIndex].end();
				++IVarVecIter) {

				int OutsideDefSSANum = (*IVarVecIter).BIVIncomingSSANum;
				if (OutsideDefSSANum != SMP_SSA_UNINIT) {
					// We have a valid basic induction variable.
					if (nullptr == (*IVarVecIter).BIVLimitExpr) {
						// This BIV was not analyzed in the first loop,
						//  because it is not involved in the loop exit
						//  compare-and-branch. Its limit must therefore be obtained
						//  using the iteration count expr times its stride plus its init value.
						STARSExpression *InitExpr = (*IVarVecIter).BIVInitExpr;
						bool StoppedOnIV = false;
						set<int> LoopRegHashes;
						if (nullptr == InitExpr) {
							STARS_ea_t OutsideDefAddr = (*IVarVecIter).BIVIncomingDefAddr;
							assert(STARS_BADADDR != OutsideDefAddr);
							// if (!(STARS_IsBlockNumPseudoID(OutsideDefAddr) || STARS_IsSSAMarkerPseudoID(OutsideDefAddr))) {
							if (!(STARS_IsBlockNumPseudoID(OutsideDefAddr))) {
								SMPInstr *OutsideDefInst = this->GetInstFromAddr(OutsideDefAddr);
								assert(nullptr != OutsideDefInst);
								STARSOpndTypePtr BIVDefOp = (*IVarVecIter).BasicInductionVar.InductionVar.GetOp();
								InitExpr = OutsideDefInst->CreateDefExpr(BIVDefOp);
								// Replace constants and simplify the init expression.
								assert(nullptr != InitExpr);
								InitExpr->EvaluateConsts();
								bool ParentChanged = InitExpr->SimplifyDriver();
								if (VerboseOutput) {
									SMP_msg("INFO: LOOP: Init EXPR for secondary BIV:");
									InitExpr->Dump(0);
								}
								// Record the InitExpr in the IV structure.
								(*IVarVecIter).BIVInitExpr = InitExpr;
							}
							else {
								SMP_msg("ERROR: LOOP: Cannot analyze BIV with OutsideDefAddr of %llx ", (uint64_t)OutsideDefAddr);
								this->DumpFuncNameAndAddr();
								continue;
							}
						} // end if nullptr == InitExpr

						// Next, we need an expr for the loop limit test.
						STARSExpression *LimitExpr = this->CreateSecondaryBIVLimitExpr(LoopIndex, *IVarVecIter);
						if (nullptr != LimitExpr) {
							LimitExpr->EvaluateConsts();
							bool ParentChanged = LimitExpr->SimplifyDriver();
							if (VerboseOutput) {
								SMP_msg("INFO: LOOP: Limit EXPR for secondary BIV:");
								LimitExpr->Dump(0);
							}
							// Record the LimitExpr in the IV structure.
							(*IVarVecIter).BIVLimitExpr = LimitExpr;
						}
						else {
							SMP_msg("ERROR: LOOP: Failure to create secondary BIV LimitExpr ");
							this->DumpFuncNameAndAddr();
						}
					}
				}
			} // end for each IVarVecIter in current loop index
		}
	} // end for each loop index: loop 2, secondary BIV range analysis

	// Third loop: Create InitExpr and LimitExpr for all dependent induction vars.
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
#if 0
		if (this->LoopAnalysisProblems[LoopIndex])
			continue; // analysis failed in the first loop
#endif
		if (VerboseOutput) {
			SMP_msg("INFO: Analyzing DIVs for function %s loop %zu\n", this->GetFuncName(), LoopIndex);
		}
		size_t IVarVecIndex = 0; // for debugging breakpoints
		for (STARSInductionVarFamilyIter IVarVecIter = this->LoopInductionVars[LoopIndex].begin();
			IVarVecIter != this->LoopInductionVars[LoopIndex].end();
			++IVarVecIter, ++IVarVecIndex) {
			int OutsideDefSSANum = (*IVarVecIter).BIVIncomingSSANum;
			if (OutsideDefSSANum != SMP_SSA_UNINIT) {
				// We have a valid basic induction variable.
				//  See if it had InitExpr and LimitExpr analyzed successfully.
				InductionVarFamily IVFamily = (*IVarVecIter);
				if (nullptr == IVFamily.BIVInitExpr) {
					continue;
				}
				// We have a valid and analyzed BIV.
				//  See if it has any dependent induction vars.
				for (size_t DIVIndex = 0; DIVIndex < IVFamily.DependentInductionVars.size(); ++DIVIndex) {
					STARSExpression *DIVInitExpr = this->CreateDIVInitExpr(LoopIndex, IVFamily, DIVIndex);
					bool success = (nullptr != DIVInitExpr);
					if (!success) {
						SMP_msg("ERROR: LOOP: CreateDIVInitExpr failed for loop %zu DIVIndex %zu\n", LoopIndex, DIVIndex);
						continue;
					}
					STARSExpression *DIVLimitExpr = nullptr;
					if (nullptr != IVFamily.BIVLimitExpr) {
						DIVLimitExpr = this->CreateDIVLimitExpr(LoopIndex, IVFamily, DIVIndex);
						success = (nullptr != DIVLimitExpr);
						if (!success) {
							SMP_msg("ERROR: LOOP: CreateDIVLimitExpr failed for loop %zu DIVIndex %zu ", LoopIndex, DIVIndex);
							this->DumpFuncNameAndAddr();
						}
					}
#if 0
					bool changed = false;
					success = this->ReplaceAllBIVsWithExprs(LoopIndex, DIVInitExpr, true, changed);
					if (!success) {
						SMP_msg("ERROR: LOOP: ReplaceAllBIVsWithExprs on InitExpr failed for loop %zu DIVIndex %zu\n", LoopIndex, DIVIndex);
						continue;
					}
					success = this->ReplaceAllBIVsWithExprs(LoopIndex, DIVLimitExpr, false, changed);
					if (!success) {
						SMP_msg("ERROR: LOOP: ReplaceAllBIVsWithExprs on LimitExpr failed for loop %zu DIVIndex %zu\n", LoopIndex, DIVIndex);
						continue;
					}
#endif
					DIVInitExpr->EvaluateConsts();
					(void) DIVInitExpr->SimplifyDriver();
					if (nullptr != DIVLimitExpr) {
						DIVLimitExpr->EvaluateConsts();
						(void)DIVLimitExpr->SimplifyDriver();
					}
					// Record InitExpr and LimitExpr for DIV.
					IVFamily.DependentInductionVars[DIVIndex].DIVInitExpr = DIVInitExpr;
					IVFamily.DependentInductionVars[DIVIndex].DIVLimitExpr = DIVLimitExpr;
					if (VerboseOutput) {
						SMP_msg("INFO: DIV Init Expr for DIVIndex = %zu\n", DIVIndex);
						DIVInitExpr->Dump(0);
						if (nullptr != DIVLimitExpr) {
							SMP_msg("INFO: DIV Limit Expr for DIVIndex = %zu\n", DIVIndex);
							DIVLimitExpr->Dump(0);
						}
					}
				}
			}
		} // end for all induction var iters
	} // end for all loops: loop 3, DIV range analysis

	// Fourth loop: Create the memory expressions for memory writes and reads in each loop.
	set<STARS_ea_t> StaticMemWriteAddrsEmitted; // annotations emitted; avoid duplication
	set<STARS_ea_t> StaticMemReadAddrsEmitted; // annotations emitted; avoid duplication
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			// Try to determine the memory write address range in the loop.
		STARSInductionVarFamilyIter IVarVecIter = this->LoopAnalyzedBIVIters[LoopIndex];
		// bool AttemptAnalysis = (!this->LoopAnalysisProblems[LoopIndex] && (IVarVecIter != this->LoopInductionVars[LoopIndex].end()));
		bool AttemptAnalysis = (IVarVecIter != this->LoopInductionVars[LoopIndex].end());
		if (!AttemptAnalysis)
			continue; // analysis failed in the first loop

		STARSExprSet MemWriteExprs;
		set<int> LoopRegHashes = this->LoopRegHashSets[LoopIndex];
		vector<set<STARS_ea_t> > StackPtrCopiesVector;
		STARSMemWriteExprsList MemWriteExprWidths;
		bool PositiveIncrement = IsPositiveIncrementBIV((*IVarVecIter).BasicInductionVar);
		bool MemRangeSuccess = this->CreateSPARKMemoryWriteRangeExpr(LoopIndex, true, LoopRegHashes, MemWriteExprs, MemWriteExprWidths, StackPtrCopiesVector);
		if (MemRangeSuccess) {
			this->LoopRegHashSets[LoopIndex] = LoopRegHashes;
			SMP_msg("INFO: LOOP: Computed %zu LoopMemWriteExprs for loop %d in function at %llx\n",
				MemWriteExprs.size(), LoopIndex, (uint64_t) this->GetFirstFuncAddr());
			size_t MemListIndex = 0; // for debugging
			for (STARSMemWriteExprListIter MemListIter = MemWriteExprWidths.begin(); MemListIter != MemWriteExprWidths.end(); ++MemListIter, ++MemListIndex) {
				STARSExprSetIter ExprIter = (*MemListIter).first;
				STARSExpression *MemAddressRangeExpr = (*ExprIter); 
				pair<STARSExprSetIter, bool> InsertResult = this->LoopMemWriteRangeExprs[LoopIndex].insert(MemAddressRangeExpr);
				if (VerboseOutput) {
					SMP_msg("INFO: LoopMemWrite EXPR:");
					MemAddressRangeExpr->Dump(0);
					if (MemAddressRangeExpr->IsStackPtrRegUsed()) {
						SMP_msg("INFO: LoopMemWrite EXPR is stack-based.\n");
					}
				}
				// Create the lower and upper bounds of the memory writing range.
				STARSExpression *MemWriteLowerBound = MemAddressRangeExpr->Clone(); // deep copy, for editing
				bool changed = false;
				MemWriteLowerBound->ResetAllProcessed();
				bool InitSuccess = this->ReplaceAllIVsWithExprs(LoopIndex, MemWriteLowerBound, true, changed);
				bool LimitSuccess = false;
				bool success = InitSuccess; // InitSuccess is mandatory for annotations
				if (success) {
					STARSExpression *MemWriteUpperBound = MemAddressRangeExpr->Clone(); // deep copy, for editing
					MemWriteUpperBound->ResetAllProcessed();
					LimitSuccess = this->ReplaceAllIVsWithExprs(LoopIndex, MemWriteUpperBound, false, changed);
					if (InitSuccess) { // InitSuccess is mandatory for annotations.
						if (VerboseOutput) {
							SMP_msg("INFO: Successfully replaced IVs with InitExpr for loop %d in function at %llx LimitSuccess: %d\n",
								LoopIndex, (uint64_t) this->GetFirstFuncAddr(), LimitSuccess);
						}
						(void) MemWriteLowerBound->SimplifyDriver();
						if (LimitSuccess) {
							(void) MemWriteUpperBound->SimplifyDriver();
						}
						else {
							// We will emit annotations that have a useful starting address for the
							//  memory range written, and a code that indicates the range limit is unknown.
							delete MemWriteUpperBound;
							MemWriteUpperBound = nullptr;
							this->LoopAnalysisProblems[LoopIndex] = true;
							SMP_msg("INFO: Failure to ReplaceIVsWithLimitExpr for write MemListIndex %zu loop %zu in function at %llx\n",
								MemListIndex, LoopIndex, (uint64_t) this->GetFirstFuncAddr());
						}
						if (PositiveIncrement) {
							STARSExprBoundsPair InsertVal(MemWriteLowerBound, MemWriteUpperBound);
							pair<STARSExprBoundsIter, bool> InsertResult = this->LoopMemWriteBoundsExprs[LoopIndex].insert(InsertVal);
						}
						else {
							STARSExprBoundsPair InsertVal(MemWriteUpperBound, MemWriteLowerBound);
							pair<STARSExprBoundsIter, bool> InsertResult = this->LoopMemWriteBoundsExprs[LoopIndex].insert(InsertVal);
						}
						int StackPtrCopiesVecIndex = (*MemListIter).second.second;
						if (VerboseOutput) {
							assert(nullptr != MemWriteLowerBound);
							SMP_msg("INFO: MemWriteLowerBound EXPR:");
							MemWriteLowerBound->Dump(0);
							if (nullptr != MemWriteUpperBound) {
								SMP_msg("INFO: MemWriteUpperBound EXPR:");
								MemWriteUpperBound->Dump(0);
							}
						}

						// Clone the lower and upper bounds expr for annotations and inheritance into callers.
						STARSExpression *MemWriteLowerBoundExpanded = MemWriteLowerBound->Clone();
						STARSExpression *MemWriteUpperBoundExpanded = nullptr;
						if (nullptr != MemWriteUpperBound)
							MemWriteUpperBoundExpanded = MemWriteUpperBound->Clone();

						// For inheritance into the callers of this function, we want
						//  to expand the expressions beyond loop boundaries, hopefully
						//  as far as incoming args or constant initializers.
						set<STARS_ea_t> NewStackPtrCopiesSet;
						success = this->ExpandExprToInArgs(LoopIndex, MemWriteLowerBoundExpanded, true, NewStackPtrCopiesSet);
						if (success && LimitSuccess) {
							LimitSuccess = this->ExpandExprToInArgs(LoopIndex, MemWriteUpperBoundExpanded, false, NewStackPtrCopiesSet);
						}
						if (success) {
							set<STARS_ea_t> UnionedStackPtrCopiesSet;
							insert_iterator<set<STARS_ea_t> > EndIter = std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin());
							if (0 <= StackPtrCopiesVecIndex) { // vector has entry for a set
								EndIter = merge(StackPtrCopiesVector[StackPtrCopiesVecIndex].begin(),
									StackPtrCopiesVector[StackPtrCopiesVecIndex].end(),
									NewStackPtrCopiesSet.begin(), NewStackPtrCopiesSet.end(),
									std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin()));
							}
							else { // just copy the NewStackPtrCopiesSet
								EndIter = std::copy(NewStackPtrCopiesSet.begin(), NewStackPtrCopiesSet.end(),
									std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin()));
							}

							// Substitute global min and max values for global mem reads that
							//  are still in the expr, and see if we can simplify as a result.
							bool FinalChange = MemWriteLowerBoundExpanded->SubstituteGlobalAndArgRangeValues(true);
							if (FinalChange)
								(void) MemWriteLowerBoundExpanded->SimplifyDriver();
							if (LimitSuccess) {
								FinalChange = MemWriteUpperBoundExpanded->SubstituteGlobalAndArgRangeValues(false);
								if (FinalChange)
									(void) MemWriteUpperBoundExpanded->SimplifyDriver();
							}

							// To assist in stack transformation security defenses, if the 
							//  memory range expression is stack-relative, we emit an annotation
							//  that identifies the memory range written to at InstAddr.
							// The annotation should look like:
							//  [hex InstAddr] [InstSize] INSTR STACKMEMRANGE MIN [ESP-k] MAX [ESP-j] ZZ
							set<STARS_ea_t> StaticMemLeaAddrSet;
							bool StackAnnotation = false;
							if (LimitSuccess && MemWriteLowerBoundExpanded->IsStackPtrPlusOffset() && MemWriteUpperBoundExpanded->IsStackPtrPlusOffset()) {
								this->EmitStackMemRangeAnnotations(MemWriteLowerBoundExpanded, MemWriteUpperBoundExpanded, PositiveIncrement, UnionedStackPtrCopiesSet, StaticMemLeaAddrSet);
								StackAnnotation = true;
							}
							if (!StaticMemLeaAddrSet.empty()) {
								SMP_msg("INFO: StaticMemLeaAddrSet size (writes): %zu \n", StaticMemLeaAddrSet.size());
							}

							STARSExprBoundsPair InsertExprPair;
							if (PositiveIncrement) {
								InsertExprPair.first = MemWriteLowerBoundExpanded;
								InsertExprPair.second = MemWriteUpperBoundExpanded;
							}
							else {
								InsertExprPair.first = MemWriteUpperBoundExpanded;
								InsertExprPair.second = MemWriteLowerBoundExpanded;
							}
							pair<STARSExprBoundsIter, bool> InsertResult = this->LoopMemWriteBoundsExprsExpanded[LoopIndex].insert(InsertExprPair);

							// Does the mem write depend on an InArg?
							if (MemWriteLowerBoundExpanded->UsesInArgReg() || (LimitSuccess && MemWriteUpperBoundExpanded->UsesInArgReg())) {
								this->HasLoopInArgMemWrites = true;
							}

							// To assist in defenses related to static global memory, output ranges
							//  of memory written when the Init address has simplified to a constant
							if (MemWriteLowerBoundExpanded->IsConstExpr()) {
								SMPInstr *MemWriteInst = MemWriteLowerBoundExpanded->GetOriginalParentInst();
								STARS_ea_t MemWriteAddr = MemWriteInst->GetAddr();
								if (StaticMemWriteAddrsEmitted.find(MemWriteAddr) == StaticMemWriteAddrsEmitted.cend()) {
									if (LimitSuccess && (MemWriteAddr != MemWriteUpperBoundExpanded->GetOriginalParentInst()->GetAddr())) {
										SMP_msg("ERROR: LOOP: Constant lower and upper bound exprs don't have same InstAddr for MemListIndex %zu loop %zu ",
											MemListIndex, LoopIndex);
										this->DumpFuncNameAndAddr();
									}
									else {
										int InstSize = MemWriteInst->GetSize();
										STARS_uval_t LowerBoundVal = MemWriteLowerBoundExpanded->GetConstLeftOperand()->GetImmedValue();
										STARS_uval_t UpperBoundVal = (STARS_uval_t) 0xfffffffffffffffe; // Code for unknown limit value
										if (LimitSuccess
											&& (nullptr != MemWriteUpperBoundExpanded->GetConstLeftOperand())
											&& (MemWriteUpperBoundExpanded->GetConstLeftOperand()->IsImmedOp())) {
											UpperBoundVal = MemWriteUpperBoundExpanded->GetConstLeftOperand()->GetImmedValue();
										}
										FILE *InfoAnnotFile = global_STARS_program->GetInfoAnnotFile();
										SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR STATICMEMWRITE MIN ", (uint64_t)MemWriteAddr, InstSize);
										if (LowerBoundVal == UpperBoundVal) {
											// A simple read from a static mem address with no index registers has
											//  the same expr for lower bound and upper bound. Use the ByteWidth to 
											//  expand the upper bound.
											uint16_t ByteWidth = MemWriteLowerBoundExpanded->GetConstLeftOperand()->GetByteWidth();
											UpperBoundVal += (STARS_uval_t)ByteWidth;
										}
										if (LowerBoundVal < UpperBoundVal) {
											SMP_fprintf(InfoAnnotFile, "%llx LIMIT %llx ZZ %s\n",
												(uint64_t) LowerBoundVal, (uint64_t) UpperBoundVal, MemWriteInst->GetDisasm());
										}
										else {
											SMP_fprintf(InfoAnnotFile, "%llx LIMIT %llx ZZ %s\n",
												(uint64_t) UpperBoundVal, (uint64_t) LowerBoundVal, MemWriteInst->GetDisasm());
										}
										(void) StaticMemWriteAddrsEmitted.insert(MemWriteAddr);
									}
								}
							}
							else if (!StackAnnotation) {
								// LowerBoundExpanded expr was neither static mem nor stack.
								SMP_msg("INFO: LOOP: Write expr neither static mem nor stack mem: \n");
								if (VerboseOutput) {
									MemWriteLowerBoundExpanded->Dump(0);
								}
							}
						}
						else {
							SMP_msg("ERROR: LOOP: ExpandExprToInArg failure for MemListIndex %zu loop %zu ",
								MemListIndex, LoopIndex);
							this->DumpFuncNameAndAddr();
							delete MemWriteLowerBoundExpanded;
							delete MemWriteUpperBoundExpanded;
							this->LoopAnalysisProblems[LoopIndex] = true;
						}
					}
					else {
						SMP_msg("ERROR: LOOP: Failure to ReplaceIVsWithInitExpr (writes) for MemListIndex %zu loop %d ",
							MemListIndex, LoopIndex);
						this->DumpFuncNameAndAddr();
						delete MemWriteLowerBound;
						if (nullptr != MemWriteUpperBound)
							delete MemWriteUpperBound;
						this->LoopAnalysisProblems[LoopIndex] = true;
					}
				}
				else {
					SMP_msg("ERROR: LOOP: Failure to ReplaceIVsWithInitExpr (writes) for MemListIndex %zu loop %d ",
						MemListIndex, LoopIndex);
					this->DumpFuncNameAndAddr();
					if (VerboseOutput) {
						MemWriteLowerBound->Dump(0);
					}
					delete MemWriteLowerBound;
					this->LoopAnalysisProblems[LoopIndex] = true;
				}
			} // end for all entries in MemWriteExprWidths
		}
		else {
			if (this->LoopWritesMemory[LoopIndex]) {
				SMP_msg("ERROR: LOOP: Failure to CreateSPARKMemRangeExpr for loop %d ", LoopIndex);
				this->DumpFuncNameAndAddr();
				this->LoopAnalysisProblems[LoopIndex] = true;
			}
			else {
				SMP_msg("INFO: LOOP: No memory writes in loop %d ", LoopIndex);
				this->DumpFuncNameAndAddr();
			}
		}

		// Now, do memory read range analysis and produce memory read range annotations.
		STARSExprSet MemReadExprs;
		STARSMemWriteExprsList MemReadExprWidths;
		StackPtrCopiesVector.clear();
		MemRangeSuccess = this->CreateSPARKMemoryReadRangeExprs(LoopIndex, false, LoopRegHashes, MemReadExprs, MemReadExprWidths, StackPtrCopiesVector);
		if (MemRangeSuccess) {
			SMP_msg("INFO: LOOP: Computed %zu LoopMemReadExprs for loop %d in function at %llx\n",
				MemReadExprs.size(), LoopIndex, (uint64_t) this->GetFirstFuncAddr());
			size_t MemListIndex = 0; // for debugging and diagnostic output
			for (STARSMemWriteExprListIter MemListIter = MemReadExprWidths.begin(); MemListIter != MemReadExprWidths.end(); ++MemListIter, ++MemListIndex) {
				STARSExprSetIter ExprIter = (*MemListIter).first;
				STARSExpression *MemAddressRangeExpr = (*ExprIter);
				if (VerboseOutput) {
					SMP_msg("INFO: LOOP: LoopMemRead EXPR:");
					MemAddressRangeExpr->Dump(0);
					if (MemAddressRangeExpr->IsStackPtrRegUsed()) {
						SMP_msg("INFO: LOOP: LoopMemRead EXPR is stack-based.\n");
					}
				}
				// Create the lower and upper bounds of the memory reading range.
				STARSExpression *MemReadLowerBound = MemAddressRangeExpr->Clone(); // deep copy, for editing
				bool changed = false;
				MemReadLowerBound->ResetAllProcessed();
				bool success = this->ReplaceAllIVsWithExprs(LoopIndex, MemReadLowerBound, true, changed);
				if (success) {
					(void) MemReadLowerBound->SimplifyDriver();
					STARSExpression *MemReadUpperBound = MemAddressRangeExpr->Clone(); // deep copy, for editing
					MemReadUpperBound->ResetAllProcessed();
					bool LimitSuccess = this->ReplaceAllIVsWithExprs(LoopIndex, MemReadUpperBound, false, changed);
					if (LimitSuccess) {
						(void) MemReadUpperBound->SimplifyDriver();
					}
					else {
						// We will emit annotations that have a useful starting address for the
						//  memory range read, and a code that indicates the range limit is unknown.
						delete MemReadUpperBound;
						MemReadUpperBound = nullptr;
						SMP_msg("INFO: Failure to ReplaceIVsWithLimitExpr for read MemListIndex %zu loop %zu in function at %llx\n",
							MemListIndex, LoopIndex, (uint64_t) this->GetFirstFuncAddr());
					}

					int StackPtrCopiesVecIndex = (*MemListIter).second.second;
					if (VerboseOutput) {
						SMP_msg("INFO: LOOP: MemReadLowerBound EXPR:");
						MemReadLowerBound->Dump(0);
						if (LimitSuccess) {
							SMP_msg("INFO: LOOP: MemReadUpperBound EXPR:");
							MemReadUpperBound->Dump(0);
						}
					}

					// Clone the lower and upper bounds expr for inheritance into callers (future SPARK).
					STARSExpression *MemReadLowerBoundExpanded = MemReadLowerBound->Clone();
					STARSExpression *MemReadUpperBoundExpanded = nullptr;
					if (LimitSuccess)
						MemReadUpperBoundExpanded = MemReadUpperBound->Clone();

					// For annotation output, we want to expand the expressions beyond loop boundaries,
					//  hopefully as far as incoming args or constant initializers.
					set<STARS_ea_t> NewStackPtrCopiesSet;
					success = this->ExpandExprToInArgs(LoopIndex, MemReadLowerBoundExpanded, true, NewStackPtrCopiesSet);
					if (LimitSuccess) {
						LimitSuccess = this->ExpandExprToInArgs(LoopIndex, MemReadUpperBoundExpanded, false, NewStackPtrCopiesSet);
					}
					if (success) {
						if (VerboseOutput) {
							SMP_msg("INFO: LOOP: MemReadLowerBoundExpanded EXPR:");
							MemReadLowerBoundExpanded->Dump(0);
							if (LimitSuccess) {
								SMP_msg("INFO: LOOP: MemReadUpperBoundExpanded EXPR:");
								MemReadUpperBoundExpanded->Dump(0);
							}
						}

						set<STARS_ea_t> UnionedStackPtrCopiesSet;
						insert_iterator<set<STARS_ea_t> > EndIter = std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin());
						if (0 <= StackPtrCopiesVecIndex) { // vector has entry for a set
							EndIter = merge(StackPtrCopiesVector[StackPtrCopiesVecIndex].begin(),
								StackPtrCopiesVector[StackPtrCopiesVecIndex].end(),
								NewStackPtrCopiesSet.begin(), NewStackPtrCopiesSet.end(),
								std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin()));
						}
						else { // just copy the NewStackPtrCopiesSet
							EndIter = std::copy(NewStackPtrCopiesSet.begin(), NewStackPtrCopiesSet.end(),
								std::inserter(UnionedStackPtrCopiesSet, UnionedStackPtrCopiesSet.begin()));
						}

						// Substitute global min and max values for global mem reads that
						//  are still in the expr, and see if we can simplify as a result.
						bool FinalChange = MemReadLowerBoundExpanded->SubstituteGlobalAndArgRangeValues(true);
						if (FinalChange)
							(void) MemReadLowerBoundExpanded->SimplifyDriver();
						if (LimitSuccess) {
							FinalChange = MemReadUpperBoundExpanded->SubstituteGlobalAndArgRangeValues(false);
							if (FinalChange)
								(void) MemReadUpperBoundExpanded->SimplifyDriver();
						}

						// To assist in stack transformation security defenses, if the 
						//  memory range expression is stack-relative, we emit an annotation
						//  that identifies the memory range written to at InstAddr.
						// The annotation should look like:
						//  [hex InstAddr] [InstSize] INSTR STACKMEMRANGE MIN [ESP-k] MAX [ESP-j] ZZ
						set<STARS_ea_t> StaticMemLeaAddrSet;
						bool StackAnnotation = false;
						if (LimitSuccess && MemReadLowerBoundExpanded->IsStackPtrPlusOffset() && MemReadUpperBoundExpanded->IsStackPtrPlusOffset()) {
							this->EmitStackMemRangeAnnotations(MemReadLowerBoundExpanded, MemReadUpperBoundExpanded, PositiveIncrement, UnionedStackPtrCopiesSet, StaticMemLeaAddrSet);
							StackAnnotation = true;
						}
						if (!StaticMemLeaAddrSet.empty()) {
							SMP_msg("INFO: LOOP: StaticMemLeaAddrSet size (reads): %zu \n", StaticMemLeaAddrSet.size());
						}

						// To assist in defenses related to static global memory, output ranges
						//  of memory written when the lower and upper bounds have simplified to
						//  constants.
						if (MemReadLowerBoundExpanded->IsConstExpr()) {
							SMPInstr *MemReadInst = MemReadLowerBoundExpanded->GetOriginalParentInst();
							STARS_ea_t MemReadAddr = MemReadInst->GetAddr();
							if (StaticMemReadAddrsEmitted.find(MemReadAddr) == StaticMemReadAddrsEmitted.cend()) {
								if (LimitSuccess && (MemReadAddr != MemReadUpperBoundExpanded->GetOriginalParentInst()->GetAddr())) {
									SMP_msg("ERROR: LOOP: (reads) Constant lower and upper bound exprs don't have same InstAddr for loop %zu ", LoopIndex);
									this->DumpFuncNameAndAddr();
								}
								else {
									int InstSize = MemReadInst->GetSize();
									STARS_uval_t LowerBoundVal = MemReadLowerBoundExpanded->GetConstLeftOperand()->GetImmedValue();
									STARS_uval_t UpperBoundVal = (STARS_uval_t)0xfffffffffffffffe; // Code for unknown limit value
									if (LimitSuccess
										&& (nullptr != MemReadUpperBoundExpanded->GetConstLeftOperand())
										&& (MemReadUpperBoundExpanded->GetConstLeftOperand()->IsImmedOp())) {
										UpperBoundVal = MemReadUpperBoundExpanded->GetConstLeftOperand()->GetImmedValue();
									}
									FILE *InfoAnnotFile = global_STARS_program->GetInfoAnnotFile();
									// NOTE: Replace WRITE with READ or RANGE in annotation
									SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR STATICMEMREAD MIN ", (uint64_t) MemReadAddr, InstSize);
									if (LowerBoundVal == UpperBoundVal) {
										// A simple read from a static mem address with no index registers has
										//  the same expr for lower bound and upper bound. Use the ByteWidth to 
										//  expand the upper bound.
										uint16_t ByteWidth = MemReadLowerBoundExpanded->GetConstLeftOperand()->GetByteWidth();
										UpperBoundVal += (STARS_uval_t) ByteWidth;
									}
									if (LowerBoundVal < UpperBoundVal) {
										SMP_fprintf(InfoAnnotFile, "%llx LIMIT %llx ZZ %s\n",
											(uint64_t) LowerBoundVal, (uint64_t) UpperBoundVal, MemReadInst->GetDisasm());
									}
									else {
										SMP_fprintf(InfoAnnotFile, "%llx LIMIT %llx ZZ %s\n",
											(uint64_t) UpperBoundVal, (uint64_t) LowerBoundVal, MemReadInst->GetDisasm());
									}
									(void) StaticMemReadAddrsEmitted.insert(MemReadAddr);
								}
							}
						}
						else if (!StackAnnotation) {
							// LowerBoundExpanded expr was neither static mem nor stack.
							SMP_msg("INFO: LOOP: Read expr neither static mem nor stack mem: \n");
							if (VerboseOutput) {
								MemReadLowerBoundExpanded->Dump(0);
							}
						}
					}
					else {
						SMP_msg("ERROR: LOOP: (reads) ExpandExprToInArg failure for loop %zu ", LoopIndex);
						this->DumpFuncNameAndAddr();
						delete MemReadLowerBoundExpanded;
						delete MemReadUpperBoundExpanded;
						// this->LoopAnalysisProblems[LoopIndex] = true;
					}
				}
				else {
					SMP_msg("ERROR: LOOP: Failure to ReplaceIVsWithInitExpr (reads) for loop %d ", LoopIndex);
					this->DumpFuncNameAndAddr();
					delete MemReadLowerBound;
					// this->LoopAnalysisProblems[LoopIndex] = true;
					// this->LoopWritesGlobalStaticMemory[LoopIndex] = false; // reset
				}
			}
		}
		else {
			if (this->LoopReadsMemory[LoopIndex]) {
				SMP_msg("ERROR: LOOP: Failure to CreateSPARKMemRangeExpr (reads) for loop %d in function at %llx\n",
					LoopIndex, (uint64_t) this->GetFirstFuncAddr());
				// this->LoopAnalysisProblems[LoopIndex] = true;
				// this->LoopWritesGlobalStaticMemory[LoopIndex] = false; // reset
			}
			else {
				SMP_msg("INFO: LOOP: No memory reads in loop %d in function at %llx\n",
					LoopIndex, (uint64_t) this->GetFirstFuncAddr());
			}
		}
		MemReadExprWidths.clear();
		MemReadExprs.clear();

	} // end for all loops: loop4, memory range analysis

	// Prepare for later SPARK Ada output.
	this->DetectInArgRegsNeededForMemWriteExprs();
	this->ExpandLoopRegHashExprs();

	// Last chance to get annotations for memory accesses:
	//  See if we can trace an index reg for an indexed
	//  StaticMemOp back to a constant initial value.
	for (SMPBasicBlock *CurrBlock : this->RPOBlocks) {
		if (CurrBlock->HasIndirectMemRead() || CurrBlock->HasIndirectMemWrite()) {
			for (vector<SMPInstr *>::const_iterator InstIter = CurrBlock->GetFirstConstInst();
				InstIter != CurrBlock->GetLastConstInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (CurrInst->IsMarkerInst())
					continue;

				STARS_ea_t InstAddr = CurrInst->GetAddr();
				STARSOpndTypePtr MemOp = nullptr;
				// See if we have a memory access with no annotation emitted.
				if (CurrInst->HasIndirectMemoryRead()) {
					if (StaticMemReadAddrsEmitted.find(InstAddr) == StaticMemReadAddrsEmitted.cend()) {
						MemOp = CurrInst->GetMemUse();
					}
				}
				else if (CurrInst->HasIndirectMemoryWrite()) {
					if (StaticMemWriteAddrsEmitted.find(InstAddr) == StaticMemWriteAddrsEmitted.cend()) {
						MemOp = CurrInst->GetMemDef();
					}
				}
				if (MemOp != nullptr) {
					int BaseReg, IndexReg;
					uint16_t Scale;
					STARS_ea_t Offset;
					MDExtractAddressFields(MemOp, BaseReg, IndexReg, Scale, Offset);
					STARS_uval_t BaseRegValue = 0;
					STARS_uval_t IndexRegValue = 0;
					bool LocalName;
					int SSANum;
					bool success1 = true;
					bool success2 = true;
					STARSOpndTypePtr DefMoveOp = nullptr;
					if (STARS_x86_R_none != BaseReg) {
						STARSOpndTypePtr BaseRegOp = CurrInst->MakeRegOpnd((STARS_regnum_t)BaseReg);
						STARSDefUseIter UseIter = CurrInst->FindUse(BaseRegOp);
						assert(UseIter != CurrInst->GetLastUse());
						LocalName = CurrBlock->IsLocalName(BaseRegOp);
						SSANum = UseIter->GetSSANum();
						success1 = CurrBlock->GetUltimateInitValue(BaseRegOp, InstAddr, SSANum, LocalName, DefMoveOp, BaseRegValue);
					}
					if (success1 && (STARS_x86_R_none != IndexReg)) {
						STARSOpndTypePtr IndexRegOp = CurrInst->MakeRegOpnd((STARS_regnum_t)IndexReg);
						STARSDefUseIter UseIter = CurrInst->FindUse(IndexRegOp);
						assert(UseIter != CurrInst->GetLastUse());
						LocalName = CurrBlock->IsLocalName(IndexRegOp);
						SSANum = UseIter->GetSSANum();
						success2 = CurrBlock->GetUltimateInitValue(IndexRegOp, InstAddr, SSANum, LocalName, DefMoveOp, IndexRegValue);
						if (success2 && (Scale > 0)) {
							IndexRegValue = (IndexRegValue << Scale);
						}
					}
					if (success1 && success2) {
						STARS_ea_t FinalAddr = Offset + (STARS_ea_t)BaseRegValue + (STARS_ea_t)IndexRegValue;
						if (IsImmedGlobalAddress(FinalAddr)) {
							size_t InstSize = CurrInst->GetSize();
							STARS_uval_t UpperBoundVal = (STARS_uval_t)0xfffffffffffffffe; // Code for unknown limit value
							if (CurrInst->HasIndirectMemoryWrite()) {
								SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR STATICMEMWRITE MIN %llx LIMIT %llx ZZ %s\n",
									(uint64_t)InstAddr, InstSize, (uint64_t)FinalAddr, (uint64_t)UpperBoundVal, CurrInst->GetDisasm());
							}
							else {
								assert(CurrInst->HasIndirectMemoryRead());
								SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR STATICMEMREAD MIN %llx LIMIT %llx ZZ %s\n",
									(uint64_t)InstAddr, InstSize, (uint64_t)FinalAddr, (uint64_t)UpperBoundVal, CurrInst->GetDisasm());
							}
						}
					}
				}
			} // end for all insts in block
		}
	} // end for all blocks

	if (!global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
		// Not inheriting expressions across call boundaries, so
		//  deallocate memory.
		this->DestroyLoopExprs();
	}
	return;
} // end of SMPFunction::AnalyzeLoopIterations()

// Helper for SSA subscript renumbering: return the next SSA number for the global name
//  and increment the SSACounter to prepare the next number. Push the returned number onto
//  the SSAStack for the global name.
int SMPFunction::SSANewNumber(std::size_t GlobNameIndex) {
	int Subscript = this->SSACounter[GlobNameIndex];
	++(this->SSACounter[GlobNameIndex]);
	// We will keep the highest SSANum in the MaxDirectStackAccessSSANum. Could get
	//  more precise and track reg and stack SSANum maxima separately, but that would
	//  slow down this method because we don't know what kind of operand GlobNameIndex
	//  refers to.
	if (this->GetMaxStackSSANum() < this->SSACounter[GlobNameIndex]) {
		this->SetMaxStackSSANum(this->SSACounter[GlobNameIndex]);
		assert(this->SSACounter[GlobNameIndex] < 65500);
	}
	this->SSAStack[GlobNameIndex].push_back(Subscript);
	return Subscript;
} // end of SMPFunction::SSANewNumber()

#define STARS_TRACK_STACK_DEF_ADDRS 1

// Main helper for SSA subscript renumbering. Renumber within block throughout its phi
//  functions, then its DEFs and USEs, then its phi successors. Recurse then on all
//  successors in the dominator tree.
void SMPFunction::SSARename(int BlockNumber) {
	assert(0 <= BlockNumber);
	assert(BlockNumber < this->BlockCount);

	SMPBasicBlock *CurrBlock = this->RPOBlocks[(size_t) BlockNumber];

	bool DumpFlag = false;
#if SMP_DEBUG_DATAFLOW_VERBOSE
	DumpFlag |=	(0 == strcmp("main", this->GetFuncName()));
	DumpFlag |= (0 == strcmp("dohanoi", this->GetFuncName()));
	DumpFlag |= (0 == strcmp("uw_frame_state_for", this->GetFuncName()));
	DumpFlag |= (0 == strcmp("_IO_sputbackc", this->GetFuncName()));
#endif

	if (DumpFlag) SMP_msg("Entered SSARename for block number %d\n", BlockNumber);

	// For each phi function at the top of the block, rename the DEF of the phi function
	//  using SSANewNumber() on the global name index.
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	list<SMPPhiFunction> TempPhiList;
	int GlobalNameIndex;
	for (CurrPhi = CurrBlock->GetFirstPhi(); CurrPhi != CurrBlock->GetLastPhi(); ++CurrPhi) {
		STARSOpndTypePtr PhiDefOp = CurrPhi->GetAnyOp();
		GlobalNameIndex = CurrPhi->GetIndex();
		assert(0 <= GlobalNameIndex);
		int NewSSANum = this->SSANewNumber((std::size_t) GlobalNameIndex);

		// Cannot change the C++ STL set item directly, as sets might become unordered.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetSSADef(NewSSANum);
		TempPhiList.push_back(TempPhi);

		bool RegOpFlag = PhiDefOp->IsRegOp();
		if (RegOpFlag || MDIsDirectStackAccessOpnd(PhiDefOp, this->UsesFramePointer())) {
			if (DumpFlag && RegOpFlag && PhiDefOp->MatchesReg(STARS_x86_R_ax)) {
				SMP_msg("New EAX Phi Def SSANum: %d Block %d\n", NewSSANum, BlockNumber);
			}
			// Map the final SSA number to the block number.
			if (RegOpFlag) {
				int DefHashValue = HashGlobalNameAndSSA(PhiDefOp, NewSSANum);
				pair<int, STARS_ea_t> DefMapEntry(DefHashValue, (STARS_PSEUDO_ID_MIN + CurrBlock->GetNumber()));
				pair<map<int, STARS_ea_t>::iterator, bool> MapReturnValue = this->GlobalDefAddrBySSA.insert(DefMapEntry);
				assert(MapReturnValue.second);
			}
#if STARS_TRACK_STACK_DEF_ADDRS
			else {
				int64_t DefHashValue = HashGlobalStackNameAndSSA(PhiDefOp, NewSSANum, this->UsesFramePointer());
				pair<int64_t, STARS_ea_t> DefMapEntry(DefHashValue, ((int64_t)(STARS_PSEUDO_ID_MIN + CurrBlock->GetNumber())));
				pair<map<int64_t, STARS_ea_t>::iterator, bool>  MapReturnValue = this->GlobalStackDefAddrBySSA.insert(DefMapEntry);
				assert(MapReturnValue.second);
			}
#endif
		}
	}
	// Go back through the Phi function set and replace the items that need to be updated.
	list<SMPPhiFunction>::iterator TempIter;
	for (TempIter = TempPhiList.begin(); TempIter != TempPhiList.end(); ++TempIter) {
		// Use the op_t from the first phi use, because they are all the same.
		bool Erased = CurrBlock->ErasePhi(TempIter->GetPhiRef(0).GetOp());
		assert(Erased);
		// Now we can add back the phi function that had the DEF SSA number changed.
		bool Added = CurrBlock->AddPhi(*TempIter);
		assert(Added);
	}
	TempPhiList.clear();
	if (DumpFlag) SMP_msg("Processed phi functions at top.\n");

	// For each instruction in the block, rename all global USEs and then all global DEFs.
	vector<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
		CurrInst = (*InstIter);
		set<DefOrUse, LessDefUse>::iterator CurrUse = CurrInst->GetFirstUse();
		STARS_ea_t InstAddr = CurrInst->GetAddr(); // for debugging break points
		while (CurrUse != CurrInst->GetLastUse()) {
			// See if Use is a global name.
			STARSOpndTypePtr UseOp = CurrUse->GetOp();
			STARSOpndSetIter GlobIter = this->GlobalNames.find(UseOp);
			if (GlobIter != this->GlobalNames.end()) { // found it
				unsigned int GlobIndex = ExtractGlobalIndex(*GlobIter);
				if (GlobIndex > this->SSAStack.size()) {
					// Get some debug info out to the log file before we crash.
					SMP_msg("FATAL ERROR: Bad GlobIndex: %u at %llx ", GlobIndex, (uint64_t) InstAddr);
					this->DumpFuncNameAndAddr();
					exit(EXIT_FAILURE);
				}
				// Set the SSA number for this use to the top of stack SSA # (back())
				int NewSSANum;
				if (this->SSAStack[GlobIndex].empty()) {
					// No top of stack entry to read.
#if SMP_DEBUG_UNINITIALIZED_SSA_NAMES
					if (!CurrInst->MDIsPopInstr() && (UseOp->IsRegOp())) {
						// POP uses the stack offset and generates spurious
						//  uninitialized variable messages for [esp+0].
						SMP_msg("WARNING: function %s : Use of uninitialized variable: ",
							this->GetFuncName());
						SMP_msg(" Variable: ");
						PrintListOperand(*GlobIter);
						SMP_msg(" Block number: %d Address: %llx Instruction: %s\n", BlockNumber,
							(uint64_t) CurrInst->GetAddr(), CurrInst->GetDisasm());
					}
#endif
					NewSSANum = SMP_SSA_UNINIT;
				}
				else {
					NewSSANum = this->SSAStack[GlobIndex].back();
				}
				CurrUse = CurrInst->SetUseSSA(UseOp, NewSSANum);
				if (DumpFlag && (UseOp->IsRegOp()) && UseOp->MatchesReg(STARS_x86_R_ax)) {
					SMP_msg("New EAX Use SSANum: %d at %llx\n", NewSSANum, (uint64_t) CurrInst->GetAddr());
				}
			}
			++CurrUse;
		} // end for all USEs
		set<DefOrUse, LessDefUse>::iterator CurrDef = CurrInst->GetFirstDef();
		while (CurrDef != CurrInst->GetLastDef()) {
			// See if Def is a global name.
			STARSOpndTypePtr DefOp = CurrDef->GetOp();
			STARSOpndSetIter GlobIter = this->GlobalNames.find(DefOp);
			if (GlobIter != this->GlobalNames.end()) { // found it
				unsigned int GlobIndex = ExtractGlobalIndex(*GlobIter);
				// Set the SSA number for this DEF to the SSANewNumber top of stack
				int NewSSANum = this->SSANewNumber(GlobIndex);
				CurrDef = CurrInst->SetDefSSA(DefOp, NewSSANum);
				bool RegOpFlag = DefOp->IsRegOp();
				if (RegOpFlag || MDIsDirectStackAccessOpnd(DefOp, this->UsesFramePointer())) {
					STARS_ea_t DefAddr = InstAddr;
					if (DumpFlag && RegOpFlag && DefOp->MatchesReg(STARS_x86_R_ax)) {
						SMP_msg("New EAX Def SSANum: %d at %llx\n", NewSSANum, (uint64_t) DefAddr);
					}

					// Map the final SSA number to the DEF address.
					if (RegOpFlag) {
						int DefHashValue = HashGlobalNameAndSSA(DefOp, NewSSANum);
						pair<int, STARS_ea_t> DefMapEntry(DefHashValue, DefAddr);
						pair<map<int, STARS_ea_t>::iterator, bool> MapReturnValue = this->GlobalDefAddrBySSA.insert(DefMapEntry);
						assert(MapReturnValue.second);
					}
#if STARS_TRACK_STACK_DEF_ADDRS
					else {
						int64_t DefHashValue = HashGlobalStackNameAndSSA(DefOp, NewSSANum, this->UsesFramePointer());
						pair<int64_t, STARS_ea_t> DefMapEntry(DefHashValue, DefAddr);
						pair<map<int64_t, STARS_ea_t>::iterator, bool> MapReturnValue = this->GlobalStackDefAddrBySSA.insert(DefMapEntry);
						assert(MapReturnValue.second);
					}
#endif
				}
			}
			++CurrDef;
		} //  end for all DEFs
	} // end for all instructions
	if (DumpFlag) SMP_msg("Processed all instructions.\n");

	// For all control flow graph (not dominator tree) successors, fill in the current
	//  (outgoing) SSA number in the corresponding USE slot in the phi function, for all
	//  global names appearing in phi functions.
	list<SMPBasicBlock *>::iterator SuccIter;
	for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
		// What position in the Preds list of this successor is CurrBlock?
		int ListPos = (*SuccIter)->GetPredPosition(BlockNumber);
		assert(0 <= ListPos);

		// Go through all phi functions in this successor. At ListPos position in the
		//  incoming arguments for that phi function, set the SSA number to the SSA number
		//  in the top of stack entry for the global name associated with that phi function.
		set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
		for (CurrPhi = (*SuccIter)->GetFirstPhi(); CurrPhi != (*SuccIter)->GetLastPhi(); ++CurrPhi) {
			int GlobIndex = CurrPhi->GetIndex();
			int CurrSSA;
			if (this->SSAStack[GlobIndex].empty()) {
				// No top of stack entry to read.
#if SMP_DEBUG_UNINITIALIZED_SSA_NAMES
				SMP_msg("WARNING: function %s : Path to use of uninitialized variable: ",
					this->GetFuncName());
				SMP_msg(" Variable: ");
				PrintListOperand(CurrPhi->GetAnyOp());
				SMP_msg(" Block number: %d Successor block number: %d\n", BlockNumber,
					(*SuccIter)->GetNumber());
#endif
				CurrSSA = SMP_SSA_UNINIT;
			}
			else {
				CurrSSA = this->SSAStack[GlobIndex].back();  // fetch from top of stack
			}
			SMPPhiFunction TempPhi = (*CurrPhi);
			TempPhi.SetSSARef(ListPos, CurrSSA);
			TempPhiList.push_back(TempPhi);
			if (DumpFlag && (BlockNumber >= 3) && (BlockNumber <= 4)) {
				SMP_msg("BlockNumber: %d  ListPos: %d\n", BlockNumber, ListPos);
			}
		} // end for all phi functions in successor
		// Go back through the Phi function set and replace the items that need to be updated.
		for (TempIter = TempPhiList.begin(); TempIter != TempPhiList.end(); ++TempIter) {
#if 0
			if (DumpFlag && (BlockNumber >= 3) && (BlockNumber <= 4)) {
				SMP_msg("Special before phi dump:\n");
				set<SMPPhiFunction, LessPhi>::iterator FoundPhi;
				FoundPhi = (*SuccIter)->FindPhi(TempIter->GetAnyOp());
				FoundPhi->Dump();
			}
#endif
			// Use the op_t from the first phi use, because they are all the same.
			bool Erased = (*SuccIter)->ErasePhi(TempIter->GetPhiRef(0).GetOp());
			assert(Erased);
			// Now we can add back the phi function that had one SSA number changed.
			bool Added = (*SuccIter)->AddPhi(*TempIter);
			assert(Added);
			if (DumpFlag && (BlockNumber >= 3) && (BlockNumber <= 4)) {
				SMP_msg("Special after phi dump:\n");
				set<SMPPhiFunction, LessPhi>::iterator FoundPhi;
				FoundPhi = (*SuccIter)->FindPhi(TempIter->GetAnyOp());
				FoundPhi->Dump();
			}
		}
		TempPhiList.clear();
	} // end for all successors of CurrBlock
	if (DumpFlag) SMP_msg("Processed successor phi functions.\n");

	// For each operand in the LiveOut set, create a mapping from operand to SSA number.
	for (STARSOpndSetIter LiveOutIter = CurrBlock->GetFirstLiveOut(); LiveOutIter != CurrBlock->GetLastLiveOut(); ++LiveOutIter) {
		STARSOpndTypePtr LiveOutOp = (*LiveOutIter);
		unsigned int GlobIndex = LiveOutOp->GetOpGlobalIndex();
		int CurrSSA;
		if (this->SSAStack[GlobIndex].empty()) {
			// No top of stack entry to read.
			SMP_msg("WARNING: Uninit var, Block: %d GlobIndex: %d Operand:", BlockNumber, GlobIndex);
			PrintOperand(LiveOutOp);
			SMP_msg("\n");
			CurrSSA = 0; // SSA Marker rung will DEF the uninitialized variable
		}
		else {
			CurrSSA = this->SSAStack[GlobIndex].back();  // fetch from top of stack
		}
		pair<map<STARSOpndTypePtr, int, STARSLessOp>::iterator, bool> InsertResult = CurrBlock->AddLiveOutSSAMap(LiveOutOp, CurrSSA);
		assert(InsertResult.second);
	}

	// For each successor in the dominator tree, recurse.
	for (int ChildBlockNum : this->DomTree[BlockNumber].second) {
			this->SSARename(ChildBlockNum);
	}
	if (DumpFlag) SMP_msg("Finished recursion.\n");

	// Pop off all SSAStack entries pushed during this block. I.e. for each global name,
	//  pop its SSAStack once per DEF and once per phi function in this block.
	for (CurrPhi = CurrBlock->GetFirstPhi(); CurrPhi != CurrBlock->GetLastPhi(); ++CurrPhi) {
		GlobalNameIndex = CurrPhi->GetIndex();
		this->SSAStack[(size_t) GlobalNameIndex].pop_back();
	}
	if (DumpFlag) SMP_msg("Popped off entries due to phi functions.\n");
	for (InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
		set<DefOrUse, LessDefUse>::iterator CurrDef;
		CurrInst = (*InstIter);
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			// See if DEF is a global name.
			STARSOpndSetIter GlobIter = this->GlobalNames.find(CurrDef->GetOp());
			if (GlobIter != this->GlobalNames.end()) { // found it
				unsigned int GlobIndex = ExtractGlobalIndex(*GlobIter);
				this->SSAStack[(size_t) GlobIndex].pop_back();
			}
		} //  end for all DEFs
	} // end for all instructions
	if (DumpFlag) { 
		SMP_msg("Popped off entries due to instructions.\n");
	}

	return;
} // end of SMPFunction::SSARename()

// Main driver of SSA subscript renumbering.
void SMPFunction::SSARenumber(void) {
	bool DumpFlag = false;
	this->SetMaxRegSSANum(0);
	this->SetMaxStackSSANum(0);
	this->MaxLocalSSANum = 0;
#if 0
	DumpFlag |= (0 == strcmp("_IO_sputbackc", this->GetFuncName()));
#endif

	if (this->GlobalNames.empty())
		return;  // no names to renumber

	// Initialize stacks and counters of SSA numbers.
	std::size_t GlobIndex;
	assert(0 == this->SSACounter.size());
	for (GlobIndex = 0; GlobIndex < this->GlobalNames.size(); ++GlobIndex) {
		list<int> DummyList;
		this->SSACounter.push_back(0);
		this->SSAStack.push_back(DummyList);
	}

	// Recurse through the dominator tree starting with node 0.
	this->SSARename(0);
	if (DumpFlag)
		this->Dump(false);
	SMP_msg("INFO: Max SSANum used for global names: %d\n", this->GetMaxStackSSANum());
	return;
} // end of SMPFunction::SSARenumber()

// Analyze system calls for memory overwriting safety
void SMPFunction::AnalyzeSystemCalls(void) {
	if (this->HasUnresolvedIndirectCalls())  // already going to be UNSAFE
		return;

	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
		if (CurrBlock->HasCallInstruction()) {
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if ((INDIR_CALL == CurrInst->GetDataFlowType()) && CurrInst->MDIsSystemCall()) {
					STARSOpndTypePtr EAXOpnd = CurrInst->MakeRegOpnd(STARS_x86_R_ax);
					STARSDefUseIter UseIter = CurrInst->FindUse(EAXOpnd);
					if (UseIter != CurrInst->GetLastUse()) {
						// EAX encodes the syscall number
						bool GlobalName = this->IsGlobalName(EAXOpnd);
						STARS_ea_t EAXDefAddr = STARS_BADADDR;
						if (GlobalName) {
							EAXDefAddr = this->GetGlobalDefAddr(EAXOpnd, UseIter->GetSSANum());
						}
						else {
							EAXDefAddr = CurrBlock->GetDefAddrFromUseAddr(EAXOpnd, CurrInst->GetAddr(), UseIter->GetSSANum(), true);
						}
						bool UnresolvedCallee = ((STARS_BADADDR == EAXDefAddr) || (STARS_IsBlockNumPseudoID(EAXDefAddr))
							|| (STARS_IsSSAMarkerPseudoID(EAXDefAddr)));
						SMP_msg("INFO: System call number def addr at %llx\n", (uint64_t) EAXDefAddr);
						STARS_uval_t SysCallNumber = 0;
						if (!UnresolvedCallee) {
							// We have an instruction addr in EAXDefAddr.
							SMPInstr *EAXDefInst = this->GetInstFromAddr(EAXDefAddr);
							if (EAXDefInst == nullptr)
								UnresolvedCallee = true;

							STARSDefUseIter UseIter;
							if (!UnresolvedCallee) {
								UseIter = EAXDefInst->GetFirstUse(); 
								// check that there's at least 1 use.
								UnresolvedCallee = (UseIter == EAXDefInst->GetLastUse());
							}
							if (!UnresolvedCallee) {
								SMP_msg("INFO: System call set of EAX.  Use found.  at: %llx\n", (uint64_t) EAXDefAddr);
								STARSOpndTypePtr UseOp = UseIter->GetOp();

								// if UseOp is nullptr or not an immediate, callee cannot be resolved.
								if (UseOp == nullptr)
									UnresolvedCallee = true;
								else
									UnresolvedCallee = (!UseOp->IsImmedOp());
				
								// Resolve if UseOp is an immediate.
								if (!UnresolvedCallee) {
									SysCallNumber = UseOp->GetImmedValue();
									SMP_msg("INFO: System call set of EAX.  Use was immed. %d  at: %llx\n", SysCallNumber, (uint64_t) EAXDefAddr);
									++UseIter;

									// check there is not a second use.
									UnresolvedCallee = (UseIter != EAXDefInst->GetLastUse());
								}
							}
							if (!UnresolvedCallee) {
								SMP_msg("INFO: System call set of EAX determined solid. number: %u at: %llx\n", 
									(unsigned int) SysCallNumber, (uint64_t) EAXDefAddr);
							}
						}
						if (!UnresolvedCallee) { // we have a good SysCallNumber
#define CGC_RECEIVE_SYSCALL 3
							UnresolvedCallee = ((CGC_RECEIVE_SYSCALL == SysCallNumber) && global_stars_interface->IsCGCBinary());
							// NOTE: Analyze Linux system calls also.
						}
						if (UnresolvedCallee) {
							this->SetHasUnresolvedIndirectCalls();
							return;
						}
					}
				} // end if system call
			} // end for all insts in block
		} // end if block has call instruction
	} // end for all blocks
} // end of SMPFunction::AnalyzeSystemCalls()

// Find variables that are USEd before they are DEFed
void SMPFunction::DetectUninitializedVars(void) {
	if (this->StackPtrAnalysisSucceeded()) {
		bool UseFP = this->UsesFramePointer();
		bool Binary32 = (global_STARS_program->GetSTARS_ISA_Bitwidth() <= 32);
		FILE *WarningsFile = global_STARS_program->GetUninitVarFile();
		assert(nullptr != WarningsFile);
		// Look at stack vars with negative offsets in the SSA Marker inst
		list<SMPInstr *>::iterator InstIter = this->GetFirstInstIter();
		SMPInstr *MarkerInst = (*InstIter);
		assert(MarkerInst->IsMarkerInst());
#if 0   // MarkerInst only has normalized SSA names, never needed normalization
		assert(MarkerInst->AreDefsNormalized());
#endif
		for (STARSDefUseIter DefIter = MarkerInst->GetFirstDef(); DefIter != MarkerInst->GetLastDef(); ++DefIter) {
			STARSOpndTypePtr DefOp = DefIter->GetOp();
			if (MDIsDirectStackAccessOpnd(DefOp, UseFP)) {
				STARS_ea_t Offset = DefOp->GetAddr();
				int SignedOffset = (int) Offset;
				if (0 > SignedOffset) {
					SMP_msg("WARNING: Uninitialized stack var, offset %d, func at %llx\n", SignedOffset, (uint64_t) this->GetFirstFuncAddr());
				}
			}
			else if (Binary32 && DefOp->IsRegOp()) {
				// Pushes to save regs that are restored later will create a USE before DEF. Screen those out.
				if (!this->IsRegPreserved((size_t)DefOp->GetReg())) {
					SMP_msg("WARNING: Uninitialized reg, reg # %d, func at %llx\n", DefOp->GetReg(), (uint64_t) this->GetFirstFuncAddr());
				}
			}
		}
	}

	return;
} // end of SMPFunction::DetectUninitializedVars()

// Emit debugging output for analyzing time spent in InferTypes() ?
#define SMP_ANALYZE_INFER_TYPES_TIME 0

// Main driver for the type inference system.
void SMPFunction::InferTypes(bool FirstIter) {
	// The type inference system is an iteration over four analysis steps, until
	//  a fixed point is reached:
	// 1) Within an instruction, set types of operators based on the operator type,
	//     the operand types, and the instruction type category, and propagate the
	//     type of the SMP_ASSIGN operator to its DEF.
	// 2) Propagate the type of a DEF along its SSA chain to all USEs of that SSA name.
	// 3) If all USEs of an SSA name have the same type, but the DEF has no type,
	//     then infer that the DEF must have the same type.
	// 4) If all references to a memory location have the same type, mark that memory
	//     location as having that type, if no aliasing occurs.
	//
	// The type inference system will mark DEFs and USEs in each instruction's DEF and USE
	//  sets with an inferred type. This inference on USEs is not conclusive for other USEs
	//  outside of that instruction. For example, a pointer could be read in from memory
	//  and used as a pointer, then hashed using an arithmetic operation. If the arithmetic
	//  operation always treats its source operands as NUMERIC and produces a NUMERIC
	//  result, e.g. SMP_BITWISE_XOR, then the USE of that pointer is NUMERIC within
	//  this xor instruction. If the DEF at the beginning of the SSA chain for the pointer
	//  is eventually marked as POINTER, then all USEs in the chain will be marked POINTER
	//  as well (see step 2 above). This inconsistency along the USE chain is perfectly
	//  acceptable in our type system. It is important to mark the USEs according to how
	//  we observe them being used, because consistent USEs will propagate back up to
	//  the DEF in step 3 above.

	bool changed;
	bool NewChange = false;
#if SMP_ANALYZE_INFER_TYPES_TIME
	bool DebugFlag2 = false;
	DebugFlag2 |= (0 == strcmp("Option", this->GetFuncName()));
	long NewChangeCount;
	long IterationCount = 0;
#endif
#if SMP_DEBUG_TYPE_INFERENCE
	bool DebugFlag = false;
	DebugFlag |= (0 == strcmp("__libc_csu_init", this->GetFuncName()));
#endif
	list<SMPInstr *>::iterator InstIter;
	SMPInstr *CurrInst;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	set<DefOrUse, LessDefUse>::iterator NextDef;
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	STARS_ea_t InstAddr;

#if SMP_DEBUG_TYPE_INFERENCE
	if (DebugFlag) {
		this->Dump(false);
	}
#endif
	// One time only: Set the types of immediate values, flags register, stack and frame
	//  pointers, and floating point registers.
	if (FirstIter) {
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			CurrInst = (*InstIter);
#if SMP_DEBUG_TYPE_INFERENCE
			if (DebugFlag) {
				SMP_msg("SetImmedTypes for inst at %x: %s\n", CurrInst->GetAddr(), CurrInst->GetDisasm());
			}
#endif
			CurrInst->SetImmedTypes(this->UseFP);
			// Infer signedness, bit width, and other info from the nature of the instruction
			//  (e.g. loads from stack locations whose signedness has been inferred earlier
			//  in FindOutGoingArgSize(), or inherently signed arithmetic opcodes like signed
			//  or unsigned multiplies and divides).
			if (this->HasGoodFGStackTable()) {
				CurrInst->MDSetWidthSignInfo(this->UseFP);
			}
		} // end for each inst
		// Check for signedness inferences from conditional branches at the end of blocks.
		for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
			CurrBlock = (*BlockIter);
			CurrBlock->MarkBranchSignedness();
			if (global_STARS_program->ShouldSTARSShadowCriticalArgs()) {
				(void) CurrBlock->MarkTaintWarningInArgs();
			}
		}

		// Find counter variables (e.g. init to zero or small constant, then just add or subtract small
		//  constant values. These cannot be POINTER and can be marked as NUMERIC.
		this->FindCounterVariables();
	} // end if (FirstIter)

	// Iterate until no more changes: set types in DEF and USE lists based on RTL
	//  operators and the instruction category, SSA DEF-USE chains, etc.
	do {
#if SMP_ANALYZE_INFER_TYPES_TIME
		if (DebugFlag2)
			++IterationCount;
#endif
#if 0
		do {
#endif
			changed = false;
#if SMP_ANALYZE_INFER_TYPES_TIME
			if (DebugFlag2)
				NewChangeCount = 0;
#endif
			// Step one: Infer types within instructions, context free.
			// Step two, propagating DEF types to all USEs, happens within step one
			//  whenever a DEF type is set for the first time.
			for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
				CurrInst = (*InstIter);
#if SMP_DEBUG_TYPE_INFERENCE
				if (DebugFlag) SMP_msg("Inferring types for %s\n", CurrInst->GetDisasm());
#endif
				NewChange = CurrInst->InferTypes();
				changed = (changed || NewChange);
#if SMP_ANALYZE_INFER_TYPES_TIME
				if (DebugFlag2 && NewChange) {
					STARS_ea_t InstAddr = CurrInst->GetAddr();
					++NewChangeCount;
				}
#endif
			}
#if SMP_ANALYZE_INFER_TYPES_TIME
			if (DebugFlag2) {
				SMP_msg(" InferTypes iteration: %ld NewChangeCount: %ld \n", IterationCount, NewChangeCount);
			}
#endif
#if 0
		} while (changed);
#endif
#if SMP_DEBUG_TYPE_INFERENCE
		if (DebugFlag) SMP_msg("Finished type inference steps 1 and 2.\n");
#endif
		// Step three: If all USEs of an SSA name have the same type, but the DEF has no
		//  type, then infer that the DEF must have the same type.
		this->TypedDefs = 0;
		this->UntypedDefs = 0;
		this->TypedPhiDefs = 0;
		this->UntypedPhiDefs = 0;
		// This step of the type inference might converge faster if we used a reverse iterator
		//  to go through the instructions, because we could infer a DEF, propagate it to
		//  the right hand side by making SMPInstr::InferOperatorType() public and calling it
		//  on the SMP_ASSIGN operator after we set the type of the left hand side (DEF). Any
		//  additional DEF inferences would be triggered mostly in the upwards direction by
		//  setting the type of one or more USEs in the current instruction. How much time gain
		//  could be achieved by doing this sequence is questionable.  !!!!****!!!!****
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			CurrInst = (*InstIter);
			InstAddr = CurrInst->GetAddr();
			// Find any DEF that still has type UNINIT.
			CurrDef = CurrInst->GetFirstDef();
			while (CurrDef != CurrInst->GetLastDef()) {
				// Set erase() and insert() are needed to change types of DEFs, so
				//  get hold of the next iterator value now.
				NextDef = CurrDef;
				++NextDef;
				NewChange = false;
				if (UNINIT != CurrDef->GetType()) {
					++(this->TypedDefs);
				}
				else {
					STARSOpndTypePtr DefOp = CurrDef->GetOp();
					bool MemDef = (! DefOp->IsRegOp());
					bool AliasedMemWrite = (MemDef && CurrDef->HasIndirectWrite());
					++(this->UntypedDefs);
					if (MDIsIndirectMemoryOpnd(DefOp, this->UseFP)  // relax this?
#if 0
						|| (DefOp->IsStaticMemOp())
#endif
						|| AliasedMemWrite) {
						// Don't want to infer along DEF-USE chains for indirect
						//  memory accesses until we have alias analysis.
						++CurrDef;
						continue;
					}
					STARS_ea_t DefAddr = InstAddr;
					// Call inference method based on whether it is a block-local
					//  name or a global name.
					CurrBlock = CurrInst->GetBlock();
					if (CurrBlock->IsLocalName(DefOp)) {
						STARSOpndSetIter NameIter;
						NameIter = CurrBlock->FindLocalName(DefOp);
						assert(CurrBlock->GetLastLocalName() != NameIter);
						unsigned int LocIndex = ExtractGlobalIndex(*NameIter);
						NewChange = CurrBlock->InferLocalDefType(DefOp, LocIndex, DefAddr);
						if (NewChange) {
							--(this->UntypedDefs);
							++(this->TypedDefs);
						}
						changed = (changed || NewChange);
					}
					else {
						// global name
						bool CallInst = ((CALL == CurrInst->GetDataFlowType())
							|| (INDIR_CALL == CurrInst->GetDataFlowType()));
						int DefSSANum = CurrDef->GetSSANum();
						SMPOperandType DefType = UNINIT;
						DefType = this->InferGlobalDefType(DefOp,
							DefSSANum, CurrBlock, CallInst, DefAddr);
						if (IsNotEqType(UNINIT, DefType)) {
							CurrDef = CurrInst->SetDefType(DefOp, DefType);
							--(this->UntypedDefs);
							++(this->TypedDefs);
							NewChange = true;
							// If we have one or more USEs of type POINTER or CODEPTR and the
							//  other USEs are UNINIT, then InferGlobalDefType() will
							//  infer that it is a POINTER/CODEPTR. We want to propagate POINTER/CODEPTR
							//  to all the USEs now.
							if (IsDataPtr(DefType) || IsCodePtr(DefType)) {
								this->ResetProcessedBlocks();
								CurrInst->GetBlock()->PropagateGlobalDefType(DefOp, DefType, DefSSANum, IsMemOperand(DefOp), false);
							}
						}
						changed = (changed || NewChange);
					} // end if local name ... else ...
				} // end if (UNINIT != CurrDef->GetType()) .. else ...
				CurrDef = NextDef;
			} // end while all DEFs in the DEF set
		} // end for all instructions
#if SMP_DEBUG_TYPE_INFERENCE
		if (DebugFlag) SMP_msg("Finished type inference step 3.\n");
#endif

		for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
			CurrBlock = (*BlockIter);
			changed |= CurrBlock->InferAllPhiDefTypes();
		}

#if SMP_DEBUG_TYPE_INFERENCE
		if (DebugFlag) SMP_msg("Finished unconditional phi type inference.\n");
#endif

#if SMP_CONDITIONAL_TYPE_PROPAGATION
		if (!changed) { // Try conditional type propagation
			changed |= this->ConditionalTypePropagation();
#if SMP_DEBUG_TYPE_INFERENCE
			if (DebugFlag) {
				SMP_msg("changed = %d after conditional type propagation.\n", changed);
			}
#endif
		}
#endif

	} while (changed);

	// With type inference finished, infer signedness from the types, e.g.
	//  POINTER and CODEPOINTER types must be UNSIGNED.
	if (FirstIter) { // Don't want profiler-dependent signedness in the system yet.
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			(*InstIter)->InferSignednessFromSMPTypes(this->UsesFramePointer());
		}
	}

	// Record the meet of all register types that reach RETURN instructions.
	(void) this->MDFindReturnTypes();
	return;
} // end of SMPFunction::InferTypes()
