/*
 * SMPFunction.cpp - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

//
// SMPFunction.cpp
//
// This module performs the fundamental data flow analyses needed for the
//   SMP project (Software Memory Protection) at the function level.
//


#include "SMPFunction_support.h"

// populate FuncLoopsByBlock and FuncBlocksByLoop data structures for DetectLoops().
void SMPFunction::FindLoopBlocksFromTailToHeader(const size_t LoopNumber, const int HeadBlockNum, list<SMPBasicBlock *>::iterator BlockIter, int &DoubleTailFollowBlockNum) {
	list<list<SMPBasicBlock *>::iterator> BlockWorkList;
	BlockWorkList.push_back(BlockIter);
	SMPBasicBlock *WorkBlock;
	SMPBasicBlock *PredBlock;
	list<SMPBasicBlock *>::iterator WorkIter;
	list<SMPBasicBlock *>::iterator PredIter;
	bool WritesMemoryStaticOrIndirect = false;
	bool ReadsMemoryStaticOrIndirect = false;
	bool DoubleTailFlag = (*BlockIter)->IsDoubleLoopTailBlock();
	do {
		WorkIter = BlockWorkList.front();
		BlockWorkList.pop_front();
		WorkBlock = (*WorkIter);
		if (WorkBlock->HasIndirectMemWrite() || WorkBlock->HasStaticMemWrite()) {
			WritesMemoryStaticOrIndirect = true;
		}
		if (WorkBlock->HasIndirectMemRead() || WorkBlock->HasStaticMemRead()) {
			ReadsMemoryStaticOrIndirect = true;
		}
		int WorkBlockNum = WorkBlock->GetNumber();
		assert(WorkBlockNum != SMP_BLOCKNUM_UNINIT);
		assert(WorkBlockNum < ((int) this->Blocks.size()));

		if (!(WorkBlock->IsProcessed())) {
			assert(LoopNumber < this->LoopCount);
			this->FuncLoopsByBlock[(std::size_t) WorkBlockNum].SetBit(LoopNumber);
			this->FuncBlocksByLoop[LoopNumber].SetBit((size_t)WorkBlockNum);
			WorkBlock->SetProcessed(true);

			// Add unprocessed predecessors to the work list until we reach the loop head.
			if (WorkBlockNum != HeadBlockNum) {
				for (PredIter = WorkBlock->GetFirstPred(); PredIter != WorkBlock->GetLastPred(); ++PredIter) {
					PredBlock = (*PredIter);
					bool AlreadyProcessed = PredBlock->IsProcessed();
					if (!AlreadyProcessed) {
						BlockWorkList.push_back(PredIter);
					}
				}
			}
		}
	} while (!BlockWorkList.empty());
	this->LoopWritesMemory.push_back(WritesMemoryStaticOrIndirect);
	this->LoopReadsMemory.push_back(ReadsMemoryStaticOrIndirect);

	if (DoubleTailFlag) {
		// Out of the two successors to a double tail block,
		//  one successor will be the head of the inner loop
		//  and the other successor will be the head of the outer loop.
		// Therefore, the only valid FollowBlockNum for both loops is
		//  the loop exit successor for the outer loop.
		// However, the outer loop has two cases:
		//  1. It is a top-testing loop, in which case we only need to find
		//     the successor to the outer loop header block that is outside the
		//     outer loop.
		//  2. It is a middle-testing loop, so we must find some other FollowBlockNum
		//     for the outer loop and use it for the inner loop, also.
		// Unfortunately, we are in the middle of the process of distinguishing
		//  top-testing from middle-testing loops. But we can simplify both cases
		//  into one with this observation: As we trace from the double tail block
		//  back up to the outer loop header, any block successor that is not in the
		//  outer loop is the FollowBlockNum for the outer loop, and hence for the
		//  inner loop, also. When we are processing the inner loop, we will mistakenly
		//  identify the header of the outer loop as the FollowBlockNum. Immediately
		//  after doing so, this method will be called with the outer loop number, and
		//  the correct FollowBlockNum will be written. The caller of this method
		//  will record this second FollowBlockNum for both loops.
		list<size_t> BlockList;
		this->BuildLoopBlockList(LoopNumber, BlockList);
		for (size_t BlkNum : BlockList) {
			SMPBasicBlock *LoopBlock = this->GetBlockByNum(BlkNum);
			list<SMPBasicBlock *>::const_iterator SuccIter = LoopBlock->GetSuccNotInLoop(LoopNumber);
			if (SuccIter != LoopBlock->GetLastConstSucc()) {
				DoubleTailFollowBlockNum = (*SuccIter)->GetNumber();
				break;
			}
		}
	}

	return;
} // end of SMPFunction::FindLoopBlocksFromTailToHeader()

// Detect which blocks are in which loops and populate FuncLoopsByBlock and FuncBlocksByLoop data structures.
void SMPFunction::DetectLoops(void) {
	map<int, int> DoubleTailBlockNumToFollowBlockMap; // keep track of tricky FollowBlockNum
													// analysis results for DoubleTailBlocks
	bool DebugFlag = (0x13f264 == this->GetFirstFuncAddr());
	this->LoopExitBlockIgnored.resize(STARS_MAX_LOOPS_PER_FUNC, false);

	std::size_t LoopNumber = 0;
	for (list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		if (CurrBlock->IsLoopTailBlock()) {
			// For each loop tail block, get its loop header block number (the target
			//  block for its back edge). Then traverse the CFG upwards, recording all
			//  of the basic block numbers that lie between the loop tail and loop head,
			//  along with the tail and head.
			// Along the way, record whether the loop writes to static memory or has
			//  an indirect memory write.
			this->ResetProcessedBlocks();
			int TailBlockNum = CurrBlock->GetNumber();
			int HeadBlockNum = CurrBlock->GetLoopHeaderNumber();
			assert((TailBlockNum != SMP_BLOCKNUM_UNINIT) && (HeadBlockNum != SMP_BLOCKNUM_UNINIT));
			bool FoundHeader = this->IsBlockNumLoopHeader(HeadBlockNum, LoopNumber);
			assert(FoundHeader);
			int DoubleTailFollowBlockNum = SMP_BLOCKNUM_UNINIT;
			this->FindLoopBlocksFromTailToHeader(LoopNumber, HeadBlockNum, BlockIter, DoubleTailFollowBlockNum);
			// If double tail block, trace outer block.
			if (CurrBlock->IsDoubleLoopTailBlock()) {
				HeadBlockNum = CurrBlock->GetOuterLoopHeaderNumberForDoubleTailBlock();
				this->ResetProcessedBlocks();
				// Get new LoopNumber.
				bool FoundHeader = this->IsBlockNumLoopHeader(HeadBlockNum, LoopNumber);
				assert(FoundHeader);
				this->FindLoopBlocksFromTailToHeader(LoopNumber, HeadBlockNum, BlockIter, DoubleTailFollowBlockNum);
				DoubleTailBlockNumToFollowBlockMap[TailBlockNum] = DoubleTailFollowBlockNum;
			}

		} // end if tail block
	} // end for all blocks

	// Now, classify the loops as top-testing, bottom-testing, or middle-testing, and
	//  find the follow block number for the loops.
	STARSBitSet LoopAlreadyClassified;   // prevent redundant work on loops with multiple tail blocks
	LoopAlreadyClassified.AllocateBits(this->GetNumLoops());
	for (size_t BlockNum = 0; BlockNum < this->GetNumBlocks(); ++BlockNum) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		if (CurrBlock->IsLoopTailBlock()) {
			int TailBlockNum = CurrBlock->GetNumber();
			int HeadBlockNum = CurrBlock->GetLoopHeaderNumber();
			assert((TailBlockNum != SMP_BLOCKNUM_UNINIT) && (HeadBlockNum != SMP_BLOCKNUM_UNINIT));
			bool FoundHeader = this->IsBlockNumLoopHeader(HeadBlockNum, LoopNumber);
			assert(FoundHeader);
			// Examine the head and tail blocks to determine the loop type. CurrBlock is the tail block.
			SMPBasicBlock *HeadBlock = this->GetBlockByNum((std::size_t) HeadBlockNum);
			// 1996 Cristina Cifuentes paper, "Structuring Decompiled Graphs," has a simple formula
			//  for classifying loops as top-testing, bottom-testing, or infinite (with possible middle-testing exit).
			//  I have modified it here to account for possible bottom-testing loop that begins with an if-then, because
			//  we do not create empty loop header blocks for loops (we use existing blocks with actual instructions).
			//  TOP-TESTING: Header has more than one successor, with at least one successor outside the loop, while
			//    the tail block has only one successor (the header block).
			//  BOTTOM-TESTING: Header has all successors in the loop (usually just one, but might be if-then-else, etc.),
			//    while the tail block has more than one successor, with at least one outside the loop.
			//  INFINITE or MIDDLE_TESTING: Header has all successors in the loop, tail block has one successor (the header block).
			//  As a truth table, there are logically four possibilities:
			//
			//     HEADER BLOCK                        TAIL BLOCK                         LOOP TYPE
			//     All successors in the loop          Header is only successsor          Infinite or middle-testing
			//     All successors in the loop          >=1 successors outside the loop    Bottom-testing
			//     >=1 successors outside the loop     Header is only successor           Top-testing
			//     >=1 successors outside the loop     >=1 successors outside the loop    Middle-testing with conditional exits at top and bottom
			//
			//  In Ada (to which we might decompile), these four loops look like:
			//    MIDDLE-TESTING or INFINITE:
			//    loop
			//      ...
			//      exit when condition;  [absence of exits would make this an infinite loop]
			//      ...
			//    end loop;
			//
			//    MIDDLE-TESTING with conditional exits at top and bottom:
			//     loop
			//       exit when condition_1;
			//       ...
			//       exit when condition_2;
			//     end loop;
			//
			//     TOP-TESTING:
			//     while condition loop
			//       ...
			//     end loop;
			//
			//     BOTTOM-TESTING:
			//     loop
			//       ...
			//       exit when condition;
			//     end loop;
			//
			//     If the top-testing loop has a finite, constant number of iterations, we could use the for-loop instead of the while-loop.
			bool DoubleTailFlag = CurrBlock->IsDoubleLoopTailBlock();
			int HeaderExitFollowNum = SMP_BLOCKNUM_UNINIT, TailExitFollowNum = SMP_BLOCKNUM_UNINIT;
			bool HeaderBlockExitsLoop = this->DoesBlockExitLoop(LoopNumber, HeadBlock, HeaderExitFollowNum);
			bool TailBlockExitsLoop = this->DoesBlockExitLoop(LoopNumber, CurrBlock, TailExitFollowNum);
			if (DebugFlag) {
				SMP_msg("DEBUG: DetectLoops(): LoopNum %d HeaderExits %d TailExits %d DoubleTail %d ",
					LoopNumber, HeaderBlockExitsLoop, TailBlockExitsLoop, DoubleTailFlag);
				this->DumpFuncNameAndAddr();
			}
			if (DoubleTailFlag) {
				// FollowBlockNum should be in our local map.
				map<int, int>::const_iterator MapIter = DoubleTailBlockNumToFollowBlockMap.find(TailBlockNum);
				assert(MapIter != DoubleTailBlockNumToFollowBlockMap.cend());
				TailExitFollowNum = (*MapIter).second;
				HeaderExitFollowNum = TailExitFollowNum; // simplify code below
			}

			if (!LoopAlreadyClassified.GetBit(LoopNumber)) {
				this->ClassifyLoop(LoopNumber, HeaderExitFollowNum, TailExitFollowNum, CurrBlock, HeadBlock, HeaderBlockExitsLoop, TailBlockExitsLoop);
				LoopAlreadyClassified.SetBit(LoopNumber);
				if (this->LoopExitingBlocksVector.size() <= LoopNumber) {
					// Loop must have had no exits.
					int LoopType = this->LoopTypesByLoopNum[LoopNumber];
					SMP_msg("INFO: LOOP: Infinite loop, no exiting blocks, loop number %zu ", LoopNumber);
					if (this->LoopExitBlockIgnored[LoopNumber] && (STARS_BOTTOM_TESTING_LOOP == LoopType)) {
						SMP_msg("INFO: Loop %zu with ignored exit blocks reclassified to infinite loop.", LoopNumber);
						this->LoopTypesByLoopNum[LoopNumber] = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
						LoopType = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
					}
					assert(STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == LoopType);
					this->DumpFuncNameAndAddr();
					vector<struct LoopExitingBlockInfo> NewVector;
					this->LoopExitingBlocksVector.push_back(NewVector);
					assert(this->LoopExitingBlocksVector.size() > LoopNumber);
				}
			}

			if (DoubleTailFlag) {
				// Classify the outer loop.
				HeadBlockNum = CurrBlock->GetOuterLoopHeaderNumberForDoubleTailBlock();
				HeadBlock = this->GetBlockByNum((std::size_t) HeadBlockNum);
				size_t OuterLoopNumber;
				bool FoundHeader = this->IsBlockNumLoopHeader(HeadBlockNum, OuterLoopNumber);
				assert(FoundHeader);
				if (!LoopAlreadyClassified.GetBit(OuterLoopNumber)) {
					this->ClassifyLoop(OuterLoopNumber, HeaderExitFollowNum, TailExitFollowNum, CurrBlock, HeadBlock, HeaderBlockExitsLoop, TailBlockExitsLoop);
					LoopAlreadyClassified.SetBit(OuterLoopNumber);
				}
			}
		} // end if tail block
		if (!this->HasStructuredControlFlow()) {
			SMP_msg("WARNING: After ClassifyLoop() function %s at %llx is detected as unstructured.\n",
				this->GetFuncName(), (uint64_t) this->GetFirstFuncAddr());
		}
	} // end for all blocks

	return;
} // end of SMPFunction::DetectLoops()

// fill LoopNests data structure.
void SMPFunction::ComputeLoopNests(void) {
	// A loop contains another loop if the header block of the first loop
	//  dominates the header block of the second loop.
	for (size_t LoopIndex1 = 0; LoopIndex1 < this->LoopCount; ++LoopIndex1) {
		for (size_t LoopIndex2 = 0; LoopIndex2 < this->LoopCount; ++LoopIndex2) {
			if (LoopIndex1 != LoopIndex2) {
				int HeaderBlock1 = this->LoopHeadBlockNumbers[LoopIndex1];
				int HeaderBlock2 = this->LoopHeadBlockNumbers[LoopIndex2];
				this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				if (this->DoesBlockDominateBlock(HeaderBlock1, HeaderBlock2)) {
					// Loop 2 is nested inside loop 1.
					this->LoopNests[LoopIndex1].SetBit(LoopIndex2);
				}
			}
		}
	}
	return;
} // end of SMPFunction::ComputeLoopNests()

// Emit info on loop nesting and preheader blocks.
//  Annotations look like:
//  [hex func ID] [func size] LOOP [loop #] FIRSTINST [hex inst ID at beginning of loop header] PREHEADER [hex inst ID] BLOCKLIST [hex inst IDs] ZZ INNERLOOPS [loop #'s] ZZ
void SMPFunction::EmitLoopNestAnnotations(void) {
	if (0 < this->LoopCount) {
		FILE *LoopAnnotFile = global_STARS_program->GetLoopsAnnotFile();
		SMP_fprintf(LoopAnnotFile, "%18llx %6zu LOOPCOUNT %zu \n", (uint64_t) this->GetFuncID(), this->FuncInfo->GetFuncSize(), this->LoopCount);
		for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			int HeaderBlockNum = this->LoopHeadBlockNumbers[LoopIndex];
			SMPBasicBlock *HeaderBlock = this->GetBlockByNum((size_t) HeaderBlockNum);
			STARS_ea_t FirstHeaderInstAddr = HeaderBlock->GetFirstNonMarkerAddr();
			int PreheaderBlockNum = this->FindLoopPreheaderBlockNum(LoopIndex);
			STARS_ea_t PreheaderFirstInstAddr = STARS_BADADDR;
			if (PreheaderBlockNum >= 0) {
				PreheaderFirstInstAddr = this->GetBlockByNum(PreheaderBlockNum)->GetFirstNonMarkerAddr();
			}
			SMP_fprintf(LoopAnnotFile, "%18llx %6zu LOOP %zu FIRSTINST %18llx PREHEADER %18llx BLOCKLIST ",
				(uint64_t) this->GetFuncID(), this->FuncInfo->GetFuncSize(), LoopIndex, (uint64_t) FirstHeaderInstAddr, (uint64_t) PreheaderFirstInstAddr);

			// Emit the first inst ID for each block in the loop.
			list<size_t> LoopBlockList;
			this->BuildLoopBlockList(LoopIndex, LoopBlockList);
			for (size_t BlockNum : LoopBlockList) {
				SMP_fprintf(LoopAnnotFile, "%18llx ", (uint64_t) this->GetBlockByNum(BlockNum)->GetFirstNonMarkerAddr());
			}
			SMP_fprintf(LoopAnnotFile, "ZZ INNERLOOPS ");

			// Emit the loop indices for inner loops.
			for (size_t LoopIndex2 = 0; LoopIndex2 < this->LoopCount; ++LoopIndex2) {
				if (this->LoopNests[LoopIndex].GetBit(LoopIndex2)) {
					SMP_fprintf(LoopAnnotFile, "%zu ", LoopIndex2);
				}
			}
			SMP_fprintf(LoopAnnotFile, "ZZ \n");
		}
	}
	return;
} // end of SMPFunction::EmitLoopNestAnnotations()

// return -1 if no preheader (immediate predecessor that dominates header, has only header as successor)
int SMPFunction::FindLoopPreheaderBlockNum(size_t LoopIndex) {
	int PreheaderBlockNum = SMP_BLOCKNUM_UNINIT;
	assert(LoopIndex < this->LoopCount);
	int HeaderBlockNum = this->LoopHeadBlockNumbers[LoopIndex];
	assert(0 <= HeaderBlockNum);
	SMPBasicBlock *HeaderBlock = this->GetBlockByNum((size_t) HeaderBlockNum);
	for (list<SMPBasicBlock *>::const_iterator PredIter = HeaderBlock->GetFirstConstPred();
		PredIter != HeaderBlock->GetLastConstPred();
		++PredIter) {
		SMPBasicBlock *PredBlock = (*PredIter);
		if (1 == PredBlock->GetNumSuccessors()) {
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			if (this->DoesBlockDominateBlock(PredBlock->GetNumber(), HeaderBlockNum)) {
				PreheaderBlockNum = PredBlock->GetNumber();
				break;
			}
		}
	}

	return PreheaderBlockNum;
} // end of SMPFunction::FindLoopPreheaderBlockNum()


void SMPFunction::DetectLoopFollowBlock(const size_t LoopNumber) {
	list<size_t> LoopBlockNums;
	this->BuildLoopBlockList(LoopNumber, LoopBlockNums);
	int HeadBlockNum = this->LoopHeadBlockNumbers[LoopNumber];

	for (size_t BlockNum : LoopBlockNums) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();
			assert(LoopNumber < this->FuncBlocksByLoop.size());
			assert(SuccBlockNum >= 0);
			assert(((size_t) SuccBlockNum) < this->GetNumBlocks());
			assert(((size_t)SuccBlockNum) < this->FuncBlocksByLoop[LoopNumber].GetNumBits());
			if (!this->FuncBlocksByLoop[LoopNumber].GetBit((size_t) SuccBlockNum)) {
				// Successor is not in loop.

				// Record exiting block info.
				struct LoopExitingBlockInfo CurrExitInfo;
				CurrExitInfo.ExitingBlockNum = (int)BlockNum;
				CurrExitInfo.TargetBlockNum = SuccBlockNum;
				CurrExitInfo.CompareTestDecInst = nullptr;
 
				CurrExitInfo.ReferencesBIV = false;

				// Can we trace the conditional branch of the tail and/or header blocks
				//  to a compare, test, decrement or other inst that can be cleanly analyzed for
				//  our loop iterations count expression?
				SMPInstr *CompareOrTestInst = nullptr;
				SMPInstr *DecInst = nullptr;
				STARS_ea_t DecAddr = STARS_BADADDR;
				if (CurrBlock->HasConditionalBranch()) {
					SMPoperator BranchOper = SMP_NULL_OPERATOR;
					SMPoperator SetBitOperator = SMP_NULL_OPERATOR;
					CompareOrTestInst = CurrBlock->FindBranchCompareOrTest(BranchOper, SetBitOperator, DecAddr, false, SMP_SSA_UNINIT, CurrBlock->GetRevInstCBegin());
					if (STARS_BADADDR != DecAddr) {
						assert(STARS_IsNotPseudoInstID(DecAddr));
						DecInst = this->GetInstFromAddr(DecAddr);
						assert(nullptr != DecInst);
						CurrExitInfo.CompareTestDecInst = DecInst;
					}
					else if (nullptr != CompareOrTestInst) {
						CurrExitInfo.CompareTestDecInst = CompareOrTestInst;
					}
				}
				// Record whether branch is a short circuit expr branch.
				// Due to phase ordering, this has not been determined until SPARK analysis in 
				//  MarkSpecialNumericErrorCases(), so it must be filled in later.
				CurrExitInfo.ShortCircuitExpr = false;

				// If first exiting block, expand outer vector, else push onto inner vector.
				if (LoopNumber < this->LoopExitingBlocksVector.size()) {
					this->LoopExitingBlocksVector[LoopNumber].push_back(CurrExitInfo);
				}
				else {
					// Must have had an infinite loop, or we are processing loops out of order.
					while (this->LoopExitingBlocksVector.size() < LoopNumber) {
						// Pad out with empty vectors.
						vector<struct LoopExitingBlockInfo> NewVector;
						this->LoopExitingBlocksVector.push_back(NewVector);
					}
					vector<struct LoopExitingBlockInfo> NewVector;
					NewVector.push_back(CurrExitInfo);
					this->LoopExitingBlocksVector.push_back(NewVector);
				}

				// Record loop follow block info.
				this->UpdateLoopFollowBlockNum(HeadBlockNum, SuccBlockNum);
			}
		}
	}

	return;
} // end of SMPFunction::DetectLoopFollowBlock()

void SMPFunction::ClassifyLoop(size_t LoopNumber, int HeaderExitFollowNum, int TailExitFollowNum, SMPBasicBlock *CurrBlock, SMPBasicBlock *HeadBlock, bool HeaderBlockExitsLoop, bool TailBlockExitsLoop) {
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();
	this->DetectLoopFollowBlock(LoopNumber);
	if (HeaderBlockExitsLoop && HeadBlock->HasIndirectJump()) {
		this->SetHasUnstructuredCFG(); // indirect jump head loop exit
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to indirect jump head loop exit at %llx for loop %d ",
				(uint64_t)CurrBlock->GetLastAddr(), LoopNumber);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
	}
	else if (TailBlockExitsLoop && CurrBlock->HasIndirectJump()) {
		this->SetHasUnstructuredCFG(); // indirect jump tail loop exit 
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to indirect jump tail loop exit at %llx for loop %d ",
				(uint64_t)CurrBlock->GetLastAddr(), LoopNumber);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
	}

	if (TailBlockExitsLoop) {
		if (HeaderBlockExitsLoop) {
			// Conditional exits at top and bottom of loop.
			this->LoopTypesByLoopNum[LoopNumber] = STARS_TOP_AND_BOTTOM_TESTING_LOOP;
			if (TailExitFollowNum != HeaderExitFollowNum) {
				if (this->HasGoodLoopFollowBlocks || VerboseOutput) {
					if (!this->PrintedSPARKUnstructuredMsg || VerboseOutput) {
						SMP_msg("INFO: SPARK: Might be unstructured due to conflicting head and tail loop follow block nums for loop %d : %d and %d ",
							LoopNumber, HeaderExitFollowNum, TailExitFollowNum);
						this->DumpFuncNameAndAddr();
					}
				}
#if 0
				this->PrintedSPARKUnstructuredMsg = true;
				this->HasGoodLoopFollowBlocks = false; // inconsistent head and tail follow nodes, e.g. multi-level loop exit
#endif
			}
		}
		else {
			// Normal bottom-testing loop.
			this->LoopTypesByLoopNum[LoopNumber] = STARS_BOTTOM_TESTING_LOOP;
		}
		this->LoopTestBlocksByLoopNum[LoopNumber] = CurrBlock->GetNumber();
		this->UpdateLoopFollowBlockNum(HeadBlock->GetNumber(), TailExitFollowNum);
	}
	else {
		if (HeaderBlockExitsLoop) {
			// Normal top-testing while or for loop.
			this->LoopTypesByLoopNum[LoopNumber] = STARS_TOP_TESTING_LOOP;
			this->LoopTestBlocksByLoopNum[LoopNumber] = HeadBlock->GetNumber();
			this->UpdateLoopFollowBlockNum(HeadBlock->GetNumber(), HeaderExitFollowNum);
		}
		else {
			// Infinite or middle-testing loop.
			this->LoopTypesByLoopNum[LoopNumber] = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
			this->LoopTestBlocksByLoopNum[LoopNumber] = SMP_BLOCKNUM_UNINIT;
		}
	}

#if 1
	// See if we want to reclassify the loop according to which COND_BRANCH controls maximum iterations.
	//  A loop could have an induction var that controls iteration count, but also have a possible early
	//  exit based on some condition. The early exit does not control maximum iterations. Often the early
	//  exit is not a simple induction var test, so it fails to match up with a compare or test inst.

	// Select the most analyzeable test block when we have more than one exiting block.

	// First, pad out the vector allocation if needed.
	while (this->LoopExitingBlocksVector.size() <= LoopNumber) {
		// Loop must have had no exits.
		int LoopType = this->LoopTypesByLoopNum[LoopNumber];
		SMP_msg("INFO: LOOP: Infinite loop, no exiting blocks, loop number %zu loop type %d ", LoopNumber, LoopType);
		this->DumpFuncNameAndAddr();
		if (this->LoopExitBlockIgnored[LoopNumber] && (STARS_BOTTOM_TESTING_LOOP == LoopType)) {
			SMP_msg("INFO: Loop %zu with ignored exit blocks reclassified to infinite loop.", LoopNumber);
			this->LoopTypesByLoopNum[LoopNumber] = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
			LoopType = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
		}
		if (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP != LoopType) { // about to assert and die
			this->Dump(false);
			global_STARS_program->SetSPARKFlag(true); // Make DotCFG file be emitted.
			this->DumpDotCFG();
			SMP_msg("FATAL ERROR: Loop with no exiting blocks is of wrong type.\n");
		}
		assert(STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == LoopType);

		vector<struct LoopExitingBlockInfo> NewVector;
		this->LoopExitingBlocksVector.push_back(NewVector);
	}

	if (1 < this->LoopExitingBlocksVector[LoopNumber].size()) {
		for (size_t VecIndex = 0; VecIndex < this->LoopExitingBlocksVector[LoopNumber].size(); ++VecIndex) {
			if (nullptr != this->LoopExitingBlocksVector[LoopNumber][VecIndex].CompareTestDecInst) {
				int GoodExitBlockNum = this->LoopExitingBlocksVector[LoopNumber][VecIndex].ExitingBlockNum;
				if (GoodExitBlockNum > this->LoopTestBlocksByLoopNum[LoopNumber]) {
					SMP_msg("INFO: LOOP: Changing loop test block from %d to traceable block %d for loop %zu type %d ",
						this->LoopTestBlocksByLoopNum[LoopNumber], GoodExitBlockNum, LoopNumber, this->LoopTypesByLoopNum[LoopNumber]);
					this->DumpFuncNameAndAddr();
					this->LoopTestBlocksByLoopNum[LoopNumber] = GoodExitBlockNum;
				}
			}
			else if (VerboseOutput) {
				SMP_msg("INFO: LOOP: Skipping untraceable loop test block %d for loop %zu ",
					this->LoopTestBlocksByLoopNum[LoopNumber], LoopNumber);
			}
		}
	}
#endif

	return;
} // end of SMPFunction::ClassifyLoop()

// return true if block can exit the loop.
bool SMPFunction::DoesBlockExitLoop(std::size_t LoopNumber, SMPBasicBlock *LoopBlock, int &FollowBlockNum) {
	bool FoundExitSuccessor = false;
	int LoopBlockNum = LoopBlock->GetNumber();

	for (list<SMPBasicBlock *>::iterator SuccIter = LoopBlock->GetFirstSucc(); SuccIter != LoopBlock->GetLastSucc(); ++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		int SuccBlockNum = SuccBlock->GetNumber();
		if (!(this->IsBlockInLoop(SuccBlockNum, LoopNumber))) {
			if (!LoopBlock->IsDoubleLoopTailBlock()) { // normal case
				FoundExitSuccessor = true;
				FollowBlockNum = SuccBlockNum;
				bool MultiLoopBreak = this->DetectMultiLevelLoopBreak(LoopBlockNum, SuccBlockNum);
				if (MultiLoopBreak) {
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("ERROR: SPARK: Unstructured due to multi-level loop break in Loop %zu at block %d ", LoopNumber, LoopBlockNum);
						this->DumpFuncNameAndAddr();
						this->PrintedSPARKUnstructuredMsg = true;
					}
					this->SetHasUnstructuredCFG(); // multi-level loop break
				}
				break;
			}
			else { // Double tail block
				FoundExitSuccessor = true;
				FollowBlockNum = SMP_BLOCKNUM_UNINIT; // will be ignored by DetectLoops() caller
				break;
			}
		}
	}

	return FoundExitSuccessor;
} // end of SMPFunction::DoesBlockExitLoop()

// Is ExitTargetBlockNum more than 1 loop nesting level away from LoopBlockNum?
bool SMPFunction::DetectMultiLevelLoopBreak(const int LoopBlockNum, const int ExitTargetBlockNum) const {
	assert(0 <= LoopBlockNum);
	assert(0 <= ExitTargetBlockNum);
	assert((size_t)LoopBlockNum < this->FuncLoopsByBlock.size());
	assert((size_t)ExitTargetBlockNum < this->FuncLoopsByBlock.size());
	bool MultiLevelBreak = false;
	int StartBlockNum = this->FuncLoopsByBlock[(size_t)LoopBlockNum].FindLowestBitSet();
	int EndBlockNum = this->FuncLoopsByBlock[(size_t)LoopBlockNum].FindHighestBitSet();
	uint32_t LeftBehindCounter = 0;
	for (int BlockNum = StartBlockNum; BlockNum <= EndBlockNum; ++BlockNum) {
		if (this->FuncLoopsByBlock[(size_t)LoopBlockNum].GetBit((size_t)BlockNum)) {
			if (!this->FuncLoopsByBlock[(size_t)ExitTargetBlockNum].GetBit((size_t)BlockNum)) {
				++LeftBehindCounter;
			}
		}
	}
	MultiLevelBreak = (2 <= LeftBehindCounter);
#if 0
	size_t LoopNestingLevel = this->FuncLoopsByBlock[(size_t)LoopBlockNum].CountSetBits();
	size_t ExitTargetNestingLevel = this->FuncLoopsByBlock[(size_t)ExitTargetBlockNum].CountSetBits();
	STARSBitSet CommonLoops = STARSBitSetIntersection(this->FuncLoopsByBlock[(size_t)LoopBlockNum], this->FuncLoopsByBlock[(size_t)ExitTargetBlockNum]);
	if (LoopNestingLevel >(1 + ExitTargetNestingLevel)) {
		MultiLevelBreak = true;
	}
	else if (CommonLoops.CountSetBits() != ExitTargetNestingLevel) {
		// Some sort of spaghetti code; passing the first test should indicate
		//  only <= 1 differences between the two loop nests, with the common loop bits
		//  being identical to the ExitTarget loop bits.
		MultiLevelBreak = true;
	}
#endif

	return MultiLevelBreak;
} // end of SMPFunction::DetectMultiLevelLoopBreak()

// Populate entries for LoopNum in LoopExitTargetsReachabilitySets and MultiLoopExitTargetsConvergenceMap
void SMPFunction::ComputeLoopFollowNodesReachability(const std::size_t LoopNum) {
	// Without following back edges, record which blocks can be reached from 
	//  each block that is the target of a loop exit (which are found in LoopFollowNodes[LoopNum]).
	assert(LoopNum < this->GetNumLoops());
	assert(LoopNum < this->LoopExitTargets.size());
	int LoopHeadBlockNum = this->LoopHeadBlockNumbers[LoopNum];
	bool Unstructured = false;
	for (int ExitTargetBlockNum : this->LoopExitTargets[LoopNum]) {
		STARSBitSet Reachable;
		Reachable.AllocateBits(2 * this->GetNumBlocks()); // allow room for cloning
		this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		int LimitBlockNum = (int)(1 + (2 *this->GetNumBlocks()));
		this->MarkReachableBlocks(ExitTargetBlockNum, (int)LoopNum, LoopHeadBlockNum, true, LimitBlockNum, Reachable);
		if (Reachable.IsAnyBitSet()) {
			pair<int, STARSBitSet> MapItem(ExitTargetBlockNum, Reachable);
			this->LoopExitTargetsReachabilitySets[LoopNum].insert(MapItem);
		}
		else {
			Unstructured = true;
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to no loop head dominance of exit target block %d from loop %zu ", ExitTargetBlockNum, LoopNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
			this->SetHasUnstructuredCFG(); // loop head does not dominate exit target
			break;
		}
	} // end for all ExitTargetBlocks.

	if (!Unstructured) {
		// Find out where the multiple follow nodes converge, if they do. If they all hit separate return blocks,
		//  we use the code of -2 to signify it.
		// Case 1: Reachability bitsets all converge. Convergence point is follow block.
		// Case 2: No common convergence point in reachability bitsets.
		// Case 2a: this->MultiLoopExitTargetsConvergenceMap[LoopNum] has size()==1, so convergence
		//          is only possible at one point, at the dominance frontier of the loop head block,
		//          and the dominance relation enforcement in MarkReachableBlocks() prevented any
		//          reachability trace from reaching that block. Leave that convergence block as our answer.
		// Case 2b: if size()==0, use the -2 code to signify valid nonconvergence. Probably have returns cutting
		//          off at least one reachability trace. NOTE: Might want to explicitly detect k returns and
		//          look for a single convergence of the other N-k traces and use that convergence point.
		// Case 2c: if size() > 1, unstructured.
		STARSBitSet ReachableIntersection;
		ReachableIntersection.AllocateBits(2 * this->GetNumBlocks()); // allow room for cloning
		bool Initialized = false;
		assert(!this->LoopExitTargetsReachabilitySets[LoopNum].empty());
		for (int ExitTargetBlockNum : this->LoopExitTargets[LoopNum]) {
			if (!Initialized) {
				ReachableIntersection = this->LoopExitTargetsReachabilitySets[LoopNum][(size_t)ExitTargetBlockNum];
				Initialized = true;
			}
			else {
				ReachableIntersection = STARSBitSetIntersection(ReachableIntersection, this->LoopExitTargetsReachabilitySets[LoopNum][(size_t)ExitTargetBlockNum]);
			}
		}
		if (ReachableIntersection.IsAnyBitSet()) { // case 1; convergence
			int FollowBlockNum = ReachableIntersection.FindLowestBitSet();
			this->MultiLoopExitTargetsConvergenceMap[LoopNum].clear();
			this->MultiLoopExitTargetsConvergenceMap[LoopNum].insert(FollowBlockNum);
			// Now, trim the paths we followed by redoing MarkReachableBlocks() with a StopBlockNum == FollowBlockNum.
			//  Paths that go through the FollowBlockNum and continue downward are not useful in SPARK translation
			//  because we stop emitting SPARK code in the loop procedure when we see the FollowBlock in this convergent case.
			// We can eliminate some error condition checks that were already performed above.
			for (int ExitTargetBlockNum : this->LoopExitTargets[LoopNum]) {
				STARSBitSet Reachable;
				Reachable.AllocateBits(2 * this->GetNumBlocks()); // allow room for cloning
				this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				int LimitBlockNum = (int)(1 + (2 * this->GetNumBlocks()));
				bool ExitTargetIsFollowBlock = (ExitTargetBlockNum == FollowBlockNum);
				this->MarkReachableBlocks(ExitTargetBlockNum, (int)LoopNum, LoopHeadBlockNum, true, LimitBlockNum, Reachable, FollowBlockNum);
				assert(Reachable.IsAnyBitSet() || ExitTargetIsFollowBlock); 
				int OldHighestBit = this->LoopExitTargetsReachabilitySets[LoopNum][ExitTargetBlockNum].FindHighestBitSet();
				int OldLowestBit = this->LoopExitTargetsReachabilitySets[LoopNum][ExitTargetBlockNum].FindLowestBitSet();
				assert(ExitTargetIsFollowBlock || (Reachable.FindLowestBitSet() >= OldLowestBit)); // No diffs before OldLowestBit
				for (int BitNum = OldLowestBit + 1; BitNum <= OldHighestBit; ++BitNum) {
					if (!Reachable.GetBit((size_t)BitNum)) {
						// Reset any bit not set in Reachable with StopBlockNum trace. (Might not be set anyway.)
						this->LoopExitTargetsReachabilitySets[LoopNum][ExitTargetBlockNum].ResetBit((size_t)BitNum);
					}
				}
			} // end for all ExitTargetBlocks.
		}
		else if (1 < this->MultiLoopExitTargetsConvergenceMap[LoopNum].size()) { // case 2c
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to no convergence of multiple follow nodes from loop %zu ", LoopNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
			this->SetHasUnstructuredCFG(); // multi-exit loop has no single convergence point of follow nodes.
			// TODO: If the multiple termination blocks all loop back to the same outer loop head block, this
			//  is structured code and can be translated. !!!!****!!!!
		}
		else if (this->MultiLoopExitTargetsConvergenceMap[LoopNum].empty()) { // case 2b
			this->MultiLoopExitTargetsConvergenceMap[LoopNum].insert(SMP_BLOCKNUM_COMMON_RETURN);
		}
		else {
			assert(1 == this->MultiLoopExitTargetsConvergenceMap[LoopNum].size()); // case 2a
		}
	}

	if (this->HasStructuredControlFlow() && (1 < this->GetNumLoops())) {
		// Check for exits of outer enclosing loop that are found in the exit target blocks and the
		//  blocks that are reachable from them. These loop exits complicate SPARK Ada translation.
		STARSBitSet BlocksProcessed;
		BlocksProcessed.AllocateBits(2 * this->GetNumBlocks()); // allow room for cloning
		for (pair<int, STARSBitSet> MapItem : this->LoopExitTargetsReachabilitySets[LoopNum]) {
			int ExitTargetBlockNum = MapItem.first;
			size_t ExitTargetLoopLevel = this->FuncLoopsByBlock[(size_t)ExitTargetBlockNum].CountSetBits();
			if (0 < ExitTargetLoopLevel) { // Still in outer loop
				int LowestBlockNumReached = MapItem.second.FindLowestBitSet();
				int HighestBlockNumReached = MapItem.second.FindHighestBitSet();
				for (int BlockNum = LowestBlockNumReached; BlockNum <= HighestBlockNumReached; ++BlockNum) {
					if (MapItem.second.GetBit((size_t)BlockNum)) { // reached by current exit target block
						if (!BlocksProcessed.GetBit((size_t)BlockNum)) { // avoid duplicates when paths merge
							BlocksProcessed.SetBit((size_t)BlockNum);
							SMPBasicBlock *ReachedBlock = this->GetBlockByNum((size_t)BlockNum);
							size_t ReachedBlockLoopLevel = this->FuncLoopsByBlock[(size_t)BlockNum].CountSetBits();
							if (ReachedBlock->DoesBlockExitLoop() && (ReachedBlockLoopLevel == ExitTargetLoopLevel)) { // already out of inner loop; must exit outer loop
								if (!this->PrintedSPARKUnstructuredMsg) {
									SMP_msg("ERROR: SPARK: Unstructured due to outer loop exit from multiple exit targets block %d from loop %zu ", BlockNum, LoopNum);
									this->DumpFuncNameAndAddr();
									this->PrintedSPARKUnstructuredMsg = true;
								}
								this->SetHasUnstructuredCFG(); // multi-exit loop has a need to pass boolean vars as args, alter translation of outer loop
								break; // stop after first problem found
							}
						}
					}
				} // end for all blocks in the reachable range
			}
		} // end for all loop exit targets
	}

	return;
} // end of SMPFunction::ComputeLoopFollowNodesReachability()

// Set bits of current BlockNum and its descendants in CFG that are dominated by HeadBlockNum. Stop tracing if StopBlockNum is seen.
// NOTE: HeadBlockNum can be any dominating block num, with LoopNum < 0 indicating we are not in a loop.
void SMPFunction::MarkReachableBlocks(const int BlockNum, const int LoopNum, const int HeadBlockNum, const bool TestDominance, const int LimitBlockNum, STARSBitSet &ReachableSet, const int StopBlockNum) {
	if (ReachableSet.GetBit((size_t)BlockNum))
		return; // avoid looping and duplication of work

	if (BlockNum == StopBlockNum)
		return; // Don't trace through StopBlockNum.

	bool ProperDominance = true;
	bool UseLimitBlock = false;
	if (TestDominance) {
		if (0 > LoopNum) { // conditional; use LimitBlockNum obtained from DomFrontier (cheaper than dominance test).
			ProperDominance = (BlockNum < LimitBlockNum);
			UseLimitBlock = (LimitBlockNum > 0);
		}
		else {
			ProperDominance = this->DoesBlockDominateBlock(HeadBlockNum, BlockNum);
		}
	}

	SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t)BlockNum);
	assert(nullptr != CurrBlock);
	if (UseLimitBlock) {
		// Every block in this branch of the conditional, until we reach LimitBlock, is reachable.
		//  Save time by not reaching the rest of the code below in this method.
		ReachableSet.SetBit((size_t)BlockNum);
		for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();
			if ((SuccBlockNum > BlockNum) && (SuccBlockNum != StopBlockNum)) { // filter out back edges, avoid StopBlockNum.
				this->MarkReachableBlocks(SuccBlockNum, LoopNum, HeadBlockNum, TestDominance, LimitBlockNum, ReachableSet, StopBlockNum);
			}
		}
		return;
	}

	int ContainingLoopNum = -1;
	// If (LoopNum < 0), we are analyzing a conditional statement. But it could be contained in a loop, and we
	//  don't want to exit that containing loop.
	if (0 > LoopNum) { // conditional statement
		ContainingLoopNum = this->GetInnermostLoopNum(HeadBlockNum);
	}
	bool ConditionalInsideLoop = (0 <= ContainingLoopNum);

	if (ProperDominance) {
		ReachableSet.SetBit((size_t)BlockNum); // Initialize to itself being reachable

		for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();
			// Weed out back edges and StopBlock.
			if ((SuccBlockNum > BlockNum) && (SuccBlockNum != StopBlockNum)) {
				// It is OK to exit an inner loop into multiple blocks that are part of an enclosing
				//  outer loop, but as we trace downwards from each such block, we cannot exit the outer
				//  loop. In the decompilation process, we will only translate blocks within the outer loop
				//  that were reachable from inner loop exit targets. E.g.:
				//  
				//  loop  -- loop 1
				//    ...
				//    loop  -- loop 0
				//      ...
				//      exit when (STARSExitLoop0ToBlock7);
				//      ...
				//      exit when (STARSExitLoop0ToBlock9);
				//    end loop;  -- end loop 0
				//    if (STARSExitLoop0ToBlock7) then
				//       [code for block 7]
				//    end if;
				//    if (STARSExitLoop0ToBlock9) then
				//       [code for block 9]
				//    end if;
				//  end loop; -- end loop 1
				//  [code for block 10]
				//  ...
				//  Following down from block 9, we might take a LOOP_EXIT or INVERTED_LOOP_EXIT
				//   and reach block 10. But we don't want to follow that path while translating, because
				//   the inner loop is lifted into a separate loop procedure and only a call to it exists
				//   in the outer loop:
				//
				//  loop  -- loop 1, outer loop
				//    ...
				//    call innerloop0;
				//    if (STARSExitLoop0ToBlock7) then
				//       [code for block 7]
				//    end if;
				//    if (STARSExitLoop0ToBlock9) then
				//       [code for block 9]
				//    end if;
				//  end loop; -- end loop 1
				//
				// This code for loop 1 is a separate procedure, called from the main body of the 
				//  enclosing procedure:
				//
				// procedure foo(argument list)
				//   ...
				//   call outerloop1;
				//   [code for block 10]
				//  ...
				//
				// We see that code that logically follows the "end loop;" for the outer loop 1
				//  is placed after the "call outerloop1;" line. Likewise, the multiple exit targets
				//  for innerloop0 get placed after the "call innerloop0;" line. So the loops themselves
				//  are separate procedures, and the calls to them do not occur in the same procedures;
				//  the call to outerloop1 is from foo() while the call to innerloop0 is from outerloop1.
				// Whether there is a single exit target or multiple exit targets, their SPARK Ada translation
				//  follows the call to their loop. We see that [code for block 10] does not end up in the
				//  same procedure as block 9, so the reachability of block 10 from block 9 is not relevant to
				//  our translation. Therefore we end our reachability marking when we are about to exit another
				//  loop level. Such a multi-level trace can be detected by the existing method that detects
				//  multi-level loop breaks, which are unstructured and need to be structured separately.
				bool InsideSomeLoop = (0 <= LoopNum) || ConditionalInsideLoop;
				bool BadLoopBreak = (!InsideSomeLoop) ? false : this->DetectMultiLevelLoopBreak(HeadBlockNum, SuccBlockNum);
				if (!BadLoopBreak) {
					bool LocalDominanceComputed = false; // expensive; only perform on demand
					bool LocalDominance = false;
					bool SuccInsideContainingLoop = (ConditionalInsideLoop && this->IsBlockInLoop(SuccBlockNum, (size_t)ContainingLoopNum));
					bool ConditionalExitingLoop = (SuccInsideContainingLoop && SuccBlock->IsLoopTailBlock() && (this->GetInnermostLoopNum(SuccBlockNum) == ContainingLoopNum));
					if (ConditionalExitingLoop) {
						if (!LocalDominanceComputed) {
							LocalDominance = this->DoesBlockDominateBlock(HeadBlockNum, SuccBlockNum);
							LocalDominanceComputed = true;
						}
						// Don't go beyond the tail block. Its successors can only be loop exit targets or the loop header.
						if (LocalDominance) {
							ReachableSet.SetBit((size_t)SuccBlockNum); // Loop tail block is reachable, but don't recurse.
						}
						continue;
					}
					else if (ConditionalInsideLoop && (!SuccInsideContainingLoop)) {
						// We are already outside the containing loop if the passed-in BlockNum exits the loop.
						continue; // Don't mark a containing loop exit target as reachable
					}
					if (!LocalDominanceComputed) {
						LocalDominance = this->DoesBlockDominateBlock(HeadBlockNum, SuccBlockNum);
						LocalDominanceComputed = true;
					}
					if (LocalDominance) {
						this->MarkReachableBlocks(SuccBlockNum, LoopNum, HeadBlockNum, TestDominance, LimitBlockNum, ReachableSet, StopBlockNum);
					}
					else {
						// Terminating because we are escaping the region dominated by the loop head block.
						//  This SuccBlockNum is a candidate for the true loop follow block, where we will resume
						//  decompilation after we have processed all of the targets of loop exits and their
						//  reachable blocks.
						// NOTE: If the ExitTargetBlock (BlockNum) is not dominated by LoopHeadBlockNum,
						//  then we should not be looking at its successors in the first place.
						if ((0 <= LoopNum) && (0 > StopBlockNum)) { // Loop, but not second pass
							if (this->DoesBlockDominateBlock(HeadBlockNum, BlockNum)) {
								this->MultiLoopExitTargetsConvergenceMap[(size_t)LoopNum].insert(SuccBlockNum);
							}
						}
					}
				}
			}
		} // end for all successors
	}

	return;
} // end of SMPFunction::MarkReachableBlocks()

// How many of the multiple exit targets from LoopNum reach BlockNum?
uint32_t SMPFunction::GetReachableSourcesCount(const size_t TCFGBlockNum, const size_t LoopNum) const {
	assert(LoopNum < this->GetNumLoops());
	assert(TCFGBlockNum < this->GetNumTCFGBlocks());

	uint32_t ReachableSourcesCounter = 0;

	if (this->LoopExitTargets[LoopNum].size() > 1) {
		assert(!this->LoopExitTargetsReachabilitySets[LoopNum].empty());
		for (pair<int, STARSBitSet> MapItem : this->LoopExitTargetsReachabilitySets[LoopNum]) {
			if (MapItem.second.GetBit(TCFGBlockNum)) {
				++ReachableSourcesCounter;
			}
		}
	}

	return ReachableSourcesCounter;
} // end of SMPFunction::GetReachableSourcesCount()

// Downward path from BlockNum to Sentinel-reachable-block without hitting SentinelBlockNum?
bool SMPFunction::CanBlockBypassBlock(const int BlockNum, const int SentinelBlockNum, const int LimitBlockNum, const STARSBitSet &SentinelReachableBlocks) const {
	bool BypassDetected = false;

	if ((0 <= BlockNum) && (BlockNum <= LimitBlockNum) && (BlockNum != SentinelBlockNum)) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum((size_t)BlockNum);
		if (!CurrBlock->IsSCCPVisited()) {
			CurrBlock->SetSCCPVisited(true);
			if (SentinelReachableBlocks.GetBit((size_t)BlockNum)) {
				BypassDetected = true;
			}
			else {
				const list<SMPBasicBlock *> BlockSuccessors = CurrBlock->GetConstSuccList();
				for (const SMPBasicBlock *SuccBlock : BlockSuccessors) {
					int SuccBlockNum = SuccBlock->GetNumber();
					if (SuccBlockNum > BlockNum) { // no back edges
						if (this->CanBlockBypassBlock(SuccBlockNum, SentinelBlockNum, LimitBlockNum, SentinelReachableBlocks)) {
							BypassDetected = true;
							break;
						}
					}
				}
			}
		}
	}

	return BypassDetected;
} // end of SMPFunction::CanBlockBypassBlock()

// Collect a set of loop-variant DEFs with the inst IDs of the DEFs.
void SMPFunction::DetectLoopInvariantDEFs(void) {
	bool UseFP = this->UsesFramePointer();
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		set<STARS_ea_t> DummySet;
		this->LoopInvariantDEFs.push_back(DummySet); // allocate set in LoopInvariantDEFs vector

		list<size_t> BlockList;
		this->BuildLoopBlockList(LoopIndex, BlockList);

		assert(LoopIndex < ((size_t) this->LoopHeadBlockNumbers.size()));
		int HeadBlockNum = this->LoopHeadBlockNumbers[LoopIndex];

		list<size_t>::const_iterator BlockNumIter;
		for (BlockNumIter = BlockList.cbegin(); BlockNumIter != BlockList.cend(); ++BlockNumIter) {
			size_t BlockNum = (*BlockNumIter);
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);

			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
				if (DefIter == CurrInst->GetLastDef())
					continue;  // only the flags were defined
				STARSOpndTypePtr DefOp = DefIter->GetOp();
				// Only care about non-control-flow statements.
				if ((DEFAULT == CurrInst->GetDataFlowType()) && CurrInst->IsAnalyzeable() && MDIsDataFlowOpnd(DefOp, UseFP)) {
					STARS_ea_t DefAddr = CurrInst->GetAddr();
					// Iterate through USEs and find variant/invariant names.
					bool VariantUseFound = false;
					// Examine all USEs. A DEF is only loop invariant if the USEs that helped define it are also loop invariant.
					for (set<DefOrUse, LessDefUse>::iterator UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
						STARSOpndTypePtr UseOp = UseIter->GetOp();
						bool DirectStackAccess = MDIsDirectStackAccessOpnd(UseOp, UseFP);
						bool UnanalyzableUse;
						if (global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
							// Include direct stack accesses for SPARK analyses
							UnanalyzableUse = (UseOp->IsMemOp() && (!DirectStackAccess))
								|| (UseOp->GetReg() == STARS_x86_R_ip); // instruction pointer a.k.a. program counter
						}
						else {
							// Assume memory operands are loop variant until we have full alias analysis.
							// NOTE: This can cause mysterious annotation differences between SPARK mode
							//  and non-SPARK mode, as the LoopInVariantDefs are used in induction var analysis.
							UnanalyzableUse = UseOp->IsMemOp() || (UseOp->GetReg() == STARS_x86_R_ip);
						}
						if (UnanalyzableUse) {
							VariantUseFound = true;
							break;
						}
						else if (UseOp->IsImmedOp() || UseOp->IsVoidOp()) {
							; // immediates are always loop-invariant
						}
						else if (UseOp->MDIsSpecialRegOpType()) {
							// Probably should do data flow analysis on these just like regular regs in the future. !!!!****!!!!
							VariantUseFound = true;
							break;
						}
						else if (UseOp->IsRegOp() || DirectStackAccess) {
							// Find out if DEF is in a Phi function for the loop header, or otherwise outside the loop.
							bool LocalName = CurrBlock->IsLocalName(UseOp);
							// Invariant tests:
							// 1. If LocalName, UseOp is defined in the loop: VARIANT. For GlobalName, go to step 2.
							// 2. Find DEF of this global USE.
							//  a. If DEF is inside the loop, including internal block Phi functions, VARIANT.
							//  b. If DEF is outside the loop, this USE is loop-invariant, continue to analyze remaining USEs.
							//  c. If DEF is in loop header block Phi function, check all Phi USEs. If any Phi USE for
							//      the DEF of UseOp is inside the loop, then UseOp is VARIANT, else continue to remaining USEs.
							int UseSSANum = UseIter->GetSSANum();
							STARS_ea_t UseDefAddr = CurrBlock->GetDefAddrFromUseAddr(UseOp, DefAddr, UseSSANum, LocalName);
							bool UseDefInPhiFunction = STARS_IsBlockNumPseudoID(UseDefAddr);
							if (UseDefInPhiFunction) {// found in Phi DEF; DefAddr encodes block #
								size_t UseDefBlockNum = STARS_GetBlockNumFromPseudoID(UseDefAddr);
								// Is UseDefBlockNum inside our loop?
								if (this->IsBlockInLoop(UseDefBlockNum, LoopIndex)) {
									// Phi DEF is inside loop, but if it is in the loop header block, it is
									//  still possible that the Phi DEF is loop-invariant in the special case
									//  where the loop header block is the join point for a previous if-then-else
									//  and the Phi function is joining loop-invariant values, as opposed to the
									//  common case in which a loop header block Phi function joins an incoming value
									//  to a loop-variant value. So, we must examine the loop variance of the Phi USEs
									//  only if UseDefBlockNum is the loop header block number.
									if (HeadBlockNum != ((int) UseDefBlockNum)) {
										VariantUseFound = true;
										break;
									}
									else { // header block Phi function; must examine all Phi USEs
										SMPBasicBlock *PhiBlock = this->GetBlockByNum(UseDefBlockNum);
										set<SMPPhiFunction, LessPhi>::iterator PhiIter = PhiBlock->FindPhi(UseOp);
										assert(PhiIter != PhiBlock->GetLastPhi());
										size_t PhiLimit = PhiIter->GetPhiListSize();
										for (size_t PhiIndex = 0; PhiIndex < PhiLimit; ++PhiIndex) {
											int PhiUseSSANum = PhiIter->GetUseSSANum(PhiIndex);
											STARS_ea_t PhiUseDefAddr = this->GetGlobalDefAddr(UseOp, PhiUseSSANum);
											if (STARS_IsBlockNumPseudoID(PhiUseDefAddr)) { // found in Phi DEF; addr encodes block #
												// DEF of current Phi USE found in previous block Phi function; must be loop-variant
												//  if the DEF block is inside our loop, i.e. header block of current loop has a Phi USE
												//  whose DEF is some join point inside the loop, like an if-then-else join point.
												size_t PhiUseDefBlockNum = STARS_GetBlockNumFromPseudoID(PhiUseDefAddr);
												if (this->IsBlockInLoop(PhiUseDefBlockNum, LoopIndex)) {
													VariantUseFound = true;
													break;
												}
											}
											else if (STARS_BADADDR == PhiUseDefAddr) {
												VariantUseFound = true;
												break;
											}
											else { // found instruction addr
												SMPBasicBlock *PhiUseDefBlock = this->GetBlockFromInstAddr(PhiUseDefAddr);
												size_t PhiUseDefBlockNum = PhiUseDefBlock->GetNumber();
												if (this->IsBlockInLoop(PhiUseDefBlockNum, LoopIndex)) {
													// Phi is normal case of joining incoming DEF to a DEF inside the loop.
													//  Current Phi Use is the one defined inside loop.
													VariantUseFound = true;
													break;
												}
											}
										} // end for all Phi USEs
									}
								}
							} // end if UseDefInPhiFunction
							else {
								bool UseDefInvariant = (this->LoopInvariantDEFs[LoopIndex].find(UseDefAddr) != this->LoopInvariantDEFs[LoopIndex].end());
								if (UseDefInvariant) {
									; // we already know the current USE is loop-invariant for the current loop
									  //  because we found the DEF in the LoopInvariantDEFs set
								}
								else if (LocalName) { // block-local DEFs in a loop are always loop-variant
									VariantUseFound = true;
									break;
								}
								else { // we have an instruction ID, not a block number, for UseDefAddr
									SMPBasicBlock *UseDefBlock = this->GetBlockFromInstAddr(UseDefAddr);
									size_t UseDefBlockNum = UseDefBlock->GetNumber();
									if (this->IsBlockInLoop(UseDefBlockNum, LoopIndex)) {
										VariantUseFound = true;
										break;
									}
								}
							}
						}
						else {
							SMP_msg("ERROR: Unknown operand type at %llx in DetectLoopInvariantDEFs ", (uint64_t) DefAddr);
							this->DumpFuncNameAndAddr();
							VariantUseFound = true;
							break;
						}
					} // end for all USEs

					if (!VariantUseFound) {
						pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->LoopInvariantDEFs[LoopIndex].insert(DefAddr);
						++LoopInvariantDEFCount;
#if STARS_DEBUG_LOOP_INVARIANTS
						SMP_msg("INFO: Invariant DEF at %llx LoopHeadAddr: %llx\n", (uint64_t) DefAddr,
							(uint64_t) this->GetBlockByNum(HeadBlockNum)->GetFirstAddr());
#endif
					}
				}
			} // end for all insts in block
		} // end for all blocks in loop
	} // end for all loops in function

	int CurrentBlockNum = SMP_BLOCKNUM_UNINIT;
	list<size_t> LoopList;
	for (size_t LoopNum = 0; LoopNum < this->LoopCount; ++LoopNum) {
		// Detect DEFs that are invariant for all loops containing the DEF.
		for (set<STARS_ea_t>::const_iterator DefAddrIter = this->LoopInvariantDEFs[LoopNum].cbegin();
			DefAddrIter != this->LoopInvariantDEFs[LoopNum].cend(); ++DefAddrIter) {
			STARS_ea_t CurrDefAddr = (*DefAddrIter);
			SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(CurrDefAddr);
			assert(nullptr != CurrBlock);
			if (CurrBlock->GetNumber() != CurrentBlockNum) { // cached LoopList is stale
				CurrentBlockNum = CurrBlock->GetNumber();
				LoopList.clear();
				this->BuildLoopList((int) CurrentBlockNum, LoopList);
			}
			bool InvariantForAllLoops = true;
			for (list<size_t>::const_iterator LoopListIter = LoopList.cbegin(); LoopListIter != LoopList.cend(); ++LoopListIter) {
				size_t LoopNum = (*LoopListIter);
				if (this->LoopInvariantDEFs[LoopNum].find(CurrDefAddr) == this->LoopInvariantDEFs[LoopNum].end()) {
					InvariantForAllLoops = false; // One loop containing CurrDefAddr does not have CurrDefAddr as invariant
					break;
				}
			} // end for all loops in LoopList for current block
			if (InvariantForAllLoops) {
				vector<SMPInstr *>::iterator InstIter = CurrBlock->GetInstIterFromAddr(CurrDefAddr);
				assert(InstIter != CurrBlock->GetLastInst());
				SMPInstr *CurrInst = (*InstIter);
				STARSDefUseIter DefIter = CurrInst->GetFirstNonFlagsDef();
				assert(DefIter != CurrInst->GetLastDef());
				DefIter = CurrInst->SetDefLoopInvariant(DefIter->GetOp());
			}
		} // end for all addresses in current loop
	} // end for all loops

	return;
} // end of SMPFunction::DetectLoopInvariantDEFs()

// Track UseOp to its DEF, see if loop-invariant for LoopIndex
bool SMPFunction::IsUseLoopInvariantDEF(std::size_t LoopIndex, const STARSOpndTypePtr &UseOp, SMPInstr *UseInst) {
	bool UseDefInvariant = false;

	if (!this->LoopInvariantDEFs.empty() && (!this->LoopInvariantDEFs[LoopIndex].empty())) {
		if (!UseOp->IsMemOp()) {
			STARSOpndTypePtr SearchOp = CloneIfNecessary(UseOp, this->UsesFramePointer());
			CanonicalizeOpnd(SearchOp);
			bool LocalName = UseInst->GetBlock()->IsLocalName(SearchOp);
			STARSDefUseIter UseIter = UseInst->FindUse(SearchOp);
			if (UseIter != UseInst->GetLastUse()) {
				int UseSSANum = UseIter->GetSSANum();
				STARS_ea_t UseDefAddr = UseInst->GetBlock()->GetDefAddrFromUseAddr(SearchOp, UseInst->GetAddr(), UseSSANum, LocalName);
				UseDefInvariant = (this->LoopInvariantDEFs[LoopIndex].find(UseDefAddr) != this->LoopInvariantDEFs[LoopIndex].end());
			}
		}
		else {
			// Check all address regs in the memory operand
			bool UseFP = this->UsesFramePointer();
			int BaseReg, IndexReg;
			uint16_t ScaleFactor;
			STARS_ea_t offset;
			UseOp->MDExtractAddressFields(BaseReg, IndexReg, ScaleFactor, offset);
			if ((BaseReg != STARS_x86_R_none) && (!MDIsStackPtrReg(BaseReg, UseFP))) {
				STARSOpndTypePtr BaseOp = UseInst->MakeRegOpnd((STARS_regnum_t)BaseReg);
				STARSDefUseIter UseIter = UseInst->FindUse(BaseOp);
				if (UseIter != UseInst->GetLastUse()) {
					bool LocalName = UseInst->GetBlock()->IsLocalName(BaseOp);
					int UseSSANum = UseIter->GetSSANum();
					STARS_ea_t UseDefAddr = UseInst->GetBlock()->GetDefAddrFromUseAddr(BaseOp, UseInst->GetAddr(), UseSSANum, LocalName);
					UseDefInvariant = (this->LoopInvariantDEFs[LoopIndex].find(UseDefAddr) != this->LoopInvariantDEFs[LoopIndex].end());
				}
				if (!UseDefInvariant) {
					return false;
				}
			}
			if ((IndexReg != STARS_x86_R_none) && (!MDIsStackPtrReg(IndexReg, UseFP))) {
				STARSOpndTypePtr IndexOp = UseInst->MakeRegOpnd((STARS_regnum_t)IndexReg);
				STARSDefUseIter UseIter = UseInst->FindUse(IndexOp);
				if (UseIter != UseInst->GetLastUse()) {
					bool LocalName = UseInst->GetBlock()->IsLocalName(IndexOp);
					int UseSSANum = UseIter->GetSSANum();
					STARS_ea_t UseDefAddr = UseInst->GetBlock()->GetDefAddrFromUseAddr(IndexOp, UseInst->GetAddr(), UseSSANum, LocalName);
					UseDefInvariant = (this->LoopInvariantDEFs[LoopIndex].find(UseDefAddr) != this->LoopInvariantDEFs[LoopIndex].end());
				}
			}
		}
	}
	return UseDefInvariant;
} // end of SMPFunction::IsUseLoopInvariantDEF()

// find loop # corresponding to header block num
size_t SMPFunction::FindLoopNumFromHeadBlockNum(int LoopHeaderBlockNum) const {
	for (size_t LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		if (this->LoopHeadBlockNumbers[LoopIndex] == LoopHeaderBlockNum) {
			return LoopIndex;
		}
	}
	assert(false); // should always find the loop if given a proper head block number
}

// Detect basic and dependent loop induction variable families for all loops
void SMPFunction::DetectLoopInductionVars(void) {
	bool UseFP = this->UsesFramePointer();
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();

	// Initialize induction var vector with entries that have BIV*0 and SMP_SSA_UNINIT as markers
	this->LoopInductionVars.resize(this->LoopCount);
	struct InductionVarTriple DummyBIV;
	SMPInstr *FirstInst = this->GetInstFromAddr(this->GetFirstFuncAddr());
	STARSOpndTypePtr ZeroOp = FirstInst->MakeImmediateOpnd(0);
	DefOrUse MultUse(ZeroOp, UNKNOWN);
	DummyBIV.Multiplier = MultUse;
	struct InductionVarFamily DummyFamily;
	DummyFamily.BIVIncomingSSANum = SMP_SSA_UNINIT;
	DummyFamily.BasicInductionVar = DummyBIV;
	DummyFamily.BIVInitExpr = nullptr;
	DummyFamily.BIVLimitExpr = nullptr;
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		this->LoopInductionVars[LoopIndex].push_back(DummyFamily);
	}

	// Go through the Phi functions in the header blocks for all loops to search for basic induction var candidates.
	for (size_t LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		// A basic induction variable meets the following criteria:
		//  1. Defined by a Phi function in the loop header block.
		//  2. Phi function has two Phi USEs, one from a DEF outside the loop, one from a DEF inside the loop.
		//  3. The DEF inside the loop is a simple +/-/*/negate function of the Phi DEF and loop invariant values.
		// Note that some texts say "constants" instead of "loop invariant values" in #3, but that is too restrictive.
		// We will limit ourselves to basic induction var registers for the present time.
		// UPDATE: We allow more than 2 Phi USEs, as long as each inside-the-loop Phi USE is the
		//  same induction var arithmetic.
		int HeaderBlockNum = this->LoopHeadBlockNumbers[LoopIndex];
		bool FoundBIV = false;
		SMPBasicBlock *HeaderBlock = this->GetBlockByNum((size_t) HeaderBlockNum);
		// Look at all register Phi functions in the HeaderBlock.
		for (PhiSetIter PhiIter = HeaderBlock->GetFirstPhi(); PhiIter != HeaderBlock->GetLastPhi(); ++PhiIter) {
			struct InductionVarTriple BIVTriple;
			struct InductionVarFamily CurrentFamily;
			STARSOpndTypePtr PhiDefOp = PhiIter->GetAnyOp();
			size_t PhiUseListSize = PhiIter->GetPhiListSize();
			CurrentFamily.BIVInitExpr = nullptr;
			CurrentFamily.BIVLimitExpr = nullptr;
			if (!MDIsDataFlowOpnd(PhiDefOp, UseFP))
				continue;
			bool InsideDEFFound = false;
			bool OutsideDEFFound = false;
			bool InsideDEFIsInst = false;
			size_t InsideBIVFoundCount = 0;
			STARS_ea_t InsideDefAddr = STARS_BADADDR;
			STARS_ea_t OutsideDefAddr = STARS_BADADDR;
			int OutsideSSANum = -1;
			STARS_ea_t UseDefAddr = STARS_BADADDR;
			int PhiUseDefBlockNum1 = SMP_BLOCKNUM_UNINIT;
			for (size_t PhiUseIndex = 0; PhiUseIndex < PhiUseListSize; ++PhiUseIndex) {
				bool CurrentDEFIsInst = false;
				DefOrUse PhiUse = PhiIter->GetPhiRef(PhiUseIndex);
				int PhiUseSSANum = PhiIter->GetUseSSANum(PhiUseIndex);
				if (0 > PhiUseSSANum) // bad path or uninitialized Phi USE
					break;
				UseDefAddr = this->GetGlobalDefAddr(PhiDefOp, PhiUseSSANum);
				assert(STARS_BADADDR != UseDefAddr);
				if (STARS_IsSSAMarkerPseudoID(UseDefAddr))
					PhiUseDefBlockNum1 = 0;
				else if (STARS_IsBlockNumPseudoID(UseDefAddr))
					PhiUseDefBlockNum1 = STARS_GetBlockNumFromPseudoID(UseDefAddr);
				else {
					CurrentDEFIsInst = true;
					PhiUseDefBlockNum1 = this->GetBlockNumFromInstAddr(UseDefAddr);
				}
				if (this->IsBlockInLoop(PhiUseDefBlockNum1, LoopIndex)) {
					InsideDEFFound = true;
					InsideDefAddr = UseDefAddr;
					InsideDEFIsInst = CurrentDEFIsInst;
					if (!InsideDEFIsInst)
						break; // Cannot analyze internal Phi DEFs right now.
					CurrentFamily.BIVInsideLoopDefAddrs.push_back(InsideDefAddr);
				}
				else {
					OutsideDEFFound = true;
					OutsideSSANum = PhiUseSSANum;
					OutsideDefAddr = UseDefAddr;
				}
			} // end for all Phi USEs
			if (!(InsideDEFFound && OutsideDEFFound))
				continue;
			if (!InsideDEFIsInst)
				continue; // Need linear function statement DEF in loop, not another Phi function DEF.
			if ((1 + CurrentFamily.BIVInsideLoopDefAddrs.size()) != PhiUseListSize)
				continue; // need one outside DEF, rest inside DEFs

			// We have satisfied criteria 1 and 2 if we reach this point. Now for criterion 3.
			//  For each InsideDefAddr, is there an inst that USEs the Phi DEF and DEFs the
			//  inside Phi USE using simple induction var arithmetic?
			for (size_t InsideIndex = 0; InsideIndex < CurrentFamily.BIVInsideLoopDefAddrs.size(); ++InsideIndex) {
				InsideDefAddr = CurrentFamily.BIVInsideLoopDefAddrs[InsideIndex];
				SMPInstr *InsideDefInst = this->GetInstFromAddr(InsideDefAddr);
				STARSDefUseIter UseIter = InsideDefInst->FindUse(PhiDefOp);
				if ((UseIter != InsideDefInst->GetLastUse()) && (UseIter->GetSSANum() == PhiIter->GetDefSSANum())) {
					STARSOpndTypePtr SecondOp = nullptr;
					SMPoperator BIVOperator = SMP_NULL_OPERATOR;
					STARSDefUseIter SecondOpDefIter;
					if (InsideDefInst->IsBasicInductionVarArithmetic(SecondOp, BIVOperator)) {
						// Last question: Is the "y" operand in x := x op y loop-invariant?
						bool BIVarFound = ((nullptr != SecondOp) && SecondOp->IsImmedOp());
						int SecondOpSSANum = SMP_SSA_UNINIT;
						if (!BIVarFound && MDIsDataFlowOpnd(SecondOp, UseFP) && this->IsGlobalName(SecondOp)) {
							STARSOpndTypePtr SearchOp = CloneIfNecessary(SecondOp, UseFP);
							CanonicalizeOpnd(SearchOp);
							STARSDefUseIter SecondOpUseIter = InsideDefInst->FindUse(SearchOp);
							assert(SecondOpUseIter != InsideDefInst->GetLastUse());
							SecondOpSSANum = SecondOpUseIter->GetSSANum();

							// Last chance to detect BIV: Is SecondOp a USE of a loop-invariant DEF?
							//  Note that block-local names are loop-variant and are already excluded.
							SMPBasicBlock *InsideDefBlock = this->GetBlockByNum((size_t)PhiUseDefBlockNum1);
							assert(nullptr != InsideDefBlock);
							STARS_ea_t SecondOpDefAddr = InsideDefBlock->GetDefAddrFromUseAddr(SearchOp, InsideDefAddr, SecondOpUseIter->GetSSANum(), false);
							assert(STARS_BADADDR != SecondOpDefAddr);
							int SecondOpDefBlockNum = SMP_BLOCKNUM_UNINIT;
							bool SecondOpDEFIsInst = false;
							if (STARS_IsSSAMarkerPseudoID(SecondOpDefAddr)) { // SSA Marker inst
								SecondOpDefBlockNum = 0;
								assert(SecondOpSSANum == 0);
							}
							else if (STARS_IsBlockNumPseudoID(SecondOpDefAddr)) {
								SecondOpDefBlockNum = STARS_GetBlockNumFromPseudoID(SecondOpDefAddr);
							}
							else {
								SecondOpDEFIsInst = true;
								SecondOpDefBlockNum = this->GetBlockNumFromInstAddr(SecondOpDefAddr);
							}
							if (this->IsBlockInLoop(SecondOpDefBlockNum, LoopIndex)) {
								if (SecondOpDEFIsInst) {
									// See if SecondOpDEF inst is on the loop-invariant DEFs list.
									if (this->LoopInvariantDEFs[LoopIndex].find(SecondOpDefAddr) != this->LoopInvariantDEFs[LoopIndex].end()) {
										BIVarFound = true;
									}
								}
							}
							else { // DEF of SecondOp is outside the loop entirely
								BIVarFound = true;
							}
						}
						if (BIVarFound) {
							DefOrUse PhiDef(PhiDefOp, PhiIter->GetDefType(), PhiIter->GetDefSSANum());
							BIVTriple.InductionVar = PhiDef;
							BIVTriple.SubtractAddend = (SMP_SUBTRACT == BIVOperator);
							if ((SMP_ADD == BIVOperator) || (SMP_SUBTRACT == BIVOperator)) {
								// Adding an operand to itself is a strength reduction of multiplying by two.
								if (IsEqOpIgnoreBitwidth(PhiDefOp, SecondOp)) {
									if (SMP_ADD == BIVOperator) { // SMP_SUBTRACT would be zeroing out, not a progression.
										// Make triple: 2*InductionVar + 0
										STARSOpndTypePtr MultiplierOp = InsideDefInst->MakeImmediateOpnd(2);
										DefOrUse MultUse(MultiplierOp, NUMERIC);
										BIVTriple.Multiplier = MultUse;
										STARSOpndTypePtr AddendOp = InsideDefInst->MakeImmediateOpnd(0);
										DefOrUse AddendUse(AddendOp, NUMERIC);
										BIVTriple.Addend = AddendUse;
									}
								}
								else {
									// Make triple: 1*InductionVar + SecondOp
									DefOrUse AddendUse(SecondOp, NUMERIC, SecondOpSSANum);
									BIVTriple.Addend = AddendUse;
									STARSOpndTypePtr MultiplierOp = InsideDefInst->MakeImmediateOpnd(1);
									DefOrUse MultUse(MultiplierOp, NUMERIC);
									BIVTriple.Multiplier = MultUse;
								}
							}
							else if ((SMP_S_LEFT_SHIFT == BIVOperator) || (SMP_U_LEFT_SHIFT == BIVOperator)) {
								// Must be a shift left by constant number of bits, same as multiply.
								assert(SecondOp->IsImmedOp());
								// Make triple: (1 << SecondOp)*InductionVar + 0
								STARS_uval_t ShiftCount = SecondOp->GetImmedValue();
								STARS_uval_t MultiplierValue = (1 << ShiftCount);
								STARSOpndTypePtr AddendOp = InsideDefInst->MakeImmediateOpnd(0);
								DefOrUse AddendUse(AddendOp, NUMERIC);
								BIVTriple.Addend = AddendUse;
								STARSOpndTypePtr MultiplierOp = InsideDefInst->MakeImmediateOpnd(MultiplierValue);
								DefOrUse MultUse(MultiplierOp, NUMERIC, SecondOpSSANum);
								BIVTriple.Multiplier = MultUse;
							}
							else {
								assert((SMP_U_MULTIPLY == BIVOperator) || (SMP_S_MULTIPLY == BIVOperator));
								// Make triple: SecondOp*InductionVar + 0
								STARSOpndTypePtr AddendOp = InsideDefInst->MakeImmediateOpnd(0);
								DefOrUse AddendUse(AddendOp, NUMERIC);
								BIVTriple.Addend = AddendUse;
								DefOrUse MultUse(SecondOp, NUMERIC, SecondOpSSANum);
								BIVTriple.Multiplier = MultUse;
							}
							if (0 < InsideBIVFoundCount) { // not first BIV candidate
								// Must match previous candidates.
								if (EqualInductionVars(BIVTriple, CurrentFamily.BasicInductionVar)) {
									// All is well; no need to update.
									FoundBIV = true;
									++InsideBIVFoundCount;
								}
								else {
									FoundBIV = false;
									SMP_msg("ERROR: Non-matching BIV candidates in loop %d ", LoopIndex);
									this->DumpFuncNameAndAddr();
									SMP_msg("Current BIV candidate: ");
									DumpInductionVar(BIVTriple);
									SMP_msg("\nExisting BIV candidate: ");
									DumpInductionVar(CurrentFamily.BasicInductionVar);
									SMP_msg("\n");
									if (VerboseOutput) {
										this->Dump(false);
										this->DumpDotCFG();
									}
									break;
								}
							}
							else { // First BIV candidate
								CurrentFamily.BIVIncomingSSANum = OutsideSSANum;
								CurrentFamily.BIVIncomingDefAddr = OutsideDefAddr;
								CurrentFamily.BasicInductionVar = BIVTriple;
								FoundBIV = true;
								++InsideBIVFoundCount;
							}
							if (FoundBIV && VerboseOutput) {
								SMP_msg("INFO: Basic Induction Var for func at %llx for Loop %d :", (uint64_t) this->GetFirstFuncAddr(), LoopIndex);
								DumpInductionVar(BIVTriple);
							}
						}
					} // end if BasicInductionVarArithmetic found
				} // end if we found USE of Phi DEF
			} // end for all values of InsideIndex
			if (InsideBIVFoundCount == CurrentFamily.BIVInsideLoopDefAddrs.size()) {
				this->LoopInductionVars[LoopIndex].push_back(CurrentFamily);
				SMP_msg("INFO: BIVFoundCount success for func at %llx for Loop %d \n", (uint64_t) this->GetFirstFuncAddr(), LoopIndex);
			}
		} // end for all Phi functions in current loop header block
		if (!FoundBIV) {
			SMP_msg("ERROR: LOOP: BIV not found for loop %d at %llx ", LoopIndex, (uint64_t) HeaderBlock->GetFirstAddr());
			this->DumpFuncNameAndAddr();
			++STARS_LoopInductionVarIDFailures;
		}
		else {
			++STARS_LoopInductionVarIDSuccesses;
		}
	} // end for all loops

#if 1
	// Look for dependent induction variables, which are linear functions of other induction variables.
	list<size_t> BlockList;
	for (size_t LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		this->BuildLoopBlockList(LoopIndex, BlockList);
		for (list<size_t>::const_iterator BlockIter = BlockList.cbegin(); BlockIter != BlockList.cend(); ++BlockIter) {
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(*BlockIter);
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				STARSOpndTypePtr Mult1 = nullptr, Mult2 = nullptr, Add1 = nullptr, Add2 = nullptr;
				SMPoperator RhsOperator = SMP_NULL_OPERATOR;

				if (CurrInst->IsDependentInductionVarArithmetic(Mult1, Mult2, Add1, Add2, RhsOperator)) {
					// See if one of Mult1, Mult2, Add1 or Add2 is a basic induction variable.
					struct DependentInductionVar DIV;
					this->FindDependentInductionVar(LoopIndex, DIV, Add1, Add2, Mult1, Mult2, RhsOperator, CurrInst);
					if (VerboseOutput && (STARS_BADADDR != DIV.DIVDefAddr)) {
						SMP_msg("INFO: Dependent Induction Variable for loop # %d found at %llx\n", LoopIndex, (uint64_t)CurrInst->GetAddr());
						PrintOperand(DIV.DIV.GetOp());
						SMP_msg("\n");
						DumpInductionVar(DIV.IVExpr);
					}
				}
			}
		}

		BlockList.clear();
	}
#endif

	return;
} // end of SMPFunction::DetectLoopInductionVars()

struct IVTracker {
	STARSBitSet CurrSSAClosure;
	bool ValidIVOpsOnly;
};

list<struct IVTracker>::iterator FindSSAClosure(list<struct IVTracker> &IVTrackerList, const int SSANum1, const int SSANum2) {
	list<struct IVTracker>::iterator MatchingIter;
	for (MatchingIter = IVTrackerList.begin(); MatchingIter != IVTrackerList.end(); ++MatchingIter) {
		if ((*MatchingIter).CurrSSAClosure.GetBit((size_t)SSANum1)
			|| (*MatchingIter).CurrSSAClosure.GetBit((size_t)SSANum2)) {
			return MatchingIter;
		}
	}

	return IVTrackerList.end();
} // end of FindSSAClosure()

// Find SSAClosure from PhiIter for loop; return true if IV definition is satisfied
bool SMPFunction::FindIVLiveRangeClosure(const size_t LoopIndex, const PhiSetIter PhiIter, STARSBitSet &SSAClosure) {
	bool ValidIVOperations = false;
	list<size_t> BlockNumList;
	this->BuildLoopBlockList(LoopIndex, BlockNumList);
	int HeaderBlockNum = this->LoopHeadBlockNumbers[LoopIndex];

	// Start the closure with the SSA numbers in the Phi function.
	list<struct IVTracker> IVTrackerList; // for all working closures except the Phi closure.
	struct IVTracker PhiIVTracker;
	PhiIVTracker.ValidIVOpsOnly = true;
	assert(0 <= this->GetMaxStackSSANum());
	PhiIVTracker.CurrSSAClosure.AllocateBits((size_t) this->GetMaxStackSSANum());
	bool RegSpilled = false;
	bool RegUnspilled = false;

	STARSOpndTypePtr PhiOp = PhiIter->GetAnyOp();
	// For all instructions in the loop, find SSA nums for the PhiOp, 
	//  whether USE or DEF. Accumulate closures, merge closures when
	//  they overlap, track whether operations are acceptable for an IV.
	// Detect spills to memory and ignore ensuing USEs/DEFs until the item
	//  is unspilled from that memory location.
	for (const size_t BlockNum : BlockNumList) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		// Get Phi function DEF and USE SSA numbers.
		PhiSetIter CurrPhiIter = CurrBlock->FindPhi(PhiOp);
		if (CurrPhiIter != CurrBlock->GetLastPhi()) {
			// See if this set of SSA nums overlaps the PhiIVTracker, or this is the header block.
			bool MainPhiTracker = (BlockNum == ((size_t) HeaderBlockNum));
			if (!MainPhiTracker) {
				for (size_t PhiUseIndex = 0; PhiUseIndex < CurrPhiIter->GetPhiListSize(); ++PhiUseIndex) {
					int SSANum = CurrPhiIter->GetUseSSANum(PhiUseIndex);
					if (PhiIVTracker.CurrSSAClosure.GetBit((size_t)SSANum)) {
						MainPhiTracker = true;
						break;
					}
				}
			}
			if (MainPhiTracker) {
				int SSANum = CurrPhiIter->GetDefSSANum();
				PhiIVTracker.CurrSSAClosure.SetBit((size_t)SSANum);
				for (size_t PhiUseIndex = 0; PhiUseIndex < CurrPhiIter->GetPhiListSize(); ++PhiUseIndex) {
					SSANum = CurrPhiIter->GetUseSSANum(PhiUseIndex);
					PhiIVTracker.CurrSSAClosure.SetBit((size_t)SSANum);
				}
			}
			else {
				// Find in another tracking set, or create a new one.
				int SSANum1 = CurrPhiIter->GetUseSSANum(0);
				int SSANum2 = CurrPhiIter->GetUseSSANum(1);
				list<struct IVTracker>::iterator TrackIter = FindSSAClosure(IVTrackerList, SSANum1, SSANum2);
				if (TrackIter != IVTrackerList.end()) {
					int SSANum = CurrPhiIter->GetDefSSANum();
					(*TrackIter).CurrSSAClosure.SetBit((size_t)SSANum);
					for (size_t PhiUseIndex = 0; PhiUseIndex < CurrPhiIter->GetPhiListSize(); ++PhiUseIndex) {
						SSANum = CurrPhiIter->GetUseSSANum(PhiUseIndex);
						(*TrackIter).CurrSSAClosure.SetBit((size_t)SSANum);
					}
				}
				else { // Create new tracking set.
					struct IVTracker NewTrackingSet;
					NewTrackingSet.CurrSSAClosure.AllocateBits((size_t) this->GetMaxStackSSANum());
					NewTrackingSet.ValidIVOpsOnly = true;
					int SSANum = CurrPhiIter->GetDefSSANum();
					NewTrackingSet.CurrSSAClosure.SetBit((size_t)SSANum);
					for (size_t PhiUseIndex = 0; PhiUseIndex < CurrPhiIter->GetPhiListSize(); ++PhiUseIndex) {
						SSANum = CurrPhiIter->GetUseSSANum(PhiUseIndex);
						NewTrackingSet.CurrSSAClosure.SetBit((size_t)SSANum);
					}
					IVTrackerList.push_back(NewTrackingSet);
				}
			}
		} // end if found in current block phi functions

		if (!CurrBlock->IsVarKill(PhiOp))
			continue; // save time

		bool InSpillRange = false; // Waiting for unspill?
		STARSOpndTypePtr SpillMemOp = nullptr;
		for (vector<SMPInstr *>::const_iterator InstIter = CurrBlock->GetFirstConstInst();
			InstIter != CurrBlock->GetLastConstInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);

			STARSDefUseIter UseIter = CurrInst->FindUse(PhiOp);
			if (UseIter != CurrInst->GetLastUse()) { // If so, search for spill.
				STARSOpndTypePtr CurrSpillMemOp = CurrInst->GetMemDef();
				STARSOpndTypePtr SourceOp = nullptr;
				if ((nullptr != CurrSpillMemOp) && CurrInst->IsSimpleCopy(SourceOp)) { // store to memory
					if (IsEqOpIgnoreBitwidth(SourceOp, PhiOp)) { // spill of PhiOp to memory
						InSpillRange = true;
						SpillMemOp = CurrSpillMemOp;
						continue; 
					}
				}
			}

			STARSDefUseIter DefIter = CurrInst->FindDef(PhiOp);
			if (DefIter != CurrInst->GetLastDef()) {
				int DefSSANum = DefIter->GetSSANum();

				// Is this an unspill from memory?
				STARSOpndTypePtr SourceOp = nullptr;
				if (InSpillRange) {
					if (CurrInst->IsSimpleCopy(SourceOp)) {
						if (IsEqOp(SourceOp, SpillMemOp)) { // unspill from previous location
							InSpillRange = false;
							PhiIVTracker.CurrSSAClosure.SetBit((size_t)DefSSANum);
						}
					}
					continue; // ignore if in spill range; if unspill, already handled by SetBit()
				}

				// See if PhiOp is also a USE, find its closure set.
				STARSDefUseIter UseIter = CurrInst->FindUse(PhiOp);
				int UseSSANum = DefSSANum; // lets us use FindSSAClosure() later
				if (UseIter != CurrInst->GetLastUse()) {
					UseSSANum = UseIter->GetSSANum();
				}
				if ((0 > UseSSANum) || (0 > DefSSANum))
					return false;
				STARSOpndTypePtr RightOp = nullptr;
				SMPoperator CurrOperator;
				bool CurrSpilled = false;
				bool CurrUnspilled = false;
				bool CurrValidIVOp = CurrInst->IsAcceptableBIVOperation(CurrSpilled, CurrUnspilled) || CurrInst->IsBasicInductionVarArithmetic(RightOp, CurrOperator);
				// Most common case is we only have one closure.
				if (PhiIVTracker.CurrSSAClosure.GetBit((size_t)UseSSANum) || PhiIVTracker.CurrSSAClosure.GetBit((size_t)DefSSANum)) {
					// Make sure both bits are set.
					PhiIVTracker.CurrSSAClosure.SetBit((size_t)UseSSANum);
					PhiIVTracker.CurrSSAClosure.SetBit((size_t)DefSSANum);
					RegSpilled = (CurrSpilled || RegSpilled);
					RegUnspilled = (CurrUnspilled || RegUnspilled);
					if (!CurrValidIVOp)
						PhiIVTracker.ValidIVOpsOnly = false;
				}
				else {
					// Locate the matching closure, or create a new one.
					list<struct IVTracker>::iterator MatchingIter = FindSSAClosure(IVTrackerList, DefSSANum, UseSSANum);
					if (MatchingIter == IVTrackerList.end()) {
						// Need new closure set.
						struct IVTracker NewIVClosure;
						NewIVClosure.CurrSSAClosure.AllocateBits((size_t) this->GetMaxStackSSANum());
						NewIVClosure.ValidIVOpsOnly = CurrValidIVOp;
						IVTrackerList.push_back(NewIVClosure);
					}
					else { // Matched a closure
						// Ensure both SSA Num bits are set.
						(*MatchingIter).CurrSSAClosure.SetBit((size_t) DefSSANum);
						(*MatchingIter).CurrSSAClosure.SetBit((size_t) UseSSANum);
						// Update valid IV ops flag if needed.
						(*MatchingIter).ValidIVOpsOnly = (*MatchingIter).ValidIVOpsOnly && CurrValidIVOp;
					}
				}		
			}
		} // end for all inst iters in block
	} // end for all blocks in loop

	// See if any of the closures need to be merged.
	list<struct IVTracker>::iterator CurrIter = IVTrackerList.begin();
	while (CurrIter != IVTrackerList.end()) {
		list<struct IVTracker>::iterator NextIter = CurrIter;
		++NextIter;
		while (NextIter != IVTrackerList.end()) {
			STARSBitSet IntersectionSet = STARSBitSetIntersection((*CurrIter).CurrSSAClosure, (*NextIter).CurrSSAClosure);
			if (IntersectionSet.IsAnyBitSet()) { // some intersection found
				// Merge list entries.
				(*CurrIter).CurrSSAClosure.UnionSets((*NextIter).CurrSSAClosure);
				(*CurrIter).ValidIVOpsOnly = ((*CurrIter).ValidIVOpsOnly && (*NextIter).ValidIVOpsOnly);
				NextIter = IVTrackerList.erase(NextIter);
			}
			else {
				++NextIter;
			}
		}
		++CurrIter;
	}

	// Finally, see if the Phi closure needs to be merged with any of the remaining closures.
	CurrIter = IVTrackerList.begin();
	while (CurrIter != IVTrackerList.end()) {
		STARSBitSet IntersectionSet = STARSBitSetIntersection((*CurrIter).CurrSSAClosure, PhiIVTracker.CurrSSAClosure);
		if (IntersectionSet.IsAnyBitSet()) { // some intersection found
			// Merge list entries.
			PhiIVTracker.CurrSSAClosure.UnionSets((*CurrIter).CurrSSAClosure);
			PhiIVTracker.ValidIVOpsOnly = (PhiIVTracker.ValidIVOpsOnly && (*CurrIter).ValidIVOpsOnly);
		}
		++CurrIter;
	}

	// Should we demand perfect matching of RegSpilled and RegUnspilled? Track them in
	//  each element of the IVTrackerList?
	ValidIVOperations = PhiIVTracker.ValidIVOpsOnly;
	SSAClosure = PhiIVTracker.CurrSSAClosure;

	return ValidIVOperations;
} // end of SMPFunction::FindIVLiveRangeClosure()

// Detect basic and dependent loop induction variable families for all loops.
//  Use analyses suitable for binaries, in which register assignment could
//  make use of a register in a live range that does not overlap the live range
//  of the induction var operations, e.g.:
//
//  phi function for reg at loop header
//  possible use of phi def
//  spill reg to stack location
//  re-use reg in all kinds of ways, not as an IV
//  restore reg from spill location
//  do induction var arithmetic on reg
//
//  Textbook definitions of an IV are based on source code and don't
//  have to deal with register assignment, split live ranges, etc.
//  A textbook algorithm will disqualify the reg as an IV due to
//  the live range "re-use reg in all kinds of ways, not as an IV" above.
void SMPFunction::DetectLoopInductionVars2(void) {
	bool UseFP = this->UsesFramePointer();
	bool VerboseOutput = global_stars_interface->VerboseLoopsMode();
	bool AlreadyDumped = false; // prevent multiple debug dumps

	// Initialize induction var vector with entries that have BIV*0 and SMP_SSA_UNINIT as markers
	this->LoopInductionVars.resize(this->LoopCount);
	struct InductionVarTriple DummyBIV;
	SMPInstr *FirstInst = this->GetInstFromAddr(this->GetFirstFuncAddr());
	STARSOpndTypePtr ZeroOp = FirstInst->MakeImmediateOpnd(0);
	DefOrUse MultUse(ZeroOp, UNKNOWN);
	DummyBIV.Multiplier = MultUse;
	struct InductionVarFamily DummyFamily;
	DummyFamily.BIVIncomingSSANum = SMP_SSA_UNINIT;
	DummyFamily.BasicInductionVar = DummyBIV;
	DummyFamily.BIVInitExpr = nullptr;
	DummyFamily.BIVLimitExpr = nullptr;
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		this->LoopInductionVars[LoopIndex].push_back(DummyFamily);
	}

	// Go through the Phi functions in the header blocks for all loops to search for basic induction var candidates.
	for (size_t LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		int HeaderBlockNum = this->LoopHeadBlockNumbers[LoopIndex];
		bool FoundBIV = false;
		list<size_t> BlockNumList;
		this->BuildLoopBlockList(LoopIndex, BlockNumList);

		SMPBasicBlock *HeaderBlock = this->GetBlockByNum((size_t)HeaderBlockNum);
		// Look at all register Phi functions in the HeaderBlock.
		for (PhiSetIter PhiIter = HeaderBlock->GetFirstPhi(); PhiIter != HeaderBlock->GetLastPhi(); ++PhiIter) {
			STARSOpndTypePtr PhiDefOp = PhiIter->GetAnyOp();
			if (!MDIsDataFlowOpnd(PhiDefOp, UseFP))
				continue;

			STARSBitSet SSAClosure;
			SSAClosure.AllocateBits((size_t) this->GetMaxStackSSANum());
			if (!FindIVLiveRangeClosure(LoopIndex, PhiIter, SSAClosure))
				continue;

			struct InductionVarTriple BIVTriple;
			struct InductionVarFamily CurrentFamily;
			size_t PhiUseListSize = PhiIter->GetPhiListSize();
			CurrentFamily.BIVInitExpr = nullptr;
			CurrentFamily.BIVLimitExpr = nullptr;
			CurrentFamily.BIVIncomingSSANum = SMP_SSA_UNINIT;
			size_t InsideBIVFoundCount = 0;
			STARS_ea_t InsideDefAddr = STARS_BADADDR;
			STARS_ea_t OutsideDefAddr = STARS_BADADDR;
			int OutsideSSANum = -1;
			STARS_ea_t UseDefAddr = STARS_BADADDR;

			// Find outside the loop DEF.
			for (size_t index = 0; index < PhiUseListSize; ++index) {
				int SSANum = PhiIter->GetUseSSANum(index);
				STARS_ea_t DefAddr = this->GetGlobalDefAddr(PhiDefOp, SSANum);
				SMPBasicBlock *DefBlock = this->GetBlockFromInstAddr(DefAddr);
				if (!this->IsBlockInLoop(DefBlock->GetNumber(), LoopIndex)) {
					OutsideDefAddr = DefAddr;
					OutsideSSANum = SSANum;
					break;
				}
			}

			if (0 > OutsideSSANum)
				continue; // not valid IV

			for (const size_t BlockNum : BlockNumList) {
				SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
				if (!CurrBlock->IsVarKill(PhiDefOp))
					continue; // save time

				for (vector<SMPInstr *>::const_iterator InstIter = CurrBlock->GetFirstConstInst();
					InstIter != CurrBlock->GetLastConstInst(); ++InstIter) {
					SMPInstr *CurrInst = (*InstIter);
					STARS_ea_t InstAddr = CurrInst->GetAddr();
					STARSDefUseIter DefIter = CurrInst->FindDef(PhiDefOp);
					if (DefIter != CurrInst->GetLastDef()) {
						int DefSSANum = DefIter->GetSSANum();
						assert(0 <= DefSSANum);
						if (!SSAClosure.GetBit((size_t) DefSSANum))
							continue; // not part of IV live range
						CurrentFamily.BIVInsideLoopDefAddrs.push_back(InstAddr);

						STARSOpndTypePtr SecondOp = nullptr;
						SMPoperator BIVOperator = SMP_NULL_OPERATOR;
						STARSDefUseIter SecondOpDefIter;
						if (CurrInst->IsBasicInductionVarArithmetic(SecondOp, BIVOperator)) {
							// Last question: Is the "y" operand in x := x op y loop-invariant?
							bool BIVarFound = ((nullptr != SecondOp) && SecondOp->IsImmedOp());
							int SecondOpSSANum = SMP_SSA_UNINIT;
							if (!BIVarFound && MDIsDataFlowOpnd(SecondOp, UseFP) && this->IsGlobalName(SecondOp)) {
								STARSOpndTypePtr SearchOp = CloneIfNecessary(SecondOp, UseFP);
								CanonicalizeOpnd(SearchOp);
								STARSDefUseIter SecondOpUseIter = CurrInst->FindUse(SearchOp);
								assert(SecondOpUseIter != CurrInst->GetLastUse());
								SecondOpSSANum = SecondOpUseIter->GetSSANum();

								// Last chance to detect BIV: Is SecondOp a USE of a loop-invariant DEF?
								//  Note that block-local names are loop-variant and are already excluded.
								STARS_ea_t SecondOpDefAddr = CurrBlock->GetDefAddrFromUseAddr(SearchOp, InsideDefAddr, SecondOpUseIter->GetSSANum(), false);
								assert(STARS_BADADDR != SecondOpDefAddr);
								int SecondOpDefBlockNum = SMP_BLOCKNUM_UNINIT;
								bool SecondOpDEFIsInst = false;
								if (STARS_IsSSAMarkerPseudoID(SecondOpDefAddr)) { // SSA Marker inst
									SecondOpDefBlockNum = 0;
									assert(SecondOpSSANum == 0);
								}
								else if (STARS_IsBlockNumPseudoID(SecondOpDefAddr)) {
									SecondOpDefBlockNum = STARS_GetBlockNumFromPseudoID(SecondOpDefAddr);
								}
								else {
									SecondOpDEFIsInst = true;
									SecondOpDefBlockNum = this->GetBlockNumFromInstAddr(SecondOpDefAddr);
								}
								if (this->IsBlockInLoop(SecondOpDefBlockNum, LoopIndex)) {
									if (SecondOpDEFIsInst) {
										// See if SecondOpDEF inst is on the loop-invariant DEFs list.
										if (this->LoopInvariantDEFs[LoopIndex].find(SecondOpDefAddr) != this->LoopInvariantDEFs[LoopIndex].end()) {
											BIVarFound = true;
										}
									}
								}
								else { // DEF of SecondOp is outside the loop entirely
									BIVarFound = true;
								}
							}
							if (BIVarFound) {
								DefOrUse PhiDef(PhiDefOp, PhiIter->GetDefType(), PhiIter->GetDefSSANum());
								BIVTriple.InductionVar = PhiDef;
								BIVTriple.SubtractAddend = (SMP_SUBTRACT == BIVOperator);
								if ((SMP_ADD == BIVOperator) || (SMP_SUBTRACT == BIVOperator)) {
									// Adding an operand to itself is a strength reduction of multiplying by two.
									if (IsEqOpIgnoreBitwidth(PhiDefOp, SecondOp)) {
										if (SMP_ADD == BIVOperator) { // SMP_SUBTRACT would be zeroing out, not a progression.
											// Make triple: 2*InductionVar + 0
											STARSOpndTypePtr MultiplierOp = CurrInst->MakeImmediateOpnd(2);
											DefOrUse MultUse(MultiplierOp, NUMERIC);
											BIVTriple.Multiplier = MultUse;
											STARSOpndTypePtr AddendOp = CurrInst->MakeImmediateOpnd(0);
											DefOrUse AddendUse(AddendOp, NUMERIC);
											BIVTriple.Addend = AddendUse;
										}
									}
									else {
										// Make triple: 1*InductionVar + SecondOp
										DefOrUse AddendUse(SecondOp, NUMERIC, SecondOpSSANum);
										BIVTriple.Addend = AddendUse;
										STARSOpndTypePtr MultiplierOp = CurrInst->MakeImmediateOpnd(1);
										DefOrUse MultUse(MultiplierOp, NUMERIC);
										BIVTriple.Multiplier = MultUse;
									}
								}
								else if ((SMP_S_LEFT_SHIFT == BIVOperator) || (SMP_U_LEFT_SHIFT == BIVOperator)) {
									// Must be a shift left by constant number of bits, same as multiply.
									assert(SecondOp->IsImmedOp());
									// Make triple: (1 << SecondOp)*InductionVar + 0
									STARS_uval_t ShiftCount = SecondOp->GetImmedValue();
									STARS_uval_t MultiplierValue = (1 << ShiftCount);
									STARSOpndTypePtr AddendOp = CurrInst->MakeImmediateOpnd(0);
									DefOrUse AddendUse(AddendOp, NUMERIC);
									BIVTriple.Addend = AddendUse;
									STARSOpndTypePtr MultiplierOp = CurrInst->MakeImmediateOpnd(MultiplierValue);
									DefOrUse MultUse(MultiplierOp, NUMERIC, SecondOpSSANum);
									BIVTriple.Multiplier = MultUse;
								}
								else {
									assert((SMP_U_MULTIPLY == BIVOperator) || (SMP_S_MULTIPLY == BIVOperator));
									// Make triple: SecondOp*InductionVar + 0
									STARSOpndTypePtr AddendOp = CurrInst->MakeImmediateOpnd(0);
									DefOrUse AddendUse(AddendOp, NUMERIC);
									BIVTriple.Addend = AddendUse;
									DefOrUse MultUse(SecondOp, NUMERIC, SecondOpSSANum);
									BIVTriple.Multiplier = MultUse;
								}
								if (0 < InsideBIVFoundCount) { // not first BIV candidate
									// Must match previous candidates.
									if (EqualInductionVars(BIVTriple, CurrentFamily.BasicInductionVar)) {
										// All is well; no need to update.
										FoundBIV = true;
										++InsideBIVFoundCount;
									}
									else {
										FoundBIV = false;
										SMP_msg("ERROR: Non-matching BIV candidates in loop %d ", LoopIndex);
										this->DumpFuncNameAndAddr();
										SMP_msg("Current BIV candidate: ");
										DumpInductionVar(BIVTriple);
										SMP_msg("\nExisting BIV candidate: ");
										DumpInductionVar(CurrentFamily.BasicInductionVar);
										SMP_msg("\n");
										if (VerboseOutput && (!AlreadyDumped)) {
											this->Dump(false);
											this->DumpDotCFG();
											AlreadyDumped = true;
										}
										break;
									}
								}
								else { // First BIV candidate
									CurrentFamily.BIVIncomingSSANum = OutsideSSANum;
									CurrentFamily.BIVIncomingDefAddr = OutsideDefAddr;
									CurrentFamily.BasicInductionVar = BIVTriple;
									FoundBIV = true;
									++InsideBIVFoundCount;
								}
								if (FoundBIV && VerboseOutput) {
									SMP_msg("INFO: Basic Induction Var for func at %llx for Loop %d :", (uint64_t) this->GetFirstFuncAddr(), LoopIndex);
									DumpInductionVar(BIVTriple);
								}
							}
						} // end if BasicInductionVarArithmetic found
					}
				} // end for all insts in block
			} // end for all blocks in loop
			if (0 <= CurrentFamily.BIVIncomingSSANum) { // valid IV family
				CurrentFamily.BIVSSAClosure = SSAClosure;
				this->LoopInductionVars[LoopIndex].push_back(CurrentFamily);
				SMP_msg("INFO: BIVFoundCount success for func at %llx for Loop %d \n", (uint64_t) this->GetFirstFuncAddr(), LoopIndex);
			}
		} // end for all Phi functions in header block
		if (!FoundBIV) {
			SMP_msg("ERROR: LOOP: BIV not found for loop %d at %llx ", LoopIndex, (uint64_t)HeaderBlock->GetFirstAddr());
			this->DumpFuncNameAndAddr();
			++STARS_LoopInductionVarIDFailures;
		}
		else {
			++STARS_LoopInductionVarIDSuccesses;
		}
	} // end for all loops

	// Look for dependent induction variables, which are linear functions of other induction variables.
	list<size_t> BlockList;
	for (size_t LoopIndex = 0; LoopIndex < this->LoopHeadBlockNumbers.size(); ++LoopIndex) {
		this->BuildLoopBlockList(LoopIndex, BlockList);
		for (size_t BlockIndex : BlockList) {
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				STARSOpndTypePtr Mult1 = nullptr, Mult2 = nullptr, Add1 = nullptr, Add2 = nullptr;
				SMPoperator RhsOperator = SMP_NULL_OPERATOR;

				if (CurrInst->IsDependentInductionVarArithmetic(Mult1, Mult2, Add1, Add2, RhsOperator)) {
					// See if one of Mult1, Mult2, Add1 or Add2 is a basic induction variable.
					struct DependentInductionVar DIV;
					this->FindDependentInductionVar(LoopIndex, DIV, Add1, Add2, Mult1, Mult2, RhsOperator, CurrInst);
					if (VerboseOutput && (STARS_BADADDR != DIV.DIVDefAddr)) {
						SMP_msg("INFO: Dependent Induction Variable for loop # %d found at %llx\n", LoopIndex, (uint64_t)CurrInst->GetAddr());
						PrintOperand(DIV.DIV.GetOp());
						SMP_msg("\n");
						DumpInductionVar(DIV.IVExpr);
					}
				}
			}
		}

		BlockList.clear();
	}

	return;
} // end of SMPFunction::DetectLoopInductionVars2()

// Is CurrOp a BIV or DIV for loop with LoopIndex? return iterator and position in family if true,
//  where we signal BIV with FamilyIndex == 0 and DIV with FamilyIndex == 1+vectorindex.
//  Do not modify ListIter or FamilyIndex if we are returning false.
bool SMPFunction::IsLoopInductionVar(std::size_t LoopIndex, STARSOpndTypePtr &CurrOp, SMPInstr *UseInst, STARSInductionVarFamilyIter &ListIter, std::size_t &FamilyIndex) {
	STARSInductionVarFamilyList::iterator TempIter;
	int CurrOpSSANum;
	STARSDefUseIter UseIter = UseInst->FindUse(CurrOp);
	if (UseIter == UseInst->GetLastUse())
		return false;
	else if (this->LoopInductionVars.size() <= LoopIndex)
		return false; // haven't done induction var analysis for this loop yet
	else
		CurrOpSSANum = UseIter->GetSSANum();

	for (TempIter = this->LoopInductionVars[LoopIndex].begin(); TempIter != this->LoopInductionVars[LoopIndex].end(); ++TempIter) {
		if (0 > TempIter->BIVIncomingSSANum)  // uninitialized entry
			continue;
		size_t TempIndex = 0;
		if ((CurrOpSSANum == TempIter->BasicInductionVar.InductionVar.GetSSANum()) && IsEqOpIgnoreBitwidth(CurrOp, TempIter->BasicInductionVar.InductionVar.GetOp())) {
			// CurrOp matches the basic induction var.
			ListIter = TempIter;
			FamilyIndex = TempIndex;
			return true;
		}
		else {
			for (TempIndex = 1; TempIndex <= TempIter->DependentInductionVars.size(); ++TempIndex) {
				int DIVSSANum = TempIter->DependentInductionVars[TempIndex - 1].DIV.GetSSANum();
				if ((CurrOpSSANum == DIVSSANum) && IsEqOpIgnoreBitwidth(CurrOp, TempIter->DependentInductionVars[TempIndex - 1].DIV.GetOp())) {
					ListIter = TempIter;
					FamilyIndex = TempIndex;
					return true;
				}
			}
		}
	}
	return false;
} // end of SMPFunction::IsLoopInductionVar()

// Is CurrOp a BIV or DIV for loop with LoopIndex? return iterator and position in family if true,
//  where we signal BIV with FamilyIndex == 0 and DIV with FamilyIndex == 1+vectorindex.
//  Do not modify ListIter or FamilyIndex if we are returning false.
// Same as IsLoopInductionVar() without requiring any SSA nums to match, i.e. is CurrOp an IV.
bool SMPFunction::IsLoopInductionVarForAnySSANum(std::size_t LoopIndex, const STARSOpndTypePtr &CurrOp, STARSInductionVarFamilyList::iterator &ListIter, std::size_t &FamilyIndex) {
	assert(this->LoopInductionVars.size() > LoopIndex);
	STARSInductionVarFamilyList::iterator TempIter;

	for (TempIter = this->LoopInductionVars[LoopIndex].begin(); TempIter != this->LoopInductionVars[LoopIndex].end(); ++TempIter) {
		if (0 > TempIter->BIVIncomingSSANum)  // uninitialized entry
			continue;
		size_t TempIndex = 0;
		if (IsEqOpIgnoreBitwidth(CurrOp, TempIter->BasicInductionVar.InductionVar.GetOp())) {
			// CurrOp matches the basic induction var.
			ListIter = TempIter;
			FamilyIndex = TempIndex;
			return true;
		}
		else {
			for (TempIndex = 1; TempIndex <= TempIter->DependentInductionVars.size(); ++TempIndex) {
				if (IsEqOpIgnoreBitwidth(CurrOp, TempIter->DependentInductionVars[TempIndex - 1].DIV.GetOp())) {
					ListIter = TempIter;
					FamilyIndex = TempIndex;
					return true;
				}
			}
		}
	}
	return false;
} // end of SMPFunction::IsLoopInductionVarForAnySSANum()

// For CurrOp in loop nest including UseInst, return iterator and position in family if true 
bool SMPFunction::IsLoopNestInductionVar(const STARSOpndTypePtr &CurrOp, SMPInstr *UseInst, STARSInductionVarFamilyIter &ListIter, std::size_t &FamilyIndex, int &LoopIndex) {
	bool success = false;
	int UseBlockNum = UseInst->GetBlock()->GetNumber();
	int InnerLoopIndex = this->GetInnermostLoopNum(UseBlockNum);
	int OuterLoopIndex = this->GetOutermostLoopNum(UseBlockNum);
	if ((0 <= InnerLoopIndex) && (0 <= OuterLoopIndex)) { // if block is in a loop
		if (InnerLoopIndex > OuterLoopIndex) { // swap directions
			for (int TempLoopIndex = InnerLoopIndex; !success && (TempLoopIndex >= OuterLoopIndex); --TempLoopIndex) {
				success = this->IsLoopInductionVarForAnySSANum((size_t)TempLoopIndex, CurrOp, ListIter, FamilyIndex);
				if (success)
					LoopIndex = TempLoopIndex;
			}
		}
		else {
			for (int TempLoopIndex = InnerLoopIndex; !success && (TempLoopIndex <= OuterLoopIndex); ++TempLoopIndex) {
				success = this->IsLoopInductionVarForAnySSANum((size_t)TempLoopIndex, CurrOp, ListIter, FamilyIndex);
				if (success)
					LoopIndex = TempLoopIndex;
			}
		}
	}
	return success;
} // end of SMPFunction::IsLoopNestInductionVar()

// pass in results of DefInst::IsDependentInductionVarArithmetic()
// Fill in DIV if DefInst creates a dependent induction var, else set DIV.DIVDefAddr to STARS_BADADDR to signal failure.
void SMPFunction::FindDependentInductionVar(std::size_t LoopIndex, struct DependentInductionVar &DIV, STARSOpndTypePtr Add1, STARSOpndTypePtr Add2, STARSOpndTypePtr Mult1, STARSOpndTypePtr Mult2, SMPoperator RhsOperator, SMPInstr *DefInst) {
	// Convert operands to constant operands if possible.
	bool NonConstAdd1 = true;
	bool NonConstAdd2 = true;
	bool NonConstMult1 = true;
	bool NonConstMult2 = true;
	DIV.DIVInitExpr = nullptr;
	DIV.DIVLimitExpr = nullptr;
	STARSOpndTypePtr IVOp = nullptr;
	size_t IVCounter = 0;
	STARSInductionVarFamilyIter IVFamilyIter;
	size_t FamilyIndex = 0;
	size_t OperandCount = 0;
	size_t LoopInvariantOperandCount = 0;
	bool UseFP = this->UsesFramePointer();
	STARS_ea_t DefAddr = DefInst->GetAddr();

	// We cannot handle operands that are not registers, immediates, or stack operands.
	if ((nullptr != Add1) && (!Add1->IsImmedOp()) && (!MDIsDataFlowOpnd(Add1, UseFP))) {
		DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
		return;
	}
	else if ((nullptr != Add2) && (!Add2->IsImmedOp()) && (!MDIsDataFlowOpnd(Add2, UseFP))) {
		DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
		return;
	}
	else if ((nullptr != Mult1) && (!Mult1->IsImmedOp()) && (!MDIsDataFlowOpnd(Mult1, UseFP))) {
		DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
		return;
	}
	else if ((nullptr != Mult2) && (!Mult2->IsImmedOp()) && (!MDIsDataFlowOpnd(Mult2, UseFP))) {
		DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
		return;
	}

	// Screen out induction variables already found by seeing if the DefInst address matches
	//  the InsideDefAddr of an existing induction variable. We don't want to say that a BIV
	//  is also a DIV that is a function of itself.
	for (size_t index = 0; index < this->LoopInductionVars.size(); ++index) {
		if (!this->LoopInductionVars[index].empty()) {
			STARSInductionVarFamilyList::const_iterator IVIter;
			for (IVIter = this->LoopInductionVars[index].cbegin(); IVIter != this->LoopInductionVars[index].cend(); ++IVIter) {
				if (IVIter->BIVIncomingSSANum >= 0) {
					vector<STARS_ea_t>::const_iterator AddrIter = std::find(IVIter->BIVInsideLoopDefAddrs.cbegin(), IVIter->BIVInsideLoopDefAddrs.cend(), DefAddr);
					bool FoundIVDefAddr = (AddrIter != IVIter->BIVInsideLoopDefAddrs.cend());
					if (FoundIVDefAddr) {
						DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
						return;
					}
				}
			}
		}
	}

	if (nullptr != Add1) {
		++OperandCount;
		if (Add1->IsImmedOp() || this->ReplaceLoopRegWithConst(LoopIndex, Add1, DefInst)) {
			NonConstAdd1 = false;
			++LoopInvariantOperandCount;
		}
		else {
			// Determine if Add1 is an induction variable or a loop-invariant DEF.
			if (this->IsLoopInductionVar(LoopIndex, Add1, DefInst, IVFamilyIter, FamilyIndex)) {
				IVOp = Add1;
				++IVCounter;
			}
			else if (this->IsUseLoopInvariantDEF(LoopIndex, Add1, DefInst)) {
				++LoopInvariantOperandCount;
			}
		}
	}
	if (nullptr != Add2) {
		++OperandCount;
		if (Add2->IsImmedOp() || this->ReplaceLoopRegWithConst(LoopIndex, Add2, DefInst)) {
			NonConstAdd2 = false;
			++LoopInvariantOperandCount;
		}
		else {
			// Determine if Add2 is an induction variable or a loop-invariant DEF.
			if (this->IsLoopInductionVar(LoopIndex, Add2, DefInst, IVFamilyIter, FamilyIndex)) {
				IVOp = Add2;
				++IVCounter;
			}
			else if (this->IsUseLoopInvariantDEF(LoopIndex, Add2, DefInst)) {
				++LoopInvariantOperandCount;
			}
		}
	}
	if (nullptr != Mult1) {
		++OperandCount;
		if (Mult1->IsImmedOp() || this->ReplaceLoopRegWithConst(LoopIndex, Mult1, DefInst)) {
			NonConstMult1 = false;
			++LoopInvariantOperandCount;
		}
		else {
			// Determine if Mult1 is an induction variable or a loop-invariant DEF.
			if (this->IsLoopInductionVar(LoopIndex, Mult1, DefInst, IVFamilyIter, FamilyIndex)) {
				IVOp = Mult1;
				++IVCounter;
			}
			else if (this->IsUseLoopInvariantDEF(LoopIndex, Mult1, DefInst)) {
				++LoopInvariantOperandCount;
			}
		}
	}
	if (nullptr != Mult2) {
		++OperandCount;
		if (Mult2->IsImmedOp() || this->ReplaceLoopRegWithConst(LoopIndex, Mult2, DefInst)) {
			NonConstMult2 = false;
			++LoopInvariantOperandCount;
		}
		else {
			// Determine if Mult2 is an induction variable or a loop-invariant DEF.
			if (this->IsLoopInductionVar(LoopIndex, Mult2, DefInst, IVFamilyIter, FamilyIndex)) {
				IVOp = Mult2;
				++IVCounter;
			}
			else if (this->IsUseLoopInvariantDEF(LoopIndex, Mult2, DefInst)) {
				++LoopInvariantOperandCount;
			}
		}
	}

	if (1 == IVCounter) { // must find exactly one induction var among the operands.
		if (!(NonConstAdd1 || NonConstAdd2)) { // Two constants; fold them by adding
			STARS_uval_t SumValue = Add1->GetImmedValue() + Add2->GetImmedValue();
			Add1 = DefInst->MakeImmediateOpnd(SumValue);
			Add2 = nullptr;
			--OperandCount;
			--LoopInvariantOperandCount;
			NonConstAdd2 = true;
		}
		if (!(NonConstMult1 || NonConstMult2)) { // Two constants; fold them by multiplying
			STARS_uval_t ProductValue = Mult1->GetImmedValue() * Mult2->GetImmedValue();
			Mult1 = DefInst->MakeImmediateOpnd(ProductValue);
			Mult2 = nullptr;
			--OperandCount;
			--LoopInvariantOperandCount;
			NonConstMult2 = true;
		}
		if ((OperandCount > 3) || (OperandCount != (IVCounter + LoopInvariantOperandCount))) {
			// Not the right operands for DIV linear function a * x + b.
			DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
		}
		else { // construct DIV and add it to the right family.
			if ((IVOp != Mult1) && (IVOp != Mult2)) {
				// Problem. We need a * IV + b and we have something like a * b + IV.
				DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
			}
			else {
				if (IVOp == Mult1) {
					STARSOpndTypePtr SearchOp = CloneIfNecessary(Mult1, UseFP);
					CanonicalizeOpnd(SearchOp);
					STARSDefUseIter IVUseIter = DefInst->FindUse(SearchOp);
					assert(IVUseIter != DefInst->GetLastUse());
					DIV.IVExpr.InductionVar = (*IVUseIter);
					// Make Mult2 into a DefOrUse
					if (MDIsDataFlowOpnd(Mult2, UseFP)) {
						STARSOpndTypePtr SearchOp2 = CloneIfNecessary(Mult2, UseFP);
						CanonicalizeOpnd(SearchOp2);
						STARSDefUseIter Mult2UseIter = DefInst->FindUse(SearchOp2);
						assert(Mult2UseIter != DefInst->GetLastUse());
						DIV.IVExpr.Multiplier = (*Mult2UseIter);
					}
					else {
						assert(Mult2->IsImmedOp());
						DefOrUse Mult2Use(Mult2);
						DIV.IVExpr.Multiplier = Mult2Use;
					}
				}
				else {
					assert(IVOp == Mult2);
					STARSOpndTypePtr SearchOp = CloneIfNecessary(Mult2, UseFP);
					CanonicalizeOpnd(SearchOp);
					STARSDefUseIter IVUseIter = DefInst->FindUse(SearchOp);
					assert(IVUseIter != DefInst->GetLastUse());
					DIV.IVExpr.InductionVar = (*IVUseIter);
					// Make Mult1 into a DefOrUse
					if (MDIsDataFlowOpnd(Mult1, UseFP)) {
						STARSOpndTypePtr SearchOp1 = CloneIfNecessary(Mult1, UseFP);
						CanonicalizeOpnd(SearchOp1);
						STARSDefUseIter Mult1UseIter = DefInst->FindUse(SearchOp1);
						assert(Mult1UseIter != DefInst->GetLastUse());
						DIV.IVExpr.Multiplier = (*Mult1UseIter);
					}
					else {
						assert(Mult1->IsImmedOp());
						DefOrUse Mult1Use(Mult1);
						DIV.IVExpr.Multiplier = Mult1Use;
					}
				}
				if ((nullptr == Add1) && (nullptr != Add2)) {
					Add1 = Add2;
					Add2 = nullptr;
				}
				assert(nullptr == Add2); // should have been folded into Add1
				// Make Add1 into a DefOrUse
				if (nullptr == Add1) {
					// Make a zero addend
					STARSOpndTypePtr ZeroAddend = DefInst->MakeImmediateOpnd(0);
					DefOrUse ZeroUse(ZeroAddend);
					DIV.IVExpr.Addend = ZeroUse;
					DIV.IVExpr.SubtractAddend = false;
				}
				else if (MDIsDataFlowOpnd(Add1, UseFP)) {
					STARSOpndTypePtr SearchOp1 = CloneIfNecessary(Add1, UseFP);
					CanonicalizeOpnd(SearchOp1);
					STARSDefUseIter Add1UseIter = DefInst->FindUse(SearchOp1);
					assert(Add1UseIter != DefInst->GetLastUse());
					DIV.IVExpr.Addend = (*Add1UseIter);
					DIV.IVExpr.SubtractAddend = false;
				}
				else {
					assert(Add1->IsImmedOp());
					STARS_sval_t SignedImmedValue = (STARS_sval_t) Add1->GetImmedValue();
					if (SignedImmedValue >= 0) {
						DefOrUse Add1Use(Add1);
						DIV.IVExpr.Addend = Add1Use;
						DIV.IVExpr.SubtractAddend = false;
					}
					else { // Make negative addend a subtrahend
						SignedImmedValue = (-SignedImmedValue);
						STARSOpndTypePtr SubtrahendOp = DefInst->MakeImmediateOpnd((STARS_uval_t) SignedImmedValue);
						DefOrUse SubtrahendUse(SubtrahendOp);
						DIV.IVExpr.Addend = SubtrahendUse;
						DIV.IVExpr.SubtractAddend = true;
					}
				}
				STARSDefUseIter DIVDefIter = DefInst->GetFirstNonFlagsDef();
				if (DIVDefIter != DefInst->GetLastDef()) {
					DIV.DIVDefAddr = DefAddr;
					DIV.DIV = (*DIVDefIter);
					// DIV is complete, so store it in the IV family.
					IVFamilyIter->DependentInductionVars.push_back(DIV);
				}
				else {
					// Must have inst that DEFs only the flags, so we cannot determine DIV.
					DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
				}
			}
		}
	}
	else {
		DIV.DIVDefAddr = STARS_BADADDR; // signal failure to find DIV
	}

	return;
} // end of SMPFunction::FindDependentInductionVar()

// If RegOp USE in UseInst is an SCCP constant, create an immediate operand for it.
bool SMPFunction::ReplaceLoopRegWithConst(std::size_t LoopIndex, STARSOpndTypePtr &RegOp, SMPInstr *UseInst) {
	bool Replaced = false;
	if (RegOp->IsRegOp()) {
		STARSOpndTypePtr SearchOp = CloneIfSubwordReg(RegOp);
		CanonicalizeOpnd(SearchOp);
		STARSDefUseIter UseIter = UseInst->FindUse(SearchOp);
		assert(UseIter != UseInst->GetLastUse());
		int UseSSANum = UseIter->GetSSANum();
		SMPBasicBlock *CurrBlock = UseInst->GetBlock();
		int UseHashIndex = HashGlobalNameAndSSA(SearchOp, UseSSANum);
		bool LocalName = CurrBlock->IsLocalName(SearchOp);
		STARSSCCPMapIter ConstIter;
		bool ConstFound = this->FindSCCPConstIter(CurrBlock, UseHashIndex, LocalName, ConstIter);
		if (ConstFound) {
			STARS_uval_t ImmedValue = ConstIter->second.ConstValue;
			RegOp = UseInst->MakeImmediateOpnd(ImmedValue);
			Replaced = true;
		}
	}

	return Replaced;
} // end of SMPFunction::ReplaceRegWithConst()

// Add edge expr to EdgeExpressions and EdgeExprMap.
void SMPFunction::AddEdgeExpression(const int SourceBlockNum, const int TargetBlockNum, const struct LoopComparison GuardExpr) {
	assert(0 <= SourceBlockNum);
	assert(0 <= TargetBlockNum);
	uint32_t EdgeHashIndex = (((uint32_t) SourceBlockNum << 16) | ((uint32_t) TargetBlockNum));
	pair<uint32_t, size_t> InsertPair(EdgeHashIndex, this->EdgeExpressions.size());
	this->EdgeExpressions.push_back(GuardExpr);
	pair<map<uint32_t, size_t>::iterator, bool> InsertResult = this->EdgeExprMap.insert(InsertPair);
	assert(InsertResult.second); // edges should be unique pairs of block numbers
	return;
} // end of SMPFunction::AddEdgeExpression()

// Compute expression that causes each edge to be taken
bool SMPFunction::ComputeEdgeExpressions(void) {
	// We compute the expression that causes an edge to be taken.
	// An edge is represented by a uint32_t object whose upper 16 bits
	//  are the block number of the source edge, and lower 16 bits are
	//  the block number of the destination edge. This uint32_t is used
	//  as the index into the hash table EdgeExprMap, which returns a 
	//  subscript into the EdgeExpressions vector. So, as we determine
	//  the expression that gates each edge, we just push it onto the vector
	//  and add the hash into the hash table.
	//
	// Algorithm:
	// for each block
	//   for each conditional outgoing edge that is not a loop back edge
	//      Extract the edge expression using Block->FindBranchCompareOrTest() and Inst->GetCompareOrTestExprAndReduceSelfTest()
	//      Add the expression to the vector & hash table for the branch taken edge
	//      Add the negation of the expression to the vector & hash table for the fall-through edge
	//   end for
	//   for each computed switch jump edge in that special case
	//      Extract the switch index expression (e.g. "RAX == 4") as done in translation to SPARK code
	//      Add the expression to the vector and hash table for the edge connecting indirect branch to destination
	//   end for
	// end for

	bool success = true;
	for (size_t BlockNum = 0; BlockNum < this->RPOBlocks.size(); ++BlockNum) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockNum];
		if (CurrBlock->HasConditionalBranch()) {
			int BranchBlockNum = CurrBlock->GetNumber();
			struct LoopComparison BranchExpr;

			list<SMPBasicBlock *>::const_iterator BranchTakenIter = CurrBlock->GetCondNonFallThroughSucc();
			bool ValidTakenEdge = (BranchTakenIter != CurrBlock->GetLastConstSucc());
			list<SMPBasicBlock *>::const_iterator BranchNotTakenIter = CurrBlock->GetFallThroughSucc();
			bool ValidFallThroughEdge = (BranchNotTakenIter != CurrBlock->GetLastConstSucc());
			int TakenBlockNumber = ValidTakenEdge ? ((*BranchTakenIter)->GetNumber()) : SMP_BLOCKNUM_UNINIT;
			int NotTakenBlockNumber = ValidFallThroughEdge ? ((*BranchNotTakenIter)->GetNumber()) : SMP_BLOCKNUM_UNINIT;
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			bool TakenBackEdge = ValidTakenEdge && this->DoesBlockDominateBlock(TakenBlockNumber, BranchBlockNum);
			this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
			bool FallThroughBackEdge = ValidFallThroughEdge && this->DoesBlockDominateBlock(NotTakenBlockNumber, BranchBlockNum);

			// Some idiotic code has a block ending with jz $+2, i.e. jump-on-zero to the very next
			//  instruction, else fall through to the same instruction. Perhaps this is the result of an
			//  optimization that was not followed by proper clean-up of the code. The result is that 
			//  there is no need to compute an expression for the edge, as the edge is always taken.
			bool LoneSuccessorCase = (TakenBlockNumber == NotTakenBlockNumber);
			if (LoneSuccessorCase)
				continue;

			// Separate back edges from all others.
			//  An edge from block A to block B is a back edge if B dominates A.
			//  If the branch-taken edge is a back edge, we invert the branch expression
			//  and record the fall-through edge. If the fall-through edge is a back edge,
			//  we have a compiler-optimized code layout that requires us to record only
			//  the branch-taken edge.

			SMPoperator BranchOperator = SMP_NULL_OPERATOR;
			SMPoperator SetBitOperator = SMP_NULL_OPERATOR;
			STARS_ea_t DecrementAddr = STARS_BADADDR;
			SMPInstr *CompareOrTestInst = CurrBlock->FindBranchCompareOrTest(BranchOperator, SetBitOperator, DecrementAddr, false, SMP_SSA_UNINIT, CurrBlock->GetRevInstCBegin());
			if (nullptr != CompareOrTestInst) {
				CompareOrTestInst->GetCompareOrTestExprAndReduceSelfTest(BranchExpr);
				if (SMP_NULL_OPERATOR == BranchExpr.CompareOperator) {
					// GetCompareOrTestExprAndReduceSelfTest() failed.
					success = false;
					break;
				}
				BranchExpr.CompareOperator = BranchOperator;
				BranchExpr.CompareAddr = CompareOrTestInst->GetAddr();
			}
			else if (STARS_BADADDR != DecrementAddr) {
				// Found a decrement rather than a compare or test.
				SMPInstr *DecrementInst = this->GetInstFromAddr(DecrementAddr);
				assert(nullptr != DecrementInst);
				BranchExpr.CompareAddr = DecrementAddr;
				BranchExpr.CompareOperator = BranchOperator;
				if (IsRelationalOperator(BranchOperator)) {
					// Branch will depend on how decremented value compares to zero. Transform as above.
					STARSDefUseIter DecrementDEF = DecrementInst->GetFirstNonFlagsDef();
					BranchExpr.Operand1 = (*DecrementDEF);
					STARSOpndTypePtr ZeroOp = DecrementInst->MakeImmediateOpnd(0);
					DefOrUse ZeroUse(ZeroOp);
					BranchExpr.Operand2 = ZeroUse;
				}
				else {
					SMP_msg("ERROR: COND_BRANCH: Decrement with BranchOperator %d in test block %d DecrementAddr %llx ",
						BranchOperator, BranchBlockNum, (uint64_t)DecrementAddr);
					this->DumpFuncNameAndAddr();
					success = false;
					break;
				}
			}
			else {
				SMP_msg("ERROR: COND_BRANCH: Could not find compare or test in test block %d ", BranchBlockNum);
				this->DumpFuncNameAndAddr();
				success = false;
				break;
			}

			// At this point, BranchExpr should be valid. Unless a loop back edge is
			//  detected, we need to record the BranchExpr for the branch-taken edge
			//  and the inverse of the BranchExpr for the fall-through edge.
			// We have to deal with special cases in which we branch outside the current function
			//  (might be incorrect grouping of instructions into functions) or even fall-through
			//  to another function (also might be incorrect disassembly).
			int BranchLoopNum = this->GetInnermostLoopNum(BranchBlockNum);
			if (ValidTakenEdge && (!TakenBackEdge)) {
				// Add guard expression for the taken edge.
				if (BranchLoopNum >= 0) {
					BranchExpr.ExitsLoop = (!this->IsBlockInLoop(TakenBlockNumber, BranchLoopNum));
				}
				else {
					BranchExpr.ExitsLoop = false;
				}
				this->AddEdgeExpression(BranchBlockNum, TakenBlockNumber, BranchExpr);
			}
			if (ValidFallThroughEdge && (!FallThroughBackEdge)) {
				if (BranchLoopNum >= 0) {
					BranchExpr.ExitsLoop = (!this->IsBlockInLoop(NotTakenBlockNumber, BranchLoopNum));
				}
				else {
					BranchExpr.ExitsLoop = false;
				}
				// Add inverted guard expression for the fall-through edge.
				struct LoopComparison InvertedBranchExpr = BranchExpr;
				InvertedBranchExpr.CompareOperator = InvertRelationalOperator(BranchExpr.CompareOperator);
				this->AddEdgeExpression(BranchBlockNum, NotTakenBlockNumber, InvertedBranchExpr);
			}
		}
		else if (CurrBlock->HasIndirectJump()) {
			// stub out for now
			SMP_msg("ERROR: INDIR_JUMP: Not ready to analyze ");
			this->DumpFuncNameAndAddr();
			success = false;
			break;
		}
	}

	return success;
} // end of SMPFunction::ComputeEdgeExpressions()

// Fill InArgConstValues with SCCP values from call sites.
void SMPFunction::ComputeInArgConstValues(void) {
	bool DebugFlag = (0x402760 == this->GetFirstFuncAddr());
	set<STARS_uval_t> DummySet;
	for (size_t i = 0; i < this->GetIncomingArgCount(); ++i) {
		this->InArgConstValues.push_back(DummySet);
	}

	for (const STARS_ea_t CallSiteAddr : this->AllCallSites) {
		int CallingBlockNum;
		SMPFunction *CallingFunc = this->GetProg()->FindFunctionAndBlockFromInstID(CallSiteAddr, CallingBlockNum);
		if (nullptr != CallingFunc) {
			assert(SMP_BLOCKNUM_UNINIT != CallingBlockNum);
			SMPInstr *CallInst = CallingFunc->GetInstFromAddr(CallSiteAddr);
			assert(nullptr != CallInst);

			// Iterate through incoming args.
			// Go through the DEFs in the SSA marker inst and see which of them are incoming args.
			list<SMPInstr *>::iterator MarkerInstIter = this->Instrs.begin();
			SMPInstr *MarkerInst = (*MarkerInstIter);
			unsigned short ArgCount = 0;
			if (MarkerInst->IsMarkerInst()) {
				std::size_t StandardByteSize = (MD_NORMAL_MACHINE_BITWIDTH / 8);
				STARSBitSet ArgsProcessed;
				ArgsProcessed.AllocateBits(STARS_MAX_ARGS_PER_FUNC);
				for (set<DefOrUse, LessDefUse>::iterator DefIter = MarkerInst->GetFirstDef(); DefIter != MarkerInst->GetLastDef(); ++DefIter) {
					STARSOpndTypePtr DefOp = DefIter->GetOp();
					std::size_t ArgIndex = 0;
					if (global_STARS_program->GetSTARS_ISA_Bitwidth() == 64) {
						if (DefOp->IsRegOp()) {
							STARS_regnum_t RegNum = DefOp->GetReg();
							for (list<STARS_regnum_t>::iterator ArgRegIter = global_STARS_program->GetFirstArgumentReg(); ArgRegIter != global_STARS_program->GetLastArgumentReg(); ++ArgRegIter) {
								if (RegNum == (*ArgRegIter)) { // found an incoming argument
									++ArgCount;
									STARS_SCCP_Const_Struct ConstStruct;
									CallInst->SCCPFetchConstUseValue(DefOp, ConstStruct);
									if (STARS_CONST_HAS_VALUE == ConstStruct.ConstType) {
										if (ArgIndex < ((size_t) this->InArgCount)) {
											(void) this->InArgConstValues[ArgIndex].insert(ConstStruct.ConstValue);
											if (DebugFlag) {
												SMP_msg("INFO: InArg value of %llx for ArgIndex %zu in func %s\n",
													(uint64_t) ConstStruct.ConstValue, ArgIndex, this->GetFuncName());
											}
										}
									}
								}
								++ArgIndex;
							}
						}
					}
					else { // 32 bits
						// Current model (gcc) is writing outargs to stack. 
						if (MDIsDirectStackAccessOpnd(DefOp, this->UsesFramePointer())) {
							// Ensure that we are dealing with an incoming arg and not a USE-before-DEF local var.
							int SignedAddr = (int)DefOp->GetAddr();
							if (0 < SignedAddr) {
								ArgIndex = (DefOp->GetAddr() / StandardByteSize) - 1; // -1 to skip over return address
								if (ArgIndex < STARS_MAX_ARGS_PER_FUNC) {
									if (!ArgsProcessed.GetBit(ArgIndex)) {
										++ArgCount;
										ArgsProcessed.SetBit(ArgIndex);
										STARS_SCCP_Const_Struct ConstStruct;
										CallInst->SCCPFetchConstUseValue(DefOp, ConstStruct);
										if (STARS_CONST_HAS_VALUE == ConstStruct.ConstType) {
											if (ArgIndex < ((size_t) this->InArgCount)) {
												(void) this->InArgConstValues[ArgIndex].insert(ConstStruct.ConstValue);
											}
										}
									}
									else { // Not the first time to see arg; could be processing an arg one byte at a time, for example.
										// Do not increment ArgCount for duplicate processing.
										;
									}
								}
								else {
									SMP_msg("ERROR: Incoming ArgIndex of %u in SMPFunction::ComputeInArgConstValues ", ArgIndex);
									this->DumpFuncNameAndAddr();
								}
							}
						}
					}
				} // end for all DEFs in marker inst
			} // end if marker inst
		}
	} // end for each CallSiteAddr

	return;
} // end of SMPFunction::ComputeInArgConstValues()

// Create loop limit expr: BIVOpnd relationaloperator Opnd2
STARSExpression* SMPFunction::CreateLimitExpr(const std::size_t &LoopIndex, const struct InductionVarFamily &IVFamily, const struct LoopComparison &LoopCompExpr) {
	STARSOpndTypePtr BIVOp = IVFamily.BasicInductionVar.InductionVar.GetOp();
	SMPoperator RelationalOperator = LoopCompExpr.CompareOperator;
	bool BIVIsLeftOp = IsEqOpIgnoreBitwidth(LoopCompExpr.Operand1.GetOp(), BIVOp);
	bool BIVIsRightOp = false;
	if (!BIVIsLeftOp)
		BIVIsRightOp = IsEqOpIgnoreBitwidth(LoopCompExpr.Operand2.GetOp(), BIVOp);
	if ((SMP_NULL_OPERATOR == RelationalOperator) || (STARS_BADADDR == LoopCompExpr.CompareAddr)) {
		return nullptr;
	}

	SMPInstr *CompareInst = this->GetInstFromAddr(LoopCompExpr.CompareAddr);
	assert(nullptr != CompareInst);
	STARSExpression *LimitExpr = new STARSExpression();
	LimitExpr->SetParentFunc(this);
	LimitExpr->SetParentInst(CompareInst);
	LimitExpr->SetOriginalParentInst(CompareInst);

	// Assume that BIV will be on the left side, swap sides later if not.
	LimitExpr->SetOperator(RelationalOperator);
	LimitExpr->SetLeftOperand(LoopCompExpr.Operand1.GetOp());
	LimitExpr->SetLeftUseAddr(LoopCompExpr.CompareAddr);
	LimitExpr->SetLeftSSANum(LoopCompExpr.Operand1.GetSSANum());
	LimitExpr->SetRightOperand(LoopCompExpr.Operand2.GetOp());
	LimitExpr->SetRightUseAddr(LoopCompExpr.CompareAddr);
	LimitExpr->SetRightSSANum(LoopCompExpr.Operand2.GetSSANum());

	// Deal with the possibility that the comparison instruction involves a copy of the BIV. If so, expanding
	//  the comparison expression will not produce two sub-trees; it will just replace operands on the path to the BIV.
	bool LhsBIV = false;
	bool RhsBIV = false;
	if (!(BIVIsLeftOp || BIVIsRightOp)) {
		// See if expanding the loop comparison expr, while stopping on encountering the BIV, will succeed.
		bool StoppedOnIV = false;
		bool changed = false;
		set<int> DummyLoopRegHashes;
		set<STARS_ea_t> StackPtrCopySet;
		int DepthCounter = 0;
		if (LimitExpr->ExpandExpr(LimitExpr->GetParentInst()->GetAddr(), LoopIndex, false, true, false, false, false, DummyLoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
			BIVIsLeftOp = (nullptr != LimitExpr->GetConstLeftOperand()) && (!LimitExpr->HasLeftSubTree()) && IsEqOpIgnoreBitwidth(LimitExpr->GetConstLeftOperand(), BIVOp);
			if (!BIVIsLeftOp) {
				BIVIsRightOp = (nullptr != LimitExpr->GetConstRightOperand()) && (!LimitExpr->HasRightSubTree()) && IsEqOpIgnoreBitwidth(LimitExpr->GetConstRightOperand(), BIVOp);
			}
		}
		else {
			SMP_msg("ERROR: ExpandExpr() failure in CreateLimitExpr() attempt to find BIV for loop %zu ", LoopIndex);
			this->DumpFuncNameAndAddr();
		}
	}

	if (!(BIVIsLeftOp || BIVIsRightOp)) {
		delete LimitExpr;
		return nullptr;
	}

	if (BIVIsRightOp) {
		// Get the Phi DEF addr for the BIV.
		STARS_ea_t UseDefAddr = this->GetGlobalDefAddr(LimitExpr->GetConstRightOperand(), LimitExpr->GetRightSSANum());
		LimitExpr->SetRightPreLoopDefAddr(UseDefAddr);

		// Swap LoopCompExpr operands and invert the relational operator
		//  to normalize LimitExpr to always have BIVOp on the left.
		LimitExpr->SetOperator(InvertRelationalOperatorForExprSwap(RelationalOperator));
		LimitExpr->SwapSides();
	}
	else {
		assert(BIVIsLeftOp);
		// Get the Phi DEF addr for the BIV.
		STARS_ea_t UseDefAddr = this->GetGlobalDefAddr(LimitExpr->GetConstLeftOperand(), LimitExpr->GetLeftSSANum());
		LimitExpr->SetLeftPreLoopDefAddr(UseDefAddr);
	}

	return LimitExpr;
} // end of SMPFunction::CreateLimitExpr()

// Create loop limit expr: BIVOpnd_InitExpr + (stride * IterationCountExpr)
STARSExpression *SMPFunction::CreateSecondaryBIVLimitExpr(const std::size_t &LoopIndex, const struct InductionVarFamily &IVFamily) {
	if (nullptr == this->LoopIterationsCountExprs[LoopIndex])
		return nullptr;

	SMPInstr *FirstInsideDefInst = this->GetInstFromAddr(IVFamily.BIVInsideLoopDefAddrs[0]);
	if (nullptr == FirstInsideDefInst) {
		return nullptr;
	}
	// Limit to multiplier of 1 initially. Generalize and search for SCCP constants values later.
	bool ImmedMultiplier = IVFamily.BasicInductionVar.Multiplier.GetOp()->IsImmedOp();
	STARS_uval_t MultiplierValue = ImmedMultiplier ? IVFamily.BasicInductionVar.Multiplier.GetOp()->GetImmedValue() : 0;
	if (!ImmedMultiplier || (1 != MultiplierValue)) {
		SMP_msg("ERROR: BIV has multiplier of value %llu in loop %zu ", (uint64_t) MultiplierValue);
		this->DumpFuncNameAndAddr();
		return nullptr;
	}
	STARSOpndTypePtr StrideOp = IVFamily.BasicInductionVar.Addend.GetOp()->clone();
	if (!StrideOp->IsImmedOp()) {
		SMP_msg("ERROR: BIV does not have constant addend.\n");
		return nullptr;
	}

	// If addend (stride) is subtracted, make it a negative immediate value.
	if (IVFamily.BasicInductionVar.SubtractAddend) {
		STARS_uval_t StrideValue = StrideOp->GetImmedValue();
		STARS_sval_t SignedStrideValue = (STARS_sval_t) StrideValue;
		SignedStrideValue = (0 - SignedStrideValue);
		StrideOp = FirstInsideDefInst->MakeImmediateOpnd((STARS_uval_t) SignedStrideValue);
	}

	STARSExpression *LimitExpr = new STARSExpression();
	LimitExpr->SetOriginalParentInst(FirstInsideDefInst);
	LimitExpr->SetParentInst(FirstInsideDefInst);
	LimitExpr->SetParentFunc(this);
	LimitExpr->SetOperator(SMP_ADD);
	LimitExpr->SetLeftTree(IVFamily.BIVInitExpr->Clone());
	STARSExpression *LimitRHS = new STARSExpression();
	LimitRHS->SetOriginalParentInst(FirstInsideDefInst);
	LimitRHS->SetParentInst(FirstInsideDefInst);
	LimitRHS->SetParentFunc(this);
	LimitRHS->SetOperator(SMP_U_MULTIPLY);
	LimitRHS->SetLeftOperand(StrideOp);
	LimitRHS->SetLeftUseAddr(FirstInsideDefInst->GetAddr());
	assert(nullptr != this->LoopIterationsCountExprs[LoopIndex]);
	LimitRHS->SetRightTree(this->LoopIterationsCountExprs[LoopIndex]->Clone());
	LimitExpr->SetRightTree(LimitRHS);

	return LimitExpr;
} // end of SMPFunction::CreateSecondaryBIVLimitExpr()

// Create DIV init expr as linear function of its BIV init expr
STARSExpression *SMPFunction::CreateDIVInitExpr(const std::size_t &LoopIndex, const struct InductionVarFamily &IVFamily, const std::size_t DIVIndex) {
	struct DependentInductionVar DIV = IVFamily.DependentInductionVars[DIVIndex];
	STARSOpndTypePtr MulOp = DIV.IVExpr.Multiplier.GetOp();
	bool ConstMul = MulOp->IsImmedOp();
	STARSExpression *DIVExpr = new STARSExpression();
	bool UnityMul = ConstMul && (1 == MulOp->GetImmedValue());
	DIVExpr->SetParentFunc(this);
	assert(DIV.DIVDefAddr < STARS_PSEUDO_ID_MIN);
	SMPInstr *DIVDefInst = this->GetInstFromAddr(DIV.DIVDefAddr);
	assert(nullptr != DIVDefInst);
	DIVExpr->SetOriginalParentInst(DIVDefInst);
	DIVExpr->SetParentInst(DIVDefInst);
	STARSDefUseIter IVUseIter = DIVDefInst->FindUse(DIV.IVExpr.InductionVar.GetOp());
	assert(IVUseIter != DIVDefInst->GetLastUse());
	int IVSSANum = IVUseIter->GetSSANum();
	if (DIV.IVExpr.SubtractAddend) {
		DIVExpr->SetOperator(SMP_SUBTRACT);
	}
	else {
		DIVExpr->SetOperator(SMP_ADD);
	}
	if (UnityMul) { // Typical, simple case.
		// DIV = IV +/- Addend.
		DIVExpr->SetLeftOperand(DIV.IVExpr.InductionVar.GetOp());
		DIVExpr->SetLeftUseAddr(DIV.DIVDefAddr);
		DIVExpr->SetLeftSSANum(IVSSANum);
		DIVExpr->SetRightOperand(DIV.IVExpr.Addend.GetOp());
		DIVExpr->SetRightUseAddr(DIV.DIVDefAddr);
	}
	else { // MulOp is not 1.
		// DIV = (MulOp * IV) +/- Addend.
		DIVExpr->SetRightOperand(DIV.IVExpr.Addend.GetOp());
		DIVExpr->SetRightUseAddr(DIV.DIVDefAddr);
		STARSExpression *DIVExprLeftTree = new STARSExpression();
		DIVExprLeftTree->SetParentFunc(this);
		DIVExprLeftTree->SetOriginalParentInst(DIVDefInst);
		DIVExprLeftTree->SetParentInst(DIVDefInst);
		DIVExprLeftTree->SetOperator(SMP_U_MULTIPLY);
		DIVExprLeftTree->SetLeftOperand(DIV.IVExpr.InductionVar.GetOp());
		DIVExprLeftTree->SetLeftUseAddr(DIV.DIVDefAddr);
		DIVExprLeftTree->SetLeftSSANum(IVSSANum);
		DIVExprLeftTree->SetRightOperand(MulOp);
		DIVExprLeftTree->SetRightUseAddr(DIV.DIVDefAddr);
		DIVExpr->SetLeftTree(DIVExprLeftTree);
	}

	return DIVExpr;
} // end of SMPFunction::CreateDIVInitExpr()

// Create DIV limit expr as linear function of its BIV limit expr
STARSExpression *SMPFunction::CreateDIVLimitExpr(const std::size_t &LoopIndex, const struct InductionVarFamily &IVFamily, const std::size_t DIVIndex) {
	if (nullptr == IVFamily.BIVLimitExpr)
		return nullptr;

	struct DependentInductionVar DIV = IVFamily.DependentInductionVars[DIVIndex];
	STARSOpndTypePtr MulOp = DIV.IVExpr.Multiplier.GetOp();
	bool ConstMul = MulOp->IsImmedOp();
	STARSExpression *DIVExpr = new STARSExpression();
	bool UnityMul = ConstMul && (1 == MulOp->GetImmedValue());
	DIVExpr->SetParentFunc(this);
	assert(DIV.DIVDefAddr < STARS_PSEUDO_ID_MIN);
	SMPInstr *DIVDefInst = this->GetInstFromAddr(DIV.DIVDefAddr);
	assert(nullptr != DIVDefInst);
	DIVExpr->SetOriginalParentInst(DIVDefInst);
	DIVExpr->SetParentInst(DIVDefInst);
	STARSDefUseIter IVUseIter = DIVDefInst->FindUse(DIV.IVExpr.InductionVar.GetOp());
	assert(IVUseIter != DIVDefInst->GetLastUse());
	int IVSSANum = IVUseIter->GetSSANum();
	if (DIV.IVExpr.SubtractAddend) {
		DIVExpr->SetOperator(SMP_SUBTRACT);
	}
	else {
		DIVExpr->SetOperator(SMP_ADD);
	}
	if (UnityMul) { // Typical, simple case.
		// DIV = IV +/- Addend.
		DIVExpr->SetLeftTree(IVFamily.BIVLimitExpr->Clone());
		DIVExpr->SetRightOperand(DIV.IVExpr.Addend.GetOp());
		DIVExpr->SetRightUseAddr(DIV.DIVDefAddr);
	}
	else { // MulOp is not 1.
		// DIV = (MulOp * IV) +/- Addend.
		DIVExpr->SetRightOperand(DIV.IVExpr.Addend.GetOp());
		DIVExpr->SetRightUseAddr(DIV.DIVDefAddr);
		STARSExpression *DIVExprLeftTree = new STARSExpression();
		DIVExprLeftTree->SetParentFunc(this);
		DIVExprLeftTree->SetOriginalParentInst(DIVDefInst);
		DIVExprLeftTree->SetParentInst(DIVDefInst);
		DIVExprLeftTree->SetLeftTree(IVFamily.BIVLimitExpr->Clone());
		DIVExprLeftTree->SetRightOperand(MulOp);
		DIVExprLeftTree->SetRightUseAddr(DIV.DIVDefAddr);
		DIVExpr->SetLeftTree(DIVExprLeftTree);
	}

	return DIVExpr;
} // end of SMPFunction::CreateDIVLimitExpr()

// Create loop iteration count expr
STARSExpression* SMPFunction::CreateIterationsExpr(std::size_t LoopIndex, const struct InductionVarFamily &IVFamily, bool PositiveIncrement, STARSExpression *InitExpr, STARSExpression *LimitExpr) {
	// For now, we will limit ourselves to BIVars with a loop update of BIV *= 1 + constant.
	bool SimpleBIV = (IVFamily.BasicInductionVar.Multiplier.GetOp()->IsImmedOp()
		&& (IVFamily.BasicInductionVar.Multiplier.GetOp()->GetImmedValue() == 1)
		&& IVFamily.BasicInductionVar.Addend.GetOp()->IsImmedOp());
	if (!SimpleBIV) {
		SMP_msg("ERROR: CreateIterationsExpr() did not find SimpleBIV for loop %u ", LoopIndex);
		this->DumpFuncNameAndAddr();
		return nullptr;
	}

	STARS_uval_t IncDecValue = IVFamily.BasicInductionVar.Addend.GetOp()->GetImmedValue();
	// bool PositiveIncrement = (!IVFamily.BasicInductionVar.SubtractAddend);
	bool BottomTestingLoop = DoesLoopTestAtBottom(this->LoopTypesByLoopNum[LoopIndex]);

	// For loops with a BIV that counts upward by 1:
	//
	// TopTesting   BranchExitsLoop    RelationalOperator     IterationCount
	// ----------   ---------------    ------------------     --------------
	//     Y               Y                   >              (Limit + 1 - Init) / 1
	//     Y               Y                   >=             (Limit - Init) / 1
	//     Y               Y                   <              undefined
	//     Y               Y                   <=             undefined
	//     Y               Y                   ==             (Limit - Init) / 1
	//     Y               Y                   !=             undefined
	//     Y               N                   >              undefined
	//     Y               N                   >=             undefined
	//     Y               N                   <              (Limit - Init) / 1
	//     Y               N                   <=             (Limit + 1 - Init) / 1
	//     Y               N                   ==             undefined
	//     Y               N                   !=             (Limit - Init) / 1
	//     N               Y                   >              (Limit + 1 - Init) / 1
	//     N               Y                   >=             (Limit - Init) / 1
	//     N               Y                   <              undefined
	//     N               Y                   <=             undefined
	//     N               Y                   ==             (Limit - Init) / 1
	//     N               Y                   !=             undefined
	//     N               N                   >              undefined
	//     N               N                   >=             undefined
	//     N               N                   <              (Limit - Init) / 1
	//     N               N                   <=             (Limit + 1 - Init) / 1
	//     N               N                   ==             undefined
	//     N               N                   !=             (Limit - Init) / 1
	//
	// We see that certain COND_BRANCH relational operators add 1 to the iteration count.
	// Certain combinations make no sense, e.g. counting up by 1 from Init to Limit while exiting
	// the loop when BIV < Limit. Some of these cases could be zero-trip loops that we will analyze
	// in greater depth later.
	// NOTE: There is no difference between top-testing and bottom-testing loops in the table.
	//  That is because the only difference in practice is the possibility that a top-testing loop
	//  executes zero iterations, while the bottom-testing version of the loop would execute one
	//  iteration in that case. If we have security properties that depend on identifying zero-trip
	//  loops, we will need a separate analysis later.
	// NOTE: In these tables, we assume that == and != are used properly. There is the possibility
	//  that someone could write code such as: for (i = 0; i != limit; i += 8) and have a limit
	//  that is not a multiple of 8. Again, we will need a special-case analysis if we want to detect
	//  the possibility of such an infinite loop.

	// For loops with a BIV that counts downward by 1:
	//
	// TopTesting   BranchExitsLoop    RelationalOperator     IterationCount
	// ----------   ---------------    ------------------     --------------
	//     Y               Y                   >              undefined
	//     Y               Y                   >=             undefined
	//     Y               Y                   <              (Init + 1 - Limit) / 1
	//     Y               Y                   <=             (Init - Limit) / 1
	//     Y               Y                   ==             (Init - Limit) / 1
	//     Y               Y                   !=             undefined
	//     Y               N                   >              (Init - Limit) / 1
	//     Y               N                   >=             (Init + 1 - Limit) / 1
	//     Y               N                   <              undefined
	//     Y               N                   <=             undefined
	//     Y               N                   ==             undefined
	//     Y               N                   !=             (Init - Limit) / 1
	//     N               Y                   >              undefined
	//     N               Y                   >=             undefined
	//     N               Y                   <              (Init + 1 - Limit) / 1
	//     N               Y                   <=             (Init - Limit) / 1
	//     N               Y                   ==             (Init - Limit) / 1
	//     N               Y                   !=             undefined
	//     N               N                   >              (Init - Limit) / 1
	//     N               N                   >=             (Init + 1 - Limit) / 1
	//     N               N                   <              undefined
	//     N               N                   <=             undefined
	//     N               N                   ==             undefined
	//     N               N                   !=             (Init - Limit) / 1

	// We can generalize the formulae above to allow increment or decrement by values other than 1.
	//  Whenever we add 1 in the numerator in the table, just add the increment or decrement value.
	//  Whenever we divide by 1 in the denominator, divide by the increment or decrement value.

	// Weed out undefined cases.
	bool Undefined = false;
	assert(LoopIndex < this->LoopComparisonExprs.size());
	SMPoperator BranchOperator = this->LoopComparisonExprs[LoopIndex].CompareOperator;
	if (PositiveIncrement) {
		if (this->LoopComparisonExprs[LoopIndex].ExitsLoop) {
			Undefined = ((BranchOperator != SMP_GREATER_THAN) && (BranchOperator != SMP_GREATER_EQUAL)
				&& (BranchOperator != SMP_ABOVE) && (BranchOperator != SMP_ABOVE_EQUAL) && (BranchOperator != SMP_EQUAL));
		}
		else {
			Undefined = ((BranchOperator != SMP_LESS_THAN) && (BranchOperator != SMP_LESS_EQUAL)
				&& (BranchOperator != SMP_BELOW) && (BranchOperator != SMP_BELOW_EQUAL) && (BranchOperator != SMP_NOT_EQUAL));
		}
	}
	else { // Decrement on each iteration
		if (this->LoopComparisonExprs[LoopIndex].ExitsLoop) {
			Undefined = ((BranchOperator != SMP_LESS_THAN) && (BranchOperator != SMP_LESS_EQUAL)
				&& (BranchOperator != SMP_BELOW) && (BranchOperator != SMP_BELOW_EQUAL) && (BranchOperator != SMP_EQUAL));
		}
		else {
			Undefined = ((BranchOperator != SMP_GREATER_THAN) && (BranchOperator != SMP_GREATER_EQUAL)
				&& (BranchOperator != SMP_ABOVE) && (BranchOperator != SMP_ABOVE_EQUAL) && (BranchOperator != SMP_NOT_EQUAL));
		}
	}
	if (Undefined) {
		SMP_msg("ERROR: CreateIterationsExpr() found Undefined terminating relational operator for loop %u operator %s PosIncr: %d BottomTesting: %d ", 
			LoopIndex, OperatorText[BranchOperator], PositiveIncrement, BottomTestingLoop);
		this->DumpFuncNameAndAddr();
		return nullptr;
	}

	STARSOpndTypePtr IncDecOp = InitExpr->GetParentInst()->MakeImmediateOpnd(IncDecValue);
	STARS_sval_t LimitIncrease = 0;
	if (PositiveIncrement) {
		this->LoopIncrementValue[LoopIndex] = (STARS_sval_t)IncDecValue;
		if ((BranchOperator == SMP_LESS_EQUAL) || (BranchOperator == SMP_GREATER_THAN)
			|| (BranchOperator == SMP_BELOW_EQUAL) || (BranchOperator == SMP_ABOVE)) {
			LimitIncrease += IncDecValue; // only because we weeded out some undefined cases; change if we redefine those cases.
		}
	}
	else { // Decrement
		this->LoopIncrementValue[LoopIndex] = (0 - ((STARS_sval_t)IncDecValue));
		if ((BranchOperator == SMP_LESS_THAN) || (BranchOperator == SMP_GREATER_EQUAL)
			|| (BranchOperator == SMP_BELOW) || (BranchOperator == SMP_ABOVE_EQUAL)) {
			LimitIncrease += IncDecValue;
		}
	}

	bool NeedLimitIncrease = (0 != LimitIncrease);
	this->LoopExecutesWithLimitValue[LoopIndex] = NeedLimitIncrease;

	// At this point, the iteration count is (Limit + LimitIncrease - Init) / IncDecValue for the SimpleBIV cases
	//  with PositiveIncrement, and (Init + LimitIncrease - Limit) / IncDecValue for the Decrement case.
	//  We will omit the division by 1 case, of course, although SimplifyExpr() should remove it.
	// If Init is a constant, then we can fold (LimitIncrease - Init) or (Init + LimitIncrease) into a constant operand.
	STARSExpression *IterationsExpr = new STARSExpression();
	IterationsExpr->SetParentInst(LimitExpr->GetParentInst());
	IterationsExpr->SetOriginalParentInst(LimitExpr->GetOriginalParentInst());
	IterationsExpr->SetParentFunc(this);
	bool LimitIncreaseFolded = false;
	if (PositiveIncrement) {
		if ((InitExpr->GetOperator() == SMP_ASSIGN) && InitExpr->GetConstLeftOperand()->IsImmedOp()) {
			// Only LeftOperand is valid, and it is a constant.
			STARS_sval_t InitValue = (STARS_sval_t) InitExpr->GetConstLeftOperand()->GetImmedValue();
			LimitIncrease -= InitValue;
			STARSOpndTypePtr RightOp = InitExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t) LimitIncrease);
			IterationsExpr->SetRightOperand(RightOp);
			LimitIncreaseFolded = true;
			NeedLimitIncrease = false; // already in IterationsExpr
		}
		else { // non-constant InitExpr
			if (InitExpr->GetOperator() == SMP_ASSIGN) { // only an operand
				IterationsExpr->SetRightOperand(InitExpr->GetLeftOperand());
				IterationsExpr->SetRightUseAddr(InitExpr->GetLeftUseAddr());
				IterationsExpr->SetRightSSANum(InitExpr->GetLeftSSANum());
				IterationsExpr->SetRightPreLoopDefAddr(InitExpr->GetLeftPreLoopDefAddr());
			}
			else {
				IterationsExpr->SetRightTree(InitExpr);
			}
		}
		if (NeedLimitIncrease) {
			IterationsExpr->SetOperator(SMP_SUBTRACT);  // (LimitExpr RHS + LimitIncrease) - InitExpr
			STARSOpndTypePtr RightOp = InitExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t)LimitIncrease);
			STARSExpression *LeftExpr = new STARSExpression();
			LeftExpr->SetParentInst(LimitExpr->GetParentInst());
			LeftExpr->SetOriginalParentInst(LimitExpr->GetOriginalParentInst());
			LeftExpr->SetParentFunc(this);
			LeftExpr->SetRightOperand(RightOp);
			LeftExpr->SetOperator(SMP_ADD); // LimitExpr + LimitIncrease is the LeftExpr
			// If LimitExpr right side (limit) is just an operand, make it the LeftOperand for LeftExpr, else make it the LeftTree for LeftExpr.
			if (!LimitExpr->HasRightSubTree()) { // just RightOperand
				LeftExpr->SetLeftOperand(LimitExpr->GetRightOperand());
				LeftExpr->SetLeftUseAddr(LimitExpr->GetRightUseAddr());
				LeftExpr->SetLeftSSANum(LimitExpr->GetRightSSANum());
				LeftExpr->SetLeftPreLoopDefAddr(LimitExpr->GetRightPreLoopDefAddr());
			}
			else {
				LeftExpr->SetLeftTree(LimitExpr->GetRightTree());
			}
			IterationsExpr->SetLeftTree(LeftExpr);
		}
		else if (LimitIncreaseFolded) { // just need LimitExpr right hand side as left tree or operand.
			IterationsExpr->SetOperator(SMP_ADD);  // LimitExpr RHS + (folded LimitIncrease - InitExpr)
			// If LimitExpr RHS is just an operand, make it the LeftOperand, else make it the LeftTree.
			if (!LimitExpr->HasRightSubTree()) { // just RightOperand
				IterationsExpr->SetLeftOperand(LimitExpr->GetRightOperand());
				IterationsExpr->SetLeftUseAddr(LimitExpr->GetRightUseAddr());
				IterationsExpr->SetLeftSSANum(LimitExpr->GetRightSSANum());
				IterationsExpr->SetLeftPreLoopDefAddr(LimitExpr->GetRightPreLoopDefAddr());
			}
			else {
				IterationsExpr->SetLeftTree(LimitExpr->GetRightTree());
			}
		}
#if 0
		else { // never folded, no LimitIncrease, just LimitExpr - InitExpr
			IterationsExpr->SetOperator(SMP_SUBTRACT);
			IterationsExpr->SetLeftTree(LimitExpr);
		}
#else
		else { // never folded, no LimitIncrease, just LimitExprRHS - InitExpr
			IterationsExpr->SetOperator(SMP_SUBTRACT);
			if (!LimitExpr->HasRightSubTree()) { // just RightOperand
				IterationsExpr->SetLeftOperand(LimitExpr->GetRightOperand());
				IterationsExpr->SetLeftUseAddr(LimitExpr->GetRightUseAddr());
				IterationsExpr->SetLeftSSANum(LimitExpr->GetRightSSANum());
				IterationsExpr->SetLeftPreLoopDefAddr(LimitExpr->GetRightPreLoopDefAddr());
			}
			else {
				IterationsExpr->SetLeftTree(LimitExpr->GetRightTree());
			}
		}
#endif
	} // end if PositiveIncrement
	else { // decrement case
		if ((InitExpr->GetOperator() == SMP_ASSIGN) && InitExpr->GetConstLeftOperand()->IsImmedOp()) {
			// Only LeftOperand is valid, and it is a constant.
			STARS_sval_t InitValue = (STARS_sval_t) InitExpr->GetConstLeftOperand()->GetImmedValue();
			LimitIncrease += InitValue;
			STARSOpndTypePtr LeftOp = InitExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t) LimitIncrease);
			IterationsExpr->SetLeftOperand(LeftOp);
			LimitIncreaseFolded = true;
			NeedLimitIncrease = false; // already in IterationsExpr
		}
		else { // non-constant InitExpr
			if (!NeedLimitIncrease) { // LHS of IterationsExpr is just InitExpr
				if (InitExpr->GetOperator() == SMP_ASSIGN) { // only an operand
					IterationsExpr->SetLeftOperand(InitExpr->GetLeftOperand());
					IterationsExpr->SetLeftUseAddr(InitExpr->GetLeftUseAddr());
					IterationsExpr->SetLeftSSANum(InitExpr->GetLeftSSANum());
					IterationsExpr->SetLeftPreLoopDefAddr(InitExpr->GetLeftPreLoopDefAddr());
				}
				else {
					IterationsExpr->SetLeftTree(InitExpr);
				}
			}
			else { // need LHS tree of (InitExpr + LimitIncrease)
				STARSExpression *LeftTree = new STARSExpression();
				LeftTree->SetParentInst(InitExpr->GetParentInst());
				LeftTree->SetOriginalParentInst(InitExpr->GetOriginalParentInst());
				LeftTree->SetParentFunc(this);
				if (InitExpr->GetOperator() == SMP_ASSIGN) { // only an operand
					LeftTree->SetLeftOperand(InitExpr->GetLeftOperand());
					LeftTree->SetLeftUseAddr(InitExpr->GetLeftUseAddr());
					LeftTree->SetLeftSSANum(InitExpr->GetLeftSSANum());
					LeftTree->SetLeftPreLoopDefAddr(InitExpr->GetLeftPreLoopDefAddr());
				}
				else {
					LeftTree->SetLeftTree(InitExpr);
				}
				LeftTree->SetOperator(SMP_ADD);
				STARSOpndTypePtr LeftRightOp = InitExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t) LimitIncrease);
				LeftTree->SetRightOperand(LeftRightOp);
				IterationsExpr->SetLeftTree(LeftTree);
			}
		}
		// Now we just need LimitExpr right hand side as right tree or operand.
		IterationsExpr->SetOperator(SMP_SUBTRACT);  // (folded InitExpr + LimitIncrease) - LimitExpr RHS
		// If LimitExpr RHS is just an operand, make it the RightOperand, else make it the RightTree.
		if (!LimitExpr->HasRightSubTree()) { // just RightOperand
			IterationsExpr->SetRightOperand(LimitExpr->GetRightOperand());
			IterationsExpr->SetRightUseAddr(LimitExpr->GetRightUseAddr());
			IterationsExpr->SetRightSSANum(LimitExpr->GetRightSSANum());
			IterationsExpr->SetRightPreLoopDefAddr(LimitExpr->GetRightPreLoopDefAddr());
		}
		else {
			IterationsExpr->SetRightTree(LimitExpr->GetRightTree());
		}
	} // end .. else decrement case

	// We want to simplify the resulting expression before we divide by IncDecValue.
	//  The division otherwise might have to have complex simplifications. E.g. if we
	//  have InitExpr of RSP-200 and LimitExpr of RSP-100 and IncDecValue of 4, we prefer
	//  to simplify ((RSP - 100) - (RSP - 200)) to 100 and then return 100 / 4. Our 
	//  caller will call SimplifyExpr() on the expr we return.
	assert(nullptr != IterationsExpr);
	bool Simplified = IterationsExpr->SimplifyDriver();
	if (IncDecValue != 1) {
		// Divide by IncDecOp.
		STARSExpression *NewIterExpr = new STARSExpression();
		NewIterExpr->SetParentInst(LimitExpr->GetParentInst());
		NewIterExpr->SetOriginalParentInst(LimitExpr->GetOriginalParentInst());
		NewIterExpr->SetParentFunc(this);
		NewIterExpr->SetOperator(SMP_U_DIVIDE);
		NewIterExpr->SetRightOperand(IncDecOp);
		NewIterExpr->SetRightUseAddr(this->LoopComparisonExprs[LoopIndex].CompareAddr);
		NewIterExpr->SetLeftTree(IterationsExpr);
		return NewIterExpr;
	}
	else {
		return IterationsExpr;
	}
} // end of SMPFunction::CreateIterationsExpr()

// Create a Memory range expression set for all static, indirect or indexed memory writes in the loop.
//  Return true on success, false on failure.
bool SMPFunction::CreateSPARKMemoryWriteRangeExpr(size_t LoopIndex, bool RecordLoopRegs, set<int> &LoopRegHashes, STARSExprSet &MemWriteExprs, STARSMemWriteExprsList &MemWriteExprWidths, vector<set<STARS_ea_t> > &StackPtrCopiesVector) {
	STARSExpression *MemoryRangeExpr = nullptr;

	// Iterate through loop instructions and find all indirect or indexed memory writes.
	list<size_t> LoopBlockList;
	this->BuildLoopBlockList(LoopIndex, LoopBlockList);
	bool success = false;
	bool OneFailure = false;
	int StackPtrCopiesVectorIndex = 0;
	// Triple: STARSExprSetIter, ByteWidth, index into vector of StackPtrCopiesVector
	uint16_t ByteWidth;

	for (size_t BlockNum : LoopBlockList) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		if (CurrBlock->HasIndirectMemWrite() || CurrBlock->HasStaticMemWrite()) {
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (CurrInst->MDHasAnyRepeatPrefix() && CurrInst->MDIsPossibleStringLoopingOpcode()) {
					// Looping string opcodes are dealt with separately.
					continue;
				}
				if (CurrInst->HasIndirectMemoryWrite() || CurrInst->HasStaticMemWrite()) {
					STARS_ea_t InstAddr = CurrInst->GetAddr();
					this->LoopWritesMemory[LoopIndex] = true;
					STARSOpndTypePtr MemDefOp = CurrInst->GetMemDef()->clone();
					CurrInst->MDGetUnnormalizedOp(MemDefOp);
					assert(MemDefOp->IsMemOp());
					// Start with the current memory addressing expression.
					MemoryRangeExpr = this->CreateMemoryAddressExpr(MemDefOp, CurrInst);
					if (nullptr == MemoryRangeExpr) {
						OneFailure = true; // failure
						SMP_msg("ERROR: CreateSPARKMemoryWriteRangeExpr() abandoned due to CreateMemoryAddressExpr() failure at %llx ", (uint64_t) InstAddr);
						this->DumpFuncNameAndAddr();
						continue;
					}
					// Expand and simplify the addressing expression.
					MemoryRangeExpr->EvaluateConsts();
					bool ParentChanged = MemoryRangeExpr->SimplifyDriver();
					ByteWidth = MemDefOp->GetByteWidth();
					bool HasGlobalStaticMemOp = MemDefOp->IsStaticMemOp();
					this->LoopWritesGlobalStaticMemory[LoopIndex] = (HasGlobalStaticMemOp || this->LoopWritesGlobalStaticMemory[LoopIndex]);

					

#if 0
					// Make a clone in case we succeed in expansion and need to trace incoming values to the loop.
					STARSExpression *MemoryRangeExprClone = MemoryRangeExpr->Clone();
#endif
					bool StoppedOnIV = false;
					bool changed = false;
					set<STARS_ea_t> StackPtrCopySet;
					int DepthCounter = 0;
					success = MemoryRangeExpr->ExpandExpr(MemoryRangeExpr->GetParentInst()->GetAddr(), LoopIndex, false, true, true, RecordLoopRegs, false, LoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter);
					if (!success) {
						delete MemoryRangeExpr;
						MemoryRangeExpr = nullptr;
						SMP_msg("ERROR: CreateSPARKMemoryWriteRangeExpr() error due to ExpandExpr() failure on loop %zu InstAddr %llx ",
							LoopIndex, (uint64_t) InstAddr);
						this->DumpFuncNameAndAddr();
						OneFailure = true; // failure
						continue; 
					}
					MemoryRangeExpr->EvaluateConsts();
					ParentChanged = MemoryRangeExpr->SimplifyDriver();
					// Handle multiple memory writes in one loop by keeping a set and combining write exprs later.
					pair<STARSExprSetIter, bool> InsertResult = MemWriteExprs.insert(MemoryRangeExpr);
					if (InsertResult.second) { // New expr in set
						int TempIndex = StackPtrCopiesVectorIndex;
						if (StackPtrCopySet.empty()) {
							TempIndex = -1; // don't need to push empty sets onto vector
						}
						else {
							StackPtrCopiesVector.push_back(StackPtrCopySet);
							++StackPtrCopiesVectorIndex;
						}
						pair<size_t, int> InsertPair2((size_t) ByteWidth, TempIndex);
						pair<STARSExprSetIter, pair<size_t, int> > InsertValue(InsertResult.first, InsertPair2);
						MemWriteExprWidths.push_back(InsertValue);
						// StackPtrCopySet.clear();
					}			
					if (MemoryRangeExpr->AreAllRegsLoopInvariant()) {
						// We need to place this memory write expr into the function-level non-loop expr container.
						//  This container is InArgsUsedInMemWrites, which should already have this expr, so there
						//  is no further work for us to do here. The exception is when the memory write expr is
						//  a global static mem write, which was not necessarily traced back to an InArg. We will
						//  store those in the InArgsUsedInMemWrites container.
						SMP_msg("INFO: Loop-invariant mem expr inside loop at %llx\n", InstAddr);
						if (!HasGlobalStaticMemOp) {
							continue;
						}
						else {
							pair<STARSExprSetIter, bool> InsertResult = this->InArgsUsedInMemWrites[LoopIndex + 1].insert(MemoryRangeExpr);
							if (InsertResult.second) {
								uint32_t ExprSerialNumber = MemoryRangeExpr->GetSerialNumber();
								pair<uint32_t, size_t> InsertValue(ExprSerialNumber, ByteWidth);
								this->InArgsUsedInMemWriteByteWidths[LoopIndex + 1].push_back(InsertValue);
							}
						}
					}
				}
			} // end for all instructions in block
		} // end if indirect memory write in current block
	} // end for all blocks

	if (global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
		// Combine exprs like (RDI+2) and (RDI+3) into a single expr of greater width for proof simplicity.
		this->CombineMemoryExprs(MemWriteExprWidths, MemWriteExprs, StackPtrCopiesVector);
	}

	success = (0 < MemWriteExprs.size());
	if (OneFailure)
		this->LoopAnalysisProblems[LoopIndex] = true;

	return success;
} // end of SMPFunction::CreateSPARKMemoryWriteRangeExpr()

// Create a Memory range expression set for all static, indirect or indexed memory reads in the loop.
//  Return true on success, false on failure.
bool SMPFunction::CreateSPARKMemoryReadRangeExprs(size_t LoopIndex, bool RecordLoopRegs, set<int> &LoopRegHashes, STARSExprSet &MemReadExprs, STARSMemWriteExprsList &MemReadExprWidths, vector<set<STARS_ea_t> > &StackPtrCopiesVector) {
	STARSExpression *MemoryRangeExpr = nullptr;

	// Iterate through loop instructions and find all static, indirect or indexed memory reads.
	list<size_t> LoopBlockList;
	this->BuildLoopBlockList(LoopIndex, LoopBlockList);
	bool success = false;
	bool OneFailure = false;
	int StackPtrCopiesVectorIndex = 0;
	// Triple: STARSExprSetIter, ByteWidth, index into vector of StackPtrCopiesVector
	uint16_t ByteWidth;

	for (size_t BlockNum : LoopBlockList) {
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockNum);
		if (CurrBlock->HasIndirectMemRead() || CurrBlock->HasStaticMemRead()) {
			this->LoopReadsMemory[LoopIndex] = true;
#if 0
			if (CurrBlock->HasStaticMemRead()) {
#endif
				for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
					SMPInstr *CurrInst = (*InstIter);
					if (CurrInst->MDHasAnyRepeatPrefix() && CurrInst->MDIsPossibleStringLoopingOpcode()) {
						// Looping string opcodes are dealt with separately.
						continue;
					}
					if (CurrInst->HasStaticMemRead() || CurrInst->HasIndirectMemoryRead()) {
						STARS_ea_t InstAddr = CurrInst->GetAddr();
						STARSOpndTypePtr MemUseOp = CurrInst->GetMemUse()->clone();
						CurrInst->MDGetUnnormalizedOp(MemUseOp);
						assert(MemUseOp->IsMemOp());

#if 0
						// For supporting move_globals, we only care about true static memory operands
						//  with static offsets in the operand.
						STARS_ea_t Displacement = MemUseOp->GetAddr();
						if (!IsImmedGlobalAddress(Displacement))
							continue;
#endif

						// Start with the current memory addressing expression.
						MemoryRangeExpr = this->CreateMemoryAddressExpr(MemUseOp, CurrInst);
						if (nullptr == MemoryRangeExpr) {
							OneFailure = true; // failure
							SMP_msg("ERROR: LOOP: CreateSPARKMemoryReadRangeExprs() abandoned due to CreateMemoryAddressExpr() failure at %llx ", (uint64_t)InstAddr);
							this->DumpFuncNameAndAddr();
							continue;
						}
						// Expand and simplify the addressing expression.
						MemoryRangeExpr->EvaluateConsts();
						bool ParentChanged = MemoryRangeExpr->SimplifyDriver();
						ByteWidth = MemUseOp->GetByteWidth();
						bool HasGlobalStaticMemOp = MemUseOp->IsStaticMemOp();
						this->LoopReadsGlobalStaticMemory[LoopIndex] = (HasGlobalStaticMemOp || this->LoopReadsGlobalStaticMemory[LoopIndex]);


						bool StoppedOnIV = false;
						bool changed = false;
						set<STARS_ea_t> StackPtrCopySet;
						int DepthCounter = 0;
						success = MemoryRangeExpr->ExpandExpr(MemoryRangeExpr->GetParentInst()->GetAddr(), LoopIndex, false, true, true, RecordLoopRegs, true, LoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter);
						if (!success) {
							delete MemoryRangeExpr;
							MemoryRangeExpr = nullptr;
							SMP_msg("ERROR: LOOP: CreateSPARKMemoryReadRangeExpr() error due to ExpandExpr() failure on loop %zu InstAddr %llx ",
								LoopIndex, (uint64_t)InstAddr);
							this->DumpFuncNameAndAddr();
							OneFailure = true; // failure
							continue;
						}
						MemoryRangeExpr->EvaluateConsts();
						ParentChanged = MemoryRangeExpr->SimplifyDriver();
						// Handle multiple memory reads in one loop by keeping a set and combining write exprs later.
						pair<STARSExprSetIter, bool> InsertResult = MemReadExprs.insert(MemoryRangeExpr);
						if (InsertResult.second) { // New expr in set
							int TempIndex = StackPtrCopiesVectorIndex;
							if (StackPtrCopySet.empty()) {
								TempIndex = -1; // don't need to push empty sets onto vector
							}
							else {
								StackPtrCopiesVector.push_back(StackPtrCopySet);
								++StackPtrCopiesVectorIndex;
							}
							pair<size_t, int> InsertPair2((size_t)ByteWidth, TempIndex);
							pair<STARSExprSetIter, pair<size_t, int> > InsertValue(InsertResult.first, InsertPair2);
							MemReadExprWidths.push_back(InsertValue);
							// StackPtrCopySet.clear();
						}
#if 0	// seems SPARK oriented; does not appear useful for memory annotations
						if (MemoryRangeExpr->AreAllRegsLoopInvariant()) {
							// We need to place this memory write expr into the function-level non-loop expr container.
							//  This container is InArgsUsedInMemWrites, which should already have this expr, so there
							//  is no further work for us to do here. The exception is when the memory write expr is
							//  a global static mem write, which was not necessarily traced back to an InArg. We will
							//  store those in the InArgsUsedInMemWrites container.
							SMP_msg("INFO: Loop-invariant mem expr inside loop at %llx\n", InstAddr);
							if (!HasGlobalStaticMemOp) {
								continue;
							}
							else {
								pair<STARSExprSetIter, bool> InsertResult = this->InArgsUsedInMemWrites[LoopIndex + 1].insert(MemoryRangeExpr);
								if (InsertResult.second) {
									pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, ByteWidth);
									this->InArgsUsedInMemWriteByteWidths[LoopIndex + 1].push_back(InsertValue);
								}
							}
						}
#endif
					}
				} // end for all instructions in block
#if 0
			} // end if static memory read in current block
#endif
		} // end if static or indirect memory read in current block
	} // end for all blocks

#if 0	// not yet emitting SPARK annotations for memory reads
	if (global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
		// Combine exprs like (RDI+2) and (RDI+3) into a single expr of greater width for proof simplicity.
		this->CombineMemoryExprs(MemReadExprWidths, MemReadExprs, StackPtrCopiesVector);
	}
#endif

	success = (!MemReadExprs.empty());
#if 0	// for SPARK use only?
	if (OneFailure)
		this->LoopAnalysisProblems[LoopIndex] = true;
#endif

	return success;
} // end of SMPFunction::CreateSPARKMemoryReadRangeExprs()

// create expression for the memory address computation in MemDefOp
STARSExpression* SMPFunction::CreateMemoryAddressExpr(const STARSOpndTypePtr &MemDefOp, SMPInstr *WriteInst) {
	STARSExpression *MemoryAddressExpr = nullptr;
	STARS_ea_t WriteAddr = WriteInst->GetAddr();
	if (MemDefOp->GetSegReg() == STARS_x86_R_cs && (!MemDefOp->IsStaticMemOp())) {
		SMP_msg("ERROR: CreateMemoryAddressExpr() fails on CS-relative MemDefOp at %llx ", (uint64_t) WriteAddr);
		this->DumpFuncNameAndAddr();
		return MemoryAddressExpr; // cannot handle IP-relative memory writes, and should only happen for reads
	}

	int BaseReg, IndexReg;
	uint16_t ScaleFactor;
	STARS_ea_t offset;
	MDExtractAddressFields(MemDefOp, BaseReg, IndexReg, ScaleFactor, offset);

	// Construct an expression based on which address elements are present.
	STARS_regnum_t SegReg = MemDefOp->GetSegReg();
	bool HasIndexReg = (STARS_x86_R_none != IndexReg);
	bool IsScaled = HasIndexReg && (0 != ScaleFactor);
	bool HasBaseReg = (STARS_x86_R_none != BaseReg);
	bool HasOffset = (0 != offset);
	bool HasSegReg = (SegReg != STARS_x86_R_none);
	bool IPRelative = (SegReg == MD_CODE_SEGMENT_REG);

	STARSOpndTypePtr IndexRegOp = nullptr;
	STARSDefUseIter IndexUseIter = WriteInst->GetLastUse();
	if (HasIndexReg) {
		IndexRegOp = WriteInst->MakeRegOpnd((STARS_regnum_t) IndexReg);
		CanonicalizeOpnd(IndexRegOp);
		IndexUseIter = WriteInst->FindUse(IndexRegOp);
		assert(IndexUseIter != WriteInst->GetLastUse());
	}
	STARSOpndTypePtr BaseRegOp = nullptr;
	STARSDefUseIter BaseUseIter = WriteInst->GetLastUse();
	if (HasBaseReg) {
		BaseRegOp = WriteInst->MakeRegOpnd((STARS_regnum_t) BaseReg);
		CanonicalizeOpnd(BaseRegOp);
		BaseUseIter = WriteInst->FindUse(BaseRegOp);
		assert(BaseUseIter != WriteInst->GetLastUse());
	}
	STARSOpndTypePtr OffsetOp = nullptr;
	if (HasOffset) {
		OffsetOp = WriteInst->MakeImmediateOpnd((STARS_uval_t) offset);
		OffsetOp->SetByteWidth(MemDefOp->GetByteWidth());
	}


	MemoryAddressExpr = new STARSExpression();
	MemoryAddressExpr->SetParentInst(WriteInst); // also sets OriginalParentInst
	// MemoryAddressExpr->SetOriginalParentInst(WriteInst);
	MemoryAddressExpr->SetParentFunc(this);
	if (IsScaled) { // has IndexReg and ScaleFactor
		if (HasBaseReg || HasOffset) {
			// Root operator should be SMP_ADD, right tree is scaled IndexReg, LeftTree is BaseReg and/or offset
			MemoryAddressExpr->SetOperator(SMP_ADD);
			STARSExpression *RightExpr = new STARSExpression();
			RightExpr->SetParentInst(WriteInst);
			RightExpr->SetOriginalParentInst(WriteInst);
			RightExpr->SetParentFunc(this);
			RightExpr->SetOperator(SMP_U_MULTIPLY);
			RightExpr->SetLeftOperand(IndexRegOp);
			RightExpr->SetLeftUseAddr(WriteAddr);
			RightExpr->SetLeftSSANum(IndexUseIter->GetSSANum());
			RightExpr->SetRightOperand(WriteInst->MakeImmediateOpnd((STARS_uval_t) (1 << ScaleFactor)));
			MemoryAddressExpr->SetRightTree(RightExpr);
			if (HasBaseReg && HasOffset) { // has both; need to add them together
				STARSExpression *LeftExpr = new STARSExpression();
				LeftExpr->SetParentInst(WriteInst);
				LeftExpr->SetOriginalParentInst(WriteInst);
				LeftExpr->SetParentFunc(this);
				LeftExpr->SetOperator(SMP_ADD);
				LeftExpr->SetLeftOperand(BaseRegOp);
				LeftExpr->SetLeftUseAddr(WriteAddr);
				LeftExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
				LeftExpr->SetRightOperand(OffsetOp);
				LeftExpr->SetRightUseAddr(WriteAddr);
				MemoryAddressExpr->SetLeftTree(LeftExpr);
			}
			else { // BaseReg or offset, not both; just need a LeftOperand
				if (HasBaseReg) {
					MemoryAddressExpr->SetLeftOperand(BaseRegOp);
					MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
					MemoryAddressExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
				}
				else {
					MemoryAddressExpr->SetLeftOperand(OffsetOp);
					MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
				}
			}
		}
		else { // scaled IndexReg, nothing else
			MemoryAddressExpr->SetOperator(SMP_U_MULTIPLY);
			MemoryAddressExpr->SetLeftOperand(IndexRegOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
			MemoryAddressExpr->SetLeftSSANum(IndexUseIter->GetSSANum());
			MemoryAddressExpr->SetRightOperand(WriteInst->MakeImmediateOpnd((STARS_uval_t)(1 << ScaleFactor)));
			MemoryAddressExpr->SetRightUseAddr(WriteAddr);
		}
	}
	else if (HasIndexReg) { // unscaled IndexReg
		if (HasBaseReg && HasOffset) { // has both; need to add them together: (BaseReg + offset) + IndexReg
			STARSExpression *LeftExpr = new STARSExpression();
			LeftExpr->SetParentInst(WriteInst); // also sets OriginalParentInst
			// LeftExpr->SetOriginalParentInst(WriteInst);
			LeftExpr->SetParentFunc(this);
			LeftExpr->SetOperator(SMP_ADD);
			LeftExpr->SetLeftOperand(BaseRegOp);
			LeftExpr->SetLeftUseAddr(WriteAddr);
			LeftExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
			LeftExpr->SetRightOperand(OffsetOp);
			LeftExpr->SetRightUseAddr(WriteAddr);
			MemoryAddressExpr->SetOperator(SMP_ADD);
			MemoryAddressExpr->SetLeftTree(LeftExpr);
			MemoryAddressExpr->SetRightOperand(IndexRegOp);
			MemoryAddressExpr->SetRightUseAddr(WriteAddr);
			MemoryAddressExpr->SetRightSSANum(IndexUseIter->GetSSANum());
		}
		else { // BaseReg or offset, or neither, not both; just need a LeftOperand
			if (HasBaseReg) { // BaseReg + IndexReg
				MemoryAddressExpr->SetOperator(SMP_ADD);
				MemoryAddressExpr->SetLeftOperand(BaseRegOp);
				MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
				MemoryAddressExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
				MemoryAddressExpr->SetRightOperand(IndexRegOp);
				MemoryAddressExpr->SetRightUseAddr(WriteAddr);
				MemoryAddressExpr->SetRightSSANum(IndexUseIter->GetSSANum());
			}
			else if (HasOffset) {  // IndexReg + offset
				MemoryAddressExpr->SetOperator(SMP_ADD);
				MemoryAddressExpr->SetLeftOperand(IndexRegOp);
				MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
				MemoryAddressExpr->SetLeftSSANum(IndexUseIter->GetSSANum());
				MemoryAddressExpr->SetRightOperand(OffsetOp);
				MemoryAddressExpr->SetRightUseAddr(WriteAddr);
			}
			else { // neither BaseReg nor offset, just IndexReg
				MemoryAddressExpr->SetOperator(SMP_ASSIGN); // signal that only LeftOperand is valid
				MemoryAddressExpr->SetLeftOperand(IndexRegOp);
				MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
				MemoryAddressExpr->SetLeftSSANum(IndexUseIter->GetSSANum());
			}
		}
	}
	else if (HasBaseReg) { // BaseReg, no IndexReg
		if (HasOffset) { // BaseReg + offset
			MemoryAddressExpr->SetOperator(SMP_ADD);
			MemoryAddressExpr->SetLeftOperand(BaseRegOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
			MemoryAddressExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
			MemoryAddressExpr->SetRightOperand(OffsetOp);
			MemoryAddressExpr->SetRightUseAddr(WriteAddr);
		}
		else { // just BaseReg
			MemoryAddressExpr->SetOperator(SMP_ASSIGN); // signal that only LeftOperand is valid
			MemoryAddressExpr->SetLeftOperand(BaseRegOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
			MemoryAddressExpr->SetLeftSSANum(BaseUseIter->GetSSANum());
		}
	}
	else { // just offset or SegReg or SegReg+offset
		if (!(HasOffset || HasSegReg)) {
			// Code might be unreachable debug path code, not removed
			//  due to low level of compiler optimization.
			SMP_msg("SERIOUS WARNING: (Unreachable?) mem offset zero at %llx\n",
				(uint64_t) WriteAddr);
			WriteInst->Dump();
			HasOffset = true; // offset is zero
			OffsetOp = WriteInst->MakeImmediateOpnd((STARS_uval_t) offset);
			OffsetOp->SetByteWidth(MemDefOp->GetByteWidth());
		}
		MemoryAddressExpr->SetOperator(SMP_ASSIGN); // signal that only LeftOperand is valid

		if (HasOffset && HasSegReg && (!IPRelative)) {
			MemoryAddressExpr->SetOperator(SMP_ADD);
			STARSOpndTypePtr SegRegOp = WriteInst->MakeRegOpnd(MemDefOp->GetSegReg());
			MemoryAddressExpr->SetLeftOperand(SegRegOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
			MemoryAddressExpr->SetRightOperand(OffsetOp);
			MemoryAddressExpr->SetRightUseAddr(WriteAddr);
		}
		else if (HasOffset) {
			MemoryAddressExpr->SetLeftOperand(OffsetOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
		}
		else { // only seg reg
			STARSOpndTypePtr SegRegOp = WriteInst->MakeRegOpnd(MemDefOp->GetSegReg());
			MemoryAddressExpr->SetLeftOperand(SegRegOp);
			MemoryAddressExpr->SetLeftUseAddr(WriteAddr);
		}
	}

	return MemoryAddressExpr;
} // end of SMPFunction::CreateMemoryAddressExpr()

// Utility function to find an iterator in a vector of iterators.
bool IsIterInVector(STARSMemWriteExprListIter ListIter, vector<STARSMemWriteExprListIter> IterVector) {
	bool Found = false;
	for (size_t i = 0; i < IterVector.size(); ++i) {
		if (IterVector[i] == ListIter) {
			Found = true;
			break;
		}
	}
	return Found;
} // end of IsIterInVector()

// Combine items in the expr set that have common regs, e.g. (RDI+1) and (RDI-2), into fewer set entries
void SMPFunction::CombineMemoryExprs(STARSMemWriteExprsList &MemWriteExprWidths, STARSExprSet &MemWriteExprs, vector<set<STARS_ea_t> > &StackPtrCopies) {
	if (1 >= MemWriteExprWidths.size())
		return; // cannot combine exprs unless we have two or more

	// Start with the first list item and scan for overlapping memory exprs.
	//  Restrict to simple assignment exprs like [RDI+1] and [RDI-1], not relational exprs like [RDI+RCX+1].
	//  Implement a methodical, simple search for matching exprs, combine exprs, repeat to end of list.
	STARSMemWriteExprsList NewMemWriteExprWidths;
	STARSExprSet NewExprSet; // hold combined exprs until we are ready to re-write the whole set
	STARSMemWriteExprListIter FirstListIter = MemWriteExprWidths.begin();
	vector<STARSMemWriteExprListIter> ItersToDiscard; // iters that match FirstIter
	while (FirstListIter != MemWriteExprWidths.end()) {
		STARSMemWriteExprListIter SecondListIter = FirstListIter;
		// Advance SecondListIter to the next iter location that is not already marked to be discarded.
		do {
			++SecondListIter;
		} while ((SecondListIter != MemWriteExprWidths.end()) && IsIterInVector(SecondListIter, ItersToDiscard));
		STARSExprSetIter FirstSetIter = (*FirstListIter).first;
		bool MatchSeen = false;
		STARS_sval_t MinOffset = 0, MaxOffset = 0; // MaxOffset is one beyond the last byte offset written
		STARS_sval_t FirstOffset; // save the ImmedValue of the FirstSetIter expr
		int FirstStackPtrCopiesVecIndex = (*FirstListIter).second.second;
		while (SecondListIter != MemWriteExprWidths.end()) {
			STARSExprSetIter SecondSetIter = (*SecondListIter).first;
			if ((*FirstSetIter)->IsEqualExprExceptingImmedOpnds(*SecondSetIter)) {
				ItersToDiscard.push_back(SecondListIter);
				// Keep track of minimum and maximum offsets seen from address reg.
				STARS_uval_t ImmedValue = 0;
				if (!MatchSeen) { // SecondSetIter is first match we have seen
					// Initialize MinOffset and MaxOffset
					if ((*FirstSetIter)->ExtractImmedValue(ImmedValue)) {
						STARS_sval_t SignedImmedValue = (STARS_sval_t) ImmedValue;
						MinOffset = SignedImmedValue;
						MaxOffset = SignedImmedValue + (STARS_sval_t) (*FirstListIter).second.first;
					}
					else {
						MinOffset = 0;
						MaxOffset = (STARS_sval_t)(*FirstListIter).second.first; // the size
					}
					FirstOffset = MinOffset;
					MatchSeen = true;
				}
				// See if SecondSetIter's expr has an extension of the min-max range.
				if ((*SecondSetIter)->ExtractImmedValue(ImmedValue)) {
					STARS_sval_t SignedImmedValue = (STARS_sval_t) ImmedValue;
					STARS_sval_t UpperBound = SignedImmedValue + (STARS_sval_t)(*SecondListIter).second.first; // offset + size
					if (SignedImmedValue < MinOffset)
						MinOffset = SignedImmedValue;
					if (UpperBound > MaxOffset)
						MaxOffset = UpperBound;
				}
				else { // no immediate value => offset of zero
					STARS_sval_t UpperBound = (STARS_sval_t) (*SecondListIter).second.first; // offset zero + size
					if (UpperBound > MaxOffset)
						MaxOffset = UpperBound;
				}
			} // end if ((*FirstSetIter)->IsEqualExprExceptingImmedOpnds(*SecondSetIter)) 

			int SecondStackPtrCopiesVecIndex = (*SecondListIter).second.second;
			if (0 <= SecondStackPtrCopiesVecIndex) { // non-empty set of stack ptr copy addrs
				// Merge stack ptr copy set into first set
				if (0 <= FirstStackPtrCopiesVecIndex) { // first set is also non-empty
					set<STARS_ea_t>::iterator SecondCopiesIter = StackPtrCopies[SecondStackPtrCopiesVecIndex].begin();
					while (SecondCopiesIter != StackPtrCopies[SecondStackPtrCopiesVecIndex].end()) {
						STARS_ea_t CopyAddr = (*SecondCopiesIter);
						StackPtrCopies[FirstStackPtrCopiesVecIndex].insert(CopyAddr);
						++SecondCopiesIter;
					}
				}
				else { // first set empty, second set non-empty
					// Just use the second set
					FirstStackPtrCopiesVecIndex = SecondStackPtrCopiesVecIndex;
					(*FirstListIter).second.second = FirstStackPtrCopiesVecIndex;
				}
			}
			++SecondListIter;
		} // end while (SecondListIter != MemWriteExprWidths.end()) 
		if (MatchSeen) {
			// We need to broaden the range of the FirstListIter mem write, and discard the matching iters.
			STARSExpression *NewFirstExpr = nullptr;
			if (MinOffset != FirstOffset) {
				// Cloned expr needs its ImmedValue changed to MinOffset.
				if (FirstOffset == 0) {
					// Original expr probably had no immediate operand. Create one.
					NewFirstExpr = new STARSExpression();
					// Make left tree be the original expr
					NewFirstExpr->SetLeftTree((*FirstSetIter)->Clone());
					NewFirstExpr->SetParentFunc(this);
					NewFirstExpr->SetParentInst(NewFirstExpr->GetLeftTree()->GetParentInst());
					NewFirstExpr->SetOriginalParentInst(NewFirstExpr->GetParentInst());
					NewFirstExpr->SetLeftPreLoopDefAddr(NewFirstExpr->GetLeftTree()->GetLeftPreLoopDefAddr());
					if (0 > MinOffset) {
						NewFirstExpr->SetOperator(SMP_SUBTRACT);
						// Make right operand be a new ImmedOp with value == absolute value of MinOffset.
						STARS_sval_t AbsValue = (0 - MinOffset);
						NewFirstExpr->SetRightOperand(NewFirstExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t) AbsValue));
					}
					else { // 0 < MinOffset
						NewFirstExpr->SetOperator(SMP_ADD);
						// Make right operand be a new ImmedOp with value == MinOffset.
						NewFirstExpr->SetRightOperand(NewFirstExpr->GetParentInst()->MakeImmediateOpnd((STARS_uval_t) MinOffset));
					}
				}
				else { // FirstOffset != 0
					// First expr had an immediate operand; find it and change its value.
					NewFirstExpr = (*FirstSetIter)->Clone();
					bool UpdatedImmed = NewFirstExpr->UpdateImmedValue((STARS_uval_t) FirstOffset, (STARS_uval_t) MinOffset);
					assert(UpdatedImmed);
				}
			}
			else { // MinOffset == FirstOffset
				// Just need to clone first expr and set width to new range
				NewFirstExpr = (*FirstSetIter)->Clone();
			}
			// NewFirstExpr now has the correct MinOffset.
			pair<STARSExprSetIter, bool> InsertResult = NewExprSet.insert(NewFirstExpr);
			assert(InsertResult.second); // All exprs that are equal except for offsets are separated and unique
			size_t NewRange = ((size_t) (MaxOffset - MinOffset));
			pair<size_t, int> InsertPair2(NewRange, FirstStackPtrCopiesVecIndex);
			pair<STARSExprSetIter, pair<size_t, int> > InsertValue(InsertResult.first, InsertPair2);
			NewMemWriteExprWidths.push_back(InsertValue);
		} // end if (MatchSeen)
		else { // !MatchSeen
			// FirstSetIter expr was unique.
			pair<STARSExprSetIter, bool> InsertResult = NewExprSet.insert(*FirstSetIter); // keep old expr
			assert(InsertResult.second); // All exprs that are equal except for offsets are separated and unique
			size_t OldRange = (*FirstListIter).second.first; // keep the old width
			int OldCopiesVecIndex = (*FirstListIter).second.second; // keep the old index
			pair<size_t, int> InsertPair2(OldRange, OldCopiesVecIndex);
			pair<STARSExprSetIter, pair<size_t, int> > InsertValue(InsertResult.first, InsertPair2);
			NewMemWriteExprWidths.push_back(InsertValue);
		}
		// Now move on to the next unique base expr.
		do {
			++FirstListIter;
		} while ((FirstListIter != MemWriteExprWidths.end()) && IsIterInVector(FirstListIter, ItersToDiscard));
	} // end while ((FirstListIter != MemWriteExprWidths.end()) && (1 < MemWriteExprWidths.size()))
	
	// If we combined any exprs, replace our arguments with the new expr set and new list of widths and ExprSetIters.
	if (MemWriteExprWidths.size() != NewMemWriteExprWidths.size()) {
		// Some combining happened. Copy new set and new list to arguments.
		assert(MemWriteExprWidths.size() > NewMemWriteExprWidths.size());
		MemWriteExprs.clear();
		MemWriteExprs.insert(NewExprSet.begin(), NewExprSet.end());
		MemWriteExprWidths.clear();
		// Synchronize the iters in MemWriteExprWidths to the new MemWriteExprs locations.
		STARSMemWriteExprsList::const_iterator NewListIter;
		for (NewListIter = NewMemWriteExprWidths.cbegin(); NewListIter != NewMemWriteExprWidths.cend(); ++NewListIter) {
			STARSExprSetIter NewSetIter = (*NewListIter).first;
			size_t NewWidth = (*NewListIter).second.first;
			// Find the matching expr that we just inserted into MemWriteExprs.
			STARSExprSetIter OldSetIter = MemWriteExprs.find(*NewSetIter);
			assert(OldSetIter != MemWriteExprs.end());
			pair<size_t, int> InsertPair2(NewWidth, (*NewListIter).second.second);
			pair<STARSExprSetIter, pair<size_t, int> > InsertValue(OldSetIter, InsertPair2);
			MemWriteExprWidths.push_back(InsertValue);
		}
	}

	return;
} // end of SMPFunction::CombineMemoryExprs()

// Replace occurrences of IVFamily.BIV in CurrExpr with BIV InitExpr or LimitExpr
bool SMPFunction::ReplaceBIVWithExpr(std::size_t LoopIndex, const struct InductionVarFamily &IVFamily, STARSExpression *CurrExpr, bool InitCase) {
	bool success = false;

	// Traverse CurrExpr to find the BIV from IVFamily.
	if (!CurrExpr->HasLeftSubTree()) {
		if ((CurrExpr->GetLeftSSANum() == IVFamily.BasicInductionVar.InductionVar.GetSSANum())
			&& IsEqOp(CurrExpr->GetConstLeftOperand(), IVFamily.BasicInductionVar.InductionVar.GetOp())) {
			// Matched. Replace CurrExpr->LeftOperand with InitExpr or LimitExpr->GetRightTree().
			//  For LimitExpr, the LeftOperand is compared to the RightOperand or RightTree, so the right side is the limit.
			if (InitCase) {
				STARSExpression *ReplaceExpr = this->LoopIterationsInitExprs[LoopIndex];
				STARSExpression *FinalExpr = ReplaceExpr->Clone();
				CurrExpr->SetLeftTree(FinalExpr);
				CurrExpr->SetLeftOperand(nullptr);
			}
			else { // LimitExpr case
				if (this->LoopIterationsLimitExprs[LoopIndex]->HasRightSubTree()) {
					STARSExpression *ReplaceExpr = this->LoopIterationsLimitExprs[LoopIndex]->GetRightTree();
					STARSExpression *FinalExpr = ReplaceExpr->Clone();
					CurrExpr->SetLeftTree(FinalExpr);
					CurrExpr->SetLeftOperand(nullptr);
				}
				else { // just a RightOperand, so replace CurrExpr->LeftOperand with LimitExpr->RightOperand
					CurrExpr->SetLeftOperand(this->LoopIterationsLimitExprs[LoopIndex]->GetRightOperand());
					CurrExpr->SetLeftUseAddr(this->LoopIterationsLimitExprs[LoopIndex]->GetRightUseAddr());
					CurrExpr->SetLeftSSANum(this->LoopIterationsLimitExprs[LoopIndex]->GetRightSSANum());
					CurrExpr->SetLeftPreLoopDefAddr(this->LoopIterationsLimitExprs[LoopIndex]->GetRightPreLoopDefAddr());
				}
			}
			success = true;
		}
	}
	else {
		success = this->ReplaceBIVWithExpr(LoopIndex, IVFamily, CurrExpr->GetLeftTree(), InitCase); // recurse
	}
	if (!CurrExpr->HasRightSubTree()) {
		if ((CurrExpr->GetRightSSANum() == IVFamily.BasicInductionVar.InductionVar.GetSSANum())
			&& IsEqOp(CurrExpr->GetConstRightOperand(), IVFamily.BasicInductionVar.InductionVar.GetOp())) {
			// Matched. Replace CurrExpr->RightOperand with InitExpr or LimitExpr->GetRightTree() or LimitExpr->GetRightOperand().
			//  For LimitExpr, the LeftOperand is compared to to the RightOperand or RightTree, so the right side is the limit.
			if (InitCase) {
				STARSExpression *ReplaceExpr = this->LoopIterationsInitExprs[LoopIndex];
				STARSExpression *FinalExpr = ReplaceExpr->Clone();
				CurrExpr->SetRightTree(FinalExpr);
				CurrExpr->SetRightOperand(nullptr);
			}
			else { // LimitExpr case
				if (this->LoopIterationsLimitExprs[LoopIndex]->HasRightSubTree()) {
					STARSExpression *ReplaceExpr = this->LoopIterationsLimitExprs[LoopIndex]->GetRightTree();
					STARSExpression *FinalExpr = ReplaceExpr->Clone();
					CurrExpr->SetRightTree(FinalExpr);
					CurrExpr->SetRightOperand(nullptr);
				}
				else { // just a RightOperand, so replace CurrExpr->RightOperand with LimitExpr->RightOperand
					CurrExpr->SetRightOperand(this->LoopIterationsLimitExprs[LoopIndex]->GetRightOperand());
					CurrExpr->SetRightUseAddr(this->LoopIterationsLimitExprs[LoopIndex]->GetRightUseAddr());
					CurrExpr->SetRightSSANum(this->LoopIterationsLimitExprs[LoopIndex]->GetRightSSANum());
					CurrExpr->SetRightPreLoopDefAddr(this->LoopIterationsLimitExprs[LoopIndex]->GetRightPreLoopDefAddr());
				}
			}
			success = true;
		}
	}
	else {
		success |= this->ReplaceBIVWithExpr(LoopIndex, IVFamily, CurrExpr->GetRightTree(), InitCase); // recurse
	}

	return success;
} // end of SMPFunction::ReplaceBIVWithExpr()

// Replace all BIVs in CurrExpr with lower or upper limit (depending on InitCase) exprs for BIV
bool SMPFunction::ReplaceAllBIVsWithExprs(std::size_t LoopIndex, STARSExpression *CurrExpr, bool InitCase, bool &changed, set<int> &RegsReplaced) {
	bool success = true;
	bool UseFP = this->UsesFramePointer();
	size_t FamilyIndex;
	STARSInductionVarFamilyIter ListIter;

	// Traverse CurrExpr to find the induction vars.
	if (!CurrExpr->HasLeftSubTree()) {
		const STARSOpndTypePtr CurrOp = CurrExpr->GetConstLeftOperand();
		if (MDIsDataFlowOpnd(CurrOp, UseFP)) {
			SMPInstr *UseInst = CurrExpr->GetParentInst();
			int CurrSSANum = CurrExpr->GetLeftSSANum();
			int IVLoopIndex = -1;
			if (this->IsLoopNestInductionVar(CurrOp, UseInst, ListIter, FamilyIndex, IVLoopIndex)) {
				if (0 == FamilyIndex) { // BIV, not DIV
					struct InductionVarFamily IVFamily = (*ListIter);
					if ((CurrSSANum == IVFamily.BasicInductionVar.InductionVar.GetSSANum())
						&& IsEqOp(CurrOp, IVFamily.BasicInductionVar.InductionVar.GetOp())) {
						// Matched. Replace CurrExpr->LeftOperand with InitExpr or LimitExpr->GetRightTree()
						//  from the BIV matched.
						//  For LimitExpr, the LeftOperand is compared to the RightOperand or RightTree, so the right side is the limit.

						// Avoid infinite recursion by seeing if this operand has already been replaced.
						bool AlreadyReplaced = false;
						if (CurrOp->IsRegOp()) {
							int RegHashIndex = HashGlobalNameAndSSA(CurrOp, CurrSSANum);
							if (RegsReplaced.find(RegHashIndex) != RegsReplaced.cend()) {
								success = true;
								AlreadyReplaced = true;
							}
							else {
								(void) RegsReplaced.insert(RegHashIndex);
							}
						}
						if (!AlreadyReplaced) {
							if (InitCase) {
								if (nullptr == IVFamily.BIVInitExpr) {
									assert(0 <= IVLoopIndex);
									success = false;
									SMP_msg("ERROR: null IVFamily.BIVInitExpr: IVFamily Dump follows.\n");
									DumpInductionVarFamily(IVFamily);
								}
								else {
									STARSExpression *ReplaceExpr = IVFamily.BIVInitExpr;

									// The InitExpr for an IV can be part of the InitExpr
									//  for a memory access. Avoid infinite recursion by seeing
									//  if we have matching exprs.
									if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
										STARSExpression *FinalExpr = ReplaceExpr->Clone();
										CurrExpr->SetLeftTree(FinalExpr);
										CurrExpr->SetLeftOperand(nullptr);
										CurrExpr->SetParentInst(FinalExpr->GetParentInst());
										CurrExpr->GetLeftTree()->SetProcessed(); // cut off infinite replacement
										changed = true;
										success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed, RegsReplaced); // recurse
									}
									else { // Exprs are equal
										success = true; // no more replacing can be done
									}
								}
							}
							else { // LimitExpr case
								if (nullptr == IVFamily.BIVLimitExpr) {
									assert(0 <= IVLoopIndex);
									success = false;
									SMP_msg("ERROR: null IVFamily.BIVLimitExpr: IVFamily Dump follows.\n");
									DumpInductionVarFamily(IVFamily);
								}
								else {
									// NOTE: There are two cases. If BIVLimitExpr is for the BIV that terminates
									//  the loop, then it will have a relational operator, e.g. RCX < k. If BIVLimitExpr
									//  is for a secondary BIV, it will have the form RSI + k, which means "incoming value
									//  of RSI plus k," where we need the InitExpr for RSI to determine the actual limit.
									if (IsRelationalOperator(IVFamily.BIVLimitExpr->GetOperator())) {
										if (IVFamily.BIVLimitExpr->HasRightSubTree()) {
											STARSExpression *ReplaceExpr = IVFamily.BIVLimitExpr->GetRightTree();

											// The LimitExpr for an IV can be part of the LimitExpr
											//  for a memory access. Avoid infinite recursion by seeing
											//  if we have matching exprs.
											if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
												// We are about to recurse into a subtree of ReplaceExpr, if
												//  it has any. Avoid recursion by seeing if CurrExpr matches
												//  a subtree.
												bool LeftMatch = (ReplaceExpr->HasLeftSubTree() &&
													(!((*(ReplaceExpr->GetLeftTree()) < *CurrExpr) || (*CurrExpr < *(ReplaceExpr->GetLeftTree())))));
												if (LeftMatch) {
													success = true; // further replacing will recurse infinitely
												}
												else {
													STARSExpression *FinalExpr = ReplaceExpr->Clone();
													CurrExpr->SetLeftTree(FinalExpr);
													CurrExpr->SetLeftOperand(nullptr);
													CurrExpr->SetParentInst(FinalExpr->GetParentInst());
													CurrExpr->GetLeftTree()->SetProcessed(); // cut off infinite replacement
													changed = true;
													success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed, RegsReplaced); // recurse
												}
											}
											else { // Exprs are equal
												success = true; // no more replacing can be done
											}
										}
										else { // just a RightOperand, so replace CurrExpr->LeftOperand with BIVLimitExpr->RightOperand
											bool SameOpnds = ((CurrExpr->GetLeftSSANum() == IVFamily.BIVLimitExpr->GetRightSSANum())
												&& IsEqOp(CurrExpr->GetConstLeftOperand(), IVFamily.BIVLimitExpr->GetConstRightOperand()));
											if (!SameOpnds) {
												CurrExpr->SetLeftOperand(IVFamily.BIVLimitExpr->GetRightOperand());
												CurrExpr->SetLeftUseAddr(IVFamily.BIVLimitExpr->GetRightUseAddr());
												CurrExpr->SetLeftSSANum(IVFamily.BIVLimitExpr->GetRightSSANum());
												CurrExpr->SetLeftPreLoopDefAddr(IVFamily.BIVLimitExpr->GetRightPreLoopDefAddr());
												success = true;
												CurrExpr->SetParentInst(IVFamily.BIVLimitExpr->GetParentInst());
												changed = true;
											}
											else { // Opnds are equal
												success = true; // no more replacing can be done
											}
										}
									}
									else {
										// Secondary BIV case. Simplest approach is to get the BIVInitExpr
										//  and use it to replace the left operand as a new left tree.
										STARSExpression *ReplaceExpr = IVFamily.BIVInitExpr;

										if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
											// We are about to recurse into a subtree of ReplaceExpr, if
											//  it has any. Avoid recursion by seeing if CurrExpr matches
											//  a subtree.
											bool LeftMatch = (ReplaceExpr->HasLeftSubTree() && 
												(!((*(ReplaceExpr->GetLeftTree()) < *CurrExpr) || (*CurrExpr < *(ReplaceExpr->GetLeftTree())))));
											if (LeftMatch) {
												success = true; // further replacing will recurse infinitely
											}
											else {
												STARSExpression *FinalExpr = ReplaceExpr->Clone();
												CurrExpr->SetLeftTree(FinalExpr);
												CurrExpr->SetLeftOperand(nullptr);
												CurrExpr->SetParentInst(FinalExpr->GetParentInst());
												CurrExpr->GetLeftTree()->SetProcessed(); // cut off infinite replacement
												changed = true;
												success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed, RegsReplaced); // recurse
											}
										}
										else { // Exprs are equal
											success = true; // no more replacing can be done
										}
									}
								}
							} // end if (InitCase) ... else ...
						}
					}
					else {
						// Alternative: Should we not insist on SSANum match? !!!!****!!!!
						success = false;
						SMP_msg("ERROR: CurrSSANum = %d does not match; IVFamily dump follows. ", CurrSSANum);
						this->DumpFuncNameAndAddr();
						DumpInductionVarFamily(IVFamily);
					}
				}
			}
		}
	}
	else { // left tree, not left operand
		success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed, RegsReplaced); // recurse
	}
	if (!success)
		return success;

	if (!CurrExpr->HasRightSubTree()) {
		const STARSOpndTypePtr CurrOp = CurrExpr->GetConstRightOperand();
		if (MDIsDataFlowOpnd(CurrOp, UseFP)) {
			SMPInstr *UseInst = CurrExpr->GetParentInst();
			int CurrSSANum = CurrExpr->GetRightSSANum();
			int IVLoopIndex = -1;
			if (this->IsLoopNestInductionVar(CurrOp, UseInst, ListIter, FamilyIndex, IVLoopIndex)) {
				if (0 == FamilyIndex) { // BIV, not DIV
					struct InductionVarFamily IVFamily = (*ListIter);
					if ((CurrSSANum == IVFamily.BasicInductionVar.InductionVar.GetSSANum())
						&& IsEqOp(CurrOp, IVFamily.BasicInductionVar.InductionVar.GetOp())) {
						// Matched. Replace CurrExpr->RightOperand with InitExpr or LimitExpr->GetRightTree() or LimitExpr->GetRightOperand().
						//  For LimitExpr, the LeftOperand is compared to to the RightOperand or RightTree, so the right side is the limit.

						// Avoid infinite recursion by seeing if this operand has already been replaced.
						bool AlreadyReplaced = false;
						if (CurrOp->IsRegOp()) {
							int RegHashIndex = HashGlobalNameAndSSA(CurrOp, CurrSSANum);
							if (RegsReplaced.find(RegHashIndex) != RegsReplaced.cend()) {
								success = true;
								AlreadyReplaced = true;
							}
							else {
								(void)RegsReplaced.insert(RegHashIndex);
							}
						}
						if (!AlreadyReplaced) {
							if (InitCase) {
								if (nullptr == IVFamily.BIVInitExpr) {
									assert(0 <= IVLoopIndex);
									success = false;
									SMP_msg("ERROR: null IVFamily.BIVInitExpr: IVFamily Dump follows. ");
									this->DumpFuncNameAndAddr();
									DumpInductionVarFamily(IVFamily);
								}
								else {
									STARSExpression *ReplaceExpr = IVFamily.BIVInitExpr;

									// The InitExpr for an IV can be part of the InitExpr
									//  for a memory access. Avoid infinite recursion by seeing
									//  if we have matching exprs.
									if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
										// We are about to recurse into a subtree of ReplaceExpr, if
										//  it has any. Avoid recursion by seeing if CurrExpr matches
										//  a subtree.
										bool LeftMatch = (ReplaceExpr->HasLeftSubTree() && 
											(!((*(ReplaceExpr->GetLeftTree()) < *CurrExpr) || (*CurrExpr < *(ReplaceExpr->GetLeftTree())))));
										if (LeftMatch) {
											success = true; // further replacing will recurse infinitely
										}
										else {
											STARSExpression *FinalExpr = ReplaceExpr->Clone();
											CurrExpr->SetRightTree(FinalExpr);
											CurrExpr->SetRightOperand(nullptr);
											CurrExpr->SetParentInst(FinalExpr->GetParentInst());
											CurrExpr->GetRightTree()->SetProcessed(); // cut off infinite replacement
											changed = true;
											success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed, RegsReplaced); // recurse
										}
									}
									else { // Exprs are equal
										success = true; // no more replacing can be done
									}
								}
							}
							else if (nullptr == IVFamily.BIVLimitExpr) { // LimitExpr case, safeguard
								assert(0 <= IVLoopIndex);
								success = false;
								SMP_msg("ERROR: null IVFamily.BIVLimitExpr: IVFamily Dump follows. ");
								this->DumpFuncNameAndAddr();
								DumpInductionVarFamily(IVFamily);
							}
							else { // LimitExpr case, safe to proceed
								// NOTE: There are two cases. If BIVLimitExpr is for the BIV that terminates
								//  the loop, then it will have a relational operator, e.g. RCX < k. If BIVLimitExpr
								//  is for a secondary BIV, it will have the form RSI + k, which means "incoming value
								//  of RSI plus k," where we need the InitExpr for RSI to determine the actual limit.
								if (IsRelationalOperator(IVFamily.BIVLimitExpr->GetOperator())) {
									if (IVFamily.BIVLimitExpr->HasRightSubTree()) {
										STARSExpression *ReplaceExpr = IVFamily.BIVLimitExpr->GetRightTree();

										// The LimitExpr for an IV can be part of the LimitExpr
										//  for a memory access. Avoid infinite recursion by seeing
										//  if we have matching exprs.
										if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
											// We are about to recurse into a subtree of ReplaceExpr, if
											//  it has any. Avoid recursion by seeing if CurrExpr matches
											//  a subtree.
											bool LeftMatch = (ReplaceExpr->HasLeftSubTree() &&
												(!((*(ReplaceExpr->GetLeftTree()) < *CurrExpr) || (*CurrExpr < *(ReplaceExpr->GetLeftTree())))));
											if (LeftMatch) {
												success = true; // further replacing will recurse infinitely
											}
											else {
												STARSExpression *FinalExpr = ReplaceExpr->Clone();
												CurrExpr->SetRightTree(FinalExpr);
												CurrExpr->SetRightOperand(nullptr);
												CurrExpr->SetParentInst(FinalExpr->GetParentInst());
												CurrExpr->GetRightTree()->SetProcessed(); // cut off infinite replacement
												changed = true;
												success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed, RegsReplaced); // recurse
											}
										}
										else { // Exprs are equal
											success = true; // no more replacing can be done
										}
									}
									else { // just a RightOperand, so replace CurrExpr->RightOperand with BIVLimitExpr->RightOperand
										bool SameOpnds = ((CurrExpr->GetRightSSANum() == IVFamily.BIVLimitExpr->GetRightSSANum())
											&& IsEqOp(CurrExpr->GetConstRightOperand(), IVFamily.BIVLimitExpr->GetConstRightOperand()));
										if (!SameOpnds) {
											CurrExpr->SetRightOperand(IVFamily.BIVLimitExpr->GetRightOperand());
											CurrExpr->SetRightUseAddr(IVFamily.BIVLimitExpr->GetRightUseAddr());
											CurrExpr->SetRightSSANum(IVFamily.BIVLimitExpr->GetRightSSANum());
											CurrExpr->SetRightPreLoopDefAddr(IVFamily.BIVLimitExpr->GetRightPreLoopDefAddr());
											CurrExpr->SetParentInst(IVFamily.BIVLimitExpr->GetParentInst());
											changed = true;
											success = true;
										}
										else { // Opnds are equal
											success = true; // no more replacing can be done
										}
									}
								}
								else {
									// Secondary BIV case. Simplest approach is to get the BIVInitExpr
									//  and use it to replace the right operand as a new right tree.
									STARSExpression *ReplaceExpr = IVFamily.BIVInitExpr;

									// The InitExpr for an IV can be part of the InitExpr
									//  for a memory access. Avoid infinite recursion by seeing
									//  if we have matching exprs.
									if (!CurrExpr->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
										// We are about to recurse into a subtree of ReplaceExpr, if
										//  it has any. Avoid recursion by seeing if CurrExpr matches
										//  a subtree.
										bool LeftMatch = (ReplaceExpr->HasLeftSubTree() &&
											(!((*(ReplaceExpr->GetLeftTree()) < *CurrExpr) || (*CurrExpr < *(ReplaceExpr->GetLeftTree())))));
										if (LeftMatch) {
											success = true; // further replacing will recurse infinitely
										}
										else {
											STARSExpression *FinalExpr = ReplaceExpr->Clone();
											CurrExpr->SetRightTree(FinalExpr);
											CurrExpr->SetRightOperand(nullptr);
											CurrExpr->SetParentInst(FinalExpr->GetParentInst());
											CurrExpr->GetRightTree()->SetProcessed(); // cut off infinite replacement
											changed = true;
											success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed, RegsReplaced); // recurse
										}
									}
									else { // Exprs are equal
										success = true; // no more replacing can be done
									}
								}
							} // end if (InitCase) ... else ...
						}
					}
					else {
						// Alternative: Should we not insist on SSANum match? !!!!****!!!!
						success = false;
						SMP_msg("ERROR: CurrSSANum = %d does not match; IVFamily dump follows. ");
						this->DumpFuncNameAndAddr();
						DumpInductionVarFamily(IVFamily);
					}
				}
			}
		}
	}
	else { // right tree, not right operand
		success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed, RegsReplaced); // recurse
	}

	return success;
} // end of SMPFunction::ReplaceAllBIVsWithExprs()

// Replace all DIVs in CurrExpr with lower or upper limit (depending on InitCase) exprs for DIV
bool SMPFunction::ReplaceAllDIVsWithExprs(std::size_t LoopIndex, STARSExpression *CurrExpr, bool InitCase, bool &changed) {
	bool success = true;
	bool UseFP = this->UsesFramePointer();
	size_t FamilyIndex;
	STARSInductionVarFamilyIter ListIter;

	// Traverse CurrExpr to find the induction vars.
	if (!CurrExpr->HasLeftSubTree()) {
		STARSOpndTypePtr CurrOp = CurrExpr->GetLeftOperand();
		if (MDIsDataFlowOpnd(CurrOp, UseFP)) {
			SMPInstr *UseInst = CurrExpr->GetParentInst();
			int CurrSSANum = CurrExpr->GetLeftSSANum();
			int IVLoopIndex = -1;
			if (this->IsLoopNestInductionVar(CurrOp, UseInst, ListIter, FamilyIndex, IVLoopIndex)) {
				if (0 < FamilyIndex) { // DIV, not BIV
					struct InductionVarFamily IVFamily = (*ListIter);
					size_t DIVIndex = (FamilyIndex - 1);
					if (CurrSSANum == IVFamily.DependentInductionVars[DIVIndex].DIV.GetSSANum()) {
						// Matched. Replace CurrExpr->LeftOperand with InitExpr or LimitExpr->GetRightTree()
						//  from the DIV matched.
						//  For LimitExpr, the LeftOperand is compared to the RightOperand or RightTree, so the right side is the limit.
						if (InitCase) {
							if (nullptr == IVFamily.DependentInductionVars[DIVIndex].DIVInitExpr) {
								assert(0 <= IVLoopIndex);
								success = false;
							}
							else {
								STARSExpression *ReplaceExpr = IVFamily.DependentInductionVars[DIVIndex].DIVInitExpr;

								if (!CurrExpr->GetLeftTree()->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
									STARSExpression *FinalExpr = ReplaceExpr->Clone();
									CurrExpr->SetLeftTree(FinalExpr);
									CurrExpr->SetLeftOperand(nullptr);
									CurrExpr->SetParentInst(FinalExpr->GetParentInst());
									CurrExpr->GetLeftTree()->SetProcessed(); // cut off infinite replacement
									changed = true;
									// Recurse; could have DIV dependent on another DIV.
									success = this->ReplaceAllIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed); // recurse
								}
								else { // Exprs are equal
									success = true; // no more replacing can be done
								}
							}
						}
						else { // LimitExpr case
							if (nullptr == IVFamily.DependentInductionVars[DIVIndex].DIVLimitExpr) {
								assert(0 <= IVLoopIndex);
								success = false;
							}
							else {
								STARSExpression *ReplaceExpr = IVFamily.DependentInductionVars[DIVIndex].DIVLimitExpr;

								if (!CurrExpr->GetLeftTree()->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
									STARSExpression *FinalExpr = ReplaceExpr->Clone();
									CurrExpr->SetLeftTree(FinalExpr);
									CurrExpr->SetLeftOperand(nullptr);
									CurrExpr->SetParentInst(FinalExpr->GetParentInst());
									CurrExpr->GetLeftTree()->SetProcessed(); // cut off infinite replacement
									changed = true;
									// Recurse; could have DIV dependent on another DIV.
									success = this->ReplaceAllIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed); // recurse
								}
								else { // Exprs are equal
									success = true; // no more replacing can be done
								}
							}
						} // end if (InitCase) ... else ...
					}
					else {
						// Alternative: Should we not insist on SSANum match? !!!!****!!!!
						success = false;
					}
				}
			}
		}
	}
	else {
		success = this->ReplaceAllDIVsWithExprs(LoopIndex, CurrExpr->GetLeftTree(), InitCase, changed); // recurse
	}
	if (!success)
		return success;

	if (!CurrExpr->HasRightSubTree()) {
		const STARSOpndTypePtr CurrOp = CurrExpr->GetConstRightOperand();
		if (MDIsDataFlowOpnd(CurrOp, UseFP)) {
			SMPInstr *UseInst = CurrExpr->GetParentInst();
			int CurrSSANum = CurrExpr->GetRightSSANum();
			int IVLoopIndex = -1;
			if (this->IsLoopNestInductionVar(CurrOp, UseInst, ListIter, FamilyIndex, IVLoopIndex)) {
				if (0 < FamilyIndex) { // DIV, not BIV
					struct InductionVarFamily IVFamily = (*ListIter);
					size_t DIVIndex = (FamilyIndex - 1);
					if (CurrSSANum == IVFamily.DependentInductionVars[DIVIndex].DIV.GetSSANum()) {
						// Matched. Replace CurrExpr->RightOperand with InitExpr or LimitExpr->GetRightTree() or LimitExpr->GetRightOperand().
						//  For LimitExpr, the LeftOperand is compared to to the RightOperand or RightTree, so the right side is the limit.
						if (InitCase) {
							if (nullptr == IVFamily.DependentInductionVars[DIVIndex].DIVInitExpr) {
								assert(0 <= IVLoopIndex);
								success = false;
							}
							else {
								STARSExpression *ReplaceExpr = IVFamily.DependentInductionVars[DIVIndex].DIVInitExpr;

								if (!CurrExpr->GetRightTree()->IsProcessed() &&  ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
									STARSExpression *FinalExpr = ReplaceExpr->Clone();
									CurrExpr->SetRightTree(FinalExpr);
									CurrExpr->SetRightOperand(nullptr);
									CurrExpr->SetParentInst(FinalExpr->GetParentInst());
									CurrExpr->GetRightTree()->SetProcessed(); // cut off infinite replacement
									changed = true;
									// Recurse; could have DIV dependent on another DIV.
									success = this->ReplaceAllIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed); // recurse
								}
								else { // Exprs are equal
									success = true; // no more replacing can be done
								}
							}
						}
						else if (nullptr == IVFamily.DependentInductionVars[DIVIndex].DIVLimitExpr) { // LimitExpr case, safeguard
							assert(0 <= IVLoopIndex);
							success = false;
						}
						else { // LimitExpr case, safe to proceed
							STARSExpression *ReplaceExpr = IVFamily.DependentInductionVars[DIVIndex].DIVLimitExpr;

							if (!CurrExpr->GetRightTree()->IsProcessed() && ((*ReplaceExpr < *CurrExpr) || (*CurrExpr < *ReplaceExpr))) {
								STARSExpression *FinalExpr = ReplaceExpr->Clone();
								CurrExpr->SetRightTree(FinalExpr);
								CurrExpr->SetRightOperand(nullptr);
								CurrExpr->SetParentInst(FinalExpr->GetParentInst());
								CurrExpr->GetRightTree()->SetProcessed(); // cut off infinite replacement
								changed = true;
								// Recurse; could have DIV dependent on another DIV.
								success = this->ReplaceAllIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed); // recurse
							}
							else { // Exprs are equal
								success = true; // no more replacing can be done
							}
						} // end if (InitCase) ... else ...
					}
					else {
						// Alternative: Should we not insist on SSANum match? !!!!****!!!!
						success = false;
					}
				}
			}
		}
	}
	else {
		success = this->ReplaceAllDIVsWithExprs(LoopIndex, CurrExpr->GetRightTree(), InitCase, changed); // recurse
	}

	return success;
} // end of SMPFunction::ReplaceAllDIVsWithExprs()

// wrapper to call ReplaceAllBIVsWithExprs() and ReplaceAllDIVsWithExprs()
bool SMPFunction::ReplaceAllIVsWithExprs(std::size_t LoopIndex, STARSExpression *CurrExpr, bool InitCase, bool &changed) {
	set<int> RegsReplaced;
	bool success = this->ReplaceAllBIVsWithExprs(LoopIndex, CurrExpr, InitCase, changed, RegsReplaced);
	if (!success)
		return success;
	success = this->ReplaceAllDIVsWithExprs(LoopIndex, CurrExpr, InitCase, changed); 

	return success;
} // end of SMPFunction::ReplaceAllIVsWithExprs()

// expand, replace IVs, simplify, until no more can be done
bool SMPFunction::ExpandExprToInArgs(std::size_t LoopIndex, STARSExpression *CurrExpr, bool InitCase, set<STARS_ea_t> &StackPtrCopySet) {
	bool success = true;
	bool changed = true;
	bool StoppedOnIV = true;
#if 0
	STARS_ea_t OrigUseAddr = CurrExpr->GetOriginalParentInst()->GetAddr();
#else
	STARS_ea_t OrigUseAddr = CurrExpr->GetParentInst()->GetAddr();
#endif
	set<int> LoopRegHashes;
	uint32_t IterationCount = 0;
	CurrExpr->ResetAllProcessed(); // avoid redundant or infinite work when replacing with induction vars
	while (success && changed && StoppedOnIV) {
		changed = false;
		int DepthCounter = 0;
		success = CurrExpr->ExpandExpr(OrigUseAddr, LoopIndex, false, true, false, false, InitCase, LoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter);
		if (success && StoppedOnIV) {
			success = this->ReplaceAllIVsWithExprs(LoopIndex, CurrExpr, InitCase, changed);
			if (!success) {
				SMP_msg("ERROR: LOOP: ReplaceAllIVs failure in ExpandExprToInArgs, InitCase %d iteration %u for UseAddr %llx loop %zu ",
					InitCase, IterationCount, (uint64_t)OrigUseAddr, LoopIndex);
				this->DumpFuncNameAndAddr();
			}
		}
		else {
			SMP_msg("ERROR: LOOP: ExpandExpr failure in ExpandExprToInArgs, InitCase %d iteration %u for UseAddr %llx loop %zu ",
				InitCase, IterationCount, (uint64_t) OrigUseAddr, LoopIndex);
			this->DumpFuncNameAndAddr();
		}
		if (success) {
			CurrExpr->EvaluateConsts();
			bool changed2 = CurrExpr->SimplifyDriver();
			changed = (changed || changed2);
		}
		++IterationCount;
	}

	return success;
} // end of SMPFunction::ExpandExprToInArgs()

// Fill in LoopMemRangeInArgRegsBitmap and MemRangeRegsBitmap and InArgsUsedInMemWrites
void SMPFunction::DetectInArgRegsNeededForMemWriteExprs(void) {
	bool UseFP = this->UsesFramePointer();
	bitset<1 + MD_LAST_REG_NO> CurrArgRegsUsed;
	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		STARSExprBoundsIter ExprIter;
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			if ((nullptr != LowerExpr) && (nullptr != UpperExpr)) {
				STARSExpression *CurrExpr = UpperExpr;
				if (nullptr != CurrExpr) {
					CurrExpr->ListInArgRegsUsed(CurrArgRegsUsed);
					CurrExpr = LowerExpr;
					assert(nullptr != CurrExpr);
					CurrExpr->ListInArgRegsUsed(CurrArgRegsUsed);
					this->MemRangeRegsBitmap |= CurrArgRegsUsed;
					this->LoopMemRangeInArgRegsBitmap[LoopIndex] |= CurrArgRegsUsed;
				}
			}
		} // end for all loop mem write exprs in current loop
	}

	// Look at all memory writes and find loop-invariant non-stack indirect writes.
	//  See which ones can be traced back to InArgs.
	// Also look at calls, and see which callees have MemExprs that trace back to InArgs that we pass to the callee.
	//  See if any of those MemExprs can be traced back to our own InArgs.
	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
		if (CurrBlock->HasIndirectMemWrite()) {
			// Ensure that mem write address regs are loop-invariant w.r.t. outermost loop containing the write
			int OuterLoopIndex = this->GetOutermostLoopNum((int) BlockIndex);
			// OuterLoopIndex could be -1 if CurrInst is not in a loop
			bool NotInLoop = (0 > OuterLoopIndex);
			int InnerLoopIndex = NotInLoop ? -1 : this->GetInnermostLoopNum((int) BlockIndex);
			size_t InnerLoopNumPlusOne = (size_t)(InnerLoopIndex + 1);
			list<size_t> LoopList; // fill on demand
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				// NOTE: Handle looping string opcodes later.
				if (CurrInst->HasIndirectMemoryWrite() && (!CurrInst->MDIsPrefetchOpcode()) 
					&& (!(CurrInst->MDIsPossibleStringLoopingOpcode() && CurrInst->MDHasAnyRepeatPrefix()))) {
					STARSOpndTypePtr DefOp = CurrInst->GetMemDef();
					assert(DefOp->IsMemOp());
					if (!MDIsDirectStackAccessOpnd(DefOp, UseFP)) {
						STARSDefUseSet AddressRegs;
						bool success = CurrInst->GetAddressRegs(AddressRegs);
						if (success && (1 == AddressRegs.size())) { // limit to one address reg for now
							STARS_ea_t MemWriteAddr = CurrInst->GetAddr();
#if 0						// Get rid of CurrInst->CreateMemDefAddrExpr()
							STARSExpression *DefExpr = CurrInst->CreateMemDefAddrExpr();
#else
							STARSExpression *DefExpr = this->CreateMemoryAddressExpr(DefOp, CurrInst);
#endif
							if (nullptr != DefExpr) {
								bool LoopInvariantDef = ((!NotInLoop) && this->IsUseLoopInvariantDEF(OuterLoopIndex, DefOp, CurrInst));
								bool StoppedOnIV = false;
								size_t DefWidth = (size_t) DefOp->GetByteWidth();
								if (NotInLoop || LoopInvariantDef) {
									set<int> DummyLoopRegHashes;
									set<STARS_ea_t> StackPtrCopySet;
									bool changed = false;
									int DepthCounter = 0;
									if (DefExpr->ExpandExpr(MemWriteAddr, OuterLoopIndex, false, true, false, false, false, DummyLoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
										if (!StoppedOnIV) {
											// We want a single node, an incoming arg.
											if (!DefExpr->HasLeftSubTree() && (!DefExpr->HasRightSubTree()) && (0 == DefExpr->GetLeftSSANum())) {
												STARSOpndTypePtr LeftOp = DefExpr->GetLeftOperand();
												if (MDIsDataFlowOpnd(LeftOp, UseFP)) {
													// Record in InArgsUsedInMemWrites
													pair<STARSExprSetIter, bool> InsertResult = this->InArgsUsedInMemWrites[InnerLoopNumPlusOne].insert(DefExpr);
													STARSOpndTypePtr AddrRegOp = AddressRegs.cbegin()->GetOp();
													if (InsertResult.second) {
														pair<uint32_t, size_t> InsertValue(DefExpr->GetSerialNumber(), DefWidth);
														this->InArgsUsedInMemWriteByteWidths[InnerLoopNumPlusOne].push_back(InsertValue);
													}
													// Even if this is a duplicate expr, make sure we map the InArg reg to the LeftOp reg.
													//  e.g. [rbx+16] and [rbp+8] could both expand to the same DefExpr [RDI+48]. Both RBX
													//  and RBP need to be mapped to RDI for assertions at the MemWriteAddr.
													// Save mapping of mem address reg operand to InArg operand
													if (NotInLoop) { // loop examples need to pass saved InArg arguments
														STARSInArgMapping InsertVal(MemWriteAddr, STARSOpndPair(AddrRegOp, LeftOp));
														pair<STARSInArgMap::iterator, bool> InsertResult = this->MemWriteToInArgMap.insert(InsertVal);
														assert(InsertResult.second);
													}
													else if (AddrRegOp->IsRegOp()) {
														// Update InArgs needed by loop
														if (LoopList.empty()) {
															this->BuildLoopList((int) BlockIndex, LoopList);
														}
														STARS_regnum_t InArgRegNum = AddrRegOp->GetReg();
														assert(!LoopList.empty());
														for (list<size_t>::const_iterator LoopIter = LoopList.cbegin(); LoopIter != LoopList.cend(); ++LoopIter) {
															size_t CurrLoopNum = (*LoopIter);
															this->LoopMemRangeInArgRegsBitmap[CurrLoopNum].set(InArgRegNum);
														}
													}
												}
											}
										}
										else { // StoppedOnIV
											this->SymbolicAnalysisProblems[InnerLoopNumPlusOne] = true;
											SMP_msg("ERROR: SPARK: Expr Expand() StoppedOnIV in DetectInArgRegsNeededForMemWriteExprs() at %llx ",
												(uint64_t) MemWriteAddr);
											this->DumpFuncNameAndAddr();
											pair<STARSExprSetIter, bool> InsertResult = this->StoppedOnIVNonRangeExprs[InnerLoopNumPlusOne].insert(DefExpr);
											if (InsertResult.second) { // new Expr
												pair<STARSExprSetIter, size_t> InsertValue(InsertResult.first, DefWidth);
												this->StoppedOnIVNonRangeIterWidths[InnerLoopNumPlusOne].push_back(InsertValue);
											}
										}
									}
									else {
										this->SymbolicAnalysisProblems[InnerLoopNumPlusOne] = true;
										SMP_msg("ERROR: ExpandExpr() failure in DetectInArgRegsNeededForMemWriteExprs() at %llx ",
											(uint64_t) MemWriteAddr);
										this->DumpFuncNameAndAddr();
									}
								}
							}
							else {
								SMP_msg("ERROR: CreateMemDefAddrExpr() failure in DetectInArgRegsNeededForMemWriteExprs() at %llx ",
									(uint64_t) MemWriteAddr);
								this->DumpFuncNameAndAddr();
							}
						}
					} // end if not direct stack access operand
				}
				// Build looping string mem writing exprs
				if (CurrInst->MDHasAnyRepeatPrefix() && CurrInst->MDIsPossibleStringLoopingOpcode()) {
					this->BuildLoopingStringMemExprs(CurrBlock, CurrInst);
				}
			} // end for all insts in CurrBlock
		}
	} // end for all blocks in func

	return;
} // end of SMPFunction::DetectInArgRegsNeededForMemWriteExprs()

// take LoopRegHashSets and expand into LoopMemAddrRegSourceExprs
void SMPFunction::ExpandLoopRegHashExprs(void) {
	for (size_t LoopNum = 0; LoopNum < this->GetNumLoops(); ++LoopNum) {
		if (!this->LoopRegHashSets[LoopNum].empty()) {
			// For each reg hash, find the DefExpr, expand it, record in LoopMemAddrRegSourceExprs.
			for (set<int>::const_iterator HashIter = this->LoopRegHashSets[LoopNum].cbegin(); HashIter != this->LoopRegHashSets[LoopNum].cend(); ++HashIter) {
				int DefHashValue = (*HashIter);
				STARS_regnum_t DefRegNo = (STARS_regnum_t) ExtractRegFromHash(DefHashValue);
				STARS_ea_t DefAddr = this->GetGlobalDefAddrForRegHash(DefHashValue);
				if ((STARS_BADADDR == DefAddr) || STARS_IsBlockNumPseudoID(DefAddr) || STARS_IsLiveInPseudoID(DefAddr)) {
					SMP_msg("ERROR: SPARK: ExpandLoopRegHashExprs() found DefAddr %llx for DefHashValue %llx ", (uint64_t) DefAddr, (uint64_t) DefHashValue);
					this->DumpFuncNameAndAddr();
					this->SymbolicAnalysisProblems[LoopNum + 1] = true;
				}
				else if (STARS_IsSSAMarkerPseudoID(DefAddr)) {
					// No DefInst expr to create; just an incoming reg value.
					// Create a DefExpr from scratch.
					SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
					assert(nullptr != DefInst);
					STARSOpndTypePtr DefOp = DefInst->MakeRegOpnd(DefRegNo);
					STARSExpression *DefExpr = new STARSExpression();
					DefExpr->SetParentFunc(this);
					DefExpr->SetParentInst(DefInst);
					DefExpr->SetOriginalParentInst(DefInst);
					DefExpr->SetLeftOperand(DefOp);
					DefExpr->SetLeftUseAddr(DefAddr);
					DefExpr->SetOperator(SMP_ASSIGN);
					DefExpr->SetLeftPreLoopDefAddr(DefAddr);
					DefExpr->SetLeftSSANum(0);
					pair<STARSExprSetIter, bool> InsertResult = LoopMemAddrRegSourceExprs.insert(DefExpr);
					if (InsertResult.second) { // new expr
						pair<STARS_regnum_t, STARSExprSetIter> InsertValue(DefRegNo, InsertResult.first);
						this->LoopRegSourceExprPairs[LoopNum].push_back(InsertValue);
						SMP_msg("INFO: SPARK: ExpandLoopRegHashExprs() succeeded at DefAddr %llx for DefHashValue %llx\n", (uint64_t)DefAddr, (uint64_t)DefHashValue);
					}
				}
				else { // found a regular instruction at DefAddr.
					SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
					assert(nullptr != DefInst);
					STARSOpndTypePtr DefOp = DefInst->MakeRegOpnd(DefRegNo);
					SMPRegTransfer *DefRT = DefInst->GetDefRT(DefOp);
					if (nullptr == DefRT) {
						SMP_msg("ERROR: SPARK: ExpandLoopRegHashExprs() did not find DefRT at DefAddr %llx for DefHashValue %llx ", (uint64_t) DefAddr, (uint64_t) DefHashValue);
						this->DumpFuncNameAndAddr();
						this->SymbolicAnalysisProblems[LoopNum + 1] = true;
					}
					else {
						// Create an expr from the DefRT and expand it.
						STARSExpression *DefExpr = new STARSExpression(DefRT);
						set<int> DummyLoopRegHashes;
						bool StoppedOnIV = false;
						bool changed = false;
						set<STARS_ea_t> StackPtrCopySet;
						int DepthCounter = 0;
						if (DefExpr->ExpandExpr(DefAddr, LoopNum, false, true, false, false, false, DummyLoopRegHashes, StoppedOnIV, changed, StackPtrCopySet, DepthCounter)) {
							if (!StoppedOnIV) {
								DefExpr->EvaluateConsts();
								(void) DefExpr->SimplifyDriver();
								pair<STARSExprSetIter, bool> InsertResult = LoopMemAddrRegSourceExprs.insert(DefExpr);
								if (InsertResult.second) { // new expr
									pair<STARS_regnum_t, STARSExprSetIter> InsertValue(DefRegNo, InsertResult.first);
									this->LoopRegSourceExprPairs[LoopNum].push_back(InsertValue);
									SMP_msg("INFO: SPARK: ExpandLoopRegHashExprs() succeeded at DefAddr %llx for DefHashValue %llx ", (uint64_t) DefAddr, (uint64_t) DefHashValue);
									this->DumpFuncNameAndAddr();
								}
							}
							else {
								SMP_msg("ERROR: SPARK: ExpandLoopRegHashExprs() StoppedOnIV for DefAddr %llx for DefHashValue %llx ", (uint64_t)DefAddr, (uint64_t)DefHashValue);
								this->DumpFuncNameAndAddr();
								this->SymbolicAnalysisProblems[LoopNum + 1] = true;
							}
						}
						else {
							SMP_msg("ERROR: SPARK: ExpandLoopRegHashExprs() failed ExpandExpr for DefAddr %llx for DefHashValue %llx ", (uint64_t) DefAddr, (uint64_t) DefHashValue);
							this->DumpFuncNameAndAddr();
							this->SymbolicAnalysisProblems[LoopNum + 1] = true;
						}
					}
				}
			} // end for all HashIters in current set
		} // end if current hashed reg set is not empty
	} // end for all LoopNum values
	return;
} // end of SMPFunction::ExpandLoopRegHashExprs()

// Analyze switch starting at indir jump at end of CurrBlock, or go past default jumps to indir jump; return false if not well-structured
bool SMPFunction::AnalyzeSwitchStatement(STARSTransCFGBlock *CurrTCFGBlock) {
	bool StructuredSwitch = true;

	SMPInstr *LastInst = (*(CurrTCFGBlock->GetRevInstCBegin()));
	STARS_ea_t LastAddr = LastInst->GetAddr();
	if (INDIR_JUMP != LastInst->GetDataFlowType()) {
		StructuredSwitch = false;
		SMP_msg("WARNING: Unstructured switch at %llx due to lack of INDIR_JUMP.\n", (uint64_t) LastAddr);
	}
	else { // populate the switch data structures and mark control flow types for switch-related jumps
		int OrigBlockNum = CurrTCFGBlock->GetOriginalBlockNum();
		int InnermostLoopNum = this->GetInnermostLoopNum(OrigBlockNum);
		size_t SwitchIndex = this->SwitchJumpMap.size(); // first index is zero, etc.
		assert(this->SwitchInfoArray.size() == this->SwitchJumpMap.size());
		struct SwitchTableInfo TableInfo;
		bool success = LastInst->AnalyzeSwitchInfo(TableInfo);
		if (!success) {
			StructuredSwitch = false;
			SMP_msg("WARNING: Unstructured switch at %llx due to SMPInstr::AnalyzeSwitchInfo failure.\n",
				(uint64_t) LastAddr);
		}
		else {
			// Mark the jumps to the default case and find the dominator node.
			this->FindSwitchIDom(TableInfo);

			// Find consistent follow block for all cases.
			if (this->FindSwitchStatementFollowBlock(TableInfo)) {

				// Search for a switch case that is outside the loop containing the switch statement.
				//  If found, this is an unstructured switch statement.
				if (0 <= InnermostLoopNum) {
					for (int BlockNum : TableInfo.CaseBlockNums) {
						if (!this->IsBlockInLoop(BlockNum, (size_t)InnermostLoopNum)) {
							StructuredSwitch = false;
							SMP_msg("ERROR: SPARK: Switch case block %d lies outside loop %d ", BlockNum, InnermostLoopNum);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							return StructuredSwitch;
						}
					}
				}

				// Map current INDIR_JUMP instruction address to the SwitchIndex,
				pair<STARS_ea_t, size_t> JumpMapItem(LastInst->GetAddr(), SwitchIndex);
				pair<map<STARS_ea_t, size_t>::iterator, bool> InsertResult = this->SwitchJumpMap.insert(JumpMapItem);
				assert(InsertResult.second);

				// Add TableInfo to the SwitchInfoArray.
				this->SwitchInfoArray.push_back(TableInfo);
				assert(this->SwitchInfoArray.size() == this->SwitchJumpMap.size());  // still in sync

				// Mark the jumps from the end of each case to the follow node.
				if (TableInfo.FollowNodeNum >= 0) {
					SMPBasicBlock *FollowBlock = this->GetBlockByNum((size_t)TableInfo.FollowNodeNum);
					// Every predecessor of FollowBlock that ends with a jump should have the jump marked
					//  as a CASE_BREAK_TO_FOLLOW_NODE. This will catch the default case as well.
					//  NOTE: Limit to current switch statement, in case the FollowBlock is also the follow
					//  block for an enclosing statement, e.g. if-then-case-endcase-endif with no code between
					//  the endcase-endif.
					for (list<SMPBasicBlock *>::iterator PredIter = FollowBlock->GetFirstPred(); PredIter != FollowBlock->GetLastPred(); ++PredIter) {
						SMPBasicBlock *PredBlock = (*PredIter);
						int PredBlockNum = PredBlock->GetNumber();
						if (PredBlockNum < TableInfo.IDomBlockNum) { // PredBlock precedes switch
							continue;
						}
						vector<SMPInstr *>::iterator LastInstIter = --(PredBlock->GetLastInst());
						SMPInstr *LastInst = (*LastInstIter);
						STARS_ea_t LastAddr = LastInst->GetAddr();
						SMPitype FlowType = LastInst->GetDataFlowType();
						if (JUMP == FlowType) {
							this->SetControlFlowType(LastAddr, CASE_BREAK_TO_FOLLOW_NODE);
						}
						else if ((DEFAULT == FlowType) || (CALL == FlowType)) { // fall-through to follow node could be optimization for final case
							;
						}
						else { // something odd here
							SMP_msg("ERROR: Inst at %llx should jump or fall through to switch follow node ", (uint64_t)LastAddr);
							this->DumpFuncNameAndAddr();
						}
					}
				}
				else {
					assert(SMP_BLOCKNUM_COMMON_RETURN == TableInfo.FollowNodeNum);
				}
				// Mark jumps around the default case to the INDIR_JUMP.
				if (TableInfo.DefaultJumpAddr != STARS_BADADDR) { // we have a default case
					for (STARSTransCFGBlock *PredBlock : CurrTCFGBlock->GetConstPredList()) {
						SMPInstr *LastPredInst = (*(PredBlock->GetRevInstCBegin()));
						STARS_ea_t LastPredAddr = LastPredInst->GetAddr();
						SMPitype FlowType = LastPredInst->GetDataFlowType();
						ControlFlowType LastPredCFType = this->GetControlFlowType(LastPredAddr);
						if ((LastPredCFType == FALL_THROUGH) &&((FlowType == JUMP) || (FlowType == COND_BRANCH))) {
							// FALL_THROUGH means the type has not been marked yet, e.g. not a LOOP_EXIT.
							// if-then-else marking has not happened yet (ordering is important).
							// Don't want to mark simple if-else jumps, so see if we are jumping around default case blocks
							this->TCFGDomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
							if (this->DoesTCFGBlockDominateTCFGBlock(TableInfo.IDomBlockNum, PredBlock->GetNumber())) {
								this->SetControlFlowType(LastAddr, JUMP_TO_SWITCH_INDIR_JUMP);
								if (FlowType == COND_BRANCH) {
									// Poorly structured. We could have if (cond) switch {...} but then the if-block
									//  would not be dominated by the INDIR_JUMP block, which is TableInfo.IDomBlockNum for
									//  the no-default-case switches.
									StructuredSwitch = false;
									break;
								}
							}
						}
					}
				}
				if (TableInfo.DefaultCaseBlockNum == TableInfo.FollowNodeNum) {
					StructuredSwitch = false;
					SMP_msg("ERROR: Unstructured switch at %llx due to DefaultCaseBlockNum %d == FollowNodeNum ", (uint64_t)LastAddr, TableInfo.FollowNodeNum);
					this->DumpFuncNameAndAddr();
					this->DumpDotCFG();
				}
			}
			else {
				StructuredSwitch = false;
				SMP_msg("WARNING: Unstructured switch at %llx due to SMPFunction::FindSwitchStatementFollowBlock failure.\n",
					(uint64_t) LastAddr);
			}
		}
	}

	if (StructuredSwitch) {
		SMP_msg("INFO: Structured switch at %llx\n", (uint64_t) LastInst->GetAddr());
	}
	else {
		SMP_msg("INFO: Unstructured switch at %llx\n", (uint64_t) LastInst->GetAddr());
	}

	return StructuredSwitch;
} // end of SMPFunction::AnalyzeSwitchStatement()

// Mark jumps to default case, find IDom of entire switch statement
void SMPFunction::FindSwitchIDom(struct SwitchTableInfo &TableInfo) {
	// The TableInfo.IndirJumpBlockNum gives the IDom of all non-default cases.
	//  We need to find the IDom of all blocks that jump to the default case, if any, and
	//  then the IDom of {that IDom and the indir jump block}.
	if (TableInfo.DefaultJumpAddr == STARS_BADADDR) {
		// Simplest case: No default cases, no prior blocks jump to a default case, so INDIR_JUMP block is IDom
		//  for the whole switch statement.
		TableInfo.IDomBlockNum = TableInfo.IndirJumpBlockNum;
	}
	else {
		// Find the blocks that jump to the default block. Mark the jumps while computing the IDom over all of them.
		SMPBasicBlock *DefaultCaseBlock = this->GetBlockByNum((size_t) TableInfo.DefaultCaseBlockNum);
		assert(nullptr != DefaultCaseBlock);
		list<SMPBasicBlock *>::iterator PredIter = DefaultCaseBlock->GetFirstPred();
		assert(PredIter != DefaultCaseBlock->GetLastPred());
		int IDom = (int) this->RPOBlocks.size(); // Find lowest RPO number, init to max
		while (PredIter != DefaultCaseBlock->GetLastPred()) {
			SMPBasicBlock *PredBlock = (*PredIter);
			SMPInstr *LastInst = (*(--PredBlock->GetLastInst()));
			assert(nullptr != LastInst);
			SMPitype FlowType = LastInst->GetDataFlowType();
			STARS_ea_t LastAddr = LastInst->GetAddr();
			if (JUMP == FlowType) {
				this->SetControlFlowType(LastAddr, JUMP_TO_DEFAULT_CASE);
			}
			else if (COND_BRANCH == FlowType) {
				// Detect the inverted case, in which a COND_BRANCH falls through to the default case
				//  and branches to the indirect jump block.
				int TargetBlockNum = PredBlock->GetCondNonFallThroughSuccBlockNum();
				if (TableInfo.DefaultCaseBlockNum == TargetBlockNum) {
					this->SetControlFlowType(LastAddr, JUMP_TO_DEFAULT_CASE);
				}
				else if (TableInfo.IndirJumpBlockNum == TargetBlockNum) {
					this->SetControlFlowType(LastAddr, JUMP_TO_SWITCH_INDIR_JUMP);
				}
				else {
					// Not sure what is going on.
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("ERROR: SPARK: Unstructured due to switch control flow at %llx ", (uint64_t)LastAddr);
						this->DumpFuncNameAndAddr();
						this->PrintedSPARKUnstructuredMsg = true;
					}
					this->SetHasUnstructuredCFG(); // Bad switch control flow
				}
			}
			else if ((INDIR_JUMP == FlowType) && (PredBlock->GetNumber() == TableInfo.IndirJumpBlockNum)) {
				// skip over switch table jumps to the default case
				++PredIter;
				continue;
			}
			int CurrNum = PredBlock->GetNumber();
			assert(SMP_BLOCKNUM_UNINIT != CurrNum);
			assert(CurrNum < (int)this->IDom.size());
			int CurrIDom = this->IDom[CurrNum];
			IDom = (IDom > CurrIDom) ? CurrIDom : IDom; // keep running minimum RPO IDom number
			++PredIter;
		}

		// Blocks that jump to the default case should precede and dominate the INDIR_JUMP block, but
		//  we will ensure safety here.
		IDom = (IDom > TableInfo.IndirJumpBlockNum) ? TableInfo.IndirJumpBlockNum : IDom; // keep running minimum RPO IDom number
		assert(SMP_BLOCKNUM_UNINIT != IDom);
		TableInfo.IDomBlockNum = IDom;
	}
	return;
} // end of SMPFunction::FindSwitchIDom()

// build skeleton CFG, then use in coalescing nodes to analyze expressions
void SMPFunction::BuildShadowCFG(void) {
	// First, just get block numbers into the vector entries.
	for (size_t BlockIndex = 0; BlockIndex < this->TCFGBlocks.size(); ++BlockIndex) {
		STARSCFGBlock *ShadowBlock = new STARSCFGBlock(this, (int) BlockIndex);
		this->ShadowCFGBlocks.push_back(ShadowBlock);
		// Keep track of fall-through and non-fall-through blocks inside ShadowCFG.
		STARSTransCFGBlock *CurrTCFGBlock = this->TCFGBlocks[BlockIndex];
		list<STARSTransCFGBlock *>::const_iterator SuccIter = CurrTCFGBlock->GetFallThroughSucc();
		if (SuccIter != CurrTCFGBlock->GetLastConstSucc()) {
			ShadowBlock->GetExpr()->SetFTBlockNum((*SuccIter)->GetNumber());
		}
		SuccIter = CurrTCFGBlock->GetCondNonFallThroughSucc();
		if (SuccIter != CurrTCFGBlock->GetLastConstSucc()) {
			ShadowBlock->GetExpr()->SetNFTBlockNum((*SuccIter)->GetNumber());
		}
		// Keep a deep copy in ClonedShadowCFGBlocks.
		STARSCFGBlock *ClonedShadowBlock = ShadowBlock->DeepCopyExpr();
		this->ClonedShadowCFGBlocks.push_back(ClonedShadowBlock);
	} // end for all blocks in TCFGBlocks
	
	return;
} // SMPFunction::BuildShadowCFG()

// Make LeftExpr be (LeftExpr TopOper RightExpr) or ((not LeftExpr) TopOper RightExpr
void SMPFunction::CoalesceShadowBlocks(int LeftTCFGBlockNum, int RightTCFGBlockNum, bool NegateLeft, SMPoperator TopOper, int NewFTNum, int NewNFTNum) {
	assert((0 <= LeftTCFGBlockNum) && (LeftTCFGBlockNum < (int)this->ClonedShadowCFGBlocks.size()));
	assert((0 <= RightTCFGBlockNum) && (RightTCFGBlockNum < (int)this->ClonedShadowCFGBlocks.size()));
	bool InsideLoop = this->IsTCFGBlockInAnyLoop(LeftTCFGBlockNum);
	STARSTransCFGBlock *LeftTCFGBlock = this->GetTCFGBlockByNum((size_t)LeftTCFGBlockNum);
	STARSTransCFGBlock *RightTCFGBlock = this->GetTCFGBlockByNum((size_t)RightTCFGBlockNum);
	int LeftOriginalBlockNum = LeftTCFGBlock->GetOriginalBlockNum();
	assert(0 <= LeftOriginalBlockNum);
	int InnermostLoopNum = InsideLoop ? this->GetInnermostLoopNum(LeftOriginalBlockNum) : -1;
	// Are both successor blocks in the same innermost loop? If not, this is a loop exit,
	//  Note that we could enter into a still more innermost loop as long as we don't exit the current one.
	bool FTExitsLoop = (!(this->IsTCFGBlockInLoop(NewFTNum, (size_t)InnermostLoopNum)));
	bool NFTExitsLoop = (!(this->IsTCFGBlockInLoop(NewNFTNum, (size_t)InnermostLoopNum)));
	bool LoopExitCase =  InsideLoop && (FTExitsLoop || NFTExitsLoop);
	STARSCFGBlock *LeftCFGBlock = this->ClonedShadowCFGBlocks[LeftTCFGBlockNum];
	STARSCFGBlock *RightCFGBlock = this->ClonedShadowCFGBlocks[RightTCFGBlockNum];

	assert(!LeftCFGBlock->IsCoalesced());
	assert(!RightCFGBlock->IsCoalesced());

	STARSCondExpr *NewTopExpr = new STARSCondExpr();
	NewTopExpr->SetOperator(TopOper);
	NewTopExpr->SetCFGBlock(LeftCFGBlock);
	// If either LeftExpr or RightExpr is just a block, we want to set that block
	//  as the new LeftExpr or RightExpr and not copy the whole expr with its SMP_NULL_OPERATOR.
	if (SMP_NULL_OPERATOR == LeftCFGBlock->GetExpr()->GetOperator()) {
		NewTopExpr->SetLeftExprBlockNum(LeftTCFGBlockNum, NegateLeft);
	}
	else {
		NewTopExpr->SetLeftExpr(LeftCFGBlock->GetExpr(), NegateLeft);
	}
	if (SMP_NULL_OPERATOR == RightCFGBlock->GetExpr()->GetOperator()) {
		NewTopExpr->SetRightExprBlockNum(RightTCFGBlockNum);
	}
	else {
		NewTopExpr->SetRightExpr(RightCFGBlock->GetExpr());
	}
	// Set the then-block (NFT) and else-block (FT)
	NewTopExpr->SetNFTBlockNum(NewNFTNum);
	NewTopExpr->SetFTBlockNum(NewFTNum);

	// Set the NewTopExpr as the expr for LeftBlockNum.
	LeftCFGBlock->SetExpr(NewTopExpr);

	// Mark RightBlockNum as coalesced.
	RightCFGBlock->SetCoalesced();
	RightCFGBlock->SetCoalescedBlockNum(LeftTCFGBlockNum);

	// Mark the conditional branches at the end of LeftBlockNum and RightBlockNum as short circuit branches.
	STARS_ea_t LeftLastAddr = LeftTCFGBlock->GetLastAddr();
	STARS_ea_t RightLastAddr = RightTCFGBlock->GetLastAddr();
	if (LoopExitCase) {
		if (NFTExitsLoop) {
			this->SetControlFlowType(LeftLastAddr, SHORT_CIRCUIT_LOOP_EXIT);
			this->SetControlFlowType(RightLastAddr, SHORT_CIRCUIT_LOOP_EXIT);
		}
		else {
			assert(FTExitsLoop);
			this->SetControlFlowType(LeftLastAddr, SHORT_CIRCUIT_INVERTED_LOOP_EXIT);
			this->SetControlFlowType(RightLastAddr, SHORT_CIRCUIT_INVERTED_LOOP_EXIT);
		}
	}
	else {
		this->SetControlFlowType(LeftLastAddr, SHORT_CIRCUIT_BRANCH);
		this->SetControlFlowType(RightLastAddr, SHORT_CIRCUIT_BRANCH);
	}

	this->HasShortCircuitExpr = true;
	return;
} // end of SMPFunction::CoalesceShadowBlocks()

// Find a consistent follow block for all switch statement cases and record it in TableInfo.FollowNodeNum
// return false if no consistent follow block
bool SMPFunction::FindSwitchStatementFollowBlock(struct SwitchTableInfo &TableInfo) {
	bool ConsistentFollowBlock = true;
	size_t CaseCount = TableInfo.CaseBlockNums.size(); // number of incoming edges we expect to see to the follow block
	int DefaultBlockNum = SMP_BLOCKNUM_UNINIT;
	if (STARS_BADADDR != TableInfo.DefaultJumpAddr) {
		++CaseCount;
		// See if jump table entries also go to the default address, and decrement CaseCount
		//  every time we see such an entry.
#if 0   // indirection, or not? DefaultJumpAddr seems to be the first inst in the default case.
		SMPInstr *DefJumpInst = this->GetInstFromAddr(TableInfo.DefaultJumpAddr);
		assert(nullptr != DefJumpInst);
		STARS_ea_t DefaultCaseAddr = DefJumpInst->GetJumpTarget();
#else
		STARS_ea_t DefaultCaseAddr = TableInfo.DefaultJumpAddr;
#endif
		assert(STARS_BADADDR != DefaultCaseAddr);
		DefaultBlockNum = TableInfo.DefaultCaseBlockNum;
		assert(SMP_BLOCKNUM_UNINIT != DefaultBlockNum);
		SMPBasicBlock *DefaultBlock = this->GetBlockByNum(DefaultBlockNum);
		assert(nullptr != DefaultBlock);
		DefaultBlock->SetSwitchDefaultCase();
		for (size_t CaseIndex = 0; CaseIndex < TableInfo.CaseBlockNums.size(); ++CaseIndex) {
			if (TableInfo.CaseBlockNums[CaseIndex] == DefaultBlockNum) {
				--CaseCount; // Only one edge from default case to follow block can be expected, no matter how many
								// incoming edges to the default block there are
			}
		}
	}

	// Find the immediate dominator for the switch statement.

	// Now, we define a well-structured switch statement to have a consistent follow block reached from all case headers
	//  that did not reach a dead end.
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	for (size_t CaseIndex = 0; ConsistentFollowBlock && (CaseIndex < TableInfo.CaseBlockNums.size()); ++CaseIndex) {
		int CaseBlockNum = TableInfo.CaseBlockNums[CaseIndex];
		this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		int CurrFollowBlockNum = this->FindCaseFollowBlock(CaseBlockNum);
		if (CurrFollowBlockNum == SMP_BLOCKNUM_UNINIT) { // failure
			ConsistentFollowBlock = false;
			SMP_msg("ERROR: SPARK: Unstructured switch at block %d due to SMPFunction::FindCaseFollowBlock failure at block %d.\n",
				TableInfo.IndirJumpBlockNum, TableInfo.CaseBlockNums[CaseIndex]);
		}
		else if (SMP_BLOCKNUM_COMMON_RETURN == CurrFollowBlockNum) { // dead end, e.g. return statement
			--CaseCount;
		}
		else { // found follow block num
			if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) { // first follow block found
				FollowBlockNum = CurrFollowBlockNum;
			}
			else { // check for consistency
				ConsistentFollowBlock = (FollowBlockNum == CurrFollowBlockNum);
				if (!ConsistentFollowBlock)
					SMP_msg("WARNING: Unstructured switch at block %d due to SMPFunction::FindCaseFollowBlock inconsistency at block %d.\n",
						TableInfo.IndirJumpBlockNum, TableInfo.CaseBlockNums[CaseIndex]);
			}
		}
	} // end for all cases except default
	if (ConsistentFollowBlock && (DefaultBlockNum != SMP_BLOCKNUM_UNINIT)) {
		// Check default block follow node consistency
		int CurrFollowBlockNum = this->FindCaseFollowBlock(DefaultBlockNum);
		ConsistentFollowBlock = ((SMP_BLOCKNUM_COMMON_RETURN == CurrFollowBlockNum) || (CurrFollowBlockNum == FollowBlockNum));
		if (!ConsistentFollowBlock) {
			SMP_msg("WARNING: Unstructured switch at block %d due to SMPFunction::FindCaseFollowBlock inconsistency at default block %d.\n",
				TableInfo.IndirJumpBlockNum, TableInfo.DefaultCaseBlockNum);
		}
		if (CurrFollowBlockNum < 0) {
			--CaseCount; // dead end; will not reach follow block
		}
	}

	if (ConsistentFollowBlock) {
		if (0 <= FollowBlockNum) {
			SMPBasicBlock *FollowBlock = this->GetBlockByNum((size_t)FollowBlockNum);
			assert(FollowBlock->GetNumPreds() >= CaseCount);
		}
		else {
			// All cases dead ended; ConsistentFollowBlock implies no problems.
			FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
		}
		TableInfo.FollowNodeNum = FollowBlockNum;
	}

	return ConsistentFollowBlock;
} // end of SMPFunction::FindSwitchStatementFollowBlock()

// Start at CurrBlockNum, return first blocknum not dominated by CaseBlockNum, or -2 if dead end reached first.
int SMPFunction::FindCaseFollowBlock(const int CaseBlockNum) {
	int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	SMPBasicBlock *CaseBlock = this->GetBlockByNum((size_t)CaseBlockNum);
	assert(nullptr != CaseBlock);
	int InnermostLoopNum = this->GetInnermostLoopNum(CaseBlockNum);
	uint16_t LoopBackCount, ForwardCount;
	FollowBlockNum = CaseBlock->GetSingleDomFrontierBlockNum(InnermostLoopNum, LoopBackCount, ForwardCount);
	if (LoopBackCount > 0) { // unstructured switch if case escapes back before the case head block.
		FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	}
	else if (1 < ForwardCount) { // Not a clean convergence if >1 downward DomFrontier blocks
		FollowBlockNum = SMP_BLOCKNUM_UNINIT;
	}
#if 0
	size_t NumDomFrontierBlocks = CaseBlock->GetDomFrontierSize();
	if (0 == NumDomFrontierBlocks) { // dead-end case (e.g. return block terminus)
		FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
	}
	else if (1 == NumDomFrontierBlocks) {
		FollowBlockNum = (*(CaseBlock->GetFirstDomFrontier()));
	}
	else {
		SMP_msg("ERROR: SPARK: CaseBlock %d has more than one block in DomFrontier ", CaseBlockNum);
	}
#endif

#if 0
	// We define a well-structured switch statement such that it is forbidden to have one case fall through to another.
	//  Our algorithm would need to be modified to handle fall-throughs, which would require special translation to SPARK Ada
	//  as such case fall-throughs are not permitted in SPARK Ada.

	// Algorithm: Traverse CFG from CurrBlockNum (init to CaseBlockNum) until we find a block not dominated by CaseBlockNum, or hit a dead end first.
	//  Return that block num, or -2 if dead end. All case blocks must return -2 or a consistent follow block num.
	//  The consistency criterion across case blocks will be imposed by our caller.
	//  Within a case block, all successors must return -2 or a consistent follow block; we impose that criterion here.
	SMPBasicBlock *CurrBlock = this->GetBlockByNum(CurrBlockNum);
	assert(nullptr != CurrBlock);
	if (0 == CurrBlock->GetNumSuccessors()) {
		FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN; // dead end; return statement or call to non-returning func
	}
	else {
		for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
			int SuccBlockNum = SuccBlock->GetNumber();

			// Screen out back edges.
			if (SuccBlockNum <= CurrBlockNum) // RPO numbering determines back edge
				continue;

			// Possibilities: already visited; not dominated by CaseBlockNum; or needs recursion
#if 0
			if (this->DoesBlockDominateBlock(HeaderBlockNum, SuccBlockNum)) {
#endif
				bool DomWithinCase = this->DoesBlockDominateBlock(CaseBlockNum, SuccBlockNum);

				// As long as we trace until DomWithinCase is false, we don't need to impose the IncomingEdgeCount
				//  criterion, because we are already going to impose a consistency criterion.
				if (!DomWithinCase) {
					// follow block found.
					FollowBlockNum = SuccBlockNum;
				}
				else { // DomWithinCase; need to recurse into successor block
					int SuccFollowBlockNum = this->FindCaseFollowBlock(SuccBlockNum, CaseBlockNum, HeaderBlockNum);
					// Meet function for existing FollowBlockNum and SuccFollowBlockNum
					if ((FollowBlockNum == SMP_BLOCKNUM_UNINIT) || (FollowBlockNum == SMP_BLOCKNUM_COMMON_RETURN)) {
						FollowBlockNum = SuccFollowBlockNum;
						if (FollowBlockNum == SMP_BLOCKNUM_UNINIT) { // recursion returned problem code
							break; // failure
						}
					}
					else { // already had good follow block number; just looking for consistency
						if (SuccFollowBlockNum == SMP_BLOCKNUM_UNINIT) { // failure
							FollowBlockNum = SMP_BLOCKNUM_UNINIT;
							break;
						}
						else if (SuccFollowBlockNum >= 0) { // not the -2 dead end code; found follow block
							if (SuccFollowBlockNum != FollowBlockNum) { // inconsistent answers starting at CaseBlockNum
								FollowBlockNum = SMP_BLOCKNUM_UNINIT;
								break;
							}
						}
					}
				}
#if 0
			}
			else { 
				// Only way switch can be structured is if switch follow block is shared with enclosing statement,
				//  e.g. if (condition) then { switch (var) { ... } } where the SPARK Ada "end case;" is followed
				//  immediately by "end if;". Consistency check in our caller will validate this hypothesis.
				FollowBlockNum = SuccBlockNum;
				break;
			}
#endif
		} // end for all successors
	}
#endif

	return FollowBlockNum;
} // end of SMPFunction::FindCaseFollowBlock()

// Search SwitchInfoArray for DefaultCaseAddr, return index
size_t SMPFunction::FindSwitchIndexForDefaultCaseAddr(STARS_ea_t DefaultCaseAddr) const {
	assert(STARS_BADADDR != DefaultCaseAddr);
	size_t FoundSwitchIndex = this->SwitchInfoArray.size() + 1;
	assert(!this->SwitchInfoArray.empty());

	for (size_t SwitchIndex = 0; SwitchIndex < this->SwitchInfoArray.size(); ++SwitchIndex) {
		if (this->SwitchInfoArray[SwitchIndex].DefaultJumpAddr == DefaultCaseAddr) {
			FoundSwitchIndex = SwitchIndex;
			break;
		}
	}

	assert(FoundSwitchIndex < this->SwitchInfoArray.size());
	return FoundSwitchIndex;
} // end of SMPFunction::FindSwitchIndexForDefaultCaseAddr()

// Analyze if-statements with short-circuit operators
bool SMPFunction::AnalyzeCompoundConditionalStatements(void) {
	bool Structured = true;

	// First, build a shadow of the CFG with successor block numbers.
	//  A deep-copied clone will be created in ClonedShadowCFGBlocks, and we will operate on it.
	this->BuildShadowCFG();

	// Second, iterate until no more changes: Detect compound conditional exprs, mark and coalesce.
	bool changed;
	do {
		changed = false;

		for (size_t BlockIndex = 0; BlockIndex < this->ClonedShadowCFGBlocks.size(); ++BlockIndex) {
			STARSTransCFGBlock *CurrBlock = this->TCFGBlocks[BlockIndex];
			int CurrBlockNum = CurrBlock->GetNumber();
			assert(0 <= CurrBlockNum);
			if (!this->ClonedShadowCFGBlocks[(size_t)CurrBlockNum]->IsCoalesced() && CurrBlock->HasConditionalBranch()) {
				// Candidate for beginning a compound (a.k.a. short-circuit) conditional.

				// There are four patterns that cause the next block to be coalesced into this one
				//  as a part of the same compound conditional. These are shown in Figure 6-32 of
				//  Cristina Cifuentes' Ph.D. dissertation on decompilation. With x being the condition
				//  at the end of the first block, and y the condition at the end of the second block,
				//  and "x" below meaning "take branch from block when condition x" and ditto for "y",
				//  these four CFG patterns match the short-circuit evaluation of:
				//  1. x AND y  (i.e. taking the branch from x to y and then also from y leads to a dest. block,
				//               falling through from either one leads to a common second dest. block).
				//         x---+
				//         |   |
				//         |   y---+
				//         |  /    |
				//         | /     |
				//         FT1   NFTNFT1
				//
				//  2. x OR y   (i.e. taking branch from x, or falling through to y and taking branch from y,
				//               leads to first destination block, while falling through to y and then falling through
				//               from y leads to a second destination block)
				//  3. !x OR y  (i.e. falling through from x, or taking branch from x to y and then taking
				//               branch from y, leads to a first dest. block; falling through from y leads
				//               to a second destination block)
				//  4. !x AND y (i.e. falling through from x to y and taking branch from y leads to first
				//               dest. block; taking branch from x or falling through from y leads to second
				//               destination block).
				// Numbering the blocks 1 and 2, ending with exprs x and y respectively, we can identify the four
				//  patterns strictly by the FT (fallthrough) and NFT (non-fall-through) block patterns, observing
				//  that (a) there must be only four blocks and (b) block 2 must be only a conditional expression
				//  and (c) block 2 must only have one predecessor block (block 1):
				//  1. FT(1) = FT(NFT(1))   [implies NFT(1) == 2]
				//  2. NFT(1) = NFT(FT(1))  [implies FT(1) == 2]
				//  3. FT(1) = NFT(NFT(1))  [implies NFT(1) == 2]
				//  4. NFT(1) = FT(FT(1))   [implies FT(1) == 2]
				// The resulting CFG can be structured for the four cases as follows, using Ada short-circuit operators:
				// 1. if x and then y then
				//       NFT(NFT(1))
				//    else
				//       FT(1)
				//    endif
				// 2. if x or else y then
				//       NFT(1)
				//    else
				//       FT(FT(1))
				//    endif
				// 3. if not x or else y then
				//       FT(1)
				//    else
				//       FT(NFT(1))
				//    endif
				// 4. if not x and then y then
				//       NFT(FT(1))
				//    else
				//       NFT(1)
				//    endif
				//  We note that the detection of short-circuit operators is iterative, starting from
				//   RPO block 0 and going downward in the CFG. In example 1 above, the block FT(1) in
				//   the "else" branch could actually be the start of another short-circuit expression,
				//   and likewise for every block in a "then" or "else" branch for all four examples.
				int FT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)CurrBlockNum]->GetExpr()->GetFallThroughBlockNum();
				int NFT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)CurrBlockNum]->GetExpr()->GetNonFallThroughBlockNum();
				if ((SMP_BLOCKNUM_UNINIT != FT1BlockNum) && (SMP_BLOCKNUM_UNINIT != NFT1BlockNum)) {
					STARSTransCFGBlock *FT1Block = this->TCFGBlocks[(size_t)FT1BlockNum];
					STARSTransCFGBlock *NFT1Block = this->TCFGBlocks[(size_t)NFT1BlockNum];
					// Two of the four cases have a conditional-expression-only block 2 as FT(1). See if we
					//  have either case.
					if (FT1Block->IsSingleExpression() && (1 == FT1Block->GetNumPreds())) { // might be case 2 or case 4
						int FTFT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)FT1BlockNum]->GetExpr()->GetFallThroughBlockNum();
						int NFTFT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)FT1BlockNum]->GetExpr()->GetNonFallThroughBlockNum();
						assert(SMP_BLOCKNUM_UNINIT != FTFT1BlockNum);
						assert(SMP_BLOCKNUM_UNINIT != NFTFT1BlockNum);
						if (FTFT1BlockNum == NFT1BlockNum) { // case 4
							// Coalesce FT1Block condition into CurrBlock condition
							this->CoalesceShadowBlocks(CurrBlockNum, FT1BlockNum, true, SMP_LOGICAL_AND, NFT1BlockNum, NFTFT1BlockNum);
							changed = true;
						}
						else if (NFT1BlockNum == NFTFT1BlockNum) { // case 2
							// Coalesce FT1Block condition into CurrBlock condition
							this->CoalesceShadowBlocks(CurrBlockNum, FT1BlockNum, false, SMP_LOGICAL_OR, FTFT1BlockNum, NFT1BlockNum);
							changed = true;
						}
					}
					else if (NFT1Block->IsSingleExpression() && (1 == NFT1Block->GetNumPreds())) { // might be case 1 or case 3
						int FTNFT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)NFT1BlockNum]->GetExpr()->GetFallThroughBlockNum();
						int NFTNFT1BlockNum = this->ClonedShadowCFGBlocks[(size_t)NFT1BlockNum]->GetExpr()->GetNonFallThroughBlockNum();
						if (FT1BlockNum == FTNFT1BlockNum) { // case 1
							// Coalesce NFT1Block condition into CurrBlock condition
							this->CoalesceShadowBlocks(CurrBlockNum, NFT1BlockNum, false, SMP_LOGICAL_AND, FT1BlockNum, NFTNFT1BlockNum);
							changed = true;
						}
						else if (FT1BlockNum == NFTNFT1BlockNum) { // case 3
							// Coalesce NFT1Block condition into CurrBlock condition
							this->CoalesceShadowBlocks(CurrBlockNum, NFT1BlockNum, true, SMP_LOGICAL_OR, FTNFT1BlockNum, FT1BlockNum);
							changed = true;
						}
					}
				}
			} // end if CurrBlock not coalesced and ends with conditional branch
		} // end for all blocks
	}  while (changed);

	// Third, arrange the ClonedShadowCFG into if-then-else order for later translation.
	//  We particularly want to avoid generating code that looks like:
	//  if (compound condition) then
	//     ;  // no code here
	//  else
	//     // code here
	//  endif;
	//  So, we prefer to have RPO order, with smaller block number in the then-branch.
	bool CloningNeeded = false;
	do {
		int CloningHeadTCFGBlockNum = SMP_BLOCKNUM_UNINIT; // track which head block causes a need for cloning
		for (size_t BlockIndex = 0; Structured && (BlockIndex < this->ClonedShadowCFGBlocks.size()); ++BlockIndex) {
			STARSCFGBlock *CurrCFGBlock = this->ClonedShadowCFGBlocks[BlockIndex];
			if (!CurrCFGBlock->IsCoalesced() && (SMP_NULL_OPERATOR != CurrCFGBlock->GetExpr()->GetOperator())) {
				// We have the header block of a short circuit expression.
				//  Distinguish between loop exits and if-then-else statements.
				int OrigBlockNum = CurrCFGBlock->GetOriginalBlockNum();
				assert(0 <= OrigBlockNum);
				SMPBasicBlock *OrigBlock = this->GetBlockByNum((size_t)OrigBlockNum);
				STARS_ea_t LastAddr = OrigBlock->GetLastAddr();
				ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
				if (!IsLoopExitFlow(LastCFType)) { // not SHORT_CIRCUIT_LOOP_EXIT or SHORT_CIRCUIT_INVERTED_LOOP_EXIT
					bool BitsetIfThenCase = false;
					bool BitsetOddIfThenCase = false;
					bool BitsetIfThenElseCase = false;
					int FollowBlockNum = this->FindConditionalFollowNode((int)BlockIndex, BitsetIfThenCase, BitsetOddIfThenCase, BitsetIfThenElseCase);
					CloningNeeded = (!this->SPARKCloningWorkList.empty());
					if (CloningNeeded) {
						CloningHeadTCFGBlockNum = (int)BlockIndex;
						break; // go to outer do-while loop to perform cloning and restart inner loop.
					}

					if (BitsetOddIfThenCase) {
						this->OddIfThenBranches.insert(OrigBlockNum);
					}
					if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) {
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to: Cannot find follow block for SHORT_CIRCUIT_BRANCH in block %zu ", BlockIndex);
							this->DumpFuncNameAndAddr();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						Structured = false;
						break;
					}
					else if (SMP_BLOCKNUM_COMMON_RETURN == FollowBlockNum) {
						SMP_msg("INFO: Common return follow block for SHORT_CIRCUIT_BRANCH in block %zu in %s\n",
							BlockIndex, this->GetFuncName());
						continue; // no code re-arranging to do
					}
					STARSCondExpr *CurrExpr = CurrCFGBlock->GetExpr();
					int FTBlockNum = CurrExpr->GetFallThroughBlockNum();
					int NFTBlockNum = CurrExpr->GetNonFallThroughBlockNum();
					// Check for the if-then case with no else-block.
					bool IfThenCase = false;
					if ((NFTBlockNum == FollowBlockNum) || (FTBlockNum == FollowBlockNum)) {
						IfThenCase = true;
						if (!(BitsetIfThenCase || BitsetOddIfThenCase)) {
							SMP_msg("INFO: SPARK: BitsetIfThenCase %d IfThenCase %d BitsetOddIfThenCase %d SHORT_CIRCUIT HeadBlock %zu ",
								BitsetIfThenCase, IfThenCase, BitsetOddIfThenCase, BlockIndex);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
						}
						// Ensure that we arrange it so that it translates directly to:
						//   if (compound condition) then
						//      do_something;
						//   endif;

						// Compiler optimizations that are not cleaned up can leave worthless branches, e.g.
						//  x86 jle $+2  which goes to the next instruction (2 bytes later) regardless of whether
						//  you take the branch or fall through. This can make the FT and NFT destinations the same.
						//  No need to rearrange code for this junk.
						bool WorthlessBranch = (NFTBlockNum == FTBlockNum);

						if ((FTBlockNum == FollowBlockNum) && (!WorthlessBranch)) {
							// Current structure is:
							//   if (compound condition) then
							//       goto NFTBlock;
							//   FTBlock:
							//   NFTBlock:
							//       [code here]
							//       goto FTBlock;  // all messed up
							// Need to invert the condition and swap FT and NFT blocks.
							// if (inverted compound condition) then goto FTBlock;
							//    NFTBlock:
							//    [code here]
							// endif;
							// FTBlock:
							if (FTBlockNum >= NFTBlockNum) {  // follow node is IDom of condition node, RPO # is greater than if-else blocks.
								STARSCondExpr *NewExpr = new STARSCondExpr();
								NewExpr->SetLeftExpr(CurrExpr, false);
								NewExpr->SetOperator(SMP_BITWISE_NOT);  // too much trouble to introduce new operator SMP_LOGICAL_NOT
								NewExpr->SetFTBlockNum(NFTBlockNum);
								NewExpr->SetNFTBlockNum(FollowBlockNum);
								CurrCFGBlock->SetExpr(NewExpr);
							}
							else {
								Structured = false;
								if (!this->PrintedSPARKUnstructuredMsg) {
									SMP_msg("ERROR: SPARK: Unstructured due to: In inverted SHORT_CIRCUIT_BRANCH at block %zu FTBlockNum: %d not > than NFTBlockNum: %d ",
										BlockIndex, FTBlockNum, NFTBlockNum);
									this->DumpFuncNameAndAddr();
									this->PrintedSPARKUnstructuredMsg = true;
								}
								break;
							}
						}
					}
					else {
						// if-then-else structure, but we usually want the ASM structure to be:
						//  jcond NFTBlock
						//  FTBlock here
						//  jmp FollowBlock
						//  NFTBlock:
						//  [code, then fall through to FollowBlock]
						//  FollowBlock here
						//
						// So, we want whichever block falls through to the FollowBlock to be either the NFTBlock
						//  or to be dominated by the NFTBlock (NFTBlock could start a loop or any other structure
						//  besides a simple basic block).
						if (!BitsetIfThenElseCase) {
							SMP_msg("INFO: SPARK: BitsetIfThenElseCase %d BitsetIfThenCase %d BitsetOddIfThenCase %d SHORT_CIRCUIT HeadBlock %zu ",
								BitsetIfThenElseCase, BitsetIfThenCase, BitsetOddIfThenCase, BlockIndex);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
						}

						int FTPredBlockNum = this->GetFallThroughPredBlockNum(FollowBlockNum);
						if (SMP_BLOCKNUM_UNINIT != FTPredBlockNum) {
							// Is FTPredBlockNum equal to FTBlock, or dominated by FTBlock?
							this->TCFGDomSoFar.ResetAllBits(); // Prepare for TCFGDomTree search, memoize dominated blocks seen so far.
							if (this->DoesTCFGBlockDominateTCFGBlock(FTBlockNum, FTPredBlockNum)) {
								// Need to switch FT and NFT and invert the condition.
								STARSCondExpr *NewExpr = new STARSCondExpr();
								NewExpr->SetLeftExpr(CurrExpr, false);
								NewExpr->SetOperator(SMP_BITWISE_NOT);  // too much trouble to introduce new operator SMP_LOGICAL_NOT
								NewExpr->SetFTBlockNum(NFTBlockNum);
								NewExpr->SetNFTBlockNum(FTBlockNum);
								NewExpr->SetCFGBlock(CurrCFGBlock);
								CurrCFGBlock->SetExpr(NewExpr);
							}
						}
					} // end if if-then-case else if-then-else case

					// Mark unconditional jumps and branches to the follow node in the JumpFollowNodes map.
					STARSTransCFGBlock *FollowBlock = this->TCFGBlocks[(size_t)FollowBlockNum];
					for (STARSTransCFGBlock *PredBlock : FollowBlock->GetConstPredList()) {
						int PredBlockNum = PredBlock->GetNumber();
						if (PredBlock->HasDirectJump() || PredBlock->HasConditionalBranch()) {
							assert(FollowBlock->GetNumber() == FollowBlockNum); // testing compiler insanity
							pair<int, int> MapItem(PredBlockNum, FollowBlockNum);
							pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
							if (!InsertResult.second) {
								// Were we inserting the same follow node again, or a conflicting follow node?
								map<int, int>::iterator ExistingValueIter = InsertResult.first;
								int OldFollowBlockNum = ExistingValueIter->second;
								if (OldFollowBlockNum != FollowBlockNum) {
									if (!this->PrintedSPARKUnstructuredMsg) {
										SMP_msg("ERROR: SPARK: Unstructured due to short-circuit CFG JumpFollowNodesMap conflict at %llx in block %d Old: %d New:%d ",
											(uint64_t)LastAddr, PredBlock->GetNumber(), OldFollowBlockNum, FollowBlockNum);
										this->DumpFuncNameAndAddr();
										this->PrintedSPARKUnstructuredMsg = true;
									}
									Structured = false;
									break;
								}
								else {
									SMP_msg("INFO: SPARK: short-circuit CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)LastAddr);
								}
							}
						}
					}
					// Map header block COND_BRANCH to the follow node
#if 0
					if (!IfThenCase) {
#else
					if (!BitsetOddIfThenCase) {
#endif
						STARSTransCFGBlock *HeaderBlock = this->GetTCFGBlockByNum(BlockIndex);
						pair<int, int> MapItem((int)BlockIndex, FollowBlockNum);
						pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
						if (!InsertResult.second) {
							// Were we inserting the same follow node again, or a conflicting follow node?
							map<int, int>::iterator ExistingValueIter = InsertResult.first;
							int OldFollowBlockNum = ExistingValueIter->second;
							if (OldFollowBlockNum != FollowBlockNum) {
								if (!this->PrintedSPARKUnstructuredMsg) {
									SMP_msg("ERROR: SPARK: Unstructured due to short-circuit CFG JumpFollowNodesMap conflict at %llx in block %zu Old: %d New:%d ",
										(uint64_t)LastAddr, BlockIndex, OldFollowBlockNum, FollowBlockNum);
									this->DumpFuncNameAndAddr();
									this->PrintedSPARKUnstructuredMsg = true;
								}
								Structured = false;
								break;
							}
							else {
								SMP_msg("INFO: SPARK: short-circuit CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)LastAddr);
							}
						}
					}
				} // end if SHORT_CIRCUIT_BRANCH i.e. not SHORT_CIRCUIT_LOOP_EXIT
			}  // end if block is not coalesced and has a non-null operator (implies head of compound conditional)
		} // end for all ShadowCFGBlocks

		if (Structured && CloningNeeded) {
			if (this->HasIndirectJumps()) {
				// Complicated; need to update switch tables after cloning; do later.
				Structured = false;
				SMP_msg("ERROR: SPARK: Function with switch statement needs cloning ");
				this->DumpFuncNameAndAddr();
			}
			else {
				// Perform the requested cloning and clear the cloning work list.
				this->CloningMap.clear();
				this->CloningReverseMap.clear();
				for (pair<int, pair<int, int> > CloningWorkListItem : this->SPARKCloningWorkList) {
					this->CurrSPARKCloningItem = CloningWorkListItem;
					int StartBlockNum = CloningWorkListItem.first;
					int UnlinkBlockNum = CloningWorkListItem.second.first;
					int LimitBlockNum = CloningWorkListItem.second.second;
					STARSTransCFGBlock *UnlinkBlock = this->GetTCFGBlockByNum((size_t)UnlinkBlockNum);
					bool UnlinkNotHeadBlock = (UnlinkBlockNum != CloningHeadTCFGBlockNum);
					int OriginalUnlinkBlockNum = UnlinkBlock->GetOriginalBlockNum();
#if 0
					int InnermostLoopNum = UnlinkNotHeadBlock ? -1 : this->GetInnermostLoopNum(OriginalUnlinkBlockNum);
#else
					int InnermostLoopNum = this->GetInnermostLoopNum(OriginalUnlinkBlockNum);
#endif
					this->ResetTCFGVisitedBlocks();
					this->MarkBlocksAndClone(StartBlockNum, UnlinkBlockNum, LimitBlockNum, InnermostLoopNum, true);
#if 0	// Problems will be fixed by UpdateAfterCloning(), or not. We can assert later if not fixed.
					if (1 == this->CloningMap.size()) {
						// Pred/succ links are not complete until UpdateAfterCloning() executes when we clone more than 1 block or exit the loop.
						bool SaneAfterCloning = this->AuditTCFG();
						if (!SaneAfterCloning) {
							this->DumpDotCFG();
							SMP_msg("FATAL ERROR: TCFG: AuditTCFG failed early in AnalyzeCompoundConditionalStatements() ");
							this->DumpFuncNameAndAddr();
							this->Dump(true);
						}
						assert(SaneAfterCloning);
					}
#endif
				} // end for all items in SPARKCloningWorkList

				if (this->TCFGBlocks.size() > this->TCFGDomSoFar.GetNumBits()) {
					Structured = false;
					break;
				}

				this->UpdateAfterCloning();
				bool SaneAfterCloningUpdate = this->AuditTCFG();
				if (!SaneAfterCloningUpdate) {
					this->DumpDotCFG();
					SMP_msg("FATAL ERROR: TCFG: AuditTCFG failed in AnalyzeCompoundConditionalStatements() ");
					this->DumpFuncNameAndAddr();
					this->Dump(true);
				}
				assert(SaneAfterCloningUpdate);
			}
			this->SPARKCloningWorkList.clear();
		} // end if (CloningNeeded)

	}  while (Structured && CloningNeeded);

	if (this->TCFGBlocks.size() > this->TCFGDomSoFar.GetNumBits()) {
		SMP_msg("ERROR: SPARK: Unstructured due to TCFG size bloat after cloning ");
		this->DumpFuncNameAndAddr();
		this->PrintedSPARKUnstructuredMsg = true;
		this->DumpDotCFG();
		Structured = false;
	}

	return Structured;
} // end of SMPFunction::AnalyzeCompoundConditionalStatements()

// Analyze if-then-else starting at COND_BRANCH at end of applicable blocks; return false if not well-structured
bool SMPFunction::AnalyzeConditionalStatements(void) {

	// First, mark the short-circuit compound conditionals.
	bool StructuredConditional = this->AnalyzeCompoundConditionalStatements();
	if (!StructuredConditional)
		return false;  // failed already, don't waste time

	// Algorithm is based on Figure 6-31 from Cristina Cifuentes' dissertation.
	//  We want to find innermost conditionals first, so we traverse the TCFG in post-order.
	//  When we find a COND_BRANCH, we want to determine the type of conditional (if-then or
	//  if-then-else) and find its follow block.The follow block will have the COND_BRANCH block
	//  as its TCFGIDom (usually), will have two or more predecessors, and will be the greatest such RPO-numbered block.
	//  If the COND_BRANCH goes straight to the follow block, we have an if-then; otherwise, we have if-then-else.
	//  We want to record the follow block for each COND_BRANCH (note that a block can be the follow
	//  block for more than one nested conditional).

	// Traverse CFG in post-order sequence and find COND_BRANCH statements and analyze them.

	vector<int> UnresolvedBranchBlocks;  // for tracking elsif branches until we reach the if block that dominates them
	bool CloningNeeded;
	do {
		CloningNeeded = (!this->SPARKCloningWorkList.empty()); // reset for each iteration
		UnresolvedBranchBlocks.clear(); // Get rid of stale block numbers after cloning; see if cloning solved problem
		int CloningHeadTCFGBlockNum = SMP_BLOCKNUM_UNINIT; // track which head block is causing a need for cloning
		for (vector<STARSTransCFGBlock *>::reverse_iterator BlockIter = this->TCFGBlocks.rbegin(); (!CloningNeeded) && BlockIter != this->TCFGBlocks.rend(); ++BlockIter) {
			STARSTransCFGBlock *CurrTCFGBlock = (*BlockIter);
			SMPInstr *LastInst = (*(CurrTCFGBlock->GetRevInstCBegin()));
			int CurrTCFGBlockNum = CurrTCFGBlock->GetNumber();
			STARS_ea_t LastAddr = LastInst->GetAddr();
			SMPitype FlowType = LastInst->GetDataFlowType();
			if (FlowType == COND_BRANCH) {
				ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
				bool ShortCircuitFlag = IsShortCircuitFlow(LastCFType);
				bool LoopFlowFlag = IsLoopFlow(LastCFType);
				bool ConditionalStatementAlreadyIdentified = IsConditionalStatementFlow(LastCFType);
				if (FALL_THROUGH != LastCFType) { // already part of some compound conditional, loop or switch structure, or conditional from previous iter with cloning
					// It is possible that the current block has a clone, which would have the same
					//  LastAddr and hence the LastCFType could be set without the JumpFollowNodesMap
					//  having an entry for this block number. If so, skip around the continue statements below
					//  so that we will call FindConditionalFollowNode(), etc.
					if (this->JumpFollowNodesMap.find(CurrTCFGBlockNum) != this->JumpFollowNodesMap.cend()) {
						continue;
					}
					else {
						// We could be working upwards into a short circuit expression. We only want to deal with
						//  the head block. Only the head block has entries in the JumpFollowNodesMap.
						if (!LoopFlowFlag && ShortCircuitFlag) {
							STARSCFGBlock *ClonedCFGBlock = this->GetClonedCFGBlockByNum((size_t)CurrTCFGBlockNum);
							if (!ClonedCFGBlock->IsExprHeadBlock()) {
								continue;
							}
						}
						else if (!ConditionalStatementAlreadyIdentified) {
							// Not a conditional of any kind.
							continue;
						}
					}
				}
				else if (1 == CurrTCFGBlock->GetNumSuccessors()) {
					// Stupid code like "jz $+2" that jumps to the fall-through instruction,
					//  so a block with a COND_BRANCH only has one successor. We don't need
					//  to find a conditional follow node.
					continue;
				}
				int HeadTCFGBlockNum = CurrTCFGBlockNum;
				bool BitsetIfThenCase = false;
				bool BitsetOddIfThenCase = false;
				bool BitsetIfThenElseCase = false;
				int FollowNodeNum = this->FindConditionalFollowNode(HeadTCFGBlockNum, BitsetIfThenCase, BitsetOddIfThenCase, BitsetIfThenElseCase);
				if (BitsetOddIfThenCase) {
					this->OddIfThenBranches.insert(HeadTCFGBlockNum);
				}
				CloningNeeded = (!this->SPARKCloningWorkList.empty());
				CloningHeadTCFGBlockNum = CloningNeeded ? HeadTCFGBlockNum : SMP_BLOCKNUM_UNINIT;
				if (FollowNodeNum == SMP_BLOCKNUM_UNINIT) {
					UnresolvedBranchBlocks.push_back(HeadTCFGBlockNum);
#if 1
					// We should no longer need the resolution of UnresolvedBranchBlocks now
					//  that FindConditionalFollowNode() uses the FindTerminus method as a last
					//  resort. Any time we return SMP_BLOCKNUM_UNINIT, we have failed to find
					//  a follow node and we have unstructured control flow.
					break;
#endif
				}
				else if (FollowNodeNum == SMP_BLOCKNUM_COMMON_RETURN) {
					// Insert mapping from COND_BRANCH to follow block.
					pair<int, int> MapItem(HeadTCFGBlockNum, FollowNodeNum);
					pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
					if (!InsertResult.second) {
						// Were we inserting the same follow node again, or a conflicting follow node?
						map<int, int>::iterator ExistingValueIter = InsertResult.first;
						int OldFollowBlockNum = ExistingValueIter->second;
						if (OldFollowBlockNum != FollowNodeNum) {
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to CFG JumpFollowNodesMap conflict at %llx Old: %d New:%d ",
									(uint64_t)LastAddr, OldFollowBlockNum, FollowNodeNum);
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							StructuredConditional = false;
							break;
						}
						else {
							SMP_msg("INFO: SPARK: CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)LastAddr);
						}
					}

					if (!ShortCircuitFlag) {
						this->SetControlFlowType(LastAddr, BRANCH_IF_THEN_ELSE);
					}
					// Any unresolved branches were elsif branches for the current follow node.
					for (size_t ElsifIndex = 0; ElsifIndex < UnresolvedBranchBlocks.size(); ++ElsifIndex) {
						// Insert mappping from COND_BRANCH to follow block.
						int ElsifBlockNum = UnresolvedBranchBlocks[ElsifIndex];
						SMPBasicBlock *ElsifBlock = this->GetBlockByNum(ElsifBlockNum);
						assert(nullptr != ElsifBlock);
						STARS_ea_t ElsifAddr = ElsifBlock->GetLastAddr();
						assert(STARS_BADADDR != ElsifAddr);
						pair<int, int> MapItem(ElsifBlockNum, FollowNodeNum);
						pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
						if (!InsertResult.second) {
							// Were we inserting the same follow node again, or a conflicting follow node?
							map<int, int>::iterator ExistingValueIter = InsertResult.first;
							int OldFollowBlockNum = ExistingValueIter->second;
							if (OldFollowBlockNum != FollowNodeNum) {
								if (!this->PrintedSPARKUnstructuredMsg) {
									SMP_msg("ERROR: SPARK: Unstructured due to CFG JumpFollowNodesMap conflict at %llx Old: %d New:%d ",
										(uint64_t)ElsifAddr, OldFollowBlockNum, FollowNodeNum);
									this->DumpFuncNameAndAddr();
									this->PrintedSPARKUnstructuredMsg = true;
								}
								StructuredConditional = false;
								break;
							}
							else {
								SMP_msg("INFO: SPARK CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)ElsifAddr);
							}
						}
					} // end for each UnresolvedBranchBlock
					if (!UnresolvedBranchBlocks.empty()) {
						SMP_msg("INFO: SPARK: CFG Resolved to ELSIF at %llx\n", (uint64_t)LastAddr);
					}
					UnresolvedBranchBlocks.clear();
				}
				else { // Found a good follow node.
					assert((0 < FollowNodeNum) && (FollowNodeNum < (int) this->TCFGBlocks.size()));
					STARSTransCFGBlock *FollowBlock = this->GetTCFGBlockByNum((size_t)FollowNodeNum);
					assert(nullptr != FollowBlock);
					bool IfThenCase = FollowBlock->IsBlockNumPred(HeadTCFGBlockNum);
					if (IfThenCase != (BitsetIfThenCase || BitsetOddIfThenCase)) {
						SMP_msg("INFO: SPARK: BitsetIfThenCase %d IfThenCase %d BitsetOddIfThenCase %d HeadTCFGBlock %d ",
							BitsetIfThenCase, IfThenCase, BitsetOddIfThenCase, CurrTCFGBlock->GetNumber());
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
					}
					if (IfThenCase) {
						if (!ShortCircuitFlag) {
							this->SetControlFlowType(LastAddr, BRANCH_IF_THEN);
						}
						if (!UnresolvedBranchBlocks.empty()) { // error; cannot have elsif branches for IfThenCase
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to if-then-endif at %llx has unresolved branches inside ",
									(uint64_t)LastAddr);
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							StructuredConditional = false;
							break;
						}
						// Ensure that we won't assert when trying to generate code. For the
						//  normal if-then, we fall through to the then-block and branch to
						//  the follow block. For the OddIfThenCase, we fall through to the
						//  follow block and branch to the then-block.
						int DistantBlockNum = CurrTCFGBlock->GetCondNonFallThroughSuccBlockNum();
						bool OddIfThenCase = ShortCircuitFlag ? BitsetOddIfThenCase : LastInst->IsOddIfThenCase();
						if (OddIfThenCase != BitsetOddIfThenCase) {
							SMP_msg("INFO: SPARK: BitsetIfThenCase %d OddIfThenCase %d BitsetOddIfThenCase %d HeadTCFGBlock %d ",
								BitsetIfThenCase, OddIfThenCase, BitsetOddIfThenCase, CurrTCFGBlock->GetNumber());
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
						}
						if (!OddIfThenCase) {
							if (DistantBlockNum != FollowNodeNum) {
								if (!this->PrintedSPARKUnstructuredMsg) {
									SMP_msg("ERROR: SPARK: Unstructured due to DistantBlock != FollowBlock for normal if-then at %llx ",
										(uint64_t)LastInst->GetAddr());
									this->DumpFuncNameAndAddr();
									this->PrintedSPARKUnstructuredMsg = true;
								}
								this->SetHasUnstructuredCFG(); // DistantBlock != FollowBlock for normal if-then
							}
						}
						else if (DistantBlockNum == FollowNodeNum) {
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to DistantBlock == FollowBlock for OddIfThenCase at %llx ",
									(uint64_t)LastInst->GetAddr());
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							this->SetHasUnstructuredCFG(); // DistantBlock == FollowBlock for OddIfThenCase
						}
					}
					else {
						if (!ShortCircuitFlag) {
							this->SetControlFlowType(LastAddr, BRANCH_IF_THEN_ELSE);
						}
						// Any unresolved branches were elsif branches for the current follow node.
						for (size_t ElsifIndex = 0; ElsifIndex < UnresolvedBranchBlocks.size(); ++ElsifIndex) {
							// Insert mappping from COND_BRANCH to follow block.
							int ElsifBlockNum = UnresolvedBranchBlocks[ElsifIndex];
							STARSTransCFGBlock *ElsifBlock = this->GetTCFGBlockByNum(ElsifBlockNum);
							assert(nullptr != ElsifBlock);
							STARS_ea_t ElsifAddr = ElsifBlock->GetLastAddr();
							assert(STARS_BADADDR != ElsifAddr);
							pair<int, int> MapItem(ElsifBlockNum, FollowNodeNum);
							pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
							if (!InsertResult.second) {
								// Were we inserting the same follow node again, or a conflicting follow node?
								map<int, int>::iterator ExistingValueIter = InsertResult.first;
								int OldFollowBlockNum = ExistingValueIter->second;
								if (OldFollowBlockNum != FollowNodeNum) {
									if (!this->PrintedSPARKUnstructuredMsg) {
										SMP_msg("ERROR: SPARK: Unstructured due to CFG JumpFollowNodesMap conflict at %llx Old: %d New:%d ",
											(uint64_t)ElsifAddr, OldFollowBlockNum, FollowNodeNum);
										this->DumpFuncNameAndAddr();
										this->PrintedSPARKUnstructuredMsg = true;
									}
									StructuredConditional = false;
									break;
								}
								else {
									SMP_msg("INFO: SPARK: CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)ElsifAddr);
								}
							}
							// Find out if the unresolved branch should be translated as beginning a nested if-then, or a nested
							//  if-then-else. The key is whether the COND_BRANCH goes to the follow node directly (if-then) or
							//  not (if-then-else).
							bool NestedIfThenCase = FollowBlock->IsBlockNumPred(ElsifBlockNum);
							if (NestedIfThenCase) {
								this->SetControlFlowType(ElsifAddr, BRANCH_IF_THEN);
							}
							else {
								this->SetControlFlowType(ElsifAddr, BRANCH_IF_THEN_ELSE);
							}
						} // end for each UnresolvedBranchBlock
						UnresolvedBranchBlocks.clear();
					} // end if (IfThenCase) ... else ...
					// Insert mapping from COND_BRANCH to follow block.
					pair<int, int> MapItem(HeadTCFGBlockNum, FollowNodeNum);
					pair<map<int, int>::iterator, bool> InsertResult = this->JumpFollowNodesMap.insert(MapItem);
					if (!InsertResult.second) {
						// Were we inserting the same follow node again, or a conflicting follow node?
						map<int, int>::iterator ExistingValueIter = InsertResult.first;
						int OldFollowBlockNum = ExistingValueIter->second;
						if (OldFollowBlockNum != FollowNodeNum) {
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to CFG JumpFollowNodesMap conflict at %llx Old: %d New:%d ",
									(uint64_t)LastAddr, OldFollowBlockNum, FollowNodeNum);
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
							}
							StructuredConditional = false;
							break;
						}
						else {
							SMP_msg("INFO: SPARK: CFG JumpFollowNodesMap redundant update at %llx\n", (uint64_t)LastAddr);
						}
					}
				} // end if (bad follownode) ... else ...
			} // end if COND_BRANCH
		} // end for all blocks in reverse iteration
		if (CloningNeeded) {
			if (this->HasIndirectJumps()) {
				// Complicated; need to update switch tables after cloning; do later.
				UnresolvedBranchBlocks.clear();
				StructuredConditional = false;
				SMP_msg("ERROR: SPARK: Function with switch statement needs cloning ");
				this->DumpFuncNameAndAddr();
				break;
			}
			// Perform the requested cloning and clear the cloning work list.
			this->CloningMap.clear();
			this->CloningReverseMap.clear();
			for (pair<int, pair<int, int> > CloningWorkListItem : this->SPARKCloningWorkList) {
				this->CurrSPARKCloningItem = CloningWorkListItem;
				int StartBlockNum = CloningWorkListItem.first;
				int UnlinkBlockNum = CloningWorkListItem.second.first;
				int LimitBlockNum = CloningWorkListItem.second.second;
				STARSTransCFGBlock *UnlinkBlock = this->GetTCFGBlockByNum((size_t)UnlinkBlockNum);
				bool UnlinkNotHeadBlock = (UnlinkBlockNum != CloningHeadTCFGBlockNum);
				int OriginalUnlinkBlockNum = UnlinkBlock->GetOriginalBlockNum();
#if 0
				int InnermostLoopNum = UnlinkNotHeadBlock ? -1 : this->GetInnermostLoopNum(OriginalUnlinkBlockNum);
#else
				int InnermostLoopNum = this->GetInnermostLoopNum(OriginalUnlinkBlockNum);
#endif
				this->ResetTCFGVisitedBlocks();
				this->MarkBlocksAndClone(StartBlockNum, UnlinkBlockNum, LimitBlockNum, InnermostLoopNum, true);
#if 0	// Problems will be fixed by UpdateAfterCloning(), or not. We can assert later if not fixed.
				if (1 == this->CloningMap.size()) {
					// Pred/succ links are not complete until UpdateAfterCloning() executes when we clone more than 1 block or exit the loop.
					bool SaneAfterCloning = this->AuditTCFG();
					if (!SaneAfterCloning) {
						this->DumpDotCFG();
						SMP_msg("FATAL ERROR: TCFG: AuditTCFG failed early in AnalyzeConditionalStatements() ");
						this->DumpFuncNameAndAddr();
						this->Dump(true);
					}
					assert(SaneAfterCloning);
				}
#endif
			}
			if (this->TCFGBlocks.size() > this->TCFGDomSoFar.GetNumBits()) {
				StructuredConditional = false;
				break;
			}

			this->UpdateAfterCloning();
			bool SaneAfterCloningUpdate = this->AuditTCFG();
			if (!SaneAfterCloningUpdate) {
				this->DumpDotCFG();
				SMP_msg("FATAL ERROR: TCFG: AuditTCFG failed in AnalyzeConditionalStatements() ");
				this->DumpFuncNameAndAddr();
				this->Dump(true);
			}
			assert(SaneAfterCloningUpdate);
			this->SPARKCloningWorkList.clear();
		}
	}  while (CloningNeeded);

	// We should have no unresolved elsif branches
	if (!UnresolvedBranchBlocks.empty()) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to unresolved branches inside conditional:  ");
			for (int BlockNum : UnresolvedBranchBlocks) {
				SMP_msg("%d ", BlockNum);
			}
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
			this->DumpDotCFG();
		}
		StructuredConditional = false;
	}

	if (this->TCFGBlocks.size() > this->TCFGDomSoFar.GetNumBits()) {
		SMP_msg("ERROR: SPARK: Unstructured due to TCFG size bloat after cloning ");
		this->DumpFuncNameAndAddr();
		this->PrintedSPARKUnstructuredMsg = true;
		this->DumpDotCFG();
		StructuredConditional = false;
	}

	return StructuredConditional;
} // end of SMPFunction::AnalyzeConditionalStatements()

int SMPFunction::FindConditionalFollowNode(const int HeadBlockNum, bool &IfThenCase, bool &OddIfThenCase, bool &IfThenElseCase) {
	IfThenCase = false;
	OddIfThenCase = false;
	IfThenElseCase = false;
	bool NewestIfThenCase = false;
	bool NewestOddIfThenCase = false;
	bool NewestIfThenElseCase = false;
	bool NewestShortCircuit = false;
	int ReturnFollowBlockNum = SMP_BLOCKNUM_UNINIT;

#if 0
	int OldAlgorithmFollowNode = this->FindConditionalFollowNode2(HeadBlockNum);
	ReturnFollowBlockNum = OldAlgorithmFollowNode;
	int NewAlgorithmFollowNode = this->FindConditionalFollowNode3(HeadBlockNum, IfThenCase, OddIfThenCase, IfThenElseCase);
#else // stop comparing; newest algorithm is always the best
	int NewestAlgorithmFollowNode = this->FindConditionalFollowNode4(HeadBlockNum, NewestIfThenCase, NewestOddIfThenCase, NewestIfThenElseCase, NewestShortCircuit);
	bool BadNewest = (NewestAlgorithmFollowNode == SMP_BLOCKNUM_UNINIT);
	ReturnFollowBlockNum = NewestAlgorithmFollowNode;
	IfThenCase = NewestIfThenCase;
	OddIfThenCase = NewestOddIfThenCase;
	IfThenElseCase = NewestIfThenElseCase;
#endif

#if 0
	bool BadOld = (OldAlgorithmFollowNode == SMP_BLOCKNUM_UNINIT);
	bool BadNew = (NewAlgorithmFollowNode == SMP_BLOCKNUM_UNINIT);
	if (BadOld) {
		if (BadNew) {
			if (BadNewest) {
				SMP_msg("INFO: SPARK: Failure on three versions of FindConditionalFollowNode at block %d ", HeadBlockNum);
			}
			else {
				SMP_msg("INFO: SPARK: Failure on two versions, success on newest version of FindConditionalFollowNode at block %d Follow: %d ", HeadBlockNum, NewestAlgorithmFollowNode);
			}
		}
		else {
			if (!BadNewest) {
				SMP_msg("INFO: SPARK: Failure on old version, success on two new versions of FindConditionalFollowNode at block %d Follow: %d NewestFollow: %d ", HeadBlockNum, NewAlgorithmFollowNode, NewestAlgorithmFollowNode);
				ReturnFollowBlockNum = NewAlgorithmFollowNode;
				this->DumpDotCFG();
			}
			else {
				SMP_msg("INFO: SPARK: Failure on old and newest versions, success on new version of FindConditionalFollowNode at block %d Follow: %d NewestFollow: %d ", HeadBlockNum, NewAlgorithmFollowNode, NewestAlgorithmFollowNode);
				ReturnFollowBlockNum = NewAlgorithmFollowNode;
				this->DumpDotCFG();
			}
		}
	}
	else {
		if (BadNew) {
			if (BadNewest) {
				SMP_msg("INFO: SPARK: Success on old version, failure on new and newest versions of FindConditionalFollowNode at block %d Follow: %d ", HeadBlockNum, OldAlgorithmFollowNode);
				// ReturnFollowBlockNum = OldAlgorithmFollowNode;
			}
			else {
				SMP_msg("INFO: SPARK: Success on old and newest versions, failure on new version of FindConditionalFollowNode at block %d Follow: %d NewestFollow: %d ", HeadBlockNum, OldAlgorithmFollowNode, NewestAlgorithmFollowNode);
			}
			this->DumpDotCFG();
		}
		else { // !BadOld, !BadNew
			if (!BadNewest) {
				SMP_msg("INFO: SPARK: Success on three versions of FindConditionalFollowNode at block %d ", HeadBlockNum);
				if ((OldAlgorithmFollowNode != NewAlgorithmFollowNode) || (NewAlgorithmFollowNode != NewestAlgorithmFollowNode)) {
					SMP_msg("Inconsistent answers: Old %d New %d Newest %d ", OldAlgorithmFollowNode, NewAlgorithmFollowNode, NewestAlgorithmFollowNode);
				}
			}
			else {
				SMP_msg("INFO: SPARK: Success on old and new versions, failure on newest version of FindConditionalFollowNode at block %d Follow: %d NewestFollow: %d ", HeadBlockNum, OldAlgorithmFollowNode, NewestAlgorithmFollowNode);
				this->DumpDotCFG();
			}
			ReturnFollowBlockNum = NewAlgorithmFollowNode;
		}
	}
	// Check for boolean CFG consistency.
	if ((BadNew || BadNewest) || (IfThenCase != NewestIfThenCase) || (OddIfThenCase != NewestOddIfThenCase) || (IfThenElseCase != NewestIfThenElseCase)) {
		SMP_msg("Flags: IfThen3 %d OddIfThen3 %d IfThenElse3 %d IfThen4 %d OddIfThen4 %d IfThenElse4 %d ShortCircuit4: %d ",
			IfThenCase, OddIfThenCase, IfThenElseCase, NewestIfThenCase, NewestOddIfThenCase, NewestIfThenElseCase, NewestShortCircuit);
	}
#else // stop comparing; newest algorithm is always the best
	if (BadNewest) {
		this->DumpDotCFG();
	}
	SMP_msg("INFO: SPARK: FindConditionalFollowNode4 for block %d follow %d ", HeadBlockNum, NewestAlgorithmFollowNode);
	SMP_msg("Flags: IfThen4 %d OddIfThen4 %d IfThenElse4 %d ShortCircuit4: %d ",
		NewestIfThenCase, NewestOddIfThenCase, NewestIfThenElseCase, NewestShortCircuit);
#endif
	this->DumpFuncNameAndAddr();

	return ReturnFollowBlockNum;
} // end of SMPFunction::FindConditionalFollowNode()
