/*
 * SMPFunction.cpp - <see below>.
 *
 * Copyright (c) 2000, 2001, 2010 - University of Virginia 
 *
 * This file is part of the Memory Error Detection System (MEDS) infrastructure.
 * This file may be used and modified for non-commercial purposes as long as 
 * all copyright, permission, and nonwarranty notices are preserved.  
 * Redistribution is prohibited without prior written consent from the University 
 * of Virginia.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: University of Virginia
 * e-mail: jwd@virginia.com
 * URL   : http://www.cs.virginia.edu/
 *
 * Additional copyrights 2010, 2011, 2012, 2013, 2014, 2015 by Zephyr Software LLC
 * e-mail: {clc,jwd}@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#include "SMPFunction_support.h"

// Pass types across procedure bounds, return true if types change.
bool SMPFunction::InferInterproceduralTypes(void) {
	this->MDFindIncomingTypes();
	bool change = this->Instrs.front()->InferMarkerInstTypes();

	// See if new type inference can happen.
	// Reset the TypeInferenceComplete flag for call instructions, so they can pick up new callee type info.
	if (!this->IsLeaf()) {
		for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
			if (CurrBlock->HasCallInstruction()) {
				for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
					SMPInstr *CurrInst = (*InstIter);
					SMPitype FlowType = CurrInst->GetDataFlowType();
					if ((CALL == FlowType) || (INDIR_CALL == FlowType)) {
						CurrInst->ResetTypeInferenceComplete();

						if (global_STARS_program->ShouldSTARSShadowCriticalArgs()) {
							// See if we are calling a function with critical InArgs.
							STARS_ea_t CalleeAddr = CurrInst->GetCallTarget();
							SMPFunction *CalleeFunc = nullptr;
							if ((STARS_BADADDR != CalleeAddr) && (!STARS_IsExternalFunc(CalleeAddr))) {
								CalleeFunc = this->GetProg()->FindFunction(CalleeAddr);
							}
							if (nullptr != CalleeFunc) {
								uint32_t CriticalArgPosBits = CalleeFunc->GetTaintInArgPositions();
								if (0 < CriticalArgPosBits) { // has critical InArgs
									size_t HighestBitPosSet = (size_t)HighestBitSetInUint(CriticalArgPosBits);
									for (size_t RegPos = 0; RegPos < HighestBitPosSet; ++RegPos) {
										if (CriticalArgPosBits & (1 << RegPos)) {
											STARS_regnum_t RegNum = (STARS_regnum_t)RegPos;
											STARSOpndTypePtr RegOp = CurrInst->MakeRegOpnd(RegNum);
											STARSDefUseIter UseIter = CurrInst->FindUse(RegOp);
											assert(UseIter != CurrInst->GetLastUse());
											STARS_ea_t InstAddr = CurrInst->GetAddr();
											int UseSSANum = UseIter->GetSSANum();
											bool LocalName = CurrBlock->IsLocalName(RegOp);
											STARSOpndTypePtr ArgSourceOp = nullptr;
											STARS_ea_t UltimateSourceAddr = CurrBlock->GetUltimateDefAddr(RegOp, InstAddr, UseSSANum, LocalName, true, false, ArgSourceOp);
											if (STARS_IsSSAMarkerPseudoID(UltimateSourceAddr) && (nullptr != ArgSourceOp) && ArgSourceOp->IsRegOp()) {
												size_t InArgPos = 0;
												if (global_STARS_program->GetArgRegPosition(ArgSourceOp->GetReg(), InArgPos)) {
													// Traced back to InArg reg. Is it a new trace?
													if (!this->IsCriticalInArg(InArgPos)) {
														// New discovery.
														this->MarkTaintInArgReg(InArgPos);
														change = true;
													}
												}
											}
										}
									} // end for all RegPos bits
								}
							}
						}
					}
				}
			}
		}
	}
	// Now, re-do type inference over the whole function.
	this->InferTypes(false);

	// Propagate fine-grained info.
	for (list<SMPBasicBlock *>::iterator BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		CurrBlock->PropagatePhiFGInfo();
	}
#if STARS_AGGRESSIVE_SIGNEDNESS_PROPAGATION
	change |= this->PropagateSignedness();
#endif

	change |= this->MDFindReturnTypes();

	return change;
} // end of SMPFunction::InferInterproceduralTypes()

// determine signedness and width info for all operands
void SMPFunction::InferFGInfo(void) {
	bool changed, NewChange;
	unsigned short IterCount = 0;
	list<SMPInstr *>::iterator InstIter;
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;

	do {
		changed = false;
		++IterCount;
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			NewChange = CurrInst->InferFGInfo(IterCount);
			changed = (changed || NewChange);
		}
		if (changed) {
			for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
				CurrBlock = (*BlockIter);
				CurrBlock->PropagatePhiFGInfo();
			}
		}
#if STARS_AGGRESSIVE_SIGNEDNESS_PROPAGATION
		if (!changed) {
			changed = this->PropagateSignedness();
		}
#endif
	} while (changed);

	return;
} // end of SMPFunction::InferFGInfo()

// Apply the profiler information to this function once we've inferred everything we can about it.
void SMPFunction::ApplyProfilerInformation(ProfilerInformation* pi)
{
	assert(pi);

	// If no profiler annotations are available, save time.
	if (0 == pi->GetProfilerAnnotationCount())
		return;

	SetIsSpeculative(true);	

	list<SMPInstr *>::iterator InstIter;
	
	bool DebugFlag = false;
#if SMP_DEBUG_PROFILED_TYPE_INFERENCE
	DebugFlag |= (0 == strcmp("dohanoi", this->GetFuncName()));
#endif

	// for each instruction in this function 
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		// lookup whether a load at this instruction was profiled as always numeric 
		InstructionInformation* ii = pi->GetInfo(CurrInst->GetAddr());
		if (ii && DebugFlag)
			SMP_msg("Found instruction information for %llx\n", (uint64_t) CurrInst->GetAddr());
		if (ii && ii->isNumeric()) {
#if SMP_DEBUG_PROFILED_TYPE_INFERENCE
			SMP_msg("Found instruction information for %llx and it's numeric!\n", (uint64_t) CurrInst->GetAddr());
#endif
			CurrInst->UpdateMemLoadTypes((SMPOperandType)(NUMERIC|PROF_BASE));
		}

		// lookup whether this instruction has been profiled as an indirect call
		set<STARS_ea_t> indirect_call_targets = pi->GetIndirectCallTargets(CurrInst->GetAddr());

		for (set<STARS_ea_t>::iterator ict_iter = indirect_call_targets.begin(); ict_iter != indirect_call_targets.end();	++ict_iter) {
			STARS_ea_t target = *ict_iter;
			pair<set<STARS_ea_t>::iterator, bool> InsertResult;
			if (STARS_BADADDR != target) {
				InsertResult = this->IndirectCallTargets.insert(target);
				if (InsertResult.second && (!vector_exists(target, AllCallTargets))) {
					AllCallTargets.push_back(target);
				}
			}
		}

	}
	return;
}	// end of SMPFunction::ApplyProfilerInformation

// For the UNINIT type DEF DefOp, see if all its USEs have a single type.
//  If so, set the DEF to that type and return type,
//  else return UNINIT.
// If DefAddr == STARS_BADADDR, then the DEF is in a Phi function, not an instruction.
SMPOperandType SMPFunction::InferGlobalDefType(const STARSOpndTypePtr &DefOp, int SSANum, SMPBasicBlock *DefBlock, bool CallInst, STARS_ea_t DefAddr) {
	bool DebugFlag = false;
	bool FoundNumeric = false;
	bool FoundCodePtr = false;
	bool FoundPointer = false;
	bool FoundUnknown = false;
	bool FoundUninit = false;
	bool FoundDEF;
	bool DefEscapes = true;

#if SMP_DEBUG_TYPE_INFERENCE
	DebugFlag |= (0 == strcmp("mem_init", this->GetFuncName()));
#endif

	if (DebugFlag) {
		SMP_msg("InferGlobalDefType for SSANum %d of ", SSANum);
		PrintOperand(DefOp);
		SMP_msg("\n");
	}

	vector<SMPInstr *>::iterator InstIter;

	assert(0 <= SSANum);
	set<DefOrUse, LessDefUse>::iterator CurrUse, CurrDef;
	// Go through all instructions in the block and find the instructions
	//  that have USEs of DefOp with SSANum. If all USEs in the chain have
	//  a single type (other than UNINIT), change the DEF type to match the
	//  USE type and set changed to true.
	SMPOperandType UseType = UNINIT;
	SMPOperandType PtrType = UNINIT;

	if (STARS_BADADDR == DefAddr) { // DEF is in a Phi function
		FoundDEF = true;
	}
	else { // DEF is in an instruction
		FoundDEF = false; // need to see the DefAddr first
	}

	for (InstIter = DefBlock->GetFirstInst(); DefEscapes && (InstIter != DefBlock->GetLastInst()); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if ((!FoundDEF) && (DefAddr == CurrInst->GetAddr())) {
			FoundDEF = true;
		}
		else if (FoundDEF) {
			CurrDef = CurrInst->FindDef(DefOp);
			if (CurrDef != CurrInst->GetLastDef()) {
				// Found re-DEF of DefOp.
				DefEscapes = false;
			}
		}
		// NOTE: Following instructions should be inside if (FoundDEF) condition.
		CurrUse = CurrInst->FindUse(DefOp);
		if (CurrUse != CurrInst->GetLastUse()) { // found a USE of DefOp
			if (CurrUse->GetSSANum() == SSANum) { // matched SSA number
				UseType = CurrUse->GetType();
				FoundNumeric |= (IsNumeric(UseType));
				FoundCodePtr |= (IsCodePtr(UseType));
				FoundUnknown |= (IsUnknown(UseType));
				FoundUninit |= (IsEqType(UNINIT, UseType));
				if (IsDataPtr(UseType)) {
					if (FoundPointer) {
						if (IsNotEqType(PtrType, UseType)) {
#if SMP_DEBUG_TYPE_INFERENCE
							SMP_msg("WARNING: Differing ptr types in global chain:");
							SMP_msg(" Prev: %d Current: %d %s\n", PtrType, UseType,
								CurrInst->GetDisasm());
#endif
							PtrType = POINTER;
						}
					}
					else {
						FoundPointer = true;
						PtrType = UseType;
					}
				}
			} // end if matched SSA #
		} // end if found a USE of DefOp
	} // end for all instructions

	if (DefEscapes) { // did not find re-def
		DefEscapes = DefBlock->IsLiveOut(DefOp);
	}


	if (DefEscapes) { // Need to recurse into successor blocks
		list<SMPBasicBlock *>::iterator SuccIter;
		STARS_ea_t TempAddr;
		this->ResetProcessedBlocks(); // set up recursion
		for (SuccIter = DefBlock->GetFirstSucc(); SuccIter != DefBlock->GetLastSucc(); ++SuccIter) {
			SMPBasicBlock *CurrBlock = (*SuccIter);
			set<SMPPhiFunction, LessPhi>::iterator PhiIter = CurrBlock->FindPhi(DefOp);
			TempAddr = DefAddr;
			if (PhiIter != CurrBlock->GetLastPhi()) {
				TempAddr = STARS_BADADDR;  // signals that DefOp will get re-DEFed in a Phi function.
			}
			else if (STARS_BADADDR == TempAddr) { // was STARS_BADADDR coming in to this function
				// We don't want to pass STARS_BADADDR down the recursion chain, because it will be interpreted
				//  by each successor block to mean that DefOp was a Phi USE that got re-DEFed in a Phi function
				//  within itself. Pass the dummy address that indicates LiveIn to the block.
				TempAddr = STARS_LIVEIN_PSEUDO_ID;
			}

			if (CurrBlock->IsLiveIn(DefOp)) {
				CurrBlock->InferGlobalDefType(DefOp, SSANum, TempAddr, FoundNumeric, FoundCodePtr, FoundPointer, FoundUnknown, FoundUninit, PtrType);
			}
		}
	}

	// Do we have a consistent type?
	// If we see any definite POINTER uses, we must set the DEF
	//  to type POINTER or a refinement of it.
	if (FoundPointer)
		UseType = PtrType;
	else if (FoundCodePtr && !FoundUnknown)
		UseType = CODEPTR; // Give CODEPTR priority over NUMERIC and UNINIT
	else if (FoundNumeric && !FoundUninit && !FoundUnknown)
		UseType = NUMERIC;
	else
		return UNINIT; // no POINTER, but no consistent type

	assert(UNINIT != UseType);
	if (DebugFlag) SMP_msg("Inferring global DEF of type %d\n", UseType);

	return UseType;
} // end of SMPFunction::InferGlobalDefType()

// Mark NUMERIC (and propagate) any DEF that starts at small immed. value and gets only small inc/dec operations.
void SMPFunction::FindCounterVariables(void) {
	// We define a counter variable as one that starts out as a small immediate value and then is only updated
	//  via small additions and subtractions. This cannot produce a POINTER, so it must be NUMERIC. This routine
	//  helps get the NUMERIC inference past the phi function barrier, e.g.:
	//
	//  mov eax,0    ; might be NULL POINTER or might be NUMERIC
	//  eax2 := phi(eax1, eax0)  ; still don't know type
	//  label1:    ; top of loop
	//   :
	//  add eax,4  ; increment induction variable; eax1 := eax0 + 4
	//  cmp eax,looplimit
	//  jl label1
	//
	//  Viewed in isolation, adding 4 to EAX could be a pointer operation or a numeric operation, and
	//   the same is true for initializing to zero. Viewed together, these statements obviously cannot be
	//   producing a POINTER value, as 0,4,8, etc. are not a sequence of POINTER values.

	list<SMPInstr *>::iterator InstIter;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		bool ValueFound;
		STARS_uval_t ConstValue;
		if (CurrInst->MDIsSimpleAssignment(ValueFound, ConstValue)) {
			if (ValueFound && (0 == ConstValue)) {
				// Start small: Find init to zero, then track it. Init to small values after we test.
				set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
				if (DefIter == CurrInst->GetLastDef()) {
					// Must have been a simple assignment to a flag, e.g. clc (clear the carry flag).
					continue;
				}
				STARSOpndTypePtr DefOp = DefIter->GetOp();
				if (DefOp->IsRegOp()) {
					list<pair<int, STARS_ea_t> > CounterSSANums;  // SSA numbers that are definitely counters for DefOp
					int DefSSANum = DefIter->GetSSANum();
					STARS_ea_t DefAddr = CurrInst->GetAddr();
					pair<int, STARS_ea_t> ListItem(DefSSANum, DefAddr);
					CounterSSANums.push_back(ListItem);
					SMPBasicBlock *CurrBlock = CurrInst->GetBlock();
					int BlockNum = CurrBlock->GetNumber();
					bool LocalName = CurrBlock->IsLocalName(DefOp);
					if (this->CounterVarHelper(DefOp, DefSSANum, BlockNum, LocalName, CounterSSANums)) {
						while (!CounterSSANums.empty()) {
							int CurrSSANum = CounterSSANums.front().first;
							STARS_ea_t CurrDefAddr = CounterSSANums.front().second;
							bool Propagated;
							if (LocalName) {
								Propagated = CurrBlock->PropagateLocalDefType(DefOp, NUMERIC, CurrDefAddr, CurrSSANum, false);
							}
							else {
								this->ResetProcessedBlocks();
								Propagated = CurrBlock->PropagateGlobalDefType(DefOp, NUMERIC, CurrSSANum, false, false);
							}
							CounterSSANums.pop_front();
						}
					}
				} // end if reg type
			} // end if const value of 0
		} // end if simple assignment
	} // end for all instructions

	return;
} // end of SMPFunction::FindCounterVariables()

// recursive helper for FindCounterVariables()
// Return true if we added to the DefSSANums list.
bool SMPFunction::CounterVarHelper(const STARSOpndTypePtr &DefOp, int DefSSANum, int BlockNum, bool LocalName, list<pair<int, STARS_ea_t> > &CounterSSANums) {
	bool ListExpanded = false;
	std::size_t IncomingListSize = CounterSSANums.size();
	set<int> NonEscapingRegisterHashes;

	// First, examine the Phi list to find uses of DefOp/DefSSANum.
	// Next, examine instructions to find uses of DefOp/DefSSANum. They must be counter operations if DefOp is re-defed.

	// As a first cut, we will just find the following pattern:
	// 1. Counter-style DEF reaches the end of the current block.
	// 2. Successor block is a single-block loop with counter DEF appearing as a USE in a phi function.
	// 3. Within the single-block loop, Phi DEF is used in a counter-style operation, with new DEF becoming a Phi USE at top of block.
	// We will expand this to loops that are not in a single block later.
	SMPBasicBlock *CurrBlock = this->GetBlockByNum((std::size_t) BlockNum);
	assert(nullptr != CurrBlock);
	STARS_ea_t DefAddr = CounterSSANums.front().second;
	if (CurrBlock->DoesDefReachBlockEnd(DefAddr, DefOp, DefSSANum, NonEscapingRegisterHashes)) {
		NonEscapingRegisterHashes.clear(); // Not memoizing for this use of DoesDefReachBlockEnd()
		bool LoopSuccFound = false;
		list<SMPBasicBlock *>::iterator SuccIter;
		SMPBasicBlock *SuccBlock;
		set<SMPPhiFunction, LessPhi>::iterator PhiIter;
		for (SuccIter = CurrBlock->GetFirstSucc(); SuccIter != CurrBlock->GetLastSucc(); ++SuccIter) {
			SuccBlock = (*SuccIter);
			if (SuccBlock->IsSelfLoop()) {
				PhiIter = SuccBlock->FindPhi(DefOp);
				if (PhiIter != SuccBlock->GetLastPhi()) { // Found a Phi function that could match
					std::size_t PhiSize = PhiIter->GetPhiListSize();
					for (std::size_t index = 0; index < PhiSize; ++index) {
						int PhiUseSSANum = PhiIter->GetUseSSANum(index);
						if (PhiUseSSANum == DefSSANum) {
							// Our DEF is a USE in the phi function at the top of SuccBlock. Success.
							int PhiDefSSANum = PhiIter->GetDefSSANum();
							LoopSuccFound = true;
							break;
						}
					}
				}
				if (LoopSuccFound) {
					break;
				}
			}
		}
		if (LoopSuccFound) {
			// SuccBlock points to a one-block loop with PhiIter pointing to a phi function that has
			//  DefOp/DefSSANum as a phi use, and DefOp/PhiDefSSANum as its phi def. Are the uses of
			//  DefOp/PhiDefSSANum within SuccBlock merely counter redefinitions?
			vector<SMPInstr *>::iterator InstIter;
			for (InstIter = SuccBlock->GetFirstInst(); InstIter != SuccBlock->GetLastInst(); ++InstIter) {
				set<DefOrUse, LessDefUse>::iterator DefIter, UseIter;
				SMPInstr *CurrInst = (*InstIter);
				DefIter = CurrInst->FindDef(DefOp);
				if (DefIter != CurrInst->GetLastDef()) {
					// Found a redefinition of DefOp. Is it just a counter operation that redefines DefOp?
					if (CurrInst->IsCounterOperation()) {
						// We will add the new DEF SSA # to the list of counter SSAs.
						pair<int, STARS_ea_t> CounterPair(DefIter->GetSSANum(), CurrInst->GetAddr());
						CounterSSANums.push_back(CounterPair);
						// We don't need to push the PhiDefSSANum discovered earlier, because if
						//  it follows the simple pattern of only using two counter DEFs as its USEs,
						//  one from before the loop and one from within the loop, then both of its USEs
						//  will get set to NUMERIC and propagation will occur naturally. If it does not
						//  fit this simple pattern, we don't want to force it to be NUMERIC yet.
					}
					else {
						// Problem: we redefined DefOp with a non-counter operation. We want to terminate
						//  the chain of detection of counter variables.
						break;
					}
				}
				else {
					UseIter = CurrInst->FindUse(DefOp);
					if (UseIter != CurrInst->GetLastUse()) {
						// Found USE of DefOp. See if it is a POINTER use, which would
						//  invalidate the hypothesis that DefOp is a counter.
						SMPOperandType UseType = UseIter->GetType();
						if (IsDataPtr(UseType) || IsEqType(UseType, CODEPTR)) {
							// Any apparent counter operations so far have really been pointer arithmetic.
							//  We need to restore the list to its incoming state.
							while (IncomingListSize < CounterSSANums.size()) {
								CounterSSANums.pop_back();
							}
							break; // terminate search
						}
					}
				}
			} // end for all insts in SuccBlock
		} // end if LoopSuccFound
	} // end if original pre-loop DEF reaches the end of its block

	ListExpanded = (CounterSSANums.size() > IncomingListSize);
	return ListExpanded;
} // end of SMPFunction::CounterVarHelper()

#define SMP_SIMPLE_CONDITIONAL_TYPE_PROPAGATION 1
#if SMP_SIMPLE_CONDITIONAL_TYPE_PROPAGATION
// The simple form of conditional type propagation observes that we
//  simply need to apply the meet operator over Phi function USEs and
//  then propagate any DEF type changes using PropagateGlobalDefType().
//  The outermost iteration over all type inference methods in InferTypes()
//  will take care of all the propagation that is handled by the work list
//  processing in the textbook algorithm.
// Iteration convergence might be slower in the simple approach, but the code
//  is much simpler to debug.
bool SMPFunction::ConditionalTypePropagation(void) {
	bool changed = false;
	SMPBasicBlock *CurrBlock;
	vector<SMPBasicBlock *>::iterator CurrRPO;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;

	for (CurrRPO = this->RPOBlocks.begin(); CurrRPO != this->RPOBlocks.end(); ++CurrRPO) {
		CurrBlock = *CurrRPO;
		SMPOperandType MeetType;
		CurrPhi = CurrBlock->GetFirstPhi();
		while (CurrPhi != CurrBlock->GetLastPhi()) {
			STARSOpndTypePtr DefOp = CurrPhi->GetAnyOp();
			bool IsMemOp = (! DefOp->IsRegOp());
			MeetType = CurrPhi->ConditionalMeetType(CurrBlock);
			CurrPhi = CurrBlock->FindPhi(DefOp); // maybe stale, so re-find; could be changed by propagation in ConditionalMeetType()

			// Here we use a straight equality test, not our macros,
			//  because we consider it a change if the MeetType is
			//  profiler derived and the DEFType is not.
			if (MeetType != CurrPhi->GetDefType()) {
				// Change the DEF type to the MeetType and propagate.
				CurrPhi = CurrBlock->SetPhiDefType(DefOp, MeetType);
				changed = true;
				this->ResetProcessedBlocks();
				changed |= CurrBlock->PropagateGlobalDefType(DefOp,
					MeetType, CurrPhi->GetDefSSANum(), IsMemOp, false);
			}
			++CurrPhi;
		} // end for all phi functions in the current block
	} // end for all blocks

	return changed;
} // end of SMPFunction::ConditionalTypePropagation()

#else  // not SMP_SIMPLE_CONDITIONAL_TYPE_PROPAGATION

// Apply the SCC (Sparse Conditional Constant) propagation algorithm to
//  propagate types starting from unresolved Phi DEFs.
bool SMPFunction::ConditionalTypePropagation(void) {
	bool changed = false;

	// Collections of Phi functions and instructions that have a DEF
	//  with type UNINIT for the current global name.
	map<int, set<SMPPhiFunction, LessPhi>::iterator> UninitDEFPhis;
	vector<list<SMPInstr>::iterator> UninitDEFInsts;

	// Work lists of Phi functions and instructions that need to be processed
	//  according to the SCC algorithm.
	list<map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator> PhiWorkList;
	list<vector<list<SMPInstr>::iterator>::iterator> InstWorkList;

	// Iterate through all global names that are either (1) registers
	//  or (2) stack locations in SAFE functions.
	STARSOpndSetIter CurrGlob;
	for (CurrGlob = this->GetFirstGlobalName(); CurrGlob != this->GetLastGlobalName(); ++CurrGlob) {
		STARSOpndTypePtr GlobalOp = *CurrGlob;
		list<SMPBasicBlock>::iterator CurrBlock;
		vector<list<SMPBasicBlock>::iterator>::iterator CurrRPO;
		if (MDIsIndirectMemoryOpnd(GlobalOp, this->UseFP))
			continue; // need alias analysis to process indirect accesses
		if ((! GlobalOp->IsRegOp())
			&& (!((this->GetReturnAddressStatus() == FUNC_SAFE) && MDIsStackAccessOpnd(GlobalOp, this->UseFP))))
			continue; // not register, not safe stack access

		// Set up a map (indexed by SSANum) of iterators to Phi functions
		//  for the current global name that have UNINIT as the Phi DEF type.
		UninitDEFPhis.clear();
		UninitDEFInsts.clear();
		for (CurrRPO = this->RPOBlocks.begin(); CurrRPO != this->RPOBlocks.end(); ++CurrRPO) {
			CurrBlock = *CurrRPO;
			set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
			CurrPhi = CurrBlock->FindPhi(GlobalOp);
			if (CurrPhi != CurrBlock->GetLastPhi()) {
				// Found Phi function for current global name.
				if (IsEqType(CurrPhi->GetDefType(), UNINIT)) {
					// Phi DEF is UNINIT; add Phi to the map.
					pair<int, set<SMPPhiFunction, LessPhi>::iterator> TempPair(CurrPhi->GetDefSSANum(), CurrPhi);
					bool Inserted = false;
					map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator WhereIns;
					pair<map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator, bool> Result(WhereIns, Inserted);
					Result = UninitDEFPhis.insert(TempPair);
					assert(Result.second == true);
				}
			}
		} // end for all blocks

		// If any Phi DEF had UNINIT as its type, set up a vector of
		//  iterators to instructions that have UNINIT as the DEF type
		//  for the current global name.
		if (UninitDEFPhis.empty())
			continue;
		list<SMPInstr *>::iterator InstIter;
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			set<DefOrUse, LessDefUse>::iterator CurrDef = CurrInst->FindDef(GlobalOp);
			if (CurrDef != CurrInst->GetLastDef()) {
				// Found DEF of current global name.
				if (IsEqType(UNINIT, CurrDef->GetType())) {
					UninitDEFInsts.push_back(CurrInst);
				}
			}
		} // end for all instructions

		// Put all UNINIT Phi DEFs that have at least one USE
		//  that is not UNINIT onto the PhiWorkList.
		map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator CurrUnPhi;
		for (CurrUnPhi = UninitDEFPhis.begin(); CurrUnPhi != UninitDEFPhis.end(); ++CurrUnPhi) {
			pair<int, set<SMPPhiFunction, LessPhi>::iterator> PhiDefPair(*CurrUnPhi);
			if (PhiDefPair.second->HasTypedUses()) {
				PhiWorkList.push_back(CurrUnPhi);
			}
		}

		// Iterate until both work lists are empty:
		while (!(PhiWorkList.empty() && InstWorkList.empty())) {
			// Process Phi items first.
			while (!PhiWorkList.empty()) {
				// If applying the meet operator over the Phi USE types
				//  would produce a new DEF type, change the DEF type and
				//  propagate it, adding Phi functions and instructions that
				//  received the propagated type to their respective work lists.
				map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator MapIter;
				MapIter = PhiWorkList.front();
				PhiWorkList.pop_front();  // remove from work list
				pair<int, set<SMPPhiFunction, LessPhi>::iterator> PhiDefPair;
				PhiDefPair.first = MapIter->first;
				PhiDefPair.second = MapIter->second;
				set<SMPPhiFunction, LessPhi>::iterator CurrPhi = PhiDefPair.second;
				SMPOperandType MeetType = CurrPhi->ConditionalMeetType();
				// Here we use a straight equality test, not our macros,
				//  because we consider it a change if the MeetType is
				//  profiler derived and the DEFType is not.
				if (MeetType == CurrPhi->GetDefType())
					continue;
				// At this point, we need to set the DEFType to the MeetType
				//  and propagate the change. We have a map of all the
				//  critical Phi functions for this global name, as well
				//  as a vector of the relevant instructions for this name.
				CurrPhi->SetDefType(MeetType);
				changed = true;
				int DefSSANum = CurrPhi->GetDefSSANum();
				map<int, set<SMPPhiFunction, LessPhi>::iterator>::iterator PhiIter;
				vector<list<SMPInstr>::iterator>::iterator InstIter;
				// Propagate to Phi functions first.
				for (PhiIter = UninitDEFPhis.begin(); PhiIter != UninitDEFPhis.end(); ++PhiIter) {
					if (DefSSANum == PhiIter->first)
						continue;  // Skip the Phi that we just changed
					for (std::size_t index = 0; index < PhiIter->second->GetPhiListSize(); ++index) {
						if (DefSSANum == PhiIter->second->GetUseSSANum(index)) {
							// Matched SSA # to USE. Propagate new type.
							PhiIter->second->SetRefType(index, MeetType);
							// Add this phi function to the work list.
							PhiWorkList.push_back(PhiIter);
						}
					}
				}
#define SMP_COND_TYPE_PROP_TO_INSTS 0
#if SMP_COND_TYPE_PROP_TO_INSTS
				// Propagate to instructions with uninit DEFs of global name.
				//  The idea is that the instructions that hold up type propagation
				//  are the ones that USE and then DEF the same global name.
				//  For example, "increment EAX" has to know the type of
				//  the USE of EAX in order to set the type of the DEF.
#endif
			} // end while the PhiWorkList is not empty
#if SMP_COND_TYPE_PROP_TO_INSTS
			// The PhiWorkList is empty at this point, so process
			//  instructions on the InstWorkList.
#endif
		} // end while both work lists are not empty

	} // end for all global names
	return changed;
} // end of SMPFunction::ConditionalTypePropagation()
#endif  // end if SMP_SIMPLE_CONDITIONAL_TYPE_PROPAGATION else ...

// Propagate signedness FG info from DEFs to USEs whenever there is no USE sign info.
bool SMPFunction::PropagateSignedness(void) {
	bool changed = false;
#if STARS_AGGRESSIVE_SIGNEDNESS_PROPAGATION
	map<int, struct FineGrainedInfo>::iterator UseFGIter, DefFGIter;
	list<SMPBasicBlock *>::iterator BlockIter;
	for (UseFGIter = this->GlobalUseFGInfoBySSA.begin(); UseFGIter != this->GlobalUseFGInfoBySSA.end(); ++UseFGIter) {
		struct FineGrainedInfo UseFG = UseFGIter->second;
		if (0 == (UseFG.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS)) {
			// No signedness info. Propagate any signedness info from DEF.
			int UseHashValue = UseFGIter->first;
			unsigned short DefSignMask = this->GetDefSignMiscInfo(UseHashValue);
			DefSignMask &= FG_MASK_SIGNEDNESS_BITS;
			if (0 != DefSignMask) {
				// DEF has signedness info.
				UseFGIter->second.SignMiscInfo |= DefSignMask;
				changed = true;
			}
		}
	}
	// See if we have DEF signedness info for DEFs with no corresponding USE map entries.
	for (DefFGIter = this->GlobalDefFGInfoBySSA.begin(); DefFGIter != this->GlobalDefFGInfoBySSA.end(); ++DefFGIter) {
		struct FineGrainedInfo DefFG = DefFGIter->second;
		unsigned short DefSignMask = (DefFG.SignMiscInfo & FG_MASK_SIGNEDNESS_BITS);
		if (0 != DefSignMask) {
			// Has signedness info. See if USE has no entry.
			int DefHashValue = DefFGIter->first;
			UseFGIter = this->GlobalUseFGInfoBySSA.find(DefHashValue);
			if (UseFGIter == this->GlobalUseFGInfoBySSA.end()) {
				this->UpdateUseSignMiscInfo(DefHashValue, DefSignMask);
				changed = true;
			}
		}
	}
	// Do the same processsing for block-local registers.
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		bool NewChange = (*BlockIter)->PropagateDEFSignedness();
		changed = changed || NewChange;
	}
#endif
	return changed;
} // end of SMPFunction::PropagateSignedness()

void SMPFunction::UpdateLoopFollowBlockNum(int LoopHeadBlockNum, int FollowBlockNum) {
	assert(SMP_BLOCKNUM_UNINIT != LoopHeadBlockNum);
	size_t LoopNum = this->FindLoopNumFromHeadBlockNum(LoopHeadBlockNum);
	int OldFollowNum = SMP_BLOCKNUM_UNINIT;
	bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, OldFollowNum);
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();
	bool ConvergenceComputed = (MultiExitLoop && (this->MultiLoopExitTargetsConvergenceMap.find((int)LoopNum) != this->MultiLoopExitTargetsConvergenceMap.cend()));

	// Safeguard against infinite loops. If our follow node is inside our loop,
	//  we want to leave SMP_BLOCKNUM_UNINIT as the follow node. Ditto, if the follow node
	//  dominates our LoopHeadBlockNum, SPARK translation will loop infinitely.
	this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
	bool FollowDominatesHead = this->DoesBlockDominateBlock(FollowBlockNum, LoopHeadBlockNum);
	if (FollowDominatesHead) {
		SMP_msg("ERROR: SPARK: UpdateLoopFollowBlockNum() ignoring follow block %d which dominates loop head block %d in loop %zu ",
			FollowBlockNum, LoopHeadBlockNum, LoopNum);
		this->DumpFuncNameAndAddr();
		this->LoopExitBlockIgnored[LoopNum] = true;
		return;
	}
	bool FollowInsideLoop = this->IsBlockInLoop(FollowBlockNum, LoopNum);
	if (FollowInsideLoop) {
		SMP_msg("ERROR: SPARK: UpdateLoopFollowBlockNum() ignoring follow block %d inside loop with head block %d in loop %zu",
			FollowBlockNum, LoopHeadBlockNum, LoopNum);
		this->DumpFuncNameAndAddr();
		this->DumpDotCFG();
		this->LoopExitBlockIgnored[LoopNum] = true;
		return;
	}
	bool MultiLoopBreak = this->DetectMultiLevelLoopBreak(LoopHeadBlockNum, FollowBlockNum);
	if (MultiLoopBreak) {
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to multi-level loop break in Loop %zu to block %d ", LoopNum, FollowBlockNum);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
		this->SetHasUnstructuredCFG(); // multi-level loop break
	}

	if (OldFollowNum == SMP_BLOCKNUM_UNINIT) {
		bool EarlyPhaseMultiTarget = (MultiExitLoop && !ConvergenceComputed);
		assert((1 == this->LoopExitTargets[LoopNum].size()) || EarlyPhaseMultiTarget);
		if (!EarlyPhaseMultiTarget) {
			// If convergence has not been computed yet, we could still be processing loops
			//  in ClassifyLoop() and we don't want to clear the exit targets we have found.
			this->LoopExitTargets[LoopNum].clear();
		}
		SMPBasicBlock *FollowBlock = this->GetBlockByNum((size_t)FollowBlockNum);
		if (FollowBlock->IsOnlyInternalDirectJump()) {
			// Skip forward to target of the direct jump. Jump is only for code layout reasons,
			//  not higher-level structuring.
			SMPBasicBlock *FollowTargetBlock = (*(FollowBlock->GetFirstConstSucc()));
			int NewFollowBlockNum = FollowTargetBlock->GetNumber();
			SMP_msg("INFO: SPARK: Replacing JumpOnly loop follow block %d with target block %d", FollowBlockNum, NewFollowBlockNum);
			this->DumpFuncNameAndAddr();
			FollowBlockNum = NewFollowBlockNum;
		}
		this->LoopExitTargets[LoopNum].insert(FollowBlockNum);
	}
	else if (OldFollowNum != FollowBlockNum) {
		if (SMP_BLOCKNUM_UNINIT == FollowBlockNum) {
			if (this->HasGoodLoopFollowBlocks || VerboseOutput) {
				if (!this->PrintedSPARKUnstructuredMsg || VerboseOutput) {
					SMP_msg("INFO: SPARK: Might be unstructured due to conflicting loop follow block nums for loop %d : %d and %d ",
						LoopNum, OldFollowNum, FollowBlockNum);
					this->DumpFuncNameAndAddr();
				}
			}
#if 0
			this->PrintedSPARKUnstructuredMsg = true;
			this->HasGoodLoopFollowBlocks = false; // conflicting loop follow block nums
#endif
		}
		else {
			// Conflicting follow block numbers
			// If we are just skipping past a block that has no instructions
			//  besides an unconditional jump to the new follow block, then that
			//  is fine. Otherwise, we have an unstructured CFG.
			SMPBasicBlock *OldFollowBlock = this->GetBlockByNum(OldFollowNum);
			SMPBasicBlock *NewFollowBlock = this->GetBlockByNum(FollowBlockNum);
			assert(nullptr != OldFollowBlock);
			assert(nullptr != NewFollowBlock);
			bool OldJumpOnly = OldFollowBlock->IsOnlyInternalDirectJump();
			bool success = false;
			if (OldJumpOnly) {
				list<SMPBasicBlock *>::const_iterator SuccIter = OldFollowBlock->GetFirstConstSucc();
				assert(SuccIter != OldFollowBlock->GetLastConstSucc());
				if (FollowBlockNum == (*SuccIter)->GetNumber()) {
					SMP_msg("INFO: Replaced DirectJumpOnly loop follow block %d with %d for head block %d in %s\n", OldFollowNum,
						FollowBlockNum, LoopHeadBlockNum, this->GetFuncName());
					success = true;
					// Remove jump-only block num from exit targets set so that we don't create a
					//  false appearance of a multi-exit-target loop.
					set<int>::const_iterator OldExitTargetIter = this->LoopExitTargets[LoopNum].find(OldFollowNum);
					if (OldExitTargetIter != this->LoopExitTargets[LoopNum].cend()) {
						(void) this->LoopExitTargets[LoopNum].erase(OldExitTargetIter);
					}
				}
			}
			if (!success) {
				if (this->HasGoodLoopFollowBlocks || VerboseOutput) {
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("INFO: SPARK: Might be unstructured due to conflicting loop follow block nums for loop %d : %d and %d ",
							LoopNum, OldFollowNum, FollowBlockNum);
						this->DumpFuncNameAndAddr();
					}
				}
				if (this->HasGoodLoopFollowBlocks) {
#if 0
					this->PrintedSPARKUnstructuredMsg = true;
					this->HasGoodLoopFollowBlocks = false; // Conflicting loop follow blocks
#endif
					if (VerboseOutput)
						this->Dump(false);
					this->DumpDotCFG();
				}
			}
			this->LoopExitTargets[LoopNum].insert(FollowBlockNum);
		}
	}
	return;
} // end of SMPFunction::UpdateLoopFollowBlockNum()

// return -1 if HeaderBlockNum is not a loop header block #
int SMPFunction::GetLoopNumFromHeaderBlockNum(const int HeaderBlockNum) const {
	int LoopNumberFound = -1;
	for (size_t LoopNum = 0; LoopNum < this->LoopHeadBlockNumbers.size(); ++LoopNum) {
		if (this->LoopHeadBlockNumbers[LoopNum] == HeaderBlockNum) {
			LoopNumberFound = (int) LoopNum;
			break;
		}
	}
	return LoopNumberFound;
} // end of SMPFunction::GetLoopNumFromHeaderBlockNum()

// Distinguish between code that must be translated as a C/C++ loop "continue" statement, and
//  code in a loop that could just be an optimized form of an if-then statement inside the loop.
//  See detailed examples in comments below.
bool SMPFunction::FindLoopContinuePattern(const int BranchBlockNum) {
	// There are two cases of a COND_BRANCH at the end of a block inside a loop that
	//  both target the header block of the loop. In the first case, the fall-through
	//  exits the loop. This case is simple and the COND_BRANCH is marked as a LOOP_BACK.
	//  In the second case, the fall-through remains inside the innermost loop containing
	//  the COND_BRANCH block. While the BranchBlockNum basic block is still a loop tail block,
	//  as is any block that has a back edge to the loop header, translation to SPARK Ada
	//  can be tricky.
	//
	// This second case has two subcases with respect to SPARK Ada translation. In the simpler
	//  subcase, the COND_BRANCH is merely guarding the code from the COND_BRANCH to the end of
	//  the loop:
	//
	//  loop
	//    if (cond1) then
	//       [code 1 here]
	//       return; or break; or call non-returning function;
	//    else
	//       [code 2 here]
	//       if (!cond2) then
	//          loop back to loop header;
	//       endif;
	//       [code 3 here]
	//    endif;
	//  endloop;
	//
	//  This code is just a compiler-optimized form of the following source code:
	//
	//  loop
	//    if (cond1) then
	//       [code 1 here]
	//       return; or break; or call non-returning function;
	//    else
	//       [code 2 here]
	//       if (cond2) then
	//          [code 3 here]
	//       endif;
	//    endif;
	//  endloop;
	//
	//  If compiled without optimization, the final endif would need to be
	//   a jump back to the loop header. To reduce jumps and branches, the
	//   compiler can take advantage of the fact that it already has to have
	//   a COND_BRANCH at the test of the innermost if-then and reverse that
	//   cond2 into !cond2 and make the target be the loop header block. 
	//
	// The problem here is that the sequence in which a COND_BRANCH either 
	//  takes a back edge to the loop header or falls through to code that is
	//  still within the loop looks like a C/C++ loop continue statement, and
	//  the SPARK Ada 2012 and Ada 2017 language standards still lack a loop
	//  continue statement. The Ada 2017 rationale document says that control
	//  flow would be less readable with a continue statement, and they actually
	//  recommend putting a label before the endloop and using an explicit goto
	//  statement (!) to reach it when you want to perform a loop continue.
	//  SPARK Ada does not permit use of the goto statement, being a subset of
	//  Ada designed for formal analysis.
	//
	// Contrast the above code with a true continue statement:
	//
	//  loop
	//    if (cond1) then
	//       [code 1 here]
	//    else
	//       [code 2 here]
	//       if (cond2) then
	//          continue;
	//       endif;
	//		 [code 3 here]
	//    endif;
	//    [code 4 here]
	//  endloop;
	//
	//  The else branch can be restructured to put [code 3] inside an if-then
	//   that is contained in the else branch. But [code 4] is reachable by the
	//   then branch as well as the else branch, and it cannot be contained within
	//   the else branch by restructuring. As with all code restructurings, we must
	//   either clone code or introduce a boolean flag to keep track of our path, e.g.:
	//
	//  loop
	//    bool ElseBranchContinueTaken = false;
	//    if (cond1) then
	//       [code 1 here]
	//    else
	//       [code 2 here]
	//       if (cond2) then
	//          ElseBranchContinueTaken = true;
	//       endif;
	//       if (!ElseBranchContinueTaken) then
	//		    [code 3 here]
	//       endif;
	//    endif;
	//    if (!ElseBranchContinueTaken) then
	//       [code 4 here]
	//    endif;
	//  endloop;
	//
	//  So, one subcase has complicated problems in SPARK Ada translation, while the other
	//   has a condition that can be inverted and become an if-then statement.
	//
	//  We can distinguish between these two subcases using dominator tree information.
	//   Beginning at the COND_BRANCH in question, we want to answer the following question:
	//   Following only forward edges, can you reach a basic block that is not dominated by
	//   the THEN (or ELSE) block yet is still within the innermost loop containing the COND_BRANCH
	//   block? The [code 4] block in the complex subcase is an example. This is a true loop continue
	//   case that requires complex translation. If we fail to find such a block, then this code
	//   pattern is NOT a loop continue statement and our SPARK Ada translation just needs to know
	//   not to treat it like a LOOP_BACK that falls out of the loop.

	// Another tricky case involves a basic block that is a tail block for an outer loop
	//  and also a header block for an inner loop. If the block is nothing but a compare-and-branch
	//  then it can be handled as a loop continue statement with additional boolean flags added.
	//  But if it is not a single-expression block (nothing but compare and branch) then the code
	//  other than the compare-and-branch actually has to be part of overlapping loops and cannot
	//  be structured. We can weed out this case before we start tracing the dominator tree for the
	//  more common case described above.
	const SMPBasicBlock *BranchBlock = this->GetBlockByNum((size_t)BranchBlockNum);
	bool OverlappingLoops = (BranchBlock->IsLoopHeaderBlock() && BranchBlock->IsLoopTailBlock() && (!BranchBlock->IsSelfLoop()));
	if (OverlappingLoops) {
		if (BranchBlock->IsSingleExpression()) {
			SMP_msg("INFO: Overlapping loops single-expr loop continue special case at block %d in %s\n", BranchBlockNum, this->GetFuncName());
			return true;
		}
		else {
			this->SetHasUnstructuredCFG(); // Overlapping loops
			if (!this->PrintedSPARKUnstructuredMsg) {
				SMP_msg("ERROR: SPARK: Unstructured due to Overlapping loops non-single-expr unstructured CFG case at block %d ", BranchBlockNum);
				this->DumpFuncNameAndAddr();
				this->PrintedSPARKUnstructuredMsg = true;
			}
			return false;
		}
	}
	const int InnerMostLoopIndex = this->GetInnermostLoopNum(BranchBlockNum);
	this->ResetProcessedBlocks();

	return this->LoopContinueHelper(BranchBlockNum, InnerMostLoopIndex, BranchBlock);
} // end of SMPFunction::FindLoopContinuePattern()

// Recursive helper for FindLoopContinuePattern(). See lengthy comments in that method.
bool SMPFunction::LoopContinueHelper(const int BranchBlockNum, const int InnermostLoopNum, const SMPBasicBlock *CurrentBlock) {
	// Find all forward edges from CurrentBlock. See which ones stay within InnermostLoopNum, then
	//  test the dominator tree info.
	bool FoundContinuePattern = false;
	int CurrBlockNum = CurrentBlock->GetNumber();

	for (list<SMPBasicBlock *>::const_iterator SuccIter = CurrentBlock->GetFirstConstSucc(); SuccIter != CurrentBlock->GetLastConstSucc(); ++SuccIter) {
		SMPBasicBlock *SuccBlock = (*SuccIter);
		int SuccBlockNum = SuccBlock->GetNumber();
		if ((SuccBlockNum > CurrBlockNum) && (!SuccBlock->IsProcessed())) { // avoid back edges
			SuccBlock->SetProcessed(true);
			if (this->IsBlockInLoop(SuccBlockNum, InnermostLoopNum)) {
				// Have we escaped the branch we started in, which can be detected
				//  by the fact that BranchBlockNum does not dominate SuccBlock?
				this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				if (!this->DoesBlockDominateBlock(BranchBlockNum, SuccBlockNum)) {
					// SuccBlock is in innermost loop with BranchBlockNum, but we
					//  cannot enclose it with an if-then inside the BranchBlockNum
					//  code sequence, so it is a true loop continue pattern.
					FoundContinuePattern = true;
					break;
				}
				else {
					// Recurse into SuccBlock.
					if (this->LoopContinueHelper(BranchBlockNum, InnermostLoopNum, SuccBlock)) {
						// Found one block definitely part of a loop continue code pattern.
						FoundContinuePattern = true;
						break;
					}
				}
			}
		}
	} // end for all successors

	return FoundContinuePattern;
} // end of SMPFunction::LoopContinueHelper()


// Detect and mark special cases before emitting numeric error and other security annotations.
void SMPFunction::MarkSpecialNumericErrorCases(void) {
	list<SMPBasicBlock *>::iterator BlockIter;
	vector<SMPInstr *>::reverse_iterator RevInstIter;
	SMPInstr *CurrInst;
	bool DebugFlag = (0 == strcmp("sub_8063BE0", this->GetFuncName()));
	string FuncName(this->GetFuncName());
	bool StartupFunc = IsStartupFuncName(FuncName);
	bool DeepLoopAnalysesRequested = global_STARS_program->ShouldSTARSPerformDeepLoopAnalyses();
	bool SPARKTranslationRequested = global_STARS_program->ShouldSTARSTranslateToSPARKAda();
	bool SPARKVerboseMode = global_stars_interface->VerboseSPARKMode();

	set<int> NonEscapingRegisterHashes; // memoization optimization: set of register/SSA# hashes that do not reach end of block

	// Special-case preparatory analyses.
	for (SMPBasicBlock *CurrBlock : this->Blocks) {
		CurrBlock->AnalyzePrepForNumericAnnotations();
	}

	// Per the decompilation dissertation by Cristina Cifuentes, we scan the code in postorder sequence
	//  to find switch statements from innermost nesting outwards. Then we search for loop control flow,
	//  then we search for if-else control flow last. The control flow graph structuring algorithm requires
	//  a definite order because it is not a finite Church-Rosser system. Because we want to identify jumps
	//  to and around the default case that often precedes the switch indirect jump, and we do not want a LOOP_EXIT
	//  to be identified as a JUMP_TO_DEFAULT_CASE (which is not part of the Cifuentes analysis scheme), we 
	//  do the loop analysis before the switch analysis.
	if (this->HasStructuredControlFlow() && this->HasUnresolvedIndirectJumps()) { // redundant ???
		this->SetHasUnstructuredCFG(); // Unresolved indirect jump
	}

	if (this->HasStructuredControlFlow()) {
		this->InitializeTCFG();
		if (DeepLoopAnalysesRequested || SPARKTranslationRequested) {
			this->ComputeTCFGIDoms();
			this->ComputeTCFGDomFrontiers();
			this->BuildTCFGDominatorTree();
		}
	}

	// Detect loop-related control flow first, then switches, then simple if-else control flow.
	if ((this->LoopCount > 0) && this->HasReducibleControlFlow() && this->HasStructuredControlFlow()) {
		for (size_t BlockIndex = 0; (BlockIndex < this->RPOBlocks.size()) && this->HasStructuredControlFlow(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->GetBlockByNum(BlockIndex);
			vector<SMPInstr *>::const_reverse_iterator InstIter = CurrBlock->GetRevInstCBegin();
			CurrInst = (*InstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			SMPitype CurrDataFlowType = CurrInst->GetDataFlowType();

			// Detect special case of a C/C++ loop continue statement. See comments in SMPFunction::FindLoopContinuePattern().
			bool OverlappingLoops = (CurrBlock->IsLoopHeaderBlock() && CurrBlock->IsLoopTailBlock() && (!CurrBlock->IsSelfLoop()));
			if (OverlappingLoops) {
				this->SetHasUnstructuredCFG(); // might handle the loop continue case (SingleExpr) below in future.
				if (CurrBlock->IsSingleExpression()) {
					SMP_msg("INFO: Overlapping loops single-expr loop continue special case at block %zu ", BlockIndex);
					this->DumpFuncNameAndAddr();
				}
				else {
					this->SetHasUnstructuredCFG(); // Overlapping loops
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("ERROR: SPARK: Unstructured due to Overlapping loops non-single-expr unstructured CFG case at block %d ", BlockIndex);
						this->DumpFuncNameAndAddr();
						this->PrintedSPARKUnstructuredMsg = true;
					}
				}
				break; // no point in continuing if we cannot analyze loop nor translate to SPARK Ada
			}

			// Find all conditional and unconditional jumps within the function.
			if (((JUMP == CurrDataFlowType) || (COND_BRANCH == CurrDataFlowType)) && (!CurrInst->IsFixedCallJump())) {
				bool InvertedExit = false;
				if (CurrInst->IsLoopExitStatement(InvertedExit)) {
					STARS_ea_t JumpTarget = CurrInst->GetJumpTarget();
					if ((STARS_BADADDR == JumpTarget) || (!this->IsInstIDInFunc(JumpTarget))) {
						this->SetHasUnstructuredCFG(); // Bad jump target
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to bad jump target at %llx ", (uint64_t)InstAddr);
							this->DumpFuncNameAndAddr();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						break; // no point in continuing if we cannot translate to SPARK Ada
					}

					if (this->FindMultiLoopContinue((int) BlockIndex)) {
						this->SetHasUnstructuredCFG(); // Multi-loop continue
						break; // no point in continuing if we cannot translate to SPARK Ada
					}

					int FollowBlockNum;
					if (!InvertedExit) {
						this->SetControlFlowType(InstAddr, LOOP_EXIT);
						SMPBasicBlock *TargetBlock = this->GetBlockFromInstAddr(JumpTarget);
						assert(nullptr != TargetBlock);
						FollowBlockNum = TargetBlock->GetNumber();
						if (TargetBlock->IsOnlyInternalDirectJump()) {
							(*(TargetBlock->GetFirstInst()))->SetSPARKTranslated(true); // Will likely be skipped, might produce null inst
							FollowBlockNum = (*(TargetBlock->GetFirstConstSucc()))->GetNumber();
						}
					}
					else {
						this->SetControlFlowType(InstAddr, INVERTED_LOOP_EXIT);
						list<SMPBasicBlock *>::const_iterator FallThroughIter = CurrBlock->GetFallThroughSucc();
						assert(FallThroughIter != CurrBlock->GetLastConstSucc());
						SMPBasicBlock *FallThroughBlock = (*FallThroughIter);
						FollowBlockNum = FallThroughBlock->GetNumber();
						if (FallThroughBlock->IsOnlyInternalDirectJump()) {
							(*(FallThroughBlock->GetFirstInst()))->SetSPARKTranslated(true); // Will likely be skipped, might produce null inst
							int NewFollowBlockNum = (*(FallThroughBlock->GetFirstConstSucc()))->GetNumber();
							SMP_msg("INFO: Special case of INVERTED_LOOP_EXIT from block %zu to %d to %d in %s\n",
								BlockIndex, FollowBlockNum, NewFollowBlockNum, this->GetFuncName());
							FollowBlockNum = NewFollowBlockNum;
						}
					}

					if (CurrBlock->IsLoopTailBlock()) {
						int HeaderBlockNum = CurrBlock->GetLoopHeaderNumber();
						this->UpdateLoopFollowBlockNum(HeaderBlockNum, FollowBlockNum);
						this->MarkLoopContinueCases(CurrBlock, HeaderBlockNum, BlockIndex, InstAddr);
					}
					else {
						int InnerLoopNum = this->GetInnermostLoopNum((int)BlockIndex);
						if (0 <= InnerLoopNum) {
							int HeaderBlockNum = this->LoopHeadBlockNumbers[InnerLoopNum];
							if (0 < HeaderBlockNum) {
								this->UpdateLoopFollowBlockNum(HeaderBlockNum, FollowBlockNum);
							}
						}
					}
				}
				else if (CurrBlock->IsLoopTailBlock()) {
					// We must be looping back to the loop header.
					if (CurrBlock->HasInvertedLoopBack()) { // falling through to loop header
						this->SetControlFlowType(InstAddr, INVERTED_LOOP_BACK);
					}
					else { // normal case; branching to loop header
						this->SetControlFlowType(InstAddr, LOOP_BACK);
					}
					if (COND_BRANCH == CurrDataFlowType) {
						int HeaderBlockNum = CurrBlock->GetLoopHeaderNumber();
						if (!CurrBlock->IsDoubleLoopTailBlock()) { // normal case
							this->MarkLoopContinueCases(CurrBlock, HeaderBlockNum, BlockIndex, InstAddr);
						}
						else { // double tail block
							// Block loops back to outer loop header, falls through to inner loop header.
							//  The FollowBlockNum is the block # of the successor of the outer loop header
							//   that does not enter the outer loop, i.e. it avoids the loop. This is the
							//   FollowBlockNum for both loops.
							int OuterHeadBlockNum = CurrBlock->GetOuterLoopHeaderNumberForDoubleTailBlock();
							assert(SMP_BLOCKNUM_UNINIT != OuterHeadBlockNum);
							int LoopIndex = this->GetLoopNumFromHeaderBlockNum(OuterHeadBlockNum);
							assert(0 <= LoopIndex);
							SMPBasicBlock *OuterHeadBlock = this->GetBlockByNum(OuterHeadBlockNum);
							list<SMPBasicBlock *>::const_iterator SuccIter = OuterHeadBlock->GetSuccNotInLoop((size_t) LoopIndex);
							int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
							if (SuccIter != OuterHeadBlock->GetLastConstSucc()) {
								FollowBlockNum = (*SuccIter)->GetNumber();
							}
							this->UpdateLoopFollowBlockNum(HeaderBlockNum, FollowBlockNum);
							this->UpdateLoopFollowBlockNum(OuterHeadBlockNum, FollowBlockNum);
						}
					}
				}
				else if (CurrBlock->IsOptimizedTopLoopTest() && (COND_BRANCH == CurrDataFlowType)) {
					this->SetControlFlowType(InstAddr, LOOP_BACK);
				}
				else if (JUMP == CurrDataFlowType) {
					// Search for jumps into optimized top-testing loops, which are optimized by compilers to
					//  transform the top-test to a bottom-test, and the loop is entered by jumping down to
					//  the bottom-test block, which is both a loop test block and a loop header block. As we
					//  do not reach this point if the jump is a LOOP_BACK or LOOP_EXIT, we only have to guard
					//  against false detection when an if-else jump within the loop happens to go to the loop
					//  test block.
					STARS_ea_t JumpTarget = CurrInst->GetJumpTarget();
					if ((STARS_BADADDR == JumpTarget) || (!this->IsInstIDInFunc(JumpTarget))) {
						if (!this->PrintedSPARKUnstructuredMsg) {
							SMP_msg("ERROR: SPARK: Unstructured due to bad jump target at %llx ", (uint64_t)InstAddr);
							this->DumpFuncNameAndAddr();
							this->PrintedSPARKUnstructuredMsg = true;
						}
						this->SetHasUnstructuredCFG(); // Bad jump target
						break; // no point in continuing if we cannot translate to SPARK Ada
					}
					else {
						SMPBasicBlock *TargetBlock = this->GetBlockFromInstAddr(JumpTarget);
						int TargetBlockNum = TargetBlock->GetNumber();
						assert((0 <= TargetBlockNum) && (TargetBlockNum < (int) this->RPOBlocks.size()));
						// See if TargetBlockNum is both a test block and a header block for a loop, and get
						//  that loop number if so.
						if (TargetBlock->IsLoopHeaderBlock()) {
							int HeaderBlockNum = TargetBlock->GetNumber();
							assert(0 <= HeaderBlockNum);
							int LoopNum = this->GetLoopNumFromHeaderBlockNum(HeaderBlockNum);
							assert(0 <= LoopNum);  // -1 return value should not happen for a true header block
							if (this->LoopTestBlocksByLoopNum[LoopNum] == TargetBlockNum) { // header & test block at once
								// Now, see if the current instruction is contained within the loop number; if so,
								//  this is not an entry into an optimized top-testing loop, but is simply control flow
								//  within the loop that happens to end up at the tail/test block.
								if (!this->IsBlockInLoop(CurrBlock->GetNumber(), (size_t)LoopNum)) {
									this->SetControlFlowType(InstAddr, JUMP_INTO_LOOP_TEST);
									TargetBlock->SetOptimizedTopLoopTest();
								}
							}
						}
					}
				}
			} // end if JUMP or COND_BRANCH on CurrInst
		} // end for all blocks
	} // end if LoopCount > 0 and structured CFG

	if (DeepLoopAnalysesRequested || SPARKTranslationRequested) {
		for (size_t LoopNum = 0; LoopNum < this->GetNumLoops(); ++LoopNum) {
			if (1 < this->LoopExitTargets[LoopNum].size()) {
				set<int> DummySet;
				this->MultiLoopExitTargetsConvergenceMap[LoopNum] = DummySet;
				this->ComputeLoopFollowNodesReachability(LoopNum);
			}
		}
		if (!this->HasGoodLoopFollowBlocks) {
			this->SetHasUnstructuredCFG(); // Bad loop follow blocks
			SMP_msg("ERROR: SPARK: Function %s is unstructured due to bad loop follow blocks.\n",
				this->GetFuncName());
		}

		if (!StartupFunc) {
			if (this->HasStructuredControlFlow() && this->HasIndirectJumps()) {
				// Traverse CFG in post-order sequence and find indirect jumps and analyze the switch statements
				for (vector<STARSTransCFGBlock *>::reverse_iterator TCFGBlockIter = this->TCFGBlocks.rbegin(); TCFGBlockIter != this->TCFGBlocks.rend(); ++TCFGBlockIter) {
					STARSTransCFGBlock *CurrTCFGBlock = (*TCFGBlockIter);
					if (CurrTCFGBlock->HasIndirectJump()) {
						bool StructuredSwitch = this->AnalyzeSwitchStatement(CurrTCFGBlock);
						if (!StructuredSwitch) {
							this->SetHasUnstructuredCFG(); // Unstructured switch
							if (!this->PrintedSPARKUnstructuredMsg) {
								SMP_msg("ERROR: SPARK: Unstructured due to unstructured switch statements ");
								this->DumpFuncNameAndAddr();
								this->PrintedSPARKUnstructuredMsg = true;
								if (SPARKTranslationRequested) {
									this->DumpDotCFG();
									if (SPARKVerboseMode) {
										this->Dump(true); // skeleton dump to analyze CFG
									}
								}
							}
							break; // give up as soon as CFG is unstructured
						}
					}
				}
			}

			// Now we have identified all switch, loop-back and loop-exit jumps. The remaining jumps
			//  must be part of if-then, if-then-else, and if-then-elsif ... constructs.
			if (this->HasStructuredControlFlow() && SPARKTranslationRequested) {
				bool StructuredConditionals = this->AnalyzeConditionalStatements();
				if (!StructuredConditionals) {
					this->SetHasUnstructuredCFG(); // Unstructured conditionals
					if (!this->PrintedSPARKUnstructuredMsg) {
						SMP_msg("ERROR: SPARK: Unstructured due to miscellaneous CondFollowNode error ");
						this->PrintedSPARKUnstructuredMsg = true;
					}
					else {
						SMP_msg("ERROR: SPARK: Unstructured COND_BRANCH statements ");
					}
					this->DumpFuncNameAndAddr();
#if 0
					this->Dump(false);
#endif
				}
				else if (this->HasStructuredControlFlow()) {
					this->FindGuardedLoops();
				}
			}
		}
	}

	if ((this->LoopCount == 0) && (1 < this->GetNumBlocks())) {
		return;
	}

	// Loop through blocks and detect tight loops of hashing arithmetic.
	//  We also include single-block functions of hashing arithmetic, as sometimes the loop
	//  is unrolled and the entire function is a hash.
#define STARS_MIN_HASH_BLOCK_SIZE 20

	for (SMPBasicBlock *CurrBlock : this->Blocks) {
		int BlockNum = CurrBlock->GetNumber();
#if 0
		if (CurrBlock->IsLoopTailBlock() && CurrBlock->IsLoopHeaderBlock()) {
#else
		if (this->IsBlockInAnyLoop(BlockNum) || ((1 == this->GetNumBlocks()) && (STARS_MIN_HASH_BLOCK_SIZE <= this->Instrs.size()))
			|| this->FuncHasHashingCode()) {
#endif
			// We have a block in a loop, func with previous hash code, or a one-block function. This is the simple case we want
			//  to start with, as hash functions we have observed are tight loops of arithmetic computations.
			//  The next question is whether we can find the kind of shift/rotate that is common to hashing, plus
			//  at least one addition to the result of the shift/rotate.
			bool ShiftFound = false;
			bool AddFound = false;
			set<DefOrUse, LessDefUse>::iterator DefIter;
			NonEscapingRegisterHashes.clear();

			for (RevInstIter = CurrBlock->GetRevInstBegin(); RevInstIter != CurrBlock->GetRevInstEnd(); ++RevInstIter) {
				CurrInst = (*RevInstIter);
				if ((!ShiftFound) && CurrInst->MDIsHashingArithmetic()) {
					// If the operand being shifted is never used in any assignment or arithmetic
					//  except as an address register computation within a memory operand, then the
					//  shifted value does not reach the top of the loop and get shifts accumulated.
					//  In that case, we are not dealing with a shift-and-add type of hash function.
					//  So, do not claim success unless the later addition DEF reaches the end
					//  of the block.
					DefIter = CurrInst->GetFirstNonFlagsDef();
					STARS_ea_t DefAddr = CurrInst->GetAddr();
					STARSOpndTypePtr DefOp = DefIter->GetOp();
					STARS_ea_t AdditionAddr = STARS_BADADDR;
					ShiftFound = CurrBlock->IsDefInvolvedInAddition(DefAddr, DefOp, AdditionAddr);
					if (ShiftFound) {
						SMPInstr *AdditionInst = this->GetInstFromAddr(AdditionAddr);
						DefIter = AdditionInst->GetFirstNonFlagsDef();
						STARSOpndTypePtr AddDefOp = DefIter->GetOp();
						AddFound = CurrBlock->DoesDefReachBlockEnd(AdditionAddr, AddDefOp, DefIter->GetSSANum(), NonEscapingRegisterHashes);
						if (AddFound) {
							break;
						}
						else {
							// Reset ShiftFound and look for a different shift.
							ShiftFound = false;
						}
					}
				}
			}
			if (ShiftFound && AddFound) {
				// We found a tight hashing loop. Mark all the overflowing and underflowing opcodes as benign.
				//  NOTE: We could do loop-variant analysis to ensure that the shifted and added values are actually
				//  changing within the loop, but if they are not, they are probably not exploitable overflows anyway,
				//  and the loop-invariant overflow would happen on every loop iteration based on initial values, which
				//  is a pattern we have never seen for this kind of code.
				this->SetHasHashingCode(true);
				vector<SMPInstr *>::iterator ForwardInstIter;
				for (ForwardInstIter = CurrBlock->GetFirstInst(); ForwardInstIter != CurrBlock->GetLastInst(); ++ForwardInstIter) {
					CurrInst = (*ForwardInstIter);
					if (CurrInst->MDIsOverflowingOpcode() || CurrInst->MDIsUnderflowingOpcode() || CurrInst->MDIsLoadEffectiveAddressInstr()) {
						CurrInst->SetHashOperation();
					}
				}
			}
		} // end if loop header and loop tail
	} // end for all blocks

	NonEscapingRegisterHashes.clear();

	if ((0 < this->GetNumLoops()) && this->HasStructuredControlFlow()) {
		// See if any loop exiting blocks are short circuit conditional expressions,
		//  and fill in the flag in the loop exiting block info.
		// See if a loop BIV is used in the compare or test or decrement inst, fill in flag.
		for (size_t LoopIndex = 0; LoopIndex < this->GetNumLoops(); ++LoopIndex) {
			for (size_t ExitInfoIndex = 0; ExitInfoIndex < this->LoopExitingBlocksVector[LoopIndex].size(); ++ExitInfoIndex) {
				if (this->HasShortCircuitConditional()) {
					int ExitingBlockNum = this->LoopExitingBlocksVector[LoopIndex][ExitInfoIndex].ExitingBlockNum;
					assert(0 <= ExitingBlockNum);
					STARSCFGBlock *ExitingCFGBlock = this->GetCFGBlockByNum((size_t)ExitingBlockNum);
					bool ShortCircuitBlock = ExitingCFGBlock->IsCoalesced() || ExitingCFGBlock->IsExprHeadBlock();
					this->LoopExitingBlocksVector[LoopIndex][ExitInfoIndex].ShortCircuitExpr = ShortCircuitBlock;
				}

				STARSInductionVarFamilyIter ListIter;
				size_t FamilyIndex;
				SMPInstr *CompareOrTestInst = this->LoopExitingBlocksVector[LoopIndex][ExitInfoIndex].CompareTestDecInst;
				if (nullptr != CompareOrTestInst) {
					if (CompareOrTestInst->IsDecrementRTL()) {
						STARSDefUseIter DefIter = CompareOrTestInst->GetFirstNonFlagsDef(); // really a DecInst
						assert(DefIter != CompareOrTestInst->GetLastDef());
						STARSOpndTypePtr DefOp = DefIter->GetOp();
						if (this->IsLoopInductionVarForAnySSANum(LoopIndex, DefOp, ListIter, FamilyIndex)) {
							this->LoopExitingBlocksVector[LoopIndex][ExitInfoIndex].ReferencesBIV = true;
						}
					}
					else {
						for (STARSDefUseConstIter UseIter = CompareOrTestInst->GetFirstConstUse(); UseIter != CompareOrTestInst->GetLastConstUse(); ++UseIter) {
							STARSOpndTypePtr UseOp = UseIter->GetOp();
							if (this->IsLoopInductionVarForAnySSANum(LoopIndex, UseOp, ListIter, FamilyIndex)) {
								this->LoopExitingBlocksVector[LoopIndex][ExitInfoIndex].ReferencesBIV = true;
								break;
							}
						} // end for all USEs
					}
				}
			} // end for all ExitInfo structs for current loop
		} // end for all loops
	}

	// See which loops have exactly one good candidate ExitBlockInfo entries.
	for (size_t LoopIndex = 0; LoopIndex < this->GetNumLoops(); ++LoopIndex) {
		uint32_t GoodCount = 0; // not ShortCircuit, traces back to compare or test or decrement inst
		uint32_t VeryGoodCount = 0; // same, but also references a loop BIV
		bool VeryGoodCandidateFound = false;
		for (const struct LoopExitingBlockInfo CurrExitInfo : this->LoopExitingBlocksVector[LoopIndex]) {
			if ((nullptr != CurrExitInfo.CompareTestDecInst) && (!CurrExitInfo.ShortCircuitExpr)) {
				++GoodCount;
				if (CurrExitInfo.ReferencesBIV) {
					++VeryGoodCount;
					if (!VeryGoodCandidateFound) {
						VeryGoodCandidateFound = true;
						// We will use the first very good candidate.
						if (this->LoopTestBlocksByLoopNum[LoopIndex] != CurrExitInfo.ExitingBlockNum) {
							SMP_msg("INFO: LOOP: Changing loop test block from %d to %d in loop %zu ",
								this->LoopTestBlocksByLoopNum[LoopIndex], CurrExitInfo.ExitingBlockNum, LoopIndex);
							this->DumpFuncNameAndAddr();
							this->DumpDotCFG();
							this->LoopTestBlocksByLoopNum[LoopIndex] = CurrExitInfo.ExitingBlockNum;
							if (this->LoopHeadBlockNumbers[LoopIndex] == CurrExitInfo.ExitingBlockNum) {
								this->LoopTypesByLoopNum[LoopIndex] = STARS_TOP_TESTING_LOOP;
							}
							else if (this->GetBlockByNum((size_t)CurrExitInfo.ExitingBlockNum)->IsLoopTailBlock()) {
								this->LoopTypesByLoopNum[LoopIndex] = STARS_BOTTOM_TESTING_LOOP;
							}
							else {
								this->LoopTypesByLoopNum[LoopIndex] = STARS_INFINITE_OR_MIDDLE_TESTING_LOOP;
							}
						}
					}
				}
			}
		} // end for all ExitInfo structs for current loop
		if ((1 != GoodCount) || (1 != VeryGoodCount)) {
			SMP_msg("WARNING: LOOP: Good candidate ExitBlockInfo count is %lu VeryGood count is %lu for loop %zu ",
				GoodCount, VeryGoodCount, LoopIndex);
			this->DumpFuncNameAndAddr();
		}
	} // end for all loops

	return;
} // end of SMPFunction::MarkSpecialNumericErrorCases()

// Refactored helper for LOOP_EXIT and LOOP_BACK cases in MarkSpecialNumericErrorCases().
void SMPFunction::MarkLoopContinueCases(const SMPBasicBlock *CurrBlock, const int HeaderBlockNum, const std::size_t BlockIndex, const STARS_ea_t InstAddr) {
	list<SMPBasicBlock *>::const_iterator SuccIter = CurrBlock->GetCondOtherSucc(HeaderBlockNum);
	if (SuccIter != CurrBlock->GetLastConstSucc()) {
		int LoopIndex = this->GetLoopNumFromHeaderBlockNum(HeaderBlockNum);
		assert(0 <= LoopIndex);
		int FollowBlockNum = (*SuccIter)->GetNumber();
		if (!this->IsBlockInLoop(FollowBlockNum, (size_t)LoopIndex)) {
			this->UpdateLoopFollowBlockNum(HeaderBlockNum, FollowBlockNum);
		}
		else if (this->FindLoopContinuePattern((int)BlockIndex)) {
			SMP_msg("WARNING: SPARK: LOOP_CONTINUE case detected in block %d FollowBlock %d loop %d ",
				CurrBlock->GetNumber(), FollowBlockNum, LoopIndex);
			this->DumpFuncNameAndAddr();
			this->SetHasUnstructuredCFG(); // loop continue detected
			this->DumpDotCFG();
			this->SetControlFlowType(InstAddr, LOOP_CONTINUE);
		}
		else {
			// LOOP_BACK, but not LOOP_CONTINUE. Translation will need to be aware
			//  that it needs to continue translating at FollowBlockNum in order
			//  to avoid omitting code, unlike LOOP_BACK blocks that fall out of the
			//  loop, as FollowBlockNum is still in the loop.
			SMP_msg("INFO: SPARK: NON_CONTINUE LOOP_BACK case detected in block %d FollowBlock %d of %s\n",
				CurrBlock->GetNumber(), FollowBlockNum, this->GetFuncName());
			this->LoopBackNonExitingBranches.insert(InstAddr);
		}
	}
	else {
		// Fall-through to next function, probably IDA Pro problem in func ID.
		this->SetHasUnstructuredCFG(); // fall-through to next func
		if (!this->PrintedSPARKUnstructuredMsg) {
			SMP_msg("ERROR: SPARK: Unstructured due to fall-through from %llx is not in function ",
				(uint64_t)InstAddr);
			this->DumpFuncNameAndAddr();
			this->PrintedSPARKUnstructuredMsg = true;
		}
	}
	return;
} // end of SMPFunction::MarkLoopContinueCases()

// Use the SSA marker inst to record incoming arg types in member InArgTypes.
void SMPFunction::GatherIncomingArgTypes(void) {
	// Go through the DEFs in the SSA marker inst and see which of them are incoming args.
	list<SMPInstr *>::iterator MarkerInstIter = this->Instrs.begin();
	SMPInstr *MarkerInst = (*MarkerInstIter);
	unsigned short ArgCount = 0;
	if (MarkerInst->IsMarkerInst()) {
		std::size_t StandardByteSize = (MD_NORMAL_MACHINE_BITWIDTH / 8);
		STARSBitSet ArgsProcessed;
		ArgsProcessed.AllocateBits(STARS_MAX_ARGS_PER_FUNC);
		for (set<DefOrUse, LessDefUse>::iterator DefIter = MarkerInst->GetFirstDef(); DefIter != MarkerInst->GetLastDef(); ++DefIter) {
			STARSOpndTypePtr DefOp = DefIter->GetOp();
			std::size_t ArgIndex = 0;
			SMPOperandType ArgType = DefIter->GetType();
			if (IsEqType(ArgType, CODEPTR)) {
				this->HasInArgCodePointer = true;
			}
			else if (IsDataPtr(ArgType)) {
				this->HasInArgDataPointer = true;
			}
			if (global_STARS_program->GetSTARS_ISA_Bitwidth() == 64) {
				if (DefOp->IsRegOp()) {
					STARS_regnum_t RegNum = DefOp->GetReg();
					for (list<STARS_regnum_t>::iterator ArgRegIter = global_STARS_program->GetFirstArgumentReg(); ArgRegIter != global_STARS_program->GetLastArgumentReg(); ++ArgRegIter) {
						if (RegNum == (*ArgRegIter)) { // found an incoming argument
							++ArgCount;
							this->InArgTypes[ArgIndex] = (unsigned short) ArgType;
							if (ArgIndex > this->MaxInArgIndex) {
								this->MaxInArgIndex = ArgIndex;
							}
							if (global_STARS_program->ShouldSTARSShadowFuncPtrs()) {
								STARSOpndSetIter GlobIter = this->GlobalNames.find(DefOp);
								if (GlobIter != this->GlobalNames.end()) {
									unsigned int GlobNameIndex = ExtractGlobalIndex(*GlobIter);
									pair<unsigned int, size_t> InsertValue(GlobNameIndex, ArgIndex);
									pair<map<unsigned int, size_t>::iterator, bool> InsertResult = this->GlobalNameIndexMapToInArgIndex.insert(InsertValue);
									assert(InsertResult.second); // should never be already present

									// Map ArgIndex to operand.
									pair<size_t, STARSOpndTypePtr> ArgMapValue(ArgIndex, DefOp);
									pair<map<size_t, STARSOpndTypePtr>::iterator, bool> ArgMapInsertResult = this->InArgIndexMapToOperand.insert(ArgMapValue);
									assert(InsertResult.second); // should never be already present
								}
							}
							break;
						}
						++ArgIndex;
					}
				}
			}
			else { // 32 bits
				// Current model is writing outargs to stack. 
				if (MDIsDirectStackAccessOpnd(DefOp, this->UsesFramePointer())) {
					// Ensure that we are dealing with an incoming arg and not a USE-before-DEF local var.
					int SignedAddr = (int) DefOp->GetAddr();
					if (0 < SignedAddr) {
						ArgIndex = (DefOp->GetAddr() / StandardByteSize) - 1; // -1 to skip over return address
						if (ArgIndex < STARS_MAX_ARGS_PER_FUNC) {
							if (!ArgsProcessed.GetBit(ArgIndex)) {
								++ArgCount;
								ArgsProcessed.SetBit(ArgIndex);
								this->InArgTypes[ArgIndex] = (unsigned short) ArgType;
								if (ArgIndex > this->MaxInArgIndex) {
									this->MaxInArgIndex = ArgIndex;
								}
								if (global_STARS_program->ShouldSTARSShadowFuncPtrs()) {
									STARSOpndSetIter GlobIter = this->GlobalNames.find(DefOp);
									if (GlobIter != this->GlobalNames.end()) {
										unsigned int GlobNameIndex = ExtractGlobalIndex(*GlobIter);
										pair<unsigned int, size_t> InsertValue(GlobNameIndex, ArgIndex);
										pair<map<unsigned int, size_t>::iterator, bool> InsertResult = this->GlobalNameIndexMapToInArgIndex.insert(InsertValue);
										assert(InsertResult.second); // should never be already present

										// Map ArgIndex to operand.
										pair<size_t, STARSOpndTypePtr> ArgMapValue(ArgIndex, DefOp);
										pair<map<size_t, STARSOpndTypePtr>::iterator, bool> ArgMapInsertResult = this->InArgIndexMapToOperand.insert(ArgMapValue);
										assert(InsertResult.second); // should never be already present
									}
								}
							}
							else { // Not the first time to see arg; could be processing an arg one byte at a time, for example.
								// Do not increment ArgCount for duplicate processing. MaxInArgIndex does not need updating, either.
								bool TypeErrorFlag = false;
								this->InArgTypes[ArgIndex] = (unsigned short) SMPTypeMeet((SMPOperandType) this->InArgTypes[ArgIndex], ArgType, TypeErrorFlag);
								if (TypeErrorFlag) {
									SMP_msg("ERROR: TypeMeet error in GatherIncomingArgType() ");
									this->DumpFuncNameAndAddr();
								}
							}
						}
						else {
							SMP_msg("ERROR: Incoming ArgIndex of %u in SMPFunction::GatherIncomingArgTypes ", ArgIndex);
							this->DumpFuncNameAndAddr();
						}
					}
				}
			}
		} // end for all DEFs in marker inst
	} // end if marker inst

	this->InArgCount = ArgCount;

	return;
} // end of SMPFunction::GatherIncomingArgTypes()

// Find all copies of incoming args and record in bitsets
void SMPFunction::TraceIncomingArgs(void) {
	// Start with tracing CODEPTR inargs.
	if (this->HasInArgCodePointer || this->HasInArgDataPointer) {
		// Allocate vector of bitsets to mark copies of InArgs.
		size_t GlobNameCount = this->NumGlobalNames();
		size_t SSALimit = this->GetMaxStackSSANum();
		if (SSALimit < (size_t) this->GetMaxRegSSANum()) {
			SSALimit = (size_t) this->GetMaxRegSSANum();
		}
		for (size_t index = 0; index < GlobNameCount; ++index) {
			STARSBitSet TempBitSet;
			TempBitSet.AllocateBits(SSALimit);
			this->InArgPointerCopies.push_back(TempBitSet);
		}
		for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
			CurrBlock->InitPointerCopies(this->MaxLocalSSANum);
		}
		this->TraceInArgPointers();
	}
	return;
} // end of SMPFunction::TraceIncomingArgs()

// helper for TraceIncomingArgs(), focused on CODEPTR and POINTER types
void SMPFunction::TraceInArgPointers(void) {
	// Mark the InArgs that are code or data pointers.
	vector<SMPInstr *>::iterator FirstInstIter = this->RPOBlocks[0]->GetFirstInst();
	SMPInstr *FirstInst = (*FirstInstIter);
	assert(nullptr != FirstInst);
	assert(FirstInst->IsMarkerInst());
	for (size_t index = 0; index < this->NumGlobalNames(); ++index) {
		set<STARS_ea_t> EmptySet;
		this->InArgPointerCopyAddrs.push_back(EmptySet);
	}
	for (set<DefOrUse, LessDefUse>::iterator DefIter = FirstInst->GetFirstDef(); DefIter != FirstInst->GetLastDef(); ++DefIter) {
		SMPOperandType DefType = DefIter->GetType();
		if ((CODEPTR == DefIter->GetType()) || IsDataPtr(DefType)) {
			int DefSSANum = DefIter->GetSSANum();
			assert(0 == DefSSANum); // all DEFs in marker inst are SSA # zero
			STARSOpndTypePtr DefOp = DefIter->GetOp();
			STARSOpndSetIter GlobIter = this->GlobalNames.find(DefOp);
			unsigned int NameIndex;
			if (GlobIter != this->GlobalNames.end()) {
				NameIndex = ExtractGlobalIndex(*GlobIter);
				assert(NameIndex < this->InArgPointerCopies.size());
				this->InArgPointerCopies[NameIndex].SetBit(0); // set SSANum 0 bit
				pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->InArgPointerCopyAddrs[NameIndex].insert(STARS_SSA_MARKER_PSEUDO_ID);
				assert(InsertResult.second);
			}
			else { // block-local SSA name; rare for marker inst
				this->RPOBlocks[0]->SetInArgPointerCopy(DefOp, DefSSANum);
			}
		}
	}

	// Iterate through blocks in reverse post-order to mark copies.
	//  NOTE: Extend to more complicated cases with phi functions, etc., iterating until no more changes in the future. !!!!****!!!!
	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
		assert(nullptr != CurrBlock);
		// Iterate through instructions in the current block and find the simple copy RTLs.
		for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARSOpndTypePtr SourceOp;
			if (CurrInst->IsSimpleCopy(SourceOp)) {
				if ((nullptr != SourceOp) && (!SourceOp->IsVoidOp())) {
					bool SourceOpInArgCopy = false;
					CanonicalizeOpnd(SourceOp);
					STARSDefUseIter UseIter = CurrInst->FindUse(SourceOp);
					assert(UseIter != CurrInst->GetLastUse());
					int UseSSANum = UseIter->GetSSANum();

					STARSOpndSetIter GlobIter = this->GlobalNames.find(SourceOp);
					if (GlobIter != this->GlobalNames.end()) {
						unsigned int SourceNameIndex = ExtractGlobalIndex(*GlobIter);
						assert(SourceNameIndex < this->InArgPointerCopies.size());
						SourceOpInArgCopy = this->InArgPointerCopies[SourceNameIndex].GetBit(UseSSANum);
					}
					else { // local SSA name
						SourceOpInArgCopy = CurrBlock->IsInArgPointerCopy(SourceOp, UseSSANum);
					}

					if (SourceOpInArgCopy) {
						// We are copying an InArgPointerCopy to a DEF.
						STARSDefUseIter DefIter = CurrInst->GetFirstDef();
						assert(DefIter != CurrInst->GetLastDef());
						int DefSSANum = DefIter->GetSSANum();
						STARSOpndTypePtr DestOp = DefIter->GetOp();
						STARS_ea_t InstAddr = CurrInst->GetAddr();

						STARSOpndSetIter GlobIter = this->GlobalNames.find(DestOp);
						if (GlobIter != this->GlobalNames.end()) {
							unsigned int DestNameIndex = ExtractGlobalIndex(*GlobIter);
							assert(DestNameIndex < this->InArgPointerCopies.size());
							this->InArgPointerCopies[DestNameIndex].SetBit(DefSSANum);
							pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->InArgPointerCopyAddrs[DestNameIndex].insert(InstAddr);
							assert(InsertResult.second);
						}
						else { // DestOp is local SSA name
							CurrBlock->SetInArgPointerCopy(DestOp, DefSSANum);
						}
					}
				}
			} // end if IsSimpleCopy()
		} // end for all insts in block
	} // end for all blocks in RPOBlocks
	return;
} // end of SMPFunction::TraceInArgPointers()

// Fetch SCCP constant iter into block- or func-level constants. True only if const value.
bool SMPFunction::FindSCCPConstIter(SMPBasicBlock *CurrBlock, int RegHashValue, bool LocalName, STARSSCCPMapIter &ConstIter) {
	bool ConstSizeFound = false;
	if (LocalName) {
		ConstIter = CurrBlock->FindLocalConstValue(RegHashValue);
		if (ConstIter != CurrBlock->GetLastLocalConstValueIter()) {
			if (ConstIter->second.ConstType == STARS_CONST_HAS_VALUE) {
				ConstSizeFound = true;
			}
		}
	}
	else {
		ConstIter = this->FindConstValue(RegHashValue);
		if (ConstIter != this->GetLastConstValueIter()) {
			if (ConstIter->second.ConstType == STARS_CONST_HAS_VALUE) {
				ConstSizeFound = true;
			}
		}
	}
	return ConstSizeFound;
} // end of SMPFunction::GetLastConstValueIter()

// Analyze buffer sizes and the use of buffers by vulnerable library funcs.
void SMPFunction::AnalyzeBufferUses(void) {
	if (this->HasCallToMalloc() && (!this->MallocCallInArgsMap.empty()) && (!this->BufferCallInArgsMap.empty()))
	{
		// We have both mallocs and calls to vulnerable library funcs that use buffers.
		//  It is possible that there are mallocs of constant size buffers, and one or more
		//  constant size buffers could be used by a vulnerable library function like strcpy().
		//  As a preliminary step, trace back from each buffer pointer arg used by vulnerable
		//  library funcs to the ultimate DefAddr. This tracing goes through moves and would stop
		//  at a return value from a call, e.g. the return value from malloc().

		// Map from (ArgOp, LibCallAddr) to UltimateDefAddr.
		//  Hopefully, some of the UltimateDefAddr values will be addresses of malloc()
		//  calls and we can link them to the MallocCallInArgsMap we already built.
		//  We can pre-screen and keep only the UltimateDefAddr values that are malloc() calls
		//  and determine later whether those malloc(size) calls had a constant size.
		STARSDefinitionLinkSet PointerArgCallDefToUltimateDefLinkSet;
		map<STARS_ea_t, STARSDefinitionSet >::const_iterator BufferCallArgsIter;
		for (BufferCallArgsIter = this->BufferCallInArgsMap.cbegin(); BufferCallArgsIter != this->BufferCallInArgsMap.cend(); ++BufferCallArgsIter) {
			STARS_ea_t LibCallAddr = BufferCallArgsIter->first;
			// Iterate through each item in the STARSDefinitionSet for the LibCallAddr.
			const STARSDefinitionSet CurrDefnSet = BufferCallArgsIter->second;
			for (STARSDefnSetIter DefnIter = CurrDefnSet.cbegin(); DefnIter != CurrDefnSet.cend(); ++DefnIter) {
				STARS_ea_t ArgDefAddr = DefnIter->second;
				SMPInstr *ArgDefInst = this->GetInstFromAddr(ArgDefAddr);
				int ArgSSANum = ArgDefInst->GetFirstNonFlagsDef()->GetSSANum();
				STARSOpndTypePtr ArgOp = DefnIter->first;
				STARSOpndTypePtr DefMoveOp;
				bool LocalName = ArgDefInst->GetBlock()->IsLocalName(ArgOp);
				STARS_ea_t UltimateArgAddr = ArgDefInst->GetBlock()->GetUltimateDefAddr(ArgOp, ArgDefAddr, ArgSSANum, LocalName, true, true, DefMoveOp);
				// Is UltimateArgAddr a call malloc() inst?
				if (this->MallocCallInArgsMap.find(UltimateArgAddr) != this->MallocCallInArgsMap.cend()) {
					// Success. Record for later analysis of constant size arg to malloc(size).
					// Make a pseudo-definition of the ArgOp and the LibCallAddr. We won't
					//  care about the ArgDefAddr when emitting annotations.
					STARSDefinition PseudoArgDef(ArgOp, LibCallAddr);
					pair<STARSDefinition, STARS_ea_t> LinkValue(PseudoArgDef, UltimateArgAddr);
					pair<STARSDefinitionLinkSet::iterator, bool> InsertResult = PointerArgCallDefToUltimateDefLinkSet.insert(LinkValue);
					if (!InsertResult.second) {
						SMP_msg("SERIOUS WARNING: AnalyzeBufferUses DefnLink conflict at %llx\n", (uint64_t) UltimateArgAddr);
					}
				}
			}
		}

		// for each entry in MallocCallInArgsMap, see if the InArg to malloc
		//  (i.e. the size argument) is a constant detected in the SCCP analysis.
		for (map<STARS_ea_t, STARS_ea_t>::const_iterator MallocIter = this->MallocCallInArgsMap.cbegin();
			MallocIter != this->MallocCallInArgsMap.cend();
			++MallocIter) {
			STARS_ea_t SizeArgAddr = MallocIter->second;
			SMPInstr *SizeArgPassInst = this->GetInstFromAddr(SizeArgAddr);
			assert(nullptr != SizeArgPassInst);
			STARSDefUseIter SizeArgIter = SizeArgPassInst->GetFirstNonFlagsDef();
			STARSOpndTypePtr SizeArgOp = SizeArgIter->GetOp();
			if (SizeArgOp->IsRegOp()) { // should always be the case for x86-64
				STARSOpndTypePtr SearchOp = CloneIfSubwordReg(SizeArgOp);
				CanonicalizeOpnd(SearchOp);
				bool LocalName = SizeArgPassInst->GetBlock()->IsLocalName(SearchOp);
				int SizeArgHashValue = HashGlobalNameAndSSA(SearchOp, SizeArgIter->GetSSANum());
				STARSSCCPMapIter SizeConstIter;
				STARS_uval_t SizeValue;
				bool ConstSizeFound = this->FindSCCPConstIter(SizeArgPassInst->GetBlock(), 
					SizeArgHashValue, LocalName, SizeConstIter);
				if (ConstSizeFound) { // malloc() of constant size
					SizeValue = SizeConstIter->second.ConstValue;
					// See if any vulnerable lib func call uses the buffer allocated
					//  by this malloc().
					// We will do a linear search of the PointerArgCallDefToUltimateDefLinkSet, which
					//  should be small, as it only has links between vulnerable library call pointer
					//  args and malloc() return values.
					STARSDefinitionLinkSet::const_iterator LinkIter = PointerArgCallDefToUltimateDefLinkSet.cbegin();
					STARS_ea_t CurrMallocAddr = MallocIter->first;
					while (LinkIter != PointerArgCallDefToUltimateDefLinkSet.cend()) {						
						STARS_ea_t MallocCallAddr = LinkIter->second;
						if (CurrMallocAddr == MallocCallAddr) {
							// Matched. Keep track of info needed in annotations: the argument position,
							//  zero-based, and the constant buffer size, linked to the lib func call addr.
							STARS_ea_t LibFuncCallAddr = LinkIter->first.second;
							size_t ArgPos;
							STARS_regnum_t ArgRegNum = LinkIter->first.first->GetReg();
							STARSBufferCallArgMap::iterator BufferArgMapIter = this->BufferCallConstBufferArgsMap.find(LibFuncCallAddr);
							bool found = global_STARS_program->GetArgRegPosition(ArgRegNum, ArgPos);
							if (found) {
								pair<size_t, STARS_uval_t> InsertVal(ArgPos, SizeValue);
								if (BufferArgMapIter != this->BufferCallConstBufferArgsMap.end()) {
									// Already has entry with some other args. Insert into sorted list.
									list<pair<size_t, STARS_uval_t> >::iterator ArgListIter = BufferArgMapIter->second.begin();
									bool inserted = false;
									while (ArgListIter != BufferArgMapIter->second.end()) {
										size_t CurrArgPos = (*ArgListIter).first;
										if (CurrArgPos > ArgPos) {
											(void) BufferArgMapIter->second.insert(ArgListIter, InsertVal);
											inserted = true;
											break;
										}
										else if (CurrArgPos == ArgPos) {
											// Duplicate search for some reason?
											inserted = true;
											break;
										}
										++ArgListIter;
									} // end of while (ArgListIter != BufferArgMapIter->second.end())
									if (!inserted) {
										// Needs to go at the end of the list.
										BufferArgMapIter->second.push_back(InsertVal);
									}
								}
								else {
									// Create new entry.
									list<pair<size_t, STARS_uval_t> > NewList;
									NewList.push_back(InsertVal);
									this->BufferCallConstBufferArgsMap[LibFuncCallAddr] = NewList;
								}
							}
							else {
								SMP_msg("ERROR: Argument position not found in AnalyzeBufferUses, ArgRegNum: %d ", (int) ArgRegNum);
								this->DumpFuncNameAndAddr();
							}
						}
						++LinkIter;
					} // end of while (LinkIter != PointerArgCallDefToUltimateDefLinkSet.cend())
				}
			}
		}
	}
	return;
} // end of SMPFunction::AnalyzeBufferUses()

// Remove non-global SSA names from RefSet.
void SMPFunction::RemoveLocalRefs(STARSDefUseSet &RefSet) {
	STARSDefUseIter RefIter = RefSet.begin();
	while (RefIter != RefSet.end()) {
		STARSOpndTypePtr RefOp = RefIter->GetOp();
		if (!this->IsGlobalName(RefOp)) {
			STARSDefUseIter NextRefIter = RefIter;
			++NextRefIter;
			RefSet.erase(RefIter);
			RefIter = NextRefIter;
		}
		else {
			++RefIter;
		}
	}

	return;
} // end of SMPFunction::RemoveLocalRefs()

// Is CriticalOp already represented in AlreadyShadowed set?
bool SMPFunction::IsAlreadyShadowed(const ShadowPoint &CriticalOp) {
	bool found = false;

	for (ShadowSet::const_iterator AlreadyIter = this->AlreadyShadowed.cbegin(); AlreadyIter != this->AlreadyShadowed.cend(); ++AlreadyIter) {
		if (AlreadyIter->first == CriticalOp.first) {
			// Addresses match. Check SSA numbers in case we have a Phi DEF pseudo-address (multiple USE SSA #s).
			assert(AlreadyIter->second < this->TempShadowList.GetSize());
			assert(CriticalOp.second < this->TempShadowList.GetSize());
			if (this->TempShadowList.GetRefNum(AlreadyIter->second)->GetSSANum() == this->TempShadowList.GetRefNum(CriticalOp.second)->GetSSANum()) {
				// Addresses and SSA numbers match. Check operands pointed to.
				if (IsEqOp(this->TempShadowList.GetRefNum(AlreadyIter->second)->GetOp(),
					this->TempShadowList.GetRefNum(CriticalOp.second)->GetOp())) {
					found = true;
					break;
				}
			}
		}
	}
	return found;
} // end of SMPFunction::IsAlreadyShadowed()

// Trace CriticalOp via copies back to ShadowAddr, return false if no valid means of shadowing it.
//  Add new CriticalOp shadow-checking points to NewCriticalOps, which will be shadowed with separate indices.
//  If TracingMemWrite, terminate early when we can prove that CriticalOp was unsafe to use as address reg in memory write.
bool SMPFunction::FindShadowingPoint2(const ShadowPoint CriticalOp, const bool TracingMemWrite, ShadowSet &ShadowAddrSet, bool &MemUnsafe, ShadowSet &NewCriticalOps, bool &NonConstSourceFound, set<STARS_uval_t> &ConstValues) {
	bool UnsafeChain = false;
	STARS_ea_t ShadowCheckAddr = CriticalOp.first;
	STARSOpndTypePtr ShadowCheckUseOp = this->TempShadowList.GetRefNum(CriticalOp.second)->GetOp();

	bool DataFlowOpnd = MDIsDataFlowOpnd(ShadowCheckUseOp, this->UsesFramePointer());
	bool Duplicate = this->IsAlreadyShadowed(CriticalOp);
	if (TracingMemWrite && (!DataFlowOpnd))
		return false;
	else if ((!DataFlowOpnd) || Duplicate) {
		// end the recursion.
		bool ImmedOp = ShadowCheckUseOp->IsImmedOp();
		if (!Duplicate && (!ImmedOp))
			NonConstSourceFound = true;
		else if (ImmedOp) {
			// Save in the set of constant values we traced back to.
			pair<set<STARS_uval_t>::iterator, bool> InsertResult = ConstValues.insert(ShadowCheckUseOp->GetImmedValue());
			assert(InsertResult.first != ConstValues.end());
		}
		return true;  // if recursion has not started, ShadowAddrSet will be empty and no annotations will be emitted.
	}
	else {
		this->AlreadyShadowed.insert(CriticalOp);
	}

	bool ValidShadowing = true;
	bool ShadowPointFound = false;
	STARSOpndTypePtr UseOp = ShadowCheckUseOp;
	bool UseFP = this->UsesFramePointer();

	do {
		int CheckBlockNum;
		if (STARS_IsBlockNumPseudoID(ShadowCheckAddr)) {
			CheckBlockNum = STARS_GetBlockNumFromPseudoID(ShadowCheckAddr);
		}
		else {
			CheckBlockNum = this->GetBlockNumFromInstAddr(ShadowCheckAddr);
		}
		assert(0 <= CheckBlockNum);
		assert(CheckBlockNum < (decltype(CheckBlockNum))this->RPOBlocks.size());
		SMPBasicBlock *CurrBlock = this->RPOBlocks[(size_t) CheckBlockNum];

		bool LocalName = (!this->IsGlobalName(UseOp));
		STARS_ea_t CurrShadowAddr = CurrBlock->GetDefAddrFromUseAddr(UseOp, ShadowCheckAddr, this->TempShadowList.GetRefNum(CriticalOp.second)->GetSSANum(), LocalName);
		if (STARS_BADADDR == CurrShadowAddr) {
			ValidShadowing = false;
			break; // cannot follow chain any higher, but previous chain info is still valid.
		}
		if (STARS_IsBlockNumPseudoID(CurrShadowAddr)) { // must have a Phi DEF
			int ShadowBlockNum = STARS_GetBlockNumFromPseudoID(CurrShadowAddr);
			assert((0 <= ShadowBlockNum) && ((decltype(ShadowBlockNum))this->GetNumBlocks() > ShadowBlockNum));
			CurrBlock = this->RPOBlocks[(size_t) ShadowBlockNum];
			PhiSetIter PhiIter = CurrBlock->FindPhi(UseOp);
			assert(PhiIter != CurrBlock->GetLastPhi());
			// For each Phi USE, recurse.
			ShadowSet TempShadowAddrSet; // only union these into ShadowAddrSet if all Phi USE recursions succeed.

			for (size_t PhiIndex = 0; PhiIndex < PhiIter->GetPhiListSize(); ++PhiIndex) {
				int PhiUseSSANum = PhiIter->GetUseSSANum(PhiIndex);
				if (0 > PhiUseSSANum) {
					SMP_msg("ERROR: FindShadowingPoint2() cannot trace uninitialized Phi USE at CurrShadowAddr %p ", CurrShadowAddr);
					this->DumpFuncNameAndAddr();
					ValidShadowing = false;
					break;
				}
				this->TempShadowList.SetRef(UseOp->clone(), PhiIter->GetUseType(PhiIndex), PhiUseSSANum);
				ShadowPoint PhiUsePoint(CurrShadowAddr, this->TempShadowList.GetSize() - 1);
				ValidShadowing = this->FindShadowingPoint2(PhiUsePoint, TracingMemWrite, TempShadowAddrSet, MemUnsafe, NewCriticalOps, NonConstSourceFound, ConstValues);
				if (!ValidShadowing) {
					break;  // we must succeed on all Phi USEs
				}
			}
			if (ValidShadowing) { // All Phi USEs succeeded
				//  Remove current CriticalOp from ShadowAddrSet (if present); let the PhiUseOps logically take its place
				ShadowSet::iterator OldIter = ShadowAddrSet.find(CriticalOp);
				if (OldIter != ShadowAddrSet.end()) { // found it
					OldIter = ShadowAddrSet.erase(OldIter);
				}
				// Insert the shadow points for all the Phi USEs now that all Phi USE recursions succeeded.
#if 0
				ShadowAddrSet.insert(TempShadowAddrSet.begin(), TempShadowAddrSet.end());
#else
				for (ShadowSet::const_iterator TempIter = TempShadowAddrSet.cbegin(); TempIter != TempShadowAddrSet.cend(); ++TempIter) {
					pair<ShadowSet::iterator, bool> InsertResult = ShadowAddrSet.insert(*TempIter);
				}
#endif
			}
			break; // success if ValidShadowing is true, failure otherwise
		}
		else if (STARS_IsSSAMarkerPseudoID(CurrShadowAddr)) {
			NonConstSourceFound = true;
			if (MDIsStackAccessOpnd(UseOp, this->UsesFramePointer())) {
				// We need to guard against the corner case of a stack location that is
				//  apparently used before it is defined. We cannot emit FPTRSHADOW [RSP-128]
				//  at the top of the function, as that is a garbage value before the stack frame
				//  is allocated.
				int SignedOffset = (int) UseOp->GetAddr();
				if (SignedOffset < 0) {
					ValidShadowing = false;
					break;
				}
			}
			list<SMPInstr *>::iterator FirstInstIter = this->GetFirstInstIter();
			SMPInstr *CurrInst = (*FirstInstIter);
			STARSDefUseIter NewDefIter = CurrInst->FindDef(UseOp);
			assert(NewDefIter != CurrInst->GetLastDef());
			if (CurrInst->IsLastInBlock()) {
				// Bizarre corner case that might never happen; first block has only SSA marker. We cannot shadow here.
				ValidShadowing = false;
				break;
			}
			// We want to replace our current CriticalOp with the DEF at the addr of the inst after the SSA marker.
			//  Remove current CriticalOp from ShadowAddrSet (if present) and add new DEF in its place.
			ShadowSet::iterator OldIter = ShadowAddrSet.find(CriticalOp);
			if (OldIter != ShadowAddrSet.end()) { // found it
				OldIter = ShadowAddrSet.erase(OldIter);
			}
			++FirstInstIter; // move to first real inst
			CurrShadowAddr = (*FirstInstIter)->GetAddr();
			this->TempShadowList.SetRef(NewDefIter->GetOp(), NewDefIter->GetType(), NewDefIter->GetSSANum());
			ShadowPoint NextCriticalOp(CurrShadowAddr, this->TempShadowList.GetSize() - 1);

			// If we are tracing, see if chain is maybe-aliased.
			if (NewDefIter->HasIndirectWrite()) {
				MemUnsafe = true;
			}

			pair<ShadowSet::iterator, bool> InsertResult = ShadowAddrSet.insert(NextCriticalOp);
			if (InsertResult.second) { // not already shadowing that DEF
				ShadowPointFound = true;
			}
			// We are finished following the chain. Cannot go higher than the SSA Marker inst.
			break;
		}
		else { // must be an instruction address
			SMPInstr *CurrInst = this->GetInstFromAddr(CurrShadowAddr);
			assert(nullptr != CurrInst);
			if (DEFAULT != CurrInst->GetDataFlowType()) {
				ValidShadowing = false;
				break; // cannot follow chain any higher, but previous chain info is still valid.
			}
			STARSOpndTypePtr NewUseOp = nullptr;
			// If we have a conditional move instruction, then there will be two USEs. One is the
			//  operand that will be copied to the DEF if the condition is true, and the other is
			//  the operand that represents the unchanged DEF if the copy does not happen, e.g.
			//  if (guard) RDX5 := RAX3;  will have USEs RAX3 and RDX4, where RDX4 is the value
			//  that becomes RDX5 when the guard is not true. We need to treat this instruction as
			//  a Phi DEF with two Phi USEs and trace both USEs.
			if (CurrInst->MDIsConditionalMoveInstr()) {
				// For each USE other than the flags register, recurse.
				ShadowSet TempShadowAddrSet; // only union these into ShadowAddrSet if all Phi USE recursions succeed.
				for (STARSDefUseIter UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
					int UseSSANum = UseIter->GetSSANum();
					if (0 > UseSSANum) {
						SMP_msg("ERROR: FindShadowingPoint2() cannot trace uninitialized CondMove USE at CurrShadowAddr %p ", CurrShadowAddr);
						this->DumpFuncNameAndAddr();
						ValidShadowing = false;
						break;
					}
					STARSOpndTypePtr CondMoveUseOp = UseIter->GetOp()->clone();
					bool FlagsReg = (CondMoveUseOp->IsRegOp() && (CondMoveUseOp->MatchesReg(MD_FLAGS_REG)));
					if (!FlagsReg) {
						this->TempShadowList.SetRef(CondMoveUseOp, UseIter->GetType(), UseSSANum);
						ShadowPoint UsePoint(CurrShadowAddr, this->TempShadowList.GetSize() - 1);
						ValidShadowing = this->FindShadowingPoint2(UsePoint, TracingMemWrite, TempShadowAddrSet, MemUnsafe, NewCriticalOps, NonConstSourceFound, ConstValues);
						if (!ValidShadowing) {
							break;  // we must succeed on all conditional move USEs
						}
					}
				}
				if (ValidShadowing) { // All USEs succeeded
					//  Remove current CriticalOp from ShadowAddrSet (if present); let the UseOps logically take its place
					ShadowSet::iterator OldIter = ShadowAddrSet.find(CriticalOp);
					if (OldIter != ShadowAddrSet.end()) { // found it
						OldIter = ShadowAddrSet.erase(OldIter);
					}
					// Insert the shadow points for all the USEs now that all USE recursions succeeded.
#if 0
					ShadowAddrSet.insert(TempShadowAddrSet.begin(), TempShadowAddrSet.end());
#else
					for (ShadowSet::const_iterator TempIter = TempShadowAddrSet.cbegin(); TempIter != TempShadowAddrSet.cend(); ++TempIter) {
						pair<ShadowSet::iterator, bool> InsertResult = ShadowAddrSet.insert(*TempIter);
					}
#endif
				}
				break; // success if ValidShadowing is true, failure otherwise
			}
			else if (!CurrInst->IsSimpleCopy(NewUseOp)) {
				NonConstSourceFound = true;
				// We could have something like [rsp+32] := [rsp+32] + 8.
				//  The DEF is the DEF of our CriticalOp, and we cannot trace any farther for this shadowing index.
				//  But the USE of the same operand could be a vulnerable DEF-USE chain that needs its own
				//  shadowing index.
				// We need to shadow the DEF right AFTER this instruction and then add a new chain to NewCriticalOps
				//  starting with the USE.

				// First, if we are tracing memory writes for safety, the chain has become unanalyzeable for safety
				//  due to the non-copy operation, unless it is a stack pointer copy.
				if (TracingMemWrite) {
					if (CurrInst->MDIsStackPointerCopy(this->UsesFramePointer())) {
						// Stack pointer copy, but not a simple copy. Must be load-effective-address copy.
						STARSOpndTypePtr LeaMemOp = CurrInst->GetLeaMemUseOp();
						if ((nullptr != LeaMemOp) && (!LeaMemOp->IsVoidOp())) {
							ValidShadowing = MDIsDirectStackAccessOpnd(LeaMemOp, this->UsesFramePointer());
						}
						else {
							ValidShadowing = true;
						}
					}
					else {
						ValidShadowing = false;
					}
					break;
				}
				if (CurrInst->IsLastInBlock()) {
					// Cannot insert shadowing instrumentation before next inst because next inst is reachable by
					//  some path other than fall-through from the current inst. DEF cannot be shadowed.
					ValidShadowing = false;
					break; // exit; we will shadow, if possible, just not as far as this DEF.
				}
				CurrBlock = CurrInst->GetBlock();  // move from ShadowCheckAddr up to CurrShadowAddr
				vector<SMPInstr *>::iterator NextInstIter = CurrBlock->GetInstIterFromAddr(CurrShadowAddr);
				assert(NextInstIter != CurrBlock->GetLastInst());
				++NextInstIter;
				assert(NextInstIter != CurrBlock->GetLastInst());

				//  Remove current CriticalOp from ShadowAddrSet (if present) and add UseOp in its place.
				ShadowSet::iterator OldIter = ShadowAddrSet.find(CriticalOp);
				if (OldIter != ShadowAddrSet.end()) { // found it
					OldIter = ShadowAddrSet.erase(OldIter);
				}
				STARSDefUseIter ShadowDefIter = CurrInst->FindDef(UseOp);
				assert(ShadowDefIter != CurrInst->GetLastDef());
				STARS_ea_t NextInstAddr = (*NextInstIter)->GetAddr();
				this->TempShadowList.SetRef(ShadowDefIter->GetOp(), ShadowDefIter->GetType(), ShadowDefIter->GetSSANum());
				pair<ShadowSet::iterator, bool> InsertResult = ShadowAddrSet.insert(ShadowPoint(NextInstAddr, this->TempShadowList.GetSize() - 1));
				if (!InsertResult.second) { // already following that chain
					break;  // nothing new to trace
				}
				ShadowPointFound = true;
				// If we are tracing, see if chain is maybe-aliased.
				if (ShadowDefIter->HasIndirectWrite()) {
					MemUnsafe = true;
				}
				// If we are starting a NewCriticalOps chain, we need to start with a USE of the same SSA name as the DEF.
				STARSDefUseIter NewUseIter = CurrInst->FindUse(UseOp);
				if (NewUseIter != CurrInst->GetLastUse()) {
					// We have a starting point for a new shadowing operation.
					this->TempShadowList.SetRef(NewUseIter->GetOp(), NewUseIter->GetType(), NewUseIter->GetSSANum());
					ShadowPoint NewCritOp(CurrShadowAddr, this->TempShadowList.GetSize() - 1);
					if (!this->IsAlreadyShadowed(NewCritOp)) {
						pair<ShadowSet::iterator, bool> InsertResult = NewCriticalOps.insert(NewCritOp);
						// We don't care whether that was a new insertion or not.
					}
				}
				break; // done with current chain
			}
			else { // simple copy instruction
				assert(nullptr != NewUseOp);
				if (NewUseOp->IsRegOp()) {
					CanonicalizeOpnd(NewUseOp);
				}
				STARSDefUseIter DefIter = CurrInst->GetFirstDef();
				assert(DefIter != CurrInst->GetLastDef());

				// In the mem-write tracing case, if we copied from a memory op that is not just a direct single-addr-reg or stack
				//  access, then we cannot analyze the chain any further. If we copied from a loop-variant value,
				//  then the memory write is unsafe.
				if (TracingMemWrite) {
					if (this->IsBlockInAnyLoop(CurrInst->GetBlock()->GetNumber()) && (!DefIter->IsInvariantForAllLoops())) {
						ValidShadowing = false;
						break;
					}
					else if (NewUseOp->IsMemOp()) {
						if (!MDIsDirectStackAccessOpnd(NewUseOp, this->UsesFramePointer())) {
							// The TracingMemWrite case differs from the shadowing case here.
							//  The shadowing case would try (and fail) to trace [eax] while the
							//  TracingMemWrite case is concerned with tracing eax, not [eax]. If eax is a 
							//  loop-invariant value and satisfies the other criteria for memory
							//  write safety, then it should point to the return address only if
							//  it is a copy of an unsafe pointer that was stored to memory. That
							//  other unsafe pointer will be detected elsewhere in the function, making
							//  the whole function unsafe, which is all that matters in the TracingMemWrite
							//  case. So, static mem ops and phrase ops with no offset and no unsafety
							//  in their address regs should be traced through their address regs.
#define STARS_AGGRESSIVE_MEM_TRACING 1
#if STARS_AGGRESSIVE_MEM_TRACING
							if (NewUseOp->IsStaticMemOp()) {
								NonConstSourceFound = true;
								ValidShadowing = true;
								break;
							}
							else if (NewUseOp->IsMemNoDisplacementOp()) {
								int BaseReg, IndexReg;
								uint16_t ScaleFactor;
								STARS_ea_t DummyOffset = 0;
								NewUseOp->MDExtractAddressFields(BaseReg, IndexReg, ScaleFactor, DummyOffset);
								assert(0 == DummyOffset);
								bool SingleAddressReg = ((IndexReg == STARS_x86_R_none) != (BaseReg == STARS_x86_R_none)); // logical XOR
								if (SingleAddressReg && (0 == ScaleFactor)) {
									STARS_regnum_t AddressReg;
									if (IndexReg == STARS_x86_R_none) {
										AddressReg = MDCanonicalizeSubReg((STARS_regnum_t) BaseReg);
									}
									else {
										AddressReg = MDCanonicalizeSubReg((STARS_regnum_t) IndexReg);
									}
									NewUseOp = CurrInst->MakeRegOpnd(AddressReg); // fall through to code to analyze NewUseOp
								}
								else {
									ValidShadowing = false;
									break;
								}
							}
#else
							ValidShadowing = false;
							break;
#endif
						}
					}
				}

				STARSDefUseIter NewUseIter = CurrInst->FindUse(NewUseOp);
				assert(NewUseIter != CurrInst->GetLastUse());
				// We want to replace our current CriticalOp with the NewUseOp at this addr.
				//  Remove current CriticalOp from ShadowAddrSet (if present) and add NewUseOp in its place.
				ShadowSet::iterator OldIter = ShadowAddrSet.find(CriticalOp);
				if (OldIter != ShadowAddrSet.end()) { // found it
					OldIter = ShadowAddrSet.erase(OldIter);
				}
				this->TempShadowList.SetRef(NewUseIter->GetOp(), NewUseIter->GetType(), NewUseIter->GetSSANum());
				ShadowPoint NextCriticalOp(CurrShadowAddr, this->TempShadowList.GetSize() - 1);
				pair<ShadowSet::iterator, bool> InsertResult = ShadowAddrSet.insert(NextCriticalOp);
				if (!InsertResult.second) { // already following that chain
					break;  // nothing new to trace
				}
				ShadowPointFound = true;
				// If we are tracing, see if chain is maybe-aliased.
				if (DefIter->HasIndirectWrite()) {
					MemUnsafe = true;
				}
				if (!MDIsDataFlowOpnd(NewUseOp, this->UsesFramePointer())) {
					// We are finished following the chain.
					if (!(NewUseOp->IsImmedOp())) {
						NonConstSourceFound = true;
					}
					else {
						// Save in the set of constant values we traced back to.
						pair<set<STARS_uval_t>::iterator, bool> InsertResult = ConstValues.insert(NewUseOp->GetImmedValue());
						assert(InsertResult.first != ConstValues.end());
					}
					break;
				}
				else { // recurse on NextCriticalOp
					ValidShadowing = this->FindShadowingPoint2(NextCriticalOp, TracingMemWrite, ShadowAddrSet, MemUnsafe, NewCriticalOps, NonConstSourceFound, ConstValues);
					break; // recursion handled rest of the chain
				}
			} // end if not SimpleCopy else ...
		} // end if blocknum else if SSA marker inst else [regular inst addr]
	} while (ValidShadowing);

	// Success if we found any shadow points before chain had to terminate in non-mem-write case.
	//  In mem-write case, must be ValidShadowing.
	bool success = TracingMemWrite ? ValidShadowing : (ValidShadowing || ShadowPointFound);
	if (!success) {
		NonConstSourceFound = true;
	}
	return success;
} // end of SMPFunction::FindShadowingPoint2()

// Try to find safe indirect memory writes
bool SMPFunction::AnalyzeMemWriteSafety(void) {
	bool FoundSafeMemWrites = false;
	for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
		if (CurrBlock->HasIndirectMemWrite()) {
			// Find all instructions with indirect mem writes.
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				bool LoopingWrite = CurrInst->MDHasAnyRepeatPrefix();
				if (CurrInst->HasIndirectMemoryWrite() && (!LoopingWrite)) {
					STARS_ea_t InstAddr = CurrInst->GetAddr();
					STARSOpndTypePtr DefOp = CurrInst->GetMemDef();
					int BaseReg, IndexReg;
					uint16_t ScaleFactor;
					STARS_ea_t offset;
					MDExtractAddressFields(DefOp, BaseReg, IndexReg, ScaleFactor, offset);
					// In order to be safe, we need to start conservatively with a single address reg
					//  that is not scaled and has no offset other than a stack offset.
					//  We want to be less conservative for StaticMemOps, e.g. 0x8049170[ebx]; if we can
					//  prove that [ebx] would be safe, consider the whole operand safe.
					bool SingleAddressReg = ((IndexReg == STARS_x86_R_none) != (BaseReg == STARS_x86_R_none)); // logical XOR
					if (SingleAddressReg && (0 == ScaleFactor)) {
						bool DirectStackOperand = MDIsDirectStackAccessOpnd(DefOp, this->UsesFramePointer());
						if (!DirectStackOperand) {
#if 0
							bool NonStaticOffset = (offset != 0) && (!DefOp->IsStaticMemOp());
							bool NonStackOffset = NonStaticOffset && (!DirectStackOperand);
							if (!NonStackOffset) {
#endif
								// It is possible that the memory write is safe. We must trace its address reg back
								//  to simple copies of loop-invariant DEFs or to an incoming arg to declare it safe
								//  from overwriting our return address on the stack.
								STARS_regnum_t AddressReg;
								if (IndexReg == STARS_x86_R_none) {
									AddressReg = MDCanonicalizeSubReg((STARS_regnum_t) BaseReg);
								}
								else {
									AddressReg = MDCanonicalizeSubReg((STARS_regnum_t) IndexReg);
								}
								STARSOpndTypePtr UseOp = CurrInst->MakeRegOpnd(AddressReg);
								STARSDefUseIter UseIter = CurrInst->FindUse(UseOp);
								assert(UseIter != CurrInst->GetLastUse());
								this->TempShadowList.clear();
								this->TempShadowList.SetRef(UseOp, UseIter->GetType(), UseIter->GetSSANum());
								STARS_ea_t ShadowCheckAddr = CurrInst->GetAddr();
								ShadowPoint CriticalOp(ShadowCheckAddr, 0); // Checking point at indirect write
								list<ShadowPoint> WorkList;
								WorkList.push_back(CriticalOp);
								this->AlreadyShadowed.clear();
								ShadowSet ShadowUses;      // dummy set of addr+USE pairs to shadow values that will be checked at ShadowCheckAddr
								ShadowSet NewCriticalOps;  // dummy worklist of new shadow USEs and their addresses
								bool MemUnsafe = false;
								bool NonConstSourceFound = false;
								set<STARS_uval_t> ConstValues;
								bool SafeWrite = this->FindShadowingPoint2(CriticalOp, true, ShadowUses, MemUnsafe, NewCriticalOps, NonConstSourceFound, ConstValues);
								if (SafeWrite) {
									assert(NewCriticalOps.empty());  // FindShadowingPoint2() should stop before adding to this set
									STARSDefUseIter DefIter = CurrInst->FindDef(DefOp);
									assert(DefIter != CurrInst->GetLastDef());
									DefIter = CurrInst->SetDefSafeMemWrite(DefOp);
									assert(DefIter != CurrInst->GetLastDef());
									FoundSafeMemWrites = true;
									SMP_msg("INFO: Found safe indirect write in AnalyzeMemWriteSafety at %p\n", InstAddr);
								}
#if 0
							}
#endif
						}
					}
				} // end if inst has indirect mem write
			} // end for all instructions in block
		} // end if block has indirect mem write
	} // end for all blocks in function

	return FoundSafeMemWrites;
} // end of SMPFunction::AnalyzeMemWriteSafety()

// Determine inst ID set that func could return to, including tail call issues; return true if set changes
bool SMPFunction::ComputeReturnTargets(bool FirstIteration) {
	bool changed = false;
	if ((!(this->FuncReturnsToCaller())) || this->IsLinkerStub()) {
		this->ReturnTargetsComputed = true;
	}
	else if (!this->ReturnTargetsComputed) {
		bool IncompleteCaller = false;
		if (FirstIteration) {
			for (set<STARS_ea_t>::const_iterator CallSitesIter = this->AllCallSites.cbegin(); CallSitesIter != this->AllCallSites.cend(); ++CallSitesIter) {
				// Get the fall-through from the calling inst
				STARS_ea_t CallAddr = (*CallSitesIter);
				// Weed out recursive calls that are jumps or branches.
				int CallBlockNum = this->GetBlockNumFromInstAddr(CallAddr);
				if (0 <= CallBlockNum) {
					// CallAddr is in the current function.
					SMPInstr *CallInst = this->RPOBlocks[CallBlockNum]->FindInstr(CallAddr);
					assert(nullptr != CallInst);
					if (CallInst->IsJumpOrBranchInstr()) {
						continue; // internal jump or branch recursion; no return points
					}
				}
				// Get Func info of the caller.
				STARS_Function_t *FuncInfo = SMP_get_func(CallAddr);
				if (nullptr == FuncInfo) {
					SMP_msg("SERIOUS WARNING: ComputeReturnTargets: Call location %llx not in a function.\n",
						(uint64_t) CallAddr);
					this->SetIsCalledFromOrphanedCode();
					// IncompleteCaller = true;
					continue;
				}
				// Get first addr in func and map to SMPFunction *.
				STARS_ea_t FirstAddrOfCaller = FuncInfo->get_startEA();
				SMPFunction *CallingFunc = this->GetProg()->FindFunction(FirstAddrOfCaller);
				assert(nullptr != CallingFunc);
				SMPInstr *CallingInst = CallingFunc->GetInstFromAddr(CallAddr);
				// Tail calls from foo() to bar() lead to all return targets of foo() being
				//  added to bar()'s return targets. 
				bool TailCallFlag = CallingInst->IsTailCall();
				if (TailCallFlag) {
					// Get return targets of tail caller, as we return to them.
					if (CallingFunc->AreReturnTargetsComputed()) {
						const set<STARS_ea_t> CallerReturnTargets = CallingFunc->GetReturnTargets();
						set<STARS_ea_t>::const_iterator TargIter = CallerReturnTargets.cbegin();
						while (TargIter != CallerReturnTargets.cend()) {
							pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->ReturnTargets.insert(*TargIter);
							++TargIter;
							changed |= InsertResult.second;
							InsertResult = this->TailReturnTargets.insert(*TargIter);
						}
					}
					else {
						IncompleteCaller = true;
						UnresolvedCallers.insert(CallingFunc);
					}
					if (CallingFunc->IsPossibleIndirectCallTarget()) {
						IncompleteCaller = true;
						UnresolvedCallers.insert(CallingFunc);
					}
					if (CallingFunc->IsCalledFromOrphanedCode() || CallingFunc->IsTailCallChainFromOrphanedCode()) {
						this->SetIsTailCallChainFromOrphanedCode();
					}
				}
				else {
					STARS_ea_t FallThroughAddr = CallingInst->GetFallThroughAddr();
					bool BadFallThrough = ((STARS_BADADDR == FallThroughAddr) || (STARS_LIVEIN_PSEUDO_ID == FallThroughAddr));
					if (BadFallThrough) { // error case; could not find fall-through
						this->ReturnTargetsComputed = true;
						UnresolvedCallers.clear();
						return false;
					}
					else { // Got good FallThroughAddr
						pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->ReturnTargets.insert(FallThroughAddr);
						changed |= InsertResult.second;
					}
				}
			}
		}
		else { // not FirstIteration; just look at unresolved tail call sites
			set<SMPFunction *>::iterator TailCallIter = UnresolvedCallers.begin();
			while (TailCallIter != UnresolvedCallers.end()) {
				SMPFunction *CallingFunc = (*TailCallIter);
				// Get return targets of tail caller, as we return to them.
				if (CallingFunc->AreReturnTargetsComputed()) {
					const set<STARS_ea_t> CallerReturnTargets = CallingFunc->GetReturnTargets();
					set<STARS_ea_t>::const_iterator TargIter = CallerReturnTargets.cbegin();
					while (TargIter != CallerReturnTargets.cend()) {
						pair<set<STARS_ea_t>::iterator, bool> InsertResult = this->ReturnTargets.insert(*TargIter);
						++TargIter;
						changed |= InsertResult.second;
					}
					TailCallIter = UnresolvedCallers.erase(TailCallIter);
				}
				else {
					IncompleteCaller = true;
					++TailCallIter;
				}
				if (CallingFunc->IsPossibleIndirectCallTarget()) {
					IncompleteCaller = true;
					UnresolvedCallers.insert(CallingFunc);
				}
				if (CallingFunc->IsCalledFromOrphanedCode() || CallingFunc->IsTailCallChainFromOrphanedCode()) {
					this->SetIsTailCallChainFromOrphanedCode();
				}
			}
		}
		this->ReturnTargetsComputed = (!IncompleteCaller);
	}
	else {
		UnresolvedCallers.clear();
	}

	return changed;
} // SMPFunction::ComputeReturnTargets()

// Emit all annotations for the function, including all per-instruction
//  annotations.
void SMPFunction::EmitAnnotations(FILE *AnnotFile, FILE *InfoAnnotFile) {
	// Emit annotation for the function as a whole.
	list<SMPBasicBlock *>::iterator BlockIter;
	SMPBasicBlock *CurrBlock;
	bool FuncHasProblems = ((!this->AnalyzedSP) || (!this->HasGoodRTLs()) || (this->HasUnresolvedIndirectCalls())
		|| (this->HasUnresolvedIndirectJumps()) || (this->HasSharedChunks()));

	if (this->StaticFunc) {
		SMP_fprintf(AnnotFile,	"%18llx %6zu FUNC LOCAL  %s ", (uint64_t) this->GetStartAddr(),
			this->Size, this->GetFuncName());
	}
	else {
		SMP_fprintf(AnnotFile,	"%18llx %6zu FUNC GLOBAL %s ", (uint64_t) this->GetStartAddr(),
			this->Size, this->GetFuncName());
	}
	FuncType RetAddrStatus = this->GetReturnAddressStatus();
	switch (RetAddrStatus)
	{
		case FUNC_UNKNOWN:
		{
			SMP_fprintf(AnnotFile, "FUNC_UNKNOWN ");
			break;
		}
		case FUNC_SAFE:
		{
			SMP_fprintf(AnnotFile, "FUNC_SAFE ");
			break;
		}
		case FUNC_UNSAFE:
		{
			SMP_fprintf(AnnotFile, "FUNC_UNSAFE ");
			break;
		}
		case FUNC_SAFE_IF_CALLEES_ARE_SAFE:
		{
			// Should only happen in FastDeadRegs mode.
			assert(global_STARS_program->ShouldSTARSDoFastDeadRegs());
			SMP_fprintf(AnnotFile, "FUNC_SAFE ");
			break;
		}
		default:
		{
			SMP_msg("FATAL ERROR: Return address status is %d ", (int)RetAddrStatus);
			this->DumpFuncNameAndAddr();
			assert(false);
		}
	}
	if (this->UseFP) {
		SMP_fprintf(AnnotFile, "USEFP ");
	}
	else {
		SMP_fprintf(AnnotFile, "NOFP ");
	}
	if (this->FuncInfo->HasReturnPoints() /*does_return()*/) {
		SMP_fprintf(AnnotFile, "RET ");
	}
	else {
		SMP_fprintf(AnnotFile, "NORET ");
	}

	if (this->IsLeaf())
		SMP_fprintf(AnnotFile, "FUNC_LEAF ");
	// Store the first return instruction's address
	// NOTE: This was useless info and we just print an address to keep the line format the same for easy parsing.
	SMP_fprintf(AnnotFile, "%18llx ", (uint64_t)(this->FuncInfo->get_startEA() + this->FuncInfo->GetFuncSize() - 1));

	if (this->IsLibFunc())
		SMP_fprintf(AnnotFile, "LIBRARY ");
	SMP_fprintf(AnnotFile, "\n");

	if (this->IsMultiEntry()) {
		global_STARS_program->IncrementMultiEntryFuncCounter();
	}

	// Emit annotations about how to restore register values
	SMP_fprintf(AnnotFile, "%18llx %6d FUNC FRAMERESTORE ", (uint64_t) this->GetStartAddr(), 0);
	for (int i = STARS_x86_R_ax; i <= global_STARS_program->GetSTARS_MD_LAST_SAVED_REG_NUM(); i++) {
		SMP_fprintf(AnnotFile, "%d %d %d ", i, this->SavedRegLoc[i], this->ReturnRegTypes[i]);
	}
	SMP_fprintf(AnnotFile, "ZZ\n");

	// Print type left in the return register.
	if (MD_RETURN_VALUE_REG != STARS_x86_R_none) {
		SMP_fprintf(InfoAnnotFile, "%18llx %6u FUNC RETURNTYPE ", (uint64_t) this->GetStartAddr(), this->Size);
		SMP_fprintf(InfoAnnotFile, "%s %d\n", MDGetRegNumName(MD_RETURN_VALUE_REG, global_STARS_program->GetSTARS_ISA_Bytewidth()), this->ReturnRegTypes[MD_RETURN_VALUE_REG]);
	}
	// Print types of incoming arguments, if any.
	if (this->GetIncomingArgCount() > 0) {
		std::size_t ArgIndex;
		std::size_t ArgLimit = this->MaxInArgIndex;
		assert((this->MaxInArgIndex + 1) >= this->InArgCount);
		if (ArgLimit >= STARS_MAX_ARGS_PER_FUNC) {
			ArgLimit = STARS_MAX_ARGS_PER_FUNC - 1;
		}
		SMP_fprintf(InfoAnnotFile, "%18llx %6u FUNC INARGS %4u  ", (uint64_t) this->GetStartAddr(), this->Size,
			ArgLimit + 1);
		for (ArgIndex = 0; ArgIndex <= ArgLimit; ++ArgIndex) {
			SMP_fprintf(InfoAnnotFile, "ARG%u %u ", ArgIndex, this->InArgTypes[ArgIndex]);
		}
		SMP_fprintf(InfoAnnotFile, "\n");
	}

	SMP_fprintf(AnnotFile, "%18llx %6d FUNC MMSAFENESS ", (uint64_t) this->GetStartAddr(), 0);
	if (!IsSpecSafe())
		SMP_fprintf(AnnotFile, "UNSAFE\n");
	else if (!IsSafe())
		SMP_fprintf(AnnotFile, "SPECSAFE\n");
	else {
		assert(IsSafe());
		SMP_fprintf(AnnotFile, "SAFE\n");
	}

	// If function has problems that limited our analyses, emit an information annotation so that
	//  other tools can be aware of which analyses will be sound.
	if (FuncHasProblems) {
		SMP_fprintf(InfoAnnotFile,	"%18llx %6zu FUNC PROBLEM %s ", (uint64_t) this->GetStartAddr(),
			this->Size, this->GetFuncName());
		if (!this->AnalyzedSP) {
			SMP_fprintf(InfoAnnotFile, "STACKANALYSIS ");
		}
		if (this->HasSharedChunks()) {
			SMP_fprintf(InfoAnnotFile, "CHUNKS ");
		}
		if (this->HasUnresolvedIndirectJumps()) {
			SMP_fprintf(InfoAnnotFile, "JUMPUNRESOLVED ");
		}
		if (this->HasUnresolvedIndirectCalls()) {
			SMP_fprintf(InfoAnnotFile, "CALLUNRESOLVED ");
		}
		if (!this->HasGoodRTLs()) {
			SMP_fprintf(InfoAnnotFile, "BADRTLS ");
		}
		SMP_fprintf(InfoAnnotFile, "\n");
	}

	// Print annotations identifying pointer args to vulnerable library
	//  functions (e.g. strcpy()) that trace back to malloc(size) where
	//  size is constant, which permits run-time instrumentation or
	//  transformation (e.g. transforming strcpy() to strncpy(), strcat()
	//  to strncat(), etc.)
	if (!this->BufferCallConstBufferArgsMap.empty()) {
		STARSBufferCallArgMap::const_iterator BufferArgsIter;
		for (BufferArgsIter = this->BufferCallConstBufferArgsMap.cbegin(); BufferArgsIter != this->BufferCallConstBufferArgsMap.cend(); ++BufferArgsIter) {
			STARS_ea_t LibCallAddr = BufferArgsIter->first;
			list<pair<size_t, STARS_uval_t> >::const_iterator ArgIter;
			SMPInstr *LibCallInst = this->GetInstFromAddr(LibCallAddr);
			// Put all constant buffer size ptrargs on same annotation line
			SMP_fprintf(InfoAnnotFile, "%18llx %6zu CALL BUFFERSIZE ARGS ", (uint64_t) LibCallAddr,
				LibCallInst->GetSize());
			for (ArgIter = BufferArgsIter->second.cbegin(); ArgIter != BufferArgsIter->second.cend(); ++ArgIter) {
				size_t ArgPos = ArgIter->first;
				STARS_uval_t BufSize = ArgIter->second;
				SMP_fprintf(InfoAnnotFile, "%3zu %llu ", ArgPos, (uint64_t) BufSize);
			}
			SMP_fprintf(InfoAnnotFile, "ZZ %s\n", LibCallInst->GetDisasm());
		}
	}

	if (!global_STARS_program->ShouldSTARSPerformFullAnalysis()) {
		list<SMPInstr *>::iterator InstIter = Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
		++InstIter;  // skip marker instruction
#endif
		bool AllocSeen = false; // Reached LocalVarsAllocInstr yet?
		bool DeallocTrigger = false;
		for ( ; InstIter != Instrs.end(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARS_ea_t addr = CurrInst->GetAddr();
			if (CurrInst->MDIsFloatNop()) {
				SMP_msg("WARNING: FloatNop not used as marker instruction at %llx\n", (uint64_t) addr);
			}
			SMP_fprintf(AnnotFile, "%18llx %6zu INSTR BELONGTO %llx \n",
				(uint64_t) addr, CurrInst->GetSize(), (uint64_t) GetStartAddr());
			if (this->LocalVarsAllocInstr == addr) {
				AllocSeen = true;
				if (this->NeedsStackReferent)
					this->EmitStackFrameAnnotations(AnnotFile, CurrInst);
				else {
					int OptType = CurrInst->GetOptType(); 
					if (5 == OptType) { // ADD or SUB
						// Prevent mmStrata from extending the caller's stack frame
						//  to include the new allocation.
						SMP_fprintf(AnnotFile, "%18llx %6d INSTR LOCAL SafeFrameAlloc %s \n",
							(uint64_t) addr, -1, CurrInst->GetDisasm());
					}
					else if (CurrInst->MDIsPushInstr()) {
						SMP_fprintf(AnnotFile, "%18llx %6d INSTR LOCAL NoWarn %s \n",
							(uint64_t) addr, -3, CurrInst->GetDisasm());
					}
					// mmStrata ignores the DATAREF annotations anyway, so even though
					//  they are not needed, emit them for use by Strata and other tools
					//  in other projects besides MEDS.
					this->EmitStackFrameAnnotations(AnnotFile, CurrInst);
				}
			}
			// If this is the instruction which deallocated space
			//  for local variables, we set a flag to remind us to 
			//  emit an annotation on the next instruction.
			// mmStrata wants the instruction AFTER the
			//  deallocating instruction, so that it processes
			//  the deallocation after it happens. It inserts
			//  instrumentation before an instruction, not
			//  after, so it will insert the deallocating
			//  instrumentation before the first POP of callee-saved regs,
			//  if there are any, or before the return, otherwise.
			if (addr == this->LocalVarsDeallocInstr) {
				DeallocTrigger = true;
			}
			else if (DeallocTrigger) { // Time for annotation
				SMP_fprintf(AnnotFile,	"%18llx %6lu DEALLOC STACK esp - %lu %s\n", (uint64_t) addr,
					(unsigned long) this->GetLocalVarsSize(), (unsigned long) this->GetLocalVarsSize(), CurrInst->GetDisasm());
				DeallocTrigger = false;
			}
			CurrInst->EmitAnnotations(this->UseFP, AllocSeen, this->NeedsStackReferent, AnnotFile, InfoAnnotFile);

			// Emit annotations for IDA Pro switch table analyses.
			SMPitype CurrDataFlow = CurrInst->GetDataFlowType();
			if ((CurrDataFlow == INDIR_JUMP) && (!this->IsLinkerStub())) { // exclude PLT stub INDIR_CALL
				if (!CurrInst->IsTailCall()) { // jmp [reg] is not a switch jump if it is a tail call.
					global_STARS_program->PrintAllCodeToCodeXrefs(addr, CurrInst->GetSize(), false, false);
				}
			}
			// print all addresses seen in this instruction                          
			global_STARS_program->PrintAllAddressTakenXrefs(addr, CurrInst->GetSize());

		} // end for all instructions

		// Free memory for the memory-constrained reduced analysis case.
		for (SMPInstr *CurrInst : this->Instrs) {
			if (nullptr != CurrInst) delete CurrInst;
		}

		for (SMPBasicBlock *CurrBlock : this->Blocks) {
			if (nullptr != CurrBlock) delete CurrBlock;
		}

		this->Instrs.clear();
		this->Blocks.clear();
		return;
	} // end if reduced analysis

#if 0 // Move to Phase 7A
	// Find and mark special cases that will affect the integer error and other security annotations.
	this->MarkSpecialNumericErrorCases();
#endif

	// Emit program CFG xref/indirect branch target annotations.
	if (this->FuncReturnsToCaller() && this->AreReturnTargetsComputed()) {
		this->EmitReturnTargetAnnotations();
	}
	else {
		if (!this->AreReturnTargetsComputed())
			SMP_msg("INFO: Return targets were not computed for func at %llx\n", (uint64_t) this->GetFirstFuncAddr());
		else if (!this->FuncReturnsToCaller())
			SMP_msg("INFO: Return targets not computed for non-returning func at %llx\n", (uint64_t) this->GetFirstFuncAddr());
	}

	// Loop through all instructions in the function.
	// Output optimization annotations for those
	//  instructions that do not require full computation
	//  of their memory metadata by the Memory Monitor SDT.
	list<std::size_t> LoopList; // for current block
	int CurrBlockNum = SMP_BLOCKNUM_UNINIT;
	list<SMPInstr *>::iterator InstIter = Instrs.begin();
#if SMP_USE_SSA_FNOP_MARKER
	++InstIter;  // skip marker instruction
#endif
	bool AllocSeen = false; // Reached LocalVarsAllocInstr yet?
	bool DeallocTrigger = false;
	bool PrefetchInstSeenLast = false; // inst before current inst was a prefetch
	bool UndefinedOpcodeSeenLast = false; // inst before current inst was an undefined opcode
	for (; InstIter != Instrs.end(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		STARS_ea_t addr = CurrInst->GetAddr();
		if (CurrInst->MDIsFloatNop()) {
			SMP_msg("WARNING: FloatNop not used as marker instruction at %llx\n", (uint64_t) addr);
		}
		CurrBlock = CurrInst->GetBlock();
		int BlockNum = CurrBlock->GetNumber();
		if (BlockNum != CurrBlockNum) {
			CurrBlockNum = BlockNum;
			if (0 < this->LoopCount) {
				LoopList.clear();
				this->BuildLoopList(BlockNum, LoopList);
			}
		}

		SMP_fprintf(AnnotFile, "%18llx %6zu INSTR BELONGTO %llx \n",
			(uint64_t) addr, CurrInst->GetSize(), (uint64_t) GetStartAddr());
		SMPitype CurrDataFlow = CurrInst->GetDataFlowType();
		bool IndirCallFlag = (CurrDataFlow == INDIR_CALL);
		if ((CurrDataFlow == INDIR_JUMP) || IndirCallFlag) {
			bool CompleteIndirCall = false;
			if (IndirCallFlag && (!this->ConstantIndirCalls.empty())) {
				CompleteIndirCall = (this->ConstantIndirCalls.find(addr) != this->ConstantIndirCalls.end());
			}
			global_STARS_program->PrintAllCodeToCodeXrefs(addr, CurrInst->GetSize(), IndirCallFlag, CompleteIndirCall);
		}
		// print any addresses that appear in this instruction                    
		global_STARS_program->PrintAllAddressTakenXrefs(addr, CurrInst->GetSize());

		if (this->LocalVarsAllocInstr == addr) {
			AllocSeen = true;
			if (this->NeedsStackReferent)
				this->EmitStackFrameAnnotations(AnnotFile, CurrInst);
			else {
				int OptType = CurrInst->GetOptType(); 
				if (5 == OptType) { // ADD or SUB
					// Prevent mmStrata from extending the caller's stack frame
					//  to include the new allocation.
					SMP_fprintf(AnnotFile, "%18llx %6d INSTR LOCAL SafeFrameAlloc %s \n",
						(uint64_t) addr, -1, CurrInst->GetDisasm());
				}
				else if (CurrInst->MDIsPushInstr()) {
					SMP_fprintf(AnnotFile, "%18llx %6d INSTR LOCAL NoWarn %s \n",
						(uint64_t) addr, -3, CurrInst->GetDisasm());
				}
				// mmStrata ignores the DATAREF annotations anyway, so even though
				//  they are not needed, emit them for use by Strata and other tools
				//  in other projects besides MEDS.
				this->EmitStackFrameAnnotations(AnnotFile, CurrInst);
			}
		}
		// If this is the instruction which deallocated space
		//  for local variables, we set a flag to remind us to 
		//  emit an annotation on the next instruction.
		// mmStrata wants the instruction AFTER the
		//  deallocating instruction, so that it processes
		//  the deallocation after it happens. It inserts
		//  instrumentation before an instruction, not
		//  after, so it will insert the deallocating
		//  instrumentation before the first POP of callee-saved regs,
		//  if there are any, or before the return, otherwise.
		if (addr == this->LocalVarsDeallocInstr) {
			DeallocTrigger = true;
		}
		else if (DeallocTrigger) { // Time for annotation
			SMP_fprintf(AnnotFile,	"%18llx %6lu DEALLOC STACK esp - %lu %s\n", (uint64_t) addr,
				(unsigned long) this->GetLocalVarsSize(), (unsigned long) this->GetLocalVarsSize(), CurrInst->GetDisasm());
			DeallocTrigger = false;
		}

#ifndef SMP_REDUCED_ANALYSIS
		if (this->StackPtrAnalysisSucceeded() && this->HasGoodRTLs() && !this->HasUnresolvedIndirectJumps()) {
			CurrInst->EmitTypeAnnotations(this->UseFP, AllocSeen, this->NeedsStackReferent, AnnotFile, InfoAnnotFile);
#ifdef STARS_EMIT_NUMERIC_ERROR_ANNOTS
			CurrInst->EmitIntegerErrorAnnotations(InfoAnnotFile, LoopList);
#endif
		}
		else
#endif
			CurrInst->EmitAnnotations(this->UseFP, AllocSeen, this->NeedsStackReferent, AnnotFile, InfoAnnotFile);

		if (CurrInst->HasReturnOpcode() && this->GetReturnAddressStatus() == FUNC_SAFE)
			CurrInst->EmitSafeReturn(AnnotFile);

		// Emit IBT annotations for instructions that fit computed-goto patterns in libc/glibc, such
		//  as prefetch instructions and the instructions that follow them (computed goto often chooses
		//  between going to the prefetch or jumping just past it, and IDA Pro cannot analyze these libc
		//  macro-generated computed gotos even if they are not orphaned code). Likewise, an undefined opcode
		//  often separates an indirect jump and its first target, so inst after undefined opcode is IBT.
		bool EmitIBTAnnotation = (PrefetchInstSeenLast || UndefinedOpcodeSeenLast);
		if (CurrInst->MDIsPrefetchOpcode()) {
			PrefetchInstSeenLast = true;
			UndefinedOpcodeSeenLast = false;
			EmitIBTAnnotation = true;
		}
		else if (CurrInst->MDIsUndefinedOpcode()) {
			UndefinedOpcodeSeenLast = true;
			PrefetchInstSeenLast = false;
		}
		else {
			PrefetchInstSeenLast = false;
			UndefinedOpcodeSeenLast = false;
		}
		if (EmitIBTAnnotation) {
			global_STARS_program->PrintUnknownCodeXref(addr, CurrInst->GetSize(), ZST_COMPUTEDGOTO);
		}
	}  // end for all instructions

	// Loop through all basic blocks and emit profiling request annotations
	//  for those blocks that have unsafe memory writes in them.
	this->SafeBlocks = 0;
	this->UnsafeBlocks = 0;
	for (BlockIter = this->Blocks.begin(); BlockIter != this->Blocks.end(); ++BlockIter) {
		CurrBlock = (*BlockIter);
		if (CurrBlock->MaybeAliasedWrite()) {
			++(this->UnsafeBlocks);
#if SMP_OPTIMIZE_BLOCK_PROFILING
			vector<SMPInstr *>::iterator CurrInst;
			CurrInst = CurrBlock->GetFirstInst();
			if ((*CurrInst)->IsMarkerInst())
				++CurrInst;
			STARS_ea_t addr = (*CurrInst)->GetAddr();
			SMP_fprintf(AnnotFile,	"%18llx %6d BLOCK PROFILECOUNT %s\n", (uint64_t)addr,
				(*CurrInst)->GetSize(), (*CurrInst)->GetDisasm());
#endif
		}
		else {
			++(this->SafeBlocks);
		}
	}
	
	if (global_STARS_program->ShouldSTARSShadowFuncPtrs() || global_STARS_program->ShouldSTARSMaximizeCFGImprovement()) {
		// Emit func pointer shadowing and/or CFG-improving annotations.
		this->EmitFuncPtrShadowingAnnotations2(InfoAnnotFile);
	}
	if (global_STARS_program->ShouldSTARSShadowCriticalArgs()) {
		// Emit shadowing annotations.
		this->EmitArgShadowingAnnotations(InfoAnnotFile);
	}

	LoopList.clear();

	return;
} // end of SMPFunction::EmitAnnotations()

// Emit Indirect Branch Target destinations for return instructions in this func.
void SMPFunction::EmitReturnTargetAnnotations(void) {
	std::bitset<1 + STARS_MAX_REGNUM_FOR_BITSETS> ReturnDeadRegs;
	bool RetDeadRegsComputed = false;

	if (!this->ReturnTargets.empty()) {
		bool ReturnProcessed = false;
		bool CompleteTargets = true;
		bool OrphanCodeProblem = (this->IsCalledFromOrphanedCode() || this->IsTailCallChainFromOrphanedCode());
		for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
			if (CurrBlock->HasReturn()) {
				vector<SMPInstr *>::reverse_iterator LastInstIter = CurrBlock->GetRevInstBegin();
				SMPInstr *ReturnInst = (*LastInstIter);
				STARS_ea_t ReturnInstAddr = ReturnInst->GetAddr();
				if (!ReturnInst->HasReturnOpcode()) {
					continue; // must be tail call with RETURN data flow type
				}
				size_t RetInstSize = 1;
				bool success = (!ReturnTargets.empty());
				if (success)
					ReturnProcessed = true;
				set<STARS_ea_t>::const_iterator RetTargIter;
				for (const STARS_ea_t RetTargetAddr : this->ReturnTargets) {
					bool TailCallFlag = (this->TailReturnTargets.find(RetTargetAddr) != this->TailReturnTargets.end());
					success &= global_STARS_program->PrintReturnInstXref(ReturnInstAddr, RetTargetAddr, RetInstSize, TailCallFlag);
				}
				if (success && (!(this->IsPossibleIndirectCallTarget() || this->MultipleEntryPoints || OrphanCodeProblem))) {
					global_STARS_program->PrintCodeToCodeXrefComplete(ReturnInstAddr, RetInstSize, this->ReturnTargets.size(), ZST_RETURN);
					++STARS_ReturnSetComplete;
					bool RetDeadRegsProblem = false;
					if (!RetDeadRegsComputed) {
						// Compute intersection of dead regs bitsets for each return target.
						for (const STARS_ea_t RetTargetAddr : this->ReturnTargets) {
							STARS_Function_t *FuncInfo = SMP_get_func(RetTargetAddr);
							if (nullptr == FuncInfo) {
								RetDeadRegsProblem = true;
								SMP_msg("ERROR: Return target ID %llx not in any function ", (uint64_t) RetTargetAddr);
								this->DumpFuncNameAndAddr();
								break; // cannot produce valid set
							}
							// Get first addr in func and map to SMPFunction *.
							STARS_ea_t FirstAddrOfCaller = FuncInfo->get_startEA();
							SMPFunction *CallingFunc = this->GetProg()->FindFunction(FirstAddrOfCaller);
							assert(nullptr != CallingFunc);
							SMPInstr *RetTargetInst = CallingFunc->GetInstFromAddr(RetTargetAddr);
							assert(nullptr != RetTargetInst);
							if (!RetDeadRegsComputed) { // First target, initialize by OR-ing into empty bitset.
								ReturnDeadRegs |= RetTargetInst->GetDeadRegsSet();
								RetDeadRegsComputed = true;
							}
							else { // Not first target, AND into bitset to get regs dead at all call sites.
								ReturnDeadRegs &= RetTargetInst->GetDeadRegsSet();
							}
						}
					}
					if (RetDeadRegsProblem) {
						ReturnDeadRegs.reset();
					}
					ReturnInst->SetDeadRegsSet(ReturnDeadRegs);
				}
				else {
					CompleteTargets = false;
					++STARS_ReturnSetIncomplete;
					if (!success)
						++STARS_ReturnSetIncompleteAddrRange;
					if (this->IsPossibleIndirectCallTarget())
						++STARS_ReturnSetIncompleteAddrTaken;
					if (this->MultipleEntryPoints)
						++STARS_ReturnSetIncompleteMultiEntry;
					if (this->IsCalledFromOrphanedCode())
						++STARS_ReturnSetIncompleteCalledFromOrphanCode;
					if (this->IsTailCallChainFromOrphanedCode())
						++STARS_ReturnSetIncompleteTailCallChainFromOrphanCode;
				}
			}
		} // end for all basic blocks in function
		if (CompleteTargets) {
			if (ReturnProcessed)
				++STARS_FuncReturnSetComplete;
		}
		else {
			++STARS_FuncReturnSetIncomplete;
		}
	}
	return;
} // end of SMPFunction::EmitReturnTargetAnnotations()

// return -1 if CopyInstAddr not found in InArgPointerCopyAddrs, InArg position # otherwise
int SMPFunction::FindInArgNumFromCopyAddr(STARS_ea_t CopyInstAddr) {
	int InArgPositionNum = -1; // error code
	for (size_t index = 0; index < this->InArgPointerCopyAddrs.size(); ++index) {
		set<STARS_ea_t>::const_iterator FindIter = this->InArgPointerCopyAddrs[index].find(CopyInstAddr);
		if (FindIter != this->InArgPointerCopyAddrs[index].cend()) { // found it; index holds GlobalNameIndex
			map<unsigned int, size_t>::const_iterator MapIter = this->GlobalNameIndexMapToInArgIndex.find((unsigned int) index);
			if (MapIter != this->GlobalNameIndexMapToInArgIndex.cend()) { // found it
				InArgPositionNum = (int) MapIter->second;
			}
			break;
		}
	}
	return InArgPositionNum;
} // end of SMPFunction::FindInArgNumFromCopyAddr()

// Debug output dump.
void SMPFunction::Dump(const bool SkeletonDump) {
	list<SMPBasicBlock *>::iterator CurrBlock;
	SMP_msg("Debug dump for function: %s\n", this->GetFuncName());
	if (STARS_BADADDR != this->LocalVarsAllocInstr) {
		SMP_msg("UseFP: %d  LocalVarsAllocInstr: %llx Block: %d\n", this->UseFP,
			(uint64_t) this->LocalVarsAllocInstr, this->GetBlockFromInstAddr(this->LocalVarsAllocInstr)->GetNumber());
	}
	else {
		SMP_msg("UseFP: %d\n", this->UseFP);
	}
	if (!this->ReturnTargets.empty()) {
		SMP_msg("Return targets: ");
		for (STARS_ea_t RetTargetAddr : this->ReturnTargets) {
			SMP_msg("%llx ", (uint64_t) RetTargetAddr);
		}
		SMP_msg("\n");
	}
	for (std::size_t index = 0; index < this->IDom.size(); ++index) {
		SMP_msg("IDOM for %zu: %d\n", index, this->IDom[index]);
	}
	for (std::size_t index = 0; index < this->DomTree.size(); ++index) {
		SMP_msg("DomTree for %zu: ", index);
		for (int DominatedBlockNum : this->DomTree[index].second) {
				SMP_msg("%d ", DominatedBlockNum);
		}
		SMP_msg("\n");
	}
	if (!SkeletonDump) {
		SMP_msg("Global names: \n");
		for (STARSOpndTypePtr GlobalName : this->GlobalNames) {
			SMP_msg("index: %u ", ExtractGlobalIndex(GlobalName));
			PrintListOperand(GlobalName);
			SMP_msg("\n");
		}
		SMP_msg("Blocks each name is defined in: \n");
		for (std::size_t index = 0; index < this->BlocksDefinedIn.size(); ++index) {
			SMP_msg("Name index: %zu Blocks: ", index);
			for (int BlockNum : this->BlocksDefinedIn[index]) {
				SMP_msg("%d ", BlockNum);
			}
			SMP_msg("\n");
		}
	}

	SMP_msg("Control flow types: \n");
	// Print the conditional control flows.
	for (pair<int, int> JumpNodePair : this->JumpFollowNodesMap) {
		STARSTransCFGBlock *TCFGBlock = this->GetTCFGBlockByNum(JumpNodePair.first);
		assert(nullptr != TCFGBlock);
		STARS_ea_t BranchAddr = TCFGBlock->GetLastAddr();
		map<STARS_ea_t, uint16_t>::const_iterator CFIter = this->ControlFlowMap.find(BranchAddr);
		if (CFIter != this->ControlFlowMap.cend()) {
			ControlFlowType LastCFType = (ControlFlowType)CFIter->second;
			if ((LastCFType == BRANCH_IF_THEN) || (LastCFType == BRANCH_IF_THEN_ELSE) || (LastCFType == SHORT_CIRCUIT_BRANCH)) {
				int FollowBlockNum = JumpNodePair.second;
				SMP_msg("Branch addr: %llx  CFType: %s Block: %d FollowBlock: %d\n",
					(uint64_t)CFIter->first, CFTTypeStrings[CFIter->second], this->GetBlockFromInstAddr(CFIter->first)->GetNumber(), FollowBlockNum);
			}
		}
	}
	// Now get the loop control flows.
	for (map<STARS_ea_t, uint16_t>::const_iterator CFIter = this->ControlFlowMap.cbegin(); CFIter != this->ControlFlowMap.cend(); ++CFIter) {
		ControlFlowType LastCFType = (ControlFlowType)CFIter->second;
		if (IsLoopFlow(LastCFType)) {
			SMP_msg("Branch addr: %llx  CFType: %s Block: %d\n", (uint64_t) CFIter->first, CFTTypeStrings[CFIter->second], this->GetBlockFromInstAddr(CFIter->first)->GetNumber());
		}
	}

	if (this->LoopCount > 0) {
		SMP_msg("Block numbers by loop number:\n");
		for (size_t LoopNum = 0; LoopNum < this->LoopCount; ++LoopNum) {
			SMP_msg("Loop %d : Loop Type: %d :", LoopNum, this->LoopTypesByLoopNum[LoopNum]);
			for (size_t BlockNum = 0; BlockNum < (size_t)this->BlockCount; ++BlockNum) {
				if (this->FuncBlocksByLoop[LoopNum].GetBit(BlockNum)) {
					SMP_msg(" %zu", BlockNum);
				}
			}
			if (!this->LoopExitTargets.empty()) {
				int FollowBlockNum = SMP_BLOCKNUM_UNINIT;
				bool MultiExitLoop = (1 < this->LoopExitTargets[LoopNum].size());
				if (!MultiExitLoop) {
					SMP_msg(" Follow block number: %d", *(this->LoopExitTargets[LoopNum].cbegin()));
				}
				else {
					SMP_msg(" Exit target block numbers: \n");
					for (int FollowBlockNum : this->LoopExitTargets[LoopNum]) {
						SMP_msg("%d ", FollowBlockNum);
						SMP_msg("Reachable descendants: ");
						if (!this->LoopExitTargetsReachabilitySets[LoopNum].empty()) {
							map<int, STARSBitSet>::const_iterator MapIter = this->LoopExitTargetsReachabilitySets[LoopNum].find(FollowBlockNum);
							if (MapIter != this->LoopExitTargetsReachabilitySets[LoopNum].cend()) {
								const STARSBitSet &CurrBitSet = this->LoopExitTargetsReachabilitySets[LoopNum][FollowBlockNum];
								CurrBitSet.DumpBitsSet();
							}
						}
						SMP_msg("\n");
					}
					map<int, set<int> >::const_iterator MapIter = this->MultiLoopExitTargetsConvergenceMap.find((int)LoopNum);
					if (MapIter != this->MultiLoopExitTargetsConvergenceMap.cend()) {
						SMP_msg(" Exit target convergence block numbers: ");
						for (int ConvergenceBlockNum : MapIter->second) {
							SMP_msg("%d ", ConvergenceBlockNum);
						}
						SMP_msg("\n");
					}
				}
			}
			SMP_msg("\n");
		}
		if (!this->LoopInductionVars.empty()) {
			SMP_msg("Induction vars by loop number:\n");
			for (size_t LoopNum = 0; LoopNum < this->LoopCount; ++LoopNum) {
				STARSInductionVarFamilyList IV = this->LoopInductionVars[LoopNum];
				for (STARSInductionVarFamilyList::const_iterator IVIter = IV.cbegin(); IVIter != IV.cend(); ++IVIter) {
					if ((*IVIter).BIVIncomingSSANum >= 0) {
						SMP_msg("\nLoop %d IVFamily: ", LoopNum);
						DumpInductionVarFamily(*IVIter);
					}
				}
			}
			SMP_msg("\n");
		}
		if (!this->LoopToGuardMap.empty()) {
			SMP_msg("Loop guards: ");
			for (pair<STARS_ea_t, STARS_ea_t> LoopGuardPair : this->LoopToGuardMap) {
				SMP_msg("\nLoop at %llx has guard at %llx", (uint64_t)LoopGuardPair.first, (uint64_t)LoopGuardPair.second);
			}
			SMP_msg("\n");
		}
	}

	if (!this->EdgeExprMap.empty()) {
		SMP_msg(" Edge expressions:\n");
		for (pair<uint32_t, size_t> EdgeMapEntry : this->EdgeExprMap) {
			uint32_t EdgeHashIndex = EdgeMapEntry.first;
			size_t EdgeVectorIndex = EdgeMapEntry.second;
			uint32_t TargetBlockNum = EdgeHashIndex & 0x0000ffff;
			uint32_t SourceBlockNum = ((EdgeHashIndex & 0xffff0000) >> 16);
			SMP_msg("  From block %u to block %u : ", SourceBlockNum, TargetBlockNum);
			assert(EdgeVectorIndex < this->EdgeExpressions.size());
			DumpLoopComparison(this->EdgeExpressions[EdgeVectorIndex]);
		}
		SMP_msg("\n");
	}
	
	if (global_STARS_program->ShouldSTARSTranslateToSPARKAda()) {
		for (size_t BlockIndex = 0; BlockIndex < this->ClonedShadowCFGBlocks.size(); ++BlockIndex) {
			STARSCFGBlock *CFGBlock = this->GetClonedCFGBlockByNum(BlockIndex);
			if (CFGBlock->IsCoalesced()) {
				int OrigBlockNum = CFGBlock->GetOriginalBlockNum();
				int OrigCoalescedBlockNum = CFGBlock->GetCoalescedBlockNum();
				SMP_msg("SPARK: Block %d has been coalesced into CFG block %d.\n", OrigBlockNum, OrigCoalescedBlockNum);
			}
		}
		if (!this->OddIfThenBranches.empty()) {
			SMP_msg("SPARK: OddIfThenCase branch block numbers: ");
			for (int BranchBlockNum : this->OddIfThenBranches) {
				SMP_msg("%d ", BranchBlockNum);
			}
			SMP_msg("\n");
		}
	}

	for (CurrBlock = this->Blocks.begin(); CurrBlock != this->Blocks.end(); ++CurrBlock) {
		// Dump out the block number and data flow sets before the instructions.
		(*CurrBlock)->Dump(SkeletonDump);
	}

	SMP_msg("End of debug dump for function: %s\n", this->GetFuncName());
	return;
} // end of SMPFunction::Dump()

// Dump CFG representation for processing by "dot" program.
void SMPFunction::DumpDotCFG(void) {
	bool SPARKMode = global_STARS_program->ShouldSTARSTranslateToSPARKAda();
	if (SPARKMode && (!this->PrintedCFG)) { // avoid redundant work
#if 0	// Permit redundancy because STARST file changes as cloning occurs.
		this->PrintedCFG = true; 
#endif
		string DotFileName(global_STARS_program->GetRootFileName());
		DotFileName += ".STARS.";
		char FuncBuf[20] = { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };
		// itoa(this->GetFirstFuncAddr(), FuncBuf, 16);
		SMP_snprintf(FuncBuf, 18, "%x", this->GetFirstFuncAddr());
		string FuncName("func");
		FuncName += string(FuncBuf);
		string FileExtension(".gv");
		DotFileName += FuncName;
		DotFileName += FileExtension;
		FILE *DotFile = SMP_fopen(DotFileName.c_str(), "w");
		if (nullptr == DotFile) {
			SMP_msg("ERROR: Cannot open dot file %s for CFG dump.\n", DotFileName.c_str());
			return;
		}

		SMP_fprintf(DotFile, "digraph %s {\n", FuncName.c_str());

		for (SMPBasicBlock *CurrBlock : this->RPOBlocks) {
			size_t NumSuccs = CurrBlock->GetNumSuccessors();
			if (NumSuccs > 0) {
				int CurrBlockNum = CurrBlock->GetNumber();
				bool CondBranch = CurrBlock->HasConditionalBranch();
				bool JumpBlock = CurrBlock->HasDirectJump();
				bool IndirJumpBlock = CurrBlock->HasIndirectJump();
				bool UnoptimizedBranch = ((1 == NumSuccs) && CondBranch); // e.g. jz $+2, branches to fall-through.
				bool ShortCircuitHead = (CondBranch && CurrBlock->IsShortCircuitHead());
				bool ShortCircuitNonHead = CurrBlock->IsShortCircuitNonHead();
				if (UnoptimizedBranch) {
					int SuccBlockNum = (*(CurrBlock->GetFirstConstSucc()))->GetNumber();
					// Two possibilities for a CondBranch with only one successor. One is
					//  the UnoptimizedBranch case, and the other is the conditional tail
					//  call or other cases of a branch to another function. We mark the former
					//  with a red edge, the latter with an orange edge.
					if (CurrBlock->HasBranchToOtherFunc()) {
						SMP_fprintf(DotFile, "  %d -> %d [color=orange]\n", CurrBlockNum, SuccBlockNum);
					}
					else {
						SMP_fprintf(DotFile, "  %d -> %d [color=red]\n", CurrBlockNum, SuccBlockNum);
					}
				}
				else if (ShortCircuitHead || ShortCircuitNonHead) {
					assert(CondBranch || JumpBlock);
					for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
						bool SuccNonHead = SuccBlock->IsShortCircuitNonHead();
						int SuccBlockNum = SuccBlock->GetNumber();
						int BranchTargetBlockNum = JumpBlock ? SuccBlockNum : CurrBlock->GetCondNonFallThroughSuccBlockNum();
						if (SuccNonHead) { // Show short circuit edge with green line
							SMP_fprintf(DotFile, "  %d -> %d [color=green]\n", CurrBlockNum, SuccBlockNum);
						}
						else if (SuccBlockNum == BranchTargetBlockNum) {
							SMP_fprintf(DotFile, "  %d -> %d [color=blue]\n", CurrBlockNum, SuccBlockNum);
						}
						else {
							string ColorString = JumpBlock ? "purple" : "black";
							SMP_fprintf(DotFile, "  %d -> %d [color=%s]\n", CurrBlockNum, SuccBlockNum, ColorString.c_str());
						}
					}
				}
				else if (CondBranch || JumpBlock) { // Color the non-fall-through edge
					for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
						int SuccBlockNum = SuccBlock->GetNumber();
						int BranchTargetBlockNum = JumpBlock ? SuccBlockNum : CurrBlock->GetCondNonFallThroughSuccBlockNum();
						string ColorString = JumpBlock ? "purple" : "black";
						if (SuccBlockNum == BranchTargetBlockNum) {
							if (CondBranch) {
								ColorString = "blue";
								assert(!JumpBlock);
							}
						}
						SMP_fprintf(DotFile, "  %d -> %d [color=%s]\n", CurrBlockNum, SuccBlockNum, ColorString.c_str());
					}
				}
				else { // No special cases
					string ColorString = IndirJumpBlock ? "purple" : "black";
					for (SMPBasicBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
						SMP_fprintf(DotFile, "  %d -> %d [color=%s]\n", CurrBlockNum, SuccBlock->GetNumber(), ColorString.c_str());
					}
				}
			}
		} // end for all blocks
		SMP_fprintf(DotFile, "}\n");
		(void)SMP_fclose(DotFile);
	} // end for all blocks

	if (SPARKMode && (this->TCFGBlocks.size() != this->RPOBlocks.size())) {
		this->PrintedTCFG = true; // We don't check this because multiple clonings might need printing
		string DotFileName(global_STARS_program->GetRootFileName());
		DotFileName += ".STARST."; // add T for TCFG file
		char FuncBuf[20] = { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };
		// itoa(this->GetFirstFuncAddr(), FuncBuf, 16);
		SMP_snprintf(FuncBuf, 18, "%x", this->GetFirstFuncAddr());
		string FuncName("func");
		FuncName += string(FuncBuf);
		string FileExtension(".gv");
		DotFileName += FuncName;
		DotFileName += FileExtension;
		FILE *DotFile = SMP_fopen(DotFileName.c_str(), "w");
		if (nullptr == DotFile) {
			SMP_msg("ERROR: Cannot open dot file %s for CFG dump.\n", DotFileName.c_str());
			return;
		}

		SMP_fprintf(DotFile, "digraph %s {\n", FuncName.c_str());

		for (STARSTransCFGBlock *CurrBlock : this->TCFGBlocks) {
			size_t NumSuccs = CurrBlock->GetNumSuccessors();
			if (NumSuccs > 0) {
				int CurrBlockNum = CurrBlock->GetNumber();
				int CurrOrigBlockNum = CurrBlock->GetOriginalBlockNum();
				bool CondBranch = CurrBlock->HasConditionalBranch();
				bool JumpBlock = CurrBlock->HasDirectJump();
				bool UnoptimizedBranch = ((1 == NumSuccs) && CondBranch); // e.g. jz $+2, branches to fall-through.
				bool BranchOutsideFunc = UnoptimizedBranch && CurrBlock->HasBranchToOtherFunc();
				string ColorString = JumpBlock ? "purple" : "black";
				if (BranchOutsideFunc) {
					ColorString = "orange";
				}
				else if (UnoptimizedBranch) {
					ColorString = "red";
				}
				if (CurrBlockNum == CurrOrigBlockNum) {
					for (STARSTransCFGBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
						int SuccBlockNum = SuccBlock->GetNumber();
						int SuccOrigBlockNum = SuccBlock->GetOriginalBlockNum();
						if (!UnoptimizedBranch && (SuccBlockNum == CurrBlock->GetCondNonFallThroughSuccBlockNum())) {
							ColorString = "blue";
						}
						if (SuccBlockNum == SuccOrigBlockNum) {
							SMP_fprintf(DotFile, "  %d -> %d [color=%s]\n", CurrBlockNum, SuccBlockNum, ColorString.c_str());
						}
						else {
							SMP_fprintf(DotFile, "  %d -> %d.%d [color=%s]\n", CurrBlockNum, SuccBlockNum, SuccOrigBlockNum, ColorString.c_str());
						}
					}
				}
				else { // Make original block nums visible
					for (STARSTransCFGBlock *SuccBlock : CurrBlock->GetConstSuccList()) {
						int SuccBlockNum = SuccBlock->GetNumber();
						int SuccOrigBlockNum = SuccBlock->GetOriginalBlockNum();
						if (!UnoptimizedBranch && (SuccBlockNum == CurrBlock->GetCondNonFallThroughSuccBlockNum())) {
							ColorString = "blue";
						}
						if (SuccBlockNum == SuccOrigBlockNum) {
							SMP_fprintf(DotFile, "  %d.%d -> %d [color=%s]\n", CurrBlockNum, CurrOrigBlockNum, SuccBlockNum, ColorString.c_str());
						}
						else {
							SMP_fprintf(DotFile, "  %d.%d -> %d.%d [color=%s]\n", CurrBlockNum, CurrOrigBlockNum, SuccBlockNum, SuccOrigBlockNum, ColorString.c_str());
						}
					}
				}
			}
		}
		SMP_fprintf(DotFile, "}\n");
		(void)SMP_fclose(DotFile);

		// Dump out the ClonedShadowCFG so we can check its sanity against the TCFG.
		string DotFileNameC(global_STARS_program->GetRootFileName());
		DotFileNameC += ".STARSC."; // add T for TCFG file
		string FuncNameC("func");
		FuncNameC += string(FuncBuf);
		DotFileNameC += FuncNameC;
		DotFileNameC += FileExtension;
		FILE *DotFileC = SMP_fopen(DotFileNameC.c_str(), "w");
		if (nullptr == DotFileC) {
			SMP_msg("ERROR: Cannot open dot file %s for ShadowCFG dump.\n", DotFileNameC.c_str());
			return;
		}

		SMP_fprintf(DotFileC, "digraph %s {\n", FuncNameC.c_str());

		for (STARSCFGBlock *CurrBlock : this->ClonedShadowCFGBlocks) {
			int CurrBlockNum = CurrBlock->GetBlockNum();
			STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum((size_t)CurrBlockNum);
			if (CurrTCFGBlock->GetNumSuccessors() > 0) {
				if (CurrBlock->IsExprHeadBlock()) { // just print FT and NFT blocks; coalesced expr blocks are omitted
					int FTBlockNum = CurrBlock->GetExpr()->GetFallThroughBlockNum();
					int NFTBlockNum = CurrBlock->GetExpr()->GetNonFallThroughBlockNum();
					SMP_fprintf(DotFileC, "  %d -> %d [color=green]\n", CurrBlockNum, FTBlockNum);
					SMP_fprintf(DotFileC, "  %d -> %d [color=green]\n", CurrBlockNum, NFTBlockNum);
				}
				else if (!CurrBlock->IsCoalesced()) { // just print TCFG
					for (STARSTransCFGBlock *SuccBlock : CurrTCFGBlock->GetConstSuccList()) {
						SMP_fprintf(DotFileC, "  %d -> %d\n", CurrBlockNum, SuccBlock->GetNumber());
					}
				}
			}
		}
		SMP_fprintf(DotFileC, "}\n");
		(void)SMP_fclose(DotFileC);
	}

	return;
} // end of SMPFunction::DumpDotCFG()

#define STARS_DEBUG_DUPLICATE_SEARCHES 1

// Is DefOp+DefSSANum at DefAddr used as address reg or as source operand in memory write?
bool SMPFunction::IsDefUsedInUnsafeMemWrite(STARSOpndTypePtr DefOp, int DefSSANum, STARS_ea_t DefAddr) {
	bool FoundMemWriteUse = false;

	list<pair<pair<STARSOpndTypePtr, int>, STARS_ea_t> > DefWorkList;
	pair<STARSOpndTypePtr, int> InitialSearchDef(DefOp, DefSSANum);
	pair<pair<STARSOpndTypePtr, int>, STARS_ea_t> InitialItem(InitialSearchDef, DefAddr);
	DefWorkList.push_back(InitialItem);
	set<pair<STARS_ea_t, int>, LessSearchOperand> AlreadySearchedSet;

	do {  // process DefWorkList
		pair<pair<STARSOpndTypePtr, int>, STARS_ea_t> HeadItem = DefWorkList.front();
		DefOp = HeadItem.first.first;
		DefSSANum = HeadItem.first.second;
		DefAddr = HeadItem.second;
		DefWorkList.pop_front();
		assert(!DefOp->IsMemOp());

		if (!(MDIsStackPtrReg(DefOp->GetReg(), this->UsesFramePointer()))) {
			// Ensure that we don't have a data dependence loop that causes
			//  duplicate searches.
			int DefHashIndex = HashGlobalNameAndSSA(DefOp, DefSSANum);
			pair<STARS_ea_t, int> SearchItem(DefAddr, DefHashIndex);
			pair<set<pair<STARS_ea_t, int>, LessSearchOperand>::iterator, bool> InsertResult = AlreadySearchedSet.insert(SearchItem);
			if (InsertResult.second) { // not previously in the AlreadySearchedSet
				// Find DefBlock and get the recursion started.
				SMPBasicBlock *DefBlock = this->GetBlockFromInstAddr(DefAddr);
				assert(nullptr != DefBlock);
				this->ResetProcessedBlocks();
				FoundMemWriteUse = DefBlock->IsDefUsedInUnsafeMemWrite(DefWorkList, DefOp, DefSSANum, DefAddr);
			}
#if STARS_DEBUG_DUPLICATE_SEARCHES
			else {
				SMP_msg("INFO: Duplicate search avoided at %llx for DefHashIndex %x\n", (uint64_t) DefAddr, DefHashIndex);
			}
#endif
		}

	} while (!(FoundMemWriteUse || DefWorkList.empty()));

	DefWorkList.clear();
	AlreadySearchedSet.clear();
	return FoundMemWriteUse;
} // end of SMPFunction::IsDefUsedInUnsafeMemWrite()

// Can the return address of any caller be read or written directly from this function?
bool SMPFunction::IsCallerReturnAddressReadOrWritten(void) {
	set<STARS_ea_t>::iterator CallerIter;
	bool CallerVulnerable = false;
	for (CallerIter = this->AllCallSources.begin(); CallerIter != this->AllCallSources.end(); ++CallerIter) {
		STARS_ea_t CallerFirstAddr = (*CallerIter); // we store first addrs of funcs that call us, not the call inst addrs, in AllCallSources.
		SMPFunction *CallerFunc = this->GetProg()->FindFunction(CallerFirstAddr);
		if ((nullptr != CallerFunc) && CallerFunc->StackPtrAnalysisSucceeded()) {
			if (CallerFunc->GetLocalVarsSize() < ((STARS_asize_t) this->GetMaxDirectStackAccessDelta())) {
				// We are reading or writing beyond the caller stack frame, perhaps just to
				//  the saved regs above the caller stack frame, but maybe to the caller's
				//  return address.
				SMP_msg("INFO: Func at %llx accesses above the local vars of caller func at %llx\n", 
					(uint64_t) this->GetFirstFuncAddr(), (uint64_t) CallerFunc->GetFirstFuncAddr());
				CallerVulnerable = true;
			}
		}
	}
	return CallerVulnerable;
} // end of SMPFunction::IsCallerReturnAddressReadOrWritten()

// Is UseOp+UseSSANum at UseAddr a safe write?
bool SMPFunction::IsAddressRegSafe(const STARSOpndTypePtr &UseOp, STARS_ea_t UseAddr, int UseSSANum) {
	bool SafeReg = false;
	if (!this->HasGoodSSAForm() || (0 > UseSSANum))
		return SafeReg;

	bool GlobalName = this->IsGlobalName(UseOp);
	SMPBasicBlock *UseBlock = this->GetBlockFromInstAddr(UseAddr);
	STARSOpndTypePtr DefMoveOp = nullptr;
	STARS_ea_t DefAddr = UseBlock->GetUltimateDefAddr(UseOp, UseAddr, UseSSANum, !GlobalName, true, false, DefMoveOp);

	// To be safe, the UseAddr needs to track back to a move instruction that is a stack pointer copy,
	//  or a load effective address using a direct stack address, or an incoming arg or incoming reg
	//  that can be similarly traced to a direct stack address in all callers.
	bool LiveIntoFunc = STARS_IsSSAMarkerPseudoID(DefAddr);
	bool PhiDef = STARS_IsBlockNumPseudoID(DefAddr);
	bool BadDefAddr = (STARS_BADADDR == DefAddr);
	bool LiveIntoBlock = STARS_IsLiveInPseudoID(DefAddr);
	if (LiveIntoFunc) {
		if (!(this->PossibleIndirectCallTarget || this->MultipleEntryPoints)) {
			// TODO: Search all callers to analyze their uses of DefOp.
			;
		}
	}
	else if (!(PhiDef || BadDefAddr || LiveIntoBlock)) {
		SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
		assert(nullptr != DefInst);
		bool StackPtrCopy = DefInst->MDIsStackPointerCopy(this->UsesFramePointer());
		if (StackPtrCopy) {
			// Safe unless we have "lea reg,[esp+ecx]" so search for index reg
			STARSOpndTypePtr LeaMemOp = DefInst->GetLeaMemUseOp();
			if ((nullptr != LeaMemOp) && (!LeaMemOp->IsVoidOp())) {
				if (MDIsDirectStackAccessOpnd(LeaMemOp, this->UsesFramePointer())) {
					SafeReg = true;
				}
			}
			else {
				// Not lea inst; safe stack pointer copy
				SafeReg = true;
			}
		}
	}

	return SafeReg;
} // end of SMPFunction::IsAddressRegSafe()

// Analyzes the function to see if the return address can be marked as safe 
void SMPFunction::MarkFunctionSafe() {
#if SMP_DEBUG_FUNC
	SMP_msg(" Analyzing function %s and isLeaf = %d \n ", this->GetFuncName(), this->IsLeaf());
#endif
	bool HasCallTargets = false;
	bool HasStackPointerCopy = false;
	bool HasStackPointerPush = false;
	bool HasIndirectGlobalWrite = false;
	bool WritesAboveLocalFrame = false;		// Direct writes above local frame
	bool WritesAboveLocalFrameIndirect = false;	// Indirect writes above local frame
	bool HasIndexedStackWrite = false;
	bool HasUnsafeIndirectWrite = false;
	bool IsIndirectCallTarget = this->PossibleIndirectCallTarget; // could be called indirectly
	bool IsTailCallTarget = false; // could be called by jump instruction used as tail call
	bool MakesTailCall = false; // problem for Fast Returns, not for return address safety
	bool HasNoCallers = this->AllCallSources.empty();

	if (this->HasGoodRTLs()) {
		this->SetReturnAddressStatus(FUNC_SAFE);
		this->SetFuncSafe(true);
	}
	else { // don't process instructions if we have any bad RTLs
		this->SetReturnAddressStatus(FUNC_UNSAFE);
		this->SetUnsafeForFastReturns(true, UNSAFE_RETURN_ADDRESS);
		this->SetFuncSafe(false);
		this->SetSpecFuncSafe(false);
		this->SetNeedsFrame(true);
		this->SetSpecNeedsFrame(true);
		return;
	}

	if (!this->AllCallTargets.empty()) {
		HasCallTargets = true;
	}

#if SMP_USE_SWITCH_TABLE_INFO
	if (this->HasUnresolvedIndirectJumps()) {
#else
	if (this->IndirectJumps) {
#endif
#if SMP_DEBUG_FUNC
		SMP_msg("Function %s marked as unsafe due to indirect jumps\n", this->GetFuncName());
#endif
	}

	STARS_ea_t FirstAddr = this->GetFirstFuncAddr();

#if 0
	STARS_Function_t *FuncPtr = SMP_get_func(this->GetStartAddr());
	if (nullptr == FuncPtr) {
		SMP_msg("FATAL ERROR in SMPFunction::MarkFunctionSafe; no func ptr\n");
	}
	assert(nullptr != FuncPtr);


	bool TargetSuccess = FuncPtr->AnalyzeInstAsCallTarget(this, IsIndirectCallTarget, IsTailCallTarget);

	if (!IsIndirectCallTarget) {
		IsIndirectCallTarget = this->GetProg()->IsCodeXrefFromData(FirstAddr);
	}

	this->PossibleIndirectCallTarget |= IsIndirectCallTarget;
	this->PossibleTailCallTarget = IsTailCallTarget;


	this->DetectMultiEntryFunction();
#endif

	bool NonLoopingLeafFunc = (this->IsLeaf() && (0 == this->LoopCount) && this->HasGoodSSAForm());

	list<SMPInstr *>::iterator Instructions = Instrs.begin();
	SMPInstr *CurrInst;
#if SMP_USE_SSA_FNOP_MARKER
	++Instructions;  // skip marker instruction
#endif

	// While processing the stack pointer writes, the prologue code for
	//  saving the frame pointer register and allocating local variables needs to be
	//  handled.
	bool SaveEBP = false;
	bool XferESPtoEBP = false;
	for ( ; Instructions != Instrs.end(); ++Instructions) {
		CurrInst = (*Instructions);
		STARS_ea_t address = CurrInst->GetAddr();
		if (!CurrInst->IsAnalyzeable())
			continue;
#if SMP_VERBOSE_DEBUG_FUNC 
		SMP_msg(" Total number of defs for this instruction %d\n", CurrInst->NumDefs());
#endif
		if (!SaveEBP) { // still looking for "push ebp"
			if (CurrInst->MDIsPushInstr() && CurrInst->RegOpndMatches(0, MD_FRAME_POINTER_REG)) {
				SaveEBP = true;
				continue;
			}
		}
		else if (!XferESPtoEBP) { // found "push ebp", looking for "mov ebp,esp"
			if ((CurrInst->GetIDAOpcode() == STARS_NN_mov)
					&& (CurrInst->GetFirstDef()->GetOp()->MatchesReg(MD_FRAME_POINTER_REG))
					&& (CurrInst->GetFirstUse()->GetOp()->MatchesReg(MD_STACK_POINTER_REG))) {
				XferESPtoEBP = true;
				continue;
			}
		}
		if ((CurrInst->IsTailCall()) || (CurrInst->IsCondTailCall())) {
			// Moved up here because a tail call can be the point at which the stack
			//  frame returns to its prior state, making it the DeallocInstr.
			MakesTailCall = true;
		}
		if (address == this->LocalVarsAllocInstr ||	address == this->LocalVarsDeallocInstr) {
			continue;
		}

		if (CurrInst->MDIsStackPointerCopy(this->UseFP)) {
			// We have four cases for the register DEF that receives the stack pointer copy.
			// 1. The DEF is used (as a source reg or as an address reg) in an unsafe memory write.
			// 2. The DEF is live-in to an indirect call; who knows how it is used in the callee.
			// 3. The DEF is live-in to a direct call.
			// 4. None of the above; DEF does not reach any call or unsafe memory write.
			// Cases 1 and 2 are unsafe. We could be writing into the stack frame on top of a return
			//  address, unless we have more precise analysis of memory writes.
			// Case 3: DEF will either be used unsafely in the direct callee or not. We want to 
			//  leave our current status as FUNC_SAFE_IF_CALLEES_SAFE. Unsafe writes in the callee
			//  will be detected when the callee is analyzed.
			// Case 4 is safe.
			// As a result, we make our IsDefUsedInUnsafeMemWrite() search method return true for
			//  both case 1 and case 2 (conservatively assuming that indirect callees use the DEF
			//  in memory writes).
			// If the stack pointer is copied into a memory location instead of a register, it is unsafe.
			set<DefOrUse, LessDefUse>::iterator DefIter = CurrInst->GetFirstNonFlagsDef();
			assert(DefIter != CurrInst->GetLastDef());
			STARSOpndTypePtr CopyOp = DefIter->GetOp();
			int DefSSANum = DefIter->GetSSANum();
			if ((CopyOp->IsRegOp()) && this->HasGoodSSAForm() && (!HasStackPointerCopy) && (!(this->IsDefUsedInUnsafeMemWrite(CopyOp, DefSSANum, address)))) {
				SMP_msg("INFO: Safe stack pointer copy at %llx\n", (uint64_t) address);
			}
			else {
				HasStackPointerCopy = true;
				if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
					// If an lea instruction loads an address above
					//  the stack frame, we must assume that writes
					//  above the stack frame could occur.
					STARSOpndTypePtr TempOp = CurrInst->GetLeaMemUseOp();
					if (this->AccessAboveLocalFrame(TempOp, CurrInst->AreDefsNormalized(), address, false))
						WritesAboveLocalFrameIndirect = true;
				}
#if 1
				SMP_msg(" Function %s marked as unsafe due to stack pointer copy ", this->GetFuncName());
				SMP_msg("%s %llx \n", CurrInst->GetDisasm(), (uint64_t) CurrInst->GetAddr());
#endif
			}
		}
		if (CurrInst->MDIsPushInstr()) {
			// not exactly sure how to handle this instruction
			// for the moment if it's a push on a esp or usefp & ebp
			// mark as unsafe
			if (CurrInst->RegOpndMatches(0, MD_STACK_POINTER_REG) || 	 
					(this->UseFP && CurrInst->RegOpndMatches(0, MD_FRAME_POINTER_REG))) {
				HasStackPointerPush = true;
#if SMP_DEBUG_FUNC 
				SMP_msg(" Function %s marked as unsafe due to push on ebp or esp outside of function header \n", this->GetFuncName());	
				SMP_msg("%s %x\n", CurrInst->GetDisasm(), CurrInst->GetAddr());
#endif
			}
			continue;
		}
		if (CurrInst->MDIsPopInstr() || CurrInst->HasReturnOpcode()) {
			// ignore pops and returns for the moment
			 continue;
		}
		if (CurrInst->HasDestMemoryOperand()) {
			STARSOpndTypePtr Operand = CurrInst->GetMemDef();
			int BaseReg;
			int IndexReg;
			uint16_t ScaleFactor;
			STARS_ea_t offset;
			if (Operand->IsStaticMemOp()) {
				HasIndirectGlobalWrite = CurrInst->HasIndirectMemoryWrite();
			}
			else if (Operand->IsMemDisplacementOp()) {
				if (MDIsStackAccessOpnd(Operand, this->UsesFramePointer())) {
					if (!CurrInst->HasIndirectMemoryWrite()) {
						bool tempWritesAboveLocalFrame = this->WritesAboveLocalFrame(Operand, CurrInst->AreDefsNormalized(), address);
						WritesAboveLocalFrame |= tempWritesAboveLocalFrame;
					}
					else {
						bool tempWritesAboveLocalFrameIndirect = this->IndexedWritesAboveLocalFrame(Operand);

						/* separate indirect writes to this frame from indirect writes to another frame */
						if (tempWritesAboveLocalFrameIndirect) {
							WritesAboveLocalFrameIndirect = true;
						}
						else {
							HasIndexedStackWrite = true;
						}
					}
				}
#if 0
				else if (!this->HasUnsafeIndirectWrites && (!CurrInst->IsWriteSafetyAnalyzed())) { 
					// not yet unsafe due to indirect writes; analyze
					bool LoopingWrite = CurrInst->MDHasAnyRepeatPrefix();
					MDExtractAddressFields(Operand, BaseReg, IndexReg, ScaleFactor, offset);
					if (!LoopingWrite && (ScaleFactor == 0) && (!((BaseReg != STARS_x86_R_none) && (IndexReg != STARS_x86_R_none)))) {
						// Non-scaled, base or index (implied by MemDisplacementOp()) but not both.
						STARS_regnum_t AddrReg = (STARS_regnum_t) BaseReg;
						if (IndexReg != STARS_x86_R_none) {
							AddrReg = (STARS_regnum_t) IndexReg;
						}
						AddrReg = MDCanonicalizeSubReg(AddrReg);
						STARSOpndTypePtr AddrRegOp = CurrInst->MakeRegOpnd(AddrReg);
						STARSDefUseIter UseIter = CurrInst->FindUse(AddrRegOp);
						assert(UseIter != CurrInst->GetLastUse());
						bool SafeAddrReg = this->IsAddressRegSafe(AddrRegOp, address, UseIter->GetSSANum());
						if (SafeAddrReg) {
							SMP_msg("INFO: Found safe indirect memory write at %p\n", address);
						}
						else {
							this->HasUnsafeIndirectWrites = true;
						}
					}
					else {
						this->HasUnsafeIndirectWrites = true;
					}
				}
#else
				else if (!this->HasUnsafeIndirectWrites) {
					STARSDefUseIter DefIter = CurrInst->FindDef(Operand);
					assert(DefIter != CurrInst->GetLastDef());
					this->HasUnsafeIndirectWrites = (!DefIter->IsSafeMemWriteDef());
				}
#endif
			}
			else if (Operand->IsMemNoDisplacementOp()) {
				// so phrase is of the form [BASE_REG + IND ]
				// if the index register is missing just make sure that
				// the displacement is below stack frame top
				if (MDIsStackAccessOpnd(Operand, this->UsesFramePointer())) {
					if (!CurrInst->HasIndirectMemoryWrite()) {
						/* addressing mode is *esp or *ebp */
						continue;
					}
					else {
						HasIndexedStackWrite = true;
					}
				}
				else if (!this->HasUnsafeIndirectWrites) { // not yet unsafe due to indirect writes; analyze
#if 0
					if (!CurrInst->IsWriteSafetyAnalyzed()) {
						bool LoopingWrite = CurrInst->MDHasAnyRepeatPrefix();
						MDExtractAddressFields(Operand, BaseReg, IndexReg, ScaleFactor, offset);
						if (!LoopingWrite && (ScaleFactor == 0) && (!((BaseReg != STARS_x86_R_none) && (IndexReg != STARS_x86_R_none)))) {
							// Non-scaled, base or index (implied by MemNoDisplacementOp()) but not both.
							STARS_regnum_t AddrReg = (STARS_regnum_t)BaseReg;
							if (IndexReg != STARS_x86_R_none) {
								AddrReg = (STARS_regnum_t)IndexReg;
							}
							AddrReg = MDCanonicalizeSubReg(AddrReg);
							STARSOpndTypePtr AddrRegOp = CurrInst->MakeRegOpnd(AddrReg);
							STARSDefUseIter UseIter = CurrInst->FindUse(AddrRegOp);
							assert(UseIter != CurrInst->GetLastUse());
							bool SafeAddrReg = this->IsAddressRegSafe(AddrRegOp, address, UseIter->GetSSANum());
							if (SafeAddrReg) {
								SMP_msg("INFO: Found safe indirect memory write at %p\n", address);
							}
							else {
								this->HasUnsafeIndirectWrites = true;
							}
						}
						else {
							this->HasUnsafeIndirectWrites = true;
						}
					}
#else
					STARSDefUseIter DefIter = CurrInst->FindDef(Operand);
					assert(DefIter != CurrInst->GetLastDef());
					this->HasUnsafeIndirectWrites = (!DefIter->IsSafeMemWriteDef());
#endif
				}
			}
			else {
				SMP_msg("FATAL ERROR: Unknown memory operand type in MarkFunctionSafe at %llx ", (uint64_t) address);
				this->DumpFuncNameAndAddr();
				this->Dump(false);
				this->DumpDotCFG();
				assert(false);
			}
		} // end if HasMemDestOperand() in current instruction
		// Find instructions that take the address of a callee-saved reg, return address, or stack-inarg.
		if (CurrInst->MDIsLoadEffectiveAddressInstr()) {
			STARSOpndTypePtr LeaMemOp = CurrInst->GetLeaMemUseOp();
			if (MDIsDirectStackAccessOpnd(LeaMemOp, UseFP)) {
				// Exclude the case in which the stack pointer register is being adjusted with
				//  an lea opcode, e.g. lea esp,[ebp-k] deallocates the stack frame to prepare
				//  for popping the saved regs.
				STARSDefUseIter LeaDefIter = CurrInst->GetFirstNonFlagsDef();
				STARSOpndTypePtr LeaDefOp = LeaDefIter->GetOp();
				if (!LeaDefOp->MatchesReg(MD_STACK_POINTER_REG)) {
					StackAccessType AccessVal = this->GetStackAccessType(LeaMemOp, CurrInst->AreDefsNormalized(), address, false);
					FILE *AlarmFile = global_STARS_program->GetAlarmFile();
					if (STARS_STACK_CALLEE_SAVED_REG == AccessVal) {
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR CALLEESAVEDREGADDRESSTAKEN \n", (uint64_t)address, CurrInst->GetSize());
					}
					else if (STARS_STACK_RETURN_ADDRESS == AccessVal) {
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR RETURNADDRADDRESSTAKEN \n", (uint64_t)address, CurrInst->GetSize());
					}
					else if (STARS_STACK_INARG == AccessVal) {
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR INARGADDRESSTAKEN \n", (uint64_t)address, CurrInst->GetSize());
					}
					else if (STARS_STACK_OUTARG == AccessVal) {
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR OUTARGADDRESSTAKEN \n", (uint64_t) address, CurrInst->GetSize());
					}
					else if (STARS_STACK_LOCAL_FRAME == AccessVal) {
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR LOCALVARADDRESSTAKEN \n", (uint64_t) address, CurrInst->GetSize());
					}
					else {
						assert(STARS_STACK_UNKNOWN == AccessVal);
						SMP_fprintf(AlarmFile, "%18llx %6d INSTR UNKNOWNSTACKADDRESSTAKEN \n", (uint64_t) address, CurrInst->GetSize());
					}
				}
			}
		}
	} // end for all instructions

	// We can reset this->HasUnsafeIndirectWrites for non-looping leaf functions.
	//  The only way to overwrite the return address in such a function is to directly
	//  write to it, which is detected in the AccessesReturnAddress boolean flag below.
	if (NonLoopingLeafFunc && this->HasUnsafeIndirectWrites) {
		this->HasUnsafeIndirectWrites = false;
		SMP_msg("INFO: Resetting HasUnsafeIndirectWrites for non-looping leaf func at %llx\n", (uint64_t) this->GetFirstFuncAddr());
	}

	// For mmStrata bounds checking of the stack frame, we don't care
	//  about indirect writes unless they are to the stack.
	bool SpecUnsafe = (HasStackPointerCopy || HasStackPointerPush || HasIndexedStackWrite || this->HasSharedChunks()
		|| this->HasUnresolvedIndirectJumps());
	bool Unsafe = SpecUnsafe || this->HasUnresolvedIndirectCalls();

	this->SetFuncSafe(!Unsafe);
	this->SetSpecFuncSafe(!SpecUnsafe);

	unsigned short FastReturnStatus = this->GetFastReturnStatus();
#if 0
	bool AccessesReturnAddress = (0 != (FastReturnStatus & (RETURN_ADDRESS_WRITE | RETURN_ADDRESS_READ)));
#else  // reading the return address is not unsafe in itself
	bool AccessesReturnAddress = (0 != (FastReturnStatus & RETURN_ADDRESS_WRITE));
#endif
	if (WritesAboveLocalFrame) {
		// Detect safe and dangerous cases.
		if (!AccessesReturnAddress) {
			// No direct reads or writes of the return address.
			// See if we can read or write the return address of any caller.
			WritesAboveLocalFrame = this->IsCallerReturnAddressReadOrWritten(); // reset to false if safe
		}
	}

	this->WritesAboveRA = WritesAboveLocalFrameIndirect;  // currently unused
	this->SafeCallee = (!Unsafe) && (!WritesAboveLocalFrameIndirect) && this->AnalyzedSP && (!WritesAboveLocalFrame);
	this->SpecSafeCallee = (!SpecUnsafe) && (!WritesAboveLocalFrameIndirect) && this->AnalyzedSP && (!WritesAboveLocalFrame);
	this->SetNeedsFrame(Unsafe);
	this->SetSpecNeedsFrame(SpecUnsafe);

	this->HasIndirectWrites = this->HasIndirectWrites || this->HasUnsafeIndirectWrites || (HasIndexedStackWrite || WritesAboveLocalFrameIndirect || HasIndirectGlobalWrite);

#if 1
	bool UnsafeReturnAddr = (Unsafe || AccessesReturnAddress || this->HasUnsafeIndirectWrites || (!this->AnalyzedSP) || this->MultipleEntryPoints);
#else
	bool UnsafeReturnAddr = (Unsafe || AccessesReturnAddress || WritesAboveLocalFrameIndirect || HasIndirectGlobalWrite
		|| HasIndirectWrite || (!this->AnalyzedSP) || this->MultipleEntryPoints);
#endif
	// We have conditions that cause fast returns to be unsafe even though the return address is safe.
	if (this->PossibleIndirectCallTarget) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it is indirect call target.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, INDIRECTLY_CALLED);
	}
	if (this->PossibleTailCallTarget) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it is tail call target.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, TAIL_CALL_TARGET);
	}
	if (HasNoCallers) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it has no callers.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, NO_CALLERS);
	}
	if (MakesTailCall) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it makes a tail call.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, MAKES_TAIL_CALL);
	}
	if (this->MultipleEntryPoints) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it has multiple entry points.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, MULTIPLE_ENTRY_POINTS);
	}
	if (this->HasUnresolvedIndirectJumps()) {
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because it has unresolved indirect jumps.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, UNRESOLVED_INDIR_JUMP);
	}

	if (UnsafeReturnAddr) {
#if STARS_CONSERVATIVE_FAST_RETURNS
		SMP_msg("INFO: Function at %llx becoming unsafe for fast returns because its return addr is UNSAFE.\n", (uint64_t) this->GetFirstFuncAddr());
		this->SetUnsafeForFastReturns(true, UNSAFE_RETURN_ADDRESS);
#endif

		this->SetReturnAddressStatus(FUNC_UNSAFE);
#if SMP_DEBUG_FUNC_SAFETY
		SMP_msg("UNSAFE function %s ", this->GetFuncName());
		SMP_msg("StackPtrCopy: %d StackPtrPush: %d IndirectGlobal: %d ",
			HasStackPointerCopy, HasStackPointerPush, HasIndirectGlobalWrite);
		SMP_msg("WritesAboveFrame: %d AccessesRA: %d IndirectStack: %d IndirectWrite: %d ",
			WritesAboveLocalFrame, AccessesReturnAddress, HasIndexedStackWrite, this->HasUnsafeIndirectWrites);
		SMP_msg("AnalyzedSP: %d UnresolvedCalls: %d UnresolvedJumps: %d SharedChunks: %d IsLeaf: %d ",
			this->AnalyzedSP, this->HasUnresolvedIndirectCalls(), this->HasUnresolvedIndirectJumps(),
			this->HasSharedChunks(), this->IsLeaf());
		SMP_msg("IndirCallTarget: %d TailCallTarget: %d HasNoCallers: %d MultiEntry: %d\n", 
			this->PossibleIndirectCallTarget, this->PossibleTailCallTarget, HasNoCallers, this->MultipleEntryPoints);
#endif
	}
	else if (HasCallTargets) {
		this->SetReturnAddressStatus(FUNC_SAFE_IF_CALLEES_ARE_SAFE);
	}

#if SMP_DEBUG_FUNC
	if (this->GetReturnAddressStatus() == FUNC_SAFE)
		SMP_msg("Function %s is SAFE\n", GetFuncName());
	else if (this->GetReturnAddressStatus() == FUNC_UNSAFE)
		SMP_msg("Function %s is UNSAFE\n", GetFuncName());
	else if (this->GetReturnAddressStatus() == FUNC_SAFE_IF_CALLEES_ARE_SAFE)
		SMP_msg("Function %s is SAFE_IF_CALLEES_ARE_SAFE\n", GetFuncName());

	if (!Unsafe) 
		SMP_msg("Function %s is mmSAFE\n", GetFuncName());
	else 
		SMP_msg("Function %s is mmUNSAFE\n", GetFuncName());

	if (!SpecUnsafe) 
		SMP_msg("Function %s is Speculatively mmSAFE\n", GetFuncName());
	else 
		SMP_msg("Function %s is Speculatively mmUNSAFE\n", GetFuncName());

#endif
	return;
} // end of SMPFunction::MarkFunctionSafe()

// Return TCFG block # of first block in downward path that is not a jump-only block.
int SMPFunction::FindEndOfJumpOnlyBlockChain(const int StartingTCFGBlockNum) const {
	STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum((size_t)StartingTCFGBlockNum);
	if (!CurrTCFGBlock->IsOnlyInternalDirectJump()) {
		return StartingTCFGBlockNum;
	}
	else {
		STARSTransCFGBlock *SuccTCFGBlock = *(CurrTCFGBlock->GetFirstConstSucc());
		int SuccTCFGBlockNum = SuccTCFGBlock->GetNumber();
		if (SuccTCFGBlockNum > StartingTCFGBlockNum) { // not a back edge
			return this->FindEndOfJumpOnlyBlockChain(SuccTCFGBlockNum);
		}
		else {
			return StartingTCFGBlockNum;
		}
	}
} // end of SMPFunction::FindEndOfJumpOnlyBlockChain()

// Perform look-ahead steps needed before translation to SPARK Ada.
void SMPFunction::PreProcessForSPARKAdaTranslation(void) {
	bool HasLoops = (this->LoopCount > 0);
	bool UseFP = this->UsesFramePointer();

	for (vector<SMPBasicBlock *>::iterator BlockIter = this->RPOBlocks.begin(); BlockIter != this->RPOBlocks.end(); ++BlockIter) {
		SMPBasicBlock *CurrBlock = (*BlockIter);
		if (CurrBlock->HasCallInstruction()) {
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				if (CurrInst->GetDataFlowType() == CALL) {
					string CalledFuncName = CurrInst->GetTrimmedCalledFunctionName();
					if (IsLibFuncName(CalledFuncName)) {
						pair<set<string>::iterator, bool> InsertResult = this->Program->InsertLibraryFuncName(CalledFuncName);
						// !!!!****!!!! Need to find setuid() arg source (e.g. RDI comes from [RBP-xx]) and save it for Ada pragma assert
						//  and pragma Loop_Invariants.
					}
				}
			}
		}
		size_t BlockNum = (size_t) CurrBlock->GetNumber();
		vector<SMPInstr *>::const_reverse_iterator LastInstIter = CurrBlock->GetRevInstCBegin();
		SMPInstr *LastInst = (*LastInstIter);
		if (LastInst->IsBranchToOtherFunc()) {
			SMP_msg("ERROR: SPARK: Bad CFG (branch outside function at %llx in block %zu) ", (uint64_t)LastInst->GetAddr(), BlockNum);
			this->DumpFuncNameAndAddr();
			this->SetHasUnstructuredCFG(); // branch outside func
		}

		if (HasLoops && CurrBlock->HasMemoryWrite() && this->IsBlockInAnyLoop((int) BlockNum)) {
			for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
				SMPInstr *CurrInst = (*InstIter);
				STARS_ea_t InstAddr = CurrInst->GetAddr();
				STARSOpndTypePtr MemDefOp = CurrInst->GetMemDef();
				STARS_ea_t offset;
				size_t DataSize;
				bool UsedFramePointer, IndexedAccess, SignedMove, UnsignedMove;
				if ((nullptr != MemDefOp) && this->MDGetStackOffsetAndSize(CurrInst, MemDefOp, this->MinStackAccessOffset, offset, DataSize, UsedFramePointer,
					IndexedAccess, SignedMove, UnsignedMove)) {
					// NOTE: Deal with indexed accesses later.
					int SignedOffset = (int) offset;
					assert(0 <= SignedOffset);  // before adding min stack delta, which is negative
					SignedOffset += this->MinStackAccessOffset;

					for (size_t j = 0; j < DataSize; ++j) { // offset has zero-based index into negative offset vectors
						for (size_t LoopIndex = 0; LoopIndex < (size_t) this->LoopCount; ++LoopIndex) {
							if (this->FuncLoopsByBlock[BlockNum].GetBit(LoopIndex)) {
								bool OutOfRange = (SignedOffset < 0) ? ((offset + j) >= this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits())
									: ((SignedOffset + j) >= this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits());
								if (OutOfRange) {
									if (IndexedAccess) {
										// Index can bring offset back into range. Ignore.
										SMP_msg("WARNING: SPARK: Imprecise range on indexed access at %llx ", (uint64_t) InstAddr);
										this->DumpFuncNameAndAddr();
										continue;
									}
									else {
										SMP_msg("ERROR: SPARK: Imprecise range on non-indexed access at %llx ", (uint64_t)InstAddr);
										this->DumpFuncNameAndAddr();
										continue;
									}
								}
								if (SignedOffset < 0) {
									this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].SetBit(offset + j);
								}
								else {
									this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].SetBit(SignedOffset + j);
								}
							}
						} // end for all loops
					} // end for all bytes in DataSize of current stack write
				}
			} // end for all instructions in current block
		}
	} // end for all blocks

	return;
} // end of SMPFunction::PreProcessForSPARKAdaTranslation()

// update {Nega, Posi}tiveStackBytesWrittenByLoop[LoopNum]
void SMPFunction::UpdateStackBytesWrittenByLoop(STARS_sval_t FinalStackPtrOffset, size_t MemWidth, size_t LoopNum) {
	assert(LoopNum < this->GetNumLoops());
	STARS_sval_t SignedOffset = FinalStackPtrOffset - this->MinStackAccessOffset;
	size_t ArrayIndex;
	if (FinalStackPtrOffset >= 0) {
		ArrayIndex = (size_t) FinalStackPtrOffset;
		assert(ArrayIndex < this->PositiveOffsetStackBytesWrittenByLoop[LoopNum].GetNumBits());
	}
	else {
		ArrayIndex = (size_t) SignedOffset;
		assert(ArrayIndex < this->NegativeOffsetStackBytesWrittenByLoop[LoopNum].GetNumBits());
	}

	for (size_t j = 0; j < MemWidth; ++j) { // offset has zero-based index into negative or positive offset vector
		if (FinalStackPtrOffset < 0)
			this->NegativeOffsetStackBytesWrittenByLoop[LoopNum].SetBit(ArrayIndex + j);
		else
			this->PositiveOffsetStackBytesWrittenByLoop[LoopNum].SetBit(ArrayIndex + j);
	} // end for all bytes in MemWidth of current stack write
	return;
} // end of SMPFunction::UpdateStackBytesWrittenByLoop()

#if 0
// specify mem ranges that loop changes; return true if anything is printed
bool SMPFunction::EmitSPARKLoopMemRangePostCondition(FILE *HeaderFile, FILE *BodyFile, STARS_ea_t LoopAddr, bool PostPrintStarted) {
	SMPBasicBlock *LoopHeaderBlock = this->GetBlockFromInstAddr(LoopAddr);
	int HeaderBlockNum = LoopHeaderBlock->GetNumber();
	int LoopNum = this->GetLoopNumFromHeaderBlockNum(HeaderBlockNum);
	assert(0 <= LoopNum);
	size_t LoopIndex = (size_t) LoopNum;
	size_t LoopNumPlusOne = LoopIndex + 1;
	STARS_sval_t IncomingStackDelta = (*LoopHeaderBlock->GetFirstInst())->GetStackPtrOffset();
	assert(0 >= IncomingStackDelta);
	IncomingStackDelta = (0 - IncomingStackDelta); // convert to positive offset to correct callee mem exprs
	bool StackWritten = (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet() || this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet());
	bool IndirectStackWrites = this->LoopMemExprsExpandToStackOffsets[LoopIndex];
	bool CalleeWrites = this->LoopHasCalleeMemWrites[LoopIndex];
	bool HasArgs = this->DoesLoopHaveArgs(LoopNum);
	StackWritten = (StackWritten || IndirectStackWrites);
	bool LoopInvariantWrites = (0 < this->GetNumInArgsUsedInMemWrites(LoopNumPlusOne));

	// Emit memory writing range post-conditions and nearly identical loop invariants.
	bool NeedsOutput = (StackWritten || HasArgs || CalleeWrites || LoopInvariantWrites);
	if (NeedsOutput) {
		if (!PostPrintStarted) {
			SMP_fprintf(HeaderFile, "\tPost => ");
		}
		if (StackWritten) {
			this->EmitSPARKStackMemRangePostCondition(HeaderFile, LoopIndex, LoopAddr, PostPrintStarted);
		}
		if (HasArgs) {
			if (!StackWritten) {
				if (PostPrintStarted) {
					SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
				}
				else { // nothing written yet since "Post => "
					SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
				}
			}
			else {
				SMP_fprintf(HeaderFile, "\n\t\tand not X86.InMemoryRange(i, ");
			}
			this->LoopMemWriteLowerBoundsExprs[LoopIndex]->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
			SMP_fprintf(HeaderFile, ", ");
			this->LoopMemWriteUpperBoundsExprs[LoopIndex]->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
			SMP_fprintf(HeaderFile, ") ");
		}
		if (CalleeWrites) {
			size_t OutputCount;
			if (!(StackWritten || HasArgs)) {
				OutputCount = 0;
			}
			else {
				OutputCount = 1;
			}
			for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
				STARSExpression *CurrExpr = (*(this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].first));
				// The StackPtrCase below is now handled via the {Posi,Nega}tiveStackOffsetWritesByLoop[] vectors
				if (!CurrExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
					size_t ByteWidth = this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].second;
					if (OutputCount == 0) {
						if (PostPrintStarted) {
							SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
						else { // nothing written since "Post => "
							SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
					}
					else {
						SMP_fprintf(HeaderFile, "\n\t\tand not X86.InRange64(i, ");
					}
					// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
					//  this loop. We need to make two adjustments when printing these exprs:
					//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
					//   2. Offset the stack address with the negated incoming stack offset.
					bool StackPtrCase = CurrExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
					STARS_ea_t PreLoopDefAddr = CurrExpr->GetLeftPreLoopDefAddr();
					bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
					if (!StackPtrCase && (!PseudoAddr)) {
						// We don't want to trace callee mem exprs all the way back to this function's InArgs.
						//  We only want to trace back to the value that comes into the loop. The loop boundary
						//  crossing is recorded in the expression.
						STARSExpression *LoopBoundaryExpr = CurrExpr->Clone();
						LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
						LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
						LoopBoundaryExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs);
					}
					else {
						CurrExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs);
						if ((0 != IncomingStackDelta) && StackPtrCase) {
							// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
							SMP_fprintf(HeaderFile, " + 16#%x#", IncomingStackDelta);
						}
					}
					SMP_fprintf(HeaderFile, ", %u) ", ByteWidth);
					++OutputCount;
				}
			} // end for all elements of LoopMemAddrExprWidthsFromCallees[LoopIndex]
			SMP_fprintf(HeaderFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
		} // end if CalleeWrites
		if (LoopInvariantWrites) {
			size_t OutputCount;
			if (!(StackWritten || HasArgs)) {
				OutputCount = 0;
			}
			else {
				OutputCount = 1;
			}
			for (size_t i = 0; i < this->GetNumInArgsUsedInMemWrites(LoopNumPlusOne); ++i) {
				STARSExprSetIter CurrExprIter = this->GetInArgExprUsedInMemWrite(LoopNumPlusOne, i);
				STARSExpression *CurrExpr = (*CurrExprIter);
				// The StackPtrCase below is now handled via the {Posi,Nega}tiveStackOffsetWritesByLoop[] vectors
				if (!CurrExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
					size_t ByteWidth = this->GetInArgMemWriteWidth(LoopNumPlusOne, i);
					if (OutputCount == 0) {
						if (PostPrintStarted) {
							SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
						else { // nothing written since "Post => "
							SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
					}
					else {
						SMP_fprintf(HeaderFile, "\n\t\tand not X86.InRange64(i, ");
					}
					// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
					//  this loop. We need to make two adjustments when printing these exprs:
					//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
					//   2. Offset the stack address with the negated incoming stack offset.
					bool StackPtrCase = CurrExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
					STARS_ea_t PreLoopDefAddr = CurrExpr->GetLeftPreLoopDefAddr();
					bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
					if (!StackPtrCase && (!PseudoAddr)) {
						// We don't want to trace callee mem exprs all the way back to this function's InArgs.
						//  We only want to trace back to the value that comes into the loop. The loop boundary
						//  crossing is recorded in the expression.
						STARSExpression *LoopBoundaryExpr = CurrExpr->Clone();
						LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
						LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
						LoopBoundaryExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs);
					}
					else {
						CurrExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs);
						if ((0 != IncomingStackDelta) && StackPtrCase) {
							// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
							SMP_fprintf(HeaderFile, " + 16#%x#", IncomingStackDelta);
						}
					}
					SMP_fprintf(HeaderFile, ", %u) ", ByteWidth);
					++OutputCount;
				}
			} // end for all elements of GetNumInArgsUsedInMemWrites(LoopNumPlusOne)
			SMP_fprintf(HeaderFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
		} // end if LoopInvariantWrites

		SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i)))");

		PrintSPARKIndentTabs(BodyFile);
		if (StackWritten) {
			SMP_fprintf(BodyFile, "pragma Loop_Invariant(for all i in Unsigned64 => (if ");
			this->EmitSPARKStackMemRangePostCondition(BodyFile, LoopIndex, LoopAddr, false);
		}
		if (HasArgs) {
			if (!StackWritten) {
				SMP_fprintf(BodyFile, "pragma Loop_Invariant(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
			}
			else {
				SMP_fprintf(BodyFile, "\n\t\tand not X86.InMemoryRange(i, ");
			}
			this->LoopMemWriteLowerBoundsExprs[LoopIndex]->EmitSPARKAda(BodyFile, true, false, false, HasArgs);
			SMP_fprintf(BodyFile, ", ");
			this->LoopMemWriteUpperBoundsExprs[LoopIndex]->EmitSPARKAda(BodyFile, true, false, false, HasArgs);
			SMP_fprintf(BodyFile, ") ");
		}

		if (CalleeWrites) {
			size_t OutputCount;
			if (!(StackWritten || HasArgs)) {
				OutputCount = 0;
			}
			else {
				OutputCount = 1;
			}
			for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
				size_t ByteWidth = this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].second;
				STARSExpression *CurrExpr = (*(this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].first));
				if (!CurrExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
					if (OutputCount == 0) {
						SMP_fprintf(BodyFile, "pragma Loop_Invariant(for all i in Unsigned64 => (if not X86.InRange64(i, ");
					}
					else {
						SMP_fprintf(BodyFile, "\n\t\tand not X86.InRange64(i, ");
					}
					// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
					//  this loop. We need to make two adjustments when printing these exprs:
					//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
					//   2. Offset the stack address with the negated incoming stack offset.
					bool StackPtrCase = CurrExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
					STARS_ea_t PreLoopDefAddr = CurrExpr->GetLeftPreLoopDefAddr();
					bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
					if (!StackPtrCase && (!PseudoAddr)) {
						// We don't want to trace callee mem exprs all the way back to this function's InArgs.
						//  We only want to trace back to the value that comes into the loop. The loop boundary
						//  crossing is recorded in the expression.
						STARSExpression *LoopBoundaryExpr = CurrExpr->Clone();
						LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
						LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
						LoopBoundaryExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs);
					}
					else {
						CurrExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs);
						if ((0 != IncomingStackDelta) && StackPtrCase) {
							// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
							SMP_fprintf(BodyFile, " + 16#%x#", IncomingStackDelta);
						}
					}
					SMP_fprintf(BodyFile, ", %u) ", ByteWidth);
					++OutputCount;
				}
			} // end for all elements of MemAddrExprWidthsFromCallees[LoopIndex]
			SMP_fprintf(BodyFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
		} // end if CalleeWrites

		if (LoopInvariantWrites) {
			size_t OutputCount;
			if (!(StackWritten || HasArgs)) {
				OutputCount = 0;
			}
			else {
				OutputCount = 1;
			}
			for (size_t i = 0; i < this->GetNumInArgsUsedInMemWrites(LoopNumPlusOne); ++i) {
				STARSExprSetIter CurrExprIter = this->GetInArgExprUsedInMemWrite(LoopNumPlusOne, i);
				STARSExpression *CurrExpr = (*CurrExprIter);
				if (!CurrExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
					size_t ByteWidth = this->GetInArgMemWriteWidth(LoopNumPlusOne, i);
					if (OutputCount == 0) {
						SMP_fprintf(BodyFile, "pragma Loop_Invariant(for all i in Unsigned64 => (if not X86.InRange64(i, ");
					}
					else {
						SMP_fprintf(BodyFile, "\n\t\tand not X86.InRange64(i, ");
					}
					// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
					//  this loop. We need to make two adjustments when printing these exprs:
					//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
					//   2. Offset the stack address with the negated incoming stack offset.
					bool StackPtrCase = CurrExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
					STARS_ea_t PreLoopDefAddr = CurrExpr->GetLeftPreLoopDefAddr();
					bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
					if (!StackPtrCase && (!PseudoAddr)) {
						// We don't want to trace callee mem exprs all the way back to this function's InArgs.
						//  We only want to trace back to the value that comes into the loop. The loop boundary
						//  crossing is recorded in the expression.
						STARSExpression *LoopBoundaryExpr = CurrExpr->Clone();
						LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
						LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
						LoopBoundaryExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs);
					}
					else {
						CurrExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs);
						if ((0 != IncomingStackDelta) && StackPtrCase) {
							// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
							SMP_fprintf(BodyFile, " + 16#%x#", IncomingStackDelta);
						}
					}
					SMP_fprintf(BodyFile, ", %u) ", ByteWidth);
					++OutputCount;
				}
			} // end for all elements of GetNumInArgsUsedInMemWrites(LoopNumPlusOne)
			SMP_fprintf(BodyFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
		} // end if LoopInvariantWrites

		SMP_fprintf(BodyFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Loop_Entry(i))));\n");
	}

	if (NeedsOutput) {
		SMP_fprintf(HeaderFile, ");\n\n"); // terminate the postcondition section
	}
	else { // nothing printed
		SMP_fprintf(HeaderFile, ";\n\n"); // terminate the specification
	}

	return NeedsOutput;
} // end of SMPFunction::EmitSPARKLoopMemRangePostCondition()
#else
// specify mem ranges that loop changes; return true if anything is printed
bool SMPFunction::EmitSPARKLoopMemRangePostCondition(FILE *HeaderFile, FILE *BodyFile, STARS_ea_t LoopAddr, bool PostPrintStarted) {
	bool UseFP = this->UsesFramePointer();
	SMPBasicBlock *LoopHeaderBlock = this->GetBlockFromInstAddr(LoopAddr);
	int HeaderBlockNum = LoopHeaderBlock->GetNumber();
	int LoopNum = this->GetLoopNumFromHeaderBlockNum(HeaderBlockNum);
	assert(0 <= LoopNum);
	size_t LoopIndex = (size_t) LoopNum;
	size_t LoopNumPlusOne = LoopIndex + 1;
	STARS_sval_t IncomingStackDelta = (*LoopHeaderBlock->GetFirstInst())->GetStackPtrOffset();
	assert(0 >= IncomingStackDelta);
	IncomingStackDelta = (0 - IncomingStackDelta); // convert to positive offset to correct callee mem exprs
	SMPInstr *FirstInst = this->GetInstFromAddr(LoopAddr);
	bool StackWritten = (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet() || this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet());
	bool IndirectStackWrites = this->LoopMemExprsExpandToStackOffsets[LoopIndex];
	bool CalleeWrites = this->LoopHasCalleeMemWrites[LoopIndex];
	bool HasArgs = this->DoesLoopHaveArgs(LoopNum);
	StackWritten = (StackWritten || IndirectStackWrites);
	bool LoopInvariantWrites = (0 < this->GetNumInArgsUsedInMemWrites(LoopNumPlusOne));
	bool HasBIVInitialValue = (this->LoopAnalyzedBIVIters[LoopNum] != this->LoopInductionVars[LoopNum].end()); // successful loop analysis

	// Use sets of exprs to avoid duplication and streamline the output section.
	bool HasRangeExprs = (!this->TempRangeExprWidthIters.empty());
	bool HasNonRangeExprs = (!this->TempNonRangeExprWidthIters.empty());
	bool HasRelationalExprs = (!this->RelationalMemWriteWidths[LoopNumPlusOne].empty());

	// Besides memory writing contracts, another pragma Loop_Invariant that has a corresponding post-condition is to assert
	//  that a register is loop-invariant if it is NOT in the OutputRegs bitset for this loop,
	//  and it is in the CalleePreservedRegsByLoop bitset for this loop. Find all such registers.
	bool InvariantPreservedRegs = false;
	std::bitset<1 + MD_LAST_REG_NO> LoopPreservedRegs;
	for (size_t RegNo = 0; RegNo < MD_LAST_REG_NO; ++RegNo) {
		if (!this->OutputRegsByLoop[LoopIndex][RegNo] && this->CalleePreservedRegsByLoop[LoopIndex][RegNo]) {
			// The fact that the register is preserved by callees will force it to be In_Out. However, the
			//  absence of an OutputRegs bit means that the most recent thing to happen to it in the call chain
			//  was its restoration by a function in the callee chain. It is, therefore, in the confusing
			//  category of "loop-invariant but declared Global In_Out." We need to tell the prover that it is
			//  loop-invariant, and also emit a post-condition that helps propagate this info to callers.
			InvariantPreservedRegs = true;
			LoopPreservedRegs.set(RegNo);
		}
	}

	// Emit memory writing range post-conditions and nearly identical loop invariants.
	bool NeedsMemoryOutput = (StackWritten || HasArgs || HasRangeExprs || HasNonRangeExprs || HasRelationalExprs);
	bool NeedsOutput = (NeedsMemoryOutput || InvariantPreservedRegs);
	bool LoopRangeWritten = false;
	bool LocalPostPrintStarted = PostPrintStarted;
	if (NeedsOutput) {
		size_t OutputCount = 0;
		if (!LocalPostPrintStarted) {
			if (this->LoopHasPreconditions[LoopIndex]) {
				SMP_fprintf(HeaderFile, "\tPost => ");
			}
			else {
				// No precondition output; need to put comma and new line after Global output section.
				SMP_fprintf(HeaderFile, ",\n\tPost => ");
			}
		}
		if (InvariantPreservedRegs) {
			for (size_t RegNo = 0; RegNo < LoopPreservedRegs.size(); ++RegNo) {
				if (LoopPreservedRegs[RegNo]) {
					if (!LocalPostPrintStarted) {
						LocalPostPrintStarted = true;
					}
					else {
						SMP_fprintf(HeaderFile, "\n\tand (");
					}
					STARSOpndTypePtr RegOp = FirstInst->MakeRegOpnd((STARS_regnum_t) RegNo);
					FirstInst->PrintSPARKAdaOperand(RegOp, HeaderFile, false, UseFP, true, false, false);
					SMP_fprintf(HeaderFile, "= ");
					FirstInst->PrintSPARKAdaOperand(RegOp, HeaderFile, false, UseFP, true, true, false);
					SMP_fprintf(HeaderFile, "'Old) ");
				}
			}
		}

		if (StackWritten) {
			this->EmitSPARKStackMemRangePostCondition(HeaderFile, LoopIndex, LoopAddr, PostPrintStarted);
		}
		if (HasRangeExprs) {
			if (!StackWritten) {
				if (PostPrintStarted) {
					SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
				}
				else { // nothing written yet since "Post => " except maybe preserved regs
					SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
				}
			}
			else {
				SMP_fprintf(HeaderFile, "\n\t\tand not X86.InMemoryRange(i, ");
			}
			size_t VectorLimit = this->TempRangeExprWidthIters.size();
			for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
				STARSExpression *LowerExpr = (*this->TempRangeExprWidthIters[VectorIndex].first);
				STARSExpression *UpperExpr = (*this->TempRangeExprWidthIters[VectorIndex].second);
				this->EmitSPARKMemRange(HeaderFile, false, true, HasArgs, LowerExpr, UpperExpr, OutputCount, 0);
			}
		}

		if (HasRelationalExprs) {
			if (0 == OutputCount) {
				if (!StackWritten) {
					if (PostPrintStarted) {
						SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
					}
					else { // nothing written yet since "Post => "
						SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ");
					}
				}
				else {
					SMP_fprintf(HeaderFile, "\n\t\tand not X86.InMemoryRange(i, ");
				}
			}
			std::list<std::pair<std::size_t, std::pair<STARSExprSetIter, STARSExprSetIter> > >::const_iterator RelationalIter;
			for (RelationalIter = this->RelationalMemWriteWidths[LoopNumPlusOne].cbegin(); RelationalIter != this->RelationalMemWriteWidths[LoopNumPlusOne].cend(); ++RelationalIter) {
				size_t MemWidth = (*RelationalIter).first;
				STARSExprSetIter LowerExprIter = (*RelationalIter).second.first;
				STARSExprSetIter UpperExprIter = (*RelationalIter).second.second;
				STARSExpression *LowerExpr = (*LowerExprIter);
				STARSExpression *UpperExpr = (*UpperExprIter);
				this->EmitSPARKMemRange(HeaderFile, false, true, HasArgs, LowerExpr, UpperExpr, OutputCount, MemWidth);
			}
		} // end if (HasRelationalExprs)

		if (HasNonRangeExprs) {
			size_t VectorLimit = this->TempNonRangeExprWidthIters.size();
			for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
				// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
				//  which would require printing the same operand twice.
				size_t ByteWidth = this->TempNonRangeExprWidthIters[VectorIndex].first;
				STARSExpression *NonRangeExpr = (*this->TempNonRangeExprWidthIters[VectorIndex].second);

				// The StackPtrCase below is now handled via the {Posi,Nega}tiveStackOffsetWritesByLoop[] vectors
				if (!NonRangeExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
					if (OutputCount == 0) {
						if (PostPrintStarted) {
							SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
						else { // nothing written since "Post => "
							SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if not X86.InRange64(i, ");
						}
					}
					else {
						SMP_fprintf(HeaderFile, "\n\t\tand not X86.InRange64(i, ");
					}
					// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
					//  this loop. We need to make two adjustments when printing these exprs:
					//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
					//   2. Offset the stack address with the negated incoming stack offset.
					bool StackPtrCase = (!NonRangeExpr->HasLeftSubTree()) && NonRangeExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
					STARS_ea_t PreLoopDefAddr = NonRangeExpr->GetLeftPreLoopDefAddr();
					bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
					if (!StackPtrCase && (!PseudoAddr)) {
						// We don't want to trace callee mem exprs all the way back to this function's InArgs.
						//  We only want to trace back to the value that comes into the loop. The loop boundary
						//  crossing is recorded in the expression.
						STARSExpression *LoopBoundaryExpr = NonRangeExpr->Clone();
						LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
						LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
						LoopBoundaryExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, true, false, HasArgs, false);
					}
					else {
						NonRangeExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, true, false, HasArgs, false);
						if ((0 != IncomingStackDelta) && StackPtrCase) {
							// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
							SMP_fprintf(HeaderFile, " + 16#%x#", IncomingStackDelta);
						}
					}
					SMP_fprintf(HeaderFile, ", %u) ", ByteWidth);
					++OutputCount;
				} // end if (!NonRangeExpr->IsStackPtrPlusOffset())
			} // end for VectorIndex over all NonRangeExprWidthIters
		} // end if (HasNonRangeExprs)

		if (CalleeWrites) {
			SMP_fprintf(HeaderFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
		} // end if CalleeWrites

		if (NeedsMemoryOutput) {
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i)))");
		}

		//   ********** LOOP INVARIANTS ******************
		// Emit similar loop invariants that mimic the post-conditions.
		OutputCount = 0; // reset for loop invariants
		if (HasBIVInitialValue) { // Need loop invariant for basic induction var modulo value if increment step is not 1
			// NOTE: We could have more than one BIV. Need to expand.
			STARSInductionVarFamilyList::const_iterator BIVIter = this->LoopAnalyzedBIVIters[LoopIndex];
			assert(BIVIter != this->LoopInductionVars[LoopIndex].end());
			STARSOpndTypePtr BIVIncrementOp = BIVIter->BasicInductionVar.Addend.GetOp();
			if (BIVIncrementOp->IsImmedOp()) {
				STARS_sval_t IncrementValue = (STARS_sval_t) BIVIncrementOp->GetImmedValue();
				if (0 > IncrementValue)
					IncrementValue = (0 - IncrementValue); // make it positive by taking absolute value
				if (IncrementValue > 1) { // not unit stride
					// Print "pragma Loop_Invariant(((X86.RAX - X86.RAX'Loop_Entry) mod IncrementValue) = 0);
					STARS_ea_t BIVInitAddr = BIVIter->BIVIncomingDefAddr;
					assert(STARS_BADADDR != BIVInitAddr);
					STARSOpndTypePtr BIVDefOp = BIVIter->BasicInductionVar.InductionVar.GetOp();
					SMPInstr *BIVInitInst = this->GetInstFromAddr(BIVInitAddr);
					assert(nullptr != BIVInitInst);
					PrintSPARKIndentTabs(BodyFile);
					SMP_fprintf(BodyFile, "pragma Loop_Invariant(((");
					BIVInitInst->PrintSPARKAdaOperand(BIVDefOp, BodyFile, false, UseFP, true, false, false);
					SMP_fprintf(BodyFile, "- ");
					BIVInitInst->PrintSPARKAdaOperand(BIVDefOp, BodyFile, false, UseFP, true, true, false);
					SMP_fprintf(BodyFile, "'Loop_Entry) mod %d) = 0); -- BIV relationship to initial and increment values\n", IncrementValue);
					PrintSPARKIndentTabs(BodyFile);
					// ++OutputCount; // leave count 0 for mem range exprs below (starting new loop invariants)
				}
			}
		}	
		if (StackWritten) {
			PrintSPARKIndentTabs(BodyFile);
			SMP_fprintf(BodyFile, "pragma Loop_Invariant(for all i in Unsigned64 => (if ");
			this->EmitSPARKStackMemRangePostCondition(BodyFile, LoopIndex, LoopAddr, false);
			SMP_fprintf(BodyFile, " -- identify stack addresses changed; no other stack addrs change. ");
			++OutputCount;
		}

		this->EmitSPARKMemRangeLoopInvariants(BodyFile, LoopIndex, IncomingStackDelta, true, OutputCount);

		// Another pragma Loop_Invariant that has a corresponding post-condition is to assert
		//  that a register is loop-invariant if it is NOT in the OutputRegs bitset for this loop,
		//  and it is in the CalleePreservedRegsByLoop bitset for this loop.
		for (size_t RegNo = 0; RegNo < LoopPreservedRegs.size(); ++RegNo) {
			if (LoopPreservedRegs[RegNo]) {
				PrintSPARKIndentTabs(BodyFile);
				SMP_fprintf(BodyFile, "pragma Loop_Invariant(");
				STARSOpndTypePtr RegOp = FirstInst->MakeRegOpnd((STARS_regnum_t) RegNo);
				FirstInst->PrintSPARKAdaOperand(RegOp, BodyFile, false, UseFP, true, false, false);
				SMP_fprintf(BodyFile, "= ");
				FirstInst->PrintSPARKAdaOperand(RegOp, BodyFile, false, UseFP, true, true, false);
				SMP_fprintf(BodyFile, "'Loop_Entry); -- reg unchanged by loop, or saved and restored by callees from loop.\n");
			}
		}
#if 0
		if (InvariantPreservedRegs) {
			PrintSPARKIndentTabs(BodyFile);
		}
#endif
	} // end if (NeedsOutput)

	if (NeedsOutput) {
		SMP_fprintf(HeaderFile, ");\n\n"); // terminate the postcondition section
	}
	else { // nothing printed
		SMP_fprintf(HeaderFile, ";\n\n"); // terminate the specification
	}

	return NeedsOutput;
} // end of SMPFunction::EmitSPARKLoopMemRangePostCondition()
#endif

// Emit only the mem range Loop_Invariant (or pragma Assert, if LoopInvariant is false)
void SMPFunction::EmitSPARKMemRangeLoopInvariants(FILE *BodyFile, size_t LoopIndex, STARS_sval_t IncomingStackDelta, bool LoopInvariant, size_t &OutputCount) {
	size_t LoopNumPlusOne = LoopIndex + 1;
	bool HasRangeExprs = (!this->TempRangeExprWidthIters.empty());
	bool HasNonRangeExprs = (!this->TempNonRangeExprWidthIters.empty());
	bool HasRelationalExprs = (!this->RelationalMemWriteWidths[LoopNumPlusOne].empty());
	bool CalleeWrites = this->LoopHasCalleeMemWrites[LoopIndex];
	bool HasArgs = this->DoesLoopHaveArgs((int) LoopIndex);
	string ContractString;
	if (LoopInvariant)
		ContractString = "Loop_Invariant";
	else
		ContractString = "Assert";

	if (HasRangeExprs) {
		if (0 == OutputCount) {
			SMP_fprintf(BodyFile, "pragma %s(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ", ContractString.c_str());
		}
		else {
			SMP_fprintf(BodyFile, "\n\t\tand not X86.InMemoryRange(i, ");
		}
		size_t VectorLimit = this->TempRangeExprWidthIters.size();
		for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
			STARSExpression *LowerExpr = (*this->TempRangeExprWidthIters[VectorIndex].first);
			STARSExpression *UpperExpr = (*this->TempRangeExprWidthIters[VectorIndex].second);
			this->EmitSPARKMemRange(BodyFile, false, true, HasArgs, LowerExpr, UpperExpr, OutputCount, 0);
		}
	}
	if (HasRelationalExprs) {
		if (0 == OutputCount) {
			SMP_fprintf(BodyFile, "pragma %s(for all i in Unsigned64 => (if not X86.InMemoryRange(i, ", ContractString.c_str());
		}
		else {
			SMP_fprintf(BodyFile, "\n\t\tand not X86.InMemoryRange(i, ");
		}
		std::list<std::pair<std::size_t, std::pair<STARSExprSetIter, STARSExprSetIter> > >::const_iterator RelationalIter;
		for (RelationalIter = this->RelationalMemWriteWidths[LoopNumPlusOne].cbegin(); RelationalIter != this->RelationalMemWriteWidths[LoopNumPlusOne].cend(); ++RelationalIter) {
			size_t MemWidth = (*RelationalIter).first;
			STARSExprSetIter LowerExprIter = (*RelationalIter).second.first;
			STARSExprSetIter UpperExprIter = (*RelationalIter).second.second;
			STARSExpression *LowerExpr = (*LowerExprIter);
			STARSExpression *UpperExpr = (*UpperExprIter);
			this->EmitSPARKMemRange(BodyFile, false, true, HasArgs, LowerExpr, UpperExpr, OutputCount, MemWidth);
		}
	}

	if (HasNonRangeExprs) {
		size_t VectorLimit = this->TempNonRangeExprWidthIters.size();
		for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
			// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
			//  which would require printing the same operand twice.
			size_t ByteWidth = this->TempNonRangeExprWidthIters[VectorIndex].first;
			STARSExpression *NonRangeExpr = (*this->TempNonRangeExprWidthIters[VectorIndex].second);

			// The StackPtrCase below is now handled via the {Posi,Nega}tiveStackOffsetWritesByLoop[] vectors
			if (!NonRangeExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
				if (OutputCount == 0) {
					SMP_fprintf(BodyFile, "pragma %s(for all i in Unsigned64 => (if not X86.InRange64(i, ", ContractString.c_str());
				}
				else {
					SMP_fprintf(BodyFile, "\n\t\tand not X86.InRange64(i, ");
				}
				// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
				//  this loop. We need to make two adjustments when printing these exprs:
				//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
				//   2. Offset the stack address with the negated incoming stack offset.
				bool StackPtrCase = (!NonRangeExpr->HasLeftSubTree()) && NonRangeExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
				STARS_ea_t PreLoopDefAddr = NonRangeExpr->GetLeftPreLoopDefAddr();
				bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
				if (!StackPtrCase && (!PseudoAddr)) {
					// We don't want to trace callee mem exprs all the way back to this function's InArgs.
					//  We only want to trace back to the value that comes into the loop. The loop boundary
					//  crossing is recorded in the expression.
					STARSExpression *LoopBoundaryExpr = NonRangeExpr->Clone();
					LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
					LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
					LoopBoundaryExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs, false);
				}
				else {
					NonRangeExpr->EmitSPARKAda(BodyFile, !StackPtrCase, false, false, HasArgs, false);
					if ((0 != IncomingStackDelta) && StackPtrCase) {
						// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
						SMP_fprintf(BodyFile, " + 16#%x#", IncomingStackDelta);
					}
				}
				SMP_fprintf(BodyFile, ", %u) ", ByteWidth);
				++OutputCount;
			} // end if (!NonRangeExpr->IsStackPtrPlusOffset())
		} // end for VectorIndex over all NonRangeExprWidthIters
	} // end if (HasNonRangeExprs)

	if (CalleeWrites) {
		SMP_fprintf(BodyFile, "\n\t\tand X86.InSafeRegion64(i, X86.RSP - 8)"); // only for CalleeWrites case
	} // end if CalleeWrites

	if (0 < OutputCount) {
		SMP_fprintf(BodyFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Loop_Entry(i))));\n");
	}

	return;
} // SMPFunction::EmitSPARKMemRangeLoopInvariants()

// collect temporary containers of mem exprs for one loop.
void SMPFunction::AggregateLoopMemExprs(size_t LoopIndex) {
	size_t LoopNumPlusOne = LoopIndex + 1;
	pair<STARSExprSetIter, bool> InsertResult, InsertResult2;
	// Clear out temp containers that might have info from an enclosing loop.
	this->TempRangeExprWidthIters.clear();
	this->TempNonRangeExprWidthIters.clear();
	this->TempNonRangeExprs.clear();
	this->TempLowerBoundsExprs.clear();
	this->TempUpperBoundsExprs.clear();
	this->RelationalMemWriteWidths[LoopNumPlusOne].clear();
	this->RelationalLowerBoundExprs[LoopNumPlusOne].clear();
	this->RelationalUpperBoundExprs[LoopNumPlusOne].clear();

	if (this->NonStackFrameLoopMemWrites[LoopIndex]) { 
		STARSExprBoundsIter ExprIter;
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			if (!IsLocalStackFrameExprPair(LowerExpr, UpperExpr)) { // not just local var write
				InsertResult = this->TempLowerBoundsExprs.insert(LowerExpr);
				InsertResult2 = this->TempUpperBoundsExprs.insert(UpperExpr);
				if (InsertResult.second || InsertResult2.second) {
					// New range, because at least one insert was not a duplicate.
					pair<STARSExprSetIter, STARSExprSetIter> InsertValue(InsertResult.first, InsertResult2.first);
					this->TempRangeExprWidthIters.push_back(InsertValue);
				}
			}
		} // end for all loop memory write exprs
	}

	// Two cases. (1) From outside any loop, we call foo() and it has writes inside a loop.
	//  (2) From inside a loop, we call foo() and it has writes inside a loop.
	// Case (2) is more complex, as we need to analyze the worst-case range of mem addresses
	//  given that we have nested loops, albeit across procedure boundaries.
	// Case (1) is found in index 0 of vector this->LoopMemAddrExprsFromCalleeLoops[] while
	//  case (2) is found at index [LoopIndex + 1] in the same vector.
	list<std::pair<STARSExprSetIter, size_t> >::const_iterator MemExprIter;
	for (MemExprIter = this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].cbegin(); MemExprIter != this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].cend(); ++MemExprIter) {
		pair<STARSExprSetIter, size_t> MemExprItem = (*MemExprIter);
		size_t MemWidth = MemExprItem.second;
		STARSExprSetIter ExprSetIter = MemExprItem.first;
		if ((*ExprSetIter)->IsStackPtrPlusOffset())
			continue;
		else if ((*ExprSetIter)->IsRelationalExpr()) {
			SMP_msg("INFO: SPARK: Saved LoopMemAddrFromCallees relational expr in EmitSPARKProcPrePostMemConditions for LoopPlusOne: %zu Func: %s\n",
				LoopNumPlusOne, this->GetFuncName());
			// We have already computed the upper bound taking into account the ByteWidth, so pass
			//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
			this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, (*ExprSetIter));
			continue;
		}
		else { // non-range expr
			InsertResult = this->TempNonRangeExprs.insert(*ExprSetIter);
			if (InsertResult.second) {
				// Unique new expr.
				pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
				this->TempNonRangeExprWidthIters.push_back(InsertValue);
			}
		}
	} // end for MemExprIter from LoopMemAddrExprWidthsFromCalleeLoops

	for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
		if (this->NonStackFrameCalleeMemWrites[LoopNumPlusOne][i]) { // not just local stack frame write
			size_t MemWidth = this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].second;
			STARSExpression *CalleeExpr = (*this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].first);
			if (CalleeExpr->IsRelationalExpr()) {
				SMP_msg("INFO: SPARK: Saved MemExprFromCallees relational expr in EmitSPARKProcPrePostMemConditions for LoopPlusOne: %zu Func: %s\n",
					LoopNumPlusOne, this->GetFuncName());
				// We have already computed the upper bound taking into account the ByteWidth, so pass
				//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
				this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, CalleeExpr);
				continue;
			}
			else {
				InsertResult = this->TempNonRangeExprs.insert(CalleeExpr);
				if (InsertResult.second) {
					// Unique new expr.
					pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
					this->TempNonRangeExprWidthIters.push_back(InsertValue);
				}
			}
		}
	} // end for i in this->MemAddrExprWidthsFromCallees[LoopNumPlusOne]

	for (size_t i = 0; i < this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne].size(); ++i) {
		uint32_t ExprSerialNumber = this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].first;
		STARSExprSetIter InArgExprIter = this->FindInArgExprBySerialNumber(LoopNumPlusOne, ExprSerialNumber);
		assert(InArgExprIter != this->GetLastInArgExprSetIter(LoopNumPlusOne));
		STARSExpression *InArgExpr = (*InArgExprIter);
		size_t MemWidth = (size_t) this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].second;
		if (InArgExpr->IsRelationalExpr()) {
			SMP_msg("INFO: SPARK: Saved InArg relational expr in AggregateLoopMemExprs for LoopPlusOne: %zu Func: %s\n",
				LoopNumPlusOne, this->GetFuncName());
			// We have already computed the upper bound taking into account the ByteWidth, so pass
			//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
			this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, InArgExpr);
			continue;
		}
		else {
			uint32_t ExprSerialNumber = this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].first;
			STARSExprSetIter InArgExprIter = this->FindInArgExprBySerialNumber(LoopNumPlusOne, ExprSerialNumber);
			assert(InArgExprIter != this->GetLastInArgExprSetIter(LoopNumPlusOne));
			InsertResult = this->TempNonRangeExprs.insert((*InArgExprIter));
			if (InsertResult.second) {
				// Unique new expr.
				pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
				this->TempNonRangeExprWidthIters.push_back(InsertValue);
			}
		}
	} // end for i in current LoopIndexPlusOne size()

	return;
} // end of SMPFunction::AggregateLoopMemExprs()

#if 0
// specify stack mem ranges that loop changes
void SMPFunction::EmitSPARKStackMemRangePostCondition(FILE *HeaderFile, std::size_t LoopIndex, STARS_ea_t LoopAddr) {
	SMPInstr *FirstInst = this->GetInstFromAddr(LoopAddr);
	assert(nullptr != FirstInst);
	STARS_sval_t IncomingStackDelta = FirstInst->GetStackPtrOffset();
	bool LoopInvariantCase = (HeaderFile == global_STARS_program->GetSPARKSourceFile());
	bool FirstOutputEmitted = false;

	if (!LoopInvariantCase) {
		SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if ");
	}
	if (this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet()) {
		int FirstBitInRange = -1;
		for (size_t i = 0; i < this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits(); ++i) {
			if (this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetBit(i)) {
				if (0 > FirstBitInRange) {
					// We need to start a new range.
					FirstBitInRange = (int) i;
				}
			}
			else {
				if (0 <= FirstBitInRange) {
					// We are working on a range that is now ending.
					STARS_sval_t LowestOffset = (-IncomingStackDelta) + (STARS_sval_t) FirstBitInRange;
					STARS_sval_t LimitOffset = (-IncomingStackDelta) + (STARS_sval_t) i;
					assert(0 <= LowestOffset);
					if (FirstOutputEmitted) {
						SMP_fprintf(HeaderFile, "\n\t\tand ");
					}
					if (LoopInvariantCase) {
						if (0 != LowestOffset)
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP+%lld, X86.RSP+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
						else
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP, X86.RSP+%lld)", (int64_t)LimitOffset);
					}
					else {
						if (0 != LowestOffset)
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old+%lld, X86.RSP'Old+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
						else
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old, X86.RSP'Old+%lld)", (int64_t)LimitOffset);
					}
					FirstBitInRange = -1; // start new range
					FirstOutputEmitted = true;
				}
			}
		}
		if (0 <= FirstBitInRange) {
			// We were working on a range that is now ending at the limit.
			STARS_sval_t LowestOffset = (-IncomingStackDelta) + (STARS_sval_t) FirstBitInRange;
			STARS_sval_t LimitOffset = (-IncomingStackDelta) + (STARS_sval_t) this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits();
			assert(0 <= LowestOffset);
			if (FirstOutputEmitted) {
				SMP_fprintf(HeaderFile, "\n\t\tand ");
			}
			if (LoopInvariantCase) {
				if (0 != LowestOffset)
					SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP+%lld, X86.RSP+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
				else
					SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP, X86.RSP+%lld)", (int64_t)LimitOffset);
			}
			else {
				if (0 != LowestOffset)
					SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old+%lld, X86.RSP'Old+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
				else
					SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old, X86.RSP'Old+%lld)", (int64_t)LimitOffset);
			}
			FirstOutputEmitted = true;
		}
	}

	if (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet()) {
		int FirstBitInRange = -1;
		STARS_sval_t NumFrameBytes = (STARS_sval_t) this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits();
		for (size_t i = 0; i < this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits(); ++i) {
			if (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetBit(i)) {
				if (0 > FirstBitInRange) {
					// We need to start a new range.
					FirstBitInRange = (int) i;
				}
			}
			else {
				if (0 <= FirstBitInRange) {
					// We are working on a range that is now ending.
					STARS_sval_t LowestOffset = (-NumFrameBytes - IncomingStackDelta) + (STARS_sval_t) FirstBitInRange;
					STARS_sval_t LimitOffset = (-NumFrameBytes - IncomingStackDelta) + (STARS_sval_t) i;
					if (FirstOutputEmitted) {
						SMP_fprintf(HeaderFile, "\n\t\tand ");
					}
					if (LoopInvariantCase) {
						if (0 != LowestOffset)
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP+%lld, X86.RSP+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
						else
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP, X86.RSP+%lld)", (int64_t)LimitOffset);
					}
					else {
						if (0 != LowestOffset)
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old+%lld, X86.RSP'Old+%lld)", (int64_t)LowestOffset, (int64_t)LimitOffset);
						else
							SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old, X86.RSP'Old+%lld)", (int64_t)LimitOffset);
					}
					FirstBitInRange = -1; // start new range
					FirstOutputEmitted = true;
				}
			}
		}
		if (0 <= FirstBitInRange) {
			// We were working on a range that is now ending at the limit.
			STARS_sval_t LowestOffset = (-NumFrameBytes - IncomingStackDelta) + (STARS_sval_t) FirstBitInRange;
			if (FirstOutputEmitted) {
				SMP_fprintf(HeaderFile, "\n\t\tand ");
			}
			if (LoopInvariantCase) {
				SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP%lld, X86.RSP)", (int64_t)LowestOffset);
			}
			else {
				SMP_fprintf(HeaderFile, "not X86.InMemoryRange(i, X86.RSP'Old%lld, X86.RSP'Old)", (int64_t)LowestOffset);
			}
			FirstOutputEmitted = true;
		}
	}

	return;
} // end of SMPFunction::EmitSPARKStackMemRangePostCondition()

#else // print one byte address on the stack at a time instead of using range functions
// specify stack mem ranges that loop changes
void SMPFunction::EmitSPARKStackMemRangePostCondition(FILE *HeaderFile, std::size_t LoopIndex, STARS_ea_t LoopAddr, bool PostPrintStarted) {
	SMPInstr *FirstInst = this->GetInstFromAddr(LoopAddr);
	assert(nullptr != FirstInst);
	STARS_sval_t IncomingStackDelta = FirstInst->GetStackPtrOffset();
	bool LoopInvariantCase = (HeaderFile == global_STARS_program->GetSPARKSourceFile()); // if HeaderFile is really the body file
	size_t OutputCount = 0;  // counter; do two outputs per line

	if (!LoopInvariantCase) {
		if (PostPrintStarted) {
			SMP_fprintf(HeaderFile, "\n\tand (for all i in Unsigned64 => (if ");
		}
		else {
			SMP_fprintf(HeaderFile, " (for all i in Unsigned64 => (if ");
		}
	}
	if (this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet()) {
		int FirstBitInRange = -1;
		for (size_t i = 0; i < this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits(); ++i) {
			if (this->PositiveOffsetStackBytesWrittenByLoop[LoopIndex].GetBit(i)) {
				if (0 != OutputCount) {
					if (0 == (OutputCount % 2)) { // printed multiple of 2, start new line
						SMP_fprintf(HeaderFile, " and\n\t\t");
					}
					else {
						SMP_fprintf(HeaderFile, " and ");
					}
				}
				else {
					SMP_fprintf(HeaderFile, "\n\t\t");
				}


				STARS_sval_t Offset = (-IncomingStackDelta) + (STARS_sval_t)i;
				assert(0 <= Offset);
				if (0 != Offset)
					SMP_fprintf(HeaderFile, "(i /= X86.RSP+16#%llx#)", (uint64_t) Offset);
				else
					SMP_fprintf(HeaderFile, "(i /= X86.RSP)");
				++OutputCount;
			}
		} // end for all positive offset stack bytes written
	}

	if (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].IsAnyBitSet()) {
		STARS_sval_t NumFrameBytes = (STARS_sval_t) this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits();
		for (size_t i = 0; i < this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetNumBits(); ++i) {
			if (this->NegativeOffsetStackBytesWrittenByLoop[LoopIndex].GetBit(i)) {
				STARS_sval_t Offset = (-NumFrameBytes - IncomingStackDelta) + (STARS_sval_t) i;
				if (0 != OutputCount) {
					if (0 == (OutputCount % 2)) { // printed multiple of 2, start new line
						SMP_fprintf(HeaderFile, " and\n\t\t");
					}
					else {
						SMP_fprintf(HeaderFile, " and ");
					}
				}
				else {
					SMP_fprintf(HeaderFile, "\n\t\t");
				}

				if (0 < Offset)
					SMP_fprintf(HeaderFile, "(i /= X86.RSP+16#%llx#)", (uint64_t) Offset);
				else if (0 > Offset) 
					SMP_fprintf(HeaderFile, "(i /= X86.RSP-16#%llx#)", (uint64_t) (0 - Offset));
				else
					SMP_fprintf(HeaderFile, "(i /= X86.RSP)");
				++OutputCount;
			}
		} // end for all negative offset stack bytes written
	}

	if (this->LoopMemExprsExpandToStackOffsets[LoopIndex]) {
		STARS_sval_t StackDeltaBias = (-IncomingStackDelta);

		STARSExprBoundsIter ExprIter;
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			bool StackExprFound = ((nullptr != LowerExpr) && (nullptr != UpperExpr))
				&& (LowerExpr->IsStackPtrPlusOffset() || UpperExpr->IsStackPtrPlusOffset());
			if (StackExprFound) {
				if (0 < OutputCount) {
					SMP_fprintf(HeaderFile, "\n\t\t and not X86.InMemoryRange(i, ");
				}
				else {
					SMP_fprintf(HeaderFile, "\n\t\t not X86.InMemoryRange(i, ");
				}
				LowerExpr->EmitSPARKAda(HeaderFile, false, !LoopInvariantCase, false, false, false);
				if (StackDeltaBias > 0) {
					SMP_fprintf(HeaderFile, "+ 16#%llx# ", (uint64_t)StackDeltaBias);
				}
				SMP_fprintf(HeaderFile, ", ");
				UpperExpr->EmitSPARKAda(HeaderFile, false, !LoopInvariantCase, false, false, false);
				if (StackDeltaBias > 0) {
					SMP_fprintf(HeaderFile, "+ 16#%llx# ", (uint64_t)StackDeltaBias);
				}
				SMP_fprintf(HeaderFile, ") ");
				++OutputCount;
			}
		} // end for all loop mem write exprs
	}

	return;
} // end of SMPFunction::EmitSPARKStackMemRangePostCondition()
#endif

// emit loop invariant for basic induction var bound
void SMPFunction::EmitSPARKAdaForLoopLimit(FILE *BodyFile, STARS_ea_t LoopAddr) {
	SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(LoopAddr);
	assert(nullptr != CurrBlock);
	int LoopHeaderBlockNum = CurrBlock->GetNumber();
	int LoopNum = this->GetLoopNumFromHeaderBlockNum(LoopHeaderBlockNum);
	assert(0 <= LoopNum);
	assert(LoopNum < (int)this->LoopCount);
	size_t LoopIndex = (size_t) LoopNum;

	if ((this->LoopIterationsLimitExprs[LoopIndex] != nullptr) 
		&& (this->LoopAnalyzedBIVIters[LoopIndex] != this->LoopInductionVars[LoopIndex].end())) {
		// We analyzed the loop iterations count
	
		this->EmitSPARKLoopBIVLimits(BodyFile, LoopAddr, LoopIndex, false);

		// Print range limitation on memory writing.
		bitset<1 + MD_LAST_REG_NO> RegsPrinted; // record regs printed to avoid redundant output
		size_t DummyOutputCount = 1;
		STARSExprBoundsIter ExprIter;
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			if ((nullptr != LowerExpr) && (nullptr != UpperExpr)) {
				LowerExpr->PrintSPARKArgLocationStrings(BodyFile, true, LoopIndex, DummyOutputCount, RegsPrinted);
				UpperExpr->PrintSPARKArgLocationStrings(BodyFile, true, LoopIndex, DummyOutputCount, RegsPrinted);
			}
		} // end for all loop mem write exprs
	}
	return;
} // end of SMPFunction::EmitSPARKAdaForLoopLimit()

// emit LoopRegSourceExprPairs
void SMPFunction::EmitIncomingLoopRegExprs(FILE *OutputFile, size_t LoopNum, bool LoopInvariantSection) {
	list<pair<STARS_regnum_t, STARSExprSetIter> >::const_iterator LoopRegIter;
	bool HasArgs = this->DoesLoopHaveArgs((int) LoopNum);
	for (LoopRegIter = this->LoopRegSourceExprPairs[LoopNum].cbegin(); LoopRegIter != this->LoopRegSourceExprPairs[LoopNum].cend(); ++LoopRegIter) {
		STARS_regnum_t RegNo = (*LoopRegIter).first;
		STARSExprSetIter RegSourceIter = (*LoopRegIter).second;
		STARSExpression *RegSourceExpr = (*RegSourceIter);
		STARS_ea_t DefAddr = RegSourceExpr->GetParentInst()->GetAddr();
		SMPInstr *DefInst = this->GetInstFromAddr(DefAddr);
		assert(nullptr != DefInst);
		STARSOpndTypePtr DefOp = DefInst->MakeRegOpnd(RegNo);
		bool UseFP = this->UsesFramePointer();
		if (LoopInvariantSection) {
			PrintSPARKIndentTabs(OutputFile);
			SMP_fprintf(OutputFile, "pragma Loop_Invariant(");
			DefInst->PrintSPARKAdaOperand(DefOp, OutputFile, false, UseFP, true, false, false);
			SMP_fprintf(OutputFile, " = (");
			RegSourceExpr->EmitSPARKAda(OutputFile, true, true, false, HasArgs, true, false);
			SMP_fprintf(OutputFile, ")); -- incoming value of register.\n");
		}
		else { // must be a precondition
			SMP_fprintf(OutputFile, " and\n");
			PrintSPARKIndentTabs(OutputFile);
			SMP_fprintf(OutputFile, "\t(");
			DefInst->PrintSPARKAdaOperand(DefOp, OutputFile, false, UseFP, true, false, false);
			SMP_fprintf(OutputFile, " = (");
			RegSourceExpr->EmitSPARKAda(OutputFile, true, false, false, HasArgs, false, false);
			SMP_fprintf(OutputFile, "))");
		}
	} // end for all this->LoopRegSourceExprPairs[LoopNum]
	return;
} // end of SMPFunction::EmitIncomingLoopRegExprs()

// Emit Assumes and Loop_Invariants for loop BIV and mem ranges.
void SMPFunction::EmitSPARKLoopBIVLimits(FILE *BodyFile, STARS_ea_t LoopAddr, size_t LoopIndex, bool PragmaAssume) {
	SMPInstr *FirstInst = this->GetInstFromAddr(LoopAddr);
	STARS_sval_t IncomingStackDelta = FirstInst->GetStackPtrOffset();
	int LoopNum = (int) LoopIndex;
	bool CountdownLoop = (this->LoopIncrementValue[LoopIndex] < 0);
	bool UseFP = this->UsesFramePointer();
	bool HasArgs = this->DoesLoopHaveArgs(LoopNum);

	// Emit range limitation on the basic induction variable.
	bool SPRelative = false;
	STARSExpression *LowerLimitExpr = nullptr;

	if (PragmaAssume) {
		// Emit pragma Assume(lower_bound < upper_bound);
#if 0 // Must be sure that source was not changed after it was saved in BIVOp and before the beginning of the loop.
		// Either we emit a dangerous assumption because of a source change, or we emit a worthless tautology like
		//  (RSP - 0x2b0 ) < (RSP - 0x100). No point in this pragma assume.
		STARSExpression *UpperLimitExpr = nullptr;
		if (!CountdownLoop) {
			LowerLimitExpr = this->LoopIterationsInitExprs[LoopIndex];
			UpperLimitExpr = this->LoopIterationsLimitExprs[LoopIndex];
			// Look only at the right hand side of the relational LimitExpr.
			assert(UpperLimitExpr->IsRelationalExpr());
			UpperLimitExpr = UpperLimitExpr->CloneRHS();
		}
		else {
			LowerLimitExpr = this->LoopIterationsLimitExprs[LoopIndex];
			UpperLimitExpr = this->LoopIterationsInitExprs[LoopIndex];
			// Look only at the right hand side of the relational LimitExpr.
			assert(LowerLimitExpr->IsRelationalExpr());
			LowerLimitExpr = LowerLimitExpr->CloneRHS();
		}
		bool BothConstExprs = (LowerLimitExpr->IsConstExpr() && UpperLimitExpr->IsConstExpr());
		// No point in an Assume on two constants
		if (!BothConstExprs) {
			SPRelative = LowerLimitExpr->IsStackPtrRegUsed();
			PrintSPARKIndentTabs(BodyFile);
			SMP_fprintf(BodyFile, "pragma Assume(");
			if (SPRelative) {
				SMP_fprintf(BodyFile, "( ");
				LowerLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
				// Offset the stack pointer from its entry point value in the func that includes this loop.
				SMP_fprintf(BodyFile, " + 16#%x# )", (0 - IncomingStackDelta));
			}
			else {
				LowerLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
			}
			SMP_fprintf(BodyFile, " < ");

			SPRelative = UpperLimitExpr->IsStackPtrRegUsed();
			if (SPRelative) {
				SMP_fprintf(BodyFile, "( ");
				UpperLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
				// Offset the stack pointer from its entry point value in the func that includes this loop.
				SMP_fprintf(BodyFile, " + 16#%x# )", (0 - IncomingStackDelta));
			}
			else {
				UpperLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
			}
			if (!CountdownLoop) {
				SMP_fprintf(BodyFile, "); -- Counting upwards loop, BIV lower (initvalue) limit < upper limit.\n");
			}
			else {
				SMP_fprintf(BodyFile, "); -- Counting downwards loop, BIV lower (endvalue) limit < upper limit.\n");
			}
		}
#endif // if 0
		return; // the loop invariants will be done on a later call
	} // end if (PragmaAssume)

	// First, emit the restriction that the BIV is always >= its lower limit.
	STARSOpndTypePtr BIVOp = this->LoopAnalyzedBIVIters[LoopIndex]->BasicInductionVar.InductionVar.GetOp();
	PrintSPARKIndentTabs(BodyFile);
	SMP_fprintf(BodyFile, "pragma Loop_Invariant(");
	FirstInst->PrintSPARKAdaOperand(BIVOp, BodyFile, false, UseFP, true, false, false);
	if (!CountdownLoop) {
		SMP_fprintf(BodyFile, ">= ");
		LowerLimitExpr = this->LoopIterationsInitExprs[LoopIndex];
	}
	else {
		// Countdown loop might or might not execute with BIV == lowerbound.
		if (this->LoopExecutesWithLimitValue[LoopIndex])
			SMP_fprintf(BodyFile, ">= ");
		else
			SMP_fprintf(BodyFile, "> ");
		LowerLimitExpr = this->LoopIterationsLimitExprs[LoopIndex];
	}
	SPRelative = LowerLimitExpr->IsStackPtrRegUsed();
	if (!CountdownLoop) {
		if (SPRelative) {
			SMP_fprintf(BodyFile, "( ");
			LowerLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
			// Offset the stack pointer from its entry point value in the func that includes this loop.
			SMP_fprintf(BodyFile, " + 16#%x# )", (0 - IncomingStackDelta));
		}
		else {
			LowerLimitExpr->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
		}
		SMP_fprintf(BodyFile, "); -- Counting upwards loop; BIV relation to lower (initvalue) limit.\n");
	}
	else {
		// Countdown loop; lower limit expr is relational.
		assert(LowerLimitExpr->IsRelationalExpr());
		if (LowerLimitExpr->HasRightSubTree()) {
			SPRelative = LowerLimitExpr->GetRightTree()->IsStackPtrRegUsed();
			if (SPRelative) {
				SMP_fprintf(BodyFile, "( ");
				LowerLimitExpr->GetRightTree()->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
				// Offset the stack pointer from its entry point value in the func that includes this loop.
				SMP_fprintf(BodyFile, " + 16#%x# )", (0 - IncomingStackDelta));
			}
			else {
				LowerLimitExpr->GetRightTree()->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
			}
		}
		else { // normal case; limit is operand
			FirstInst->PrintSPARKAdaOperand(LowerLimitExpr->GetConstRightOperand(), BodyFile, false, UseFP, true, false, false);
		}
		SMP_fprintf(BodyFile, "); -- Counting downwards loop; BIV relation to lower (endvalue) limit.\n");
	}

	// See if we have a non-constant upper bound for the BIV. If so, print its value, 
	//  which will be found in the right hand side of the LimitExpr for counting-up loops
	//  and in the InitExpr for the CountdownLoop case.
	//  Otherwise, use the constant limit.
	assert(LoopIndex < this->LoopComparisonExprs.size());
	assert(this->LoopIterationsLimitExprs[LoopIndex]->IsRelationalExpr());
	PrintSPARKIndentTabs(BodyFile);
	SMP_fprintf(BodyFile, "pragma Loop_Invariant(");
	FirstInst->PrintSPARKAdaOperand(BIVOp, BodyFile, false, UseFP, true, false, false);
	if (!CountdownLoop) {
		STARSOpndTypePtr LimitOp = this->LoopComparisonExprs[LoopIndex].Operand2.GetOp();
		if (this->LoopExecutesWithLimitValue[LoopIndex])
			SMP_fprintf(BodyFile, "<= ");
		else
			SMP_fprintf(BodyFile, "< ");
		FirstInst->PrintSPARKAdaOperand(LimitOp, BodyFile, false, UseFP, true, false, false);
		SMP_fprintf(BodyFile, "); -- Counting upwards loop; BIV relation to upper (endvalue) limit.\n");
	}
	else { // CountdownLoop case has several differences
		SMP_fprintf(BodyFile, "<= ");
		this->LoopIterationsInitExprs[LoopIndex]->EmitSPARKAda(BodyFile, true, false, false, HasArgs, false, false);
		SMP_fprintf(BodyFile, "); -- Counting downwards loop; BIV relation to upper (initvalue) limit.\n");
	}

#if 0 // Must be sure that source was not changed after it was saved in BIVOp and before the beginning of the loop.
	// Now, print a version of the LimitOp that has been traced back to its source value
	//  if LimitOp is not a constant.
	if (!LimitOp->IsImmedOp()) {
		PrintSPARKIndentTabs(BodyFile);
		SMP_fprintf(BodyFile, "pragma Loop_Invariant(");
		FirstInst->PrintSPARKAdaOperand(BIVOp, BodyFile, false, UseFP, true, false, false);
		if (!CountdownLoop) {
			if (this->LoopExecutesWithLimitValue[LoopIndex])
				SMP_fprintf(BodyFile, "<= ");
			else
				SMP_fprintf(BodyFile, "< ");
		}
		else {
			if (this->LoopExecutesWithLimitValue[LoopIndex])
				SMP_fprintf(BodyFile, ">= ");
			else
				SMP_fprintf(BodyFile, "> ");
		}

		bool SPRelative = false;
		if (this->LoopIterationsLimitExprs[LoopIndex]->HasRightSubTree()) {
			SPRelative = this->LoopIterationsLimitExprs[LoopIndex]->GetRightTree()->IsStackPtrRegUsed();
			if (SPRelative)
				SMP_fprintf(BodyFile, "( ");
			this->LoopIterationsLimitExprs[LoopIndex]->GetRightTree()->EmitSPARKAda(BodyFile, true, false, false, HasArgs, true, false);
		}
		else { // RightOperand
			const STARSOpndTypePtr LimitInitOp = this->LoopIterationsLimitExprs[LoopIndex]->GetConstRightOperand();
			SPRelative = (LimitInitOp->IsRegOp() && LimitInitOp->MatchesReg(MD_STACK_POINTER_REG));
			if (SPRelative)
				SMP_fprintf(BodyFile, "( ");
			FirstInst->PrintSPARKAdaOperand(LimitInitOp, BodyFile, false, UseFP, true, false, true);
		}
		if (SPRelative) {
			// Offset the stack pointer from its entry point value in the func that includes this loop.
			SMP_fprintf(BodyFile, " + 16#%x# )", (0 - IncomingStackDelta));
		}
		if (!CountdownLoop) {
			SMP_fprintf(BodyFile, "); -- Counting upwards loop; BIV relation to upper (endvalue) limit pre-loop source.\n");
		}
		else {
			SMP_fprintf(BodyFile, "); -- Counting downwards loop; BIV relation to upper (initvalue) limit pre-loop source.\n");
		}
	} // end if (!LimitOp->IsImmedOp())
#else // Just limit the BIVOp in relation to its LoopEntry value.
	PrintSPARKIndentTabs(BodyFile);
	SMP_fprintf(BodyFile, "pragma Loop_Invariant(");
	FirstInst->PrintSPARKAdaOperand(BIVOp, BodyFile, false, UseFP, true, false, false);
	if (!CountdownLoop) {
		SMP_fprintf(BodyFile, ">= ");
	}
	else {
		SMP_fprintf(BodyFile, "<= ");
	}
	FirstInst->PrintSPARKAdaOperand(BIVOp, BodyFile, false, UseFP, true, true, false);
	SMP_fprintf(BodyFile, "'LoopEntry); -- BIVOp relation to its LoopEntry value.\n");
#endif

	return;
} // end of SMPFunction::EmitSPARKLoopBIVLimits()

// LoopIndex == -1 indicates main procedure; print warnings if proofs will fail
void SMPFunction::EmitAnalysisProblemWarnings(FILE *HeaderFile, int LoopIndex) {
	assert(-1 <= LoopIndex);
	size_t LoopNumPlusOne = (size_t)(1 + LoopIndex);
	if (LoopIndex < 0) { // regular procedure
		// Aggregate all loop problems into this procedure if this procedure is not marked already.
		if (!(this->SymbolicAnalysisProblems[0] && this->CalleeMemExprProblems[0])) {
			for (size_t LoopNum = 0; LoopNum < this->GetNumLoops(); ++LoopNum) {
				this->SymbolicAnalysisProblems[0] = (this->SymbolicAnalysisProblems[0] || this->SymbolicAnalysisProblems[LoopNum + 1] || this->LoopAnalysisProblems[LoopNum]);
				this->CalleeMemExprProblems[0] = (this->CalleeMemExprProblems[0] || this->CalleeMemExprProblems[LoopNum + 1]);
			}
		}
		if (this->SymbolicAnalysisProblems[0])
			SMP_fprintf(HeaderFile, "-- WARNING: Procedure had symbolic analysis problems.\n");
		if (this->CalleeMemExprProblems[0])
			SMP_fprintf(HeaderFile, "-- WARNING: Procedure had callee mem expr inheritance problems.\n");
	}
	else { // loop procedure
		if (this->LoopAnalysisProblems[(size_t) LoopIndex])
			SMP_fprintf(HeaderFile, "-- WARNING: Loop procedure had iteration or mem range analysis problems.\n");
		if (this->SymbolicAnalysisProblems[LoopNumPlusOne])
			SMP_fprintf(HeaderFile, "-- WARNING: Loop procedure had symbolic analysis problems.\n");
		if (this->CalleeMemExprProblems[LoopNumPlusOne])
			SMP_fprintf(HeaderFile, "-- WARNING: Loop procedure had callee mem expr inheritance problems.\n");
	}
	return;
} // end of SMPFunction::EmitAnalysisProblemWarnings()

// Emit SPARK Ada translation of function
void SMPFunction::EmitFuncSPARKAda(void) {
	if (this->IsLinkerStub()) {
		return;
	}
	bool VerboseOutput = global_stars_interface->VerboseSPARKMode();

	string AdaFuncName(this->GetFuncName());
#if STARS_EMIT_ADA_FOR_MAIN_ONLY
	if (0 != strcmp("main", AdaFuncName.c_str())) {
		return;
	}
#endif
	if (IsStartupFuncName(AdaFuncName)) {
		return;
	}

	SMP_msg("INFO: Emitting SPARK for function %s\n", AdaFuncName.c_str());
	// Convert dashes to underscores in the func name to suit Ada standards.
	replace(AdaFuncName.begin(), AdaFuncName.end(), '-', '_');

	FILE *BodyFile = global_STARS_program->GetSPARKSourceFile();
	FILE *HeaderFile = global_STARS_program->GetSPARKHeaderFile();
	size_t ByteWidth = global_STARS_program->GetSTARS_ISA_Bytewidth();

	// Emit beginning of function body.
	this->TranslatingSPARKLoop = false;
	SMP_fprintf(BodyFile, "procedure %s\nis\n", AdaFuncName.c_str());
	SMP_fprintf(BodyFile, "\tSaveStackPtr : Unsigned64 := X86.RSP with Ghost;\n");
	SMP_fprintf(BodyFile, "\tra0 : Unsigned8 := X86.ReadMem8(X86.RSP) with Ghost;\n");
	for (unsigned short i = 1; i < ByteWidth; ++i) {
		SMP_fprintf(BodyFile, "\tra%u : Unsigned8 := X86.ReadMem8(X86.RSP + %u) with Ghost;\n", i, i);
	}
	this->EmitSPARKSavedArgs(BodyFile);
	SMP_fprintf(BodyFile, "\nbegin\n");
	SMP_fprintf(BodyFile, "\tpragma Assume(X86.RSP = X86.DummyRSP);\n");

	// Emit procedure specification.
	this->EmitAnalysisProblemWarnings(HeaderFile, -1);
	SMP_fprintf(HeaderFile, "procedure %s with\n", AdaFuncName.c_str());
	if (this->AltersSPARKMemory()) {
		SMP_fprintf(HeaderFile, "\tGlobal => (In_Out => (X86.Memory, X86.CarryFlag,\n");
	}
	else {
		SMP_fprintf(HeaderFile, "\tGlobal => (In_Out => (X86.CarryFlag,\n");
	}
	if (this->HasCallToNonReturningFunc() || this->HasCalleeChainWithNonReturningFunc()) {
		SMP_fprintf(HeaderFile, "\t\tX86.OverflowFlag, X86.ParityFlag, X86.SignFlag, X86.ZeroFlag, X86.Exit_Called,\n");
	}
	else {
		SMP_fprintf(HeaderFile, "\t\tX86.OverflowFlag, X86.ParityFlag, X86.SignFlag, X86.ZeroFlag,\n");
	}

	size_t OutCount = 0;
	size_t RegCount = this->OutputRegs.count();
	bool FPStackBelowTopPrinted = false;
	for (size_t RegIndex = 0; RegIndex < this->OutputRegs.size(); ++RegIndex) {
		bool FlagsReg = (RegIndex == (size_t) MD_FLAGS_REG);
		if (FlagsReg || (!(this->OutputRegs[RegIndex] || this->CalleePreservedRegs[RegIndex]))) {
			continue; // flag, or bit not set
		}
		bool CommaPrinted = false;
		bool FPRegBelowStackTop = ((RegIndex >= STARS_x86_R_st1) && (RegIndex <= STARS_x86_R_st7));
		if (!(FPRegBelowStackTop && FPStackBelowTopPrinted)) { // avoid duplicate X86.FloatingPointStackDummy1
			if (0 == OutCount) { // first line does not need another newline
				SMP_fprintf(HeaderFile, "\t\t");
				CommaPrinted = true; // printed a comma after we printed flags
			}
			else if (0 == (OutCount % 4)) {  // print 4 regs per line
				SMP_fprintf(HeaderFile, ",\n\t\t");
				CommaPrinted = true;
			}
			++OutCount;
			if (CommaPrinted) {
				SMP_fprintf(HeaderFile, "X86.%s ", MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, (uint16_t)ByteWidth));
			}
			else {
				SMP_fprintf(HeaderFile, ", X86.%s ", MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, (uint16_t)ByteWidth));
			}
			FPStackBelowTopPrinted |= FPRegBelowStackTop;
		}
	} // end for all entries in the OutputRegs set.
	SMP_fprintf(HeaderFile, ")"); // close In_Out list parenthesis

	OutCount = 0;
	RegCount = this->InputRegs.count();
	if (RegCount > 0) {
		for (size_t RegIndex = 0; RegIndex < this->InputRegs.size(); ++RegIndex) {
			bool FlagsReg = (RegIndex == (size_t) MD_FLAGS_REG);
			if (FlagsReg || this->OutputRegs[RegIndex] || this->CalleePreservedRegs[RegIndex] || (!this->InputRegs[RegIndex])) {
				continue; // flag, or bit not set, or already handled as In_Out
			}
			bool FPRegBelowStackTop = ((RegIndex >= STARS_x86_R_st1) && (RegIndex <= STARS_x86_R_st7));
			if (!(FPRegBelowStackTop && FPStackBelowTopPrinted)) { // avoid duplicate X86.FloatingPointStackDummy1
				bool CommaPrinted = false;
				if (0 == OutCount) { // first line needs Input =>
					if (this->AltersSPARKMemory()) {
						SMP_fprintf(HeaderFile, ",\n\t\tInput => (");
					}
					else {
						SMP_fprintf(HeaderFile, ",\n\t\tInput => (X86.Memory, ");
					}
					CommaPrinted = true; // suppress comma before first register
				}
				else if (0 == (OutCount % 4)) {  // print 4 regs per line
					SMP_fprintf(HeaderFile, ",\n\t\t");
					CommaPrinted = true;
				}
				++OutCount;
				if (CommaPrinted) {
					SMP_fprintf(HeaderFile, "X86.%s ", MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, (uint16_t)ByteWidth));
				}
				else {
					SMP_fprintf(HeaderFile, ", X86.%s ", MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, (uint16_t)ByteWidth));
				}
				FPStackBelowTopPrinted |= FPRegBelowStackTop;
			}
		} // end for all entries in the InputRegs set.
		if (OutCount > 0) { // last reg needs no comma after it, but must close inner parentheses
			SMP_fprintf(HeaderFile, ")");
		}
		else if (!this->AltersSPARKMemory()) {
			SMP_fprintf(HeaderFile, ",\n\t\tInput => (X86.Memory)");
		}
	}
	SMP_fprintf(HeaderFile, "),\n"); // close Globals section parenthesis

	// Emit memory writing range pre-conditions and post-conditions.
	bool HasLoopMemWrites = ((this->LoopCount > 0) && this->MemRangeRegsBitmap.any());
	this->EmitSPARKProcPrePostMemConditions(BodyFile, HeaderFile, true);
	this->EmitSPARKProcPrePostMemConditions(BodyFile, HeaderFile, false);

	// Begin recursive descent translation with entry block.
	STARSTransCFGBlock *CurrTCFGBlock = this->TCFGBlocks[0];
	this->ResetProcessedBlocks(); // no block has been translated yet
#if 0
	int FollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, false);
#else
	int FollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
#endif
	this->EmitSPARKAdaForBlock(0, 0, FollowTCFGBlockNum, BodyFile, false, false);

	SMP_fprintf(BodyFile, "\nend %s;\n\n", AdaFuncName.c_str());
	STARS_SPARK_IndentCount = 1;

	// Translate the loops to procedures
	list<pair<int, pair<int, int> > >::iterator WorkListIter = this->SPARKLoopWorkList.begin();
	while (WorkListIter != this->SPARKLoopWorkList.end()) {
		this->TranslatingSPARKLoop = true;
		int LoopIndex = WorkListIter->first;
		int HeaderTCFGBlockNum = WorkListIter->second.first;
		FollowTCFGBlockNum = WorkListIter->second.second;
		this->ResetProcessedBlocks(); // no block has been translated yet on this WorkList iteration
		// Emit the procedure specification, up through pre-conditions, into HeaderFile and begin the proc in BodyFile.
		string ProcName = this->EmitSPARKProcForLoopHeaderBlock(LoopIndex, HeaderTCFGBlockNum, FollowTCFGBlockNum, BodyFile, HeaderFile);
		// Emit the SPARK code for the loop.
		this->EmitSPARKAdaForBlock(HeaderTCFGBlockNum, HeaderTCFGBlockNum, FollowTCFGBlockNum, BodyFile, false, true);
		// Terminate the procedure emitted for the loop.
		SMP_fprintf(BodyFile, "\nend %s;\n\n", ProcName.c_str());
		STARS_SPARK_IndentCount = 1;

		this->SPARKLoopWorkList.pop_front();
		WorkListIter = this->SPARKLoopWorkList.begin();
	}

	this->TranslatingSPARKLoop = false;

	// Audit completeness of translation.
	bool TranslationComplete = true;
	for (list<SMPInstr *>::iterator InstIter = this->GetFirstInstIter(); InstIter != this->GetLastInstIter(); ++InstIter) {
		SMPInstr *CurrInst = (*InstIter);
		if (!CurrInst->HasBeenTranslatedToSPARK() && (!CurrInst->IsMarkerInst()) && (!CurrInst->IsNop())) {
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			SMPBasicBlock *CurrBlock = this->GetBlockFromInstAddr(InstAddr);
			assert(nullptr != CurrBlock);
			SMP_fprintf(BodyFile, "ERROR: Instruction at %llx block %d not translated.\n",
				(uint64_t)InstAddr, CurrBlock->GetNumber());
			TranslationComplete = false;
		}
	}
	if (!TranslationComplete) {
		SMP_msg("ERROR: Incomplete SPARK Ada translation ");
		this->DumpFuncNameAndAddr();
		this->DumpDotCFG();
		this->Dump(true);
	}

	return;
} // end of SMPFunction::EmitFuncSPARKAda()

#if 0
// Emit mem writing ranges for pre- and post-conditions
void SMPFunction::EmitSPARKProcPrePostMemConditions(FILE *BodyFile, FILE *HeaderFile, bool PreconditionSection) {
	size_t ByteWidth = global_STARS_program->GetSTARS_ISA_Bytewidth();
	bool HasLoopMemWrites = false;
	bool HasNonLoopInArgMemWrites = false;
	bool HasCalleeMemWrites = false;
	bool HasCalleeLoopInArgMemWrites = this->CalleeUsesInArgsForLoopMemWrites();
	bool HasCalleeLoopMemWrites = false; // any source, calleed from inside or outside loop
	bool NeedsPreconditions = false;
	size_t OutputCount = 0;

	if (PreconditionSection) {
		NeedsPreconditions = this->NeedsSPARKPreconditions(HasLoopMemWrites, HasNonLoopInArgMemWrites, HasCalleeMemWrites, HasCalleeLoopMemWrites);
		if (!NeedsPreconditions) {
			return;
		}
	}
	else { // post-conditions
		// Determine booleans that drive output
		HasCalleeLoopMemWrites = this->HasMemExprsFromCalleeLoops;
		for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			if (this->NonStackFrameLoopMemWrites[LoopIndex]) {
				HasLoopMemWrites = true;
				break;
			}
		}
		HasNonLoopInArgMemWrites = (!this->InArgsUsedInMemWrites[0].empty());
		for (size_t LoopNumPlusOne = 0; LoopNumPlusOne <= this->GetNumLoops(); ++LoopNumPlusOne) {
			for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
				if (this->NonStackFrameCalleeMemWrites[LoopNumPlusOne][i]) { // not just local stack frame write
					HasCalleeMemWrites = true;
					break;
				}
			}
		}

		// Start the post-condition section with items not shared with the pre-condition logic.
		if (this->HasCalleeChainWithNonReturningFunc() || this->HasCallToNonReturningFunc()) {
			SMP_fprintf(HeaderFile, "\tPost => X86.Exit_Called or else ");
		}
		else {
			SMP_fprintf(HeaderFile, "\tPost => ");
		}
		if (this->AltersSPARKMemory()) {
			SMP_fprintf(HeaderFile, "(X86.RSP = (X86.RSP'Old + 8)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old) = X86.Memory'Old(X86.RSP'Old)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 1) = X86.Memory'Old(X86.RSP'Old + 1)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 2) = X86.Memory'Old(X86.RSP'Old + 2)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 3) = X86.Memory'Old(X86.RSP'Old + 3)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 4) = X86.Memory'Old(X86.RSP'Old + 4)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 5) = X86.Memory'Old(X86.RSP'Old + 5)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 6) = X86.Memory'Old(X86.RSP'Old + 6)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 7) = X86.Memory'Old(X86.RSP'Old + 7))");
			OutputCount = 9;
		}
		else {
			SMP_fprintf(HeaderFile, "(X86.RSP = (X86.RSP'Old + 8))");
			++OutputCount;
		}
		if (this->PreservedRegsBitmap.any()) {
			size_t RegLimit = this->PreservedRegsBitmap.size();
			if (RegLimit > STARS_x86_R_r15) {
				RegLimit = STARS_x86_R_r15; // last full-width general purpose reg
			}
			for (size_t RegIndex = 0; RegIndex <= RegLimit; ++RegIndex) {
				if (this->IsRegPreserved(RegIndex)) {
					const char *RegName = MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, ByteWidth);
					SMP_fprintf(HeaderFile, " and\n\t\t(X86.%s = X86.%s'Old)", RegName, RegName);
					++OutputCount;
				}
			}
		}
	}

	if (HasLoopMemWrites || HasNonLoopInArgMemWrites || HasCalleeMemWrites || HasCalleeLoopInArgMemWrites || HasCalleeLoopMemWrites) {
		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "\tPre => (");
		else
			SMP_fprintf(HeaderFile, "\tand \n\t(for all i in Unsigned64 => (if X86.InSafeRegion64(i, X86.RSP'Old)");
	}
	else if (!PreconditionSection) {
		if (this->IsPossibleIndirectCallTarget() || this->AltersSPARKMemory()) {
			SMP_fprintf(HeaderFile, "\tand \n\t(for all i in Unsigned64 => (if X86.InSafeRegion64(i, X86.RSP'Old)");
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
		}
		else {
			SMP_fprintf(HeaderFile, ";\n\n"); // terminate postcondition section
		}
	}

	if (!HasNonLoopInArgMemWrites) {
		if (HasLoopMemWrites) {
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, "for all i in Unsigned64 => (if X86.InMemoryRange(i, ");

			for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
				if (this->NonStackFrameLoopMemWrites[LoopIndex]) { // not just local var write
					this->EmitSPARKMemRange(HeaderFile, PreconditionSection, this->LoopMemWriteLowerBoundsExprs[LoopIndex], this->LoopMemWriteUpperBoundsExprs[LoopIndex], OutputCount);
				}
			}
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, " then\n\t\tX86.InSafeRegion64(i, X86.RSP))"); // terminate universal quantifier section
		}
		if (HasCalleeMemWrites) {
			for (size_t LoopNumPlusOne = 0; LoopNumPlusOne <= this->GetNumLoops(); ++LoopNumPlusOne) {
				this->EmitSPARKPrePostMemConditionsFromCalleeNonLoopWrites(HeaderFile, PreconditionSection, LoopNumPlusOne, OutputCount);
			}
		}
		if (HasCalleeLoopMemWrites) {
			this->EmitSPARKPrePostMemConditionsFromCalleeLoops(HeaderFile, PreconditionSection, OutputCount);
		}
		if (HasLoopMemWrites || HasCalleeMemWrites || HasCalleeLoopInArgMemWrites || HasCalleeLoopMemWrites) {
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, "),\n"); // terminate precondition section
			else // terminate the postcondition section
				SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
		}
	}
	else { // HasNonLoopInArgMemWrites is true
		if (HasLoopMemWrites) {
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, "for all i in Unsigned64 => (if X86.InMemoryRange(i, ");

			for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
				if (this->NonStackFrameLoopMemWrites[LoopIndex]) { // not just local var write
					this->EmitSPARKMemRange(HeaderFile, PreconditionSection, this->LoopMemWriteLowerBoundsExprs[LoopIndex], this->LoopMemWriteUpperBoundsExprs[LoopIndex], OutputCount);
				}
			}
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, " then\n\t\tX86.InSafeRegion64(i, X86.RSP))"); // terminate universal quantifier section
		}

		// NonLoopInArgMemWrites
		STARSExpression *PreviousExpr = nullptr;
		for (size_t LoopIndexPlusOne = 0; LoopIndexPlusOne <= this->GetNumLoops(); ++LoopIndexPlusOne) {
			for (size_t i = 0; i < this->InArgsUsedInMemWriteByteWidths[LoopIndexPlusOne].size(); ++i) {
				STARSExprSetIter InArgExprIter = this->InArgsUsedInMemWriteByteWidths[LoopIndexPlusOne][i].first;
				STARSExpression *InArgExpr = (*InArgExprIter);
				// Reduce duplicate output.
				if (nullptr == PreviousExpr) {
					PreviousExpr = InArgExpr;
				}
				else if (PreviousExpr->IsEqualExpr(InArgExpr)) {
					continue;
				}
				else {
					PreviousExpr = InArgExpr;
				}

				// Handle memory ranges in the LoopMemWrites section above.
				if (InArgExpr->IsRelationalExpr())
					continue;

				// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
				//  which would require printing the same operand twice.
				if (0 < OutputCount) {
					if (PreconditionSection)
						SMP_fprintf(HeaderFile, " and ");
					else
						SMP_fprintf(HeaderFile, "\n\t\t and not X86.InRange64(i, ");
				}
				else {
					assert(PreconditionSection);
				}
				size_t MemWidth = (size_t) this->InArgsUsedInMemWriteByteWidths[LoopIndexPlusOne][i].second;
				if (PreconditionSection) {
					InArgExpr->EmitSPARKInSafeRegion64(HeaderFile, MemWidth);
				}
				else {
					InArgExpr->EmitSPARKAda(HeaderFile, false, !PreconditionSection, false, false);
					SMP_fprintf(HeaderFile, ", %u)", MemWidth);
				}
				++OutputCount;
			} // end for i in current LoopIndexPlusOne size()
		} // end for LoopIndexPlusOne

		if (HasCalleeMemWrites) {
			for (size_t LoopNumPlusOne = 0; LoopNumPlusOne <= this->GetNumLoops(); ++LoopNumPlusOne) {
				this->EmitSPARKPrePostMemConditionsFromCalleeNonLoopWrites(HeaderFile, PreconditionSection, LoopNumPlusOne, OutputCount);
			}
		}
		if (HasCalleeLoopMemWrites) {
			this->EmitSPARKPrePostMemConditionsFromCalleeLoops(HeaderFile, PreconditionSection, OutputCount);
		}

		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "),\n"); // terminate precondition section
		else // terminate the postcondition section
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
	} // end if (!HasNonLoopInArgMemWrites) ... else ...

	return;
} // end of SMPFunction::EmitSPARKProcPrePostMemConditions()

#else
// Emit mem writing ranges for pre- and post-conditions
void SMPFunction::EmitSPARKProcPrePostMemConditions(FILE *BodyFile, FILE *HeaderFile, bool PreconditionSection) {
	size_t ByteWidth = global_STARS_program->GetSTARS_ISA_Bytewidth();
	bool HasLoopMemWrites = false;
	bool HasNonLoopInArgMemWrites = false;
	bool HasCalleeMemWrites = false;
	bool HasCalleeLoopInArgMemWrites = this->CalleeUsesInArgsForLoopMemWrites();
	bool HasCalleeLoopMemWrites = false; // any source, calleed from inside or outside loop
	bool NeedsPreconditions = false;
	size_t OutputCount = 0;

	if (PreconditionSection) {
		NeedsPreconditions = this->NeedsSPARKPreconditions(HasLoopMemWrites, HasNonLoopInArgMemWrites, HasCalleeMemWrites, HasCalleeLoopMemWrites);
		if (!NeedsPreconditions) {
			return;
		}
	}
	else { // post-conditions
		// Determine booleans that drive output
		HasCalleeLoopMemWrites = this->HasMemExprsFromCalleeLoops;
		for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			if (this->NonStackFrameLoopMemWrites[LoopIndex]) {
				HasLoopMemWrites = true;
				break;
			}
		}
		HasNonLoopInArgMemWrites = (!this->InArgsUsedInMemWrites[0].empty());
		for (size_t LoopNumPlusOne = 0; LoopNumPlusOne <= this->GetNumLoops(); ++LoopNumPlusOne) {
			for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
				if (this->NonStackFrameCalleeMemWrites[LoopNumPlusOne][i]) { // not just local stack frame write
					HasCalleeMemWrites = true;
					break;
				}
			}
		}

		// Start the post-condition section with items not shared with the pre-condition logic.
		if (this->HasCalleeChainWithNonReturningFunc() || this->HasCallToNonReturningFunc()) {
			SMP_fprintf(HeaderFile, "\tPost => X86.Exit_Called or else ");
		}
		else {
			SMP_fprintf(HeaderFile, "\tPost => ");
		}
		if (this->AltersSPARKMemory()) {
			SMP_fprintf(HeaderFile, "(X86.RSP = (X86.RSP'Old + 8)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old) = X86.Memory'Old(X86.RSP'Old)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 1) = X86.Memory'Old(X86.RSP'Old + 1)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 2) = X86.Memory'Old(X86.RSP'Old + 2)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 3) = X86.Memory'Old(X86.RSP'Old + 3)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 4) = X86.Memory'Old(X86.RSP'Old + 4)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 5) = X86.Memory'Old(X86.RSP'Old + 5)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 6) = X86.Memory'Old(X86.RSP'Old + 6)) and\n");
			SMP_fprintf(HeaderFile, "\t\t(X86.Memory(X86.RSP'Old + 7) = X86.Memory'Old(X86.RSP'Old + 7))");
			OutputCount = 9;
		}
		else {
			SMP_fprintf(HeaderFile, "(X86.RSP = (X86.RSP'Old + 8))");
			++OutputCount;
		}
		if (this->PreservedRegsBitmap.any()) {
			size_t RegLimit = this->PreservedRegsBitmap.size();
			if (RegLimit > STARS_x86_R_r15) {
				RegLimit = STARS_x86_R_r15; // last full-width general purpose reg
			}
			for (size_t RegIndex = 0; RegIndex <= RegLimit; ++RegIndex) {
				// In addition to our own PreservedRegs, a reg could be Preserved by no being written,
				//  and all callees preserving it, in which case it is In_Out but does not need to be
				//  explicitly preserved in this function and will not be in the PreservedRegs bitset.
				bool PreservedByCalleesOnly = ((!this->OutputRegs[RegIndex]) && this->CalleePreservedRegs[RegIndex]);
				if (this->IsRegPreserved(RegIndex) || PreservedByCalleesOnly) {
					const char *RegName = MDGetSPARKRegNumName((STARS_regnum_t)RegIndex, ByteWidth);
					SMP_fprintf(HeaderFile, " and\n\t\t(X86.%s = X86.%s'Old)", RegName, RegName);
					++OutputCount;
				}
			}
		}
	}

	// Use sets of exprs to avoid duplication and streamline the output section.
	STARSExprSet LowerBoundsExprs;
	STARSExprSet UpperBoundsExprs;
	STARSExprSet NonRangeExprs;
	pair<STARSExprSetIter, bool> InsertResult, InsertResult2;
	vector<pair<STARSExprSetIter, STARSExprSetIter> > RangeExprWidthIters;
	vector<pair<size_t, STARSExprSetIter> > NonRangeExprWidthIters;

	for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
		STARSExprBoundsIter ExprIter;
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			bool NonStackFrameAccess = IsLocalStackFrameExprPair(LowerExpr, UpperExpr);

			if (NonStackFrameAccess) { // not just local var write
				InsertResult = LowerBoundsExprs.insert(LowerExpr);
				InsertResult2 = UpperBoundsExprs.insert(UpperExpr);
				if (InsertResult.second || InsertResult2.second) {
					// New range, because at least one insert was not a duplicate.
					pair<STARSExprSetIter, STARSExprSetIter> InsertValue(InsertResult.first, InsertResult2.first);
					RangeExprWidthIters.push_back(InsertValue);
				}
			}
		} // end for all loop mem write exprs
	}

	// Two cases. (1) From outside any loop, we call foo() and it has writes inside a loop.
	//  (2) From inside a loop, we call foo() and it has writes inside a loop.
	// Case (2) is more complex, as we need to analyze the worst-case range of mem addresses
	//  given that we have nested loops, albeit across procedure boundaries.
	// Case (1) is found in index 0 of vector this->LoopMemAddrExprsFromCalleeLoops[] while
	//  case (2) is found at index [LoopIndex + 1] in the same vector.
	size_t IndexLimit = this->LoopMemAddrExprWidthsFromCalleeLoops.size();
	for (size_t LoopNumPlusOne = 0; LoopNumPlusOne < IndexLimit; ++LoopNumPlusOne) {
		list<std::pair<STARSExprSetIter, size_t> >::const_iterator MemExprIter;
		for (MemExprIter = this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].cbegin(); MemExprIter != this->LoopMemAddrExprWidthsFromCalleeLoops[LoopNumPlusOne].cend(); ++MemExprIter) {
			pair<STARSExprSetIter, size_t> MemExprItem = (*MemExprIter);
			size_t MemWidth = MemExprItem.second;
			STARSExprSetIter ExprSetIter = MemExprItem.first;
			if ((*ExprSetIter)->IsStackPtrPlusOffset())
				continue;
			else if ((*ExprSetIter)->IsRelationalExpr()) {
				SMP_msg("INFO: SPARK: Saved LoopMemAddrFromCallees relational expr in EmitSPARKProcPrePostMemConditions for LoopPlusOne: %zu Func: %s\n",
					LoopNumPlusOne, this->GetFuncName());
				// We have already computed the upper bound taking into account the ByteWidth, so pass
				//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
				this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, (*ExprSetIter));
				continue;
			}
			else { // non-range expr
				InsertResult = NonRangeExprs.insert(*ExprSetIter);
				if (InsertResult.second) {
					// Unique new expr.
					pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
					NonRangeExprWidthIters.push_back(InsertValue);
				}
			}
		} // end for MemExprIter from LoopMemAddrExprWidthsFromCalleeLoops

		for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopNumPlusOne].size(); ++i) {
			if (this->NonStackFrameCalleeMemWrites[LoopNumPlusOne][i]) { // not just local stack frame write
				size_t MemWidth = this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].second;
				STARSExpression *CalleeExpr = (*this->MemAddrExprWidthsFromCallees[LoopNumPlusOne][i].first);
				if (CalleeExpr->IsRelationalExpr()) {
					SMP_msg("INFO: SPARK: Saved MemExprFromCallees relational expr in EmitSPARKProcPrePostMemConditions for LoopPlusOne: %zu Func: %s\n",
						LoopNumPlusOne, this->GetFuncName());
					// We have already computed the upper bound taking into account the ByteWidth, so pass
					//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
					this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, CalleeExpr);
					continue;
				}
				else {
					InsertResult = NonRangeExprs.insert(CalleeExpr);
					if (InsertResult.second) {
						// Unique new expr.
						pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
						NonRangeExprWidthIters.push_back(InsertValue);
					}
				}
			}
		} // end for i in this->MemAddrExprWidthsFromCallees[LoopNumPlusOne]

		for (size_t i = 0; i < this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne].size(); ++i) {
			uint32_t ExprSerialNumber = this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].first;
			STARSExprSetIter InArgExprIter = this->FindInArgExprBySerialNumber(LoopNumPlusOne, ExprSerialNumber);
			assert(InArgExprIter != this->GetLastInArgExprSetIter(LoopNumPlusOne));
			STARSExpression *InArgExpr = (*InArgExprIter);
			// Handle memory ranges in the LoopMemWrites section above.
			size_t MemWidth = (size_t) this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].second;
			if (InArgExpr->IsRelationalExpr()) {
				SMP_msg("INFO: SPARK: Saved InArg relational expr in EmitSPARKProcPrePostMemConditions for LoopPlusOne: %zu Func: %s\n", 
					LoopNumPlusOne, this->GetFuncName());
				// We have already computed the upper bound taking into account the ByteWidth, so pass
				//  a ByteWidth of 0 here to avoid double-adding the ByteWidth in EmitSPARKMemRange().
				this->SplitAndSaveRelationalExpr(true, LoopNumPlusOne, 0, InArgExpr);
				continue;
			}
			else {
				uint32_t ExprSerialNumber = this->InArgsUsedInMemWriteByteWidths[LoopNumPlusOne][i].first;
				STARSExprSetIter FoundExprIter = this->FindInArgExprBySerialNumber(LoopNumPlusOne, ExprSerialNumber);
				assert(FoundExprIter != this->GetLastInArgExprSetIter(LoopNumPlusOne));
				InsertResult = NonRangeExprs.insert(*FoundExprIter);
				if (InsertResult.second) {
					// Unique new expr.
					pair<size_t, STARSExprSetIter> InsertValue(MemWidth, InsertResult.first);
					NonRangeExprWidthIters.push_back(InsertValue);
				}
			}
		} // end for i in current LoopIndexPlusOne size()
	} // end for LoopNumPlusOne

	bool HasRangeExprs = (!RangeExprWidthIters.empty());
	bool HasNonRangeExprs = (!NonRangeExprWidthIters.empty());
	bool HasRelationalExprs = false;
	assert(IndexLimit == (1 + this->GetNumLoops()));
	for (size_t LoopNumPlusOne = 0; LoopNumPlusOne < IndexLimit; ++LoopNumPlusOne) {
		if (!this->RelationalMemWriteWidths[LoopNumPlusOne].empty()) {
			HasRelationalExprs = true;
			break;
		}
	}

	if (HasRangeExprs || HasNonRangeExprs || HasRelationalExprs) { // output is needed
		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "\tPre => (");
		else
			SMP_fprintf(HeaderFile, "\tand \n\t(for all i in Unsigned64 => (if X86.InSafeRegion64(i, X86.RSP'Old)");

		if (HasRangeExprs || HasRelationalExprs) {
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, "(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
		}

		if (HasRangeExprs) {
			size_t VectorLimit = RangeExprWidthIters.size();
			for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
				STARSExpression *LowerExpr = (*RangeExprWidthIters[VectorIndex].first);
				STARSExpression *UpperExpr = (*RangeExprWidthIters[VectorIndex].second);
				this->EmitSPARKMemRange(HeaderFile, PreconditionSection, false, false, LowerExpr, UpperExpr, OutputCount, 0);
			}
		}

		if (HasRelationalExprs) {
			for (size_t LoopNumPlusOne = 0; LoopNumPlusOne < IndexLimit; ++LoopNumPlusOne) {
				std::list<std::pair<std::size_t, std::pair<STARSExprSetIter, STARSExprSetIter> > >::const_iterator RelationalIter;
				for (RelationalIter = this->RelationalMemWriteWidths[LoopNumPlusOne].cbegin(); RelationalIter != this->RelationalMemWriteWidths[LoopNumPlusOne].cend(); ++RelationalIter) {
					size_t MemWidth = (*RelationalIter).first;
					STARSExprSetIter LowerExprIter = (*RelationalIter).second.first;
					STARSExprSetIter UpperExprIter = (*RelationalIter).second.second;
					this->EmitSPARKMemRange(HeaderFile, PreconditionSection, false, false, (*LowerExprIter), (*UpperExprIter), OutputCount, MemWidth);
				}
			}
		}

		if (HasRangeExprs || HasRelationalExprs) {
			if (PreconditionSection)
				SMP_fprintf(HeaderFile, " then\n\t\tX86.InSafeRegion64(i, X86.RSP)))"); // terminate universal quantifier section
		}

		if (HasNonRangeExprs) {
			size_t VectorLimit = NonRangeExprWidthIters.size();
			for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
				// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
				//  which would require printing the same operand twice.
				if (0 < OutputCount) {
					if (PreconditionSection)
						SMP_fprintf(HeaderFile, " and ");
					else
						SMP_fprintf(HeaderFile, "\n\t\t and not X86.InRange64(i, ");
				}
				else {
					assert(PreconditionSection);
				}
				size_t MemWidth = NonRangeExprWidthIters[VectorIndex].first;
				STARSExpression *NonRangeExpr = (*NonRangeExprWidthIters[VectorIndex].second);
				if (PreconditionSection) {
					NonRangeExpr->EmitSPARKInSafeRegion64(HeaderFile, MemWidth);
				}
				else {
					NonRangeExpr->EmitSPARKAda(HeaderFile, false, !PreconditionSection, false, false, false);
					SMP_fprintf(HeaderFile, ", %u)", MemWidth);
				}
				++OutputCount;
			} // end for VectorIndex over all NonRangeExprWidthIters
		} // end if (HasNonRangeExprs)

		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "),\n"); // terminate precondition section
		else // terminate the postcondition section
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
	} // end if (HasRangeExprs || HasNonRangeExprs || HasRelationalExprs)
	else if (!PreconditionSection) {
		if (this->IsPossibleIndirectCallTarget() || this->AltersSPARKMemory()) {
			SMP_fprintf(HeaderFile, "\tand \n\t(for all i in Unsigned64 => (if X86.InSafeRegion64(i, X86.RSP'Old)");
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
		}
		else {
			SMP_fprintf(HeaderFile, ";\n\n"); // terminate postcondition section
		}
	}

#if 0
	// Temporary fix: If the only exprs were relational and were skipped above, we need to
	//  close the precondition or postcondition section.
	if ((0 == OutputCount) && (HasRangeExprs || HasNonRangeExprs)) {
		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "),\n"); // terminate precondition section
		else // terminate the postcondition section
			SMP_fprintf(HeaderFile, " then\n\t\t(X86.Memory(i) = X86.Memory'Old(i))));\n\n");
	}
#endif

	RangeExprWidthIters.clear();
	NonRangeExprWidthIters.clear();
	NonRangeExprs.clear();
	LowerBoundsExprs.clear();
	UpperBoundsExprs.clear();

	return;
} // end of SMPFunction::EmitSPARKProcPrePostMemConditions()

#endif

void SMPFunction::EmitSPARKMemRange(FILE *HeaderFile, bool PreconditionSection, bool ProcessingLoop, bool HasArgs, STARSExpression *LowerExpr, STARSExpression *UpperExpr, size_t &OutputCount, size_t MemWidth) {
	if (nullptr == LowerExpr) {
		SMP_msg("ERROR: SPARK: EmitSPARKMemRange() called with LowerExpr == nullptr ");
		this->DumpFuncNameAndAddr();
		return;
	}

	if (0 < OutputCount) {
		if (PreconditionSection)
			SMP_fprintf(HeaderFile, "\n\t\t or X86.InMemoryRange(i, ");
		else
			SMP_fprintf(HeaderFile, "\n\t\t and not X86.InMemoryRange(i, ");
	}
	bool OldSuffix = (!(ProcessingLoop || PreconditionSection));
	LowerExpr->EmitSPARKAda(HeaderFile, ProcessingLoop, OldSuffix, false, HasArgs, false, false);
	SMP_fprintf(HeaderFile, ", ");
	if (nullptr != UpperExpr) {
		UpperExpr->EmitSPARKAda(HeaderFile, ProcessingLoop, OldSuffix, false, HasArgs, false, false);
	}
	else { // Don't know upper bound.
		SMP_fprintf(HeaderFile, "16#ffffffffffffffff");
	}
	if ((0 < MemWidth) && (nullptr != UpperExpr)) {
		SMP_fprintf(HeaderFile, " + %zu) ", MemWidth);
	}
	else {
		SMP_fprintf(HeaderFile, ") ");
	}
	++OutputCount;
	return;
} // end of SMPFunction::EmitSPARKMemRange()

// helper for EmitSPARKProcPrePostMemConditions()
void SMPFunction::EmitSPARKPrePostMemConditionsFromCalleeNonLoopWrites(FILE *HeaderFile, bool PreconditionSection, size_t LoopIndexPlusOne, size_t &OutputCount) {
	for (size_t i = 0; i < this->MemAddrExprWidthsFromCallees[LoopIndexPlusOne].size(); ++i) {
		if (this->NonStackFrameCalleeMemWrites[LoopIndexPlusOne][i]) { // not just local stack frame write
			// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
			//  which would require printing the same operand twice.
			if (0 < OutputCount) {
				if (PreconditionSection)
					SMP_fprintf(HeaderFile, " and ");
				else
					SMP_fprintf(HeaderFile, "\n\t\t and not X86.InRange64(i, ");
			}
			else {
				assert(PreconditionSection);
			}
			size_t MemWidth = this->MemAddrExprWidthsFromCallees[LoopIndexPlusOne][i].second;
			if (PreconditionSection) {
				(*this->MemAddrExprWidthsFromCallees[LoopIndexPlusOne][i].first)->EmitSPARKInSafeRegion64(HeaderFile, MemWidth);
			}
			else {
				(*this->MemAddrExprWidthsFromCallees[LoopIndexPlusOne][i].first)->EmitSPARKAda(HeaderFile, false, !PreconditionSection, false, false, false);
				SMP_fprintf(HeaderFile, ", %u)", MemWidth);
			}
			++OutputCount;
		}
	} // end for all entries in MemAddrExprWidthFromCallees[LoopIndexPlusOne]
	return;
} // end of SMPFunction::EmitSPARKPrePostMemConditionsFromCalleeNonLoopWrites()

// Determine if any mem writes (or callee mem writes) are not just to the stack frame locals.
bool SMPFunction::NeedsSPARKPreconditions(bool &HasLoopMemWrites, bool &HasNonLoopInArgMemWrites, bool &HasCalleeMemWrites, bool &HasCalleeLoopMemWrites) {
	HasLoopMemWrites = ((this->LoopCount > 0) && this->MemRangeRegsBitmap.any());
	HasNonLoopInArgMemWrites = (!this->InArgsUsedInMemWrites[0].empty());
	HasCalleeMemWrites = (!this->MemAddrExprsFromCallees[0].empty());
	HasCalleeLoopMemWrites = this->HasMemExprsFromCalleeLoops;

	bool HasLoopGlobalStaticWrites = false;
#if 0  // now that we process multiple writes per loop, static write does not exclude stack writes
	if (!HasLoopMemWrites && (0 < this->LoopCount)) {
		// See if we have loop mem writes to global static memory.
		for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			if (this->LoopWritesGlobalStaticMemory[LoopIndex]) {
				HasLoopGlobalStaticWrites = true;
				HasLoopMemWrites = true;
				this->NonStackFrameLoopMemWrites[LoopIndex] = true;
				break;
			}
		}
	}
#endif

	if (HasLoopMemWrites && (!HasLoopGlobalStaticWrites)) {
		// Apply more stringent criterion: Loop has to write outside of our stack frame.
		HasLoopMemWrites = false;
		for (size_t LoopIndex = 0; LoopIndex < this->LoopCount; ++LoopIndex) {
			STARS_sval_t FinalStackDelta = 0;
			STARSExprBoundsIter ExprIter;
			for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
				STARSExpression *LowerExpr = ExprIter->first;
				STARSExpression *UpperExpr = ExprIter->second;
				if (!IsLocalStackFrameExprPair(LowerExpr, UpperExpr)) {
					this->NonStackFrameLoopMemWrites[LoopIndex] = true;
					HasLoopMemWrites = true;
					break; // needs preconditions already
				}
			} // end for all loop bounds exprs
		}
	}

	if (!HasCalleeMemWrites) { // no callee mem writes outside of this function's loop
		// Check the loop-index entries for true votes.
		for (size_t LoopIndexOffsetByOne = 1; LoopIndexOffsetByOne <= this->LoopCount; ++LoopIndexOffsetByOne) {
			if (!this->MemAddrExprsFromCallees[LoopIndexOffsetByOne].empty()) {
				HasCalleeMemWrites = true;
				break;
			}
		}
	}

	if (HasCalleeMemWrites) {
		// Apply the more stringent criterion.
		HasCalleeMemWrites = false;
		for (size_t LoopIndexOffsetByOne = 0; LoopIndexOffsetByOne <= this->LoopCount; ++LoopIndexOffsetByOne) {
			for (size_t index = 0; index < this->MemAddrExprWidthsFromCallees[LoopIndexOffsetByOne].size(); ++index) {
				STARS_sval_t FinalStackDelta = 0;
				if ((*(this->MemAddrExprWidthsFromCallees[LoopIndexOffsetByOne][index].first))->IsStackPtrOffset((STARS_sval_t) this->AllocPointDelta, FinalStackDelta)) {
					if (FinalStackDelta < 0) { // stack frame local var
						this->NonStackFrameCalleeMemWrites[LoopIndexOffsetByOne].push_back(false);
					}
					else {
						this->NonStackFrameCalleeMemWrites[LoopIndexOffsetByOne].push_back(true);
						HasCalleeMemWrites = true;
					}
				}
				else {
					this->NonStackFrameCalleeMemWrites[LoopIndexOffsetByOne].push_back(true);
					HasCalleeMemWrites = true;
				}
			}
		}
	}

	return (HasLoopMemWrites || HasNonLoopInArgMemWrites || HasCalleeMemWrites || HasCalleeLoopMemWrites);
} // end of SMPFunction::NeedsSPARKPreconditions()

string SMPFunction::GetFuncSPARKSuffixString(void) const {
	char HexAddr[12];
	SMP_snprintf(HexAddr, 11, "_%x", this->GetFirstFuncAddr());
	string FuncAddrString(HexAddr);  // e.g. "_400586"
	return FuncAddrString;
} // end of SMPFunction::GetFuncSPARKSuffixString()

// Generate loop exit flow-tracking boolean name, e.g. STARSExitLoop2ToBlock5
string SMPFunction::GenerateSPARKLoopExitBooleanName(const size_t LoopNum, const int ExitTargetTCFGBlockNum) const {
	assert(0 < ExitTargetTCFGBlockNum);
	assert(LoopNum < this->GetNumLoops());
	char LoopChars[8], TargetChars[8];
	string BoolName("STARSExitLoop");
	SMP_snprintf(LoopChars, 7, "%zu", LoopNum);
	SMP_snprintf(TargetChars, 7, "%d", ExitTargetTCFGBlockNum);
	BoolName += string(LoopChars);
	BoolName += "ToBlock";
	BoolName += string(TargetChars);

	return BoolName;
} // end of SMPFunction::GenerateSPARKLoopExitBooleanName()


// Save incoming args as locals to preserve their values
void SMPFunction::EmitSPARKSavedArgs(FILE *BodyFile) const {
	if ((0 < this->LoopCount) || (!this->InArgsUsedInMemWrites[0].empty())) {
		string FuncAddrString = this->GetFuncSPARKSuffixString();  // e.g. "_400586"
		std::bitset<1 + MD_LAST_REG_NO> EmittedRegs;
		if (0 < this->LoopCount) {
			EmittedRegs = this->MemRangeRegsBitmap;
			for (size_t i = 0; i < this->MemRangeRegsBitmap.size(); ++i) {
				if (this->MemRangeRegsBitmap[i]) {
					string CurrRegName(MDGetSPARKRegNumName((STARS_regnum_t)i, global_STARS_program->GetSTARS_ISA_Bytewidth()));
					string OutputName(CurrRegName + FuncAddrString);
#if 0
					// Need to only append "with Ghost" if the saved InArg will NOT be passed as a loop arg.
					SMP_fprintf(BodyFile, "\t%s : Unsigned64 := X86.%s with Ghost;\n", OutputName.c_str(), CurrRegName.c_str());
#else
					SMP_fprintf(BodyFile, "\t%s : Unsigned64 := X86.%s;\n", OutputName.c_str(), CurrRegName.c_str());
#endif
				}
			}
		}

		// Could also use this->LoopMemRangeInArgRegsBitMap for this next loop.
		for (size_t LoopPlusOne = 0; LoopPlusOne <= this->GetNumLoops(); ++LoopPlusOne) {
			size_t InArgLimit = this->GetNumInArgsUsedInMemWrites(LoopPlusOne);
			for (size_t ArgExprIndex = 0; ArgExprIndex < InArgLimit; ++ArgExprIndex) {
				uint32_t InArgExprSerialNumber = this->GetInArgExprUsedInMemWrite(LoopPlusOne, ArgExprIndex);
				assert(! InArgsUsedInMemWriteByteWidths[LoopPlusOne].empty());
				STARSExprSetIter InArgExprIter = this->FindInArgExprBySerialNumber(LoopPlusOne, InArgExprSerialNumber);
				assert(InArgExprIter != this->GetLastInArgExprSetIter(LoopPlusOne));
				STARS_regnum_t RegNum = (*InArgExprIter)->GetConstLeftOperand()->GetReg();
				// Don't emit an InArg save if we already did so for the InArgs used in loops above.
				if (!EmittedRegs[(size_t) RegNum]) {
					string CurrRegName(MDGetSPARKRegNumName(RegNum, global_STARS_program->GetSTARS_ISA_Bytewidth()));
					string OutputName(CurrRegName + FuncAddrString);
					SMP_fprintf(BodyFile, "\t%s : Unsigned64 := X86.%s with Ghost;\n", OutputName.c_str(), CurrRegName.c_str());
					EmittedRegs.set((size_t) RegNum);
				}
			}
		}
	}

	return;
} // end of SMPFunction::EmitSPARKSavedArgs()

// Emit loop function args, as Signature to HeaderFile or proc call to BodyFile.
void SMPFunction::EmitSPARKArgs(FILE *BodyFile, FILE *HeaderFile, bool Signature, size_t LoopIndex) const {
	if (Signature)
		assert(nullptr != HeaderFile);
	else
		assert(nullptr != BodyFile);

	char HexAddr[12];
	SMP_snprintf(HexAddr, 11, "_%x", this->GetFirstFuncAddr());
	string FuncAddrString(HexAddr);  // e.g. "_400586"
	size_t OutputCount = 0;

	for (size_t i = 0; i < this->LoopMemRangeInArgRegsBitmap[LoopIndex].size(); ++i) {
		if (this->LoopMemRangeInArgRegsBitmap[LoopIndex][i]) {
			string CurrRegName(MDGetSPARKRegNumName((STARS_regnum_t) i, global_STARS_program->GetSTARS_ISA_Bytewidth()));
			string OutputName(CurrRegName + FuncAddrString);
			if (OutputCount == 0) {
				if (Signature) {
					SMP_fprintf(HeaderFile, "%s : in Unsigned64", OutputName.c_str());
				}
				else {
					SMP_fprintf(BodyFile, "%s", OutputName.c_str());
				}
			}
			else {
				if (Signature) {
					SMP_fprintf(HeaderFile, "; %s : in Unsigned64", OutputName.c_str());
				}
				else {
					SMP_fprintf(BodyFile, ", %s", OutputName.c_str());
				}
			}
			++OutputCount;
		}
	}

	return;
} // end of SMPFunction::EmitSPARKArgs()

// Create SPARK procedure for loop starting at HeaderBlockNum
string SMPFunction::EmitSPARKProcForLoopHeaderBlock(int LoopIndex, int HeaderTCFGBlockNum, int FollowTCFGBlockNum, FILE *BodyFile, FILE *HeaderFile) {
	assert(0 <= LoopIndex);
	size_t LoopNum = (size_t) LoopIndex;
	assert(0 <= HeaderTCFGBlockNum);
	size_t LoopNumPlusOne = LoopNum + 1;
	STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum(HeaderTCFGBlockNum);
	bool HasArgs = this->DoesLoopHaveArgs(LoopIndex);
	bool MemoryInput = false;
	bool MemoryOutput = false;
	bool UseFP = this->UsesFramePointer();
	STARS_ea_t LoopAddr = CurrTCFGBlock->GetFirstNonMarkerAddr();
	bitset<1 + MD_LAST_REG_NO> LoopInputRegs, LoopOutputRegs, CalleePreservedRegs;
	int FirstTCFGBlockNum = HeaderTCFGBlockNum;

	// See if we are translating a guarded loop.
	STARS_ea_t LastAddr = CurrTCFGBlock->GetLastAddr();
	vector<SMPInstr *>::iterator InstIter = CurrTCFGBlock->GetFirstInst();
	SMPInstr *LastInst = this->GetInstFromAddr(LastAddr);
	if ((*InstIter)->IsMarkerInst())
		++InstIter;
	bool TranslatingGuardedLoop = ((*InstIter)->HasBeenTranslatedToSPARK() && (!LastInst->HasBeenTranslatedToSPARK())
		&& (BRANCH_IF_THEN == this->GetControlFlowType(LastAddr)));
	if (TranslatingGuardedLoop) {
		map<STARS_ea_t, STARS_ea_t>::const_iterator GuardLoopIter = this->GuardToLoopMap.find(LastAddr);
		assert(GuardLoopIter != this->GuardToLoopMap.cend());
		LoopAddr = GuardLoopIter->second;
		LoopInputRegs.set(MD_FLAGS_REG);
		// Skip on to next block.
		FirstTCFGBlockNum = (*(CurrTCFGBlock->GetFallThroughSucc()))->GetNumber();
	}

	// Generate procedure name for translated loop.
	std::string ProcName("ZSTLoopProc_");
	std::ostringstream AddrString;
	AddrString << std::hex << LoopAddr;
	ProcName += AddrString.str();

	bool HasNonStackPreconditions = this->LoopRequiresNonStackPreconditions(LoopNum);
	// Use sets of exprs to avoid duplication and streamline the output section.
	this->AggregateLoopMemExprs(LoopNum);
	bool HasRangeExprs = (!this->TempRangeExprWidthIters.empty());
	bool HasNonRangeExprs = (!this->TempNonRangeExprWidthIters.empty());
	bool HasRelationalExprs = (!this->RelationalMemWriteWidths[LoopNumPlusOne].empty());
	bool HasMemExprs = (HasRangeExprs || HasNonRangeExprs || HasRelationalExprs);

	bool StackWritten = (this->NegativeOffsetStackBytesWrittenByLoop[LoopNum].IsAnyBitSet() || this->PositiveOffsetStackBytesWrittenByLoop[LoopNum].IsAnyBitSet());
	bool IndirectStackWrites = this->LoopMemExprsExpandToStackOffsets[LoopNum];
	bool UsesSaveStackPtr = (StackWritten || IndirectStackWrites || HasArgs || HasMemExprs);

	this->AnalyzeLoopGlobals(FirstTCFGBlockNum, FollowTCFGBlockNum, LoopAddr, MemoryInput, MemoryOutput, LoopInputRegs, LoopOutputRegs, CalleePreservedRegs);
	MemoryOutput |= HasMemExprs;
	this->LoopWritesMemory[LoopNum] = MemoryOutput;
	this->OutputRegsByLoop[LoopNum] = LoopOutputRegs;
	this->CalleePreservedRegsByLoop[LoopNum] = CalleePreservedRegs;
	bool UsesStackPtrRegs = (LoopInputRegs[MD_STACK_POINTER_REG] || LoopOutputRegs[MD_STACK_POINTER_REG]
		|| (UseFP && (LoopInputRegs[MD_FRAME_POINTER_REG] || LoopOutputRegs[MD_FRAME_POINTER_REG])));
	if (!UsesStackPtrRegs && UsesSaveStackPtr) {
		// Stack pointer reg is only used in Ghost variable.
		LoopInputRegs.set(MD_STACK_POINTER_REG);
	}
	this->LoopUsesStackPtrRegs[LoopNum] = UsesStackPtrRegs;
	bool CalleeWrites = this->LoopHasCalleeMemWrites[LoopIndex];

	// Emit beginning of function body.
	if (HasArgs) {
		SMP_fprintf(BodyFile, "procedure %s(", ProcName.c_str());
		// Make EmitSPARKArgs write the signature to the BodyFile, not HeaderFile.
		this->EmitSPARKArgs(nullptr, BodyFile, true, LoopNum);
		SMP_fprintf(BodyFile, ")\nis\n");
	}
	else {
		SMP_fprintf(BodyFile, "procedure %s\nis\n", ProcName.c_str());
	}
	// Keep track of the RSP and saved RBP from the original context of the loop.
	vector<SMPInstr *>::iterator FirstInstIter = CurrTCFGBlock->GetFirstInst();
	SMPInstr *FirstInst = (*FirstInstIter);
	STARS_sval_t IncomingStackDelta = FirstInst->GetStackPtrOffset();
	if (UsesStackPtrRegs || UsesSaveStackPtr) {
		if (0 != IncomingStackDelta)
			SMP_fprintf(BodyFile, "\tSaveStackPtr : Unsigned64 := X86.RSP + 16#%x# with Ghost;\n", (0 - IncomingStackDelta));
		else
			SMP_fprintf(BodyFile, "\tSaveStackPtr : Unsigned64 := X86.RSP with Ghost;\n");
		if (UseFP) {
			SMP_fprintf(BodyFile, "\tsavedRBP : Unsigned64 := X86.ReadMem64(X86.RBP) with Ghost;");
		}
	}
	SMP_fprintf(BodyFile, "\nbegin\n");
	if (UsesStackPtrRegs) {
		SMP_fprintf(BodyFile, "\tpragma Assume(X86.RSP = X86.DummyRSP);\n");
	}
	if ((this->LoopIterationsLimitExprs[LoopNum] != nullptr)
		&& (this->LoopAnalyzedBIVIters[LoopNum] != this->LoopInductionVars[LoopNum].end())) {
		// We analyzed the loop iterations count
		this->EmitSPARKLoopBIVLimits(BodyFile, LoopAddr, LoopNum, true); // pragma Assume(no wrap-around on loop limits)
	}

	// Emit procedure specification.
	this->EmitAnalysisProblemWarnings(HeaderFile, LoopIndex);
	if (HasArgs) {
		SMP_fprintf(HeaderFile, "procedure %s(", ProcName.c_str());
		this->EmitSPARKArgs(nullptr, HeaderFile, true, LoopNum);
		SMP_fprintf(HeaderFile, ") with\n");
	}
	else {
		SMP_fprintf(HeaderFile, "procedure %s with\n", ProcName.c_str());
	}
	this->EmitSPARKLoopProcGlobals(BodyFile, HeaderFile, MemoryInput, MemoryOutput, LoopInputRegs, LoopOutputRegs, CalleePreservedRegs);

	// Emit pre-conditions.
	bool HasBIVInitialValue = (this->LoopAnalyzedBIVIters[LoopNum] != this->LoopInductionVars[LoopNum].end()); // successful loop analysis
	bool HasPreconditions = (HasNonStackPreconditions || (UseFP && UsesStackPtrRegs) || HasBIVInitialValue);
	this->LoopHasPreconditions[LoopNum] = HasPreconditions;
	
	if (HasPreconditions) {
		size_t OutputCount = 0;
		SMP_fprintf(HeaderFile, ",\n\tPre => ");
		if (UseFP && UsesStackPtrRegs) {
			int RBPRSPDelta = (int)(this->FramePointerStackDelta - IncomingStackDelta);
			if (RBPRSPDelta == 0) {
				SMP_fprintf(HeaderFile, "(X86.RBP = X86.RSP)");
			}
			else {
				SMP_fprintf(HeaderFile, "(X86.RBP = X86.RSP + %d)", RBPRSPDelta);
			}
			if (HasNonStackPreconditions || HasBIVInitialValue) {
				SMP_fprintf(HeaderFile, " and \n\t"); // prepare for more preconditions
			}
			++OutputCount;
		}

		bitset<1 + MD_LAST_REG_NO> RegsPrinted; // record regs printed to avoid redundant output
		if (HasBIVInitialValue) { // Need precondition for basic induction var incoming value
			// NOTE: We could have more than one BIV. Need to expand.
			STARSInductionVarFamilyList::const_iterator BIVIter = this->LoopAnalyzedBIVIters[LoopNum];
			assert(BIVIter != this->LoopInductionVars[LoopNum].end());
			STARS_ea_t BIVInitAddr = BIVIter->BIVIncomingDefAddr;
			assert(STARS_BADADDR != BIVInitAddr);
			STARSOpndTypePtr BIVDefOp = BIVIter->BasicInductionVar.InductionVar.GetOp();
			SMP_fprintf(HeaderFile, " (");
			FirstInst->PrintSPARKAdaOperand(BIVDefOp, HeaderFile, false, UseFP, true, false, false);
			SMP_fprintf(HeaderFile, "= ");
			this->LoopIterationsInitExprs[LoopNum]->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
			++OutputCount;
			if (this->LoopIterationsInitExprs[LoopNum]->IsStackPtrRegUsed()) {
				// Offset the stack pointer from its entry point value in the func that includes this loop.
				SMP_fprintf(HeaderFile, " + 16#%x# ", (0 - IncomingStackDelta));
			}
			SMP_fprintf(HeaderFile, ")");
			if (BIVDefOp->IsRegOp()) {
				RegsPrinted.set((size_t) BIVDefOp->GetReg());
			}

			// See if we have a non-constant limit for the BIV. If so, print its value, 
			//  which will be found in the right hand side of the LimitExpr.
			assert(LoopIndex < (int)this->LoopComparisonExprs.size());
			STARSOpndTypePtr LimitOp = this->LoopComparisonExprs[LoopNum].Operand2.GetOp();
			if (!LimitOp->IsImmedOp()) {
				assert(this->LoopIterationsLimitExprs[LoopNum]->IsRelationalExpr());
				SMP_fprintf(HeaderFile, " and \n\t"); // prepare for LimitOp precondition
				SMP_fprintf(HeaderFile, " (");
				FirstInst->PrintSPARKAdaOperand(LimitOp, HeaderFile, false, UseFP, true, false, false);
				SMP_fprintf(HeaderFile, "= ");
				bool SPRelative = false;
				if (this->LoopIterationsLimitExprs[LoopNum]->HasRightSubTree()) {
					this->LoopIterationsLimitExprs[LoopNum]->GetRightTree()->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
					SPRelative = this->LoopIterationsLimitExprs[LoopNum]->GetRightTree()->IsStackPtrRegUsed();
				}
				else { // RightOperand
					const STARSOpndTypePtr LimitInitOp = this->LoopIterationsLimitExprs[LoopNum]->GetConstRightOperand();
					FirstInst->PrintSPARKAdaOperand(LimitInitOp, HeaderFile, false, UseFP, true, false, false);
					SPRelative = (LimitInitOp->IsRegOp() && LimitInitOp->MatchesReg(MD_STACK_POINTER_REG));
				}
				if (SPRelative) {
					// Offset the stack pointer from its entry point value in the func that includes this loop.
					SMP_fprintf(HeaderFile, " + 16#%x# ", (0 - IncomingStackDelta));
				}
				SMP_fprintf(HeaderFile, ")");
				if (BIVDefOp->IsRegOp()) {
					RegsPrinted.set((size_t) BIVDefOp->GetReg());
				}
			} // end if (!LimitOp->IsImmedOp())
		} // end if (HasBIVInitialValue)

		if (HasNonStackPreconditions) {
			if (HasArgs && HasMemExprs) { // Need preconditions for memory writing range
				// First, print the incoming values of memory address registers that were
				//  expanded to produce a PreLoopDefAddr in the expr.
				size_t VectorLimit = this->TempNonRangeExprWidthIters.size();
				for (size_t VecIndex = 0; VecIndex < VectorLimit; ++VecIndex) {
					STARSExpression *MemExpr = (*(this->TempNonRangeExprWidthIters[VecIndex].second));
					MemExpr->PrintSPARKArgLocationStrings(HeaderFile, false, LoopNum, OutputCount, RegsPrinted);
				}
				VectorLimit = this->TempRangeExprWidthIters.size();
				for (size_t VecIndex = 0; VecIndex < VectorLimit; ++VecIndex) {
					// Print lower bound mem expr.
					STARSExpression *MemExpr = (*(this->TempRangeExprWidthIters[VecIndex].first));
					if (nullptr == MemExpr) {
						SMP_msg("ERROR: SPARK: nullptr for lower bound mem expr in EmitSPARKProcForLoopHeaderBlock() in loop %d VecIndex %zu ",
							LoopIndex, VecIndex);
						this->DumpFuncNameAndAddr();
						continue;
					}
					MemExpr->PrintSPARKArgLocationStrings(HeaderFile, false, LoopNum, OutputCount, RegsPrinted);
					// Print upper bound mem expr; if nullptr, print maximum limit.
					MemExpr = (*(this->TempRangeExprWidthIters[VecIndex].second));
					if (nullptr != MemExpr) {
						MemExpr->PrintSPARKArgLocationStrings(HeaderFile, false, LoopNum, OutputCount, RegsPrinted);
					}
					else { // Don't know upper bound.
						SMP_fprintf(HeaderFile, "16#ffffffffffffffff");
						++OutputCount;
					}
				}
			}
			// Print the incoming values of regs that got copied to address registers for mem writes.
			if (HasMemExprs && (!this->LoopRegSourceExprPairs[LoopNum].empty())) {
				this->EmitIncomingLoopRegExprs(HeaderFile, LoopNum, false);
				++OutputCount;
			}

			// Now, print the ranges of memory written (or inherited) in this loop.
			bool PreRangeOutput = (0 < OutputCount);
			bool RangesStarted = false;

			if (HasRangeExprs) {
				if (OutputCount == 0) {
					SMP_fprintf(HeaderFile, "\n\t\t(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
				}
				else {
					SMP_fprintf(HeaderFile, " and \n\t\t(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
				}
				RangesStarted = true;
				OutputCount = 0; // reset before first call to EmitSPARKMemRange()
				size_t VectorLimit = this->TempRangeExprWidthIters.size();
				for (size_t VecIndex = 0; VecIndex < VectorLimit; ++VecIndex) {
					STARSExpression *LowerExpr = (*this->TempRangeExprWidthIters[VecIndex].first);
					STARSExpression *UpperExpr = (*this->TempRangeExprWidthIters[VecIndex].second);
					if (nullptr == LowerExpr) {
						SMP_msg("ERROR: SPARK: LowerExpr nullptr, VecIndex %zu ", VecIndex);
						this->DumpFuncNameAndAddr();
					}
					else {
						this->EmitSPARKMemRange(HeaderFile, true, true, HasArgs, LowerExpr, UpperExpr, OutputCount, 0);
					}
				}
			} // end if (HasRangeExprs)

			if (HasRelationalExprs) {
				if (!RangesStarted) {
					if (PreRangeOutput) // something printed after "Pre =>" and before mem ranges
						SMP_fprintf(HeaderFile, " and \n\t\t(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
					else
						SMP_fprintf(HeaderFile, "\n\t\t(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
					RangesStarted = true;
				}

				OutputCount = 0; // reset before first call to EmitSPARKMemRange()
				std::list<std::pair<std::size_t, std::pair<STARSExprSetIter, STARSExprSetIter> > >::const_iterator RelationalIter;
				for (RelationalIter = this->RelationalMemWriteWidths[LoopNumPlusOne].cbegin(); RelationalIter != this->RelationalMemWriteWidths[LoopNumPlusOne].cend(); ++RelationalIter) {
					size_t MemWidth = (*RelationalIter).first;
					STARSExprSetIter LowerExprIter = (*RelationalIter).second.first;
					STARSExprSetIter UpperExprIter = (*RelationalIter).second.second;
					this->EmitSPARKMemRange(HeaderFile, true, true, HasArgs, (*LowerExprIter), (*UpperExprIter), OutputCount, MemWidth);
				}
			} // end if (HasRelationalExprs)

			if (HasNonRangeExprs) {
				size_t VectorLimit = this->TempNonRangeExprWidthIters.size();
				for (size_t VectorIndex = 0; VectorIndex < VectorLimit; ++VectorIndex) {
					// Easier in this loop to use X86.InRange64(i, X86.RAX, 8) than to use X86.InMemoryRange(i, X86.RAX, X86.RAX+8)
					//  which would require printing the same operand twice.
					size_t ByteWidth = this->TempNonRangeExprWidthIters[VectorIndex].first;
					STARSExpression *NonRangeExpr = (*this->TempNonRangeExprWidthIters[VectorIndex].second);

					// The StackPtrCase below is now handled via the {Posi,Nega}tiveStackOffsetWritesByLoop[] vectors
					if (!NonRangeExpr->IsStackPtrPlusOffset()) { // screen out stack frame writes
						if (!RangesStarted) {
							if (PreRangeOutput) // something printed after "Pre =>" and before mem ranges
								SMP_fprintf(HeaderFile, " and \n\t\t(for all i in Unsigned64 => (if X86.InRange64(i, ");
							else
								SMP_fprintf(HeaderFile, "\n\t\t(for all i in Unsigned64 => (if X86.InRange64(i, ");
							RangesStarted = true;
						}
						else {
							SMP_fprintf(HeaderFile, "\n\t\tor X86.InRange64(i, ");
						}
						// Some MemAddrExprs were based on the incoming stack pointer value for the function enclosing
						//  this loop. We need to make two adjustments when printing these exprs:
						//   1. Don't treat the incoming stack ptr reg as an InArg (i.e. don't make it look like an Ada argument).
						//   2. Offset the stack address with the negated incoming stack offset.
						bool StackPtrCase = (!NonRangeExpr->HasLeftSubTree()) && NonRangeExpr->GetConstLeftOperand()->MatchesReg(STARS_x86_R_sp);
						STARS_ea_t PreLoopDefAddr = NonRangeExpr->GetLeftPreLoopDefAddr();
						bool PseudoAddr = STARS_IsSSAMarkerPseudoID(PreLoopDefAddr) || STARS_IsBlockNumPseudoID(PreLoopDefAddr) || STARS_IsLiveInPseudoID(PreLoopDefAddr);
						if (!StackPtrCase && (!PseudoAddr)) {
							// We don't want to trace callee mem exprs all the way back to this function's InArgs.
							//  We only want to trace back to the value that comes into the loop. The loop boundary
							//  crossing is recorded in the expression.
							STARSExpression *LoopBoundaryExpr = NonRangeExpr->Clone();
							LoopBoundaryExpr->SetLeftOperand(LoopBoundaryExpr->FindLeftPreLoopDefOp());
							LoopBoundaryExpr->SetLeftUseAddr(LoopBoundaryExpr->GetLeftPreLoopDefAddr());
							LoopBoundaryExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs, false);
						}
						else {
							NonRangeExpr->EmitSPARKAda(HeaderFile, !StackPtrCase, false, false, HasArgs, false);
							if ((0 != IncomingStackDelta) && StackPtrCase) {
								// Traced an outarg to a callee back to a stack pointer address. Offset by IncomingStackDelta.
								SMP_fprintf(HeaderFile, " + 16#%x#", IncomingStackDelta);
							}
						}
						SMP_fprintf(HeaderFile, ", %u) ", ByteWidth);
						++OutputCount;
					} // end if (!NonRangeExpr->IsStackPtrPlusOffset())
				} // end for VectorIndex over all NonRangeExprWidthIters
			} // end if (HasNonRangeExprs)

#if 0 // old code when we only emitted the loop mem range; part of TempRangeExprs now
			SMP_fprintf(HeaderFile, " and \n\t(for all i in Unsigned64 => (if X86.InMemoryRange(i, ");
			this->LoopMemWriteLowerBoundsExprs[LoopIndex]->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
			SMP_fprintf(HeaderFile, ", ");
			this->LoopMemWriteUpperBoundsExprs[LoopIndex]->EmitSPARKAda(HeaderFile, true, false, false, HasArgs, false);
			if (0 != IncomingStackDelta) {
				SMP_fprintf(HeaderFile, "then\n\t\tX86.InSafeRegion64(i, X86.RSP+%d)))", (0 - IncomingStackDelta));
			}
			else {
				SMP_fprintf(HeaderFile, "then\n\t\tX86.InSafeRegion64(i, X86.RSP)))");
			}
#endif			
			if (CalleeWrites) {
				SMP_fprintf(HeaderFile, "then\n\t\tX86.InSafeRegion64(i, X86.RSP-8)))");
			}
			else if (HasMemExprs) {
				SMP_fprintf(HeaderFile, "then\n\t\tX86.InSafeRegion64(i, X86.RSP)))");
			}
		}
		SMP_fprintf(HeaderFile, ",\n");  // terminate pre-condition section

	} // end if (HasPreconditions)

	// NOTE: The procedure post-conditions will be emitted in SMPInstr::EmitSPARKAdaLoopInvariants()

	return ProcName;
} // end of SMPFunction::EmitSPARKProcForLoopHeaderBlock()

bool SMPFunction::LoopRequiresNonStackPreconditions(size_t LoopIndex) {
	bool LoopWritePreconditions = ((nullptr != this->LoopIterationsInitExprs[LoopIndex])
		&& (!this->LoopMemWriteBoundsExprs[LoopIndex].empty()));
	if (LoopWritePreconditions) {
		STARSExprBoundsIter ExprIter;
		bool TraceableToSimpleStackOffset = this->LoopIterationsInitExprs[LoopIndex]->IsStackPtrPlusOffset();
		for (ExprIter = this->GetFirstLoopMemWriteExprBoundsIter(LoopIndex); TraceableToSimpleStackOffset && ExprIter != this->GetLastLoopMemWriteExprBoundsIter(LoopIndex); ++ExprIter) {
			STARSExpression *LowerExpr = ExprIter->first;
			STARSExpression *UpperExpr = ExprIter->second;
			if ((nullptr != LowerExpr) && (nullptr != UpperExpr)) {
				TraceableToSimpleStackOffset = TraceableToSimpleStackOffset || LowerExpr->IsStackPtrPlusOffset()
					|| UpperExpr->IsStackPtrPlusOffset();
			}
		}
		this->LoopMemExprsExpandToStackOffsets[LoopIndex] = TraceableToSimpleStackOffset;
		LoopWritePreconditions = (!TraceableToSimpleStackOffset);
	}
#if 1
	// If we don't already need preconditions based on loop mem writes, examine
	//  looping string mem writes, inherited from callees mem writes.
	if (!LoopWritePreconditions) {
		LoopWritePreconditions = ((!this->MemAddrExprWidthsFromCallees[LoopIndex + 1].empty())
			&& (!this->LoopMemAddrExprWidthsFromCalleeLoops[LoopIndex + 1].empty())
			&& (!this->RelationalMemWriteWidths[LoopIndex + 1].empty()));
	}
#endif
	return LoopWritePreconditions;
} // end of SMPFunction::LoopRequiresNonStackPreconditions()

// Analyze reg and mem accesses in loop
void SMPFunction::AnalyzeLoopGlobals(int HeaderTCFGBlockNum, int FollowTCFGBlockNum, STARS_ea_t LoopAddr, bool &MemoryInput, bool &MemoryOutput, std::bitset<1 + MD_LAST_REG_NO> &InputRegs, std::bitset<1 + MD_LAST_REG_NO> &OutputRegs, std::bitset<1 + MD_LAST_REG_NO> &CalleePreservedRegs) {
	assert((0 <= HeaderTCFGBlockNum) && (HeaderTCFGBlockNum < (int)this->TCFGBlocks.size()));
	MemoryInput = false;
	MemoryOutput = false;

	// We need something like the UpExposed set for HeaderTCFGBlockNum and the union of the VarKill sets
	//  of all blocks in the procedure or loop. But the extra USEs and DEFs that are added to CALL 
	//  and RETURN instructions are not desired for SPARK Ada analyses, so we must go through all the
	//  instructions and exclude CALL and RETURN types and build up the input and output data flow sets.

	STARSTransCFGBlock *LoopHeaderTCFGBlock = this->GetTCFGBlockFromFirstInstAddr(LoopAddr);
	int TempLoopNum = this->GetLoopNumFromHeaderBlockNum(LoopHeaderTCFGBlock->GetOriginalBlockNum());
	assert(0 <= TempLoopNum);
	size_t LoopNum = (size_t) TempLoopNum;
	list<size_t> LoopBlockList;
	this->BuildLoopBlockList(LoopNum, LoopBlockList); // operate off original CFG blocks for this analysis

	for (list<size_t>::const_iterator BlockIter = LoopBlockList.cbegin(); BlockIter != LoopBlockList.cend(); ++BlockIter) {
		size_t BlockNum = *BlockIter;
		assert(BlockNum < this->RPOBlocks.size());
		SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockNum];
		for (vector<SMPInstr *>::iterator InstIter = CurrBlock->GetFirstInst(); InstIter != CurrBlock->GetLastInst(); ++InstIter) {
			SMPInstr *CurrInst = (*InstIter);
			STARS_ea_t InstAddr = CurrInst->GetAddr();
			SMPitype FlowType = CurrInst->GetDataFlowType();
			if ((DEFAULT != FlowType) && (JUMP != FlowType) && (INDIR_JUMP != FlowType)) {
				if ((CALL == FlowType) || (INDIR_CALL == FlowType) || (RETURN == FlowType)) {
					// Calls change the stack pointer.
					InputRegs.set((size_t) MD_STACK_POINTER_REG);
					OutputRegs.set((size_t) MD_STACK_POINTER_REG);
					// Inherit from callee. NOTE: RETURN could be tail call, INDIR_CALL could be resolved.
					STARS_ea_t CalleeAddr = CurrInst->GetCallTarget();
					if ((STARS_BADADDR != CalleeAddr) && (!STARS_IsExternalFunc(CalleeAddr))) {
						SMPFunction *CalleeFunc = this->GetProg()->FindFunction(CalleeAddr);
						if (nullptr != CalleeFunc) {
							InputRegs |= CalleeFunc->GetInputRegs();
							std::bitset<1 + MD_LAST_REG_NO> TempCalleeOutputRegs = CalleeFunc->GetOutputRegs();
							std::bitset<1 + MD_LAST_REG_NO> TempCalleePreservedRegs = CalleeFunc->GetPreservedRegs();
							std::bitset<1 + MD_LAST_REG_NO> TempCalleeChainPreservedRegs = CalleeFunc->GetCalleePreservedRegs();
							for (size_t RegNo = 0; RegNo < TempCalleeOutputRegs.size(); ++RegNo) {
								if (TempCalleeOutputRegs[RegNo]) {
									// If RegNo is Preserved, record in CalleePreservedRegs argument.
									if (TempCalleePreservedRegs[RegNo]) {
										CalleePreservedRegs.set(RegNo);
									}
									else {
										// Altered, not preserved
										OutputRegs.set(RegNo);
									}
								}
								else if (TempCalleePreservedRegs[RegNo]) { // preserved in callee, not explicitly altered
									CalleePreservedRegs.set(RegNo);
								}
								else if (TempCalleeChainPreservedRegs[RegNo]) { // preserved in callee chain, not explicitly altered
									CalleePreservedRegs.set(RegNo);
								}
							}
							MemoryOutput |= CalleeFunc->AltersSPARKMemory();
						}
					}
				}
				continue;  // exclude CALL, INDIR_CALL, RETURN with their conservative USE and DEF lists
			}
			for (STARSDefUseIter UseIter = CurrInst->GetFirstUse(); UseIter != CurrInst->GetLastUse(); ++UseIter) {
				STARSOpndTypePtr UseOp = UseIter->GetOp();
				if (UseOp->IsRegOp()) {
					STARS_regnum_t RegNo = UseOp->GetReg();
					if (!OutputRegs[(size_t) RegNo]) { // UpExposed, USE before DEF in RPO order traversal
						InputRegs.set((size_t) RegNo, true);
					}
				}
				else if (UseOp->IsMemOp()) {
					MemoryInput = true;
				}
			} // end for all USEs

			for (STARSDefUseIter DefIter = CurrInst->GetFirstDef(); DefIter != CurrInst->GetLastDef(); ++DefIter) {
				STARSOpndTypePtr DefOp = DefIter->GetOp();
				if (DefOp->IsRegOp()) {
					STARS_regnum_t RegNo = DefOp->GetReg();
					OutputRegs.set((size_t)RegNo, true);
				}
				else if (DefOp->IsMemOp()) {
					MemoryOutput = true;
				}
			} // end for all DEFs
		} // end for all insts in block
		if (!MemoryOutput && CurrBlock->HasMemoryWrite()) {
			SMP_msg("INFO: SPARK: Found new MemoryOutput using HasMemoryWrite\n");
			MemoryOutput = true;  // could be non-stack mem, not in DEFs
		}
	} // end for all blocks

	if (!OutputRegs[MD_STACK_POINTER_REG]) {
		// See if we have an implicit use of the stack pointer reg via loop mem write exprs that
		//  will expand to be stack pointer +/- offset.
		if (this->LoopMemExprsExpandToStackOffsets[LoopNum]) {
			InputRegs.set(MD_STACK_POINTER_REG);
			SMP_msg("INFO: SPARK: Loop %zu at %llx has SP reg use only implicitly.\n", LoopNum, (uint64_t) LoopAddr);
		}
	}
	LoopBlockList.clear();
	return;
} // end of SMPFunction::AnalyzeLoopGlobals()

// emit Input, Output, In_Out flow annotations
void SMPFunction::EmitSPARKLoopProcGlobals(FILE *BodyFile, FILE *HeaderFile, bool MemoryInput, bool MemoryOutput, const bitset<1 + MD_LAST_REG_NO> &InputRegs, const bitset<1 + MD_LAST_REG_NO> &OutputRegs, const bitset<1 + MD_LAST_REG_NO> &CalleePreservedRegs) {
	bool FlagsInput = false;
	bool FlagsOutput = false;
	bool UseFP = this->UsesFramePointer();

	FlagsInput = InputRegs[MD_FLAGS_REG];
	FlagsOutput = OutputRegs[MD_FLAGS_REG] || CalleePreservedRegs[MD_FLAGS_REG];

	// If a variable is Input but not Output, it gets the SPARK Ada keyword "Input"
	size_t ByteWidth = global_STARS_program->GetSTARS_ISA_Bytewidth();
	SMP_fprintf(HeaderFile, "\tGlobal => (");
	size_t TotalOutputCount = 0;
	size_t OutputCount = 0;
	bool FPStackBelowTopPrinted = false;

	for (size_t OutputIndex = 0; OutputIndex < OutputRegs.size(); ++OutputIndex) {
		if (OutputIndex == MD_FLAGS_REG)
			continue;
		bool FPRegBelowStackTop = ((OutputIndex >= STARS_x86_R_st1) && (OutputIndex <= STARS_x86_R_st7));
		if (!(FPRegBelowStackTop && FPStackBelowTopPrinted)) { // avoid duplicate X86.FloatingPointStackDummy1
			if (OutputRegs[OutputIndex] || CalleePreservedRegs[OutputIndex]) {
				if (0 == OutputCount) {
					if (0 < TotalOutputCount) {
						SMP_fprintf(HeaderFile, ",\n\t\t");  // line up with previous Global lines
					}
					SMP_fprintf(HeaderFile, "In_Out => (X86.%s", MDGetSPARKRegNumName((STARS_regnum_t)OutputIndex, (uint16_t)ByteWidth));
				}
				else {
					SMP_fprintf(HeaderFile, ", X86.%s", MDGetSPARKRegNumName((STARS_regnum_t)OutputIndex, (uint16_t)ByteWidth));
				}
				++OutputCount;
				FPStackBelowTopPrinted |= FPRegBelowStackTop;
			}
		}
	} // for all OutputRegs
	if (MemoryOutput) {
		if (0 == OutputCount) {
			SMP_fprintf(HeaderFile, "In_Out => (X86.Memory");
		}
		else {
			SMP_fprintf(HeaderFile, ", X86.Memory");
		}
		++OutputCount;
	}
	if (FlagsOutput) {
		if (0 == OutputCount) {
			if (0 < TotalOutputCount) {
				SMP_fprintf(HeaderFile, ",\n\t\t");  // line up with previous Global lines
			}
			SMP_fprintf(HeaderFile, "In_Out => (X86.CarryFlag, X86.ParityFlag, X86.SignFlag, X86.OverflowFlag, X86.ZeroFlag");
		}
		else {
			SMP_fprintf(HeaderFile, ", X86.CarryFlag, X86.ParityFlag, X86.SignFlag, X86.OverflowFlag, X86.ZeroFlag");
		}
		++OutputCount;
	}
	if (0 < OutputCount) {
		SMP_fprintf(HeaderFile, ")"); // close the In_Out open parenthesis
	}
	TotalOutputCount += OutputCount;

	OutputCount = 0;
	for (size_t InputIndex = 0; InputIndex < InputRegs.size(); ++InputIndex) {
		if (InputIndex == MD_FLAGS_REG)
			continue;
		if (InputRegs[InputIndex] && (!OutputRegs[InputIndex]) && (!CalleePreservedRegs[InputIndex])) {
			bool FPRegBelowStackTop = ((InputIndex >= STARS_x86_R_st1) && (InputIndex <= STARS_x86_R_st7));
			if (!(FPRegBelowStackTop && FPStackBelowTopPrinted)) { // avoid duplicate X86.FloatingPointStackDummy1
				if (0 == OutputCount) {
					if (0 < TotalOutputCount) {
						SMP_fprintf(HeaderFile, ",\n\t\t");  // line up with previous Global lines
					}
					SMP_fprintf(HeaderFile, "Input => (X86.%s", MDGetSPARKRegNumName((STARS_regnum_t)InputIndex, (uint16_t)ByteWidth));
				}
				else {
					SMP_fprintf(HeaderFile, ", X86.%s", MDGetSPARKRegNumName((STARS_regnum_t)InputIndex, (uint16_t)ByteWidth));
				}
				++OutputCount;
				FPStackBelowTopPrinted |= FPRegBelowStackTop;
			}
		}
	} // end for all InputRegs
	if (MemoryInput && (!MemoryOutput)) {
		if (0 == OutputCount) {
			if (0 < TotalOutputCount) {
				SMP_fprintf(HeaderFile, ",\n\t\t");  // line up with previous Global lines
			}
			SMP_fprintf(HeaderFile, "Input => (X86.Memory");
		}
		else {
			SMP_fprintf(HeaderFile, ", X86.Memory");
		}
		++OutputCount;
	}
	if (FlagsInput && (!FlagsOutput)) {
		if (0 == OutputCount) {
			SMP_fprintf(HeaderFile, "Input => (X86.CarryFlag, X86.ParityFlag, X86.SignFlag, X86.OverflowFlag, X86.ZeroFlag");
		}
		else {
			SMP_fprintf(HeaderFile, ", X86.CarryFlag, X86.ParityFlag, X86.SignFlag, X86.OverflowFlag, X86.ZeroFlag");
		}
		++OutputCount;
	}
	if (0 < OutputCount) {
		SMP_fprintf(HeaderFile, ")"); // close the Input open parenthesis
	}
	TotalOutputCount += OutputCount;

	SMP_fprintf(HeaderFile, ")"); // close the Global open parenthesis

	return;
} // end of SMPFunction::EmitSPARKLoopProcGlobals()

// recursive descent translation to SPARK Ada starting with CurrBlock, stop before Follow Block, or limit to blocks with the same LoopExitTargets reachability, stay within region dominated by DomBlockNum.
void SMPFunction::EmitSPARKAdaForBlock(int CurrTCFGBlockNum, const int DomTCFGBlockNum, const int FollowTCFGBlockNum, FILE *SPARKBodyFile, const bool ReadytoEmitSwitchDefault, const bool LoopToProc, const uint32_t ReachableSourcesCount) {
	this->DomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
	if (!this->DoesTCFGBlockDominateTCFGBlock(DomTCFGBlockNum, CurrTCFGBlockNum)) {
		return;
	}

	unsigned short IncomingTabDepth = STARS_SPARK_IndentCount;

	PrintSPARKIndentTabs(SPARKBodyFile);
	STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum((size_t)CurrTCFGBlockNum);
	int OrigCurrBlockNum = CurrTCFGBlock->GetOriginalBlockNum();
	STARSTransCFGBlock *DomTCFGBlock = this->GetTCFGBlockByNum((size_t)DomTCFGBlockNum);
	int OrigDomBlockNum = DomTCFGBlock->GetOriginalBlockNum();
	int InnerLoopNum = this->GetInnermostLoopNum(OrigCurrBlockNum);
	SMP_fprintf(SPARKBodyFile, "-- Basic block %d;\n", OrigCurrBlockNum);
	bool InsideLoop = (0 <= InnerLoopNum);
	if (LoopToProc && (!this->IsSPARKLoopInTranslationStack())) {
		// See if we are beginning to translate a guarded loop.
		SMPBasicBlock *CurrBlock = this->GetBlockByNum(OrigCurrBlockNum);
		STARS_ea_t LastAddr = CurrBlock->GetLastAddr();
		STARS_ea_t FirstAddr = CurrBlock->GetFirstNonMarkerAddr();
		ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
		SMPInstr *FirstInst = this->GetInstFromAddr(FirstAddr);
		SMPInstr *LastInst = this->GetInstFromAddr(LastAddr);
		if ((LastCFType == BRANCH_IF_THEN) && FirstInst->HasBeenTranslatedToSPARK() && (!LastInst->HasBeenTranslatedToSPARK())) {
			// We are beginning to translate a guarded loop. Only the COND_BRANCH statement
			//  at the end of CurrBlockNum needs to be translated, via EmitSPARKAdaForConditional().
			//  The rest of the block has been translated.
			//  After the recursive descent into EmitSPARKAdaForConditional(), we are done with
			//  this guarded loop.
			this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile);
			return;
		}
	}

	// Recursive descent based on the kind of block CurrBlock is.
	while ((CurrTCFGBlockNum != FollowTCFGBlockNum) && (CurrTCFGBlockNum >= 0)) {
		this->TCFGDomSoFar.ResetAllBits(); // prepare for memoization optimization
		if (!this->DoesTCFGBlockDominateTCFGBlock(DomTCFGBlockNum, CurrTCFGBlockNum)) {
			break;
		}

		CurrTCFGBlock = this->GetTCFGBlockByNum((size_t)CurrTCFGBlockNum);
		OrigCurrBlockNum = CurrTCFGBlock->GetOriginalBlockNum();

#if 1	// With the dominance test that we just passed, is this code redundant?
		if (0 < ReachableSourcesCount) {
			int RelevantLoopNum = this->GetLoopNumFromHeaderBlockNum(OrigDomBlockNum);
			if (0 > RelevantLoopNum)
				RelevantLoopNum = this->GetInnermostDominatingLoopNum(OrigCurrBlockNum);
			if ((0 <= RelevantLoopNum) && (RelevantLoopNum < (int) this->GetNumLoops()) && this->DoesLoopHaveMultipleExitTargets((size_t)RelevantLoopNum)) {
				uint32_t CurrReachableSourcesCount = this->GetReachableSourcesCount((size_t)CurrTCFGBlockNum, RelevantLoopNum);
				if (CurrReachableSourcesCount != ReachableSourcesCount) // cannot follow beyond scope of current guard Boolean
					break;
			}
		}
#endif
		SMPBasicBlock *OrigCurrBlock = this->GetBlockByNum(OrigCurrBlockNum);
		if (!OrigCurrBlock->IsReadyToTranslate())
			break;

		int ResumeTCFGBlockNum = FollowTCFGBlockNum;
		if (CurrTCFGBlock->IsLoopHeaderBlock()) {  // !!!!****!!!! Fix for funcs that start with a loop header
			int LoopNum = this->GetLoopNumFromHeaderBlockNum(OrigCurrBlockNum);
			assert(0 <= LoopNum);
			STARS_ea_t LoopAddr = CurrTCFGBlock->GetFirstNonMarkerAddr();
			bool TranslatingGuardedLoop = (this->LoopToGuardMap.find(LoopAddr) != this->LoopToGuardMap.cend());
			int NextFollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
			bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, NextFollowTCFGBlockNum);
			// Could end with a loop if we call abort, etc., in which case no follow block.
			//  Could be an infinite loop. Otherwise, we should have a follow block.
			bool FuncMightNotReturn = ((!this->FuncReturnsToCaller()) || this->HasCallToNonReturningFunc());
			bool LoopMightBeInfinite = (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == this->LoopTypesByLoopNum[LoopNum]);
			// assert((SMP_BLOCKNUM_UNINIT != NextFollowBlockNum) || FuncMightNotReturn || LoopMightBeInfinite);

			if (LoopToProc) {
				assert(TranslatingGuardedLoop || (ResumeTCFGBlockNum == NextFollowTCFGBlockNum)); // should be passed in as loop follow block number
				this->EmitSPARKAdaForLoop(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile); // recurse and resume
			}
			else { // !LoopToProc case
				bool MultipleExitTargets = this->EmitSPARKAdaLoopCall(LoopAddr, (size_t) LoopNum, SPARKBodyFile);
				ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
				pair<int, int> BlockItem(CurrTCFGBlockNum, ResumeTCFGBlockNum);
				pair<int, pair<int, int> > WorkListItem(LoopNum, BlockItem);
				this->SPARKLoopWorkList.push_back(WorkListItem);
				this->CleanUpSPARKLoopWorkList();
			}
		}
		else if (CurrTCFGBlock->IsLoopTailBlock()) {
			// assert(false);
			ControlFlowType LastCFType = CurrTCFGBlock->GetLastInstCFType();
			vector<SMPInstr *>::const_reverse_iterator LastInstIter = CurrTCFGBlock->GetRevInstCBegin();
			SMPInstr *LastInst = (*LastInstIter);
			SMPitype FlowType = LastInst->GetDataFlowType();
			STARS_ea_t LastInstAddr = LastInst->GetAddr();
			// We are at the end of a recursive descent into a loop. Translate the block and return to 
			//  EmitSPARKAdaForLoop() unless we are starting a conditional inside the loop when the LOOP_BACK
			//  branch is not taken.
			if (CurrTCFGBlock->HasLoopHeadWithInvertedExitAsSuccessor()) {
				// Last inst in block is transfer to loop head. Need to ensure that we don't
				//  emit "end loop;" before we return to our caller and emit "end if;" in order
				//  to get proper nesting of control structures. 
				if (JUMP == FlowType) {
					// One branch of an if-then-else can fall through to the inverted-exit loop head.
					//  The other branch jumps to the loop header.
					OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile);
					SPARKTranslationCFType CurrentCFType = this->SPARKControlStack.back();
					if (SPARK_LOOP == CurrentCFType) {
						// Must be called from EmitSPARKAdaForLoop(); time for "end loop;" without "end if;"
#if STARS_SPARK_CENTRALIZE_EMIT_LOOP
						;
#else
						--STARS_SPARK_IndentCount;
						PrintSPARKIndentTabs(SPARKBodyFile);
						SMP_fprintf(SPARKBodyFile, "end loop;\n\n");
#endif
						LastInst->SetSPARKTranslated(true);
					}
					else if (SPARK_ELSE_CLAUSE == CurrentCFType) {
						// Must be called from EmitSPARKAdaForConditional(); time for "end if;" and then "end loop;"
						--STARS_SPARK_IndentCount;
						PrintSPARKIndentTabs(SPARKBodyFile);
						SMP_fprintf(SPARKBodyFile, "end if;\n\n");
						--this->SPARKConditionalNestingDepth;
#if STARS_SPARK_CENTRALIZE_EMIT_LOOP
						;
#else
						--STARS_SPARK_IndentCount;
						PrintSPARKIndentTabs(SPARKBodyFile);
						SMP_fprintf(SPARKBodyFile, "end loop;\n\n");
#endif
						LastInst->SetSPARKTranslated(true);
					}
					else { // mysterious
						LastInst->EmitSPARKAda(SPARKBodyFile, false);
					}
				}
				else {
					bool StaysInLoop = this->IsNonExitingLoopBackBranch(LastInstAddr);
					if (!StaysInLoop) { // just loop-back and return to EmitSPARKAdaForLoop().
						OrigCurrBlock->EmitSPARKAdaForAllInsts(SPARKBodyFile);
					}
					else {
						// We need to emit the fall through insts, then translate COND_BRANCH to loop header
						//  into if (inverted condition) then ... for the rest of the loop body.
						OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile);
						// EmitSPARKAdaForConditional() will emit the if (inverted condition) then ... and
						//  will continue through the rest of the if-then body.
						this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile);
					}
				}
			}
			else {
				bool StaysInLoop = this->IsNonExitingLoopBackBranch(LastInstAddr);
				if (!StaysInLoop) { // just loop-back and return to EmitSPARKAdaForLoop().
					OrigCurrBlock->EmitSPARKAdaForAllInsts(SPARKBodyFile);
				}
				else {
					// We need to emit the fall through insts, then translate COND_BRANCH to loop header
					//  into if (inverted condition) then ... for the rest of the loop body.
					OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile);
					// EmitSPARKAdaForConditional() will emit the if (inverted condition) then ... and
					//  will continue through the rest of the if-then body.
					this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile);
				}
			}
			return;
		}
		else if (OrigCurrBlock->IsSwitchDefaultCase() && (!ReadytoEmitSwitchDefault)) {
			// We can encounter the default case before the INDIR_JUMP for a switch.
			//  EmitSPARKAdaForSwitch() will emit the default case last.
			int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, false);
			this->EmitSPARKAdaForSwitch(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
			ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
		}
		else { // Not a special block, or it is a ReadytoEmitSwitchDefault block.
			// Translate instructions in current block, then see if block terminating inst
			//  tells us what control structure we are entering, e.g. loop or switch-case.
			OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile); // all except last inst if it is control flow terminator for block
			vector<SMPInstr *>::const_reverse_iterator LastInstIter = CurrTCFGBlock->GetRevInstCBegin();
			SMPInstr *LastInst = (*LastInstIter);
			SMPitype FlowType = LastInst->GetDataFlowType();
			if (FlowType != DEFAULT) {
				STARS_ea_t LastAddr = LastInst->GetAddr();
				ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
				if (FlowType == INDIR_JUMP) {
					// Translate switch-case statement.
					int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, true);
					this->EmitSPARKAdaForSwitch(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
					ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
					LastInst->SetSPARKTranslated(true);
				}
				else if (FlowType == COND_BRANCH) {
					if (1 == CurrTCFGBlock->GetNumSuccessors()) {
						// Stupid unoptimized code like "jz $+2" that branches to the fall-through inst.
						//  We will just treat as a no-op.
						assert(FALL_THROUGH == LastCFType);
						LastInst->EmitSPARKAda(SPARKBodyFile, false);
						// We don't have a true COND_BRANCH, so we don't get a conditional follow node
						//  in the normal way. Just grab the lone successor block num and resume
						//  translation there.
						ResumeTCFGBlockNum = CurrTCFGBlock->GetCondNonFallThroughSuccBlockNum();
					}
					else if (LastCFType == JUMP_TO_SWITCH_INDIR_JUMP) {
						// Should fall through to default case block. Resume translation there,
						//  which will cause use to emit the entire switch statement.
						list<STARSTransCFGBlock *>::const_iterator FallThroughIter = CurrTCFGBlock->GetFallThroughSucc();
						assert(FallThroughIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH always has a fall-through
						ResumeTCFGBlockNum = (*FallThroughIter)->GetNumber();
						LastInst->SetSPARKTranslated(true);
					}
					else if (LastCFType == JUMP_TO_DEFAULT_CASE) {  
						// jump to default case is not translated directly
						// We do not follow the branch to the default case; that case will be translated when the 
						//  switch statement is translated. Instead, we go to the fall-through block. We need to
						//  translate all of the non-branching instructions that lead up to the switch statement in
						//  order to get the switch index value computed properly, e.g:
						//  cmp eax,8        ; translate, but we don't really need it in most cases
						//  jb default_case  ; no translation
						//  sub eax,8        ; need this translation
						//  cmp eax,12       ; translate, don't need it in most cases
						//  ja default_case  ; don't translate
						//  jmp [memaddr + eax*4]  ; translate to "case X86.EAX is"
						list<STARSTransCFGBlock *>::const_iterator FallThroughIter = CurrTCFGBlock->GetFallThroughSucc();
						assert(FallThroughIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH always has a fall-through
						ResumeTCFGBlockNum = (*FallThroughIter)->GetNumber();
						LastInst->SetSPARKTranslated(true);
					}
					else if (IsLoopExitFlow(LastCFType)) {
						// Middle exit, else it would be loop header or tail handled above.
						LastInst->EmitSPARKAda(SPARKBodyFile, false); // emit exit when (condition)
						// If we conditionally exit the loop but the other successor is inside the loop,
						//  we need to resume with the other successor.
						int FTBlockNum = SMP_BLOCKNUM_UNINIT;
						if (!IsShortCircuitFlow(LastCFType)) {
							list<STARSTransCFGBlock *>::const_iterator FallThroughIter = CurrTCFGBlock->GetFallThroughSucc();
							assert(FallThroughIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH always has a fall-through
							STARSTransCFGBlock *FTBlock = (*FallThroughIter);
							FTBlockNum = FTBlock->GetNumber();
							int OrigFTBlockNum = FTBlock->GetOriginalBlockNum();
							if (this->DoesEdgeExitLoop(OrigCurrBlockNum, OrigFTBlockNum)) {
								// We want to resume with the block that does NOT exit the loop.
								list<STARSTransCFGBlock *>::const_iterator NonFallThroughIter = CurrTCFGBlock->GetCondNonFallThroughSucc();
								assert(NonFallThroughIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH always has a fall-through
								int NFTBlockNum = (*NonFallThroughIter)->GetNumber();
								int OrigNFTBlockNum = (*NonFallThroughIter)->GetOriginalBlockNum();
								if ((!this->DoesEdgeExitLoop(OrigCurrBlockNum, OrigNFTBlockNum)) && (*NonFallThroughIter)->IsReadyToTranslate()) {
									ResumeTCFGBlockNum = NFTBlockNum;
								}
							}
							else if (FTBlock->IsReadyToTranslate()) {
								ResumeTCFGBlockNum = FTBlockNum;
							}
						}
						else {
							STARSCFGBlock *CFGBlock = this->GetClonedCFGBlockByNum((size_t)CurrTCFGBlockNum);
							assert(nullptr != CFGBlock);
							FTBlockNum = CFGBlock->GetExpr()->GetSuccNumInLoop((size_t)InnerLoopNum);
							// assert(OrigFTBlockNum > 0);
							// FTBlockNum = CurrTCFGBlock->FindTCFGBlockNumMatchingOrigBlockNum(OrigFTBlockNum);
							assert(FTBlockNum > 0);
							ResumeTCFGBlockNum = FTBlockNum;
						}
					}
					else {
						assert((LastCFType == BRANCH_IF_THEN) || (LastCFType == BRANCH_IF_THEN_ELSE) || (LastCFType == SHORT_CIRCUIT_BRANCH));
						// Translate if-else structure previously identified
						int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, true);
						this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
						if (0 < NextFollowTCFGBlockNum) {
							STARSTransCFGBlock *NextFollowTCFGBlock = this->GetTCFGBlockByNum((size_t)NextFollowTCFGBlockNum);
							int OrigNextFollowBlockNum = NextFollowTCFGBlock->GetOriginalBlockNum();
							bool CrossingLoopBoundary = (InsideLoop && (0 < NextFollowTCFGBlockNum) && this->DoesEdgeExitLoop(OrigCurrBlockNum, OrigNextFollowBlockNum));
							if (!CrossingLoopBoundary) {
								ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
							}
						}
					}
				}
				else if (FlowType == JUMP) {
					// Get control flow type of the jump
					assert((1 == CurrTCFGBlock->GetNumSuccessors()) || LastInst->IsBranchToOtherFunc());
					if (JUMP_INTO_LOOP_TEST == LastCFType) {
						// Optimized top-testing loop becomes a bottom-testing loop with a jump to the loop header/loop-test
						//  block for the first iteration. We translate as a top-testing loop in Ada by starting with the 
						//  target block of the jump.
						STARSTransCFGBlock *NextTCFGBlock = (*(CurrTCFGBlock->GetFirstConstSucc()));
						assert(NextTCFGBlock->IsLoopHeaderBlock());
						int NextTCFGBlockNum = NextTCFGBlock->GetNumber();
						int NextOrigBlockNum = NextTCFGBlock->GetOriginalBlockNum();
						int LoopNum = this->GetLoopNumFromHeaderBlockNum(NextOrigBlockNum);
						assert(0 <= LoopNum);
						int NextFollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
						bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, NextFollowTCFGBlockNum);
						bool LoopMightBeInfinite = (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == this->LoopTypesByLoopNum[LoopNum]);
						// assert((SMP_BLOCKNUM_UNINIT != NextFollowBlockNum) || LoopMightBeInfinite);
						if (LoopToProc) {
							this->EmitSPARKAdaForLoop(NextTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile); // recurse and return
							ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
							LastInst->SetSPARKTranslated(true);
						}
						else { // !LoopToProc case
							STARS_ea_t LoopAddr = NextTCFGBlock->GetFirstAddr();
							
							bool MultipleExitTargets = this->EmitSPARKAdaLoopCall(LoopAddr, (size_t)LoopNum, SPARKBodyFile);
							ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
							LastInst->SetSPARKTranslated(true);
							pair<int, int> BlockItem(NextTCFGBlockNum, ResumeTCFGBlockNum);
							pair<int, pair<int, int> > WorkListItem(LoopNum, BlockItem);
							this->SPARKLoopWorkList.push_back(WorkListItem);
							this->CleanUpSPARKLoopWorkList();
						}
					}
					else if ((JUMP_TO_DEFAULT_CASE == LastCFType) || (JUMP_TO_SWITCH_INDIR_JUMP == LastCFType)) {
						; // jump to default case is not translated directly; ditto for jump around default case to INDIR_JUMP
						LastInst->SetSPARKTranslated(true);
					}
					else {
						// All other cases should be handled in SMPInstr::EmitSPARKAda().
						LastInst->EmitSPARKAda(SPARKBodyFile, false);
						LastInst->SetSPARKTranslated(true);
						// Handle special case in which a JUMP just goes to the lone successor,
						//  which is equivalent to falling through.
						// if ((FALL_THROUGH == LastCFType) && (ResumeBlockNum < 0)) {
						if (FALL_THROUGH == LastCFType) {
							ResumeTCFGBlockNum = (*(CurrTCFGBlock->GetFirstConstSucc()))->GetNumber();
						}
					}
				}
				else if (FlowType == RETURN) {
					LastInst->EmitSPARKAda(SPARKBodyFile, false);
					// Handle possible tail call case
					if ((FALL_THROUGH == LastCFType) && (1 == CurrTCFGBlock->GetNumSuccessors())) {
						assert(LastInst->IsCondTailCall());
						ResumeTCFGBlockNum = (*(CurrTCFGBlock->GetFirstConstSucc()))->GetNumber();
					}
				}
				else if ((FlowType == CALL) || (FlowType == INDIR_CALL)) {
					if (!LastInst->HasBeenTranslatedToSPARK() || LastInst->IsBasicBlockTerminator()) {
						LastInst->EmitSPARKAda(SPARKBodyFile, false);
					}
					if (LastCFType == FALL_THROUGH) {
						// if ((FALL_THROUGH == LastCFType) && (ResumeBlockNum < 0)) {
						list<STARSTransCFGBlock *>::const_iterator SuccIter = CurrTCFGBlock->GetFirstConstSucc();
						if (SuccIter != CurrTCFGBlock->GetLastConstSucc()) {
							ResumeTCFGBlockNum = (*SuccIter)->GetNumber();
						}
					}
				}
			}
			else {
				// Last inst was translated by SMPBasicBlock::EmitSPARKAdaForFallThroughInsts().
				//  CurrBlock can only have one successor, a fall-through block.
				// NOTE: If CurrBlock ends with a HALT or a call to a NORET function, we can have no successors.
				size_t NumSuccessors = CurrTCFGBlock->GetNumSuccessors();
				if (1 == NumSuccessors) {
					STARSTransCFGBlock *FallThroughSucc = (*(CurrTCFGBlock->GetFirstSucc()));
					assert(nullptr != FallThroughSucc);
					ResumeTCFGBlockNum = FallThroughSucc->GetNumber();
				}
				else {
					// Nop at end of block could be for alignment, masking the control flow type of CALL
					//  or HALT before it.
					bool SpecialLastInst = (LastInst->IsNop() || LastInst->MDIsExceptionThrow());
					bool GoodCode = (((FlowType == CALL) || (FlowType == HALT) || CurrTCFGBlock->IsUnreachableBlock() || SpecialLastInst)
						&& (0 == NumSuccessors));
					if (!GoodCode) {
						SMP_msg("ERROR: SPARK: Bad control flow at end of block, inst at %llx ", (uint64_t)LastInst->GetAddr());
						this->DumpFuncNameAndAddr();
						SMP_fprintf(SPARKBodyFile, "\nERROR -- Bad control flow at end of block\n");
					}
				}
			}
		} // end if header block ... elsif tail block ... else ...
		CurrTCFGBlockNum = ResumeTCFGBlockNum;
	} // end while good CurrBlockNum

	return;
} // end of SMPFunction::EmitSPARKAdaForBlock()

// recursive descent translation of loop to SPARK Ada starting with header CurrBlock, stop before Follow Block
void SMPFunction::EmitSPARKAdaForLoop(int HeaderTCFGBlockNum, int FollowTCFGBlockNum, FILE *SPARKBodyFile) {
	if (HeaderTCFGBlockNum == FollowTCFGBlockNum)
		return;  // end recursion if Follow Block is seen

	assert(0 <= HeaderTCFGBlockNum);
	int CurrTCFGBlockNum = HeaderTCFGBlockNum;
	STARSTransCFGBlock *HeaderTCFGBlock = this->GetTCFGBlockByNum((size_t)HeaderTCFGBlockNum);
	int ResumeTCFGBlockNum = FollowTCFGBlockNum;
	this->SPARKControlStack.push_back(SPARK_LOOP);
	int OrigCurrBlockNum = HeaderTCFGBlock->GetOriginalBlockNum();
	int LoopNum = this->GetLoopNumFromHeaderBlockNum(OrigCurrBlockNum);
	assert(0 <= LoopNum);
	STARS_ea_t FirstAddr = HeaderTCFGBlock->GetFirstNonMarkerAddr();
	assert(STARS_BADADDR != FirstAddr);
	SMPInstr *FirstInst = *(HeaderTCFGBlock->GetInstIterFromAddr(FirstAddr));
	bool MultipleExitTargets = this->DoesLoopHaveMultipleExitTargets((size_t) LoopNum);

#if STARS_SPARK_CENTRALIZE_EMIT_LOOP
	// We want to print "loop" and increase indentation.
	PrintSPARKIndentTabs(SPARKBodyFile);
	SMP_fprintf(SPARKBodyFile, "loop \n");
	++STARS_SPARK_IndentCount;
	FirstInst->EmitSPARKAdaLoopInvariants(SPARKBodyFile);

	// Emit Booleans that track CFG flow if we have multiple exit targets.
	if (MultipleExitTargets) {
		// Generate the "Boolean flag = false;" declaration and initialization for each exit target block.
		for (int ExitTargetBlockNum : this->LoopExitTargets[LoopNum]) {
			string TargetString = this->GenerateSPARKLoopExitBooleanName(LoopNum, ExitTargetBlockNum);
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "%s : Boolean := false;\n", TargetString.c_str());
		}
	}
#endif

	while ((CurrTCFGBlockNum != FollowTCFGBlockNum) && (CurrTCFGBlockNum >= 0)) {
		assert((0 <= CurrTCFGBlockNum) && (CurrTCFGBlockNum < (int)this->TCFGBlocks.size()));
		STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum(CurrTCFGBlockNum);
		if (!CurrTCFGBlock->IsReadyToTranslate())
			break;

		int OrigCurrBlockNum = CurrTCFGBlock->GetOriginalBlockNum();
		bool StartingLoop = (CurrTCFGBlock->IsLoopHeaderBlock() && (CurrTCFGBlockNum == HeaderTCFGBlockNum));
		SMPBasicBlock *OrigCurrBlock = this->GetBlockByNum((size_t)OrigCurrBlockNum);

		// Recursive descent based on the kind of block CurrBlock is.
		if (CurrTCFGBlock->IsLoopHeaderBlock() && (!StartingLoop)) {
			// We have fallen through to an inner loop. Emit a call to the loop proc created later, and jump past inner loop.
			int InnerLoopNum = this->GetLoopNumFromHeaderBlockNum(OrigCurrBlockNum);
			assert(0 <= InnerLoopNum);
			int NextFollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
			bool MultiExitTargetsInnerLoop = this->GetLoopFollowBlockNum((size_t)InnerLoopNum, NextFollowTCFGBlockNum);
			bool LoopMightBeInfinite = (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == this->LoopTypesByLoopNum[InnerLoopNum]);
			// assert((SMP_BLOCKNUM_UNINIT != NextFollowTCFGBlockNum) || LoopMightBeInfinite);
			bool MultipleExitTargets = this->EmitSPARKAdaLoopCall(CurrTCFGBlock->GetFirstAddr(), (size_t) InnerLoopNum, SPARKBodyFile); 
			ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
			pair<int, int> BlockItem(CurrTCFGBlockNum, ResumeTCFGBlockNum);
			pair<int, pair<int, int> > WorkListItem(InnerLoopNum, BlockItem);
			this->SPARKLoopWorkList.push_back(WorkListItem);
			this->CleanUpSPARKLoopWorkList();
		}
		else if (CurrTCFGBlock->IsLoopTailBlock()) {
			// We are at the end of a recursive descent into a loop. Translate the block and return.
			// NOTE: What about one-block loops, where block is header block and tail block?   !!!!****!!!!
			STARS_ea_t BranchAddr = CurrTCFGBlock->GetLastAddr();
			bool StaysInLoop = this->IsNonExitingLoopBackBranch(BranchAddr);
			if (StaysInLoop) {
				// At the end of CurrBlock, emit an if-then that encloses the path from fallthrough onwards. 
				OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile);
				// The code for the if-then has scope until the end of the loop; use loop FollowBlockNum.
				this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile);
			}
			else {
				OrigCurrBlock->EmitSPARKAdaForAllInsts(SPARKBodyFile);
				break;
			}
		}
		else if (CurrTCFGBlock->IsSwitchDefaultCase()) {
			// Translate switch-case statement.
			int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, false);
			this->EmitSPARKAdaForSwitch(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
			ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
		}
		else {
			// Translate instructions in current block, then see if block terminating inst
			//  tells us what control structure we are entering, e.g. loop or switch-case.
			OrigCurrBlock->EmitSPARKAdaForFallThroughInsts(SPARKBodyFile); // all except last inst if it is control flow terminator for block
			vector<SMPInstr *>::const_reverse_iterator LastInstIter = CurrTCFGBlock->GetRevInstCBegin();
			SMPInstr *LastInst = (*LastInstIter);
			STARS_ea_t LastAddr = LastInst->GetAddr();
			SMPitype FlowType = LastInst->GetDataFlowType();
			if (LastInst->IsBasicBlockTerminator()) {
				ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
				if (FlowType == INDIR_JUMP) {
					// Translate switch-case statement.
					int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, true);
					this->EmitSPARKAdaForSwitch(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
					ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
				}
				else if (FlowType == COND_BRANCH) {
					if (LastCFType == JUMP_TO_SWITCH_INDIR_JUMP) {
						// conditional branch to INDIR_JUMP is not well-structured; should not occur
						SMP_msg("FATAL ERROR: COND_BRANCH of type JUMP_TO_SWITCH_INDIR_JUMP at %llx ", (uint64_t) LastAddr);
						this->DumpFuncNameAndAddr();
						this->DumpDotCFG();
						this->Dump(false);
						assert(LastCFType != JUMP_TO_SWITCH_INDIR_JUMP); // kablooey!
					}
					else if (IsLoopExitFlow(LastCFType)) {
						// Middle exit, else it would be loop header or tail handled above.
						LastInst->EmitSPARKAda(SPARKBodyFile, false); // emit exit when (condition)
						// We want to resume at the fall-through block, if it is in the loop,
						//  which it should be because this is not the tail block.
						if (IsShortCircuitFlow(LastCFType)) {
							// We have to follow the IsSingleExpression() descendants to get to the end of
							//  the short-circuit expression. Then we can look at its successors and find one
							//  that is in the loop, or set ResumeBlockNum to the loop follow node.
							STARSCFGBlock *CurrCFGBlock = this->GetClonedCFGBlockByNum((size_t)CurrTCFGBlockNum);
							assert(nullptr != CurrCFGBlock);
							int SuccBlockNum = CurrCFGBlock->GetExpr()->GetSuccNumInLoop((size_t)LoopNum);
							if (0 <= SuccBlockNum) {
								ResumeTCFGBlockNum = SuccBlockNum;
							}
							else {
								ResumeTCFGBlockNum = FollowTCFGBlockNum; // done with loop
							}
						}
						else {
							list<STARSTransCFGBlock *>::const_iterator SuccIter = CurrTCFGBlock->GetSuccNotInLoop((size_t)LoopNum);
							assert(SuccIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH must have fall through and non-fall-through
							int SuccBlockNum = (*SuccIter)->GetNumber();
							SuccIter = CurrTCFGBlock->GetCondOtherSucc(SuccBlockNum);
							assert(SuccIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH must have fall through and non-fall-through
							ResumeTCFGBlockNum = (*SuccIter)->GetNumber();
						}
					}
					else if ((LastCFType == BRANCH_IF_THEN) || (LastCFType == BRANCH_IF_THEN_ELSE) || (LastCFType == SHORT_CIRCUIT_BRANCH)) {
						// Translate if-else structure previously identified
						int NextFollowTCFGBlockNum = this->FindFollowBlockNum(CurrTCFGBlock, true);
						this->EmitSPARKAdaForConditional(CurrTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile);
						ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
					}
					else if (LastCFType != JUMP_TO_DEFAULT_CASE) {
						LastInst->EmitSPARKAda(SPARKBodyFile, false); // handles loop-back case
						// If this is the optimized top-testing block in which the header was moved to the bottom
						//  position of the loop, then we will have a LOOP_BACK control flow type rather than a
						//  LOOP_EXIT type for the COND_BRANCH at the end of the header block. LastInst->EmitSPARKAda()
						//  will emit an inverted-condition loop exit, but now we need to fall through into the successor
						//  block that is within our loop.
						if (OrigCurrBlock->IsOptimizedTopLoopTest()) {
							assert(LOOP_BACK == this->GetControlFlowType(LastAddr));
							int LoopNum = this->GetLoopNumFromHeaderBlockNum(OrigCurrBlockNum);
							assert(0 <= LoopNum);
							for (STARSTransCFGBlock *SuccTCFGBlock : CurrTCFGBlock->GetConstSuccList()) {
								int SuccTCFGBlockNum = SuccTCFGBlock->GetNumber();
								if (this->IsTCFGBlockInLoop(SuccTCFGBlockNum, LoopNum)) {
									ResumeTCFGBlockNum = SuccTCFGBlockNum;
									break;
								}
							}
						}
						else if (FALL_THROUGH == LastCFType) {
							size_t NumSuccessors = CurrTCFGBlock->GetNumSuccessors();
							if (1 == NumSuccessors) {
								STARSTransCFGBlock *FallThroughSucc = (*(CurrTCFGBlock->GetFirstSucc()));
								assert(nullptr != FallThroughSucc);
								int NewResumeTCFGBlockNum = FallThroughSucc->GetNumber();
								int OrigResumeBlockNum = FallThroughSucc->GetOriginalBlockNum();
								if (this->GetInnermostLoopNum(OrigResumeBlockNum) == LoopNum) {
									// Odd fall through, like a jump to the fall-through block.
									//  In any case, we are staying inside the loop, so we need to go
									//  on to the fall-through successor.
									if (ResumeTCFGBlockNum != NewResumeTCFGBlockNum) {
										SMP_msg("INFO: SPARK: Adjusting LoopNum %d resume block from %d to %d at TCFGBlock %d which is FALL_THROUGH COND_BRANCH with 1 successor ",
											LoopNum, ResumeTCFGBlockNum, NewResumeTCFGBlockNum, CurrTCFGBlockNum);
										this->DumpFuncNameAndAddr();
										ResumeTCFGBlockNum = NewResumeTCFGBlockNum;
									}
								}
							}
						}
					}
					else { // JUMP_TO_DEFAULT_CASE
						// See comments in EmitSPARKAdaForBlock() for the JUMP_TO_DEFAULT_CASE issues.
						//  Here, we are starting to encounter a switch statement inside our loop.
						list<STARSTransCFGBlock *>::const_iterator FallThroughIter = CurrTCFGBlock->GetFallThroughSucc();
						assert(FallThroughIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH always has a fall-through
						ResumeTCFGBlockNum = (*FallThroughIter)->GetNumber();
						LastInst->SetSPARKTranslated(true); // Ada "others ==>" clause subsumes all COND_BRANCHes to default case.
					}
				}
				else if (FlowType == JUMP) {
					// Get control flow type of the jump
					assert(1 == CurrTCFGBlock->GetNumSuccessors());
					if (JUMP_INTO_LOOP_TEST == LastCFType) {
						// Optimized top-testing loop becomes a bottom-testing loop with a jump to the loop header/loop-test
						//  block for the first iteration. We translate as a top-testing loop in Ada by starting with the 
						//  target block of the jump. At this level, we emit a call and resume after the inner loop.
						STARSTransCFGBlock *NextTCFGBlock = (*(CurrTCFGBlock->GetFirstSucc()));
						assert(NextTCFGBlock->IsLoopHeaderBlock());
						int NextTCFGBlockNum = NextTCFGBlock->GetNumber();
						int NextOrigBlockNum = NextTCFGBlock->GetOriginalBlockNum();
						int LoopNum = this->GetLoopNumFromHeaderBlockNum(NextOrigBlockNum);
						assert(0 <= LoopNum);
						int NextFollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
						bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, NextFollowTCFGBlockNum);
						bool LoopMightBeInfinite = (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == this->LoopTypesByLoopNum[LoopNum]);
						// assert((SMP_BLOCKNUM_UNINIT != NextFollowTCFGBlockNum) || LoopMightBeInfinite);
						bool InnerMultipleExitTargets = this->EmitSPARKAdaLoopCall(NextTCFGBlock->GetFirstAddr(), (size_t) LoopNum, SPARKBodyFile);
						ResumeTCFGBlockNum = NextFollowTCFGBlockNum;
						LastInst->SetSPARKTranslated(true);
						pair<int, int> BlockItem(NextTCFGBlockNum, ResumeTCFGBlockNum);
						pair<int, pair<int, int> > WorkListItem(LoopNum, BlockItem);
						this->SPARKLoopWorkList.push_back(WorkListItem);
						this->CleanUpSPARKLoopWorkList();
					}
					else if ((JUMP_TO_DEFAULT_CASE == LastCFType) || (JUMP_TO_SWITCH_INDIR_JUMP == LastCFType)) {
						; // jump to default case is not translated directly; ditto for jump around default case to INDIR_JUMP
						LastInst->SetSPARKTranslated(true);
					}
					else {
						// All other cases should be handled in SMPInstr::EmitSPARKAda().
						LastInst->EmitSPARKAda(SPARKBodyFile, false);
						// Handle special case in which a JUMP just goes to the lone successor,
						//  which is equivalent to falling through.
						if (FALL_THROUGH == LastCFType) {
							ResumeTCFGBlockNum = (*(CurrTCFGBlock->GetFirstConstSucc()))->GetNumber();
						}
					}
				}
				else if (FlowType == RETURN) {
					LastInst->EmitSPARKAda(SPARKBodyFile, false);
					// Handle possible tail call case
					;
				}
			} // end if LastInst is basic block terminator
			else {
				// Last inst was translated by SMPBasicBlock::EmitSPARKAdaForFallThroughInsts().
				//  CurrBlock can only have one successor, a fall-through block.
				// NOTE: If CurrBlock ends with a HALT or a call to a NORET function, we can have no successors.
				size_t NumSuccessors = CurrTCFGBlock->GetNumSuccessors();
				if (1 == NumSuccessors) {
					STARSTransCFGBlock *FallThroughSucc = (*(CurrTCFGBlock->GetFirstSucc()));
					assert(nullptr != FallThroughSucc);
					ResumeTCFGBlockNum = FallThroughSucc->GetNumber();
				}
				else {
					assert(((FlowType == CALL) || (FlowType == HALT)) && (0 == NumSuccessors));
				}
			} // end if LastInst is basic block terminator ... else ...
		} // end if loop tail block ... else ...

		// We have translated the header block, and recursed into the next control flow structure if the end of the
		//  header block indicated a new structure was beginning. Time to pick up with ResumeBlock unless it is our
		//  follow block.
		CurrTCFGBlockNum = ResumeTCFGBlockNum;
	} // end while CurrBlockNum is in the loop

	this->SPARKControlStack.pop_back();

#if STARS_SPARK_CENTRALIZE_EMIT_LOOP
	// We want to print "end loop;" and decrease indentation.
	--STARS_SPARK_IndentCount;
	PrintSPARKIndentTabs(SPARKBodyFile);
	SMP_fprintf(SPARKBodyFile, "end loop;\n");
	this->TranslatingSPARKLoop = false;

	if (MultipleExitTargets) {
		this->EmitSPARKAdaForMultipleLoopExitTargets(SPARKBodyFile, (size_t)LoopNum, HeaderTCFGBlockNum);
	}

#endif

	return;
} // end of SMPFunction::EmitSPARKAdaForLoop()

void SMPFunction::EmitSPARKAdaForMultipleLoopExitTargets(FILE *SPARKBodyFile, const size_t LoopNum, const int HeaderTCFGBlockNum) {
	// Translate blocks that are reachable from the multiple exit
	//  target blocks, including those exit target blocks, guarded
	//  by checks of the CFG Boolean flags that control access to
	//  those blocks in the original control flow.

	// Start by getting the union of all block nums reachable from the exit target blocks
	//  that are dominated by the loop head block.
	int CurrLoopFollowTCFGBlockNum;
	(void) this->GetLoopFollowBlockNum(LoopNum, CurrLoopFollowTCFGBlockNum);
	int LoopHeaderOrigBlockNum = this->LoopHeadBlockNumbers[LoopNum];
	assert(0 <= LoopHeaderOrigBlockNum);
	STARSBitSet ReachableUnion;
	ReachableUnion.AllocateBits(2 * this->GetNumBlocks()); // allow room for cloning
	bool Initialized = false;
	assert(!this->LoopExitTargetsReachabilitySets[LoopNum].empty());
	for (int ExitTargetTCFGBlockNum : this->LoopExitTargets[LoopNum]) {
		if (!Initialized) {
			ReachableUnion = this->LoopExitTargetsReachabilitySets[LoopNum][(size_t)ExitTargetTCFGBlockNum];
			Initialized = true;
		}
		else {
			ReachableUnion.UnionSets(this->LoopExitTargetsReachabilitySets[LoopNum][(size_t)ExitTargetTCFGBlockNum]);
		}
	}

	// Now, generate code for each block number in ReachableUnion, guarded based on the bits in each reachability set.
	size_t StartTCFGBlockNum = (size_t)ReachableUnion.FindLowestBitSet();
	size_t EndTCFGBlockNum = (size_t)ReachableUnion.FindHighestBitSet();
	assert(StartTCFGBlockNum < this->GetNumTCFGBlocks());
	assert(EndTCFGBlockNum < this->GetNumTCFGBlocks());

#if 0   // MultiExitTargets could go beyond follow block with a return block. Use DomFrontier of header to limit below.
	// Limit translation to stop at the follow block.
	if ((((int)EndTCFGBlockNum) >= CurrLoopFollowTCFGBlockNum) && (0 < CurrLoopFollowTCFGBlockNum)) {
		EndTCFGBlockNum = (size_t)(CurrLoopFollowTCFGBlockNum - 1);
	}
#endif

	STARSTransCFGBlock *HeaderTCFGBlock = this->GetTCFGBlockByNum((size_t)HeaderTCFGBlockNum);
	assert(nullptr != HeaderTCFGBlock);
	for (size_t TCFGBlockNum = StartTCFGBlockNum; TCFGBlockNum <= EndTCFGBlockNum; ++TCFGBlockNum) {
		if (HeaderTCFGBlock->IsBlockInDomFrontier((int)TCFGBlockNum)) {
			continue; // don't follow paths from exit targets beyond loop head dominance
		}

		STARSTransCFGBlock *CurrTCFGBlock = this->GetTCFGBlockByNum(TCFGBlockNum);
		STARSCFGBlock *CurrShadowCFGBlock = this->GetClonedCFGBlockByNum(TCFGBlockNum);
		bool EligibleBlock = (!CurrShadowCFGBlock->IsCoalesced()); // only generate code starting at short-circuit head blocks
		if (ReachableUnion.GetBit(TCFGBlockNum) && CurrTCFGBlock->IsReadyToTranslate() && EligibleBlock) {
			bool StartedGuardOutput = false;
			uint32_t ReachableSourceCount = 0;
			for (const auto &CurrReachablePair : this->LoopExitTargetsReachabilitySets[LoopNum]) {
				STARSBitSet CurrBitSet;
				CurrBitSet.AllocateBits(CurrReachablePair.second.GetNumBits());
				CurrBitSet = CurrReachablePair.second;
				if (CurrBitSet.GetBit(TCFGBlockNum)) {
					if (CurrTCFGBlock->IsLoopHeaderBlock()) {
						// We have to emit a loop call and not translate blocks in this loop.
						STARS_ea_t LoopAddr = CurrTCFGBlock->GetFirstNonMarkerAddr();
						int OriginalTCFGBlockNum = CurrTCFGBlock->GetOriginalBlockNum();
						size_t InnerLoopNum = this->GetLoopNumFromHeaderBlockNum(OriginalTCFGBlockNum);
						assert(0 <= InnerLoopNum);
						bool MultipleExitTargets = this->EmitSPARKAdaLoopCall(LoopAddr, InnerLoopNum, SPARKBodyFile);
						if (MultipleExitTargets) { // Recurse for reachable blocks from inner loop
							this->EmitSPARKAdaForMultipleLoopExitTargets(SPARKBodyFile, InnerLoopNum, TCFGBlockNum);
						}
						int ResumeTCFGBlockNum;
						(void) this->GetLoopFollowBlockNum(InnerLoopNum, ResumeTCFGBlockNum);
						pair<int, int> BlockItem(TCFGBlockNum, ResumeTCFGBlockNum);
						pair<int, pair<int, int> > WorkListItem(InnerLoopNum, BlockItem);
						this->SPARKLoopWorkList.push_back(WorkListItem);
						this->CleanUpSPARKLoopWorkList();

						// Reset the bits in the bitsets that are within the inner loop.
						for (size_t TCFGBlockIndex = 0; TCFGBlockIndex < this->GetNumTCFGBlocks(); ++TCFGBlockIndex) {
							if (CurrBitSet.GetBit(TCFGBlockIndex) || ReachableUnion.GetBit(TCFGBlockIndex)) {
								STARSTransCFGBlock *TCFGBlock = this->GetTCFGBlockByNum(TCFGBlockIndex);
								if (TCFGBlock->IsBlockInLoop(InnerLoopNum)) {
									CurrBitSet.ResetBit(TCFGBlockIndex);
									ReachableUnion.ResetBit(TCFGBlockIndex);
								}
							}
						}
					}
				}
				if (CurrBitSet.GetBit(TCFGBlockNum) && CurrTCFGBlock->IsReadyToTranslate()) {
					int ExitTargetTCFGBlockNum = CurrReachablePair.first;
					++ReachableSourceCount;
					string GuardBoolString = this->GenerateSPARKLoopExitBooleanName(LoopNum, ExitTargetTCFGBlockNum);
					if (!StartedGuardOutput) {
						PrintSPARKIndentTabs(SPARKBodyFile);
						SMP_fprintf(SPARKBodyFile, "-- Restructured CFG for multiple loop exit targets.\n");
						PrintSPARKIndentTabs(SPARKBodyFile);
						SMP_fprintf(SPARKBodyFile, "if (%s", GuardBoolString.c_str());
						StartedGuardOutput = true;
					}
					else {
						SMP_fprintf(SPARKBodyFile, " or %s", GuardBoolString.c_str());
					}
				}
			} // end for each CurrReachablePair
			if (StartedGuardOutput) {
				SMP_fprintf(SPARKBodyFile, ") then\n");
				++STARS_SPARK_IndentCount;
				this->EmitSPARKAdaForBlock((int)TCFGBlockNum, HeaderTCFGBlockNum, CurrLoopFollowTCFGBlockNum, SPARKBodyFile, false, false, ReachableSourceCount);
				--STARS_SPARK_IndentCount;
				PrintSPARKIndentTabs(SPARKBodyFile);
				SMP_fprintf(SPARKBodyFile, "end if;\n");
			}
		} // end for all reachability sets
	} // end for all blocks in range

	return;
} // end of SMPFunction::EmitSPARKAdaForMultipleLoopExitTargets()

// recursive descent translation of switch statement starting with INDIR_JUMP block, stop before Follow Block
void SMPFunction::EmitSPARKAdaForSwitch(int HeaderTCFGBlockNum, int FollowTCFGBlockNum, FILE *SPARKBodyFile) {
	assert(0 <= HeaderTCFGBlockNum);
	STARSTransCFGBlock *HeaderTCFGBlock = this->GetTCFGBlockByNum((size_t) HeaderTCFGBlockNum);
	this->SPARKControlStack.push_back(SPARK_SWITCH);

	size_t SwitchIndex;
	SMPInstr *LastInst;
	STARS_ea_t LastAddr;
	if (HeaderTCFGBlock->IsSwitchDefaultCase()) {
		// Encountered default case first, but we want to emit Ada code for it last.
		STARS_ea_t DefaultCaseAddr = HeaderTCFGBlock->GetFirstAddr();
		SwitchIndex = this->FindSwitchIndexForDefaultCaseAddr(DefaultCaseAddr);
		// Switch the HeaderBlock to the INDIR_JUMP block.
		HeaderTCFGBlockNum = this->SwitchInfoArray[SwitchIndex].IndirJumpBlockNum;
		HeaderTCFGBlock = this->GetTCFGBlockByNum(HeaderTCFGBlockNum);
		assert(nullptr != HeaderTCFGBlock);
		LastInst = (*(--(HeaderTCFGBlock->GetLastInst())));
		LastAddr = LastInst->GetAddr();
	}
	else {
		vector<SMPInstr *>::iterator LastInstIter = --(HeaderTCFGBlock->GetLastInst());
		LastInst = (*LastInstIter);

		LastAddr = LastInst->GetAddr();
		SMPitype FlowType = LastInst->GetDataFlowType();
		assert(FlowType == INDIR_JUMP);
		map<STARS_ea_t, size_t>::iterator MapIter = this->SwitchJumpMap.find(LastAddr);
		if (MapIter != this->SwitchJumpMap.end()) {
			SwitchIndex = MapIter->second;
		}
		else {
			SMP_msg("FATAL ERROR: Failed to find INDIR_JUMP addr %llx in SwitchJumpMap ", (uint64_t) LastAddr);
			this->DumpFuncNameAndAddr();
			SMP_fprintf(SPARKBodyFile, "ERROR in switch statement translation at %llx.\n", (uint64_t) LastAddr);
			assert(false);
		}
	}
	assert(SwitchIndex < this->SwitchInfoArray.size());

	struct SwitchTableInfo TableInfo = this->SwitchInfoArray[SwitchIndex];
	int NextFollowTCFGBlockNum = TableInfo.FollowNodeNum;
	assert(NextFollowTCFGBlockNum == FollowTCFGBlockNum);
	// Emit Ada case statement.
	STARS_ea_t FirstInstAddr = HeaderTCFGBlock->GetFirstNonMarkerAddr();
	SMPBasicBlock *OrigHeaderBlock = HeaderTCFGBlock->GetOriginalBlock();
	if (FirstInstAddr != LastAddr) { // Might have untranslated insts before INDIR_JUMP
		SMPInstr *FirstInst = this->GetInstFromAddr(FirstInstAddr);
		if (!FirstInst->HasBeenTranslatedToSPARK()) {
			OrigHeaderBlock->EmitSPARKAdaForAllInsts(SPARKBodyFile);
		}
		else {
			LastInst->EmitSPARKAda(SPARKBodyFile, false); // INDIR_JUMP is handled here
		}
	}
	else {
		LastInst->EmitSPARKAda(SPARKBodyFile, false); // INDIR_JUMP is handled here
	}
	OrigHeaderBlock->SetProcessed(true);
	// Loop through jump table, emitting code for each case.
	assert(!TableInfo.CaseBlockNums.empty());
	for (size_t CaseIndex = 0; CaseIndex < TableInfo.CaseBlockNums.size(); ++CaseIndex) {
		// Skip the default case if jump table entries to it are encountered.
		int CaseTCFGBlockNum = TableInfo.CaseBlockNums[CaseIndex];
		assert(0 <= CaseTCFGBlockNum);
		if (CaseTCFGBlockNum == TableInfo.DefaultCaseBlockNum) {
			continue;  // we emit the default at the end of the Ada case statement as "others =>"
		}

		// Ada cases look like:
		//  when 3 =>
		// or
		//  when 3 | 4 | 5 =>
		PrintSPARKIndentTabs(SPARKBodyFile);
		SMP_fprintf(SPARKBodyFile, "when ");
		assert(!TableInfo.IndexValue[CaseIndex].empty());
		size_t LastValueIndex = TableInfo.IndexValue[CaseIndex].size() - 1;
		for (size_t ValueIndex = 0; ValueIndex <= LastValueIndex; ++ValueIndex) {
			int Value = TableInfo.IndexValue[CaseIndex][ValueIndex];
			SMP_fprintf(SPARKBodyFile, "%d ", Value);
			if (ValueIndex < LastValueIndex) {
				SMP_fprintf(SPARKBodyFile, "| "); // OR operator before next value
			}
			else {
				SMP_fprintf(SPARKBodyFile, "=>\n");
				++STARS_SPARK_IndentCount;
			}
		} // end for all values in current case
		// Now, translate the code inside the case.
		this->EmitSPARKAdaForBlock(CaseTCFGBlockNum, HeaderTCFGBlockNum, NextFollowTCFGBlockNum, SPARKBodyFile, false, false);
		--STARS_SPARK_IndentCount;
	} // end for all cases
	// Emit the default case.
	if (SMP_BLOCKNUM_UNINIT != TableInfo.DefaultCaseBlockNum) { // we have a default case
		PrintSPARKIndentTabs(SPARKBodyFile);
		SMP_fprintf(SPARKBodyFile, "others =>\n");
		++STARS_SPARK_IndentCount;
		int DefaultTCFGIDom = this->TCFGIDom[TableInfo.DefaultCaseBlockNum];
		assert(0 <= DefaultTCFGIDom);
		this->EmitSPARKAdaForBlock(TableInfo.DefaultCaseBlockNum, DefaultTCFGIDom, NextFollowTCFGBlockNum, SPARKBodyFile, true, false);
		--STARS_SPARK_IndentCount;
	}
	// Emit the "end case;" statement
	--STARS_SPARK_IndentCount;
	PrintSPARKIndentTabs(SPARKBodyFile);
	SMP_fprintf(SPARKBodyFile, "end case;\n");

	this->SPARKControlStack.pop_back();
	// After translating the switch statement, we must be ready to resume at the switch follow node in our caller.
	return;
} // end of SMPFunction::EmitSPARKAdaForSwitch()

// recursive descent translation of if-else statement starting with COND_BRANCH block, stop before Follow Block
void SMPFunction::EmitSPARKAdaForConditional(int HeaderTCFGBlockNum, int FollowTCFGBlockNum, FILE *SPARKBodyFile) {
	assert(0 <= HeaderTCFGBlockNum);
	STARSTransCFGBlock *HeaderTCFGBlock = this->GetTCFGBlockByNum((size_t) HeaderTCFGBlockNum);
	int OrigHeaderBlockNum = HeaderTCFGBlock->GetOriginalBlockNum();
	int OrigFallThroughBlockNum = SMP_BLOCKNUM_UNINIT;
	int OrigDistantBlockNum = SMP_BLOCKNUM_UNINIT;
	SMPInstr *LastInst = (*(HeaderTCFGBlock->GetRevInstCBegin()));
	STARS_ea_t LastAddr = LastInst->GetAddr();

	// We want to avoid exiting a loop, which can happen when the FollowBlockNum is -1 or -2
	//  so that we don't have a definite termination point.
	bool UnknownFollowBlock = (0 > FollowTCFGBlockNum);
	bool InsideLoop = this->IsSPARKLoopInTranslationStack();

	// Record our incoming conditional nesting level.
	unsigned short IncomingNestingDepth = this->SPARKConditionalNestingDepth;
	unsigned short IncomingTabDepth = STARS_SPARK_IndentCount;

	// Increment conditional nesting depth. We decrement it whenever we emit "end if;".
	++this->SPARKConditionalNestingDepth;

	// We need to detect the case in which we are beginning a guarded loop, i.e.
	//  if cond then loop ... end loop; end if;
	// For guarded loops, we save the COND_BRANCH location on the loop work list
	//  and return. If the SPARKControlStack already has a LOOP entry, or if
	//  we are not translating loop instructions, we just emit a call to the loop procedure
	//  (which will be emitted later) and save the COND_BRANCH on the work list.
	//  Otherwise, we actually translate the instructions.
	map<STARS_ea_t, STARS_ea_t>::const_iterator GuardIter = this->GuardToLoopMap.find(LastAddr);
	bool SaveAndReturn = false;
	bool TranslatingGuardedLoop = (GuardIter != this->GuardToLoopMap.cend());

	if (TranslatingGuardedLoop) {
		// COND_BRANCH instruction is a guard for a loop.
		SaveAndReturn = (!this->TranslatingLoopToProc()) || this->IsSPARKLoopInTranslationStack();
		// Even if we are translating a loop to a procedure, nested loops
		//  do not get translated until they come to the head of the work list.
		//  If IsSPARKLoopInTranslationStack() returns true, we are inside an outer
		//  loop, the guarded loop is an inner loop and is translated later.

		// In SMPFunction::FindGuardedLoops(), we had two cases. The conditional branch
		//  that guards the loop could be separated from the loop header by a block of
		//  straight-line (fallthrough-only) code, or not. We need to detect this case.

		STARSTransCFGBlock *FallThroughBlock = *(HeaderTCFGBlock->GetFallThroughSucc());
		int FallThroughTCFGBlockNum = FallThroughBlock->GetNumber();
		bool InvertedCase = LastInst->IsOddIfThenCase();
		if (InvertedCase) {
			// Guard jumps to the THEN case.
			assert(FallThroughTCFGBlockNum == FollowTCFGBlockNum);
			FallThroughBlock = *(HeaderTCFGBlock->GetCondNonFallThroughSucc()); // reset so FTBlock is loop
			FallThroughTCFGBlockNum = FallThroughBlock->GetNumber();
		}
		STARS_ea_t LoopAddr = GuardIter->second;
		bool SeparatedCase = (!FallThroughBlock->IsLoopHeaderBlock()); // FTBlock falls through to loop header in SeparatedCase
		SMPBasicBlock *OrigLoopHeadBlock = this->GetBlockFromInstAddr(LoopAddr);
		SMPBasicBlock *OrigFallThroughBlock = FallThroughBlock->GetOriginalBlock();
		int LoopHeadTCFGBlockNum = FallThroughTCFGBlockNum;
		if (SeparatedCase) {
			LoopHeadTCFGBlockNum = FallThroughBlock->FindChildLoopHeaderTCFGBlockNum();
			assert(0 < LoopHeadTCFGBlockNum);
		}

		LastInst->EmitSPARKAda(SPARKBodyFile, false); // emit if (inverted condition) then

		if (SaveAndReturn) {
			this->SPARKControlStack.push_back(SPARK_THEN_CLAUSE);
			if (SeparatedCase) {
				// Emit code for fall-through instructions block.
				OrigFallThroughBlock->EmitSPARKAdaForAllInsts(SPARKBodyFile);
			}

			int OrigCurrBlockNum = OrigLoopHeadBlock->GetNumber();
			int LoopNum = this->GetLoopNumFromHeaderBlockNum(OrigCurrBlockNum);
			assert(0 <= LoopNum);
			bool MultipleExitTargets = this->EmitSPARKAdaLoopCall(LoopAddr, (size_t) LoopNum, SPARKBodyFile);
			// Put the loop header and loop follow block numbers onto the work list
			int LoopFollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
			bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t)LoopNum, LoopFollowTCFGBlockNum);
			bool LoopFollowBlocksAreReturnBlocks = (SMP_BLOCKNUM_COMMON_RETURN == LoopFollowTCFGBlockNum);
			if (MultiExitLoop && (0 < LoopFollowTCFGBlockNum)) {
				// Could still be a return block, and multiexit targets could converge on it.
				STARSTransCFGBlock *LoopFollowTCFGBlock = this->GetTCFGBlockByNum((size_t)LoopFollowTCFGBlockNum);
				LoopFollowBlocksAreReturnBlocks = LoopFollowTCFGBlock->GetOriginalBlock()->HasReturn();
			}
			assert((0 < LoopFollowTCFGBlockNum) || (LoopFollowBlocksAreReturnBlocks && MultiExitLoop));
			pair<int, int> BlockItem(LoopHeadTCFGBlockNum, LoopFollowTCFGBlockNum);
			pair<int, pair<int, int> > WorkListItem(LoopNum, BlockItem);
			this->SPARKLoopWorkList.push_back(WorkListItem);
			this->CleanUpSPARKLoopWorkList();

			// See if we have a loop follow block number that is inside this conditional
			//  and recurse into it if so. This is for the case in which the loop is entirely
			//  contained within the conditional. We need to NOT recurse into the loop follow block
			//  in the case where the conditional is entirely contained within a loop; in that case,
			//  wait until we return to EmitSPARKAdaForLoop() to go into the loop follow node.
			if (0 < LoopFollowTCFGBlockNum) {
				STARSTransCFGBlock *LoopTCFGFollowBlock = this->GetTCFGBlockByNum((size_t)LoopFollowTCFGBlockNum);
				this->TCFGDomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
				if (LoopTCFGFollowBlock->IsReadyToTranslate() && this->DoesTCFGBlockDominateTCFGBlock(HeaderTCFGBlockNum, LoopFollowTCFGBlockNum)) {
					this->EmitSPARKAdaForBlock(LoopFollowTCFGBlockNum, HeaderTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile, false, false);
				}
			}

			--STARS_SPARK_IndentCount;
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "end if;\n");
			this->SPARKControlStack.pop_back();
			--this->SPARKConditionalNestingDepth;
			assert(IncomingNestingDepth == this->SPARKConditionalNestingDepth);
			assert(IncomingTabDepth == STARS_SPARK_IndentCount);

			// Return so that our caller can resume after the if-then-loop-endloop-endif.
			return;
		}
	}

	LastInst->EmitSPARKAda(SPARKBodyFile, false); // emit if (inverted condition) then

	// We have already emitted the "if (inverted condition) then" for the COND_BRANCH instruction.
	//  For the if-then case, we emit the code for the then-block and then "end if;" and return.
	//  For the if-then-else case, we emit the code for the then-block and then "else" etc.
	ControlFlowType LastCFType = this->GetControlFlowType(LastAddr);
	bool IsLoopBack = ((LastCFType == LOOP_BACK) || (LastCFType == INVERTED_LOOP_BACK));
	bool LoopBackStaysInLoop = this->IsNonExitingLoopBackBranch(LastAddr);
	bool IfThenCase = ((LastCFType == BRANCH_IF_THEN) || (LoopBackStaysInLoop && (IsLoopBack || (LastCFType == LOOP_EXIT))));
	bool ShortCircuitCase = (LastCFType == SHORT_CIRCUIT_BRANCH);
	assert(IfThenCase || ShortCircuitCase || (LastCFType == BRANCH_IF_THEN_ELSE));
	int FallThroughTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
	int DistantTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
	list<STARSTransCFGBlock *>::const_iterator SuccIter = HeaderTCFGBlock->GetCondNonFallThroughSucc();
	if (!ShortCircuitCase) {
		assert(SuccIter != HeaderTCFGBlock->GetLastConstSucc());
		DistantTCFGBlockNum = (*SuccIter)->GetNumber();
	}
	// Find the fall-through and non-fall-through block numbers.
	if (ShortCircuitCase) {
		STARSCFGBlock *HeaderCFGBlock = this->GetClonedCFGBlockByNum((size_t)HeaderTCFGBlockNum);
		assert(nullptr != HeaderCFGBlock);
		STARSCondExpr *HeaderExpr = HeaderCFGBlock->GetExpr();
		assert(nullptr != HeaderExpr);
		FallThroughTCFGBlockNum = HeaderExpr->GetFallThroughBlockNum();
		DistantTCFGBlockNum = HeaderExpr->GetNonFallThroughBlockNum();
		IfThenCase = (DistantTCFGBlockNum == FollowTCFGBlockNum); // treat ShortCircuitCase as IfThenCase from now on
	}
	else {
		assert(2 == HeaderTCFGBlock->GetNumSuccessors());
		SuccIter = HeaderTCFGBlock->GetFallThroughSucc();
		FallThroughTCFGBlockNum = (*SuccIter)->GetNumber();
		if (LastCFType == INVERTED_LOOP_BACK) {
			// Swap the fall-through and distant blocks.
			int TempBlockNum = DistantTCFGBlockNum;
			DistantTCFGBlockNum = FallThroughTCFGBlockNum;
			FallThroughTCFGBlockNum = TempBlockNum;
		}
	}
	STARSTransCFGBlock *ThenTCFGBlock = this->TCFGBlocks[(size_t) FallThroughTCFGBlockNum];
	STARSTransCFGBlock *DistantTCFGBlock = this->TCFGBlocks[(size_t)DistantTCFGBlockNum];
	OrigFallThroughBlockNum = ThenTCFGBlock->GetOriginalBlockNum();
	OrigDistantBlockNum = DistantTCFGBlock->GetOriginalBlockNum();
	bool CheckFTBlock = false; // check for falling out of loop?
	bool CheckDistantBlock = false; // check for falling out of loop?
	bool FallThroughLoopExit = this->DoesEdgeExitLoop(OrigHeaderBlockNum, OrigFallThroughBlockNum);
	bool DistantLoopExit = this->DoesEdgeExitLoop(OrigHeaderBlockNum, OrigDistantBlockNum);
	if (IfThenCase) {
		// The common case is to jump around the ThenBlock, which can then fall-through to the FollowBlock:
		//  if (cond) then goto L1;
		//   then-block
		//  L1: 
		// However, it is also possible to jump to the ThenBlock, which then jumps or falls through to the FollowBlock:
		// if (cond) then goto L2;
		// L1: 
		//      somewhere else:  L2: then-block
		//                           goto L1
		// This could be odd code generation or hand-written code.
		// In this odd case, the FallThroughBlock has more than 1 predecessor. In the normal case, the
		//  FallThroughBlock has only one predecessor (the COND_BRANCH block falls through to it). We can use
		//  the number of predecessor blocks to distinguish the two cases.
		// NOTE: The ShortCircuitCase code earlier assumes that this odd case cannot be combined with the ShortCircuitCase.
		this->SPARKControlStack.push_back(SPARK_THEN_CLAUSE);
		// AnalyzeCompoundConditionalStatements() undoes the convolution of OddIfThenElse cases, so
		//  there is no such thing as an OddIfThenCase for a SHORT_CIRCUIT_BRANCH when it comes time
		//  to generate the SPARK Ada translation.
		bool OddIfThenCase = (!ShortCircuitCase) && LastInst->IsOddIfThenCase();
		if (!OddIfThenCase) { // normal case
			assert(LoopBackStaysInLoop || (DistantTCFGBlockNum == FollowTCFGBlockNum)); // non-fall-through block is normal follow block for if-then
			// Avoid generating code outside the loop from THEN clause if unknown follow block.
			if (InsideLoop && UnknownFollowBlock && FallThroughLoopExit) {
				PrintSPARKIndentTabs(SPARKBodyFile);
				SMP_fprintf(SPARKBodyFile, "exit;\n");
			}
			else {
				this->EmitSPARKAdaForBlock(FallThroughTCFGBlockNum, FallThroughTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile, false, (TranslatingGuardedLoop && !SaveAndReturn));
			}
		}
		else {
			// Need to generate the code for the ThenBlock == DistantBlock first.
			// Avoid generating code outside the loop from THEN clause if unknown follow block.
			if (InsideLoop && UnknownFollowBlock && DistantLoopExit) {
				PrintSPARKIndentTabs(SPARKBodyFile);
				SMP_fprintf(SPARKBodyFile, "exit;\n");
			}
			else {
				this->EmitSPARKAdaForBlock(DistantTCFGBlockNum, DistantTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile, false, (TranslatingGuardedLoop && !SaveAndReturn));
			}
		}
		this->SPARKControlStack.pop_back();
	}
	else { // !IfThenCase ; must be if-then-else
		STARSTransCFGBlock *ElseTCFGBlock = DistantTCFGBlock;
		// Avoid generating code outside the loop from THEN clause if unknown follow block.
		if (InsideLoop && UnknownFollowBlock && FallThroughLoopExit) {
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "exit;\n");
		}
		else {
			this->SPARKControlStack.push_back(SPARK_THEN_CLAUSE);
			if (UnknownFollowBlock && (!ThenTCFGBlock->IsReadyToTranslate())) {
				PrintSPARKIndentTabs(SPARKBodyFile);
				SMP_fprintf(SPARKBodyFile, "; -- block already visited \n"); // should not happen with proper structural analysis
			}
			else {
				this->EmitSPARKAdaForBlock(FallThroughTCFGBlockNum, FallThroughTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile, false, false);
			}
			this->SPARKControlStack.pop_back();
			CheckFTBlock = (InsideLoop && UnknownFollowBlock);
		}
		--STARS_SPARK_IndentCount;
		PrintSPARKIndentTabs(SPARKBodyFile);
		SMP_fprintf(SPARKBodyFile, "else\n");
		++STARS_SPARK_IndentCount;
		// Avoid generating code outside the loop from ELSE clause if unknown follow block.
		if (InsideLoop && UnknownFollowBlock && DistantLoopExit) {
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "exit;\n");
		}
		else {
			this->SPARKControlStack.push_back(SPARK_ELSE_CLAUSE);
			if (UnknownFollowBlock && (!ElseTCFGBlock->IsReadyToTranslate())) {
				PrintSPARKIndentTabs(SPARKBodyFile);
				SMP_fprintf(SPARKBodyFile, "; -- block already visited \n"); // should not happen
			}
			else {
				this->EmitSPARKAdaForBlock(DistantTCFGBlockNum, DistantTCFGBlockNum, FollowTCFGBlockNum, SPARKBodyFile, false, false);
			}
			this->SPARKControlStack.pop_back();
			CheckDistantBlock = (InsideLoop && UnknownFollowBlock);
		}
	}

	// We need to detect the odd case of falling through to a loop header with INVERTED_LOOP_EXIT,
	//  i.e. we are doing a LOOP_BACK of sorts by falling out of the conditional,
	//  in which case the "end if;" already was emitted before "end loop;" inside EmitSPARKAdaForBlock().
	bool InvertedLoopExitCase = false;
	if (!UnknownFollowBlock) {
		STARSTransCFGBlock *FollowTCFGBlock = this->GetTCFGBlockByNum((size_t)FollowTCFGBlockNum);
		assert(nullptr != FollowTCFGBlock);
		// FollowTCFGBlock will be loop header that includes conditional in InvertedLoopExitCase
		ControlFlowType FollowBlockCFType = FollowTCFGBlock->GetLastInstCFType();
		this->TCFGDomSoFar.ResetAllBits(); // Prepare for DomTree search, memoize dominated blocks seen so far.
		InvertedLoopExitCase = (IsInvertedLoopExitFlow(FollowBlockCFType) && this->DoesTCFGBlockDominateTCFGBlock(FollowTCFGBlockNum, FallThroughTCFGBlockNum));
	}
#if 1
	else if (InsideLoop) {
		this->ResetSCCPVisitedBlocks(); // prune search in FindLastBlockProcessedInChain()
		int LastBlockEmittedNum = this->FindLastBlockProcessedInChain(OrigHeaderBlockNum);
		SMPBasicBlock *LastBlockEmitted = this->GetBlockByNum((size_t) LastBlockEmittedNum);
		assert(nullptr != LastBlockEmitted);
		ControlFlowType LastCFType = LastBlockEmitted->GetLastInstCFType();
		list<SMPBasicBlock *>::const_iterator FTSuccIter = LastBlockEmitted->GetFallThroughSucc();
		if (FTSuccIter != LastBlockEmitted->GetLastConstSucc()) {
			// Ended translation with a fall-through. Was it falling through to the top of the loop?
			if (LastBlockEmittedNum > (*FTSuccIter)->GetNumber()) { // back edge
				InvertedLoopExitCase = true;
			}
		}
	}
#else
	else if (CheckDistantBlock) {
		SMPBasicBlock *DistantBlock = this->GetBlockByNum((size_t)DistantBlockNum); // will be loop header that includes conditional in InvertedLoopExitCase
		ControlFlowType DistantBlockCFType = DistantBlock->GetLastInstCFType();
		if (INVERTED_LOOP_EXIT == DistantBlockCFType) {
			SMPBasicBlock *LoopHeadBlock = *(DistantBlock->GetFallThroughSucc());
			assert(nullptr != LoopHeadBlock);
			assert(LoopHeadBlock->IsLoopHeaderBlock());
			InvertedLoopExitCase = (this->DoesBlockDominateBlock(LoopHeadBlock->GetNumber(), DistantBlockNum));
		}
	}
	if (!InvertedLoopExitCase && CheckFTBlock) {
		SMPBasicBlock *FTBlock = this->GetBlockByNum((size_t)FallThroughBlockNum); // will be loop header that includes conditional in InvertedLoopExitCase
		ControlFlowType FTBlockCFType = FTBlock->GetLastInstCFType();
		if (INVERTED_LOOP_EXIT == FTBlockCFType) {
			SMPBasicBlock *LoopHeadBlock = *(FTBlock->GetCondNonFallThroughSucc());
			assert(nullptr != LoopHeadBlock);
			assert(LoopHeadBlock->IsLoopHeaderBlock());
			InvertedLoopExitCase = (this->DoesBlockDominateBlock(LoopHeadBlock->GetNumber(), FallThroughBlockNum));
		}
	}
#endif
	// Now we just need to emit the "end if;"
	bool EndifNotYetEmitted = (IncomingNestingDepth < this->SPARKConditionalNestingDepth);

	if (!InvertedLoopExitCase) {
		if (EndifNotYetEmitted) {
			--STARS_SPARK_IndentCount;
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "end if;\n");
			--this->SPARKConditionalNestingDepth;
		}
		assert(IncomingNestingDepth == this->SPARKConditionalNestingDepth);
		assert(IncomingTabDepth == STARS_SPARK_IndentCount);
	}
	else {
		if (EndifNotYetEmitted) {
			--STARS_SPARK_IndentCount;
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "end if;\n");
			--this->SPARKConditionalNestingDepth;
		}
		else {
			SMP_msg("INFO: SPARK: Not emitting end if for guarded conditional at %llx due to InvertedLoopExitCase\n",
				(uint64_t)LastAddr);
		}
		assert(IncomingNestingDepth == this->SPARKConditionalNestingDepth);
		assert(IncomingTabDepth == STARS_SPARK_IndentCount);
		if (UnknownFollowBlock) { // odd case of falling into header block loop-back
#if STARS_SPARK_CENTRALIZE_EMIT_LOOP
			;
#else
			--STARS_SPARK_IndentCount;
			PrintSPARKIndentTabs(SPARKBodyFile);
			SMP_fprintf(SPARKBodyFile, "end loop;\n\n");
#endif
		}
	}

	return;
} // end of SMPFunction::EmitSPARKAdaForConditional()

// emit call to loop proc that will be created later starting at LoopAddr
bool SMPFunction::EmitSPARKAdaLoopCall(STARS_ea_t LoopAddr, size_t LoopIndex, FILE *SPARKBodyFile) {
	// Generate procedure name for the loop being converted into a procedure.
	// If loop starts at 0x8048ca0, we want ZSTLoopProc_8048ca0 as the proc name.
	std::string ProcName("ZSTLoopProc_");
	std::ostringstream AddrString;
	AddrString << std::hex << LoopAddr;
	ProcName += AddrString.str();
	// Emit a call to the proc name, e.g. "ZSTLoopProc_8048ca0;"
	SMP_fprintf(SPARKBodyFile, "\n");
	PrintSPARKIndentTabs(SPARKBodyFile);
	if (this->LoopMemRangeInArgRegsBitmap[LoopIndex].any()) {
		SMP_fprintf(SPARKBodyFile, "%s(", ProcName.c_str());
		this->EmitSPARKArgs(SPARKBodyFile, nullptr, false, LoopIndex);
		SMP_fprintf(SPARKBodyFile, ");\n");
	}
	else {
		SMP_fprintf(SPARKBodyFile, "%s;\n", ProcName.c_str());
	}

	bool MultipleExitTargets = (this->LoopExitTargets[LoopIndex].size() > 1);
	return MultipleExitTargets;
} // end of SMPFunction::EmitSPARKAdaLoopCall()

// Based on control flow structure of CurrBlock, find Ada follow block num; -1 if no structure besides fall-through
int SMPFunction::FindFollowBlockNum(STARSTransCFGBlock *CurrTCFGBlock, bool StartAtLastInst) {
	int FollowTCFGBlockNum = SMP_BLOCKNUM_UNINIT;
	if (!StartAtLastInst) { // StartAtLastInst => Key is to examine branch or jump, not starting a new loop or switch default case
		if (CurrTCFGBlock->IsLoopHeaderBlock()) {
			int LoopNum = this->GetLoopNumFromHeaderBlockNum(CurrTCFGBlock->GetOriginalBlockNum());
			assert(0 <= LoopNum);
			bool LoopMightBeInfinite = (STARS_INFINITE_OR_MIDDLE_TESTING_LOOP == this->LoopTypesByLoopNum[LoopNum]);
			bool MultiExitLoop = this->GetLoopFollowBlockNum((size_t) LoopNum, FollowTCFGBlockNum);
			// assert((0 <= FollowTCFGBlockNum) || LoopMightBeInfinite);
		}
		else if (CurrTCFGBlock->IsSwitchDefaultCase()) {
			STARS_ea_t DefaultCaseAddr = CurrTCFGBlock->GetFirstAddr();
			size_t SwitchIndex = this->FindSwitchIndexForDefaultCaseAddr(DefaultCaseAddr);
			FollowTCFGBlockNum = this->SwitchInfoArray[SwitchIndex].FollowNodeNum;
		}
	}
	else {
		vector<SMPInstr *>::iterator LastInstIter = --(CurrTCFGBlock->GetLastInst());
		SMPInstr *LastInst = (*LastInstIter);
		STARS_ea_t LastAddr = LastInst->GetAddr();
		SMPitype FlowType = LastInst->GetDataFlowType();
		int CurrTCFGBlockNum = CurrTCFGBlock->GetNumber();
		if (FlowType == INDIR_JUMP) { // switch statement
			map<STARS_ea_t, size_t>::iterator MapIter = this->SwitchJumpMap.find(LastAddr);
			if (MapIter != this->SwitchJumpMap.end()) {
				size_t SwitchIndex = MapIter->second;
				assert(SwitchIndex < this->SwitchInfoArray.size());
				FollowTCFGBlockNum = this->SwitchInfoArray[SwitchIndex].FollowNodeNum;
			}
			else { // leave error code of SMP_BLOCKNUM_UNINIT in FollowBlockNum
				SMP_msg("ERROR: Failed to find INDIR_JUMP addr %llx in SwitchJumpMap ", (uint64_t) LastAddr);
				this->DumpFuncNameAndAddr();
			}
		}
		else if (FlowType == COND_BRANCH) {
			ControlFlowType LastCFType = GetControlFlowType(LastAddr);
			if (IsLoopExitFlow(LastCFType)) {
				// Middle exit from loop. Follow block is fall-through.
				list<STARSTransCFGBlock *>::const_iterator SuccIter;
				if (!IsInvertedLoopExitFlow(LastCFType)) {
					SuccIter = CurrTCFGBlock->GetFallThroughSucc();
					assert(SuccIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH must have fall through and non-fall-through
					FollowTCFGBlockNum = (*SuccIter)->GetNumber();
				}
				else if (INVERTED_LOOP_EXIT == LastCFType) { // must take branch to stay in the loop for INVERTED_LOOP_EXIT or SHORT_CIRCUIT_INVERTED_LOOP_EXIT
					SuccIter = CurrTCFGBlock->GetCondNonFallThroughSucc();
					assert(SuccIter != CurrTCFGBlock->GetLastConstSucc()); // COND_BRANCH must have fall through and non-fall-through
					FollowTCFGBlockNum = (*SuccIter)->GetNumber();
				}
			}
			else {
				// Handle if-else jumps here.
				assert((LastCFType == BRANCH_IF_THEN) || (LastCFType == BRANCH_IF_THEN_ELSE) || (LastCFType == SHORT_CIRCUIT_BRANCH));
				map<int, int>::iterator MapIter = this->JumpFollowNodesMap.find(CurrTCFGBlockNum);
				if (MapIter != this->JumpFollowNodesMap.end()) { // found it
					FollowTCFGBlockNum = MapIter->second;
				}
				else {
					SMP_msg("ERROR: Failed to find COND_BRANCH entry for addr %llx TCFG block %d in JumpFollowNodesMap ",
						(uint64_t) LastAddr, CurrTCFGBlockNum);
					this->DumpFuncNameAndAddr();
				}
			}
		}
	}
	return FollowTCFGBlockNum;
} // end of SMPFunction::FindFollowBlockNum()

// return true if loop has multiple follow blocks
bool SMPFunction::GetLoopFollowBlockNum(const size_t LoopNum, int &FollowBlockNum) const {
	assert(LoopNum < this->LoopExitTargets.size());
	bool MultiExitTargetLoop = false;
	if (this->LoopExitTargets[LoopNum].empty()) {
		FollowBlockNum = SMP_BLOCKNUM_COMMON_RETURN;
	}
	else {
		FollowBlockNum = SMP_BLOCKNUM_UNINIT;
		MultiExitTargetLoop = (1 < this->LoopExitTargets[LoopNum].size());
		if (MultiExitTargetLoop) { // Multiple exit targets; their convergence point is the follow block
			map<int, set<int> >::const_iterator MapIter = this->MultiLoopExitTargetsConvergenceMap.find((int)LoopNum);
			if (MapIter != this->MultiLoopExitTargetsConvergenceMap.cend()) {
				if (1 == MapIter->second.size()) {
					FollowBlockNum = *(MapIter->second.cbegin());
				}
			}
		}
		else { // Only one exit target; it is the follow block num
			FollowBlockNum = *(this->LoopExitTargets[LoopNum].cbegin());
		}
	}

	return MultiExitTargetLoop;
} // end of SMPFunction::GetLoopFollowBlockNum()

// return true if loop has multiple exit target blocks
bool SMPFunction::DoesLoopHaveMultipleExitTargets(const size_t LoopNum) const {
	assert(LoopNum < this->LoopExitTargets.size());
	bool MultiExitTargetLoop = (1 < this->LoopExitTargets[LoopNum].size());

	return MultiExitTargetLoop;
} // end of SMPFunction::DoesLoopHaveMultipleExitTargets()

// found in LoopExitTargets[LoopNum] set?
bool SMPFunction::IsBlockLoopExitTarget(const std::size_t LoopNum, const int BlockNum) const {
	assert(LoopNum < this->LoopExitTargets.size());
	return (this->LoopExitTargets[LoopNum].find(BlockNum) != this->LoopExitTargets[LoopNum].cend());
} // end of SMPFunction::IsBlockLoopExitTarget()

// common code for different cases in EmitFuncPtrShadowingAnnotations2()
void SMPFunction::EmitShadowingHelper(FILE *InfoAnnotFile, SMPInstr *CurrInst, bool FuncPtr, STARS_ea_t CallAddr) {
	STARS_ea_t ShadowCheckAddr = CurrInst->GetAddr();
	STARS_ea_t OldShadowCheckAddr = STARS_BADADDR;
	bool GoodCallAddr = (STARS_BADADDR != CallAddr);
	bool InterruptCall = CurrInst->IsInterruptCall();
	bool ImproveCFG = global_STARS_program->ShouldSTARSMaximizeCFGImprovement();
	bool ShadowFuncPtrs = (FuncPtr && global_STARS_program->ShouldSTARSShadowFuncPtrs());
	bool ShadowArgs = !FuncPtr;
	ShadowPoint CriticalOp(ShadowCheckAddr, 0); // Checking point before indirect call
	list<ShadowPoint> WorkList;
	WorkList.push_back(CriticalOp);
	this->AlreadyShadowed.clear();

	ShadowSet ShadowUses;      // set of addr+USE pairs to shadow values that will be checked at ShadowCheckAddr
	ShadowSet NewCriticalOps;  // worklist of new shadow USEs and their addresses
	while (!WorkList.empty()) {
		CriticalOp = WorkList.front();
		WorkList.pop_front();
		ShadowCheckAddr = CriticalOp.first;
		if (ShadowCheckAddr != OldShadowCheckAddr) { // starting a new chain
			OldShadowCheckAddr = ShadowCheckAddr;
		}

		bool UnsafeCodePointerChain = false;
		if (ShadowFuncPtrs)
			++STARS_FuncPtrShadowPointsAttempted;
		bool NonConstSourceFound = false;
		set<STARS_uval_t> ConstValues;
		bool ValidShadowing = this->FindShadowingPoint2(CriticalOp, false, ShadowUses, UnsafeCodePointerChain, NewCriticalOps, NonConstSourceFound, ConstValues);
		if (!NonConstSourceFound && ValidShadowing && ImproveCFG && GoodCallAddr && (!InterruptCall)) {
			SMP_msg("INFO: Only constants found in tracing func ptr at %llx\n", (uint64_t) ShadowCheckAddr);
			FILE *XrefsFile = global_STARS_program->GetXrefsFile();
			for (set<STARS_uval_t>::const_iterator ValIter = ConstValues.cbegin(); ValIter != ConstValues.cend(); ++ValIter) {
				SMP_fprintf(XrefsFile, "%18llx %6d INSTR XREF IBT FROMIB %18llx INDIRCALL\n",
					(uint64_t) (*ValIter), CurrInst->GetSize(), (uint64_t) CallAddr);
			}
			if (!ConstValues.empty()) {
				SMP_fprintf(XrefsFile, "%18llx %6d INSTR XREF FROMIB COMPLETE %6zu INDIRCALL\n",
					(uint64_t) CallAddr, CurrInst->GetSize(), ConstValues.size());
			}
		}
		if (ValidShadowing && UnsafeCodePointerChain) {
			if (ShadowFuncPtrs)
				++STARS_FuncPtrShadowPointsSucceeded;
			if (ShadowArgs || ShadowFuncPtrs) { // not just ImproveCFG
				// Need to shadow ShadowDefs right before their addrs, validate CriticalOp USE at ShadowCheckAddr
				unsigned int CurrentShadowID = global_STARS_program->GetShadowID();
				for (ShadowSet::const_iterator ShadowIter = ShadowUses.cbegin(); ShadowIter != ShadowUses.cend(); ++ShadowIter) {
					STARS_ea_t ShadowAddr = ShadowIter->first;
					assert(STARS_BADADDR != ShadowAddr);
					STARSOpndTypePtr ShadowOp = this->TempShadowList.GetRefNum(ShadowIter->second)->GetOp();
					assert((nullptr != ShadowOp) && (!ShadowOp->IsVoidOp()));
					SMPInstr *ShadowInst = this->GetInstFromAddr(ShadowAddr);
					STARSOpndTypePtr UnnormalizedUseOp = CloneIfNecessary(ShadowOp, this->UsesFramePointer());
					if (ShadowOp->IsMemOp()) {
						// Un-normalize stack ops for annotation printing.
						ShadowInst->MDGetUnnormalizedOp(UnnormalizedUseOp);
					}

					if (ShadowFuncPtrs) {
						// Emit the annotation to shadow the InArg value at the top of the function.
						SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR FPTRSHADOW ",
							(uint64_t) ShadowAddr, ShadowInst->GetSize());
					}
					else if (ShadowArgs) {
						// Emit the annotation to shadow the InArg value at the top of the function.
						SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR ARGSHADOW ",
							(uint64_t) ShadowAddr, ShadowInst->GetSize());
					}
					AnnotPrintOperand(UnnormalizedUseOp, InfoAnnotFile, UseFP, ShadowInst->MDIsAddressing64bit());
					SMP_fprintf(InfoAnnotFile, " SHADOWID %u\n", CurrentShadowID);

					AlreadyShadowed.insert(*ShadowIter);
				} // end for all ShadowUses

				// Emit the annotation to check the shadowed value before the INDIR_CALL.
				if (!ShadowUses.empty()) {
					SMPInstr *ShadowCheckInst = this->GetInstFromAddr(ShadowCheckAddr);
					if (FuncPtr) {
						SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR FPTRCHECK ",
							(uint64_t) ShadowCheckAddr, ShadowCheckInst->GetSize());
					}
					else {
						SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR ARGCHECK ",
							(uint64_t) ShadowCheckAddr, ShadowCheckInst->GetSize());
					}
					STARSOpndTypePtr UseOp = ShadowCheckInst->GetFirstRightOperandNoNorm(); // for annotation printing
					AnnotPrintOperand(UseOp, InfoAnnotFile, UseFP, ShadowCheckInst->MDIsAddressing64bit());
					SMP_fprintf(InfoAnnotFile, " SHADOWID %u\n", CurrentShadowID);

					// Increment the unique shadowing index
					global_STARS_program->IncrementShadowID();
				}
			}
		} // end if (ValidShadowing && UnsafeCodePointerChain)
		else if (ValidShadowing) { // Must be safe from overwriting
			if (FuncPtr)
				++STARS_FuncPtrShadowPointsSafe;
		}
		else {
			if (FuncPtr)
				++STARS_FuncPtrShadowPointsFailed;
		}

		// Now we need to find shadowing points for the NewCriticalOps.
		if (!NewCriticalOps.empty()) {
			for (ShadowSet::const_iterator NewIter = NewCriticalOps.cbegin(); NewIter != NewCriticalOps.cend(); ++NewIter) {
				ShadowPoint CurrPoint = *NewIter;
				if (!this->IsAlreadyShadowed(CurrPoint)) {
					// Only add to WorkList if not already shadowed.
					WorkList.push_back(CurrPoint);
				}
			}
			NewCriticalOps.clear();
		}
		ShadowUses.clear();
	} // end while (!WorkList.empty())
	this->AlreadyShadowed.clear();

	return;
} // end of SMPFunction::EmitShadowingHelper()

// Emit annotations for func ptr shadowing defense and/or for improving CFG related to INDIR_CALLs.
void SMPFunction::EmitFuncPtrShadowingAnnotations2(FILE *InfoAnnotFile) {
	// Keep track of which annotations are requested.
	bool ImproveCFG = global_STARS_program->ShouldSTARSMaximizeCFGImprovement();
	bool ShadowFuncPtrs = global_STARS_program->ShouldSTARSShadowFuncPtrs();
	if (this->HasGoodRTLs() && (ImproveCFG || this->HasUnsafeIndirectWrites) && (!this->IsLeaf())) {
		// Candidate for shadowing function pointers, if any of three cases passes through unsafe blocks.
		for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
			if (CurrBlock->HasCallInstruction()) {
				size_t ArgNum = 0;
				bool Case3Search = false; // have not found CALL yet
				for (vector<SMPInstr *>::reverse_iterator LastInstIter = CurrBlock->GetRevInstBegin(); LastInstIter != CurrBlock->GetRevInstEnd(); ++LastInstIter) {
					SMPInstr *CurrInst = (*LastInstIter);
					// Case 1: InArg CODEPTR reaches INDIR_CALL via simple copies
					// Case 2: Internally initialized CODEPTR reaches INDIR_CALL via simple copies
					if (INDIR_CALL == CurrInst->GetDataFlowType()) {
						Case3Search = false; // Override previous CALL now that we are processing INDIR_CALL
						STARSOpndTypePtr UseOp = CurrInst->GetFirstRightOperand();
						STARSDefUseIter UseIter = CurrInst->FindUse(UseOp);
						assert(UseIter != CurrInst->GetLastUse());
						this->TempShadowList.clear();
						this->TempShadowList.SetRef(UseIter->GetOp(), UseIter->GetType(), UseIter->GetSSANum());

						this->EmitShadowingHelper(InfoAnnotFile, CurrInst, true, CurrInst->GetAddr());
					} // end if INDIR_CALL
					else if (CALL == CurrInst->GetDataFlowType()) {
						// Search for case 3: CODEPTR passed as outgoing arg.
						Case3Search = ShadowFuncPtrs;
					}
					else if (Case3Search && CurrInst->MDIsArgumentPass(ArgNum)) {
						STARSDefUseIter ArgIter = CurrInst->GetFirstNonFlagsDef();
						if ((ArgIter != CurrInst->GetLastDef()) && IsEqType(CODEPTR, ArgIter->GetType())) {
							STARSOpndTypePtr ArgOp = nullptr;
							unsigned short SignMask = 0;
							if (CurrInst->MDIsMoveInstr()) {
								ArgOp = CurrInst->GetMoveSource(); // Need a USE for FindShadowingPoint2()
							}
							else if (CurrInst->MDIsSignedLoad(SignMask)) {
								ArgOp = CurrInst->GetFirstRightOperandNoNorm();
								if (ArgOp->IsRegOp()) {
									CanonicalizeOpnd(ArgOp);
								}
								else if (MDIsStackAccessOpnd(ArgOp, this->UsesFramePointer()) && CurrInst->AreDefsNormalized()) {
									ArgOp = this->GetNormalizedOperand(CurrInst->GetAddr(), ArgOp);
								}
							}
							if ((nullptr != ArgOp) && !ArgOp->IsVoidOp()) {
								STARSDefUseIter UseIter = CurrInst->FindUse(ArgOp);
								assert(UseIter != CurrInst->GetLastUse());
								this->TempShadowList.clear();
								this->TempShadowList.SetRef(ArgOp, UseIter->GetType(), UseIter->GetSSANum());
								this->EmitShadowingHelper(InfoAnnotFile, CurrInst, true, STARS_BADADDR);
							}
						} // end if CODEPTR ArgIter
					} // end if INDIR_CALL elsif CALL elsif Case3 arg pass
				} // end for all insts in reverse order
			} // end if current block has a call inst
		} // end for all blocks

	} // end if good RTLs, unsafe indirect writes, and not a leaf func
	return;
} // end of SMPFunction::EmitFuncPtrShadowingAnnotations2()

// Emit annotations for critical argument shadowing defense
void SMPFunction::EmitArgShadowingAnnotations(FILE *InfoAnnotFile) {
	if (this->HasGoodRTLs() && this->HasUnsafeIndirectWrites && (!this->IsLeaf())) {
		// Candidate for shadowing critical args, if either of two cases passes through unsafe blocks.
		for (size_t BlockIndex = 0; BlockIndex < this->RPOBlocks.size(); ++BlockIndex) {
			SMPBasicBlock *CurrBlock = this->RPOBlocks[BlockIndex];
			if (CurrBlock->HasCallInstruction()) {
				size_t ArgNum = 0;
				bool ArgSearch = false; // have not found CALL yet
				uint32_t CriticalArgPosBits = 0;
				for (vector<SMPInstr *>::reverse_iterator LastInstIter = CurrBlock->GetRevInstBegin(); LastInstIter != CurrBlock->GetRevInstEnd(); ++LastInstIter) {
					SMPInstr *CurrInst = (*LastInstIter);
					// Case 1: CALL to critical library function.
					if (CALL == CurrInst->GetDataFlowType()) {
						// Case 1: Calling critical library function.
						string TaintFuncName = CurrInst->GetTrimmedCalledFunctionName();
						GetTaintWarningArgPositionsForCallName(TaintFuncName, CriticalArgPosBits);
						ArgSearch = (0 < CriticalArgPosBits); // found a critical call if any critical arg
						// Case 2: Callee has critical InArg (eventually passed to critical library function).
						if (!ArgSearch) {
							STARS_ea_t CalleeAddr = CurrInst->GetCallTarget();
							if ((STARS_BADADDR != CalleeAddr) && (!STARS_IsExternalFunc(CalleeAddr))) {
								SMPFunction *CalleeFunc = this->GetProg()->FindFunction(CalleeAddr);
								if (nullptr != CalleeFunc) {
									CriticalArgPosBits = CalleeFunc->GetTaintInArgPositions();
									ArgSearch = (0 < CriticalArgPosBits); // found a critical call if any critical arg
								}
							}
						}
					}
					else if (ArgSearch && CurrInst->MDIsArgumentPass(ArgNum)) {
						STARSDefUseIter ArgDefIter = CurrInst->GetFirstNonFlagsDef();
						if (ArgDefIter != CurrInst->GetLastDef()) {
							STARSOpndTypePtr ArgDefOp = ArgDefIter->GetOp();
							if (ArgDefOp->IsRegOp()) {
								if (0 != (CriticalArgPosBits & (1 << ArgNum))) {
									// ArgNum matches a bit set in CriticalArgPosBits.
									STARSOpndTypePtr ArgOp = nullptr;
									unsigned short SignMask = 0;
									if (CurrInst->MDIsMoveInstr()) {
										ArgOp = CurrInst->GetMoveSource(); // Need a USE for FindShadowingPoint2()
									}
									else if (CurrInst->MDIsSignedLoad(SignMask)) {
										ArgOp = CurrInst->GetFirstRightOperandNoNorm();
										if (ArgOp->IsRegOp()) {
											CanonicalizeOpnd(ArgOp);
										}
										else if (MDIsStackAccessOpnd(ArgOp, this->UsesFramePointer()) && CurrInst->AreDefsNormalized()) {
											ArgOp = this->GetNormalizedOperand(CurrInst->GetAddr(), ArgOp);
										}
									}
									if ((nullptr != ArgOp) && !ArgOp->IsVoidOp()) {
										STARSDefUseIter UseIter = CurrInst->FindUse(ArgOp);
										assert(UseIter != CurrInst->GetLastUse());
										this->TempShadowList.clear();
										this->TempShadowList.SetRef(ArgOp, UseIter->GetType(), UseIter->GetSSANum());
										this->EmitShadowingHelper(InfoAnnotFile, CurrInst, false, STARS_BADADDR);
									}
								}
							}
						} // end if ArgDefIter valid
					} // end if CALL elsif arg pass
				} // end for all insts in reverse order
			} // end if current block has a call inst
		} // end for all blocks

	} // end if good RTLs, unsafe indirect writes, and not a leaf func
	return;
} // end of SMPFunction::EmitArgShadowingAnnotations()

// To assist in stack transformation security defenses, if the 
//  memory range expression is stack-relative, we emit an annotation
//  that identifies the memory range accessed, at InstAddr.
// The annotation should look like:
//  [hex InstAddr] [InstSize] INSTR STACKMEMRANGE MIN [ESP-k] LIMIT [ESP-j] INSTRSPDELTA -d ZZ
// The INSTRSPDELTA is the value of RSP relative to its value on function entry, which needs
//  to be added to the other constants to obtain the entry-relative stack offsets.
//  The other constants are relative to the current SPOFFSET for the mem write instruction.
// We write the InstAddrs of static memory lea opcodes to StaticMemLeaAddrs for later processing.
void SMPFunction::EmitStackMemRangeAnnotations(STARSExpression *MemLowerBound, STARSExpression *MemUpperBound, bool PositiveIncrement, const set<STARS_ea_t> &StackPtrCopiesSet, set<STARS_ea_t> &StaticMemLeaAddrs) {
	FILE *InfoAnnotFile = global_STARS_program->GetInfoAnnotFile();
	SMPInstr *MemInst = MemLowerBound->GetOriginalParentInst();
	STARS_ea_t MemAddr = MemInst->GetAddr();
	char DummyBuf[STARS_MAXSTR];
	(void) SMP_snprintf(DummyBuf, STARS_MAXSTR - 1, "%18llx", (uint64_t) MemAddr);
	string OutputString1(DummyBuf);
	OutputString1 += " ";
	(void) SMP_snprintf(DummyBuf, STARS_MAXSTR - 1, "%6lld", (int64_t) MemInst->GetSize());
	OutputString1 += DummyBuf;
	string OutputString2(" INSTR STACKMEMRANGE MIN ");
#if 0
	std::ostringstream AddrString;
	AddrString << std::hex << MemAddr << " ";
	AddrString << std::dec << MemInst->GetSize() << " INSTR STACKMEMRANGE MIN ";
	SMP_fprintf(InfoAnnotFile, "%18llx %6zu INSTR STACKMEMRANGE MIN ", (uint64_t)MemAddr, MemInst->GetSize());
#endif
	if (PositiveIncrement) {
		MemLowerBound->StringPrintStackPtrPlusOffset(OutputString2);
	}
	else {
		MemUpperBound->StringPrintStackPtrPlusOffset(OutputString2);
	}
	OutputString2 += " LIMIT ";
	if (PositiveIncrement) {
		MemUpperBound->StringPrintStackPtrPlusOffset(OutputString2);
	}
	else {
		MemLowerBound->StringPrintStackPtrPlusOffset(OutputString2);
	}
	OutputString2 += " INSTRSPDELTA ";
	(void) SMP_snprintf(DummyBuf, STARS_MAXSTR - 1, "%lld", (int64_t) MemInst->GetStackPtrOffset());
	OutputString2 += DummyBuf;
	OutputString2 += " ZZ";

	SMP_fprintf(InfoAnnotFile, "%s%s\n", OutputString1.c_str(), OutputString2.c_str());

	// Now, emit similar annotations for any StackPtrCopy instructions that were found
	//  when the Mem exprs were ExpandExpr()-ed.
	if (!StackPtrCopiesSet.empty()) {
		for (STARS_ea_t StackPtrCopyInstAddr : StackPtrCopiesSet) {
			assert(STARS_BADADDR != StackPtrCopyInstAddr);
			SMPInstr *StackPtrCopyInst = this->GetInstFromAddr(StackPtrCopyInstAddr);
			assert(nullptr != StackPtrCopyInst);

			// Ensure that we don't have the address of a global static memory item.
			if (StackPtrCopyInst->IsGlobalStaticLoadEffectiveAddress()) {
				(void) StaticMemLeaAddrs.insert(StackPtrCopyInstAddr);
			}
			else if (StackPtrCopyInst->MDIsStackPointerCopy(this->UsesFramePointer())) {
				(void)SMP_snprintf(DummyBuf, STARS_MAXSTR - 1, "%18llx", (uint64_t)StackPtrCopyInstAddr);
				string OutputString3(DummyBuf);
				OutputString3 += " ";
				(void)SMP_snprintf(DummyBuf, STARS_MAXSTR - 1, "%6lld", (int64_t)StackPtrCopyInst->GetSize());
				OutputString3 += DummyBuf;
				// Re-use OutputString2 with new InstAddr and InstSize.
				SMP_fprintf(InfoAnnotFile, "%s%s\n", OutputString3.c_str(), OutputString2.c_str());
			}
			else {
				SMP_msg("ERROR: Inst at %llx should be lea of either stack or static mem address ", (uint64_t) StackPtrCopyInstAddr);
				this->DumpFuncNameAndAddr();
			}
		} // end for all addrs in StackPtrCopiesVector[index]
	}
	return;
} // end of SMPFunction::EmitStackMemRangeAnnotations()
